/**************************************************************************************************
 * FifoBuffer.c                                                                           R Moulton
 * Runs as a background process to support a 1 to many in, 1 out buffered fifo.  Output is to a 
 * named fifo, or via TCP/IP with this process serving either as a server or a client.  In any
 * case, partner requests each record.  Input can be from a named fifo and/or connections to a
 * listen socket.  Records received on the inbound side are buffered to disc, then output to the 
 * output fifo when requested by the reading process.  Can be used any type of EDI messages are 
 * being sent from one process to another, permitting the messages from one to be queued for the 
 * other while it is down. On invocation, par[1] is the path name of the buffered fifo.  name.put 
 * and name.get are the input and output sides of the fifo.  If name.put does not exist, there is
 * no pipe input. name.ini is the .ini file, name.mon is the monitor fifo, and name.0 and name.1 
 * are the buffer files (which may not exist at any given moment).
 *-------------------------------------------------------------------------------------------------
 * 2-4 R Moulton Initial write
 * 9-5 R Moulton Modified to use log library.
 *************************************************************************************************/
#include <errno.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stropts.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <poll.h>
#include <dirent.h>

#include "dbg.h"
#include "ini.h"
#include "alert.h"
#include "hl7.h"
#include "log.h"

/*--------------------------------------------------------------------------------------- Macros */
#define MSTR(A,B) {if(A)free(A); if(B && (A=malloc(strlen(B)+10)))strcpy(A,B);}
#define FREE(A) if(A){free(A);A=NULL;}
#define STR_EQ(A,B) !strcasecmp(A,B)
#define STR_NOT_NULL(A,B) {if(A)free(A); if(B && strcasecmp(B,"NULL") && (A=malloc(strlen(B)+10))) strcpy(A,B); }
#define LAST_CHR(A) A[strlen(A) - 1]

#define SB 11
#define EB 28

/*************************************** Function prototypes *************************************/
void setup(int numpars, char *par[]);
void read_ini(void);                                       /* Periodic checks */
void warning_check(void);
void monitor_check(void);
void log_msg(char * msg);
void setup_put_sock(void);
void setup_get_sock(void);
void close_socket(int * socket);
void clear_put_client(char client);
void check_put_socket(void);
void check_get_socket(void);
void connect_get_socket(char place);
void process_request_record(char * msg);
void process_sent_record(void);
char * internalize(char * out, char * in);
char * externalize(char * out, char * in);
void totals_check(void);
int fd_isset(int socket, short flag);
void fd_clr(int socket);
void fd_st(int socket, short flag);
int fd_num(void);

/*************************************** Working storage *****************************************/
#define TIME_HACK 20                                                      /* Length of time hack */
static char * name;                                                                 /* fifo name */
static char * input_pfn, * input_buffer;                         /* input (to this program) fifo */
static char * output_pfn, * output_msg, rtr_buffer[128];    /* output (from this program) fifo */
static char * ini_pfn, * mon_pfn, * rtr_pfn;
static char * file1_pfn, * file1_buffer;                                    /* disc buffer files */
static char * file2_pfn, * file2_buffer;
static char * read_buffer;

#define MAX_CLIENTS 20                                                /* can't be larger than 29 */
#define CLIENT clients[client]
struct _client_ {
   char * name;                /* host name of client */
   unsigned char ip[4];        /* host ip of client */
   char * buffer;              /* msg buffer for client */
   int sock;                   /* socket, if connected */
   char rts;                   /* ready to send? */
   char rtr;
   unsigned long msgs_today;   /* # of messages received since midnight */
   unsigned long msgs_hour;    /* # of messages received since top of the hour */
   unsigned long connects_today;  /* same, for connections made */
   unsigned long connects_hour;
   time_t last_msg;
   time_t last_connect;
};
static struct _client_ clients[MAX_CLIENTS];
static char client;
static char max_client = 0;
static struct _client_ gtc;

static int put_port = 0, get_port = 0, put_sock = 0, get_sock = 0;
static char connect_as = 'c';
static char host_name[MAXHOSTNAMELEN + 1];
static struct hostent *phostent, shostent;
static struct sockaddr_in put_name, get_name;
static struct sockaddr * put_ptr = (struct sockaddr*)&put_name;
static struct sockaddr * get_ptr = (struct sockaddr*)&get_name;
static size_t sl = sizeof(put_name);
static int on = 1;
#define TIMEOUT_SECONDS 60
#define MAX_POLL 25
#define POLL pfds[pfd]
static struct pollfd pfds[MAX_POLL];
static char pfd;

static FILE * debug_fp, * input_fp, * output_fp, * file1_fp, * file2_fp, * mon_fp, * rtr_fp;
static int input_fd, output_fd, rtr_fd;
static fpos_t file1_read_position, file1_delete_position;
static char file1_need_to_delete_record, file2_has_data, output_rtr = 0, output_rts = 0;
static int max_record_size, buffer_size;
static int monitoring_interval;

static char sb_eb;
static int warning_MB, warning_latency; /*MB of buffer files to issue warning at (0 = no warning) */
static int no_input, no_output;                       /* Mins without input or output to warn on */
static char * warning_email;
static char * warning_printer;
static char * warning_pager;
static int check_interval;    /* Minutes between checks */
static int warning_interval;  /* Minutes between warnings */
static time_t last_input = 0, last_output = 0;
static unsigned long input_today = 0, output_today = 0, input_hour = 0, output_hour = 0;
static unsigned long max_latency = 0;
#define MSG_SIZE 2048

static char std_ack_msg[] = "\013ACK\034";

/*----------------------------- Keep track of most recent rejected connections, if using TCP/IP */
#define MAX_INTRUDERS 5
#define INTRUDER intruders[intr]
struct _intruders_ {
   char ip[18];
   time_t at;
};
static struct _intruders_ intruders[MAX_INTRUDERS];
static char intr;

#define MAX_LATENCIES 5
#define LATENCY latencies[i]
struct _latency_ {
   unsigned long time;
   char alpha[18];
};
static struct _latency_ latencies[MAX_LATENCIES];

static time_t current_time;
struct tm * current_tm;
/**************************************************************************************************
 **************************************************************************************************
 ****************************** Main routine ******************************************************
 **************************************************************************************************
 *************************************************************************************************/
int main (int numpars, char * par[])
{  int rc;
   char * p;

   current_time = time(NULL);
   current_tm = localtime(&current_time);
   
   setup (numpars, par);                                              /* process initialization */
   logInitialize("MESSAGES");
   logInitialize("ERRORS");
   
   while (1)
   {  current_time = time(NULL);
      current_tm = localtime(&current_time);
      
      read_ini();                                                    /* Periodic checks */
      monitor_check();
      warning_check();                        
      logPurge("Messages");
      logPurge("Errors");
      totals_check();
      check_get_socket();
      
/*----------------------------------------------------------------------------- Setup for select */
      for (pfd = 0; pfd < MAX_POLL; pfd++) { POLL.fd = -1; POLL.events = POLL.revents = 0; }

      if (put_port)                                           /* If we're doing TCP/IP put */
      {  fd_st(put_sock, POLLIN);                           /* Listen socket */
         for (client=0; client < max_client; client++)
         {  if (CLIENT.sock)                              /* all clients with open sockets */
            {  fd_st(CLIENT.sock, POLLIN); 
               if (CLIENT.rts) fd_st(CLIENT.sock, POLLIN | POLLOUT);
            }
         }
      }

      if (input_pfn) fd_st(input_fd, POLLIN);            /* input fifo, if we have one */
         
/*---------------------------------- outbound side, can be TCP/IP server, TCP/IP client, or fifo */
      if (get_port)
      {  if (connect_as == 's') fd_st(get_sock, POLLIN);  /* We're server, this is listen socket */
         else connect_get_socket(1);                      /* We're client, this is connect */
         if (gtc.sock)                                    /* This is actual socket */
         {  if (gtc.rts) fd_st(gtc.sock, POLLIN | POLLOUT);
            else fd_st(gtc.sock, POLLIN);
         }
      }
      else                                                     /* Using fifo */
      {  if (!output_rtr) fd_st (rtr_fd, POLLIN);                 /* rtr inbound fifo */
         if (output_rts)  fd_st (output_fd, POLLOUT);             /* outbound data fifo */
      }
/******************************* I/O select ******************************************************/
      rc = poll(pfds, fd_num(), TIMEOUT_SECONDS * 1000);
      if (rc < 0) { dbge("poll"); continue; }
		if (!rc) {sprintf(em,"Poll timed out at %d seconds", TIMEOUT_SECONDS); dbg(em);continue; }
      
      connect_get_socket(2);
/*----------------------------------------------------------------------------------- fifo input */
      if (input_pfn && fd_isset(input_fd, POLLIN))
      {  p = fgets(input_buffer, max_record_size, input_fp);
         if (p && strlen(p)) 
         {  last_input = current_time;     
            internalize(file2_buffer,input_buffer);
            if (fputs(file2_buffer, file2_fp) < 0) dbge_exit("fputs file2", 1); 
            file2_has_data = 1;
            input_today++; input_hour++;
            dbg("Input msg");
         }
      }
      
/*--------------------------------------------------------------------------------- TCP/IP input */
      if (put_port)
      {  check_put_socket();                                                /* listen socket */
         for (client = 0; client < max_client; client++)                             /* clients */
         {  if (CLIENT.sock)
            {  if (fd_isset(CLIENT.sock, POLLERR))                     /* error, close socket */
               {  close_socket(&CLIENT.sock);
                  continue;
               }
               if (fd_isset(CLIENT.sock, POLLOUT) && CLIENT.rts)                     /* write */
               {  rc = send(CLIENT.sock, std_ack_msg, strlen(std_ack_msg), 0);
                  if (rc == -1)
                  {  close_socket(&CLIENT.sock);
                     sprintf(em,"send error on %s", CLIENT.name); dbg(em);
                  }
                  else
                  {  sprintf(em,"sent %d bytes to %s", rc, CLIENT.name); dbg(em);
                     CLIENT.rts = 0;
                  }
               }
               if (fd_isset(CLIENT.sock, POLLIN))                                     /* read */
               {  memset(read_buffer,0,max_record_size);
                  rc = recv(CLIENT.sock, read_buffer, max_record_size, 0);
                  if (rc < 0) 
                  {  sprintf(em,"%s recv", CLIENT.name); dbge(em);
                     close_socket(&CLIENT.sock);
                     continue;
                  }
                  if (!rc)
                  {  sprintf(em,"%s connection dropped by client", CLIENT.name); dbg(em);
                     close_socket(&CLIENT.sock);
                     continue;
                  }
                  sprintf(em,"recv %d bytes from %s", rc, CLIENT.name); dbg(em);
                  if (p = strchr(read_buffer,SB)) strcpy(CLIENT.buffer,p);
                  else strcat(CLIENT.buffer,read_buffer);
                  if (p = strchr(CLIENT.buffer,EB))
                  {  *(p+1) = '\0';
                     internalize(file2_buffer, CLIENT.buffer);
                     if (fputs(file2_buffer, file2_fp) < 0) dbge_exit("fputs file2", 1);
                     file2_has_data = 1;
                     memset(CLIENT.buffer,0,max_record_size);
                     CLIENT.msgs_today++;
                     CLIENT.msgs_hour++;
                     CLIENT.last_msg = time(NULL);
                     CLIENT.rts = 1;
                  }
               }
            }
         }
      }

/*-------------------------------------------------------------------------------- TCP/IP output */
      if (get_port)
      {  check_get_socket(); 
         if (gtc.sock && fd_isset(gtc.sock, POLLERR)) close_socket(&gtc.sock);
         if (gtc.sock && fd_isset(gtc.sock, POLLIN))
         {  rc = recv(gtc.sock, rtr_buffer, sizeof(rtr_buffer), 0);
            if (rc < 0) { dbge("get recv"); close_socket(&gtc.sock); }
            if (rc > 0)
            {  if ((p = strchr(rtr_buffer,SB)) && p != rtr_buffer) memmove(rtr_buffer,p,strlen(p) +1);
               if (strchr(rtr_buffer,SB) && strchr(rtr_buffer,EB)) process_request_record(rtr_buffer);
            }
         }
         if (gtc.sock && fd_isset(gtc.sock, POLLOUT))
         {  rc = send(gtc.sock, output_msg, strlen(output_msg), 0);
            if (rc < 0) { dbge("get send"); close_socket(&gtc.sock); }
            else process_sent_record();
         }
      }
/*---------------------------------------------------------------------------------- Fifo output */
      if (!get_port)
      {  if (!output_rtr && (fd_isset(rtr_fd, POLLIN)))                  /* request */
         {  p = fgets(rtr_buffer, sizeof(rtr_buffer), rtr_fp);
            if (p) process_request_record(p);
         }
/*------------------------------------------------------------------------- We are ready to send */
         if (output_rts && (fd_isset(output_fd, POLLOUT)))
         {  if (fputs(output_msg, output_fp) < 0) dbge_exit("output fputs", 1);
            process_sent_record();
         }
      }
   }
}

/**************************************************************************************************
 * process an incoming message in msg
 *************************************************************************************************/
void process_request_record(char *msg)
{  char * p;
   if (*msg == SB) memmove(msg,msg+1,strlen(msg));
   while (LAST_CHR(msg) == '\n' || LAST_CHR(msg) == EB) LAST_CHR(msg) = '\0';
/*-------------------------------------------------------------------------------------------------
 * Receive an RTR or REL and have a record to delete.  "delete" this record by replacing its first 
 * character with a "~".  Since the beginning of the record is a time hack, the first character 
 * will never otherwise be a "~".  The record being deleted is at file1_delete_position; our 
 * current read position is file1_read_p-osition.
 *-----------------------------------------------------------------------------------------------*/
   if ((STR_EQ(msg,"RTR") || STR_EQ(msg,"REL")) && file1_need_to_delete_record)
   {  if (fgetpos(file1_fp, &file1_read_position)) dbge_exit("file1 get read pos", 1);
      if (fsetpos(file1_fp, &file1_delete_position)) dbge_exit("file1 set del pos", 1);
      *file1_buffer = '~';
      if (fputs(file1_buffer, file1_fp) < 0) dbge_exit("file1 del", 1);
      if (fsetpos(file1_fp, &file1_read_position)) dbge_exit("file1 set read pos", 1);
      file1_need_to_delete_record = 0;
   }
/*-------------------------------------------------------------------------------------------------
 * Receive a SND and have a record to delete.  This indicates that the output process terminated
 * abnormally (otherwise we would have received an REL before it closed and the record to delete
 * would have already been processed.  So we will resend the record to delete.
 *-----------------------------------------------------------------------------------------------*/
   if (STR_EQ(msg,"SND") && file1_need_to_delete_record)
   {  if (fsetpos(file1_fp, &file1_delete_position)) dbge_exit("file1 set read pos", 1);
      file1_need_to_delete_record = 0;
   }
   if (STR_EQ(msg,"RTR") || STR_EQ(msg,"SND")) output_rtr = 1;
   dbg(msg);
/*-------------------------------------------------------------------------------------------------
 * Here we handle the case where we have gone rtr and are not rts. 
 *-----------------------------------------------------------------------------------------------*/
   while (output_rtr && !output_rts)                /* ready to receive but not ready to send */
   {  if (!file1_fp && !file2_has_data) break;                      /* Nothing to send at all */
      if (!file1_fp)                               /* Nothing in file1, get file 2 */
      {  if (fclose(file2_fp)) dbge_exit("fclose file2", 1);
         if (rename(file2_pfn, file1_pfn)) dbge_exit("rename file2 -> file1", 1);
         if (!(file1_fp = fopen(file1_pfn, "r+"))) dbge_exit("file1 open", 1);
         if (!(file2_fp = fopen(file2_pfn, "a" ))) dbge_exit("file2 open", 1);
         setvbuf(file1_fp, NULL, _IOLBF, 0);
         setvbuf(file2_fp, NULL, _IOLBF, 0);
         file2_has_data = 0;
         continue;
      }
/*-------------------------------------------------------------------------------------------------
 * Get next record from file1.  First we save the position to use later when we clear this record.
 *-----------------------------------------------------------------------------------------------*/
      if (fgetpos(file1_fp, &file1_delete_position)) dbge_exit("file1 get del pos", 1);
/*-------------------------------------------------------------------------- Get the next record */
      p = fgets(file1_buffer, buffer_size, file1_fp);
      if (p) 
      {  file1_need_to_delete_record = 1;         /* We will need to delete this record later */
         output_rts = 1;                          /* We are now ready to send */
         externalize(output_msg, file1_buffer);
         break;
      }
      if (ferror(file1_fp)) dbge("file1 fgets");     /* eof or error on file1 */
      if (fclose(file1_fp)) dbge("file1 fclose");
      if (remove(file1_pfn)) dbge_exit("file1 remove", 1);
      file1_fp = NULL;
      continue;
   }
}

void process_sent_record(void)
{  output_rts = output_rtr = 0;
   memset(rtr_buffer,0,sizeof(rtr_buffer));
   last_output = current_time;
   output_today++; output_hour++;
   log_msg(file1_buffer);
   dbg("sent msg");
}

/**************************************************************************************************
 * Process initialization (one time)
 *************************************************************************************************/
void setup(int numpars, char * par[])
{  char * p, a[128];
   int fd, r;
   struct stat st;
   umask(S_IROTH | S_IWOTH);
/*-------------------------------------------------------------------------------------------------
 * pfn of fifo is parameter 1.  Make that directory current and load file pfns. 
 *-----------------------------------------------------------------------------------------------*/
   if (numpars < 2) dbg_exit("fifo name nust be param 1", 1);
   
   p = strrchr(par[1],'/');          /* Find end of path */
/*-------------------------------------------------- If path exists, make that directory current */
   if (p) 
   {  *p = '\0';
      if (chdir(par[1]) == -1) dbg_exit("Can't cd to fifo directory", 1); 
      *p = '/';
   }
/*------------------------------------------------------------------ store fifo name, file names */
   if (p) MSTR(name,p+1)
   else MSTR(name,par[1])
/*--------------------------------------------------------- ini file, must exist and be readable */
   MSTR(ini_pfn,name)  strcat(ini_pfn,".ini");
   if (access(ini_pfn, F_OK)) { sprintf(em, "%s doesn't exist", ini_pfn); dbg_exit(em, 1); }
   if (access(ini_pfn, R_OK)) { sprintf(em, "%s not readable", ini_pfn); dbg_exit(em, 1); }
/*-------------------------------------------------------------------------------------- input fifo
 * This file must exist and be readable.
 *---------------------------------------------------------------------------------------------- */
   MSTR(input_pfn,name) strcat(input_pfn,".put");
   if (access(input_pfn, F_OK)) { sprintf(em,"Can't find %s", input_pfn); dbg_exit(em, 1); }
   if (access(input_pfn, R_OK)) { sprintf(em, "%s not readable", input_pfn); dbg_exit(em, 1); }
/*-------------------------------------------------------------------------------------------------
 * if output fifo and rtr exist, must be readable and writable.  If they don't exist, all output
 * will be via TCP/IP
 *---------------------------------------------------------------------------------------------- */
   MSTR(output_pfn,name) strcat(output_pfn,".get");
   if (access(output_pfn, F_OK)) { FREE(output_pfn) }
   else if (access(output_pfn, W_OK | R_OK)) { sprintf(em, "%s not writable", output_pfn); dbg_exit(em, 1); }

   if (output_pfn)
   {  MSTR(rtr_pfn,name) strcat(rtr_pfn,".rtr");
      if (access(rtr_pfn, F_OK)) { FREE(rtr_pfn) FREE(output_pfn) }
      else if (access(rtr_pfn, R_OK)) { sprintf(em, "%s not readable", rtr_pfn); dbg_exit(em, 1); }
   }
/*------------------------------------------- Monitoring fifo, if it exists, it must be writable */
   MSTR(mon_pfn,name) strcat(mon_pfn,".mon");
   if (access(mon_pfn, F_OK)) { FREE(mon_pfn) }
   else if (access(mon_pfn, W_OK)) { sprintf(em, "%s not writable", mon_pfn); dbg_exit(em, 1); }
/*--------------------------------------------------------------------------------- Buffer files */
   MSTR(file1_pfn,name) strcat(file1_pfn,".1");
   MSTR(file2_pfn,name) strcat(file2_pfn,".2");

/*-------------------------------------------------------------------------------------------------
 * First thing we do is to attempt to open the input fifo to write, non-blocking.  If this works, 
 * the MessageServer is already running, so we quit.
 *-----------------------------------------------------------------------------------------------*/
   fd = open(input_pfn, O_WRONLY | O_NONBLOCK);
   if (fd >= 0) dbg_exit("already running", 0); 

/*--------------------------------------------------------------------------- Now open the pipes */ 
   if (input_pfn)
   {  input_fd = open(input_pfn, O_RDWR);
      if (input_fd < 0) dbge_exit("open .put", 1); 
      input_fp = fdopen(input_fd, "r");
      if (!input_fp) dbge_exit("fdopen .put", 1);
   }

/*-------------------------------------------------------- We do initial load of .ini file stuff */
   read_ini();
   
/*------------------------------------------------------------------------ Set up disc buffer files
 * If file1 exists, we (1) open it to read, (2) skip over previously processed records, and (3)
 * delete it if it is logically empty.
 *-----------------------------------------------------------------------------------------------*/
   if (!access(file1_pfn, F_OK))
   {  file1_fp = fopen(file1_pfn, "r+");
      if (!file1_fp) dbge_exit("file1 fopen", 1);
      setvbuf(file1_fp, NULL, _IOLBF, 0);
      while (1)
      {  if (fgetpos(file1_fp, &file1_read_position)) dbge_exit("fgetpos", 1);
         p = fgets(file1_buffer, buffer_size, file1_fp);
         if (!p) 
         {  if (fclose(file1_fp)) dbge_exit("close file1", 1);
            if (remove(file1_pfn)) dbge_exit("remove file1", 1);
            break;
         }
         if (*file1_buffer == '~') continue;
         if (fsetpos(file1_fp, &file1_read_position)) dbge_exit("fsetpos", 1);
         break;
      }
   }
/*-------------------------------------------------------------------------------------------------
 * If file1 does not exist, or it did exist but was logically empty and therefore was deleted, we
 * check for file2.  If it exists, we rename it to become file 1 and open it -- But we don't have
 * to check it for deleted records, because until now it was only being used to write to.
 *-----------------------------------------------------------------------------------------------*/
   if (access(file1_pfn, F_OK) && !access(file2_pfn, F_OK))
   {  if (rename(file2_pfn, file1_pfn)) dbge_exit("rename file2 -> file1", 1);
      file1_fp = fopen(file1_pfn, "r+");
      if (!file1_fp) dbge_exit("file1 open", 1);
      setvbuf(file1_fp, NULL, _IOLBF, 0);      
      if (fgetpos(file1_fp, &file1_read_position)) dbge_exit("fgetpos", 1);
   }
/*--------------------------------------------- Here there is no file yet, so we NULL out the fp */
   if (access(file1_pfn, F_OK)) file1_fp = NULL;
   file1_need_to_delete_record = 0;
/*-------------------------------------------------------------------------------------------------
 * file2 is where we will put the stuff coming in.  If the file exists and has data in it, we 
 * can use it when we need data for the output.
 *-----------------------------------------------------------------------------------------------*/
   file2_has_data = 0;
   if (!access(file2_pfn, F_OK) && !stat(file2_pfn, &st) && st.st_size) file2_has_data = 1;
   file2_fp = fopen(file2_pfn, "a");
   if (!file2_fp) dbge_exit("file2 open", 1);
   setvbuf(file2_fp, NULL, _IOLBF, 0);

}

/**************************************************************************************************
 * check environment variables; done periodically 
 *************************************************************************************************/
void read_ini(void)
{  static time_t last_check = 0, ini_mtime = 0;
   static first_time = 1;
   time_t mtime;
   char * p, * h, a[1024], * host[MAX_CLIENTS +1], * ip[MAX_CLIENTS +1], *getip, new_connect_as;
   unsigned char ipn[4];
   int fd, r, i, j;
   struct stat st;
   int new_port;
   FILE * fp;
   
/*------------------------------------------------------ We run this check only after 60 seconds */
   if ((current_time - last_check) < 60) return;
   last_check = current_time;

/*------------------------- We load env only if .ini is more current than the last one we loaded */ 
   if (stat(ini_pfn, &st) < 0) dbge_exit(".ini file", 1);
   if (st.st_mtime <= ini_mtime) return;
   ini_mtime = st.st_mtime;
   
/*----------------------------------------------------------------------------------- Debug mode */
   if (!GetIniString("[Main]","Debug","NULL",a,sizeof(a),ini_pfn))dbg_exit(GetIniErrorString(),1);
   debug_fp = debug_on(a);
/*------------------------------------- Should start and stop bytes be put on outbound messages? */
   if (!GetIniString("[Main]","SBEB","Yes",a,sizeof(a),ini_pfn))dbg_exit(GetIniErrorString(),1);
   if (*a == 'n' || *a == 'N') sb_eb = 0; else sb_eb = 1;
/*----------------------------------------------------- Maximum record length to be accommodated */
   if (!GetIniInt("[Main]","MaxRecordSize",20480,&max_record_size,ini_pfn))dbg_exit(GetIniErrorString(),1);
   max_record_size += 4;                                        /* Allow for \n, null terminator */
   buffer_size = max_record_size + TIME_HACK;                   /* buffer has time hack at front */
/*-------------------------------------------- Monitoring interval in seconds (0 = no monitoring */
   if (!GetIniInt("[Monitoring]","Interval",0,&monitoring_interval,ini_pfn)) dbg_exit(GetIniErrorString(),1);
/*----------------------------------------------------------------------------- Alert thresholds */
   if (!GetIniInt("[Warning]","MB",0,&warning_MB,ini_pfn)) dbg_exit(GetIniErrorString(),1);
   if (!GetIniInt("[Warning]","Latency",0,&warning_latency,ini_pfn)) dbg_exit(GetIniErrorString(),1);
   if (!GetIniInt("[Warning]","NoInput",0,&no_input,ini_pfn)) dbg_exit(GetIniErrorString(),1);
   if (!GetIniInt("[Warning]","NoOutput",0,&no_output,ini_pfn)) dbg_exit(GetIniErrorString(),1);
/*----------------------------------------------------------------------------- Alert parameters */
   if (!GetIniString("[Warning]","Email","NULL",a,sizeof(a),ini_pfn)) dbg_exit(GetIniErrorString(),1);
   STR_NOT_NULL(warning_email,a)
   if (!GetIniString("[Warning]","Printer","NULL",a,sizeof(a),ini_pfn)) dbg_exit(GetIniErrorString(),1);
   STR_NOT_NULL(warning_printer,a)
   if (!GetIniString("[Warning]","Pager","NULL",a,sizeof(a),ini_pfn)) dbg_exit(GetIniErrorString(),1);
   STR_NOT_NULL(warning_pager,a)
   if (!GetIniInt("[Warning]","CheckInterval",5,&check_interval,ini_pfn)) dbg_exit(GetIniErrorString(),1);
   if (!GetIniInt("[Warning]","WarningInterval",30,&warning_interval,ini_pfn)) dbg_exit(GetIniErrorString(),1);
   
/*------------------------------------------ Set up buffers copy over existing contents (if any) */
   if (!(p = malloc(max_record_size))) dbg_exit("malloc error", 1);
   memset(p,0,max_record_size);
   if (input_buffer) { strcpy(p,input_buffer); free(input_buffer); }
   input_buffer = p;

   if (!(p = malloc(max_record_size))) dbg_exit("malloc error", 1);
   memset(p,0,max_record_size);
   if (output_msg) { strcpy(p,output_msg); free(output_msg); }
   output_msg = p;
   
   if (!(p = malloc(buffer_size))) dbg_exit("malloc error", 1);
   memset(p,0,buffer_size);
   if (file1_buffer) { strcpy(p,file1_buffer); free(file1_buffer); }
   file1_buffer = p;
   
   if (!(p = malloc(buffer_size))) dbg_exit("malloc error", 1);
   memset(p,0,buffer_size);
   if (file2_buffer) { strcpy(p,file2_buffer); free(file2_buffer); }
   file2_buffer = p;

   for (client = 0; client < max_client; client++)
   {  if (!(p=malloc(max_record_size))) dbg_exit("malloc error", 1);
      memset(p,0,max_record_size);
      if (CLIENT.buffer) { strcpy(p,CLIENT.buffer); free(CLIENT.buffer); }
      CLIENT.buffer = p;
   }

   if (!(p=malloc(max_record_size))) dbg_exit("malloc error", 1);
   memset(p,0,max_record_size);
   if (gtc.buffer) { strcpy(p,gtc.buffer); free(gtc.buffer); }
   gtc.buffer = p;
   
   if (read_buffer) free(read_buffer);
   read_buffer = malloc(max_record_size);

/*-------------------------------------------------------------------------------------------------
 *  .get parameters (These are read only at startup)
 *-----------------------------------------------------------------------------------------------*/
   if (first_time)
   {  first_time = 0;
      if (!GetIniString("[GetTCPIP","ConnectAs","c",a,sizeof(a),ini_pfn)) dbg_exit(GetIniErrorString(),1);
      if (*a == 'S' || *a == 's') connect_as = 's'; else connect_as = 'c';
      if (!GetIniInt("[GetTCPIP]","Port",0,&get_port,ini_pfn)) dbg_exit(GetIniErrorString(),1);
      if (!GetIniString("[GetTCPIP]","PeerName","NULL",a,sizeof(a),ini_pfn))dbg_exit(GetIniErrorString(),1);
      STR_NOT_NULL(gtc.name,a)
      if (!GetIniString("[GetTCPIP]","PeerIP","NULL",a,sizeof(a),ini_pfn))dbg_exit(GetIniErrorString(),1);
/*------------------------------------------------------ If TCP/IP and no ip, look in /etc/hosts */
      STR_NOT_NULL(getip,a)
      if (get_port && !getip)
      {  if (!(fp = fopen("/etc/hosts", "r"))) dbge_exit("fopen /ect/hosts",1);
         while (fgets(a,sizeof(a),fp))
         {  if (p = strchr(a,'#')) *p = '\0';           /* handles comment data */
            p = strtok(a," \t");                       /* ip address, if any */
            h = strtok(NULL," \t");                    /* host name, if any */
            if (!p || !h) continue;                    /* not a data line */
            if (STR_EQ(gtc.name,h))                   /* match host to file */
            {  MSTR(getip,p)                       /* load ip from file */
               break;
            }
         }
         fclose(fp);
      }
/*-------------------------------------------------------------------------- generate numeric ip */
      gtc.ip[0] = gtc.ip[1] = gtc.ip[2] = gtc.ip[3] = 0;
      if (getip) for (i=0, p=strtok(getip,"."); i < 4 && (p=strtok(NULL,".")); i++) gtc.ip[i] = atoi(p); 
/*----------------------------------------------------------------------------- We're using fifo */
      if (!get_port)
      {  output_fd = open(output_pfn, O_RDWR);
         if (output_fd < 0) dbge_exit("open .get", 1); 
         output_fp = fdopen(output_fd, "a");
         if (!output_fp) dbge_exit("fdopen .get", 1);
         setvbuf(output_fp, NULL, _IOLBF, 0);
         rtr_fd = open(rtr_pfn, O_RDWR);
         if (rtr_fd < 0) dbge_exit("open .rtr", 1); 
         rtr_fp = fdopen(rtr_fd, "r");
         if (!rtr_fp) dbge_exit("fdopen .rtr", 1);
      }
/*-------------------------------------------------------------------- TCP/IP connection for get */
      if (get_port && connect_as == 's') setup_get_sock();
/*-------------------------------------------------------------------------------------------------
 * TCP/IP .put parameters (only at startup)
 *-----------------------------------------------------------------------------------------------*/
      if (!GetIniInt("[PutTCPIP]","Port",0,&put_port,ini_pfn)) dbg_exit(GetIniErrorString(),1);
      if (put_port) setup_put_sock();
   }
/*-------------------------------------------------------------------------------------------------
 * The put client data section is read whenever the .ini is updated, as it may have changed.  But 
 * It is only relevant if we are using TCP/IP, which we would know at startup.
 *-----------------------------------------------------------------------------------------------*/
   if (!put_port) return;
/*-------------------------------------------------------------------------------------------------
 * Pull client names from the .ini file, load ip's in the table and look up others in the host
 * table
 *---------------------------------------------------------------------------------------------- */
   for (i=0; i < MAX_CLIENTS; i++) host[i] = ip[i] = NULL;                 /* Clear working area */
   
/*----------------------------------------- Load all clients, key names in the [PutClients] section */
   if (!GetIniSection("[PutClients]",a,sizeof(a),ini_pfn)) dbge_exit(GetIniErrorString(),1);
   for (p = strtok(a,"="), i=0; p && i < MAX_CLIENTS; p = strtok(NULL,"="), i++) MSTR(host[i],p)
   if (i == MAX_CLIENTS) dbg("# of clients limit exceeded");

/*---------------------------- Load ip's from the .ini file (these take precedence over /etc/hosts)
 * r flags that at least one ip was not in the ini file, so we will need to look at /etc/hosts
 *-----------------------------------------------------------------------------------------------*/
   for (i=0, r=0; host[i]; i++)
   {  if (!GetIniString("[PutClients]",host[i],"NULL",a,sizeof(a),ini_pfn)) dbge_exit(GetIniErrorString(),1);
      STR_NOT_NULL(ip[i],a)
      if (!ip[i]) r=1;
   }
/*--------------------------------------------------------- Attempt to look up ips in /etc/hosts */
   if (r)
   {  if (!(fp = fopen("/etc/hosts", "r"))) dbge_exit("fopen /ect/hosts",1);
      while (fgets(a,sizeof(a),fp))
      {  if (p = strchr(a,'#')) *p = '\0';           /* handles comment data */
         p = strtok(a," \t");                       /* ip address, if any */
         h = strtok(NULL," \t");                    /* host name, if any */
         if (!p || !h) continue;                    /* not a data line */
         for (i=0; i < MAX_CLIENTS && host[i]; i++)   /* pass hosts from .ini */
         {  if (STR_EQ(host[i],h) && !ip[i])          /* match host to file */
            {  MSTR(ip[i],p)                           /* load ip from file */
               break;
            }
         }
      }
      fclose(fp);
   }
/*------------------------------------------------------------- Note & delete clients with no ip */
   for (i=0; host[i]; i++)
   {  if (!ip[i]) 
      {  sprintf(em,"client %s, no ip", host[i]); dbg(em);
         FREE(host[i])
         for (j=i; j==i || host[j]; j++) {host[j]=host[j+1]; ip[j]=ip[j+1];}
         i--;
      }
   }

/*-------------------------------------------------------------------------------------------------
 * Match clients in table to .ini data. Clear any clients in the table which aren't in .ini, update
 * the ips of those in the table as needed
 *-----------------------------------------------------------------------------------------------*/
   for (client=0; client < max_client; client++)             /* pass existing clients */
   {  for (i=0, r=0; i < MAX_CLIENTS && host[i]; i++)        /* pass .ini clients */
      {  if (STR_EQ(CLIENT.name,host[i]))                    /* match */
         {  r = 1;                                           /* r flags that we did match */
            for (j=0, p = strtok(ip[i],"."); j < 4; p = strtok(NULL,"."), j++) ipn[j] = ( p ? atoi(p) : 0); /* numeric ip */
            for (j=0; j < 4; j++)              /* update ip, close open socket if it changes */
            {  if (CLIENT.ip[j] != ipn[j]) 
               {  if (CLIENT.sock) 
                  {  close_socket(&CLIENT.sock);              /* close socket */
                     memset(CLIENT.buffer,0,buffer_size);     /* clear buffer */
                  }
                  CLIENT.ip[j] = ipn[j];                      /* update ip in table */
               }
            }
         }
      }
      if (!r) clear_put_client(client); client--;                 /* client not in .ini file */
   }
/*-------------------------------------------------- Now we look for new clients and set them up */
   for (i=0, r=0; host[i]; i++)
   {  for (client=0; client < max_client; client++) if (STR_EQ(CLIENT.name,host[i])) { r=1; break; }
      if (r) continue;
      client = max_client; max_client++;
      MSTR(CLIENT.name,host[i]);
      for (j=0, p = strtok(ip[i],"."); j < 4; p = strtok(NULL,"."), j++) CLIENT.ip[j] = ( p ? atoi(p) : 0);
      CLIENT.buffer = malloc(max_record_size);
      memset(CLIENT.buffer,0,max_record_size);
      CLIENT.sock = 0;
      CLIENT.rts = 0;
      CLIENT.msgs_today = CLIENT.msgs_hour = 0;
      CLIENT.connects_today = CLIENT.connects_hour = 0;
      CLIENT.last_msg = CLIENT.last_connect = 0;
   }
} 

/*-------------------------------------------------------------------------------------------------
 * logs message with time hacks, computes latency
 *-----------------------------------------------------------------------------------------------*/
static HL7MSG hl7msg;
void log_msg(char * msg)
{  time_t recv, send;
   unsigned long latency;
   struct tm t;
   struct tm * tm_recv = &t, * tm_send;
   char a[80];
   FILE * fp;
   int i;
   send = time(NULL);
   tm_send = localtime(&send);
/*------------------------------------------------- Load time from time hack in front of message */
   memset(a,0,sizeof(a));
   strncpy(a,file1_buffer,20);
   tm_recv->tm_year = atoi(strtok(a,"/")) - 1900;
   tm_recv->tm_mon  = atoi(strtok(NULL,"/")) -1;
   tm_recv->tm_mday = atoi(strtok(NULL," "));
   tm_recv->tm_hour = atoi(strtok(NULL,":"));
   tm_recv->tm_min  = atoi(strtok(NULL,":"));
   tm_recv->tm_sec  = atoi(strtok(NULL,":"));
   tm_recv->tm_isdst = -1;
   recv = mktime(tm_recv);
   latency = difftime(send, recv);
   if (latency > max_latency) max_latency = latency;           /* Update max latency for warning */

/*----------------------------------------------------------------------- Update latency history */
   for (i = MAX_LATENCIES -1; i >= 0; i--)
   {  LATENCY.time = latencies[i-1].time;
      strcpy(LATENCY.alpha,latencies[i-1].alpha);
   }
   i = 0;
   LATENCY.time = latency;
   sprintf(LATENCY.alpha,"%d:%d:%d",(int) latency/3600, (int) latency/60, (int) latency % 60);

   memset(a,0,sizeof(a));
   sprintf(a,"received: %d/%d/%d %d:%d:%d  sent: %d/%d/%d %d:%d:%d  latency: %d:%d:%d\n",
      tm_recv->tm_year/100, tm_recv->tm_mon+1, tm_recv->tm_mday, tm_recv->tm_hour, tm_recv->tm_min, tm_recv->tm_sec,
      tm_send->tm_year/100, tm_send->tm_mon+1, tm_send->tm_mday, tm_send->tm_hour, tm_send->tm_min, tm_send->tm_sec,
      (int) latency/3600, (int) latency/60, (int) latency % 60);

   if (hl7ReadMsg(&hl7msg, msg, sizeof(msg)))   PrtHl7ErrorString(debug_fp, "log hl7 read"); 
   if (logHl7Msg("MESSAGES", a, &hl7msg, hl7_wrap_pars)) dbg(GetLogErrorString());
   hl7FreeMemory(&hl7msg);

}

/*-------------------------------------------------------------------------------------------------
 * Periodic check for alert warnings
 *-----------------------------------------------------------------------------------------------*/
static char wmon[1024] = "\n";    /* warning monitor report */
void warning_check(void)
{  static time_t last_check = 0;
   static time_t last_warning = 0;
   char w_MB = 0, w_latency = 0, w_no_input = 0, w_no_output = 0;
   struct tm * tm;
   struct stat st;
   unsigned int buff_size, m_latency, m_no_input, m_no_output;
   FILE * log = NULL;
   char b[20], a[128], msg[MSG_SIZE];
/*--------------------------------------------------- Quick check; If no warnings set up, return */
   if (!check_interval || (!warning_MB && !warning_latency && !no_input && !no_output)) 
   {  strcpy(wmon,"\nAlert warnings:\n   Not implemented\n");
      return;
   }    

   if (!last_check)                                             /* First call */ 
   {  strcpy(wmon,"\nAlert warnings:\n   Check not yet run\n");
      last_check = last_warning = current_time; 
      return; 
   }  
   
   if ((current_time - last_check) < (check_interval * 60)) return;        /* not time for next check yet */
   last_check = current_time;
   sprintf(em,"\nAlert warnings:\n   Checked every %d minutes\n", check_interval); strcpy(wmon,em);
   cftime(em,"   Last check at %c\n", &last_check); strcat(wmon,em);
   if (last_warning) { cftime(em,"   Last warning issued %c\n", &last_warning); strcat(wmon,em); }
/*-------------------------------------------------------------------------------------------------
 * MB warning: Compute file size from blocks & match against warning threshold
 *---------------------------------------------------------------------------------------------- */
   if (warning_MB)
   {  buff_size = 0;
      if (!stat(file1_pfn, &st)) buff_size += st.st_blocks;
      if (!stat(file2_pfn, &st)) buff_size += st.st_blocks;
      buff_size = buff_size / 2048;
      if (buff_size >= warning_MB) w_MB = 1;
      sprintf(em,"   Buffer size %d MB - Warn at %d MB", buff_size, warning_MB); strcat(wmon,em);
      if (w_MB) strcat(wmon, " *** WARNING ***"); 
      strcat(wmon,"\n");  
   }

/*-------------------------------------------------------------------------------------------------
 * Latency warning: Warn if highest latency recorded since last check hits threshold 
 *-----------------------------------------------------------------------------------------------*/
   if (warning_latency)
   {  m_latency = max_latency / 60;
      if (m_latency >= warning_latency) w_latency = 1;
      max_latency = 0;
      sprintf(em,"   High latency %d min - Warn at %d min", m_latency, warning_latency); strcat(wmon,em);
      if (w_latency) strcat(wmon, " *** WARNING ***"); 
      strcat(wmon,"\n");  
   }

/*-------------------------------------------------------------------------------------------------
 * no input warning: If last input is greater than threshold minutes ago
 *-----------------------------------------------------------------------------------------------*/
   if (no_input)
   {  m_no_input = (current_time - last_input) / 60;
      if (m_no_input >= no_input) w_no_input = 1;
      sprintf(em,"   No input %d min - Warn at %d min", m_no_input, no_input); strcat(wmon,em);
      if (w_no_input) strcat(wmon, " *** WARNING ***"); 
      strcat(wmon,"\n");  
   }

/*-------------------------------------------------------------------------------------------------
 * no output warning: similar to no input, except that this is checked only when we have something
 * to send
 *-----------------------------------------------------------------------------------------------*/
   if (no_output)
   {  m_no_output = (current_time - last_output) / 60;
      if (file1_fp || file2_has_data)
         if (m_no_output >= no_output) w_no_output = 1;
      sprintf(em,"   No output %d min - Warn at %d min", m_no_output, no_output); strcat(wmon,em);
      if (w_no_output) strcat(wmon, " *** WARNING ***"); 
      if (!file1_fp && !file2_has_data) strcat(wmon, " no data waiting");
      strcat(wmon,"\n");  
   }

/*--------------------------------------------------------------- Over threshold, issue warnings */
   if (!w_MB && !w_latency && !w_no_input && !w_no_output) return;
   if ((current_time - last_warning) < (warning_interval * 60)) return;  /* Not time for next warning yet */
   last_warning = current_time;
   tm = current_tm;

   memset(msg,0,sizeof(msg));

/*------------------------------------------------------------------------- print/e-mail message */
   sprintf(msg,
      "****************************************************\n"
      "* PACS Monitor System - Buffered Fifo warning\n"
      "* %s FIFO\n", name);
   if (w_MB) { sprintf(a,"* %d MB backlog\n", buffer_size); strcat(msg,a); }
   if (w_latency) { sprintf(a,"* message latency at %d seconds\n", m_latency); strcat(msg,a); }
   if (w_no_input) { sprintf(a,"* no input in %d minutes\n", m_no_input); strcat(msg,a); }
   if (w_no_output) { sprintf(a,"* no output in %d minutes\n", m_no_output); strcat(msg,a); }
   sprintf(a,"* %s****************************************************\n", asctime(tm)); strcat(msg,a);
   
   if (!alertSendPrintEmail(warning_printer, warning_email, "PACS Mon", msg, NULL))
   {  PrtAlertErrorString(debug_fp, "Print/Email");
   }
   if (logMsg("ERRORS", msg)) dbg(GetLogErrorString());
/*-------------------------------------------------------------------------------- Pager message */
   if (warning_pager)
   {  sprintf(msg, "PACS Mon FIFO\n%s\n", name);
      cftime(b,"%D %R\n",&current_time); strcat(msg,b);
      if (w_MB) { sprintf(a,"%d MB backlog\n", buffer_size); strcat(msg,a); }
      if (w_latency) { sprintf(a,"%d sec latency\n", m_latency); strcat(msg,a); }
      if (w_no_input) { sprintf(a,"no input %d mins\n", m_no_input); strcat(msg,a); }
      if (w_no_output) { sprintf(a,"no output %d mins\n", m_no_output); strcat(msg,a); }
      if (!alertSendEmail(warning_pager, "PACSMon", msg, NULL))
      {  PrtAlertErrorString(debug_fp, "Pager");
      }
   }
}

/**************************************************************************************************
 * Set up put server listen socket 
 *************************************************************************************************/
void setup_put_sock(void)
{		
   if (!put_port) return;
	sprintf(em,"Initializing put listen socket on port %d", put_port); dbg(em);
	
	if ((put_sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) dbge_exit("socket", 1);
	if (setsockopt(put_sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1) dbge_exit("setsockopt", 1);
	if (gethostname(host_name, sizeof(host_name)) == -1) dbge_exit("gethostname",1);
   if (!(phostent = gethostbyname(host_name))) dbge_exit("gethostbyname",1);

	shostent = *phostent;
   put_name.sin_family = shostent.h_addrtype;
   put_name.sin_port = htons(put_port);
   put_name.sin_addr = *((struct in_addr *)shostent.h_addr);

   if (bind(put_sock, put_ptr, sizeof(put_name)) == -1) dbge_exit("bind", 1);
   if (listen(put_sock, 10) == -1) dbge_exit("listen", 1);

   dbg("Server initialized\nWaiting for connection");
}

/**************************************************************************************************
 * Set up get server listen socket 
 *************************************************************************************************/
void setup_get_sock(void)
{		
   if (!get_port) return;
	sprintf(em,"Initializing get listen socket on port %d", get_port); dbg(em);
	
	if ((get_sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) dbge_exit("socket", 1);
	if (setsockopt(get_sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1) dbge_exit("setsockopt", 1);
	if (gethostname(host_name, sizeof(host_name)) == -1) dbge_exit("gethostname",1);
   if (!(phostent = gethostbyname(host_name))) dbge_exit("gethostbyname",1);

	shostent = *phostent;
   get_name.sin_family = shostent.h_addrtype;
   get_name.sin_port   = htons(get_port);
   get_name.sin_addr   = *((struct in_addr *)shostent.h_addr);

   if (bind(get_sock, get_ptr, sizeof(get_name)) == -1) dbge_exit("bind", 1);
   if (listen(get_sock, 10) == -1) dbge_exit("listen", 1);

   dbg("Server initialized\nWaiting for connection");
}

/**************************************************************************************************
 * Attempt to connect to get server on target machine.  This is done in two parts.  The first part
 * creates the socket and tries to connect, non-blocking.  If it succeeds, great.  If it generates
 * an "in progress" error, it ties into the main poll to wait for the connect to finish.  When the
 * connect finishes, the second part picks up and completes the connection.  Attempts to connect
 * are made initially at min 5 second intervals, but if repeated attempts to connect fail, the 
 * interval between attempts is gradually raised to 5 minutes.
 *************************************************************************************************/
void connect_get_socket(char place /* 1 = before poll, 2 = after poll */)
{  static char state = 1;   /* 1 = need to connect, 2 = waiting for select */
   static int seconds_to_wait = 0;
   static time_t last_connect_attempt = 0;
   int sock, i, r;
   char a[24];
   if (gtc.sock || connect_as != 'c') return;  /* Already connected, or not a client */
   
   if (place == 1 && state == 1)                 /* Need to connect, before poll */
   {  if ((current_time - last_connect_attempt) < seconds_to_wait) return;
      sock = socket(AF_INET, SOCK_STREAM, 0);
      if (sock == -1) dbge_exit("get socket", 1);
      i = fcntl(sock, F_GETFL, 0);
      fcntl(sock, F_SETFL, O_NONBLOCK|i);
      phostent = gethostbyname(gtc.name);
      if (!phostent)
      {  sprintf(a,"%d.%d.%d.%d", gtc.ip[0], gtc.ip[1], gtc.ip[2], gtc.ip[3]);
         phostent = gethostbyname(a);
      }
      shostent = *phostent;
      get_name.sin_family = shostent.h_addrtype;
      get_name.sin_port = htons(get_port);
      get_name.sin_addr = *((struct in_addr *)shostent.h_addr);
      i = sizeof(get_name);
      r = connect(sock, (struct sockaddr *)&get_name, i);
      if (r == 0){gtc.sock=sock; seconds_to_wait=0; dbg("get connect"); return;}
      if (errno == EINPROGRESS) {state = 2; dbg("get connect in progress"); }
      else
      {  close_socket(&sock);
         if (seconds_to_wait < 300) seconds_to_wait += 5; 
         last_connect_attempt = current_time;
         state = 1;
         sprintf(em, "get connect fail # %d", seconds_to_wait / 5); dbg(em);
      }
   }

   if (place == 1 && state == 2) { fd_st(sock, POLLOUT); }

   if (place == 2 && state == 2)
   {  if (fd_isset(sock, POLLERR))
      {  close_socket(&sock);
         if (seconds_to_wait < 300) seconds_to_wait += 5; 
         last_connect_attempt = current_time;
         sprintf(em, "get connect fail # %d", seconds_to_wait / 5); dbg(em);
         state = 1;
      }
      else if (fd_isset(sock, POLLOUT))
      {  gtc.sock = sock;
         seconds_to_wait = 0;
         state = 1;
         dbg("get connected");
      }
   }
}

/************************************************************************** Close/cleanup socket */
void close_socket(int * socket)
{	if (*socket > 0) {shutdown(*socket,2); close(*socket);}
	*socket = 0;
}

/*-------------------------------------------------------------------------------------------------
 * Clear client from client table - Also closes socket if open, bumps up the other clients to close
 * the empty slot and reduces the client count.
 *-----------------------------------------------------------------------------------------------*/
#define NEXT_CLIENT clients[client+1]
void clear_put_client(char client)
{  if (CLIENT.sock) close_socket(&CLIENT.sock);           /* close socket if active */
   FREE(CLIENT.name)                                  /* return allocated memory */
   FREE(CLIENT.buffer)
   while (client+1 < max_client)                          /* close up empty slot */
   {  CLIENT.name = NEXT_CLIENT.name;
      CLIENT.ip[0] = NEXT_CLIENT.ip[0];
      CLIENT.ip[1] = NEXT_CLIENT.ip[1];
      CLIENT.ip[2] = NEXT_CLIENT.ip[2];
      CLIENT.ip[3] = NEXT_CLIENT.ip[3];
      CLIENT.buffer = NEXT_CLIENT.buffer;
      CLIENT.sock = NEXT_CLIENT.sock;
      CLIENT.rts = NEXT_CLIENT.rts;
      CLIENT.msgs_today = NEXT_CLIENT.msgs_today;
      CLIENT.msgs_hour = NEXT_CLIENT.msgs_hour;
      CLIENT.connects_today = NEXT_CLIENT.connects_today;
      CLIENT.connects_hour = NEXT_CLIENT.connects_hour;
      CLIENT.last_msg = NEXT_CLIENT.last_msg;
      CLIENT.last_connect = NEXT_CLIENT.last_connect;
   }
   client = max_client--;                              /* clean up now empty last slot */
   CLIENT.name = CLIENT.buffer = 0;
   CLIENT.ip[0] = CLIENT.ip[1] = CLIENT.ip[2] = CLIENT.ip[3] = 0;
   CLIENT.rts = 0;
   CLIENT.msgs_today = CLIENT.msgs_hour = CLIENT.connects_today = CLIENT.connects_hour = 0;
   CLIENT.sock = 0;
   CLIENT.last_msg = CLIENT.last_connect = 0;
}

/*-------------------------------------------------------------------------------------------------
 * Converts messages between internal and external formats.  Coming in, strips <SB> and <EB> and
 * adds time hack.  Going out, strips time hack and adds <SB> and <EB> if they are called for.
 * Note that in and out cannot be the same memory area. in assumed to be max_record_size and out
 * assumed to be buffer_size.
 *-----------------------------------------------------------------------------------------------*/

char * internalize(char * out, char * in)
{  assert(in && out);
   cftime(out,"%Y/%m/%d %H:%M:%S ", &current_time);
   if (*in == SB) strcat(out,in+1);
   else strcat(out,in);
   if (LAST_CHR(out) == EB) LAST_CHR(out) = '\0';
   return (out);
}

char * externalize(char * out, char * in)
{  char * p = in + TIME_HACK;
   if (LAST_CHR(p) == '\n') LAST_CHR(p) = '\0';        /* Pull line feed if present */
   if (!sb_eb || (*p == SB && LAST_CHR(p) == EB)) 
   {  strcpy(out,p);                              /* packet bytes not needed, or already there */
   }
   else
   {  strcpy(out,"\013\0");                       /* Put on packet bytes */
      strcat(out,p);
      strcat(out,"\034");
   }
   strcat(out,"\n");                              /* Add line feed at end */
   return (out);
} 


/*-------------------------------------------------------------------------------------------------
 * Put Listen socket.  Read is open of new .put connection from a client
 *-----------------------------------------------------------------------------------------------*/
void check_put_socket(void)
{	int ret, new_socket;	
	char ip[18];
	unsigned char * p;
	unsigned char flag;
   
	if (!put_port || !fd_isset(put_sock, POLLIN)) return;                    
/*---------------------------------------------------------- Accept new connection on put socket */
	new_socket = accept(put_sock, (struct sockaddr*)&put_name, &sl);
	if (new_socket == -1) {	dbge("accept"); return; }
	
	ret = setsockopt (new_socket, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof (on));
	if (ret == -1) { dbge("setsockopt"); return; }

	p = (unsigned char *) &put_name.sin_addr;
	sprintf(ip, "%3d.%3d.%3d.%3d", p[0], p[1], p[2], p[3]);
	
/*-------------------------------------------------------------------------------------------------
 * Scan the clients.  The new socket's ip address must match one of them, and the connection
 * is for that distributor.  If that distributor already has a connection, we need to clear it and
 * set this up as the new one
 *-----------------------------------------------------------------------------------------------*/
	for(client = 0; client < max_client; client++)
	{	if (CLIENT.ip[0] == p[0] && CLIENT.ip[1] == p[1] && CLIENT.ip[2] == p[2] && CLIENT.ip[3] == p[3])
		{	if (CLIENT.sock) 
			{	fd_clr(CLIENT.sock);
				close_socket(&CLIENT.sock);
			}
			CLIENT.sock = new_socket;
         CLIENT.rts = 0;
         memset(CLIENT.buffer,0,max_record_size);
         CLIENT.connects_today++;
         CLIENT.connects_hour++;
         CLIENT.last_connect = time(NULL);
			sprintf(em,"Connected to %s", CLIENT.name); dbg(em);
			return;
		}
	}
	sprintf(em,"Attempt to connect from %s rejected", ip); dbg(em);
   for (intr = MAX_INTRUDERS -1; intr >= 0; intr--)
   {  strcpy(INTRUDER.ip,intruders[intr -1].ip);
      INTRUDER.at = intruders[intr -1].at;
   }
   intr = 0;
   strcpy(INTRUDER.ip,ip); 
   INTRUDER.at = time(NULL);
	return;
}
/*-------------------------------------------------------------------------------------------------
 * Get Listen socket.  Read is open of new .get connection from client
 *-----------------------------------------------------------------------------------------------*/
void check_get_socket(void)
{	int ret, new_socket;	
	char ip[18];
	unsigned char * p;
	unsigned char flag;
   
	if (!get_port || (connect_as != 's') || !fd_isset(get_sock, POLLIN)) return;                    
/*---------------------------------------------------------- Accept new connection on put socket */
	new_socket = accept(get_sock, (struct sockaddr*)&get_name, &sl);
	if (new_socket == -1) {	dbge("accept"); return; }
	
	ret = setsockopt (new_socket, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof (on));
	if (ret == -1) { dbge("setsockopt"); return; }

	p = (unsigned char *) &get_name.sin_addr;
	sprintf(ip, "%3d.%3d.%3d.%3d", p[0], p[1], p[2], p[3]);
	
/*-------------------------------------------------------------------------------------------------
 * The new socket's ip address must match the get client, and the connection is for that 
 * distributor.  If there is already a connection, clear it and set this up as the new one
 *-----------------------------------------------------------------------------------------------*/
	if (gtc.ip[0] == p[0] && gtc.ip[1] == p[1] && gtc.ip[2] == p[2] && gtc.ip[3] == p[3])
	{	if (gtc.sock) 
		{	fd_clr(gtc.sock);
			close_socket(&gtc.sock);
		}
		gtc.sock = new_socket;
      gtc.rts = 0;
      memset(gtc.buffer,0,max_record_size);
      gtc.connects_today++;
      gtc.connects_hour++;
      gtc.last_connect = time(NULL);
	   sprintf(em,"Connected to %s", gtc.name); dbg(em);
		return;
	}
	sprintf(em,"Attempt to connect from %s rejected", ip); dbg(em);
   for (intr = MAX_INTRUDERS -1; intr >= 0; intr--)
   {  strcpy(INTRUDER.ip,intruders[intr -1].ip);
      INTRUDER.at = intruders[intr -1].at;
   }
   intr = 0;
   strcpy(INTRUDER.ip,ip); 
   INTRUDER.at = time(NULL);
	return;
}

/*----------------------------------------------- Clears hourly and daily totals as required */
void totals_check(void)
{  static int current_mday = -1;
   static int current_hour = -1;
   char h = 0, d = 0;
   
   if (current_tm->tm_mday != current_mday) { d = 1; current_mday = current_tm->tm_mday; }
   if (current_tm->tm_hour != current_hour) { h = 1; current_hour = current_tm->tm_hour; }

   if (d) input_today = output_today = 0;
   if (h) input_hour  = output_hour  = 0;
   
   if (h | d)
   {  for (client = 0; client < max_client; client++)
      {  if (d) CLIENT.msgs_today = CLIENT.connects_today = 0;
         if (h) CLIENT.msgs_hour  = CLIENT.msgs_hour = 0;
      }
   }
}

/*-------------------------------------------------------------------------------------------------
 * prints out status display to monitor fifo every so often, provided that it is open
 *-----------------------------------------------------------------------------------------------*/
void ltime(char *d, time_t time) {cftime(d,"%m/%d %H:%M:%S",&time);}
   
void monitor_check(void)
{  static time_t last_monitor = 0;
   int fd, i;
   FILE * fp;
   char lmsg[15], lcon[15];
   char src_fmt[] = "%14s %7d %5d %s  %7d %5d %s\n";
   char clear_screen[] = "\033[2J\n";   /*clear screen ESC sequence (vt100) */
   
/*------------------------------------------- check for mon fifo, monitoring interval & time yet */
   if (!mon_pfn || !monitoring_interval || ((current_time - last_monitor) < monitoring_interval)) return;
   last_monitor = current_time;

/*------------------------------------------------ Open non-blocking, then open stream to append */
   fd = open(mon_pfn, O_WRONLY | O_NONBLOCK);
   if (fd < 0) return;
   
   fp = fdopen(fd, "a");
   if (!fp) dbge_exit("mon fdopen", 1);
   setvbuf(fp, NULL, _IOLBF, 0);

/*---------------------------------------------------------------- Display monitor status report */
   fprintf(fp,clear_screen);
   fprintf(fp,"%s FIFO buffer status, as of %sUpdated every %d seconds or at timeout\n\n", 
      name, asctime(current_tm), monitoring_interval);
   fprintf(fp,"                         Messages                     Connections\n"
              "               ----------------------------  ----------------------------\n"
              "                 Today  Hour  Last Message     Today  Hour  Last Connect\n"
              "--------------------------------------------------------------------------------\n"
              "Source\n");
   if (input_pfn) 
   {  ltime(lmsg, last_input);
      fprintf(fp, src_fmt, "input fifo", input_today, input_hour, lmsg, 0, 0, " ");
   }
   if (put_port)
   {  for (client = 0; client < max_client; client++)
      {  ltime(lmsg, CLIENT.last_msg);
         ltime(lcon, CLIENT.last_connect);
         fprintf(fp, src_fmt, CLIENT.name, CLIENT.msgs_today, CLIENT.msgs_hour, lmsg,
            CLIENT.connects_today, CLIENT.connects_hour, lcon);
      }
   }
   fprintf(fp,"Destination\n");
   ltime(lmsg, last_output);
   fprintf(fp,src_fmt, "output fifo", output_today, output_hour, lmsg, 0, 0, " ");

   i =0;
   if (strlen(LATENCY.alpha))
   {  fprintf(fp,"\nMost recent %d message latencies\n--------------------------------\n", MAX_LATENCIES);
      for (i = 0; i < MAX_LATENCIES; i++)
         if (strlen(LATENCY.alpha)) fprintf(fp,"%s\n", LATENCY.alpha);
   }

   intr = 0;
   if (INTRUDER.at)
   {  fprintf(fp,"\nMost recent %d intruders\n--------------------------------\n", MAX_INTRUDERS);
      for (intr = 0; intr < MAX_INTRUDERS; intr++)
         if (INTRUDER.at) fprintf(fp,"%18s %s", INTRUDER.ip, ctime(&INTRUDER.at));
   }

   fputs(wmon, fp);    /* warning check info */
   
   fclose(fp);   
}

void fd_st(int socket, short flag)
{  unsigned char pfd;
   for (pfd = 0; pfd < MAX_POLL; pfd++)
   {  if (POLL.fd == -1) POLL.fd = socket;
      if (POLL.fd == socket) { POLL.events = POLL.events | flag; break; }
   }
}

int fd_isset(int fd, short flag)
{  unsigned char pfd;
   for (pfd = 0; pfd < MAX_POLL && POLL.fd; pfd++) 
      if (POLL.fd == fd) return (POLL.revents & flag);
   return 0;
}

void fd_clr(int fd)
{  unsigned char pfd;
   for (pfd = 0; pfd < MAX_POLL && POLL.fd; pfd++) 
      if (POLL.fd == fd) POLL.revents = 0;
   return;
}

int fd_num(void)
{  int pfd;
   for (pfd = 0; pfd < MAX_POLL; pfd++) if (POLL.fd == -1) return pfd + 1; 
}
