/****************************************************************************************
 * MIRCS_dicom_broadcast                                                        R Moulton
 * A CTN based dicom_broadcast program tuned to the needs of MIR Computer Section, based
 * on dicom_broadcast, written by S Moore.  The program will broadcast all DICOM image
 * files in the specified directory and its subdirectories to the given destination.
 * Configuration information is given in the MIRCS_dicom_broadcast.ini file, which 
 * must be in the program directory. 
 * --------------------------------------------------------------------------------------
 * Compile with makeFile
 * Invocation: MIRCS_dicom_broadcast study_directory_path dest_code overide_AE_title
 ***************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>

#include <iostream>
using namespace std;
#include "ctn_os.h"
#include "dcm_api.h"

#include "MBroadcastManager.hpp"

CTNBOOLEAN silent = TRUE;
static CTNBOOLEAN verbose = FALSE;

extern "C" {
   #include "dbg.h"
   #include "ini.h"
   #include "log.h"
}

/*----------------------------- Function prototypes -----------------------------------*/
static int read_ini(void);
static char * clip_comment(char * text);
static void makeFileVector(vector<char*>& fileVector);
static void log_broadcast(int status);


/*---------------------------- macros -------------------------------------------------*/
#define FREE(A) {if(A)free(A);A=NULL;}
#define STR_EQ(A,B) !strcasecmp(A,B)
#define STR(A,B) A=strdup(B)
#define LCHR(A) A[strlen(A) - 1]
#define CLEAR(A) memset(A,0,sizeof(A))
#define CONDERR {sprintf(em,"COND=%x",cond); dbg(em); COND_PopCondition(TRUE);}
#define CLEARCOND COND_PopCondition(TRUE);


/*--------------------------------------------------------- cmd line params, .ini file */
static char * cwd;                      /* Current working directory (program dir) */
static char inipfn[256];                /* .ini file pfn */
static FILE * debug_fp;                 /* debug file pointer, or NULL if debug is off */
static char * study_directory;          /* root directory study is in */
static char * dest_code;                /* destination code, look up in .ini file */
static char * destinations;             /* destination value from .ini file */
static char * overide_AE_title = NULL;  /* use this as AE title */

void * msg_pointer = NULL;
char * msg_text = NULL;
unsigned int msg_result;

/****************************************************************************************
 ****************************************************************************************
 ********************************** main function ***************************************
 ****************************************************************************************
 ***************************************************************************************/
int main(int argc, char * argv[])
{  char * p, * ip_port;
   struct stat buf;
   void* timeStamp;
   double deltaTime;
   int status, flag;
   char a[2048], pfn[256];
   int sa = sizeof(a);
   char buffer[256];
   FILE * fp;
   char * reportMgr = "";
   char * host = "none";
   char * queuename = "test";
/*---------------------------------------------------------------------------------------
 * Program runs out of directory binary is in.  Make that current.  Then use parameter 1
 * or default to complete .ini file pfn.
 *-------------------------------------------------------------------------------------*/
   if (p = MakeProgDirCurrent(argv[0])) STR(cwd,p);
   else { PrtIniErrorString(stdout,"Startup"); return(1); }
   strcpy(inipfn,cwd); strcat(inipfn,"MIRCS_dicom_broadcast.ini");

/*------------------------------------------------------------ General .ini parameters */
   if (read_ini()) return (1);
   
/*---------------------------------------------------------------------------------------
 * First parameter is the complete path of the study directory. Second parameter must be
 * destination code in .ini file.  Third parameter is optional overide AE title.
 *-------------------------------------------------------------------------------------*/
   if (argc <= 2)
   {  printf("Usage:\nMIRCS_dicom_broadcast study_dir dest_code [overide_AE_title]\n");
      return(1);
   }
   
   STR(study_directory,argv[1]);
   if (stat(study_directory,&buf)) dbge_exit("stat study directory:", 1); 
   if (!S_ISDIR(buf.st_mode)) 
   {  sprintf(em,"%s is not a directory", study_directory); 
      dbg_exit(em, 1); 
   }
   
   STR(dest_code,argv[2]);
   if (!GetIniString("[StorageControl]", dest_code, "NOF", a, sa, inipfn) || STR_EQ(a,"NOF"))
   {  sprintf(em,"%s is not a valid destination code", dest_code);
      dbg_exit(em, 1);
   }
   STR(destinations,clip_comment(a));
   
/*---------------------------------------------------------------------------------------
 * AE title: Use dest code. Overide with code from .ini, overide with code from cmd line
 * (.ini file is dest_code=AE,AE,AE...|overide AE
 * Note that the overide AE is checked whether or not the .broadcast.map file exists, 
 * because it is not put into the .broadcast.map file
 *-------------------------------------------------------------------------------------*/
   overide_AE_title = dest_code;
   if (strchr(destinations,'|'))
   {  strtok(destinations,"|");
      STR(overide_AE_title,strtok(NULL,"| "));
      sprintf(em,"AE overide %s, from .ini file",overide_AE_title); dbg(em);
   }
   if (argc > 3) 
   {  STR(overide_AE_title,argv[3]);
      sprintf(em,"AE overide %s, from cmd line",overide_AE_title); dbg(em);
   }   

/*---------------------------------------------------------------------------------------
 * If the .broadcast.map.dest_code file does not exist in the study directory, create it.
 * It contains records for each of the destinations, of the form:
 * "dest_code\tAEtitle:IPaddress:port\n"
 * Information is retrieved from the .ini file.
 *-------------------------------------------------------------------------------------*/
   sprintf(pfn,"%s/.broadcast.map.%s", study_directory, dest_code);
   if (access(pfn,F_OK))
   {  fp = fopen(pfn,"a");
      if (!fp) dbge_exit("create .map file", 1);
      flag = 0;
      while (p = strtok_r(destinations,",",&destinations))
      {  if (GetIniString("[ApplicationEntity]", p, "NOF", a, sa, inipfn) && strcasecmp(a,"NOF"))
         {  clip_comment(a);
            ip_port = strtok(a,"|,");
            if (ip_port)
            {  sprintf(buffer,"%s\t%s:%s\n", dest_code, p, ip_port);
               fputs(buffer,fp);
               dbg(buffer);
               flag = 1;
            }
         }
      }
      fclose(fp);
      if (!flag) {sprintf(em,"No destination data for %s", dest_code); dbg_exit(em,1);}
   }
   else dbg(".map file already exists");
               

   timeStamp = ::UTL_GetTimeStamp();

   ::THR_Init();

   vector<char*> fileVector;
   makeFileVector(fileVector);

   MBroadcastManager manager;
   status = manager.broadcast(reportMgr, dest_code, overide_AE_title, study_directory,
				fileVector, host, queuename); 

   ::COND_DumpConditions(); 
   deltaTime = ::UTL_DeltaTime(timeStamp);
   ::UTL_ReleaseTimeStamp(timeStamp);
   printf ("Broadcast elapsed time (sec): %.2f\n", deltaTime);
   ::THR_Shutdown();

   log_broadcast(status);

   return status;
}

/*---------------------------------------------------------------------------------------
 * Builds the vector of files which are to be transmitted.
 * This function has been modified from the original, which worked off a wildcard on the
 * command line.  It presumes that the study_directory has series directories in it, and
 * that the files in these directories are the images which are to be sent.
 * Because of the ucb library problem, the actual location of the pfn is two characters
 * before the struct pointer
 *-------------------------------------------------------------------------------------*/
#define D_NAME (dnt->d_name - 2)
static void makeFileVector(vector<char*>& fileVector)
{  DIR * study_dp = NULL, * series_dp = NULL;
   struct dirent * dnt = NULL;
   struct stat buf;
   char *p, series_directory[256], image_file[256], pfn[256];
   int i = 0;
/*--------------------------------------------------------------- Open study directory */
   dbg(study_directory);
   study_dp = opendir(study_directory);
   if (!study_dp) dbge_exit("opendir study_directory", 1);
/*--------------------------------------------------------------- Pass study directory */
   while (dnt = readdir(study_dp))
   {  strcpy(series_directory,D_NAME);
      if (*series_directory == '.') continue;                  /* Ignore all '.' files */
      sprintf(pfn,"%s/%s", study_directory, series_directory);
      if (stat(pfn, &buf)) continue;                  /* ignore anything we can't stat */
      if (!S_ISDIR(buf.st_mode)) continue;                          /* Not a directory */
/*-------------------------------------------------------------- Open series directory */
      sprintf(em,"   series: %s",series_directory); dbg(em);
      series_dp = opendir(pfn);
      if (!series_dp) continue;
/*-------------------------------------------------------------- Pass series directory */
      while (dnt = readdir(series_dp))
      {  strcpy(image_file,D_NAME);
         if (*image_file == '.') continue;
         sprintf(pfn,"%s/%s/%s", study_directory, series_directory, image_file);
         if (stat(pfn, &buf)) continue;
         if (S_ISDIR(buf.st_mode)) continue;
         p = new char[strlen(series_directory) + strlen(image_file) + 2];
         sprintf(p,"%s/%s", series_directory, image_file);
         fileVector.push_back(p);
         i++;
         sprintf(em,"      image %4d: %s", i, image_file); dbg(em);
      }
      closedir(series_dp);
   }
   closedir(study_dp);
   sprintf(em,"\n%d total images in study", i); dbg(em);
}

int read_ini(void)
{  char a[1024], main[] = "[MAIN]";
   int sa = sizeof(a);
   
   if (access(inipfn, F_OK)) { printf("%s file not found\n", inipfn); return(1); }
   
   if (!GetIniString(main, "Debug", "stdout", a, sa, inipfn))
   {  PrtIniErrorString(stdout, "ini file error");
      return(1);
   }
   debug_fp = debug_on(a);

   if (logInitialize(inipfn)) dbg(GetLogErrorString());

   return (0);
}

/*---------------------------------------------------------------------------------------
 * Log results of broadcast. Level 2 means log all Xmits, level 1 is Errors only.
 * Includes message text to describe message and/or CTN MSG facility dump.
 *-------------------------------------------------------------------------------------*/
void log_broadcast(int status)
{  char msg[10240];
   memset(msg,0,10240);
   sprintf(msg,"Xmit %s : %s %s", study_directory, dest_code, 
      overide_AE_title ? overide_AE_title : " ");
   if (!status)
   {  logNoteLevel("XMIT", msg, 2);
      return;
   }
   sprintf(em, "\n   Return = %d", status); strcat(msg,em);
   if (msg_text) 
   {  sprintf(em," - %s\n", msg_text);
      printf("%s\n", msg_text);
      strcat(msg,em);
   }
   else strcat(msg,"\n");
   
   logMsg("XMIT",msg);
   
   if (msg_pointer)
   {  FILE * fp = logOpen("XMIT");
      if (fp) 
      {  MSG_DumpMessage(msg_pointer, fp);
         MSG_DumpMessage(msg_pointer, stdout);
         logClose();
      }
   }
   logPurge("XMIT");
}
         
   

/*--------------------------------------------- Removes comment (# to eos) from string */
char * clip_comment(char *text)
{  if(strtok(text,"#"))
      while(LCHR(text) == ' ') LCHR(text)='\0';
   return text;
}



