/***********************************************************************************
 * MIRCS_image_scrubber.c                                                  R Moulton
 * Processes images received on a distributor, governed by MIRCS_simple_storage.ini,
 * which needs to be in the same directory as the MIRCS_image_scrubber binary.
 * This program is run two ways, forked from simple storage to scrub images in a
 * single transmit, and periodically from cron to scan for and process any images
 * which have not been processed using the previous method.  Build using ctn make.
 *----------------------------------------------------------------------------------
 * invoke: MIRCS_image_scrubber dir_to_prs pid
 * if pid is non-zero, MIRCS_image_scrubber is being run in the fork mode, pid is 
 * the process id of a child MIRCS_simple_storage process which is receiving images,
 * and dir_to_prs is a directory containing img_tmp image files which are to be
 * processed.
 * if pid is zero, MIRCS_image_scrubber is being urn in the scan mode, and 
 * dir_to_prs is a directory containing std_tmp directories, each of which contains
 * img_tmp files to be processed.
 *----------------------------------------------------------------------------------
 * 06-04-13 R Moulton - initial write
 **********************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>

#include "ctn_os.h"
#include "dicom.h"
#include "utility.h"
#include "dicom_uids.h"
#include "lst.h"
#include "condition.h"
#include "dulprotocol.h"
#include "dicom_objects.h"
#include "dicom_messages.h"
#include "dicom_services.h"

#include "MIRCS_simple_storage.h"
#include "MIRCS_image_processing.h"
#include "dbg.h"
#include "ini.h"
#include "log.h"

/******************************** function prototypes *****************************/
static void process_study(char * study_directory, pid_t pid);
static char * GetString(DCM_OBJECT *obj, DCM_TAG tag, char ** value);
static char * GetStringNOF(DCM_OBJECT *obj, DCM_TAG tag, char ** value);
static void fileNameChecks(char *name);

/******************************** macros ******************************************/
#define STR(A,B) { if (A) free(A); A=strdup(B); }
#define FREE(A) if (A) { free(A); A=NULL; }
#define STR_EQ(A,B) !strcasecmp(A,B)
#define ISON(A) (!strcasecmp(A,"ON") || !strcasecmp(A,"YES"))
#define LCHR(A) A[strlen(A) -1]
#define CLEAR(A) memset(A,0,sizeof(A))

/***************************** working storage ************************************/
static char inipfn[256] = "\0";
#define INI_FILE_NAME "MIRCS_simple_storage.ini"

FILE * debug_fp = NULL;

static char * image_directory   = NULL;
static char * base_directory    = NULL;
static char * receive_error_log = NULL;
static pid_t pid;

static CTNBOOLEAN ctn_true = TRUE;

/***********************************************************************************
 ***********************************************************************************
 ********************************** main function **********************************
 ***********************************************************************************
 **********************************************************************************/
int main (int numpars, char *par[])
{  char *p;
   struct stat st;
   char a[1024];
   int i;
   int sa = sizeof(a);
   DIR * dp;
   struct dirent * de;
   if (numpars > 3 && (i=atoi(par[3])))
   {  printf("scrubber pid = %d - debug pause %d seconds\n", getpid(), i); 
      sleep(i);
   } 
/*-------------------------------------------------------------- Get ini file pfn */
   p = par[0];
   if (!getcwd(inipfn,sizeof(inipfn)))            /* start with current directory */
      dbg_exit("Invalid current dir", 1);
   if (*p == '/') strcpy(inipfn,p);                       /* Prog called absolute */
   else if (*p == '.') strcat(inipfn,p+1);              /* Prog called . relative */
   else {strcat(inipfn,"/"); strcat(inipfn,p); }          /* Prog called relative */
   *(strrchr(inipfn,'/')+1) = 0;                     /* Insert eos after last "/" */
   strcat(inipfn,INI_FILE_NAME);                            /* Add .ini file name */

/*----------------------------------------------------------- validate parameters */
   if (numpars < 3) dbg_exit("invalid calling parameters", 2);
   pid = atoi(par[2]);
/*--------------------------------------------- This is directory we are scanning */
   if (stat(par[1], &st)) dbge_exit("stat",3);
   if (!S_ISDIR(st.st_mode)) dbg_exit("par 1 not a directory", 4);
   STR(base_directory,par[1])
      
   if (!GetIniString("[MAIN]", "Debug", "NULL", a, sa, inipfn))
      dbg_exit(GetIniErrorString(), 4);
   debug_fp = debug_on(a);
/*---------------------------------------------------------- base image directory */
   GetIniString("[MAIN]", "ImageDirectory", "/distributor/imgs", a, sa, inipfn);
   STR(image_directory,a)
   
   GetIniString("[MAIN]", "ReceiveErrorLog", "NONE", a, sa, inipfn);
   if (strcasecmp(a,"NONE")) STR(receive_error_log,a)

   MIRCS_Setup(par[0], image_directory);
/*------------------------------------------------------------- single study mode */
   if (pid) 
   {  sleep(10);
      process_study(base_directory, pid);
      return 0;
   }

/*---------------- Multiple study mode, scan base directory for study directories */
   if (!(dp=opendir(base_directory))) dbge_exit("base directory open", 5);
   while (de=readdir(dp))
   {  if ((p=strstr(de->d_name,"std_")) != de->d_name) continue;
      sprintf(a,"%s/%s", base_directory, de->d_name);
      if (stat(a,&st) || !S_ISDIR(st.st_mode)) continue;
      
      process_study(a, 0);
   }
   return 0;
} 

static void process_study(char * study_directory, pid_t pid)
{  char a[1024] = "\0", *p;
   int sa = sizeof(a);
   char * fifo_pfn = NULL;
   char * ae_pfn = NULL;
   char * ae_title = NULL;
   int number_of_times_scrubbed = 0;  /* # times scrubber has been run this study */
   int xmit_status;
   FILE * ae_fp = NULL;
   int lock_fd;
   static char reset = 0;                   /* Do MIRC routines need to be reset? */
   char first_image = 1;                         /* Is this the first image file? */
   char ok_del;                                  /* Ok to delete study directory? */
   DIR * dp;
   struct dirent * de;
   struct stat st;
   char image_file_pfn[1024];
   char new_directory[1024];
   char new_file_name[1024];
   CONDITION cond = DCM_NORMAL;
   DCM_OBJECT * obj = NULL;
   char * modality = NULL;
   char * study_uid = NULL;
   char * series_uid = NULL;
   char * instance_uid = NULL;
   char * p_id = NULL;
   char * pname = NULL;
   char * accn = NULL;

/*---------------------------------------------------------------------------------
 * Each iteration of this loop makes one pass through the study directory, 
 * processing "img_*" files.
 *-------------------------------------------------------------------------------*/
   dp = opendir(study_directory);
   while (1)
/*-------------------------------- scan study directory, looking for img_* files */
   {  rewinddir(dp);
      ok_del = 1;
      while (de=readdir(dp))
      {  if (!strcmp(de->d_name,".")  || !strcmp(de->d_name,"..") || 
             !strcasecmp(de->d_name,"AE_TITLE") ||
             !strcasecmp(de->d_name,"PROCESSING")) continue;
         if ((p=strstr(de->d_name,"img_")) != de->d_name) {ok_del = 0; continue; }
         sprintf(image_file_pfn,"%s/%s", study_directory,de->d_name);
         if (stat(image_file_pfn,&st) || !S_ISREG(st.st_mode)) {ok_del = 0; continue; }

/*----------------------- we have an image file to process; Do initial processing */
         if (first_image)
         {  first_image = 0;
/*----------------------------------------------------------------------------------
 * To avoid having two instances of the scrubber trying to process the same study
 * at the same time, we have a locking fifo in the study file.  The scrubber creates
 * and deletes this file
 *--------------------------------------------------------------------------------*/
            sprintf(em,"%s/PROCESSING", study_directory); STR(fifo_pfn,em)
            if (mkfifo(fifo_pfn, S_IRWXU | S_IRWXG) && (errno != EEXIST)) 
            {  dbge("makefifo");
               goto ps_exit;
            }
/*--------------------------------------------------------------- Open file to write
 * If this works, the process is already running, and we quit
 *--------------------------------------------------------------------------------*/
	         lock_fd = open(fifo_pfn, O_WRONLY | O_NONBLOCK);
	         if (lock_fd >= 0)
	         {	close(lock_fd);
		         sprintf(em,"%s already being scrubbed", study_directory); dbg(em);
		         goto ps_exit;
	         }
/*----------------------------------------------------------------------------------
 * Now we open the file rw.  If another process is run, it will fail first test.
 *--------------------------------------------------------------------------------*/
	         lock_fd = open(fifo_pfn, O_RDWR | O_NONBLOCK);
	         if (lock_fd < 0)
	         {	sprintf(em, "Couldn't open %s", fifo_pfn); dbg(em);
		         goto ps_exit;
	         }

/*------------------------------------------------------- pull AE title from file */
            STR(ae_title,"UNKNOWN_AE")
            sprintf(em,"%s/AE_TITLE", study_directory); STR(ae_pfn,em)
            if (ae_fp=fopen(ae_pfn,"r"))
            {   if (fgets(a,sa,ae_fp))
               {  p=strtok(a,",\n");
                  if (p) STR(ae_title,p)
                  p=strtok(NULL,",\n");
                  if (p) number_of_times_scrubbed = atoi(p);
               }
            }
/*---------------------------------------------- Update number of scrubing passes */
            number_of_times_scrubbed++;
            sprintf(a,"%s,%d\n",ae_title,number_of_times_scrubbed);
            if (ae_fp=freopen(ae_pfn,"w",ae_fp)) fputs(a,ae_fp);
            fclose(ae_fp);
      
/*-------------------------------------------- Reset values in scrubbing routines */
            if (reset) MIRCS_Reset();
            reset = 1;
         }

/*----------------------------------------------------------- Load image into memory
 * If we can't do this, and there is a transfer process running, it is most likely
 * that the file has not been completely written yet, so we just loop.  If no 
 * transfer process is running, the file is probably a bad file, and should be 
 * moved to the receipt errors directory and logged.
 *--------------------------------------------------------------------------------*/
         cond = DCM_OpenFile(image_file_pfn, DCM_ORDERLITTLEENDIAN, &obj);
         if (cond != DCM_NORMAL) 
         {  cond = DCM_OpenFile(image_file_pfn, DCM_PART10FILE, &obj);
            if (cond != DCM_NORMAL) 
            {  if (pid)
               {  COND_PopCondition(ctn_true); 
                  ok_del = 0; continue; 
               }
               else
               {  sprintf(new_directory, "%s/xfer_errors/%s", image_directory, ae_title);
                  strcpy(new_file_name, tempnam(new_directory,"img_"));
                  fileNameChecks(new_directory);
                  fileNameChecks(new_file_name);
                  UTL_VerifyCreatePath(new_directory);
                  if (rename(image_file_pfn, new_file_name)) dbge("bad file rename");
                  if (receive_error_log)
                  {  int ms = 4096;
                     char * msg = malloc(ms);
                     memset(msg,0,ms);
                     sprintf(msg,"Xmit from %s couldn't open file\n   %s\n", ae_title, new_file_name);
                     if (modality)
                     {  sprintf(em," mod   : %s\n", modality); strcat(msg,em);
                        sprintf(em," study : %s\n", study_uid); strcat(msg,em);
                     }
                     else strcat(msg," No images received");
                     logMsg(receive_error_log, msg);
                     free(msg);
                  }
               }
                        
            }
         }
/*-------------------------------------------- Calculate new directory and file name
 * simple_storage does this with a naming convention file.  The distributors always
 * use the same naming convention, and are processing using data from an open TCP/IP
 * connection, which this routine may not have access to, we we do it hard wired.
 *--------------------------------------------------------------------------------*/
         GetStringNOF(obj, DCM_IDMODALITY, &modality);
         GetStringNOF(obj, DCM_RELSTUDYINSTANCEUID, &study_uid);
         GetStringNOF(obj, DCM_RELSERIESINSTANCEUID, &series_uid);
         GetStringNOF(obj, DCM_IDSOPINSTANCEUID, &instance_uid);

         sprintf(new_directory,"%s/%s/%s/%s/%s", image_directory, ae_title, modality,
            study_uid, series_uid);
         sprintf(new_file_name,"%s/%s", new_directory, instance_uid);
      
/*-------------------------------------------------------- Call MIRCS scrubbing routines
 * These may change the new directory and new file name
 *------------------------------------------------------------------------------------*/
         MIRCS_ObjectReceived(image_file_pfn, obj, new_directory, sizeof(new_directory),
            new_file_name, sizeof(new_file_name));

/*--------------------------------- rename file to new name, deleting any precedessor */
         DCM_CloseObject(&obj);                               /* Done with DCM Object */
         fileNameChecks(new_directory);                          /* replace odd chars */
         fileNameChecks(new_file_name); 
         UTL_VerifyCreatePath(new_directory);            /* make sure dest dir exists */
      
         if (!access(new_file_name,F_OK)) if(remove(new_file_name)) dbge("remove");
         if (rename(image_file_pfn, new_file_name)) { ok_del = 0; dbge("rename"); }
         COND_PopCondition(ctn_true);
      }  /* EOL for one dir entry in study dir */
/*----------------------------------- no child, or previously terminated, we are done */
      if (!pid) break;           
/*----------------------- child (xmit) not done, pause a moment and make another pass */
      if (pid != waitpid(pid, &xmit_status, WNOHANG)) { sleep(10); continue; }
/*------------------------------------------- Xmit is done.  log abnormal exit status */
      if (receive_error_log && xmit_status)
      {  int ms = 4096;
         char * msg = malloc(ms);
         memset(msg,0,ms);
         sprintf(msg,"Xmit from %s returned status %d\n", ae_title, xmit_status);
         if (obj)
         {  GetStringNOF(obj, DCM_PATID, &p_id);
            GetStringNOF(obj, DCM_PATNAME, &pname);
            GetStringNOF(obj, DCM_IDACCESSIONNUMBER, &accn);
            sprintf(em," pname : %s\n", pname); strcat(msg,em);
            sprintf(em," pid   : %s\n", p_id); strcat(msg,em);
            sprintf(em," accn  : %s\n", accn); strcat(msg,em);
            sprintf(em," mod   : %s\n", modality); strcat(msg,em);
            sprintf(em," study : %s\n", accn); strcat(msg,em);
         }
         else strcat(msg," No images received");
         logMsg(receive_error_log, msg);
         free(msg);
      }
/*--------------------------------------------- One last pass through study directory */
      pid = 0;
   }
   closedir(dp);  
   close(lock_fd);                                          /* Close the locking fifo */
/*--------------------------------------------------------------------------------------
 * If everything worked, the AE_TITLE file and the PROCESSING fifo should be the only
 * files left in the study directory, and we can now delete them and delete the study
 * directory with them
 *------------------------------------------------------------------------------------*/
   if (ok_del)
   {  sprintf(a,"%s/AE_TITLE", study_directory); remove(a);
      sprintf(a,"%s/PROCESSING", study_directory); remove(a);
      remove(study_directory);
   }
  ps_exit:
      MIRCS_EndConnection();
      FREE(fifo_pfn)
      FREE(ae_pfn)
      FREE(ae_title)
      FREE(modality)
      FREE(study_uid)
      FREE(series_uid)
      FREE(instance_uid)
      FREE(p_id)
      FREE(pname)
      FREE(accn)
      return;
}

      
/*--------------------------------------------------------------------------------------
 * This is a wrapper for DCM_GetSTring.  value should start out as a NULL char pointer.
 * It will return with allocated memory giving either the value of the requested tag or
 * NULL.  On subsequent calls with the same char *, it will free the allocated memory
 * before making the next call.  Thus you can keep as many items as you want, but the
 * reuse of the pointer will automatically handle freeing up the memory.
 * The char pointer is also returned for ease of use.
 *------------------------------------------------------------------------------------*/
static char * GetString(DCM_OBJECT *obj, DCM_TAG tag, char ** value)
{  FREE(*value)
   *value = DCM_GetString(&obj, tag);
   if (*value) while (LCHR((*value)) == ' ') LCHR((*value)) = 0;
   return *value;
}
/*--------------- As above, but returns string "NOF" instead of NULL if tag not found */
static char * GetStringNOF(DCM_OBJECT *obj, DCM_TAG tag, char ** value)
{  if (!GetString(obj, tag, value)) *value = strdup("NOF");
   return *value;
}

static void fileNameChecks(char *name){while(*name){if(*name=='^'||*name==' ')*name='_';name++;}}
