/****************************************************************************************
 * mircs_simple_storage.c                                                       R Moulton
 * Code added for MIR Computer section functionality in simple storage.  Parameters for 
 * this functionality are maintained in MIRCS_simple_storage.ini which is located in the 
 * same directory the simple_storage binary is run from.  If this ini file is not present
 * the MIRCS functionality is turned off.  NOTE: simple_storage functions as a listener
 * which forks a copy of itself for each DICOM receiving event - thus the MIRCS_Setup
 * routine will be run once, when the receiver is started, and each new DICOM reception
 * will start out with the initial values in place after this call.
 ***************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <mysql.h>

#include "ctn_os.h"
#include "dicom.h"
#include "utility.h"
#include "dicom_uids.h"
#include "lst.h"
#include "condition.h"
#include "dulprotocol.h"
#include "dicom_objects.h"
#include "dicom_messages.h"
#include "dicom_services.h"
#include "dbquery.h"
#include "iap.h"
#include "simple_storage.h"
#include "dbg.h"
#include "ini.h"
#include "hl7.h"
#include "sql.h"
#include "hl7query.h"
#include "EQCerrors.h"
#include "log.h"
#include "map.h"

#include "mircs_simple_storage.h"


/*------------------------------ function prototypes ----------------------------------*/
static void EQCFirstObject(const char * fileName, DCM_OBJECT * obj, char * newDirectory, size_t sizeof_newDirectory);
static void EQCObject(const char * fileName, DCM_OBJECT * obj, 
      char * newDirectory, size_t sizeof_newDirectory,
      char * newName, size_t sizeof_newName);
static int is_equal(char * a, char * b);
static int make_numeric(char * a);
static int is_valid_date(char * date);
static int is_valid_uid(char * uid);
static int is_numeric(char * fld);
static int is_dnumber(char * fld);
static char * trim (char * input_string);
static char * EQCGetString(DCM_OBJECT *obj, DCM_TAG tag, char ** value);
static char * EQCGetStringNOF(DCM_OBJECT *obj, DCM_TAG tag, char ** value);
static char UnknownStationCheck(DCM_OBJECT * obj);
static void repldir(char * so, char * sn, char * str);
static void RemoveGroup2(DCM_OBJECT * obj, char log, const char * fileName);
static char * chomp(char * a);
static void RemoveExcessNameComponent(char eqc);

/*----------------------------------------------------------------------------- Macros */
#define STR(A,B) { if (A) free(A); A=strdup(B); }
#define FREE(A) if (A) { free(A); A=NULL; }
#define STR_EQ(A,B) !strcasecmp(A,B)
#define ISON(A) (!strcasecmp(A,"ON") || !strcasecmp(A,"YES"))
#define LCHR(A) A[strlen(A) -1]
#define ERR "ERR"
#define LOG "LOG"
#define CLEAR(A) memset(A,0,sizeof(A))
#define DVAL(A) trim(EQCItem[EQC.A].e.d.string)
#define HVAL(A) trim(EQCItem[EQC.A].dicom)
#define FILL(A,B) {while(strlen(A) < B) strcat(A," "); }

static char * inipfn = NULL;                      /* ini file pfn */
static FILE * debug_fp = NULL;
int MIRCS_dbg_pause = 0;
int MIRCS_delete_image = 0;

static int image_received = 0;
static char first_call = 1;

static char * imgs = NULL;                          /* Images base directory */

/*---------------------------------------- Group 2 (Part 10 format) removal processing */
static char remove_group2 = 0;
static char * group2_log_name = NULL;

/*------------------------------------------------------- Send held studies processing */
#define SHDEST sh_dest[dst]
static char ** sh_dest = NULL;
/*---------------------------------------------------------------------------------------
 * This flag is set for EQC studies which are being put on hold but which need to be
 * copied out to their original directory in addition to EQChold, because they have to 
 * be sent to one or more destinations immediatly, e.g. NORTH, SOUTH
 *-------------------------------------------------------------------------------------*/
static char sh_need_to_copy;                      
static char dst, max_dst = 0;
/*---------------------------------- string appended to study dir for send held images */
#define SH_STUDY_DIR_APPEND ".sh"

/*---------------------------------------------------------------------------------------
 * PACSI query links (0 thru 5) are used to query RIS for accn or MPI related data.  Each
 * type of query can use one or more of the links.  There is an extra slot, so for loops
 * just check for < 0 to end.
 *-------------------------------------------------------------------------------------*/
#define MAX_LNK 5
#define ACCN_LNK accn_link[lnk]
#define MPI_LNK  mpi_link[lnk]
static int accn_link[MAX_LNK +2] = {-1, -1, -1, -1, -1, -1, -1};
static int mpi_link[MAX_LNK +2]  = {-1, -1, -1, -1, -1, -1, -1};
static char lnk;

/*----------------------------------------------------- Emageon Image Scrubbing and QC */
static char EmageonQCOn = 0;                       /* Are we doing Emageon QC (at all) */
static char EmageonQCStudy = 0;                /* Are we doing Emageon QC (this study) */
static char EQCLogDir[128] = "/export/home/spectrum";
static char EQCTime[24];
static int  EQCLogPurgeHour =  2;
static int  EQCLogPurgeDays = 30;
static char EQCUsePM    = 0;
static char EQCPMUser[] = "inquirer";
static char EQCUsePACSI = 0;
static int  EQCPACSIWait = -1;
static char EQCServer = 0;
static char * EQCFacility = NULL;
static char EQCHoldStudy = 0;
static struct _hl7_mpi_query_ EQCMPIQuery;
static struct _hl7_accn_query_ EQCAccnQuery;
static char * EQCNewDirectory = NULL;
static char * EQCAETitle = NULL;
static char * EQCDBDat = NULL;
static char EQCHost[24];
static char log3 =0;
static char EQCStripExamModifiers = 0;
static char EQCValidateStudyUID = 0;        /* Validate DICOM UID's */
static char EQCValidateSeriesUID = 0;
static char EQCValidateImageUID = 0;
static char EQCStrictUIDValidation = 0;
/*-------------------------------------------------------------- What studies to scrub */
static char EQCScrubCFG = 0;       /* 0=based on .ini file, 1=based on .cfg file */  
static char EQCScrubLoose = 0;     /* If scrub on .cfg file, include loose studies */
static char EQCScrubTest = 0;      /* Scrub test xmits ("test" or "TEST" in study UID) */
static char EQCLogTest = 0;        /* Log test xmits */
static char * EQCCFGFile = NULL;   /* pfn of Sname.cfg file */
static int EQC_delete_image = 0;
static int EQC_no_idqc_hold = 0;
/*---------------------------------------- Unknown station/institution name processing */
static char UnkOn = 0;                        /* Are we doing this? */
static char * UnkLogName = NULL; 
static char UnknownStationFound = 0;
/*---------------------------------------------- Are we truncating tage over max size? */
static char EQCTruncateTags = 0;
/*--------------- Are we removing ^'s over 4 in patient name and referring physicians? */
static char EQCRemoveExcessNameComponents = 0;
/*---------------------------------------------------------------------------------------
 * Here we will load a string of comma separated facility codes which are to have held
 * studies sent to the EQCweb destinations (nominally Imageweb)
 *-------------------------------------------------------------------------------------*/
static char * EQCwebOn = NULL;
/*-------------------------- Structs used to hold information from DICOM & HL7 sources */
static struct {
   char study_date;        /* (0008,0020) */
   char image_date;        /* (0008,0023) OBR 7   */
   char study_time;        /* (0008,0030) */
   char accn;              /* (0008,0050) OBR 3.1 */
   char modality;          /* (0008,0060) */
   char institution_name;  /* (0008,0080) */
   char station_name;      /* (0008,1010) */
   char description;       /* (0008,1030) OBR 4.2 */
   char pname;             /* (0010,0010) PID 5   */
   char pid;               /* (0010,0020) PID 3.1 */
   char facility;          /* (0010,0021) PID 3.4 (in mpi query)*/
   char dob;               /* (0010,0030) PID 7   */
   char sex;               /* (0010,0040) PID 8   */
   char other_id;          /* (0010,1000) PID 4   */
   char body_part;         /* (0018,0015) */
   char study_uid;         /* (0020,000d) */
   char series_uid;        /* (0020 000e) */
   char image_uid;         /* (0008 0018) */
   char visit;             /* (0038,0010) PID 18  */
   char sched_date;        /* (0040,0002) OBR 27  */
   char sched_time;        /* (0040,0003) OBR 27  */
   char exam_code;         /* (0008,1032) -> (0008,0100) OBR 4.1 */
   char exam_coding_scheme;/* (0008,1032) -> (0008,0102) IDX constant */
   char exam_code_meaning; /* (0008,1032) -> (0008,0104) OBR 4.2 */
   char r_physician_name;  /* (0008,0090) RAD 8.2 on */
   char photometric_interp; /* (0028,0004) */
} EQC;
   static struct _EQC_elements_ {
   char * ptr;    /* pointer to list above */
   char status;   /* 0=not there,1=there,2=add it,3=modify it */
   char inhl7;    /* Is this something that should be in the RAD? */
   char * dicom;  /* Value from the DICOM header */
   char * value;  /* Value from PACSMON/RAD */
   char required; /* Is this required for Emageon? */
   DCM_TAG seqTag; /* Tag of SQ this is part of, or NULL (tags in same SQ should be grouped)*/
   char err_num;   /* EQCerrors number for this item if it is invalid */
   DCM_ELEMENT e; 
} EQCItem[] = {
   { &EQC.study_date,       0, 0, NULL, NULL, 1, NULL, EQCE_STUDY_DATE,
      DCM_IDSTUDYDATE,                   DCM_DA, "study date  ", 1, DICOM_DA_LENGTH, NULL},
   { &EQC.image_date,       0, 1, NULL, NULL, 1, NULL, EQCE_IMAGE_DATE, 
      DCM_IDIMAGEDATE,                   DCM_DA, "image date  ", 1, DICOM_DA_LENGTH, NULL},
   { &EQC.study_time,       0, 0, NULL, NULL, 1, NULL, EQCE_STUDY_TIME,
      DCM_IDSTUDYTIME,                   DCM_TM, "study time  ", 1, DICOM_TM_LENGTH, NULL},
   { &EQC.accn,             0, 1, NULL, NULL, 1, NULL, EQCE_ACCESSION_NUMBER,
      DCM_IDACCESSIONNUMBER,             DCM_SH, "accession   ", 1, DICOM_SH_LENGTH, NULL},
   { &EQC.modality,         0, 0, NULL, NULL, 1, NULL, EQCE_MODALITY,
      DCM_IDMODALITY,                    DCM_CS, "modality    ", 1, DICOM_CS_LENGTH, NULL},
   { &EQC.institution_name, 0, 0, NULL, NULL, 1, NULL, EQCE_INSTITUTION_NAME,
      DCM_IDINSTITUTIONNAME,             DCM_LO, "inst name   ", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.station_name,     0, 0, NULL, NULL, 1, NULL, EQCE_STATION_NAME,
      DCM_IDSTATIONNAME,                 DCM_SH, "station name", 1, DICOM_SH_LENGTH, NULL},
   { &EQC.description,      0, 1, NULL, NULL, 0, NULL, EQCE_STUDY_DESCRIPTION,
      DCM_IDSTUDYDESCRIPTION,            DCM_LO, "description ", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.pname,            0, 1, NULL, NULL, 1, NULL, EQCE_PATIENT_NAME,
      DCM_PATNAME,                       DCM_PN, "patient name", 1, DICOM_PN_LENGTH, NULL},
   { &EQC.pid,              0, 1, NULL, NULL, 1, NULL, EQCE_PATIENT_ID,
      DCM_PATID,                         DCM_LO, "patient id  ", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.facility,         0, 1, NULL, NULL, 1, NULL, EQCE_FACILITY,
      DCM_ISSUERPATIENTID,               DCM_LO, "facility    ", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.dob,              0, 1, NULL, NULL, 1, NULL, EQCE_PATIENT_DOB,
      DCM_PATBIRTHDATE,                  DCM_DA, "patient DOB ", 1, DICOM_DA_LENGTH, NULL},
   { &EQC.sex,              0, 1, NULL, NULL, 1, NULL, EQCE_PATIENT_SEX,
      DCM_PATSEX,                        DCM_CS, "patient sex ", 1, DICOM_CS_LENGTH, NULL}, 
   { &EQC.other_id,         0, 1, NULL, NULL, 0, NULL, EQCE_OTHER_ID,
      DCM_PATOTHERIDS,                   DCM_LO, "d number    ", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.body_part,        0, 0, NULL, NULL, 0, NULL, EQCE_BODY_PART,
      DCM_ACQBODYPARTEXAMINED,           DCM_CS, "body part   ", 1, DICOM_CS_LENGTH, NULL},
   { &EQC.study_uid,        0, 0, NULL, NULL, 1, NULL, EQCE_STUDY_UID,
      DCM_RELSTUDYINSTANCEUID,           DCM_UI, "study UID   ", 1, DICOM_UI_LENGTH, NULL},
   { &EQC.series_uid,       0, 0, NULL, NULL, 1, NULL, EQCE_SERIES_UID,
      DCM_RELSERIESINSTANCEUID,          DCM_UI, "series UID  ", 1, DICOM_UI_LENGTH, NULL},
   { &EQC.image_uid,        0, 0, NULL, NULL, 1, NULL, EQCE_IMAGE_UID,
      DCM_IDSOPINSTANCEUID,              DCM_UI, "image UID   ", 1, DICOM_UI_LENGTH, NULL},
   { &EQC.visit,            0, 1, NULL, NULL, 1, NULL, EQCE_VISIT_NUMBER,
      DCM_VISADMISSIONID,                DCM_LO, "visit number", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.sched_date,       0, 1, NULL, NULL, 1, NULL, EQCE_SCHED_DATE,
      DCM_PRCSCHEDULEDPROCSTEPSTARTDATE, DCM_DA, "sched date  ", 1, DICOM_DA_LENGTH, NULL},
   { &EQC.sched_time,       0, 1, NULL, NULL, 1, NULL, EQCE_SCHED_TIME,
      DCM_PRCSCHEDULEDPROCSTEPSTARTTIME, DCM_TM, "sched time  ", 1, DICOM_TM_LENGTH, NULL},
   { &EQC.exam_code,        0, 1, NULL, NULL, 1, DCM_IDPROCEDURECODESEQUENCE, EQCE_EXAM_CODE,
      DCM_IDCODEVALUE,                   DCM_SH, "exam code   ", 1, DICOM_SH_LENGTH, NULL},
   { &EQC.exam_coding_scheme, 0, 1, NULL, NULL, 1, DCM_IDPROCEDURECODESEQUENCE, EQCE_EXAM_SCHEME,
      DCM_IDCODINGSCHEMEDESIGNATOR,      DCM_SH, "coding sch  ", 1, DICOM_SH_LENGTH, NULL},
   { &EQC.exam_code_meaning,  0, 1, NULL, NULL, 1, DCM_IDPROCEDURECODESEQUENCE, EQCE_EXAM_DESCRIPTION,
      DCM_IDCODEMEANING,                 DCM_LO, "exam code ds", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.r_physician_name, 0, 1, NULL, NULL, 1, NULL, EQCE_REF_PHYSICIAN,
      DCM_IDREFERRINGPHYSICIAN,          DCM_PN, "re physician", 1, DICOM_PN_LENGTH, NULL},
   { &EQC.photometric_interp, 0, 0, NULL, NULL, 0, NULL, EQCE_PALETTE_COLOR,
      DCM_IMGPHOTOMETRICINTERP,          DCM_CS, "photm interp", 1, DICOM_CS_LENGTH, NULL},
};
#define EQCITEM EQCItem[i]
static char max_EQCItem = sizeof(EQCItem) / sizeof(struct _EQC_elements_);
static DCM_ELEMENT wse;

/*------------------------------------------------ simple_storage error logging/alerts */
static char * receive_error_log = NULL;
#define ALDEST al_dest[al]
static char ** al_dest = NULL;
static char al, max_al;

/*------------------------------------------------------------------- MySQL processing */
static MYSQL * conn = NULL;
static MYSQL_RES * res = NULL;
static MYSQL_ROW row;
static char sql_query[2048];

/*---------------------------------------------------------------------------- Mapping */
static PMAP pmap = NULL;

/****************************************************************************************
 * MIRCS_Setup is run at program start, and performs any needed setup functions, notably 
 * checking for the presence of the .ini file and loading parameters from it.  All of the
 * parameters above are defaulted to "off", and are only turned "on" if the .ini file
 * entries say to.
 ***************************************************************************************/
void MIRCS_Setup(char * progDirectory, char * baseDirectory)
{  char ws[1024], *p = baseDirectory;
   int i, sws = sizeof(ws);
/*-------------------------------------------------------------- Must be user spectrum */
   if (!(p=getenv("USER")) || strcmp(p,"spectrum"))
   {  dbg("Must run as user spectrum");
      exit(1);
   }         
/*---------------------------------------------------- Compute pfn of images directory */
   p = baseDirectory;
   if (!getcwd(ws,sws)) return;                                   /* Current directory */
   if (*p == '/') strcpy(ws,p);                         /* param is absolute directory */
   else if (*p == '.') strcat(ws,p+1);                /* param is . relative directory */
   else { strcat(ws,"/"); strcat(ws,p); }               /* param is other relative dir */
   STR(imgs,ws)
/*------------------------------------ Compute pfn of .ini file (in program directory) */
   strcpy(ws,progDirectory);                   /* This is arg0, program invocation */
   p = strrchr(ws,'/');                        /* Find last / char */
   if (!p) {dbg("no / in arg0"); exit(1); }
   *(p+1) = 0;                                    /* strip program name (after last /) */
   strcat(ws,"MIRCS_simple_storage.ini");         /* Add ini file name */
   STR(inipfn,ws)                          /* Load to static ws */
   if (!inipfn) return;
/*---------------------------- If .ini file doesn't exist, all MIRCS mods are left off */
   if (access(inipfn, F_OK)) return; 

   if (!GetIniString("[MAIN]", "Debug", "NULL", ws, sws, inipfn)) return;
   debug_fp = debug_on(ws);

/*------------------------------------------------------------- Logging initialization */
   if (logInitialize(inipfn)) dbg(GetLogErrorString());
/*-------------------------- Number of seconds to pause after fork to allow gdb attach */
   GetIniInt("[MAIN]", "DebugPause", 0, &MIRCS_dbg_pause, inipfn);

/*---------------------------------------- Group 2 (Part 10 format) removal processing */
   if (GetIniString("[GROUP2]","Status","OFF",ws,sws,inipfn) && ISON(ws))
   {  remove_group2 = 1;
      GetIniString("[GROUP2]","LogName","NULL",ws,sws,inipfn);
      if (strcasecmp(ws,"NULL")) STR(group2_log_name,ws)
      logNote(group2_log_name,"Group 0002 removal initialized");
   }

/*--------------------------------------------------- simple_storage error logs/alerts */
   if (GetIniString("[MAIN]","ReceiveErrorLog", NULL, ws, sws, inipfn) && strlen(ws))
      STR(receive_error_log,ws)
   
   if (GetIniString("[MAIN]","ReceiveErrorAlert", "OFF", ws, sws, inipfn) && 
         strcasecmp(ws,"OFF") && strcasecmp(ws,"NO"))
   {  char a[sws]; strcpy(a,chomp(ws));
      for (p=strtok(a,","), max_al=0; p; p=strtok(NULL,",")) if (strlen(chomp(p))) max_al++;
      if (max_al)
      {  al_dest = (char **) calloc (max_al + 1, sizeof(char *));
         if (!al_dest) dbg_exit("malloc failure", 1);
         for(p=strtok(ws,","), al=0; p; p=strtok(NULL,",")) if(strlen(chomp(p))){STR(ALDEST,p) al++;}
         al_dest[max_al] = NULL;
      }
   }
   
/*---------------------------------------------------------------------------------------
 * Emageon QC image testing and scrubbing
 *-------------------------------------------------------------------------------------*/
   if (GetIniString("[EmageonQC]","Status","OFF",ws,sws,inipfn) && ISON(ws))
   {  EmageonQCOn=1;
/*------------------------------------------------------------------- host server name */
      GetIniString("[MAIN]","Hostname","NOF",ws,sws,inipfn);
      if(strcmp(ws,"NOF")) strcpy(EQCHost,ws);
      else gethostname(EQCHost,sizeof(EQCHost));
/*---------------------------------------------------------------------- Log Directory */
      if (p=getenv("LOG")) strcpy(EQCLogDir,p); strcat(EQCLogDir,"/");
      if (GetIniString("[EmageonQC]","LogDir","\0",ws,sws,inipfn))
         if (strlen(ws)) {strcat(EQCLogDir,ws); strcat(EQCLogDir,"/");}
/*------------------------------------------------------------------------ Log purging */
      GetIniInt("[EmageonQC]","LogPurgeHour", EQCLogPurgeHour, &EQCLogPurgeHour, inipfn);
      GetIniInt("[EmageonQC]","LogPurgeDays", EQCLogPurgeDays, &EQCLogPurgeDays, inipfn);
/*----------------------------------------------------------- Use this ini for sql ini */
      sqlInitialize(inipfn);
/*------------------------------------------------ Use PACS Monitor Link to find data? */
      if (GetIniString("[EmageonQC]","PACSMonitorLink", "OFF", ws, sws, inipfn) && ISON(ws)) EQCUsePM = 1;
/*------------------------------------------------------- Use PACSI Link to find data? */
      if (GetIniString("[EmageonQC]","PACSILink", "OFF", ws, sws, inipfn) && ISON(ws)) 
      {  EQCUsePACSI = 1;
         GetIniInt("[EmageonQC]","PACSIWait", EQCPACSIWait, &EQCPACSIWait, inipfn);
      }
      if (!EQCUsePM && !EQCUsePACSI) logNote("EmageonQC", "Setup: Neither PM or PACSI selected");

/*------------------------------- Second copy of held studies to send to backup viewer */
      if (GetIniString("[EmageonQC]","EQCweb", "OFF", ws, sws, inipfn) && strcmp(ws,"OFF") && strcmp(ws,"NO")) 
         STR(EQCwebOn,ws)

/*-------------------------------------------------------------- Send holds processing */
      if (GetIniString("[EmageonQC]","SendHeld", "OFF", ws, sws, inipfn) && strcasecmp(ws,"OFF") && strcasecmp(ws,"NO"))
      {  char a[sws]; strcpy(a,chomp(ws));
         for (p=strtok(a,","), max_dst=0; p; p=strtok(NULL,",")) if (strlen(chomp(p))) max_dst++;
         if (max_dst)
         {  sh_dest = (char **) calloc (max_dst, sizeof(char *));
            if (!sh_dest) dbg_exit("malloc failure", 1);
            for(p=strtok(ws,","), dst=0; p; p=strtok(NULL,",")) if(strlen(chomp(p))){STR(SHDEST,p) dst++;}
         }
      }

/*----------------------------------------------------------- PACSI accn query link(s) */
      GetIniString("[EmageonQC]","PACSIAccnQuery", "0", ws, sws, inipfn);
      for (p = strtok(ws,", "); p; p = strtok(NULL,", "))
      {  int i;
         if (is_numeric(p))
         {  i = atoi(p);
            if (i >= 0 && i <= MAX_LNK)
            {  for (lnk = 0;; lnk++)
               {  if (ACCN_LNK < 0) ACCN_LNK = i;
                  if (ACCN_LNK == i) break;
               }
            }
         }
      }
      
      GetIniString("[EmageonQC]","PACSIMPIQuery", "5", ws, sws, inipfn);
      for (p = strtok(ws,", "); p; p = strtok(NULL,", "))
      {  int i;
         if (is_numeric(p))
         {  i = atoi(p);
            if (i >= 0 && i <= MAX_LNK)
            {  for (lnk = 0;; lnk++)
               {  if (MPI_LNK == -1) MPI_LNK = i;
                  if (MPI_LNK == i) break;
               }
            }
         }
      }
                  
               
/*--------------------------------------------------- Strip modifiers from exam codes? */ 
      if (GetIniString("[EmageonQC]","StripExamModifiers","OFF",ws,sws,inipfn) && ISON(ws)) EQCStripExamModifiers=1;

/*------------------------------------------------------ truncate tags over DICOM max? */
      if (GetIniString("[EmageonQC]","TruncateTags","OFF",ws,sws,inipfn) && ISON(ws))
         EQCTruncateTags = 1;

/*------------------------ Remove ^'s over 4 in patient name and referring physicians? */
      if (GetIniString("[EmageonQC]","RemoveExcessNameComponents","OFF",ws,sws,inipfn) && ISON(ws))
         EQCRemoveExcessNameComponents = 1;

/*---------------------------------------------------------------------- Validate UIDs */
      if (GetIniString("[EmageonQC]","ValidateStudyUID" ,"OFF",ws,sws,inipfn) && ISON(ws)) EQCValidateStudyUID=1;
      if (GetIniString("[EmageonQC]","ValidateSeriesUID","OFF",ws,sws,inipfn) && ISON(ws)) EQCValidateSeriesUID=1;
      if (GetIniString("[EmageonQC]","ValidateImageUID" ,"OFF",ws,sws,inipfn) && ISON(ws)) EQCValidateImageUID=1;
      if (GetIniString("[EmageonQC]","StrictUIDValidation" ,"OFF",ws,sws,inipfn) && ISON(ws)) EQCStrictUIDValidation=1;

/*-------------------------------------------------------------- What studies to scrub */
      GetIniString("[EmageonQC]", "Scrub", "INI", ws, sws, inipfn);
      if (!strcasecmp(ws,"CFG")) EQCScrubCFG = 1;
      if (GetIniString("[EmageonQC]","ScrubLoose","OFF",ws,sws,inipfn) && ISON(ws)) EQCScrubLoose=1;
/*-------------------------------- Locate Sname.cfg and verify that it can be accessed */
      GetIniString("[EmageonQC]", "CFGFile", "Distributor.cfg", ws, sws, inipfn);
      STR(EQCCFGFile,ws)
      if (access(EQCCFGFile, F_OK))      dbg_exit("Can't find .cfg file", 1);
      else if (access(EQCCFGFile, R_OK)) dbg_exit("Can't read .cfg file", 1);

      if (GetIniString("[EmageonQC]","ScrubTest" ,"OFF",ws,sws,inipfn) && ISON(ws)) EQCScrubTest=1;

      if (GetIniString("[LOG_EMAGEONQC]","LogTestXmits","OFF",ws,sws,inipfn) && ISON(ws)) EQCLogTest=1;
      
      for (i = 0; i < max_EQCItem; i++) *(EQCITEM.ptr) = i;  /* load cross pointers */

      if (logLevel("EmageonQC", 3)) log3 = 1;
   }

/*---------------------------------------- Unknown station/institution name processing */
      if (GetIniString("[UnknownStation]","Status" ,"OFF",ws,sws,inipfn) && ISON(ws)) UnkOn=1;
      if (UnkOn)
      {  GetIniString("[UnknownStation]","LogName", "NULL", ws, sws, inipfn);
         STR(UnkLogName,ws)
      }
      
      pmap = mapSetup("[Mapping]", inipfn);
   
   return;
}

/****************************************************************************************
 * Called by MIRCS_ObjectReceived the first time the later is called
 ***************************************************************************************/
static void FirstObjectReceived(const char * fileName, DCM_OBJECT * obj,
      char * newDirectory, size_t sizeof_newDirectory, 
      char * newName, size_t sizeof_newName)
{  dbg("FirstObjectReceived called");

   if (UnkOn) UnknownStationFound = UnknownStationCheck(obj);
   if (EmageonQCOn) EQCFirstObject(fileName, obj, newDirectory, sizeof_newDirectory);
   return;
}

/****************************************************************************************
 * Routine called by simple_storage each time an object is received, and before that
 * object is permanently written out to disc. 
 * filename = pfn of temporary file object is in.
 * obj = DCM_OBJECT structure containing loaded object data.
 * newDirectory = proposed new directory name for object file (you can chg this)
 * newName = proposed new file name (you can chg this)
 ***************************************************************************************/
void MIRCS_ObjectReceived(const char * fileName, DCM_OBJECT * obj, 
      char * newDirectory, size_t sizeof_newDirectory, 
      char * newName, size_t sizeof_newName)
{  static char * old_modality = NULL; static char * new_modality = NULL;
   static char * old_study_id = NULL; static char * new_study_id = NULL;
/*----------------------- "first call" if first time or if modality or study_id change */
   EQCGetStringNOF(obj, DCM_IDMODALITY, &new_modality);
   EQCGetStringNOF(obj, DCM_RELSTUDYINSTANCEUID, &new_study_id);

   if (!old_modality || strcmp(old_modality,new_modality))
   {  first_call = 1;
      free(old_modality);
      old_modality = strdup(new_modality);
   }

   if (!old_study_id || strcmp(old_study_id,new_study_id))
   {  first_call = 1;
      free(old_study_id);
      old_study_id = strdup(new_study_id);
   }
   
/*------------------------------------ Get AE title, assuming storage convention below */
   if (!EQCAETitle)
   {  char *p, *pfn = strdup(newName);
      sprintf(em,"/%s/%s/", old_modality, old_study_id);
      if (p=strstr(pfn,em))
      {  *p = 0;
         if (p=strrchr(pfn,'/'))
         {  p++;
            STR(EQCAETitle,p)
         }
      }
      if (!EQCAETitle) EQCAETitle = strdup("NOF");
      FREE(pfn)
   }
   mapTag(pmap, 1, 1, EQCAETitle);     /*    Pass AE Title to mapping routines */
   mapTag(pmap, 2, 1, EQCHost);        /*    Pass distributor name to mapping routines */
/*---------------------------------------------------------------------------------------
 * Process maps, then update directory and file name, in case any of the values have
 * changed.  This presumes the MIRCS standard storage.convention:
 *   A           AE Title
 *   D 0008 0060 MODALITY(time_t *) time(NULL)
 *   D 0020 000D STUDYUID
 *   D 0020 000E SERIESUID
 *   F 0008 0018 INSTANCEUID 
 *-------------------------------------------------------------------------------------*/
   if (pmap) 
   {  static char *bmo = NULL, *emo = NULL, 
           *bstuid = NULL, *estuid = NULL,
           *bseuid = NULL, *eseuid = NULL,
           *bimuid = NULL, *eimuid = NULL;
      static int changes = 0;
      static CONDITION cond;
      static char * pfn = NULL;
/*----------------------------------- store initial values of items which might change */
      EQCGetStringNOF(obj, DCM_IDMODALITY, &bmo);
      EQCGetStringNOF(obj, DCM_RELSTUDYINSTANCEUID,  &bstuid);
      EQCGetStringNOF(obj, DCM_RELSERIESINSTANCEUID, &bseuid);
      EQCGetStringNOF(obj, DCM_IDSOPINSTANCEUID,     &bimuid);
/*----------------------------------------------------------------------- process maps */
      mapProcess(pmap, obj, first_call, NULL, NULL, &changes);
      EQC_delete_image = mapGetFlag(MAP_FLAG_DELETE_IMAGE);
      EQC_no_idqc_hold = mapGetFlag(MAP_FLAG_NO_IDQC_HOLD);
      if (changes)
      {  dbg("Writing out updated image file");
         pfn = malloc(strlen(fileName) + 6);
         strcpy(pfn,fileName); strcat(pfn,".tmp");
         cond = DCM_WriteFile(&obj, DCM_ORDERLITTLEENDIAN, pfn);
         if (cond != DCM_NORMAL) COND_DumpConditions();
         if (cond == DCM_NORMAL)
         {  remove(fileName);
            rename(pfn,fileName);
         }
         free(pfn);
/*--------------------------------------------------------------------- Get new values */
         EQCGetStringNOF(obj, DCM_IDMODALITY, &emo);
         EQCGetStringNOF(obj, DCM_RELSTUDYINSTANCEUID,  &estuid);
         EQCGetStringNOF(obj, DCM_RELSERIESINSTANCEUID, &eseuid);
         EQCGetStringNOF(obj, DCM_IDSOPINSTANCEUID,     &eimuid);
/*---------------------------------------------------- Update items which have changed */
         if (strcmp(bmo,emo)) 
         {  repldir(bmo, emo, newDirectory); 
            repldir(bmo, emo, newName); 
         }
         if (strcmp(bstuid,estuid)) 
         {  repldir(bstuid, estuid, newDirectory); 
            repldir(bstuid, estuid, newName); 
         }
         if (strcmp(bseuid,eseuid)) 
         {  repldir(bseuid, eseuid, newDirectory); 
            repldir(bseuid, eseuid, newName); 
         }
         if (strcmp(bimuid,eimuid)) repldir(bimuid, eimuid, newName); 
      }
   }

/*--------------------------------------------------- remove group2; log on first_call */
   RemoveGroup2(obj, first_call, fileName);
   
   if (first_call) 
   {  first_call = 0;
      FirstObjectReceived(fileName, obj, newDirectory, sizeof_newDirectory, newName, sizeof_newName);
   } 
   
   sprintf(em,"ObjectReceived called, image %d", ++image_received); dbg(em);

   if (EmageonQCOn && EmageonQCStudy) 
      EQCObject(fileName, obj, newDirectory, sizeof_newDirectory, newName, sizeof_newName);
}

void MIRCS_EndConnection(void)
{  sprintf(em,"EndConnection called, images %d", image_received); dbg(em);
} 

/****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 * Remove Group 2 (Part 10 format) processing
 ****************************************************************************************
 ****************************************************************************************
 ***************************************************************************************/

/*---------------------------------------------------------------------------------------
 * Some modalities send files over the network with Group2 data, which is invalid.  If
 * flag is set in .ini file, and group 2 is found, remove the group.  If log is set
 * (first image) and logging is on, make a log entry.
 *-------------------------------------------------------------------------------------*/
static void RemoveGroup2(DCM_OBJECT * obj, char log, const char * fileName)
{  CONDITION cond; 
   char * msg = NULL, * pfn = NULL;
   char * in = NULL, * st = NULL, * pid = NULL, * accn = NULL, * pname = NULL, 
      * modality = NULL, * uid = NULL;
   if (!remove_group2) return;                    /* Not doing this */
   cond = DCM_RemoveGroup(&obj, 0x0002);          /* Remove the group */
   if (cond == DCM_GROUPNOTFOUND) return;         /* Wasn't there */
   if (cond != DCM_NORMAL)                        /* Error */
   {  dbg("Error trying to remove Group 2");
      COND_DumpConditions(); 
      return; 
   }
/*---------------------------------- A group 2 has been found and removed (write file) */
   dbg("Grp 2 deleted, Writing out updated image file");
   pfn = malloc(strlen(fileName) + 6);
   strcpy(pfn,fileName); strcat(pfn,".tmp");
   cond = DCM_WriteFile(&obj, DCM_ORDERLITTLEENDIAN, pfn);
   if (cond != DCM_NORMAL) COND_DumpConditions();
   if (cond == DCM_NORMAL)
   {  remove(fileName);
      rename(pfn,fileName);
   }
   free(pfn);
   
   if (!log || !group2_log_name) return;            /* Not first image, or not logging */

/*---------------------------------------------------------------------------------------
 * Hopefully, removing group 2's will occur only rarely, so we do the log purge only when
 * we are ready to make a log entry.
 *-------------------------------------------------------------------------------------*/
   logPurge(group2_log_name);
/*-------------------------------------------------------------------------- Log entry */
   if (!(msg=malloc(4096))) return;     
   EQCGetStringNOF(obj, DCM_PATNAME, &pname);
   EQCGetStringNOF(obj, DCM_PATID, &pid);
   EQCGetStringNOF(obj, DCM_IDACCESSIONNUMBER, &accn);
   EQCGetStringNOF(obj, DCM_IDMODALITY, &modality);
   EQCGetStringNOF(obj, DCM_IDINSTITUTIONNAME, &in);
   EQCGetStringNOF(obj, DCM_IDSTATIONNAME, &st);
   EQCGetStringNOF(obj, DCM_RELSTUDYINSTANCEUID, &uid);
      
   sprintf(msg, "Study received from %s- Group 0002 removed\n"
               "   pid........ %s\n"
               "   accn....... %s\n"
               "   name....... %s\n"
               "   modality... %s\n"
               "   station.... %s\n"
               "   institution %s\n"
               "   study UID.. %s",
       EQCAETitle, pid, accn, pname, modality, st, in, uid);
   if (logMsg(group2_log_name, msg)) dbg(GetLogErrorString());
   FREE(pid)
   FREE(accn)
   FREE(pname)
   FREE(modality)
   FREE(in)
   FREE(st)
   FREE(uid)
   FREE(msg)
      
   return;
}

/****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 * Unknown Station/Institution name checking
 ****************************************************************************************
 ****************************************************************************************
 ***************************************************************************************/
  
/****************************************************************************************
 * Unknown Station/Institution name checking.
 * Returns 1 if this object has and unknown station/institution name, 0 if OK.
 * This is done by trying to locate the idqc station table bucket the study would be in
 * if it was put on hold for Emageon scrubbing.  If this can be found, the study is OK;
 * otherwise it is unknown.  NOTE: This is useful, but not essential, processing.  If we
 * can't query the db for some reason, we presume the station/inst name is found
 ***************************************************************************************/
static char UnknownStationCheck(DCM_OBJECT * obj)
{  char *st=NULL, *in=NULL, *pname=NULL, *pid=NULL, *accn=NULL, *modality=NULL, *uid=NULL;
   char r = 0;
   
   if (!UnkOn) return r;                    /* Not processing unknown station/inst */

/*------------------------------------------- We do not do station check on test xmits */
   EQCGetStringNOF(obj, DCM_RELSTUDYINSTANCEUID, &uid);
   if (strstr(uid,"test") || strstr(uid,"TEST"))
   {  FREE(uid)
      return r;
   }
   
   EQCGetStringNOF(obj, DCM_IDSTATIONNAME, &st);               /* pull tag values */
   EQCGetStringNOF(obj, DCM_IDINSTITUTIONNAME, &in);
   
/*----------------------------------------------------------------------- Open idqc db */
   do
   {  sqlInitialize(inipfn);
      conn = sqlOpen("EmageonQCDB", "qcprs");
      if (!conn) { PrtSqlErrorString(conn, debug_fp, "open qc db"); break; }
/*-------------------------------------- Query for bucket for this station/institution */
      CLEAR(sql_query);
      sprintf(sql_query, "SELECT dba FROM station WHERE "
            "(station = \"%s\" AND id_by = \"station\") OR "
            "(dicom_tag = \"%s\" AND id_by = \"dicom_tag\")", st, in);
      if(mysql_query(conn, sql_query)) { dbg(sql_query); dbg((char*)mysql_error(conn)); break; }
      if (!(res = mysql_store_result(conn))) { dbg(sql_query); dbg((char*)mysql_error(conn)); break; }
      if (!mysql_num_rows(res)) r = 1;  /* no rows in result means unknown station/inst */
      mysql_free_result(res);
   } while (0);

   if (conn) mysql_close(conn);
/*----------------------------------------------------------------------- Log unknowns */
   if (r && strcasecmp(UnkLogName,"NULL"))
   {  char msg[4096];     
      EQCGetStringNOF(obj, DCM_PATNAME, &pname);
      EQCGetStringNOF(obj, DCM_PATID, &pid);
      EQCGetStringNOF(obj, DCM_IDACCESSIONNUMBER, &accn);
      EQCGetStringNOF(obj, DCM_IDMODALITY, &modality);
      
      sprintf(msg, "Study received from %s- Unknown station/institution\n"
                   "   pid........ %s\n"
                   "   accn....... %s\n"
                   "   name....... %s\n"
                   "   modality... %s\n"
                   "   station.... %s\n"
                   "   institution %s\n"
                   "   study UID.. %s",
         EQCAETitle, pid, accn, pname, modality, st, in, uid);
      if (logMsg(UnkLogName, msg)) dbg(GetLogErrorString());
      FREE(pid)
      FREE(accn)
      FREE(pname)
      FREE(modality)
      FREE(uid)
   }
   FREE(in)
   FREE(st)
   return r;
}

/****************************************************************************************
 ****************************************************************************************
 ****************************************************************************************
 * Emageon QC processing
 ****************************************************************************************
 ****************************************************************************************
 ***************************************************************************************/

/****************************************************************************************
 * Emageon QC, first image processing.
 * Processing overview:
 * 1. Determines if study is to be QC'd.  We use a station name lookup, unless the station
 *    name is blank or on the "ignore" list, otherwise we use institution name lookup.  
 *    Studies which are not being QC'd are ignored.
 * 2. Determine facility code from lookup.
 * 3. Pulls relevant items from DICOM header.
 * 4. Validates pid and accn, currently by IDXRad lookup - other validations.
 * 5. Scrubs data in valid studies and passed them on to Emageon.
 * 6. Hold invalid studies for interactive QC processing.
 * 7. Passes on held studies to EQCweb directory or partial sourcing as indicated.
 ***************************************************************************************/
static void EQCFirstObject(const char * fileName, DCM_OBJECT * obj, char * newDirectory, 
      size_t sizeof_newDirectory)
{  char * p = NULL, ws[128] = "\0", msg[4096] = "\0";
   char * in = NULL, * st = NULL, * pid = NULL, * accn = NULL, * pname = NULL, 
      * modality = NULL, * uid = NULL;
   char query_ok, in_ini_file = 1;
   int i, sws = sizeof(ws);
   unsigned long ul;
   void * ctx;
   static int mpi_accn = -1;
   static char * mpi_facility = NULL;
   time_t tim = time(NULL);
   CONDITION cond;
   dbg("EQCFirstObject called");
   sh_need_to_copy = 0;
   EmageonQCStudy = 1;
/*---------------------------------------------------------------------------------------
 * test transmits (study UID contains "test") are sent periodically to monitor system
 * functionality.  The EQCScrubTest .ini flag controls whether these are scrubbed, and
 * the EQCLogTest flag controls whether they are logged.  If they are scrubbed, they are
 * always logged; otherwise a brief log entry is made, as the studies are of interest
 * only to see that they came through.
 *-------------------------------------------------------------------------------------*/
   EQCGetStringNOF(obj, DCM_RELSTUDYINSTANCEUID, &uid);
   if (strstr(uid,"test") || strstr(uid,"TEST"))
   {  if (!EQCScrubTest) 
      {  EmageonQCStudy = 0;
         if (EQCLogTest)
         {  EQCGetStringNOF(obj, DCM_IDMODALITY, &modality);
            sprintf(msg, "Xmit test received from %s\n"
                         "   modality... %s\n"
                         "   study UID.. %s", 
               EQCAETitle, modality, uid);
               if (logMsgLevel("EmageonQC", msg, 1)) dbg(GetLogErrorString());
            FREE(modality)
         }
         FREE(uid)
         return;
      }
   }
/*---------------------------------------------------------------------------------------
 * If EQCScrubCFG is not set, .ini file entries are used to determine what to scrub.  We
 * use the station name as a key, unless it is blank or on the ignore list, in which case
 * we use the institution name.  In any case, we do the look up, because the facility
 * code is in the .ini file - this will save us having to do an query the RIS for it.
 *-------------------------------------------------------------------------------------*/
   FREE(EQCFacility)
   strcpy(ws,"NOF");
   EQCGetString(obj, DCM_IDSTATIONNAME, &p);
   if (p && !strlen(p)) FREE(p)
   if (p && GetIniString("[EmageonQCIgnoreStation]",p,"NOF",ws,sws,inipfn) && strcmp(ws,"NOF")) p=NULL;
   if (!p) EQCGetString(obj, DCM_IDINSTITUTIONNAME, &p);
   if (p) { sprintf(em,"inst: %s", p); dbg(em); }
   if (!p || !GetIniString("[EmageonQCModalities]",p,"NOF",ws,sws,inipfn) || STR_EQ(ws,"NOF")) in_ini_file = 0;
   FREE(p)
   if (strlen(ws) && strcmp(ws,"NOF")) STR(EQCFacility,ws)
/*---------------------------------------------------------------------------------------
 * Now we are looking at 'real' studies.  There are two basic approaches to deciding
 * which studies to scrub.  If EQCScrubCFG is set, we determine what to scrub based on
 * the main routing file, Server.cfg, scrubbing studies which have an entry in the file.
 * If the additional EQCScrubLoose flag is set, we scrub loose studies also, meaning all
 * the real studies coming into the system are scrubbed.
 *-------------------------------------------------------------------------------------*/
   if (EQCScrubCFG)
   {  if (!EQCScrubLoose)
      {  int ret;
         EQCGetStringNOF(obj, DCM_IDMODALITY, &modality);
         sprintf(em,"grep '^%s,%s,' %s > /dev/null", EQCAETitle, modality, EQCCFGFile); 
         if(system(em)) EmageonQCStudy = 0;
      }
   }
/*---------------------------------------------------------------------------------------
 * If EQCScrubCFG is not set, .ini file entries are used to determine what to scrub. 
 *-------------------------------------------------------------------------------------*/
   else EmageonQCStudy = in_ini_file;
/*------------------------------------- Log entry for study which will not be scrubbed */
   if (!EmageonQCStudy)
   {  EQCGetStringNOF(obj, DCM_PATNAME, &pname);
      EQCGetStringNOF(obj, DCM_PATID, &pid);
      EQCGetStringNOF(obj, DCM_IDACCESSIONNUMBER, &accn);
      EQCGetStringNOF(obj, DCM_IDMODALITY, &modality);
      EQCGetStringNOF(obj, DCM_IDINSTITUTIONNAME, &in);
      EQCGetStringNOF(obj, DCM_IDSTATIONNAME, &st);
            
      sprintf(msg, "Study received from %s - not scrubbed\n"
                  "   pid........ %s\n"
                  "   accn....... %s\n"
                  "   name....... %s\n"
                  "   modality... %s\n"
                  "   station.... %s\n"
                  "   institution %s\n"
                  "   study UID.. %s",
          EQCAETitle, pid, accn, pname, modality, st, in, uid);
      if (UnknownStationFound) strcat(msg,"\n   *** Unknown Station/Institution Name ***");
/*-------------------------------------------------------------------------------- Log */
      if (logMsgLevel("EmageonQC", msg, 1)) dbg(GetLogErrorString());
      FREE(pid)
      FREE(accn)
      FREE(pname)
      FREE(modality)
      FREE(in)
      FREE(st)
      FREE(uid)
      
      return; 
   }
   dbg("study to scrub");
/*-------------------------------------------------------- Pull DICOM tags from header */
   dbg("parsing 1st DICOM image");
   for (i = 0; i < max_EQCItem; i++)
   {  STR(EQCITEM.dicom,"<NULL>")
      ctx = NULL;
/*----------------------------------------------------------- Clear previous item data */
      if (!EQCITEM.e.d.string) 
      {  EQCITEM.e.d.string = malloc (EQCITEM.e.length + 1);
         if (!EQCITEM.e.d.string) dbge("malloc");
      }    
      memset(EQCITEM.e.d.string,0,EQCITEM.e.length + 1);
      EQCITEM.status = 0;
      FREE(EQCITEM.value)
      if (EQCITEM.seqTag)      
/*-------------------------------------------------------- If item is part of sequence */
      {  LST_HEAD * seqList = NULL;
         DCM_SEQUENCE_ITEM * item = NULL;
/*--------------------------------------------- Pull the list out of the sequence item */
         cond = DCM_GetSequenceList(&obj, EQCITEM.seqTag, &seqList);
         if (cond == DCM_NORMAL)
/*-------------- Pull object item out of list (this works only for non-repeating sq's) */
         {  item = (DCM_SEQUENCE_ITEM*) LST_Head(&seqList);
            if (item && item->object)
/*--------------------------------------------------------------- Pull item out of object
 * NOTE: status of DCM_GETINCOMPLETE means that tag value is longer than the max length
 * permitted by the DICOM standard.  
 *-------------------------------------------------------------------------------------*/
            {  cond = DCM_GetElementValue(&(item->object), &EQCITEM.e, &ul, &ctx);
               if (cond == DCM_NORMAL || cond == DCM_GETINCOMPLETE)
               {  EQCITEM.status = 1;
                  STR(EQCITEM.dicom,EQCITEM.e.d.string)
               }
            }
         }
      }
      else
/*--------------------------------------------------------------------- top level tags */
      {  cond = DCM_GetElementValue(&obj, &EQCITEM.e, &ul, &ctx);
         if (cond == DCM_NORMAL || cond == DCM_GETINCOMPLETE) 
         {  EQCITEM.status = 1;
            STR(EQCITEM.dicom,EQCITEM.e.d.string)
         }
      }                                                                                                   
   }

/*---------------------------------------------------------- Facility, if in .ini file */
   if (EQCFacility)
   {  STR(EQCItem[EQC.facility].value,EQCFacility)
      strcpy(EQCItem[EQC.facility].e.d.string,EQCFacility);
      EQCItem[EQC.facility].status += 2;
   }

/*------------------------------------------------------------------ PACSI Query items */
   while (EQCUsePACSI)
   {  dbg("PACS Query");
      hl7SelectHost(NULL, NULL);
/*------------------------------------------ If no accn number, can't look anything up */
      if (!EQCItem[EQC.accn].status || !strlen(EQCItem[EQC.accn].e.d.string)) break;
/*-------------------------------------- If accn has invalid characters, can't look up */
      if (!is_numeric(EQCItem[EQC.accn].e.d.string)) break;
/*---------------------------------------------------------------------------------------
 * Accession number query: In some cases, multiple studies for the same accession number
 * may be combined in a single transmit.  If the desired accession number has already be
 * queried, we do not repeat the query, as it is time consuming.
 *-------------------------------------------------------------------------------------*/
      query_ok = 1;
      if (!EQCAccnQuery.accn || EQCAccnQuery.accn != atoi(EQCItem[EQC.accn].e.d.string))
      {  EQCAccnQuery.accn = atoi(EQCItem[EQC.accn].e.d.string);
/*--------------------------------------------------------------- Use PACSI query feed */
         for (lnk = 0; ACCN_LNK >= 0; lnk++)
         {  if (!hl7QSelectServer(ACCN_LNK, EQCPACSIWait, 1)) continue;
            sprintf(em, "PACSI Accn Query, link %d", ACCN_LNK); dbg(em);  
            if (hl7AccnQuery(&EQCAccnQuery)) 
            {  logNote("EmageonQC", GetHl7QErrorString()); 
               continue; 
            }
            break;
         }
         if (ACCN_LNK < 0) query_ok = 0;
      }
/*------------------------------------------------ If query was successful; load items */ 
      if (query_ok) 
      {  char * pp = NULL;
/*------------------------------------------------- strip modifiers from the exam code */
         if (EQCStripExamModifiers && (pp=strchr(EQCAccnQuery.exam,'/'))) *pp = 0; 
         sprintf(em,"%07d", EQCAccnQuery.accn); STR(EQCItem[EQC.accn].value,em)
         sprintf(em,"%07d", EQCAccnQuery.mpi);  STR(EQCItem[EQC.pid].value,em)
         STR(EQCItem[EQC.other_id].value,EQCAccnQuery.dnumber)
         STR(EQCItem[EQC.visit].value,EQCAccnQuery.visit)
         STR(EQCItem[EQC.pname].value,EQCAccnQuery.pname)
         STR(EQCItem[EQC.dob].value,EQCAccnQuery.dob)
         STR(EQCItem[EQC.sex].value,EQCAccnQuery.sex)
         STR(EQCItem[EQC.exam_code].value,EQCAccnQuery.exam)
         if (strlen(EQCItem[EQC.exam_code].value)) STR(EQCItem[EQC.exam_coding_scheme].value,"99IDX")
         STR(EQCItem[EQC.exam_code_meaning].value,EQCAccnQuery.exam_desc)
         STR(EQCItem[EQC.description].value,EQCAccnQuery.exam_desc)
         STR(EQCItem[EQC.r_physician_name].value,EQCAccnQuery.r_physician)
/*----------------------------------------- Split sched date & time (one field in hl7) */
         if (strlen(EQCAccnQuery.start_time) >= 8)
         {  memset(em,0,sizeof(em)); 
            memcpy(em,EQCAccnQuery.start_time,8);
            STR(EQCItem[EQC.sched_date].value,em);
            if (strlen(EQCAccnQuery.start_time) >= 12)
            {  memset(em,0,sizeof(em)); 
               memcpy(em,EQCAccnQuery.start_time+8,4);
               STR(EQCItem[EQC.sched_time].value,em);
            }
         }
      }
      
/*--------------------------- We will query for facility if it wasn't in the .ini file */
      if (EQCFacility) break;
/*------------------------------------------------------------ Have we already got it? */
      if (mpi_accn == EQCAccnQuery.accn && mpi_facility)
      {  STR(EQCItem[EQC.facility].value,mpi_facility)
         break;
      }
/*-------------------------------------------------------------- Need MPI; use Query 5 */
      if (is_numeric(EQCItem[EQC.pid].e.d.string)) EQCMPIQuery.mpi = atoi(EQCItem[EQC.pid].e.d.string);
      else break;
      for (lnk = 0; MPI_LNK >= 0; lnk++)
      {  struct _hl7_mpi_query_exam_ * x;
         int accn = atoi(EQCItem[EQC.accn].e.d.string);
         if (!hl7QSelectServer(MPI_LNK, EQCPACSIWait, 1)) continue;  
         sprintf(em, "PACSI MPI Query, link %d", MPI_LNK); dbg(em);
         if (hl7MPIQuery(&EQCMPIQuery))
         {  logNote("EmageonQC", GetHl7QErrorString());
            continue;
         }
/*------------------------------------------- Query successful, scan for facility code */
         for (i = 0, x = EQCMPIQuery.exam; i < EQCMPIQuery.num_exams; i++, x++)
         {  if (accn == x->accn)
            {  mpi_accn = accn;
               STR(EQCItem[EQC.facility].value,x->facility)
               STR(mpi_facility,x->facility)
               break;
            }
         }
         break;
      }
      break;
   }

/****************************************************************************************
 * Now we match the data from the RAD with the DICOM data.  
 ****************************************************************************************/
/*---------------------------------------------------------------------------------------
 * If the DICOM pid is a dnumber, we have the mpi from the RAD, and the dnumbers match,
 * we replace the dnumber with the RAD mpi number.
 *-------------------------------------------------------------------------------------*/
   if (EQCItem[EQC.pid].status == 1 && EQCItem[EQC.pid].value && 
       is_dnumber(EQCItem[EQC.other_id].value) &&
       STR_EQ(EQCItem[EQC.pid].e.d.string,EQCItem[EQC.other_id].value))
   {  strcpy(EQCItem[EQC.pid].e.d.string,EQCItem[EQC.pid].value);
      EQCItem[EQC.pid].status += 2;
   }
/*---------------------------------------------------------------------------------------
 * Otherwise, the pid's and accn's must be present and match, or we hold the study.
 *-------------------------------------------------------------------------------------*/
   if (EQCItem[EQC.pid].status == 0 || EQCItem[EQC.accn].status == 0 ||
       !EQCItem[EQC.pid].value || !EQCItem[EQC.accn].value ||
       !is_equal(EQCItem[EQC.pid].e.d.string,EQCItem[EQC.pid].value) ||
       !is_equal(EQCItem[EQC.accn].e.d.string,EQCItem[EQC.accn].value)) 
   {  dbg(EQCGetErrorString(EQCE_ID_ACCN_MISMATCH, "holding for - ", 1, 1)); 
      EQCHoldStudy = 1;
   }

/*---------------------------------------------------------------------------------------
 * If the status of the accn is cancelled, we hold the study.
 *-------------------------------------------------------------------------------------*/
   if (EQCAccnQuery.status && !strcasecmp(EQCAccnQuery.status,"X"))
   {  dbg(EQCGetErrorString(EQCE_ACCN_CANCELED, "holding for - ", 1, 1));
      EQCHoldStudy = 1;
   }

/*---------------------------------------------------------------------------------------
 * Now, with the exception of the pid and accn, any other data from the RAD is presumed
 * to be more accurate that corresponding DICOM data and is updated if different, 
 * provided that the study is not on hold.
 *-------------------------------------------------------------------------------------*/
   if (!EQCHoldStudy)
   {  for (i = 0; i < max_EQCItem; i++)
      {  if (i == EQC.pid || i == EQC.accn) continue;
         if (EQCITEM.value && strlen(EQCITEM.value) && !is_equal(EQCITEM.value,EQCITEM.e.d.string))
         {  strcpy(EQCITEM.e.d.string,EQCITEM.value);
            if (EQCITEM.e.representation == DCM_LO || EQCITEM.e.representation == DCM_SH)
               EQCITEM.e.length = strlen(EQCITEM.e.d.string);
            EQCITEM.status += 2;
         }
      }
   }
/*--------------------------------- If the gender is not present, we add it as unknown */
   if (!EQCItem[EQC.sex].status)
   {  strcpy(EQCItem[EQC.sex].e.d.string,"U");
      EQCItem[EQC.sex].status += 2;
   }

/*----------------------------------------------- We make date and time fields numeric */
   EQCItem[EQC.study_date].status += make_numeric(EQCItem[EQC.study_date].e.d.string);
   EQCItem[EQC.image_date].status += make_numeric(EQCItem[EQC.image_date].e.d.string);
   EQCItem[EQC.study_time].status += make_numeric(EQCItem[EQC.study_time].e.d.string);
   EQCItem[EQC.dob].status        += make_numeric(EQCItem[EQC.dob].e.d.string);
   EQCItem[EQC.sched_date].status += make_numeric(EQCItem[EQC.sched_date].e.d.string);
   EQCItem[EQC.sched_time].status += make_numeric(EQCItem[EQC.sched_time].e.d.string);

/*-------------------------------- Check for valid birth dates, which must be YYYYMMDD */
   if (EQCItem[EQC.dob].status && !is_valid_date(EQCItem[EQC.dob].e.d.string)) 
   {  sprintf(em,"%s: %s", EQCGetErrorString(EQCE_PATIENT_DOB, "holding for - ", 1, 1), 
         EQCItem[EQC.dob].e.d.string); dbg(em);
      EQCHoldStudy = 1;
   }

/*--------------------------------------------------------------------- Validate UID's */
   if (EQCValidateStudyUID && 
       EQCItem[EQC.study_uid].status && 
       !is_valid_uid(EQCItem[EQC.study_uid].e.d.string))
   {  sprintf(em,"%s: %s", EQCGetErrorString(EQCE_STUDY_UID, "holding for - ", 1, 1),
         EQCItem[EQC.study_uid].e.d.string); dbg(em);
      EQCHoldStudy = 1;
   }
   
   if (EQCValidateSeriesUID && 
       EQCItem[EQC.series_uid].status && 
       !is_valid_uid(EQCItem[EQC.series_uid].e.d.string))
   {  sprintf(em,"%s: %s", EQCGetErrorString(EQCE_SERIES_UID, "holding for - ", 1, 1),
         EQCItem[EQC.series_uid].e.d.string); dbg(em);
      EQCHoldStudy = 1;
   }

   if (EQCValidateImageUID && 
       EQCItem[EQC.image_uid].status && 
       !is_valid_uid(EQCItem[EQC.image_uid].e.d.string))
   {  sprintf(em,"%s: %s", EQCGetErrorString(EQCE_IMAGE_UID, "holding for - ", 1, 1),
         EQCItem[EQC.image_uid].e.d.string); dbg(em);
      EQCHoldStudy = 1;
   }

/*------------------------- Remove ^'s over 4 in patient name and referring physicians */
   if (EQCRemoveExcessNameComponents)
   {  RemoveExcessNameComponent(EQC.pname);
      RemoveExcessNameComponent(EQC.r_physician_name);
   }

/*--------- Emageon does not support photometric interpretation value of PALETTE COLOR */
   if (EQCItem[EQC.photometric_interp].status && 
       strstr(EQCItem[EQC.photometric_interp].e.d.string,"PALETTE COLOR"))
   {  sprintf(em, "%s: %s", EQCGetErrorString(EQCE_PALETTE_COLOR, "holding for - ", 1, 1), 
         EQCItem[EQC.photometric_interp].e.d.string); dbg(em);
      EQCHoldStudy = 1;
   }
   
/*************************** Insert other validations here *****************************/

/*------------------- Now, if any required elements are not present, we hold the study */
   for (i=0; i < max_EQCItem; i++) if (EQCITEM.required && !EQCITEM.status) 
   {  sprintf(em, "%s: %s", EQCGetErrorString(EQCITEM.err_num, "holding for - ", 1, 1), 
         EQCITEM.e.d.string); dbg(em);
      EQCHoldStudy = 1;
   }

/*--------------------------- If the RAD query failed, we add this to the error string */
   if (hl7q_query_failed) dbg(EQCGetErrorString(EQCE_QUERY_FAILED, "holding for - ", 1, 1));
   
/*---------------------------------------------------------------------------------------
 * Logging for studies which are scrubbed and perhaps, held.  Level 1 contains only id
 * info.  Level 2 prints all values taken from DICOM header and query
 *-------------------------------------------------------------------------------------*/
   sprintf(msg, "Study received from %s - scrubbed", EQCAETitle);
   if (EQCHoldStudy)
   {  strcat(msg,", errors found");
      strcat(msg, EQC_no_idqc_hold ? ", no IDQC hold" : ", study on hold");
   }
   strcat(msg,"\n");
   if (UnknownStationFound) strcat(msg,"   *** Unknown Station/Institution Name ***\n");
   if (!logLevel("EmageonQC", 2))
   {  strcat(msg,"    data item         DICOM header value                   RIS value\n");
      strcat(msg,"   ----------- -------------------------------- --------------------------------\n");
      sprintf(em,"   pid........ %s ", HVAL(pid));              FILL(em,48) strcat(msg,em);
      if (EQCItem[EQC.pid].value) strcat(msg,EQCItem[EQC.pid].value); 
      strcat(msg,"\n");
      sprintf(em,"   accn....... %s ", HVAL(accn));             FILL(em,48) strcat(msg,em);
      if (EQCItem[EQC.accn].value) strcat(msg,EQCItem[EQC.accn].value); 
      strcat(msg,"\n");
      sprintf(em,"   name....... %s ", HVAL(pname));            FILL(em,48) strcat(msg,em);
      if (EQCItem[EQC.pname].value) strcat(msg,EQCItem[EQC.pname].value); 
      strcat(msg,"\n");
      sprintf(em,"   modality... %s ", HVAL(modality));         FILL(em,48) strcat(msg,em);
      if (EQCItem[EQC.modality].value) strcat(msg,EQCItem[EQC.modality].value); 
      strcat(msg,"\n");
      sprintf(em,"   station.... %s ", HVAL(station_name));     FILL(em,48) strcat(msg,em);
      if (EQCItem[EQC.station_name].value) strcat(msg,EQCItem[EQC.station_name].value); 
      strcat(msg,"\n");
      sprintf(em,"   institution %s ", HVAL(institution_name)); FILL(em,48) strcat(msg,em);
      if (EQCItem[EQC.institution_name].value) strcat(msg,EQCItem[EQC.institution_name].value); 
      strcat(msg,"\n");
      sprintf(em,"   study UID.. %s ", HVAL(study_uid));        FILL(em,48) strcat(msg,em);
      if (EQCItem[EQC.study_uid].value) strcat(msg,EQCItem[EQC.study_uid].value); 
      strcat(msg,"\n");
      strcat(msg,EQCGetErrorsMsg(NULL, 0));
      logMsgLevel("EmageonQC", msg, 1);                    /* Main Emageon Log */
      if (EQCHoldStudy) logMsgLevel("EQChold", msg, 1);    /* qc (held only) Log */
   }
   else
   {  strcat(msg,"    data item          DICOM header value                   RIS value\n");
      strcat(msg,"   ------------ -------------------------------- --------------------------------\n");
      for (i=0; i < max_EQCItem; i++)
      {  if (EQCITEM.status)
         {  sprintf(em,"   %s ", EQCITEM.e.description); strcat(msg,em);
            sprintf(em,"%s ", EQCITEM.dicom); FILL(em,33) strcat(msg,em);
            if (EQCITEM.value) strcat(msg,EQCITEM.value); 
            strcat(msg,"\n");
         }
      }
      strcat(msg,EQCGetErrorsMsg(NULL,0));
      logMsgLevel("EmageonQC", msg, 2);                    /* Main Emageon Log */
      if (EQCHoldStudy) logMsgLevel("EQChold", msg, 2);    /* qc (held only) Log */
   }
   
/*---------------------------------------------------------------------------------------
 * If study is being held, and before we alter the directory it is going to, we check for
 * those destinations to which the study should be sent immediately, if any.  If we find
 * any, we make sure the directory exists, and create the dest.DEST.sendto files for those
 * destinations; Once we have done this, the Router won't source it, so it will go only
 * to the destinations we enter. We also set a flag (sh_need_to_copy) so that all image
 * files will be copied to this original directory in addition to the EQChold directory.
 *-------------------------------------------------------------------------------------*/
   if (EQCHoldStudy && sh_dest)
   {  char a[sizeof_newDirectory], b[sizeof_newDirectory], 
         *ae_title=NULL, *mod=NULL;
      int rc;
/*------------------------------- Pull Source and modality out of image directory path */
      strcpy(a,newDirectory);
      p = a;
      if (*p == '.') p += 2; else p += (strlen(imgs) + 1);
      if (p=strtok(p,"/"))    STR(ae_title,p)
      if (p=strtok(NULL,"/")) STR(mod,p)
      if (ae_title && mod)
/*------------------------- Pass destinations looking for sourcing to that destination */
      {  for (dst = 0; dst < max_dst; dst++)
         {  sprintf(a,"grep \"^%s,%s,\" %s | grep \",%s\"", ae_title, mod,  EQCCFGFile, SHDEST);
            if (!system(a)) /* 0 return means we found a match & need to send */
            {  strcpy(a,newDirectory);
/*------------------------------------------- cut off series directory to get to study */
               *(strrchr(a,'/')) = 0;  *(strrchr(a,'/')) = 0;  
/*----------------------------------------------------------- add send held dir append */
               strcat(a,SH_STUDY_DIR_APPEND);
               cond = UTL_VerifyCreatePath(a);
               sprintf(b,"touch %s/dest.%s.sendto", a, SHDEST);
               system(b);
               sh_need_to_copy = 1; /* We will need to copy image files */
            }
         }
      }
      FREE(ae_title)
      FREE(mod)
   }

/*---------------------------------------------------------------------------------------
 * If study is being held, alter the directory it is going to
 *-------------------------------------------------------------------------------------*/
   if (EQCHoldStudy)
   {  EQCNewDirectory = malloc(sizeof_newDirectory);
      memset(EQCNewDirectory,0,sizeof_newDirectory);
      strcpy(EQCNewDirectory,imgs);
      strcat(EQCNewDirectory,"/EQChold/");
      if (*newDirectory == '.') strcat(EQCNewDirectory,newDirectory +2);
      else strcat(EQCNewDirectory,newDirectory + strlen(imgs) +1);
   }

/*---------------------------------------------------------------------------------------
 * If study is being held, create the .dat file which will be used to update the study
 * table, and will be put in the study directory for use in rebuilding the study table.
 *-------------------------------------------------------------------------------------*/
   if (EQCHoldStudy && !EQC_no_idqc_hold)
   {  char a[2048], b[2048], *p;
      char EQCDBDat[2048];
      char time_now[24];
      char source[24];
      int i, j;
      time_t now = time(NULL);
/*------------------------------------------------------ Create the .dat file contents */
      cftime(time_now,"%Y%m%d%H%M%S", &now);
      strcpy(a, EQCNewDirectory);                       /* Pull source from new directory */
      p = strstr(a,"/EQChold/") + strlen("/EQChold/");
      strcpy(source,strtok(p,"/"));
/*------------------------------------------- Build tab delimited record of values for db */
      strcpy(EQCDBDat,"NULL\t0\t");          /* dba and QC status (0) */
         
      strcat(EQCDBDat,EQCHost);                strcat(EQCDBDat,"\t"); 
      strcat(EQCDBDat,DVAL(study_uid));        strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,source);                 strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,DVAL(modality));         strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,DVAL(station_name));     strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,DVAL(institution_name)); strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,DVAL(pid));              strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,DVAL(other_id));         strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,DVAL(pname));            strcat(EQCDBDat,"\t");
      
      if (is_valid_date(DVAL(dob))) strcat(EQCDBDat,DVAL(dob)); strcat(EQCDBDat,"\t");
      
      strcat(EQCDBDat,DVAL(sex));         strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,DVAL(visit));       strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,DVAL(accn));        strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,DVAL(description)); strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,DVAL(exam_code));   strcat(EQCDBDat,"\t");
      strcat(EQCDBDat,DVAL(body_part));   strcat(EQCDBDat,"\t");

/*--------------------------------------------------- Schedule date and time are combined */
      if (is_valid_date(DVAL(sched_date)))   /* if date is valid, put that in */
      {  strcat(EQCDBDat,DVAL(sched_date));
         strcpy(b,DVAL(sched_time));         /* pull time, truncate at decimal */
         if (p = strchr(b,'.')) *p = 0;
         while (strlen(b) < 6) strcat(b,"0"); /* tack on zeroes until 6 digits */
         strcat(EQCDBDat,b);                  /* add it */
      }
      strcat(EQCDBDat,"\t");
/*------------------------------------------------------ Study date and time are combined */
      if (is_valid_date(DVAL(study_date)))   /* if date is valid, put that in */
      {  strcat(EQCDBDat,DVAL(study_date));
         strcpy(b,DVAL(study_time));         /* pull time, truncate at decimal */
         if (p = strchr(b,'.')) *p = 0;
         while (strlen(b) < 6) strcat(b,"0"); /* tack on zeroes until 6 digits */
         strcat(EQCDBDat,b);                  /* add it */
      }
      strcat(EQCDBDat,"\t");

/*----------------------------------------------------------------- Report done (unknown) */
      strcat(EQCDBDat,"0\t");
      
/*----------------------------------------------------------- First arrive (which is now) */
      strcat(EQCDBDat,time_now); strcat(EQCDBDat,"\t");

/*----------------------------------------------------------------- Errors, in hex format */
      sprintf(em,"%u\t", EQC_error); strcat(EQCDBDat, em);
      
/*--------------------- QC done time (null), QC'ing user id, and resend dest (all blanks) */
      strcat(EQCDBDat,"0\t\t\t");
      
/*----------------------------------------------------------------------------------- EOR */
      strcat(EQCDBDat,"\n");
/*---------------------------------------------------- make sure the new directory exists */
      cond = UTL_VerifyCreatePath(EQCNewDirectory);
/*------------------------------------------------------------------------------------------
 * Create pfn for .dat file in study directory.  We take the newdirectory and scan past 
 * EQChold three levels (for source, modality, and study UID), then truncate it and add the
 * QCStudy.dat file name 
 *----------------------------------------------------------------------------------------*/
      strcpy(a, EQCNewDirectory);
      for (i=0, p=strstr(a,"/EQChold/") +9; i < 3; p = strchr(p,'/') +1, i++) ;
      *p = 0;
      strcat(a,"QCStudy.dat");
/*--------------------------------------------- Open the file output write the data to it */
      do
      {  i = open (a, O_WRONLY | O_CREAT, 0644);
         if (i < 0) { dbge(".dat file open"); break; }
         j = write (i, EQCDBDat, strlen(EQCDBDat));
         if (j < 0) dbge(".dat file write");
         close(i);
      } while (0);
/*------------------------------------------------------- Now open connection to qc db */ 
      do
      {  char need_to_load = 1;  /* flag: will we need to load the .dat file record? */
         sqlInitialize(inipfn);
         conn = sqlOpen("EmageonQCDB", "qcprs");
         if (!conn) { PrtSqlErrorString(conn, debug_fp, "open qc db"); break; }
/*-------------------------------- Pull any records for this source, modality, and UID */
         CLEAR(sql_query);
         sprintf(sql_query, "SELECT dba, qc_status, server FROM study WHERE source = \"%s\" AND "
            "modality = \"%s\" AND study_id = \"%s\"", source, DVAL(modality), DVAL(study_uid));
         if(mysql_query(conn, sql_query)) { dbg(sql_query); dbg((char*)mysql_error(conn)); break; }
         if (!(res = mysql_store_result(conn))) { dbg(sql_query); dbg((char*)mysql_error(conn)); break; }
/*--------------------------------------------------------- Look at each row in result */
         for (i = 0; i < mysql_num_rows(res); i++)         
         {  row = mysql_fetch_row(res);
            if (atoi(row[1])) break;    /* If study has been qc'd, no db updates are needed */
            if (STR_EQ(row[2],EQCHost)) {
               sprintf(em,"dba=%s, server=%s", row[0], row[2]); dbg(em);
               need_to_load = 0; /* for for server exists, won't need to load */
            }
/*---------------------------------------------------------- Update demographic values */
            CLEAR(sql_query);
            sprintf(sql_query,"UPDATE study SET pid = \"%s\", dnumber = \"%s\", pname = \"%s\", "
               "dob = \"%s\", sex = \"%s\" WHERE dba = \"%s\"", DVAL(pid), DVAL(other_id), 
               DVAL(pname), DVAL(dob), DVAL(sex), row[0]);
            if(mysql_query(conn, sql_query)) { dbg(sql_query); dbg((char*)mysql_error(conn)); break; }
         }
         mysql_free_result(res);
/*------------------------------------------------------------------ Add row if needed */
         if (need_to_load)
         {  sprintf(sql_query,"LOAD DATA LOCAL INFILE \"%s\" REPLACE INTO TABLE study"
               " FIELDS TERMINATED BY '\t' LINES TERMINATED BY '\n'", a);
            if(mysql_query(conn, sql_query)) { dbg(sql_query); dbg((char*)mysql_error(conn)); break; }

         }
      } while (0);
      mysql_close(conn);
   }

   return;
}

static void EQCObject(const char * fileName, DCM_OBJECT * obj, 
      char * newDirectory, size_t sizeof_newDirectory,
      char * newName, size_t sizeof_newName)
{  int i, count, flag = 0;
   CONDITION cond;
   char * pfn = NULL, msg[4096];
   DCM_OBJECT * ob, * seqObj = NULL;
   DCM_TAG seqObj_tag = 0;
   LST_HEAD * seqList;
   DCM_SEQUENCE_ITEM * seqItem;
   void * ctx;
   unsigned long ul;
   
/*-------------------------------------------------------------- start for log message */
   sprintf(msg,"Processing image %d\n", image_received);
   
/*------------------------------------------------------ Add or update items as needed */
   for (i = 0; i < max_EQCItem; i++)
   {  if ((EQCITEM.status < 2) && !EQCTruncateTags) continue;
/*---------------------------------------------------------------------------------------
 * For items which are part of a sequence, make the sequence if needed, and pull the 
 * DICOM object pointer from the list.  Note that the table groups objects in the same
 * SQ tag, we can just use the one we have if it is the same as the last item's.
 *-------------------------------------------------------------------------------------*/
      if (EQCITEM.seqTag && (!seqObj || (EQCITEM.seqTag != seqObj_tag)))
      {  cond = DCM_GetSequenceList(&obj, EQCITEM.seqTag, &seqList); 
         if (cond != DCM_NORMAL)
         {  DCM_ELEMENT e = {EQCITEM.seqTag, DCM_SQ, "", 1, 0, 0};
            e.d.sq = seqList = LST_Create();
            seqItem = (DCM_SEQUENCE_ITEM *) malloc(sizeof(*seqItem));
            DCM_CreateObject(&seqItem->object, 0);
            LST_Enqueue(&seqList, seqItem);
            cond = DCM_AddSequenceElement(&obj, &e);
            if (cond != DCM_NORMAL)
            {  sprintf(em,"Failed to add SQ for %s", EQCITEM.e.description); dbg(em);
               COND_DumpConditions();
               logNote("EmageonQC",em);
               continue;
            }
         }
         seqItem = (DCM_SEQUENCE_ITEM*) LST_Head(&seqList);
         seqObj = seqItem->object;
         seqObj_tag = EQCITEM.seqTag;
      }
      if (EQCITEM.seqTag) ob = seqObj; else ob = obj;
      
/*-------------------------------------------- Items which need to be added or changed */
      if (EQCITEM.status >= 2)
      {  sprintf(em,"   Modifying %s: v=%s\n", EQCITEM.e.description, EQCITEM.e.d.string); strcat(msg,em);
         flag = 1;
         cond = DCM_RemoveElement(&ob, EQCITEM.e.tag);
         cond = DCM_AddElement(&ob, &(EQCITEM.e));
         if (cond != DCM_NORMAL) 
         {  sprintf(em,"Failed to fix %s: v=%s", EQCITEM.e.description, EQCITEM.e.d.string);
            COND_DumpConditions();
            logNote("EmageonQC",em);
         }
      }
/*------------------------------------------------ Truncating Items which are too long */
      else if (EQCTruncateTags)
      {  ctx = NULL;
         if (DCM_GetElementValue(&ob, &EQCITEM.e, &ul, &ctx) == DCM_GETINCOMPLETE)
         {  flag = 1;
            cond = DCM_RemoveElement(&ob, EQCITEM.e.tag);
            cond = DCM_AddElement(&ob, &(EQCITEM.e));
            if (cond != DCM_NORMAL) 
            {  sprintf(em,"Failed to truncate %s: v=%s", EQCITEM.e.description, EQCITEM.e.d.string);
               COND_DumpConditions();
               logNote("EmageonQC",em);
            }
         } 
      }
   }
/*------------------------------------------------- If we have updated, write back out */
   if (flag)
   {  dbg("Writing out updated image file");
      pfn = malloc(strlen(fileName) + 6);
      strcpy(pfn,fileName); strcat(pfn,".tmp");
      cond = DCM_WriteFile(&obj, DCM_ORDERLITTLEENDIAN, pfn);
      if (cond != DCM_NORMAL) COND_DumpConditions();
      if (cond == DCM_NORMAL)
      {  remove(fileName);
         rename(pfn,fileName);
      }
      free(pfn);
   }

   
/*---------------------------------------------------------------------------------------
 * If study is on hold but needs to go out to specific dest(s) immediately, copy it to 
 * the 'send held' directory - NOT done for map deleted images.
 *-------------------------------------------------------------------------------------*/
      if (sh_need_to_copy && !EQC_delete_image)
      {  char new[sizeof_newDirectory];
         char old_uid[strlen(EQCItem[EQC.study_uid].e.d.string) + 5];
         char new_uid[strlen(EQCItem[EQC.study_uid].e.d.string) + strlen(SH_STUDY_DIR_APPEND) + 5];
/*----------------------------- original study uid (is part of directory and file name */
         strcpy(old_uid,EQCItem[EQC.study_uid].e.d.string); 
         chomp(old_uid);
/*------------------------------------------------- new study 'uid', with append chars */
         strcpy(new_uid,old_uid);
         strcat(new_uid,SH_STUDY_DIR_APPEND);
/*----------------------------------- modify directory to add 'send held' append chars */
         strcpy(new,newDirectory);
         repldir(old_uid, new_uid, new);
         
         cond = UTL_VerifyCreatePath(new);
/*----------------------------------- modify file name to add 'send held' append chars */
         strcpy(new,newName);
         repldir(old_uid, new_uid, new);
         
         cond = DCM_WriteFile(&obj, DCM_ORDERLITTLEENDIAN, new);
         if (cond != DCM_NORMAL) COND_DumpConditions();
      }
   
/*---------------------------------------- If study is hold, change directory and name */
   if (EQCHoldStudy) 
   {  char nd[sizeof_newDirectory];
      strcat(msg,"   Holding image\n");
      memset(nd,0,sizeof_newDirectory);
      strcpy(nd,imgs); strcat(nd,"/EQChold/");
      if (*newDirectory == '.') strcat(nd,newDirectory +2);
      else strcat(nd,newDirectory + strlen(imgs) +1);
      strcpy(newDirectory,nd);
      
      pfn = malloc(sizeof_newName);
      strcpy(pfn,imgs); strcat(pfn,"/EQChold/");
      if (*newName == '.') strcat(pfn,newName +2);
      else strcat(pfn,newName + strlen(imgs) +1);
      memset(newName,0,sizeof_newName);
      strcpy(newName,pfn);
      free(pfn);
/*---------------------------------------------------------------------------------------
 * We may also make a second copy of the image file to the EQCweb directory.  This is 
 * used to send a copy of the un-qc'd study to a backup viewer, so that it can be seen 
 * before the qc process is completed.
 *-------------------------------------------------------------------------------------*/
      if (EQCwebOn && EQCItem[EQC.facility].value && 
          strstr(EQCwebOn,EQCItem[EQC.facility].value) && !EQC_delete_image)
      {  char wd[sizeof_newDirectory], *p;
         memset(wd,0,sizeof_newDirectory);
         strcpy(wd,imgs); strcat(wd,"/EQCweb/");
         p = strstr(newDirectory,"EQChold/") + strlen("EQChold/");
         p = strchr(p,'/') + 1;
         strcat(wd,p);
         cond = UTL_VerifyCreatePath(wd);

         pfn = malloc(sizeof_newName);
         strcpy(pfn,imgs); strcat(pfn,"/EQCweb/");
         p = strstr(newName,"EQChold/") + strlen("EQChold/");
         p = strchr(p,'/') + 1;
         strcat(pfn,p);
         cond = DCM_WriteFile(&obj, DCM_ORDERLITTLEENDIAN, pfn);
         if (cond != DCM_NORMAL) COND_DumpConditions();
         free(pfn);
      }
   }
   if (log3) logMsg("EmageonQC", msg);
/*---------------------------------------------------------------------------------------
 * Inform calling program that file is not to be written if (1) a mapping indicated that
 * it is to be deleted, or (2) the study is on hold and a mapping indicated that no iqqc
 * hold is to be done.
 *-------------------------------------------------------------------------------------*/
   if (EQC_delete_image || (EQCHoldStudy && EQC_no_idqc_hold)) MIRCS_delete_image = 1;
   else MIRCS_delete_image = 0;
   
   return;
}

/*---------------------------------------------------------------------------------------
 * Is this SOP a storage class SOP? Return 1=Yes, 0=No
 *-------------------------------------------------------------------------------------*/
int MIRCS_is_storage_SOP(const char * s) {
   char a[1024], sop[256];
   int sa = sizeof(a);
   strcpy(sop,s);
   GetIniString("[StorageSOPClasses]", sop, "NOF", a, sa, inipfn);
   if (STR_EQ(a,"NOF")) return 0;
   else return 1;
}

/*---------------------------------------------------------------------------------------
 * Should this SOP class be ignored (i.e. rejected 'silently', without a log entry or an
 * alert?  Returns: 1=yes,0=no
 *-------------------------------------------------------------------------------------*/
int MIRCS_is_ignore_SOP(const char * s) {
   char a[1024], sop[256];
   int sa = sizeof(a);
   strcpy(sop,s);
   GetIniString("[IgnoreSOPClasses]", sop, "NOF", a, sa, inipfn);
   if (STR_EQ(a,"NOF")) return 0;
   else return 1;
}

/*---------------------------------------------------------------------------------------
 * Log and generate alerts for errors from simple_storage
 *-------------------------------------------------------------------------------------*/
void MIRCS_log_error(char * errm, char * AETitle, char alert) {
   char msg[4096], a[4096];
/*----------------------------------------- rolling log (with demograpics if available */
   if (receive_error_log) {
      sprintf(msg, "DICOM receiver error= %s", errm);
      if (AETitle) {
         sprintf(a,"\n   Study received from %s", AETitle); strcat(msg, a);
      }
      if (!first_call) {
         sprintf(a, "\n"
               "   pid........ %s\n"
               "   accn....... %s\n"
               "   name....... %s\n"
               "   modality... %s\n"
               "   station.... %s\n"
               "   institution %s\n"
               "   study UID.. %s",
               EQCItem[EQC.pid].dicom, EQCItem[EQC.accn].dicom, EQCItem[EQC.pname].dicom, 
               EQCItem[EQC.modality].dicom, EQCItem[EQC.station_name].dicom, 
               EQCItem[EQC.institution_name].dicom, EQCItem[EQC.study_uid].dicom);
         strcat(msg,a);
      }
      logMsg(receive_error_log,msg);
   }
/*---------------------------------------------------- e-mail and pager alerts */
   if (alert != MIRCS_NO_ALERT && al_dest) {
      char * null[1] = {NULL};
      sprintf(msg, " on %s:\n%s", EQCHost, errm);
      if (alertSend(al_dest, null, null, "DICOM rec err", NULL, msg, NULL) && 
            receive_error_log) {
         logNote(receive_error_log, (char*)GetAlertErrorString()); 
      }
   }
   /*---------------------------------------------------------------------------
    * One alert only is used for unknown SOP classes, which are then added to
    * the ignore section of the .ini file, so that if they come up again, they
    * are treated as ignore SOP classes. The SOP class should be in the error
    * message, 2 bytes past the only colon.
    *-------------------------------------------------------------------------*/
   if (alert == MIRCS_SEND_ALERT_ONCE_ONLY && strchr(errm,':')) {
      PutIniString("[IgnoreSOPClasses]", strchr(errm,':') + 2, 
            AETitle ? AETitle : " ", inipfn);
   }
}
      

   
/*----------------------------- string compare which ignores trailing spaces (like COBOL)
 * returns 1 if they are equal, 0 if not equal
 *-------------------------------------------------------------------------------------*/
static int is_equal(char * a, char * b)
{  char * at = NULL, * bt = NULL;
   int ret = 1;
   if (a) STR(at,a) else STR(at," ") while (LCHR(at) == ' ') LCHR(at) = 0;
   if (b) STR(bt,b) else STR(bt," ") while (LCHR(bt) == ' ') LCHR(bt) = 0;
   if (strcmp(at,bt)) ret = 0; 
   FREE(at)
   FREE(bt)
   return (ret);
}

/*---------------------------------------------------------------------------------------
 * Returns pointer to string without trailing spaces.  Does not alter original. Uses 
 * twenty char pointers in rotation, reallocating them.
 *-------------------------------------------------------------------------------------*/
static char * trim(char * a)
{  static char * b[20];
   static char i = -1;
   if (i == -1) for (i=0; i < 20; i++) b[i] = NULL;
   if (i >=20) i = 0;
   free(b[i]); 
   b[i] = malloc(strlen(a) + 1);
   strcpy(b[i],a);
   while (LCHR(b[i]) == ' ') LCHR(b[i]) = 0;
   return b[i++];
} 
   
/*---------------------------------------------------------------------------------------
 * Removes all characters from string except numeric digits and first decimal point.
 * Returns 2 if string was modified, 0 otherwise.
 *-------------------------------------------------------------------------------------*/
static int make_numeric(char * string)
{  int r = 0, dflag = 0;
   size_t s = strlen(string) + 1;
   char *p, *q, *ws = malloc(s);
   if (!ws) return (r);
   memset (ws,0,s);
   for (p = ws, q = string; *q; q++)
   {  if (isdigit(*q)) *(p++) = *q;
      else if (!dflag && (*q == '.')) {dflag = 1; *(p++) = *q;}
      else r = 2;
   }
   if (r) strcpy(string,ws);
   return (r);
}
   
/*---------------------------------------------------------------------------------------
 * Returns 1 if the argument string represents a valid Gregorian date in YYYYMMDD format,
 * stripping off any trailing characters after the date.  Returns 0 if the string does
 * not represent a valid Gregorian date, making no changes to the string.
 *-------------------------------------------------------------------------------------*/
static int is_valid_date(char * date)
{  int month_length[] = {0,31,28,31,30,31,30,31,31,30,31,30,31};
   int year = 0, month = 0, day = 0, i, j;
/*------------------------------------------------------------- Here we parse the date */
   for (i = 0; i < 8; i++) 
   {  if (!isdigit(*(date + i))) return (0);
      j = *(date + i) - '0';
      if (i < 4) year = year * 10 + j;
      else if (i < 6) month = month * 10 + j;
      else day = day * 10 + j;
   }
/*------------------------------------------------------ Is year involved a leap year? */
   if (!(year%4) && ((year%100) || !(year%400))) month_length[2] = 29;
/*---------------------------------------------------------------------- Validate date */
   if (month < 1 || month > 12) return (0);
   if (day < 1 || day > month_length[month]) return (0);
/*-------------------------------------------- strip off any characters after the date */
   *(date + 8) = 0;
   return (1);
}

/*--------------------------------------------------------------------------------------
 * Returns 1 if the argument string represents a valid DICOM UID, i.e., that it consists
 * of numeric parts without leading zeroes, separated by periods.  Internal numeric 
 * parts may be zero, and if so are represented by a single ddd.0.xxx. Returns 0 if the 
 * string does not represent a valid DICOM UID, making no changes to the string.
 * if EQCStrictUIDValidation is not set, we will accept 0 at the beginning of any part 
 * except part 1
 *------------------------------------------------------------------------------------*/
static int is_valid_uid(char * p)
{  char state = 1;             /* states 1=bo 1st,2=bo other,3=in portion,4=eo portion*/
   while (1)

   {  switch (state)
/*---------------------------------------------------------------------------------------
 * At bo portion other than first; If EQCStrictUIDValidation 0 OK, but then must be at eo 
 * portion. If not EQCStrictUIDValidation, 0 OK.
 *-------------------------------------------------------------------------------------*/
      {  case 2:              
            if (*p == '0') 
            {  if (EQCStrictUIDValidation) state=4;
               else state = 3; 
               break; 
            }
/*-------------------------------------------- bo 1st or non 0 portion; must have 1-9 */
         case 1:                     
            if ((*p < '1') || (*p > '9')) return (0);    
            state = 3;
            break;
/*-------------------------- within portion; eos OK, . means bo portion, or any digit */
         case 3:                       /* within portion */
            if ((*p == NULL) || (*p == ' ')) return (1);  
            if (*p == '.') { state = 2; break; } 
            if ((*p < '0') || (*p > '9')) return (0);     
            break;
/*---------------------------------------------------- must be at eo portion (or eos) */
         case 4:
            if ((*p == NULL) || (*p == ' ')) return (1);  
            if (*p == '.') { state = 2; break; }
            return (0);  
         default:                          /* Internal error */
            dbg ("internal error in is_valid_uid()");
            return (0);
      }
      p++;
   }
}
/*--------------------------------------------------------------------------------------
 * This is a wrapper for DCM_GetSTring.  value should start out as a NULL char pointer.
 * It will return with allocated memory giving either the value of the requested tag or
 * NULL.  On subsequent calls with the same char *, it will free the allocated memory
 * before making the next call.  Thus you can keep as many items as you want, but the
 * reuse of the pointer will automatically handle freeing up the memory.
 * The char pointer is also returned for ease of use.
 *------------------------------------------------------------------------------------*/
static char * EQCGetString(DCM_OBJECT *obj, DCM_TAG tag, char ** value)
{  FREE(*value)
   *value = DCM_GetString(&obj, tag);
   if (*value) while (LCHR((*value)) == ' ') LCHR((*value)) = 0;
   return *value;
}
/*--------------- As above, but returns string "NOF" instead of NULL if tag not found */
static char * EQCGetStringNOF(DCM_OBJECT *obj, DCM_TAG tag, char ** value)
{  if (!EQCGetString(obj, tag, value)) *value = strdup("NOF");
   return *value;
}

/*---------------------------------------------------------------------------------------
 * Replace first occurrence (if any) of substring1 with substring2 in a string which
 * represents a directory path or pfn.  s1 represents one directory in that path, and s2
 * a replacement for that directory name.  Thus, s1 should appear in the string as part
 * of the sequence "/s1/".  This is true even for the last directory in the path, as it
 * will be represented with a trailing "/"
 *-------------------------------------------------------------------------------------*/
static void repldir(char * s1, char * s2, char * str)
{  char * p;
   char a[strlen(s1) + 3];
   int i;
/*----------------------------------------------- Skip situations we do not deal with */
   if (!s1 || !s2 || !str || !strlen(s1) || !strlen(s2) || !strlen(str)) return;
   if (STR_EQ(s1,"NOF") || STR_EQ(s2,"NOF") || STR_EQ(str,"NOF")) return;
/*--- Is s1 present as a directory name? if so, point to first character of s1 in str */
   strcpy(a,"/"); strcat(a,s1); strcat(a,"/");
   if (!(p = strstr(str, a))) return;
   p++;
/*-------------------------- Handle shift, in case s1 and s2 are not the same length */
   i = strlen(s2) - strlen(s1);
   if (i != 0) memmove(p + strlen(s1) + i, p + strlen(s1), strlen(p) - strlen(s1) + 1);
/*----------------------------------------------------------- put s2 into the string */
   memmove(p,s2,strlen(s2));
   return;
}

/*---------------------- Remove leading and trailing spaces and new lines from string */
static char * chomp(char * a)
{  int i;
   if (!a) return a;
   i = strspn(a," ");
   memmove(a,a+i,strlen(a+i)+1);
   while (LCHR(a) == ' ' || LCHR(a) == '\n') LCHR(a) = '\0';
   return a;
}

/*-------------------------------------------- remove any excess ^'s from name string */
static void RemoveExcessNameComponent(char c)
{  char cnt = 0;
   char * p;
   if (EQCItem[c].status == 0) return;
   p = EQCItem[c].e.d.string;
   while (*p)
   {  if (*p != '^') { p++; continue; }
      cnt++;
      if (cnt < 5) { p++; continue; }
      memmove(p,p+1,strlen(p));
      continue;
   }
   if ((cnt > 4) && EQCItem[c].status == 1) EQCItem[c].status = 3;
   return;
}

/*------------------- test pid & accn, return 1 if would be valid number, 0 otherwise */
static int is_numeric(char * fld)
{  int flag = 0;
   char * p;
   if (!fld) return (0);
   for (p = fld; *p; p++) 
   {  if (isdigit(*p)) { flag = 1; continue; }
      if (!isspace(*p)) return (0);
   }
   return (flag);
}

/*---------------------------------------------------------------------------------------
 * test for field being a valid dnumber.  For our purposes, that means that the field
 * exists, starts with an upper or lower case 'd', has at least one numeric digit after
 * that, and nothing else except spaces and digits after that.
 * returns 1 if valid, 0 otherwise.
 *-------------------------------------------------------------------------------------*/
static int is_dnumber(char * fld)
{  int flag = 0;
   char *p;
   if (!fld || (*fld != 'D' && *fld != 'd')) return (0);
   for (p = fld+1; *p; p++) 
   {  if (isdigit(*p)) { flag = 1; continue; }
      if (!isspace(*p)) return (0);
   }
   return (flag);
}
