/******************************************************************************
 * EQCerrors.c                                                        R Moulton
 * Emageon QC error processing.  Allows user to set and clear EQC errors, pack
 * them for storage in a db, and generate Messages for individual errors or all
 * the errors on a particular study.
 *-----------------------------------------------------------------------------
 * 5/19/5 R Moulton original write
 *****************************************************************************/

#include <stdlib.h>
#include <string.h>

#include "EQCerrors.h"

#define CLEAR(A) memset(A,0,sizeof(A))
#define LCHR(A) A[strlen(A) -1]

/*----------------------------- Error / Message struct ----------------------*/
static struct _eqc_error_{
   char qc;            /* Include in qc display? */
   char repairable;    /* Can this error be repaired in IDQC? */
   char * dicom_tag;  /* DICOM tag, if any */
   char * msg;        /* Error message */
} eqc_errors[] = {
   { 1, 0, "(0008,0020) ", "invalid study date" },
   { 1, 1, "(0008,0023) ", "invalid content (image) date" },
   { 1, 0, "(0008,0030) ", "invalid study time" },
   { 1, 1, "(0008,0050) ", "invalid accession number" },
   { 1, 0, "(0008,0060) ", "invalid modality" },
   { 1, 0, "(0008,0080) ", "invalid institution name" },
   { 1, 0, "(0008,1010) ", "invalid station name" },
   { 1, 1, "(0008,1030) ", "invalid study description" },
   { 1, 1, "(0010,0010) ", "invalid patient name" },
   { 1, 1, "(0010,0020) ", "invalid patient ID" },
   { 1, 1, "(0010,0021) ", "invalid issuer patient id (facility)" },
   { 1, 1, "(0010,0030) ", "invalid patient DOB" },
   { 1, 1, "(0010,0040) ", "invalid patient sex" },
   { 1, 1, "(0010,1000) ", "invalid other patient id (Dnumber)" },
   { 1, 1, "(0018,0015) ", "invalid body part examined" },
   { 1, 0, "(0020,000D) ", "invalid study UID" },
   { 1, 0, "(0020,000E) ", "invalid series UID" },
   { 1, 0, "(0008,0018) ", "invalid image UID" },
   { 0, 1, "(0038,0010) ", "invalid admission ID (visit #)" },
   { 0, 1, "(0040,0002) ", "invalid procedure scheduled start date" },
   { 0, 1, "(0040,0003) ", "invalid procedure scheduled start time" },
   { 0, 1, "(0008,1032)->(0008,0100) ", "invalid exam code" },
   { 0, 1, "(0008,1032)->(0008,0102) ", "invalid exam code scheme designator" },
   { 0, 1, "(0008,1032)->(0008,0104) ", "invalid exam code description" },
   { 1, 1, NULL, "patient ID - Accession mismatch" },
   { 1, 1, NULL, "Accession is for cancelled exam" },
   { 1, 1, "(0008,0090) ", "missing referring physician name"},
   { 1, 1, NULL, "RAD query failed"},
   { 1, 0, "(0028,0004) ", "PALETTE COLOR not supported"},
};
static char max_eqce = sizeof(eqc_errors) / sizeof(struct _eqc_error_);
static char eqce;
#define EQCE eqc_errors[eqce]

static char msg[4096];

void EQCClearErrors(void) { EQC_error = 0; }

void EQCSetError(char eqce) 
{  if (eqce < 0 || eqce >= max_eqce) return;
   EQC_error |= 1UL << eqce; 
}
   
int EQCIsErrorSet(char eqce)
{  if (EQC_error & (1UL << eqce)) return 1;
   else return 0;
}

unsigned long EQCGetErrorBit(char eqce) { return (1UL << eqce); }

char * EQCGetErrorString(char eqce, char * prefix, char tag, char set)
{  CLEAR(msg);
   if (eqce < 0 || eqce >= max_eqce) 
   {  sprintf(msg,"no such EQC error number - %d", eqce);
      return msg;
   }

   if (set) EQCSetError(eqce);
   
   if (prefix && strlen(prefix)) 
   {  strcpy(msg,prefix); 
      if (LCHR(msg) != ' ') strcat(msg," "); 
   }
   if (EQCE.dicom_tag && tag) strcat(msg,EQCE.dicom_tag);
   strcat(msg, EQCE.msg);
   return msg;
}
   
char * EQCGetErrorsMsg(char * prefix_line, char tags)
{  CLEAR(msg);
   if (prefix_line && strlen(prefix_line))
   {  strcpy(msg,prefix_line);
      if (LCHR(msg) != '\n') strcat(msg,"\n");
   }
   for (eqce = 0; eqce < max_eqce; eqce++)
   {  if (EQCIsErrorSet(eqce))
      {  if (EQCE.dicom_tag && tags) strcat(msg,EQCE.dicom_tag);
         strcat(msg,EQCE.msg); strcat(msg,"\n");
      }
   }
   return msg;
}

char * EQCGetQCErrorsMsg(char f)
{  CLEAR(msg);
   for (eqce = 0; eqce < max_eqce; eqce++)
   {  if (EQCE.qc && EQCIsErrorSet(eqce)) {strcat(msg,EQCE.msg); strcat(msg,"\n");}
   }
   if (!EQCQCRepairable()) strcat(msg,"*** Can't be fixed in IDQC ***\n");
   return msg;
}

int EQCQCRepairable(void)
{  for (eqce = 0; eqce < max_eqce; eqce++) 
      if (!EQCE.repairable && EQCIsErrorSet(eqce)) return 0;
   return 1;
}
