/******************************************************************************
 * hl7.c                                                              R Moulton
 * Library routines which allow simple manipulations of HL7 messages.
 *-----------------------------------------------------------------------------
 * 9-3 R Moulton Initial write.
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "hl7.h"

/*-------------------------------- Macros --------------------------------------*/
#define MAKE_STR(A,B) if (A=malloc(strlen(B)+3)) strcpy(A,B);
#define FREE(A) if (A) { free(A); A=NULL; }
#define STR_EQ(A,B) !strcmp(A,B)
#define SEG(A) m->seg[A]
#define CURRENT m->current_segment
#define LASTCHAR(A) A[strlen(A) -1]
#define ADDCHAR(A,B) A[strlen(A)] = B

/*************************** Function Prototypes ********************************/
static int hl7PutIn(PMSG, char * value, int field, char separator);

unsigned int hl7_errno;

/*-------------------------------- Error msg strings. Match to ini_errno values */
static char * e_msg[] = 
{	"No error",
	"No message read",
	"Invalid message size",
	"Invalid text message",
	"Missing segid field",
	"At end of segments",
	"No output buffer",
	"No current segment",
	"No FILE defined",
	"Invalid print format",
	"Invalid parameter",
	"Stream I/O error",
	"Segment limit reached",
	"Invalid value",
	"Invalid ACK code"
};
static int max_e_msg = sizeof(e_msg) / sizeof(char *);
static char * no_msg = "Unknown hl7_errno value";

static int max_segment_size = 1024;
static int max_msg_size = 40960;

static char v[1025] = "\0";           /* Static return for alpha Get values */

#define TRUE   1
#define FALSE  0

#define MAX_LINE 1024
#define ST '\r'        /* Segment terminator <CR>. Can't be altered */

#define MSG_STRUCT_SIZE sizeof(struct _hl7_message_)
static char def_SB = 11;
static char def_EB = 28;
static char def_FS = '|';
static char def_CS = '^';
static char def_SS = '&';
static char def_RS = '~';
static char def_ES = '\\';

/******************************************************************************
 * Frees memory in use to store message on pointer, if any.  Sets pointer to
 * NULL.  This routine always returns OK.
 *****************************************************************************/
int hl7FreeMemory(PMSG m)
{	int i;
	hl7_errno = HL7_OK;
	if (!m) { hl7_errno = HL7_NOMSG; return (HL7_R_OK); }
	for (i = 0; i < MAX_SEGMENTS; i++) FREE(SEG(i))
	return(HL7_R_OK);
}

/******************************************************************************
 * Set up empy message structure (clear any existing message stuff)
 *****************************************************************************/
int hl7NewMsg(PMSG m)
{	if (hl7ReadMsg(m, NULL, 0) == HL7_R_FAIL) return (HL7_R_FAIL);
	hl7_errno = HL7_OK;
	return (HL7_R_OK);
}
/******************************************************************************
 * Load message from char into struct.  Clear struct first if needed. 
 * Messages may be wrapped in the SB and EB characters, may be the exact size
 * passed, or may be null terminated.
 *****************************************************************************/
int hl7ReadMsg(PMSG m, char * msg, int msg_size)
{	char segment[max_segment_size];
	int sn, sc, mc, i;
/*------------------------------------------------------ Validate parameters */
	if (msg && (msg_size <= 0 || msg_size > max_msg_size)) 
	{	hl7_errno = HL7_INVMSGSIZE; 
		return (HL7_R_FAIL); 
	}
/*----------------------------- Clear, allocate memory for message structure */
	hl7FreeMemory(m);
/*------------------------------------------------------ Load default values */
	CURRENT = 0;
	m->SB = def_SB; m->EB = def_EB; m->FS = def_FS;
	m->CS = def_CS; m->SS = def_SS; m->RS = def_RS; m->ES = def_ES;
/*-------------------------------------- No message, creates empty structure */
	if (!msg) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_OK); 
	}
	
	sn = 0;   /* Segment number we are currently building */
	sc = 0;   /* Next char position in segment buffer */
	mc = 0;   /* Next char position in msg buffer */
	memset(segment,0,sizeof(segment));  /* Initialize current segment */

/****************************************** Pass msg buffer, building struct */
	while (mc < msg_size && sn < MAX_SEGMENTS && sc < max_segment_size && msg[mc])
/*--------------------------------------- SB at beginning of message skipped */
	{	if (mc == 0 && msg[mc] == m->SB) { mc++; continue; }
/*-----------------------------------------------------------------------------
 * If beginning of first segment is MSH, then look at next few characters as
 * the delimiters, and alter message defaults if needed.  Note that the escape
 * and subcomponent separators may be omitted if they are not used, in which
 * case we set them to null and have to watch for that later.  The way we 
 * know they are not used is that we see the field separator.
 *---------------------------------------------------------------------------*/
		if (!sn && !memcmp(segment, "MSH", 3) && sc < 8)
		{	segment[sc] = msg[mc];
			switch (sc)
			{	case 3:
					m->FS = segment[sc];
					break;
				case 4:
					m->CS = segment[sc];
					break;
				case 5:
					m->RS = segment[sc];
					break;
				case 6:
					if (segment[sc] == m->FS) m->ES = '\0';
					else m->ES = segment[sc];
					break;
				case 7:
					if (segment[sc] == m->FS && segment[sc - 1] == m->FS) m->SS = '\0';
					else m->SS = segment[sc];
					break;
				default:
					break;
			}
			sc++; mc++;
			continue;
		}
/*------------------------- Escaped characters copied without interpretation */
		if (m->ES && m->ES == msg[mc])
		{	segment[sc++] = msg[mc++];
			segment[sc++] = msg[mc++];
			continue;
		}
/*------------------------------------------------------- segment terminator */
		if (msg[mc] == ST)
		{	segment[sc++] = msg[mc++];                 /* Move the ST character */
			sc++;                    /* Bump count to include a NULL terminator */
			SEG(sn) = malloc(sc);     /* Add the segment to the struct array */
			memcpy(SEG(sn),segment,sc);
			memset(segment,0,sizeof(segment));              /* Clear the buffer */
			sn++; sc = 0;            /* Counters: next segment, first character */
			continue;
		}
/*-------------------------------------------- EB not copied, but we're done */
		if (msg[mc] == m->EB) break;
/*-------------------------------------------------- Ordinary character copy */
		segment[sc++] = msg[mc++];
		continue;
	}
/*--------------------------------------------------------- Check for errors */
	if (sn == 0 || sc > 0)
	{	hl7FreeMemory(m);
		hl7_errno = HL7_INVMSG;
		return (HL7_R_OK);
	}
	hl7_errno = HL7_OK;
	return(HL7_R_OK);
}

/******************************************************************************
 * Write HL7 message from structure to txt.  msg buffer must be big enough to
 * hold entire message as NULL terminated string.  If start_end, add the start
 * and end bytes to the message.
 *****************************************************************************/
int hl7WriteMsg(PMSG m, char * msg, int msg_size, int start_end)
{	int sn, i, a;
	hl7_errno = HL7_OK;
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*-------------------------------------------- Lines? Will it fit in buffer? */
	sn = 0;
	a = 1; if (start_end) a += 2;
	for (i = 0; i < MAX_SEGMENTS; i++) 
		if (SEG(i)) 
		{	sn++;
			a += strlen(SEG(i));
		}
	if (sn == 0) {	hl7_errno = HL7_INVMSG; return (HL7_R_FAIL); }
	if (a < 10 || a > msg_size) { hl7_errno = HL7_INVMSGSIZE; return (HL7_R_FAIL); }
	
/*------------------------------------------------------------ Load message */
	memset(msg,0,msg_size);
	if (start_end) msg[0] = m->SB;
	for (i = 0; i < MAX_SEGMENTS; i++) if (SEG(i)) strcat(msg,SEG(i));
	if (start_end) msg[strlen(msg)] = m->EB;
	return (HL7_R_OK);
}

/******************************************************************************
 * Position to segment
 *****************************************************************************/
int hl7FirstSeg(PMSG m, char * segid)
{	unsigned int new;
	char a[max_segment_size];
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*----------------------------------- Is there a place to return segment id? */
	if (!segid)
	{	hl7_errno = HL7_NOSEGID;
		return (HL7_R_FAIL);
	}

	new = 0;
	hl7_errno = HL7_OK;
	if (!SEG(new)) hl7_errno = HL7_EOSEG;
	CURRENT = new;
	
	strcpy(a," ");
	if (SEG(CURRENT)) strcpy(a,SEG(CURRENT));
	strcpy(segid,strtok(a,&m->FS));
	return (HL7_R_OK);
}

int hl7NextSeg(PMSG m, char * segid)
{	unsigned int new;
	char a[max_segment_size];
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*----------------------------------- Is there a place to return segment id? */
	if (!segid)
	{	hl7_errno = HL7_NOSEGID;
		return (HL7_R_FAIL);
	}

	new = CURRENT + 1;
	hl7_errno = HL7_OK;
	if (new >= MAX_SEGMENTS || !SEG(new)) hl7_errno = HL7_EOSEG;
	else CURRENT = new;
	
	strcpy(a," ");
	if (SEG(CURRENT)) strcpy(a,SEG(CURRENT));
	strcpy(segid,strtok(a,&m->FS));
	return (HL7_R_OK);
}

int hl7PrevSeg(PMSG m, char * segid)
{	unsigned int new;
	char a[max_segment_size];
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*----------------------------------- Is there a place to return segment id? */
	if (!segid)
	{	hl7_errno = HL7_NOSEGID;
		return (HL7_R_FAIL);
	}

	new = CURRENT - 1;
	hl7_errno = HL7_OK;
	if (new < 0 || !SEG(new)) hl7_errno = HL7_EOSEG;
	else CURRENT = new;
	
	strcpy(a," ");
	if (SEG(CURRENT)) strcpy(a,SEG(CURRENT));
	strcpy(segid,strtok(a,&m->FS));
	return (HL7_R_OK);
}

int hl7LastSeg(PMSG m, char * segid)
{	unsigned int new;
	char a[max_segment_size];
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*----------------------------------- Is there a place to return segment id? */
	if (!segid)
	{	hl7_errno = HL7_NOSEGID;
		return (HL7_R_FAIL);
	}

	new = 0;
	while (new < MAX_SEGMENTS && SEG(new)) new++;
	if (new > 0) new--;
	hl7_errno = HL7_OK;
	if (!SEG(new)) hl7_errno = HL7_EOSEG;
	CURRENT = new;
	
	strcpy(a," ");
	if (SEG(CURRENT)) strcpy(a,SEG(CURRENT));
	strcpy(segid,strtok(a,&m->FS));
	return (HL7_R_OK);
}

/******************************************************************************
 * Position to segment number.  If no such segment, position to first or last
 * segment, whichever is closer.  Return actual segment number and segment id.
 *****************************************************************************/
int hl7SegNum(PMSG m, char * segid, int * segnumber)
{	int ret;
	char a[max_segment_size];
	
	if (*segnumber < 0)
	{	ret = hl7FirstSeg(m, segid);
		*segnumber = CURRENT;
	}
	else if (*segnumber > MAX_SEGMENTS || !SEG(*segnumber))
	{	ret = hl7LastSeg(m, segid);
		*segnumber = CURRENT;
	}
	else 
	{	CURRENT = *segnumber;
		strcpy(a," ");
		if (SEG(CURRENT)) strcpy(a,SEG(CURRENT));
		strcpy(segid,strtok(a,&m->FS));
		return (HL7_R_OK);
	}
}

/******************************************************************************
 * Find segment with given segment id, starting after current segment.
 * Returns OK, EOSEQ if no segment found.  Returns segment number in 3rd 
 * parameter.
 *****************************************************************************/
int hl7FindSeg(PMSG m, char * segid, int * segnumber)
{	char isegid[24];
	int ret;
	while (1)
	{	ret = hl7NextSeg(m, isegid);
		*segnumber = CURRENT;
		if (ret == HL7_R_FAIL) return (HL7_R_FAIL);
		if (hl7_errno == HL7_EOSEG) return (HL7_R_OK);
		if (!strcasecmp(segid,isegid)) return (HL7_R_OK);
	}
}

/******************************************************************************
 * Write current segment to char buffer.
 *****************************************************************************/
int hl7WriteSeg(PMSG m, char * seg, int size)
{	hl7_errno = HL7_OK;
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*----------------------------------- Is there a place to return segment id? */
	if (!seg)
	{	hl7_errno = HL7_NOSEGID;
		return (HL7_R_FAIL);
	}
/*---------------------------------------------- Is there a current segment? */
	if (CURRENT < 0 || CURRENT > MAX_SEGMENTS ||
		 !SEG(CURRENT))
	{	hl7_errno = HL7_NOSEGDEF;
		return (HL7_R_FAIL);
	}

	memset(seg,0,size);
	strcpy(seg,SEG(CURRENT));
	return (HL7_R_OK);
}

/******************************************************************************
 * Write message to stream using specified format.  Useful for logging and
 * debugging.  Message is already loaded.  Stream must be already opened for 
 * write and positioned properly. Multiple formats are supported, depending on
 * the format code.  The number of parameters is determined by the format code,
 * which is parameter 0.
 *****************************************************************************/
int hl7WriteTxt(PMSG m, FILE * fp, int par[])
{	int format = par[0], margin, line_length, indent;
	hl7_errno = HL7_OK;
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*-------------------------------------------------------------- File given? */
	if (!fp)
	{	hl7_errno = HL7_NOFILE;
		return (HL7_R_FAIL);
	}
/*---------------------------------------------------------- Validate format */
	switch (format)
/*-----------------------------------------------------------------------------
 * Simple line format, one segment per line, regardless of length.  EO segment
 * is implied in the newline.  
 * par[1] = left margin.
 *---------------------------------------------------------------------------*/
	{	case HL7_LINE:
			margin = par[1]; if (margin < 0) margin = 0;
			break;
/*-----------------------------------------------------------------------------
 * Line format, but with a maximum line length.  Each new segment starts at the
 * beginning of a new line; segments longer than the line length are wrapped to
 * subsequent line or lines, indented. EO segment is implied in the newline.
 * par[1] = margin.
 * par[2] = Line length 15-250 characters (after margin)
 * par[3] = Indent for extra lines 0-5 characters.
 *---------------------------------------------------------------------------*/
		case HL7_LINEWRAP:
			margin = par[1]; if (margin < 0) margin = 0;
			line_length = par[2]; 
			if (line_length < 15)  line_length = 15;
			if (line_length > 250) line_length = 250;
			indent = par[3];
			if (indent < 0) indent = 0;
			if (indent > 5) indent = 5;
			break;
/*-----------------------------------------------------------------------------
 * Block format, no newlines at EO segment.  EO segment character replaced with
 * <CR>.
 * par[1] = margin.
 * par[2] = line length 15-250 characters.
 *---------------------------------------------------------------------------*/
		case HL7_BLOCK:
			margin = par[1]; if (margin < 0) margin = 0;
			line_length = par[2]; 
			if (line_length < 15)  line_length = 15;
			if (line_length > 250) line_length = 250;
			break;
		default:	
			hl7_errno = HL7_INVPAR;
			return (HL7_R_FAIL);
	}
	
	if (format == HL7_LINE)
	{	int i, j;
		char a[MAX_LINE];
		for (i = 0; i < hl7NumSegs(m); i++)           /* for each segment */
		{	j = margin;                                     /* margin */
			while (j--)
			{	if (fputc(' ', fp) == EOF)
				{	hl7_errno = HL7_IOERR;
					return (HL7_R_FAIL);
				}
			}
			strcpy(a,SEG(i));
			a[strlen(a) -1] = '\n';
			if (fputs(a, fp) == EOF)
			{	hl7_errno = HL7_IOERR;
				return (HL7_R_FAIL);
			}
		}
	}

	if (format == HL7_LINEWRAP)
	{	int i, j, indent_flag;
		char * p;
		char a[MAX_LINE];
		for (i = 0; i < hl7NumSegs(m); i++)
		{	p = SEG(i);
			indent_flag = 0;
			memset(a,0,MAX_LINE);
			while (p)
			{	j = line_length - indent_flag * indent;
				if ((strlen(p) -1) <= j)
				{	strcpy(a,p); 
					a[strlen(a) -1] = '\n';
					p = NULL;
				}
				else
				{	memset(a,0,strlen(a));
					memcpy(a,p,j);
					a[strlen(a)] = '\n';
					p += j;
				}
				j = margin + indent_flag * indent;
				while (j--)
				{	if (fputc(' ', fp) == EOF)
					{	hl7_errno = HL7_IOERR;
						return (HL7_R_FAIL);
					}
				}
				if (fputs(a, fp) == EOF)
				{	hl7_errno = HL7_IOERR;
					return (HL7_R_FAIL);
				}
				indent_flag = 1;
			}
		}
	}

	if (format == HL7_BLOCK)
	{	int i, j, sz = margin + line_length;
		char * p;
		char a[MAX_LINE + 5], b[margin + line_length + 2];
		memset(b,0,sizeof(b));
		memset(b,' ',margin);
		for (i = 0; i < hl7NumSegs(m); i++)    /* Each segment */
		{	strcpy(a,SEG(i));
			a[strlen(a) - 1] = 0;             /* Replace \n with <CR> */
			strcat(a, "<CR>");
			p = a;
			while (1)
			{	if (strlen(b) == sz)          /* Complete line, output */
				{	b[strlen(b)] = '\n';
					if (fputs(b, fp) == EOF)
					{	hl7_errno = HL7_IOERR;
						return (HL7_R_FAIL);
					}
					memset(b,0,sizeof(b));        /* Reset line buffer */
					memset(b,' ',margin);
				}
				j = sz - strlen(b);             /* Space remaining in line */
				if (strlen(p) <= j)             /* Rest of this seg will fit */
				{	strcat(b,p);
					break;
				}
				memcpy(b + strlen(b),p,j);     /* Some of this seg will fit */
				p += j;
				continue;
			}
		}
		if (strlen(b) > margin)      /* EO segments, flush last line if anything in it */
		{	b[strlen(b)] = '\n';
			if (fputs(b, fp) == EOF)
			{	hl7_errno = HL7_IOERR;
				return (HL7_R_FAIL);
			}
		}	  
	}
	hl7_errno = HL7_OK;
	return (HL7_R_OK);
}

/******************************************************************************
 * Copy current segment in first msg after current segment in second msg.
 * New segment becomes destination current segment.
 *****************************************************************************/
int hl7CopySeg(PMSG src, PMSG dest)
{	int i, j, k;
/*------------------------------------------------ Is there a valid message? */
	if (!src) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*------------------------------------------------ Is there a valid message? */
	if (!dest) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*-------------------- Is there room in the destination for another segment? */
	if (hl7NumSegs(dest) == MAX_SEGMENTS)
	{	hl7_errno = HL7_SEGFULL;
		return (HL7_R_FAIL);
	}
/*------------------------------------------- Locate/Validate source segment */
	i = src->current_segment;
	if (!src->seg[i])
	{	hl7_errno = HL7_NOSEGDEF;
		return (HL7_R_FAIL);
	}
	j = dest->current_segment;
	if (dest->seg[j]) j++;
	if (dest->seg[j])	for (k = hl7NumSegs(dest); k >= j; k--) dest->seg[k+1] = dest->seg[k];
	dest->seg[j] = malloc(strlen(src->seg[i] + 1));
	strcpy(dest->seg[j],src->seg[i]);

	dest->current_segment = j;
	
	hl7_errno = HL7_OK;
	return (HL7_R_OK);
}

/******************************************************************************
 * Make new segment after current segment, using seg.  New segment becomes
 * current segment. seg should at least have segment id, but could have other
 * fields.  If supplied segment is just segment id, an FS and ST will be 
 * added.  If supplied segment has more than just segment id, and does not 
 * have terminating ST, it will be added.  Otherwise, it is up to the user
 * to insure that the supplied segment is appropriate.
 *****************************************************************************/
int hl7InsSeg(PMSG m, char * seg)
{	int j, k;
	char a[max_segment_size];
	memset(a,0,sizeof(a));
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*-------------------------------------------------------------- segment id? */
	if (!seg || !strlen(seg))
	{	hl7_errno = HL7_NOSEGID;
		return (HL7_R_FAIL);
	}
	strcpy(a,seg);
/*----------------------------------------------------- Fix up seg as needed */
	if (LASTCHAR(a) != m->FS && LASTCHAR(a) != ST) ADDCHAR(a,m->FS);
	if (LASTCHAR(a) != ST) ADDCHAR(a,ST);

/*------------------------ Is there room in the message for another segment? */
	if (hl7NumSegs(m) == MAX_SEGMENTS)
	{	hl7_errno = HL7_SEGFULL;
		return (HL7_R_FAIL);
	}
	j = CURRENT;
	if (SEG(j)) j++;
	if (SEG(j))	for (k = hl7NumSegs(m); k >= j; k--) SEG(k+1) = SEG(k);
	SEG(j) = malloc(strlen(a) + 3);
	
	strcpy(SEG(j),a);

	CURRENT = j;

	hl7_errno = HL7_OK;
	return (HL7_R_OK);
}


/******************************************************************************
 * Delete current segment.  Following segment becomes current segment, with
 * allowance for being at end of segments
 *****************************************************************************/
int hl7DelSeg(PMSG m)
{	int i, j, k;
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*------------------------------------------------------- No current segment */
	j = CURRENT;
	if (!SEG(j))
	{	hl7_errno = HL7_NOSEGDEF;
		return (HL7_R_FAIL);
	}

	free(SEG(j));
	i = hl7NumSegs(m);
	for (k = j; k < i; k++) SEG(k) = SEG(k+1);
	SEG(i) = NULL;
	if (!SEG(j) && j > 0) j--;
	CURRENT = j;

	hl7_errno = HL7_OK;
	return (HL7_R_OK);
}

/******************************************************************************
 * Returns the number of segments in the message.
 *****************************************************************************/
int hl7NumSegs(PMSG m)
{	int i = 0;
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (0); 
	}
	while (SEG(i)) i++;
	hl7_errno = HL7_OK;
	return (i);
}

/******************************************************************************
 * Returns the number of fields in the segment.  This includes the segment id.
 * See note on hl7GetFld below on field numbering.
 *****************************************************************************/
int hl7NumFlds(PMSG m)
{	int i;
	char * p;
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (0); 
	}
/*------------------------------------------------ Is there a valid segment? */
	if (!SEG(CURRENT))
	{	hl7_errno = HL7_NOSEGDEF;
		return (0);
	}

	p = SEG(CURRENT);
	i = 1;
	while (*p) if (*p++ == m->FS) i++;
	if (!memcmp(SEG(CURRENT),"MSH",3)) i++;		
	
	hl7_errno = HL7_OK;
	return (i);
}

/******************************************************************************
 * Generate a simple ack for message.  msg is the hl7 message we want to ack, 
 * the ack will be built in ack_msg (anything in it already will be cleared.)
 * ack_code is the ACK code (AA, AE, or AR; default is AA). ack_text is the
 * ack text, if any). NOTE: This simple ACK does not cover all possibilities.
 *****************************************************************************/
int hl7ACK(PMSG msg, PMSG ack_msg, char * ack_code, char * ack_text)
{	char code[3] = HL7_AA;
	char a[max_segment_size], msg_ctrl_num[24];
	char * p;
	time_t now;
	hl7_errno = HL7_OK;
/*------------------------------------- Validate message structures and code */
	if (!msg || !ack_msg)
	{	hl7_errno = HL7_NOMSG;
		return (HL7_R_FAIL);
	}
	if (ack_code)
	{	if (strcmp(ack_code,HL7_AA) && strcmp(ack_code,HL7_AE) && strcmp(ack_code,HL7_AR))
		{	hl7_errno = HL7_INVACK;
			return (HL7_R_FAIL);
		}
		else strcpy(code,ack_code);
	}
/*---------------------------------------- Clear anything already in message */
	if (hl7NewMsg(ack_msg)) return (HL7_R_FAIL);
/*------------------------------------------------------- Copy first segment */
	if (hl7FirstSeg(msg, a)) return (HL7_R_FAIL);
	if (hl7CopySeg(msg, ack_msg)) return (HL7_R_FAIL);
/*----------------------------------------------------- Time hack in field 7 */
	now = time(NULL);
	strftime(a, sizeof(a), "%Y%m%d%H%M", localtime(&now));
	if (hl7PutFld(ack_msg, a, 7));
/*----------------------------------- ORM's become ORR^O02.  Rest become ACK */
	p = hl7GetComp(msg, 9, 1, 1);
	if (!p) return (HL7_R_FAIL);
	if (STR_EQ(p,"ORM"))
	{	if (hl7PutFld(ack_msg, "ORR", 9)) return (HL7_R_FAIL);
		if (hl7PutComp(ack_msg, "O02", 9)) return (HL7_R_FAIL);
	}
	else if (hl7PutFld(ack_msg, "ACK", 9)) return (HL7_R_FAIL);	
/*-------------------------------------------------- Pull message control id */
	p = hl7GetFld(msg, 10);
	if (!p) return (HL7_R_FAIL);
	strcpy(msg_ctrl_num, p);
/*------------------------------------------------------- Insert MSA segment */
	if (hl7InsSeg(ack_msg, "MSA")) return (HL7_R_FAIL);
	if (hl7PutFld(ack_msg, code, 1)) return (HL7_R_FAIL);
	if (hl7PutFld(ack_msg, msg_ctrl_num, 2)) return (HL7_R_FAIL);
  	if (ack_text) if (hl7PutFld(ack_msg, ack_text, 3)) return (HL7_R_FAIL);
	return (HL7_R_OK);
}

/******************************************************************************
 * Get field by number.  This uses the field numbering of HL7 standard.  The
 * segment id is field 0 and the fields work up from there.  The one exception
 * is the MSH segment, in which the field separator after the segment id is
 * field 1, the characters after it and before the next field separator are
 * field 2, and then the numbering continues.  This numbering applied to all
 * the get and put field and subfield routines.
 *****************************************************************************/
char * hl7GetFld(PMSG m, int field)
{	int i, j, MSH_flag;
	char *p, * pv;
	hl7_errno = HL7_OK;
	if (field < 0) field = 0;
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (NULL); 
	}
/*------------------------------------------------ Is there a valid segment? */
	if (!SEG(CURRENT))
	{	hl7_errno = HL7_NOSEGDEF;
		return (NULL);
	}
/*-------------------------------------------------- Is this an MSH segment? */
	if (memcmp(SEG(CURRENT),"MSH",3)) MSH_flag = FALSE;
	else MSH_flag = TRUE;

/*-----------------------------------------------------------------------------
 * Special case: field 1 of an MSH segment is the Field Separator.
 *---------------------------------------------------------------------------*/
	if (MSH_flag && field == 1){v[0]=m->FS; v[1]=0; return(v);}	

	p = SEG(CURRENT);             /* pointer in segment */
	i = 0;                        /* Current field we are looking for */
	memset(v,0,sizeof(v));        /* Field we are buildling value in */
	pv = v;                       /* Pointer to field we are building value in */ 

	while (1)
	{	if (!(*p)) return (v);         /* EO Seg, rest of values empty */
		if (*p == m->FS || *p == ST)   /* EO Fld */
		{	if (i == field) return (v);     /* return if this is right one */
			if (j = pv -v) { memset(v,0,j); pv = v; }  /* Clear for next field */
			i++; p++;                      /* bump field # and pointer */
			if (MSH_flag && i == 1) i++;   /* Special case handled above */
			continue;
		}
		if (*p == m->ES) {*pv = *p; pv++; p++; } /* Escape char; copy it and next char */
		*pv = *p; pv++; p++;                  /* Field character, move it over */
	}
}

/******************************************************************************
 * Get repetition.  This uses hl7GetFld, then operates on the field. Reps are
 * numbered starting at 1.
 *****************************************************************************/
char * hl7GetRep(PMSG m, int field, int repetition)
{	char a[sizeof(v)], *pa, *pv;
	int i, j;
/*---------------------------------- First, pull the field and copy it to ws */
	if (!hl7GetFld(m, field)) return (NULL);
	strcpy(a,v);
	
	if (repetition < 1) repetition = 1;
	pa = a;                            /* pointer in field */
	memset(v,0,strlen(v));             /* Field we are building rep in */
	pv = v;                            /* pointer in building field */
	i = 1;                             /* rep we are currently looking for */

	while (1)
	{	if (!(*pa) || *pa == m->RS)       /* EO rep */
		{	if (i == repetition) return (v); /* return if this is right one */
			if (j = pv -v) { memset(v,0,j); pv = v; }  /* Clear for next field */
			if (!(*pa)) return (v);         /* EO field, all other reps empty */
			i++; pa++;                      /* bump field # and pointer */
			continue;
		}
		if (*pa == m->ES) {*pv = *pa; pv++; pa++; } /* Escape char; copy it and next char */
		*pv = *pa; pv++; pa++;
	}
}

/******************************************************************************
 * Get component.  This uses hl7GetRep, then operates on the field. Components
 * numbered starting at 1.
 *****************************************************************************/
char * hl7GetComp(PMSG m, int field, int repetition, int component)
{	char a[sizeof(v)], *pa, *pv;
	int i, j;
/*---------------------------------- First, pull the field/rep and copy it to ws */
	if (!hl7GetRep(m, field, repetition)) return (NULL);
	strcpy(a,v);
	
	if (component < 1) component = 1;
	pa = a;                            /* pointer in field */
	memset(v,0,strlen(v));             /* Field we are building comp in */
	pv = v;                            /* pointer in building field */
	i = 1;                             /* comp we are currently looking for */

	while (1)
	{	if (!(*pa) || *pa == m->CS)       /* EO component */
		{	if (i == component) return (v); /* return if this is right one */
			if (j = pv -v) { memset(v,0,j); pv = v; }  /* Clear for next field */
			if (!(*pa)) return (v);         /* EO field, all other reps empty */
			i++; pa++;                      /* bump field # and pointer */
			continue;
		}
		if (*pa == m->ES) {*pv = *pa; pv++; pa++; } /* Escape char; copy it and next char */
		*pv = *pa; pv++; pa++;
	}
}

/******************************************************************************
 * Get Subcomponent.  This uses hl7GetComp, then operates on the field.
 * Subcomponents are numbered starting at 1.
 *****************************************************************************/
char * hl7GetSub(PMSG m, int field, int repetition, int component, int subcomponent)
{	char a[sizeof(v)], *pa, *pv;
	int i, j;
/*---------------------------------- First, pull the field/rep and copy it to ws */
	if (!hl7GetComp(m, field, repetition, component)) return (NULL);
	strcpy(a,v);
	
	if (subcomponent < 1) subcomponent = 1;
	pa = a;                            /* pointer in field */
	memset(v,0,strlen(v));             /* Field we are building comp in */
	pv = v;                            /* pointer in building field */
	i = 1;                             /* comp we are currently looking for */

	while (1)
	{	if (!(*pa) || *pa == m->SS)       /* EO component */
		{	if (i == subcomponent) return (v); /* return if this is right one */
			if (j = pv -v) { memset(v,0,j); pv = v; }  /* Clear for next field */
			if (!(*pa)) return (v);         /* EO field, all other reps empty */
			i++; pa++;                      /* bump field # and pointer */
			continue;
		}
		if (*pa == m->ES) {*pv = *pa; pv++; pa++; } /* Escape char; copy it and next char */
		*pv = *pa; pv++; pa++;
	}
}

/******************************************************************************
 * Put Field in current segment, by number.  Replaces entire current contents 
 * of field, if any.  Adds any needed FS entries.  User may provide complex
 * fields as entries (reps, comp, sub), but needs to determine the correct
 * separator characters to use. Returns HL7_R_OK or HL7_R_FAIL.  See doc at
 * hl7GetFld for further info on field numbering.
 * Insuring that field contents are valid is the responsibility of the user.
 *****************************************************************************/
int hl7PutFld(PMSG m, char * value, int field)
{	int MSH_flag, i, j;
	char * pv, *ps, * max_pv, * max_ps;
	char v[max_segment_size + 2];
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*------------------------------------------------ Is there a valid segment? */
	if (!SEG(CURRENT))
	{	hl7_errno = HL7_NOSEGDEF;
		return (HL7_R_FAIL);
	}
/*-------------------------------------------------- Is this an MSH segment? */
	if (memcmp(SEG(CURRENT),"MSH",3)) MSH_flag = FALSE;
	else MSH_flag = TRUE;
	
	if (field < 0) field = 0;
	i = 0;                /* Current field number */
	memset(v,0,sizeof(v)); 
	pv = v; max_pv = pv + sizeof(v);
	ps = SEG(CURRENT); max_ps = ps + strlen(SEG(CURRENT));
/*------------------------ Locate beginning of field we are going to change */
	while (1)
/*-----------------------------------------------------------------------------
 * We found the right field.  Bump counter until it is pointing to the first
 * character after this field (which is where we will pick up copying after we
 * have put the new value in, and break out of loop. Exception is field 1 of
 * MSH, which is itself an FS.
 *---------------------------------------------------------------------------*/
	{	if (i == field)
		{	if (MSH_flag && i == 1) { ps++; break; }
			while (*ps != m->FS && *ps != ST && *ps)
			{	if (*ps == m->ES) ps++;
				ps++;
			}
			break;
		}
/*-----------------------------------------------------------------------------
 * Not the right field yet, but at EO segment.  Append an FS to the output
 * buffer, bump the field count, and try again. MSH exception is same process
 *---------------------------------------------------------------------------*/
		if (*ps == ST || !(*ps))
		{	*pv = m->FS; pv++;
			i++;
			continue;
		}
/*-----------------------------------------------------------------------------
 * Not at the right field or at EO segment.  copy this field through its
 * terminator to the buffer, so that you are pointing to the beginning of the
 * next field, bump the field count, and try again.
 *---------------------------------------------------------------------------*/
		if (MSH_flag && i == 1) 
		{ 	*pv = *ps; pv ++; ps++;
			i++;
			continue; 
		}
		while (*ps != m->FS && *ps != ST && *ps)
		{	if (*ps == m->ES) {*pv = *ps; pv++; ps++; }
			*pv = *ps; pv++; ps++;
		}
		if (*ps == m->FS) { *pv = *ps; pv++; ps++; }
		else { *pv = m->FS; pv++; }
		i++;
		continue;
	}
/*-----------------------------------------------------------------------------
 * We have copied everything preceding the field we want from the segment to 
 * the buffer.  Now insert the new field value, if we have room.
 *---------------------------------------------------------------------------*/
	if (pv + strlen(value) >= max_pv)
	{	hl7_errno = HL7_SEGFULL;
		return (HL7_R_FAIL);
	}
	strcat(v,value);
/*-----------------------------------------------------------------------------
 * Now tack on the rest of the message, if we have room.
 *---------------------------------------------------------------------------*/
	if ((strlen(v) + strlen(ps)) >= sizeof(v))
	{	hl7_errno = HL7_SEGFULL;
		return (HL7_R_FAIL);
	}
	strcat(v,ps);
/*-----------------------------------------------------------------------------
 * If resulting new segment is larger, reallocate.  Otherwise, just copy it.
 *---------------------------------------------------------------------------*/
	if (strlen(v) > strlen(SEG(CURRENT)))
	{	free(SEG(CURRENT));
		SEG(CURRENT) = malloc(strlen(v) + 1);
	}
	strcpy(SEG(CURRENT),v);
	return (HL7_R_OK);
}

/******************************************************************************
 * Put repetition, put component, and put subcomponent adds their value at the 
 * end of the existing value, putting in the appropriate separator if there is 
 * any existing value.  It is up to the user to insure that the field is built
 * in the proper order and permits the structure they are using.
 *****************************************************************************/
int hl7PutRep(PMSG m, char * value, int field)
{	int ret;
	ret = hl7PutIn(m, value, field, m->RS);
	return (ret);
}
int hl7PutComp(PMSG m, char * value, int field)
{	int ret;
	ret = hl7PutIn(m, value, field, m->CS);
	return (ret);
}
int hl7PutSub(PMSG m, char * value, int field)
{	int ret;
	ret = hl7PutIn(m, value, field, m->SS);
	return (ret);
}

int hl7PutIn(PMSG m, char * value, int field, char separator)
{	int MSH_flag, i, j, existing_value = FALSE;
	char * pv, *ps, * max_pv, * max_ps;
	char v[max_segment_size + 2];
/*------------------------------------------------ Is there a valid message? */
	if (!m) 
	{	hl7_errno = HL7_NOMSG; 
		return (HL7_R_FAIL); 
	}
/*------------------------------------------------ Is there a valid segment? */
	if (!SEG(CURRENT))
	{	hl7_errno = HL7_NOSEGDEF;
		return (HL7_R_FAIL);
	}
/*-------------------------------------------------- Is this an MSH segment? */
	if (memcmp(SEG(CURRENT),"MSH",3)) MSH_flag = FALSE;
	else MSH_flag = TRUE;
	
	if (field < 0) field = 0;
	i = 0;                /* Current field number */
	memset(v,0,sizeof(v)); 
	pv = v; max_pv = pv + sizeof(v);
	ps = SEG(CURRENT); max_ps = ps + strlen(SEG(CURRENT));
/*------------------------ Locate beginning of field we are going to change */
	while (1)
/*-----------------------------------------------------------------------------
 * We found the right field. Copy until it is pointing to the first
 * character after this field (which is where we will pick up copying after we
 * have put the new value in), and break out of loop. Exception is field 1 of
 * MSH, which is itself an FS.
 *---------------------------------------------------------------------------*/
	{	if (i == field)
		{	if (MSH_flag && i == 1) { *pv = *ps; pv++; ps++; break; }
			while (*ps != m->FS && *ps != ST && *ps)
			{	if (*ps == m->ES) { *pv = *ps; pv++; ps++; }
				existing_value = TRUE;
				*pv = *ps; pv++; ps++;
			}
			break;
		}
/*-----------------------------------------------------------------------------
 * Not the right field yet, but at EO segment.  Append an FS to the output
 * buffer, bump the field count, and try again. MSH exception is same process
 *---------------------------------------------------------------------------*/
		if (*ps == ST || !(*ps))
		{	*pv = m->FS; pv++;
			i++;
			continue;
		}
/*-----------------------------------------------------------------------------
 * Not at the right field or at EO segment.  copy this field through its
 * terminator to the buffer, so that you are pointing to the beginning of the
 * next field, bump the field count, and try again.
 *---------------------------------------------------------------------------*/
		if (MSH_flag && i == 1) 
		{ 	*pv = *ps; pv ++; ps++;
			i++;
			continue; 
		}
		while (*ps != m->FS && *ps != ST && *ps)
		{	if (*ps == m->ES) {*pv = *ps; pv++; ps++; }
			*pv = *ps; pv++; ps++;
		}
 		*pv = *ps; pv++; ps++;
		i++;
		continue;
	}
/*-----------------------------------------------------------------------------
 * We have copied everything preceding the field we want from the segment to 
 * the buffer.  Now insert the new field value, if we have room.
 *---------------------------------------------------------------------------*/
	if (pv + strlen(value) +existing_value >= max_pv)
	{	hl7_errno = HL7_SEGFULL;
		return (HL7_R_FAIL);
	}
	if (existing_value) ADDCHAR(v,separator);
	strcat(v,value);
/*-----------------------------------------------------------------------------
 * Now tack on the rest of the message, if we have room.
 *---------------------------------------------------------------------------*/
	if ((strlen(v) + strlen(ps)) >= sizeof(v))
	{	hl7_errno = HL7_SEGFULL;
		return (HL7_R_FAIL);
	}
	strcat(v,ps);
/*-----------------------------------------------------------------------------
 * If resulting new segment is larger, reallocate.  Otherwise, just copy it.
 *---------------------------------------------------------------------------*/
	if (strlen(v) > strlen(SEG(CURRENT)))
	{	free(SEG(CURRENT));
		SEG(CURRENT) = malloc(strlen(v) + 1);
	}
	strcpy(SEG(CURRENT),v);
	return (HL7_R_OK);
}

/******************************************************************************
 * return error message string for ini_errno, or print to stdout
 *****************************************************************************/
char * GetHl7ErrorString(void)
{	if (hl7_errno < max_e_msg) return e_msg[hl7_errno];
	else return no_msg;
}
void PrtHl7ErrorString(FILE * file, char * prefix)
{	if (!file) file = stdout;
   fprintf(file, "%s: %s\n", prefix, GetHl7ErrorString());
	return;
}
