#ifndef _HL7_H_
#define _HL7_H_

/******************************************************************************
 * hl7.h                                                              R Moulton
 * Header file for hl7 routines, which allow parsing, modification, and logging
 * of messages in HL7 format.  Unless otherwise stated, all int returns are 
 * 0 =OK, non-zero =Failed.  All pointer returns are pointer=OK, NULL=Failed. 
 * hl7_errno is set at each call to indicate the specifics of the error. 
 * GetHl7ErrorString will return a pointer to a string describing the error,
 * PrtHl7ErrorString will print the string to the passed FILE * or stdout. 
 * All pointers returned are to a static area. They need not be "freed", but 
 * they must be copied away before a subsequent call.
 *-----------------------------------------------------------------------------
 * 9-3 R Moulton Initial write.
 *****************************************************************************/

#define MAX_SEGMENTS 1000
struct _hl7_message_ {
	char SB;                    /* message Start Byte.  Def 11   */
	char EB;                    /* message End Byte.    Def 28   */
	char FS;                    /* Field Separator.     Def '|'  */
	char CS;                    /* Component Separator. Def '^'  */
	char SS;                    /* Subcomp Separator.   Def '&'  */
	char RS;                    /* Repetition Separator Def '~'  */
	char ES;                    /* EScape character     Def '\'  */
	unsigned int current_segment; /* Current segment pointer */
	char * seg[MAX_SEGMENTS];
};
typedef struct _hl7_message_ HL7MSG;
typedef struct _hl7_message_ * PMSG;

/*-------------------------------------------------------------- error codes */

#define HL7_R_OK   0
#define HL7_R_FAIL 1

extern unsigned int hl7_errno;

#define HL7_OK         0  /* Returns OK */
#define HL7_NOMSG      1  /* May return OK */
#define HL7_INVMSGSIZE 2
#define HL7_INVMSG     3
#define HL7_NOSEGID    4
#define HL7_EOSEG      5  /* Returns OK */
#define HL7_NOBUFFER   6
#define HL7_NOSEGDEF   7
#define HL7_NOFILE     8
#define HL7_INVFORMAT  9
#define HL7_INVPAR    10
#define HL7_IOERR     11
#define HL7_SEGFULL   12
#define HL7_INVVAL    13
#define HL7_INVACK    14

/*-------------------------------------- Formats and paramters for hl7WriteTxt
 * See hl7.c for complete documentation
 *--------------------------------------------------------------------------*/
#define HL7_LINE       0   
#define HL7_LINEWRAP   1  
#define HL7_BLOCK      2  

static int hl7_line_pars[]  = {HL7_LINE,     3};
static int hl7_wrap_pars[]  = {HL7_LINEWRAP, 3,80,3};
static int hl7_block_pars[] = {HL7_BLOCK,    3,80};

/*-------------------------------------------------------------- ACK codes */
#define HL7_AA "AA"
#define HL7_AE "AE"
#define HL7_AR "AR"

int hl7FreeMemory(PMSG);
int hl7NewMsg(PMSG);
int hl7ReadMsg(PMSG, char * msg, int size);
int hl7WriteMsg(PMSG, char * msg, int size, int start_end);
int hl7FirstSeg(PMSG, char * segid);
int hl7NextSeg(PMSG, char * segid);
int hl7PrevSeg(PMSG, char * segid);
int hl7LastSeg(PMSG, char * segid);
int hl7FindSeg(PMSG, char * segid, int * segnumber);
int hl7SegNum(PMSG, char * segid, int * segnumber);
int hl7WriteSeg(PMSG, char * seg, int size);
int hl7WriteTxt(PMSG, FILE * fp, int par[]);

int hl7CopySeg(PMSG source, PMSG destination);
int hl7InsSeg(PMSG, char * seg);
int hl7DelSeg(PMSG);
int hl7NumSegs(PMSG);
int hl7NumFlds(PMSG);

int hl7ACK(PMSG msg, PMSG ack_msg, char * ack_code, char * ack_text); 

char * hl7GetFld (PMSG, int field);
char * hl7GetRep (PMSG, int field, int repetition);
char * hl7GetComp(PMSG, int field, int repetition, int component);
char * hl7GetSub (PMSG, int field, int repetition, int component, int subcomponent);

int hl7PutFld (PMSG, char * value, int field);
int hl7PutRep (PMSG, char * value, int field);
int hl7PutComp(PMSG, char * value, int field);
int hl7PutSub (PMSG, char * value, int field);

char * GetHl7ErrorString(void);
void PrtHl7ErrorString(FILE * file, char * prefix);
int hl7SetParameter(PMSG, int paramater_id, char * value);

#endif

