/******************************************************************************
 * log.c                                                              R Moulton
 * Library routines to support standardized log file system.
 *-----------------------------------------------------------------------------
 * 5-5 R Moulton Initial write.
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <dirent.h>

#include "dbg.h"
#include "ini.h"
#include "log.h"

/*------------------------- macros ------------------------------------------*/
#define FREE(A) {free(A);A=NULL;}
#define STR_EQ(A,B) !strcasecmp(A,B)
#define STR(A,B) {free(A); A=strdup(B);}
#define LCHR(A) A[strlen(A) - 1]
#define CLEAR(A) memset(A,0,sizeof(A))

/************************* Function Prototypes *******************************/
int logSetup(char * log_name);

/*-------------------------------- Error msg strings. Match to ini_errno values */
#define LOG_R_OK   0
#define LOG_R_FAIL 1

unsigned int log_errno;

static char * e_msg[] = 
{	"No error",
   "Not enabled",
   "Low level",
   "log type turned off",
   "NULL/invalid ini pfn",
   "ini pfn not found",
   "ini pfn not read/write",
   "logInitialize not run",
   "NULL/invalid log name",
   "log directory not found",
   "log directory stat err",
   "log directory not directory",
   "log dir needs RWX permissions",
   "couldn't open log file",
   "log message NULL",
   "nothing in log message",
   "couldn't open log directory",
   "log name section error",
   "hl7WriteTxt error",
   "mesgWrite error"
};
static int max_e_msg = sizeof(e_msg) / sizeof(char *);
static char * no_msg = "Unknown log_errno value";

/**************************** Working Storage ********************************/
static char * inipfn;
static char * log_name;
static char * sname;
static char log_dir[256];
static char * prefix;
static char * suffix;
static char daily;                 /* period (0=monthly, 1=daily) */
static FILE * fp;
static char enabled;                /* Is this log enabled? */
static int level;                   /* log level */

static time_t now;
static struct tm tm;
static char zone[] = "TZ=US/Central";

/*-------- Initialize logging; set inipfn, clear any previous log file entry */
int logInitialize(char * pfn)
{  log_errno = LOG_OK;
   if (fp) { fclose(fp); fp = NULL; }
   if (!pfn || !strlen(pfn)) { log_errno = LOG_INIINV; return (LOG_R_FAIL); }
   if (access(pfn, F_OK)) { log_errno = LOG_ININOF; return (LOG_R_FAIL); }
   if (access(pfn, R_OK | W_OK)) { log_errno = LOG_ININOTRW; return (LOG_R_FAIL); }
   STR(inipfn,pfn)
   FREE(log_name)
   CLEAR(log_dir);
   FREE(prefix)
   FREE(suffix)
   return (LOG_R_OK);
}

/*----------------------------------------------- Opens log file and returns fp
 * Log file pfn is determined from current time, using log directory, prefix,
 * and suffix defined for this log name in the .ini file
 *---------------------------------------------------------------------------*/
FILE * logOpen(char *log_name)
{  char pfn[256], tstring[12] = "";
   if (logSetup(log_name) == LOG_R_FAIL) return (NULL); /* Set up log */
   if (fp) return fp;                            /* already open */
   if (daily) strftime(tstring,sizeof(tstring),"%Y%m%d", &tm);    /* daily or monthly time */
   else strftime(tstring,sizeof(tstring),"%Y%m", &tm);
   sprintf(pfn, "%s/%s%s%s", log_dir, prefix, tstring, suffix);
   if (!enabled) { log_errno = LOG_DISABLED; strcpy(pfn,"/dev/null"); }
   fp = fopen(pfn,"a");
   if (!fp)
   {  sprintf(em,"open %s", pfn); dbge(em);
      log_errno = LOG_OPEN;
      return (NULL);
   }
   return (fp);
}
/*---------------------------------------------------------- Closes log file */
FILE * logClose(void)
{  if (fp) { fclose(fp); fp = NULL; }
   if (enabled) log_errno = LOG_OK; 
   else log_errno = LOG_DISABLED;
   return (NULL);
}

/*-----------------------------------------------------------------------------
 * Generates a log note, which is a one line log entry of the form:
 * YYYY-MM-DD HH:MM:SS - note
 * The note should not have any line feeds, even at the end.
 *---------------------------------------------------------------------------*/
int logNote(char * log_name, char * note)
{  char time_hack[20], *p;
   now = time(NULL);
   localtime_r(&now, &tm);
   log_errno = LOG_OK;
   if (!note) { log_errno = LOG_NULL; return (LOG_R_FAIL); }
/*-------------- strip any trailing nl's, replace internal one's with spaces */
   while (strlen(note) && LCHR(note) == '\n') LCHR(note) = 0;
   if (!strlen(note)) { log_errno = LOG_EMPTY; return (LOG_R_FAIL); }
   while (p = strchr(note,'\n')) *p = ' ';
/*--------------------------- We get a fresh open because the pfn may change */
   logClose(); 
   if (!logOpen(log_name)) return (LOG_R_FAIL);
   strftime(time_hack, sizeof(time_hack), "%Y-%m-%d %T", &tm);
   sprintf(em,"%s - %s", time_hack, note); dbg(em);
   fprintf(fp, "%s\n", em);
   logClose();
   return (LOG_R_OK);
}

/*-----------------------------------------------------------------------------
 * Generates a log message which is a multi line log entry of the form:
 * YYYY-MM-DD HH:MM:SS --------------------------------------------------------
 *    message
 * ============================================================================
 * The msg should have internal line feeds which cause it to break reasonably,
 * Line feeds at the end will be stripped.  If an internal blank line is
 * desired, there must be at least one space on the line (i.e. "\n \n", not 
 * "\n\n", as multiple adjacent new lines will be ignored.
 *---------------------------------------------------------------------------*/
#define MSG_LEAD_CHR "-"
#define MSG_TRAIL_CHR "="
#define MSG_LINE_LENGTH 80
#define MSG_INDENT 3

int logMsg(char * log_name, char * msg)
{  char *p, *m = NULL;
   now = time(NULL);
   localtime_r(&now, &tm);
   log_errno = LOG_OK;
   if (!msg) { log_errno = LOG_NULL; return (LOG_R_FAIL); }
/*-------------------------------------------------- strip any trailing nl's */
   while (strlen(msg) && LCHR(msg) == '\n') LCHR(msg) = 0;
   if (!strlen(msg)) { log_errno = LOG_EMPTY; return (LOG_R_FAIL); }
/*--------------------------- We get a fresh open because the pfn may change */
   logClose(); 
   if (!logOpen(log_name)) return (LOG_R_FAIL);
/*------------------------- First line - time hack followed by leading chars */
   strftime(em, sizeof(em), "%Y-%m-%d %T ", &tm);
   while (strlen(em) < MSG_LINE_LENGTH) strcat(em,MSG_LEAD_CHR);
   dbg(em); fprintf(fp, "%s\n", em);
/*------------------- message lines - indent plus segment to new line or end */
   STR(m,msg)
   p = strtok(m,"\n");
   while (p)
   {  CLEAR(em); 
         while (strlen(em) < MSG_INDENT) em[strlen(em)] = ' ';
      strcat(em,p);
      dbg(em); fprintf(fp, "%s\n", em);
      p = strtok(NULL,"\n");
   }
/*----------------------------------------------- last line - trailing chars */ 
   CLEAR(em);
   while (strlen(em) < MSG_LINE_LENGTH) strcat(em,MSG_TRAIL_CHR);
   dbg(em); fprintf(fp, "%s\n", em);
   
   logClose();
   return (LOG_R_OK);
}

/*-----------------------------------------------------------------------------
 * Same as logMsg, except that an hl7 message struct is written out after msg
 *---------------------------------------------------------------------------*/
int logHl7Msg(char * log_name, char * msg, PMSG hl7, int hl7par[])
{  char *p, *m = NULL;
   int hl7_ret;
   now = time(NULL);
   localtime_r(&now, &tm);
   log_errno = LOG_OK;
   if (!msg) { log_errno = LOG_NULL; return (LOG_R_FAIL); }
   if (!strlen(msg) && !hl7) { log_errno = LOG_EMPTY; return (LOG_R_FAIL); }
/*--------------------------- We get a fresh open because the pfn may change */
   logClose(); 
   if (!logOpen(log_name)) return (LOG_R_FAIL);
/*------------------------- First line - time hack followed by leading chars */
   strftime(em, sizeof(em), "%Y-%m-%d %T ", &tm);
   while (strlen(em) < MSG_LINE_LENGTH) strcat(em,MSG_LEAD_CHR);
   dbg(em); fprintf(fp, "%s\n", em);
/*------------------- message lines - indent plus segment to new line or end */
   STR(m,msg)
   p = strtok(m,"\n");
   while (p)
   {  CLEAR(em); 
         while (strlen(em) < MSG_INDENT) em[strlen(em)] = ' ';
      strcat(em,p);
      dbg(em); fprintf(fp, "%s\n", em);
      p = strtok(NULL,"\n");
   }
/*----------------------------------------------- last line - trailing chars */ 
   CLEAR(em);
   while (strlen(em) < MSG_LINE_LENGTH) strcat(em,MSG_TRAIL_CHR);
   dbg(em); fprintf(fp, "%s\n", em);

/*-------------------------------------------------------------- hl7 message */
   hl7_ret = hl7WriteTxt(hl7, fp, hl7par);
   
   logClose();
   if (hl7_ret == HL7_R_OK) return (LOG_R_OK);
   log_errno = LOG_HL7ERR;
   return (LOG_R_FAIL);
}

/*-----------------------------------------------------------------------------
 * Purge log files based on retention data in .ini file.  Returns number of log
 * files purged if OK, -1 on error.
 *---------------------------------------------------------------------------*/
int logPurge(char * log_name)
{  int purge_count = 0, retention, purge_hour, mo = 0, da = 0, yr = 0;
   char cutoff[24] = "", *p;
   DIR * direct = NULL;
   struct dirent * entry = NULL;
   char a[256];
   int sa = sizeof(a);
   now = time(NULL);
   localtime_r(&now, &tm);
   if (logSetup(log_name) == LOG_R_FAIL) return (-1);

/*------------------------------------------- Is it time for next purge run? */
   FREE(sname) sname = malloc(strlen(log_name)+5);
   strcpy(sname,"LOG_"); strcat(sname,log_name);
   GetIniString(sname, "LastPurge", "00-00-0000", a, sa, inipfn);
   if (p = strtok(a,"-")) mo = atoi(p);
   if (p = strtok(NULL,"-")) da = atoi(p);
   if (p = strtok(NULL,"-")) yr = atoi(p);
   GetIniInt(sname, "PurgeHour", 2, &purge_hour, inipfn);
   if (yr > (tm.tm_year + 1900)) return (0);
   if ((yr == (tm.tm_year + 1900)) && (mo > (tm.tm_mon + 1))) return (0);
   if (daily && (yr == (tm.tm_year + 1900)) && (mo == (tm.tm_mon + 1)) && (da >= tm.tm_mday)) return (0);
   if (tm.tm_hour < purge_hour) return (0);
/*---------- We are going to do a purge run - update last purge in .ini file */
   sprintf(em, "%d-%d-%d", tm.tm_mon + 1, tm.tm_mday, tm.tm_year + 1900);
   PutIniString(sname, "LastPurge", em, inipfn);
   
/*------------------------------------------------------ Compute cutoff date */  
   if (daily) retention = 30; else retention = 6; 
   GetIniInt(sname, "Retention", retention, &retention, inipfn);
/*----------------------------------- Daily logs retain that many days worth */
   if (daily) 
   {  now -= retention * 86400;
      localtime_r(&now, &tm);
      strftime(cutoff, sizeof(cutoff), "%Y%m%d", &tm);
   }
/*------------------------------- Monthly logs retain that many months worth */
   else
   {  while (retention >= (tm.tm_mon +1)) { tm.tm_year--; retention -= 12;}
      tm.tm_mon -= retention;
      strftime(cutoff, sizeof(cutoff), "%Y%m", &tm);
   }  

/*------------------------------------------------------- Open log directory */
   direct = opendir(log_dir);
   if (!direct) { log_errno = LOG_DIROPEN; return (-1); }
   while (entry = readdir(direct))  /* pass directory entries */
/*-- To insure only correct files are purged, verify name length as prefix + date + suffix */
   {  if (strlen(entry->d_name) != (strlen(prefix) + strlen(suffix) + strlen(cutoff))) continue;
      if (strncmp(prefix,entry->d_name,strlen(prefix)))continue; /* check prefix & suffix*/
      if (strcmp(suffix,entry->d_name + strlen(entry->d_name) - strlen(suffix))) continue;
      if (strncmp(entry->d_name + strlen(prefix), cutoff, strlen(cutoff)) > 0) continue;
      sprintf(a,"%s/%s", log_dir, entry->d_name); remove(a);
      sprintf(em,"purging %s log file - %s", log_name, entry->d_name); dbg(em);
      purge_count++;
   }
   closedir(direct);
   return(purge_count);
}
      
/*----------------------------------------- Sets up log file, if not current */
int logSetup(char * iname)
{  char * p, a[257], sname[24] = "LOG_", name[24];
   int sa = sizeof(a);
   DIR * dp;
   log_errno = LOG_OK;
   if (!inipfn) { log_errno = LOG_NOTINIT; return (LOG_R_FAIL); }
   if (!iname || !strlen(iname)) { log_errno = LOG_NAMEINV; return (LOG_R_FAIL); }
   strcpy(name,iname);
   for (p = name; *p; p++) *p = toupper(*p);
   if (log_name && STR_EQ(name,log_name)) 
   {  if (!enabled) log_errno = LOG_DISABLED;
      return (LOG_R_OK);
   }
/*-----------------------------------------------------------------------------
 * Not current, so we clear old one, but we will not put new name in until we
 * have successfully completed all the setup steps
 *---------------------------------------------------------------------------*/
   FREE(log_name)
   if (fp) { fclose(fp); fp = NULL; }
   
   strcat(sname,name);  /* section name = LOG_name */
/*------------------------------------------------------------ log directory */
   getcwd(a,sa); strcpy(log_dir,a);    /* default, cwd */
   if (!GetIniString(sname,"Dir","",a,sa,inipfn)){log_errno=LOG_SECINV; return(LOG_R_FAIL);}
   if (*a == '/') strcpy(log_dir,a);    /* .ini had absolute path */
   else if (*a == '.' && *(a+1) == '/') strcat(log_dir,a+1); /* .ini had ./ path */
   else { strcat(log_dir,"/"); strcat(log_dir,a); } /* .ini had subdirectory name */
/*--- Now verify that log_dir is a directory with read and write permissions */
   if (access(log_dir, F_OK)) 
   {  dbge("log_dir access"); 
      log_errno = LOG_DIRNOF;
      return (LOG_R_FAIL);
   }
   dp = opendir(log_dir);
   if (dp) closedir(dp);
   else 
   {  log_errno = LOG_DIRNOTDIR;
      return (LOG_R_FAIL);
   }
   if (access(log_dir, R_OK | W_OK | X_OK))
   {  dbge("log_dir perms");
      log_errno = LOG_DIRPERMS;
      return (LOG_R_FAIL);
   }
/*---------------- Log dir is OK; Get rest of stuff (which needs no checking */
   STR(log_name,name)
   GetIniString(sname, "Prefix", "", a, sa, inipfn);
   STR(prefix,a)
   GetIniString(sname, "Suffix", ".log", a, sa, inipfn);
   STR(suffix,a)
   GetIniString(sname, "Period", "DAILY", a, sa, inipfn);
   if (STR_EQ(a,"MONTHLY")) daily = 0; else daily = 1;

   GetIniString(sname, "Status", "ON", a, sa, inipfn);
   if (STR_EQ(a,"ON") || STR_EQ(a,"YES")) enabled = 1;
   else { enabled = 0; log_errno = LOG_DISABLED; }

   GetIniInt(sname, "Level", 0, &level, inipfn);

   return (LOG_R_OK);
}
   
/******************************************************************************
 * return error message string for log_errno, or print to stdout
 *****************************************************************************/
char * GetLogErrorString(void)
{	if (log_errno < max_e_msg) return e_msg[log_errno];
	else return no_msg;
}
void PrtLogErrorString(FILE * file, char * prefix)
{	if (!file) file = stdout;
   fprintf(file, "%s: %s\n", prefix, GetLogErrorString());
	return;
}

/*-----------------------------------------------------------------------------
 * Level based calls.  These generate log entries only if the level of the log
 * is less than or equal to the log_level in the call.
 *---------------------------------------------------------------------------*/
int logNoteLevel(char * log_name, char * note, char log_level)
{  if (logSetup(log_name) == LOG_R_FAIL) return (LOG_R_FAIL);
   if (level && level < log_level) { log_errno = LOG_LEVEL; return (LOG_R_OK); }
   return (logNote(log_name, note));
}
         
int logMsgLevel(char * log_name, char * msg, char log_level)
{  if (logSetup(log_name) == LOG_R_FAIL) return (LOG_R_FAIL);
   if (level && level < log_level) { log_errno = LOG_LEVEL; return (LOG_R_OK); }
   return (logMsg(log_name, msg));
}

int logHl7MsgLevel(char * log_name, char * prefix, PMSG hl7, int par[], char log_level)
{  if (logSetup(log_name) == LOG_R_FAIL) return (LOG_R_FAIL);
   if (level && level < log_level) { log_errno = LOG_LEVEL; return (LOG_R_OK); }
   return (logHl7Msg(log_name, prefix, hl7, par));
}

/*-----------------------------------------------------------------------------
 * Returns zero if an entry at this level would not be logged, positive if it 
 * would, and negative if an error occurred.
 *---------------------------------------------------------------------------*/
int logLevel(char * log_name, char log_level)
{  if (logSetup(log_name) == LOG_R_FAIL) return -1;
   if (level && level < log_level) return 0;
   else return 1;
}

int logNoteType(char * log_name, char * note, char * type)
{  char a[1024];
   int sa = sizeof(a);
   if (logSetup(log_name) == LOG_R_FAIL) return (LOG_R_FAIL);
   GetIniString(sname, type, "ON", a, sa, inipfn);
   if (STR_EQ(a,"OFF") || STR_EQ(a,"NO")) { log_errno = LOG_TYPE; return (LOG_R_OK); }
   return (logNote(log_name, note));
}
         
int logMsgType(char * log_name, char * msg, char * type)
{  char a[1024];
   int sa = sizeof(a);
   if (logSetup(log_name) == LOG_R_FAIL) return (LOG_R_FAIL);
   GetIniString(sname, type, "ON", a, sa, inipfn);
   if (STR_EQ(a,"OFF") || STR_EQ(a,"NO")) { log_errno = LOG_TYPE; return (LOG_R_OK); }
   return (logMsg(log_name, msg));
}
