/****************************************************************************************
 * map.c                                                                        R Moulton
 * Library routines to allow simple mapping functions on DICOM headers of DCM_OBJECT.
 ***************************************************************************************/

/*---------------------------------------------------------------------- Unix includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <regex.h>

/*--------------------------------------------------------------- CTN library includes */
#include "ctn_os.h"
#include "dicom.h"
#include "utility.h"
#include "dicom_uids.h"
#include "lst.h"
#include "condition.h"
#include "dulprotocol.h"
#include "dicom_objects.h"
#include "dicom_messages.h"
#include "dicom_services.h"
#include "dbquery.h"
#include "iap.h"

/*------------------------------------------------------------- MIRCS library includes */
#include "dbg.h"
#include "ini.h"
#include "log.h"

#include "map.h"

/*------------------------------ function prototypes ----------------------------------*/
static int is_header(char * line);
static char * chomp(char * a);
static void free_pmtag(PMTAG);
static void free_pmi(PMI);
static void free_pmt(PMT);
static void free_pmlt(PMLT);
static void free_pmm(PMM);
static void free_pmap(PMAP);
static PMTAG malloc_pmtag();
static PMI   malloc_pmi();
static PMT   malloc_pmt();
static PMLT  malloc_pmlt();
static PMM   malloc_pmm();
static PMAP  malloc_pmap();
static char * GetTag(PMTAG, DCM_OBJECT * obj, char first, char level);
static void SetTag(PMTAG, DCM_OBJECT * obj, char * new_value);
static void ClearTags(PMAP, char first);


/*----------------------------------------------------------------------------- Macros */
#define STR(A,B) { if (A) free(A); A=B; if (A) A=strdup(A); }
#define FREE(A)       if (A) { free(A);       A=NULL; }
#define FREE_PMTAG(A) if (A) { free_pmtag(A); A=NULL; }
#define FREE_PMI(A)   if (A) { free_pmi(A);   A=NULL; }
#define FREE_PMT(A)   if (A) { free_pmt(A);   A=NULL; }
#define FREE_PMLT(A)  if (A) { free_pmlt(A);  A=NULL; }
#define FREE_PMM(A)   if (A) { free_pmm(A);   A=NULL; }
#define FREE_PMAP(A)  if (A) { free_pmap(A);  A=NULL; }
#define STR_EQ(A,B) !strcasecmp(A,B)
#define ISON(A) (!strcasecmp(A,"ON") || !strcasecmp(A,"YES"))
#define LCHR(A) A[strlen(A) -1]
#define CLEAR(A) memset(A,0,sizeof(A))

/*-------------------------------- Error msg strings. Match to ini_errno values */
static char * e_msg[] = 
{	"No error",
   "No .ini file name",
   ".ini file not found",
   "Can't read .ini file",
   "No .ini section name",
   "No such .ini section",
   ".ini file error",
   "mapping not on",
   "memory allocation error",
   ".map file not found",
   "Can't read .map file",
   "Can't open .map file",
   "Internal error",
   "No Tags defined in .map file",
   "'Tag' line: invalid format",
   "'Tag' line: invalid name",
   "'Tag' line: duplicate name",
   "'Tag' line: invalid group",
   "'Tag' line: invalid element",
   "Header line invalid format",
   "'LogTag' tag undefined",
   "'If' line: invalid format",
   "'If' line: tag undefined",
   "'If' line: invalid relation",
   "'If' line: invalid logical",
   "'THEN:' expected, not found",
   "'Then' line: invalid format",
   "'Then' line: tag undefined",
   "'Then' line: '=' expected, not found",
   "Incomplete map definition",
   "Error reading .map file",
   "No valid mappings",
   "'Tag' line: No such DICOM tag",
   "Invalid DICOM tag",
   "SetTag failed",
   "NULL pmap pointer",
   "NULL DICOM object pointer",
   "Requested map off or not found",
   "attempt to map tag not found", 
   "no such map flag",
   "invalid regular expression",
};
static int max_e_msg = sizeof(e_msg) / sizeof(char *);
static char * no_msg = "Unknown map_errno value";

static CONDITION cond;
static unsigned long ul;
static void * ctx = NULL;

/****************************************************************************************
 * Map Flags are used in the THEN section of a mapping in lieu of changes to DICOM header
 * tags.  The flags are set by the map, then retrieved by the calling routine using
 * mapGetFlag.
 ***************************************************************************************/
static struct _map_flag_ {
   char * keyword;
   char * log_msg;
   char flag;
} mapflags[] = {
   { "DELETE IMAGE", "Image Deleted", 0},
   { "NO HOLD"     , "No IDQC Hold",  0},
};
#define FLAG mapflags[flg]
static char max_flag = sizeof(mapflags) / sizeof(struct _map_flag_);
static char flg;
/****************************************************************************************
 * Mapping setup. Loads all the mapping structures and prepares for mapping functions.
 * Returns a pointer to the map struct if mapping is on, NULL if not (or if an error 
 * occurred.)
 ***************************************************************************************/
PMAP mapSetup(char * section, char * inipfn)
{  char a[1024], b[1034];
   int sa = sizeof(a);
/*---------------------------------------------------------------------------------------
 * struct pointers.  First is one we are building, and which may have to be cleared at
 * exit.  tag2 is used for stepping through pointer lists, and will not have to be 
 * cleared.
 *-------------------------------------------------------------------------------------*/
   PMTAG pmtag = NULL, pmtag2;
   PMI   pmi   = NULL, pmi2;
   PMT   pmt   = NULL, pmt2;
   PMLT  pmlt  = NULL, pmlt2;
   PMM   pmm   = NULL, pmm2;
   PMAP  pmap  = NULL;
   
   FILE * map_fp = NULL;
   char * map_pfn = NULL, * def_log_name = NULL;
   char * p = NULL;
   enum _state_ { dicom_tags, map_header, map_if, map_ifovr, map_then, map_off, map_eof } state;
   char valid_map, valid_tag, valid_if, valid_then, valid_flag, keep_reading;
   
   map_errno = MAP_OK;
   
/*----------------------------------------------- Validate .ini file and section name */
   if (!inipfn) { map_errno = MAP_NOINI; goto ext; }
   if (access(inipfn, F_OK)) { map_errno = MAP_ININOF; goto ext; }
   if (access(inipfn, R_OK)) { map_errno = MAP_INIREAD; goto ext; }
   if (!section) { map_errno = MAP_NOINISEC; goto ext; }
/*-------------------------------------------------------------------- Is mapping on? */
   if (!GetIniString(section, "Status", "OFF", a, sa, inipfn))
   {  if (ini_errno == INI_SECNOF) { map_errno = MAP_SECNOF; goto ext; }
      map_errno = MAP_INIERR; goto ext; 
   }
   if (!ISON(a)) { map_errno = MAP_NOTON; goto ext; }
   
/*--------------------------------------------------------- Create overall map struct */
   pmap = malloc_pmap();
   if (!pmap) { map_errno = MAP_MALLOC; goto ext; }
 
   if (!GetIniString(section, "Debug", "stdout", a, sa, inipfn))
   {  PrtIniErrorString(stdout, "ini file error");
      map_errno = MAP_INIERR; goto ext;
   }
   pmap->debug_fp = debug_on(a);

   STR(pmap->inipfn,inipfn)
   
/*--------- Default log name, can be overidden in a mapping, if nof no default logging */
   GetIniString(section,"LogName","NOF",a,sa,inipfn);
   if (strcasecmp(a,"NOF")) 
   {  STR(def_log_name,a)
      if (logInitialize(inipfn)) dbg(GetLogErrorString());
   }

/*---------------------------------------------------------------------------------------
 * .map file pfn.  By default this is mircs.map.  The pfn must be either an absolute
 * path name, or in the same directory as the .ini file
 *-------------------------------------------------------------------------------------*/
   GetIniString(section,"MapFile","mircs.map",a,sa,inipfn);
   if (a[0] != '/')
   {  strcpy(b,inipfn);
      if (p=strrchr(b,'/')) 
      {  *(p+1) = '\0'; 
         strcat(b,a);
         strcpy(a,b);
      }
   }
   
   if (access(a, F_OK)) { map_errno = MAP_MAPNOF;  goto ext; }
   if (access(a, R_OK)) { map_errno = MAP_MAPREAD; goto ext; }
   STR(map_pfn,a)
   if (!(map_fp = fopen(map_pfn,"r"))) { map_errno = MAP_MAPOPEN; goto ext; }   

/*---------------------------------------------------------------------------------------
 * .map file is processed using a "next state" paradigm. States are:
 *    dicom_tags - Processing dicom tag entries at beginning of file.
 *    map_header - Processing a new map header, from [name] up to IF:
 *    map_if     - Processing conditional lines, from IF: to last if line before THEN:
 *    map_ifovr  - We have seen the last if, and now need to find the THEN:
 *    map_then   - Processing changes, from THEN: to next map or EOF
 *    map_off    - This map is not ON, skipping
 *    map_eof    - End of file or error has been found on .map file
 * --------------------------------------------------------------------------------------
 * Flags are:
 *    valid_map  - True if one+ valid mappings have been found/loaded.
 *    valid_tag  - True if one+ valid tag entries have been found/processed.
 *    valid_if   - True is one+ valid conditional lines found/prs for this mapping.
 *    valid_then - True if one+ valid change lines found/processed for this mapping.
 *    valid_flag - True if valid keyword then line found/processed for this map.
 * --------------------------------------------------------------------------------------
 * Any errors in the .map file will result in it not being processed.  However, maps 
 * which are not ON are not parsed, just skipped to the next map or EOF
 *-------------------------------------------------------------------------------------*/

   state = dicom_tags;
   valid_tag = valid_map = 0;
   keep_reading = 1;
   while (keep_reading)
   {  CLEAR(a);
      if (!fgets(a,sa,map_fp)) 
      {  strcpy(a,"... end of file ...");
         state = map_eof;
         keep_reading = 0;
      }
      dbg(chomp(a));
/*------------------------------------------------- Skip comment lines and blank lines */
      if (a[0] == '#' || !strlen(a)) continue;

      switch (state)
      {
/****************************************************************************************
 * dicom_tags state.  Need to find one or more Tags lines, then the first mapping section
 * name [mapname].
 ***************************************************************************************/
         case dicom_tags:
/*----------------------------------------------- First mapping header line: [mapname] */
            if (is_header(a))
            {  if (valid_tag)
               {  pmm = malloc_pmm();
                  if (!pmm) { map_errno = MAP_MALLOC; goto ext; }
                  p = strtok(a+1,"]");
                  STR(pmm->name,p)
                  STR(pmm->log_name,def_log_name)
                  pmm->log_level = 1;
                  state = map_header;
                  valid_if = valid_then = 0;
                  break;
               }
               else { map_errno = MAP_NOTAGS; goto ext; }
            }
/*------------------------- Should be looking at a line like: Tag name (group,element) */
            {  char n[1024] = "\0";
               unsigned int group = 0, element = 0;
               if(sscanf(a,"Tag %s (%x,%x)",n,&group,&element) != 3){map_errno=MAP_INVTAG; goto ext;}
/*--------------------------------------------------------------------------- tag name */
               if (!strlen(n)) { map_errno = MAP_TAGNAME; goto ext; }
/*--------------------------------------------------------------- Check for duplicates */
               for (pmtag2 = pmap->pmtag; pmtag2; pmtag2 = pmtag2->pmtag)
               {  if (STR_EQ(n,pmtag2->name))
                  {  map_errno = MAP_TAGDUP; 
                     goto ext;
                  }
               }
               if (!element) { map_errno = MAP_TAGELM; goto ext; }
/*-------------------------------------------------------------------- Make tag struct */
               pmtag = malloc_pmtag();
               if (!pmtag) { map_errno = MAP_MALLOC; goto ext; }
               STR(pmtag->name,n)
               pmtag->group = group;
               pmtag->element = element;
/*------------ 0 group is non-DICOM "tags".  All the rest need to be a valid DICOM tag */
               if (pmtag->group)
               {  pmtag->e.tag = DCM_MAKETAG(group,element);
                  cond = DCM_LookupElement(&(pmtag->e));
                  if (cond != DCM_NORMAL) { map_errno = MAP_DICOMTAG; goto ext; }
               }
               else strcpy(pmtag->e.description,n);
               
/*--------------------------------------- Valid tag, add to list in overall map struct */
               valid_tag = 1;                        
               if (!pmap->pmtag) pmap->pmtag = pmtag; else
               {  for (pmtag2 = pmap->pmtag; pmtag2->pmtag; pmtag2 = pmtag2->pmtag) ;
                  pmtag2->pmtag = pmtag;
               }
               pmtag = NULL;
               break;
            }
/****************************************************************************************
 * map_header state.  We have a pmm struct, and we need to fill in the header items. 
 * All of these entries have defaults, so we don't need them, and if multiple entries are
 * present, we will just use the last one.  They are:
 *    Status=[ON|OFF].  (If OFF, we clear the struct and go to MAP_OFF state)
 *    Level=[STUDY|IMAGE]  
 *    LogName=name
 *    LogLevel=#
 *    LogMsg=message
 *    LogTags=tagname,tagname,tagname
 *
 * The LogTag entries all have to be in the tags section, and each one is entered in the 
 * map header
 *    
 * Then we should find the "IF:" line.
 ***************************************************************************************/
         case map_header:
            if (STR_EQ(a,"IF:")) { state = map_if; break; }
            if (!strchr(a,'=')) { map_errno = MAP_HDRINV; goto ext; }
            p = strtok(a,"=");
/*------------------- Status=[ON|OFF]  If not ON, clear struct and skip to next header */
            if (STR_EQ(p,"Status"))
            {  p = strtok(NULL," ");
               if (!p || !ISON(p))
               {  FREE_PMM(pmm)
                  state = map_off;
               }
               break;
            }
/*--------------------------------- Evaluate at Study(0) or Image(1) level (def study) */
            if (STR_EQ(p,"Level"))
            {  pmm->level = 0;
               p = strtok(NULL," ");
               if (p && STR_EQ(p,"IMAGE")) pmm->level = 1;
               break;
            }
/*---------------------------------------------------------------------------- LogName */
            if (STR_EQ(p,"LogName"))
            {  p = strtok(NULL," ");
               if (p && strlen(p)) STR(pmm->log_name,p)
               break;
            }
/*--------------------------------------------------------------------------- Category */
            if (STR_EQ(p,"Category"))
            {  p = strtok(NULL," ");
               if (p && strlen(p)) STR(pmm->category,p)
               break;
            }
/*------------------------------------------------------ LogLevel (integer, default 1) */
            if (STR_EQ(p,"LogLevel"))
            {  p = strtok(NULL," ");
               if (p) pmm->log_level = atoi(p);
               break;
            }
/*------------------------------------------------------------------------ Log message */
            if (STR_EQ(p,"LogMsg"))
            {  p = strtok(NULL,"\n");
               FREE(pmm->log_msg)
               if (p) STR(pmm->log_msg,p)
               break;
            }
/*------------------------------------------------------------------------------ Log Tags
 * Each one must be one of the Tags entries.  What we store is the pointer to that Tags
 * entry, from which we will get the value to print at logging time.
 * For LogTags, tags in an additional entry are added to those already there.
 *-------------------------------------------------------------------------------------*/
            if (strcasecmp(p,"LogTags")) { map_errno = MAP_HDRINV; goto ext; }
/*------------------------------------------- Process each tag in comma separated list */
            while (p=strtok(NULL,","))
/*--------------------------------------------------- Find it in tag list or error out */
            {  for (pmtag2 = pmap->pmtag; pmtag2; pmtag2 = pmtag2->pmtag)
               {  if (STR_EQ(p,pmtag2->name)) break;
               }
               if (!pmtag2) { map_errno = MAP_HDRLTAG; goto ext; }
/*--------------------------------------------------------------- Create logtag struct */
               pmlt = malloc_pmlt();
               if (!pmlt) { map_errno = MAP_MALLOC; goto ext; }
               pmlt->pmtag = pmtag2;
/*------------------------------------------------------ Add it to the map logtag list */
               if (!pmm->pmlt) pmm->pmlt = pmlt; else
               {  for (pmlt2 = pmm->pmlt; pmlt2->pmlt; pmlt2 = pmlt2->pmlt) ;
                  pmlt2->pmlt = pmlt;
               }
               pmlt = NULL;
            }
            break;
/****************************************************************************************
 * Processing lines in the IF: section.  These are of the format:
 *    TAG [==|!=|>|>=|<|<=] [TAG|"string"] [&& or ||] or
 *    TAG [=~|!~] "regular expression" [IGNORE_CASE] [&& or ||]
 * All the tags must be defined.  The last line before the "THEN:" line does not have a
 * logical && or ||. Once that is seen, the next line has to be the THEN:.  We do not
 * currently handle \" or "" in the strings or regular expressions.
 * The =~ and !~ operators (syntax borrowed from Perl) indicate a match or not match to
 * the following regular expression in quotes.  If IGNORE_CASE is present (regular expr
 * only) the computer will ignore the case of alphabetic characters when determining a
 * match.
 ***************************************************************************************/
         case map_if:
            {  char left_side[1024] = "\0", relation[1024] = "\0", 
                  right_side[1024] = "\0", and_or[1024] = "\0";
               char quoted = 0; 
               char * rest = NULL;
               int ignore_case = 0, ret;
/*---------------------------------------------------------------------------------------
 * Can't use sscanf, because the right side may be a quoted string.  parse line 
 *-------------------------------------------------------------------------------------*/
               if (p=(char *)strtok_r(a," ",&rest)) strcpy(left_side,p);
               else {map_errno = MAP_IFINV; goto ext; }
               
               if (p=(char *)strtok_r(rest," ",&rest)) strcpy(relation,p);
               else {map_errno = MAP_IFINV; goto ext; }
               chomp(rest);
               
               if (strchr(relation,'~'))
               {  if (!rest) {map_errno = MAP_IFINV; goto ext; }
                  if (rest[0] != '"') {map_errno = MAP_IFINV; goto ext; }
                  if (p=(char *)strtok_r(rest+1,"\"",&rest)){strcpy(right_side,p);}
                  else { map_errno = MAP_IFINV; goto ext; } 
                  if (p=strstr(rest,"IGNORE_CASE"))
                  {  ignore_case = REG_ICASE;
                     rest = p + strlen("IGNORE_CASE");
                  }
                  chomp(rest); 
               }
               else if (strcmp(relation,"!"))
               {  if (!rest) {map_errno = MAP_IFINV; goto ext; }
                  if (rest[0] == '"')                     /* right side is quoted str */
                  {  if (p=(char *)strtok_r(rest+1,"\"",&rest)){strcpy(right_side,p);quoted=1;}
                     else { map_errno = MAP_IFINV; goto ext; }    
                  }
                  else                                           /* Right side is tag */ 
                  {  if (p=(char *)strtok_r(rest," ",&rest))strcpy(right_side,p);
                     else { map_errno = MAP_IFINV; goto ext; } 
                  }
               }
               else CLEAR(right_side);    /* "!" relation has no right side arg */
/*------------------------------------------------------------------------------ && || */
               if (p=(char *)strtok_r(rest," ",&rest)) strcpy(and_or,p);
/*----------------------------------------------------------------- Create "if" struct */
               pmi = malloc_pmi();
               if (!pmi) { map_errno = MAP_MALLOC; goto ext; }
/*----------------------------- Left side must be a defined tag; We store it's pointer */
               for (pmtag2 = pmap->pmtag; pmtag2; pmtag2 = pmtag2->pmtag)
                  if (STR_EQ(left_side,pmtag2->name)) break;
               if (!pmtag2) { map_errno = MAP_IFTAG; goto ext; }
               pmi->left_pmtag = pmtag2;
/*---------------------------------------------------------------------------------------
 * Relations: == != > < >= < <= (and also => and =<) are mapped to lt, eq, and gt to 
 * indicate which one's of "less than", "equal to", and "greater than" are "TRUE" for 
 * this relation.  =~ and !~ are mapped only to eq.
 *-------------------------------------------------------------------------------------*/
               pmi->exists = 1; pmi->lt = pmi->eq = pmi->gt = 0;
               if (STR_EQ(relation,"==")) { pmi->eq = 1; } else
               if (STR_EQ(relation,"!=")) { pmi->lt = pmi->gt = 1; } else
               if (STR_EQ(relation,">"))  { pmi->gt = 1; } else
               if (STR_EQ(relation,"<"))  { pmi->lt = 1; } else
               if (STR_EQ(relation,">=") || STR_EQ(relation,"=>")) { pmi->gt = pmi->eq = 1; } else
               if (STR_EQ(relation,"<=") || STR_EQ(relation,"=<")) { pmi->lt = pmi->eq = 1; } else
               if (STR_EQ(relation,"=~")) { pmi->eq = 1; } else
               if (STR_EQ(relation,"!~")) { pmi->eq = 0; } else
               if (STR_EQ(relation,"!")) { pmi->exists = 0; } else
               { map_errno = MAP_IFREL; goto ext; }
/*---------------------------------------------------------------------------------------
 * Right side is a tag name or a double quoted string constant, which is compiled as a
 * regular expression for =~ and !~.  There is no right side for the "!" (not) operator.
 *-------------------------------------------------------------------------------------*/
               if (strchr(relation,'~'))
               {  pmi->right_exp = malloc(sizeof(regex_t));
                  memset(pmi->right_exp,0,sizeof(regex_t));
                  if (ret=regcomp(pmi->right_exp, right_side, REG_EXTENDED | ignore_case | REG_NOSUB))
                  {  map_errno = MAP_IEXP;
                     regerror(ret, pmi->right_exp, em, sizeof(em));
                     dbg(em);
                     goto ext;
                  }
               }
               else if (strcmp(relation,"!"))
               {  if (quoted) STR(pmi->right_side,right_side)
                  else
                  {  for (pmtag2 = pmap->pmtag; pmtag2; pmtag2 = pmtag2->pmtag)
                        if (STR_EQ(right_side,pmtag2->name)) break;
                     if (!pmtag2) { map_errno = MAP_IFTAG; goto ext; }
                     pmi->right_pmtag = pmtag2;
                  }
               }
/*---------------------------------------------------------------------------------------
 * Logical is either && or ||, except for last (or maybe only) if line, which is blank
 *-------------------------------------------------------------------------------------*/
               if (STR_EQ(and_or,"&&")) { pmi->logical = 1;  } else
               if (STR_EQ(and_or,"||")) { pmi->logical = 0;  } else
               if (!strlen(and_or))     { state = map_ifovr; } else
               { map_errno = MAP_IFREL; goto ext; }
/*---------------------------------------------------------------------------------------
 * Valid IF line, add it to the mapping
 *-------------------------------------------------------------------------------------*/
               valid_if = 1;
               if (!pmm->pmi) pmm->pmi = pmi; else
               {  for (pmi2 = pmm->pmi; pmi2->pmi; pmi2 = pmi2->pmi) ;
                  pmi2->pmi = pmi;
               }           
               pmi = NULL;
            }
            break;  
/****************************************************************************************
 * This state is for when we have found the last IF line (the one without a logical).
 * The next line must be "THEN:".
 ***************************************************************************************/
         case map_ifovr:
            if (strcasecmp(a,"THEN:")) { map_errno = MAP_NOTHEN; goto ext; }
            state = map_then;
            break;
/****************************************************************************************
 * Processing lines in the THEN: section.  These must be of the form:
 *    TAG = [TAG|"string"]
 * The left hand tag cannot be group 0000, which are the non-tag items, as we cannot set
 * them in the DICOM header.
 * We know we're done when we hit eof or the next header string, but when we do, we need
 * to store the map we are working on now and start a new one.
 *---------------------------------------------------------------------------------------
 * Alternatively, the THEN: section may contain a single line of the form:
 *    KEYWORD     - where KEYWORD matches one of the keyword entries in mapflags
 * If this is the case, the mapping returns with the corresponding flag set to 1.      
 * The calling program is responsible for accessing the flag and taking appropriate
 * action.
 ***************************************************************************************/
         case map_then:
/*------------------------------------------------------------------- next header line */
            if (is_header(a))
/*--------------------------------------- store current map if valid, error out if not */
            {  if (pmm && valid_if && valid_then)
               {  valid_map = 1;
                  if (!pmap->pmm) pmap->pmm = pmm; else
                  {  for (pmm2 = pmap->pmm; pmm2->pmm; pmm2 = pmm2->pmm) ;
                     pmm2->pmm = pmm;
                  }
                  pmm = NULL;
/*--------------------------------------------------------------------- set up new map */
                  pmm = malloc_pmm();
                  if (!pmm) { map_errno = MAP_MALLOC; goto ext; }
                  p = strtok(a+1,"]");
                  STR(pmm->name,p)
                  STR(pmm->log_name,def_log_name)
                  pmm->log_level = 1;
                  state = map_header;
                  valid_if = valid_then = 0;
                  break;
               }
               else { map_errno = MAP_INVMAP; goto ext; }
            }
/*---------------------------------------------------------------------------------------
 * special case, first and only line in in then section is keywords from mapflags
 *-------------------------------------------------------------------------------------*/
            if (valid_then == 0)
            {  valid_flag = 0;
               for (flg = 0; flg < max_flag; flg++)
               {  if (STR_EQ(a,FLAG.keyword))
                  {  pmt = malloc_pmt();
                     STR(pmt->right_side,FLAG.keyword)
                     valid_then = valid_flag = 1;
                     pmm->pmt = pmt; 
                     pmt = NULL;
                     break;
                  }
               }
               if (valid_flag) break;
            }
/*--------------------------- if flag keyword was found, no other then lines can exist */
            if (valid_flag) {map_errno = MAP_INVDEL; goto ext; }
            
/******************************************************************** normal Then line */
            {  char left_side[1024] = "\0", relation[1024] = "\0", 
                  right_side[1024] = "\0";
               char quoted = 0;
               char * rest = NULL;
/*---------------------------------------------------------------------------------------
 * parse line - Can't use sscanf, because the right side may be a quoted string  
 *-------------------------------------------------------------------------------------*/
               if (p=(char *)strtok_r(a," ",&rest)) strcpy(left_side,p);
               else {map_errno = MAP_THENINV; goto ext; }
               
               if (p=(char *)strtok_r(rest," ",&rest)) strcpy(relation,p);
               else {map_errno = MAP_THENINV; goto ext; }
               chomp(rest);
               
               if (!rest) {map_errno = MAP_THENINV; goto ext; }
               if (rest[0] == '"')                   /* right side is quoted str */
               {  if (rest[1] == '"') quoted = 1;
                  else if (p=(char *)strtok_r(rest+1,"\"",&rest)){strcpy(right_side,chomp(p));quoted=1;}
                  else { map_errno = MAP_IFINV; goto ext; }    
               }
               else                                           /* Right side is tag */ 
               {  if (p=(char *)strtok_r(rest," ",&rest))strcpy(right_side,p);
                  else { map_errno = MAP_IFINV; goto ext; } 
               }
/*----------------------------------------------------------------- Create then struct */
               pmt = malloc_pmt();
               if (!pmt) { map_errno = MAP_MALLOC; goto ext; }
/*----------------------------- Left side must be a defined tag; We store it's pointer */
               for (pmtag2 = pmap->pmtag; pmtag2; pmtag2 = pmtag2->pmtag)
                  if (STR_EQ(left_side,pmtag2->name) && pmtag2->group) break;
               if (!pmtag2) { map_errno = MAP_IFTAG; goto ext; }
               pmt->left_pmtag = pmtag2;
/*---------------------------------------------------------------------------------------
 * Relation must be =
 *-------------------------------------------------------------------------------------*/
               if (strcmp(relation,"=")) { map_errno = MAP_THENEQ; goto ext; }
/*---------------------------------------------------------------------------------------
 * Right side is either a tag name or a double quoted string constant
 *-------------------------------------------------------------------------------------*/
               if (quoted) STR(pmt->right_side,right_side)
               else
               {  for (pmtag2 = pmap->pmtag; pmtag2; pmtag2 = pmtag2->pmtag)
                     if (STR_EQ(right_side,pmtag2->name)) break;
                  if (!pmtag2) { map_errno = MAP_THENTAG; goto ext; }
                  pmt->right_pmtag = pmtag2;
               }
/*---------------------------------------------------------------------------------------
 * Valid IF line, add it to the mapping
 *-------------------------------------------------------------------------------------*/
               valid_then = 1;
               if (!pmm->pmt) pmm->pmt = pmt; else
               {  for (pmt2 = pmm->pmt; pmt2->pmt; pmt2 = pmt2->pmt) ;
                  pmt2->pmt = pmt;
               }           
               pmt = NULL;
            }
            break;  
            
/****************************************************************************************
 * This mapping is turned off.  Skip all lines until you find the next mapping, or eof
 ***************************************************************************************/
         case map_off:
/*------------------------------------------------------------------- next header line */
            if (is_header(a))
/*--------------------------------------------------------------------- set up new map */
            {  pmm = malloc_pmm();
               if (!pmm) { map_errno = MAP_MALLOC; goto ext; }
               p = strtok(a+1,"]");
               STR(pmm->name,p)
               STR(pmm->log_name,def_log_name)
               pmm->log_level = 1;
               state = map_header;
               valid_if = valid_then = 0;
            }
            break;
            
/****************************************************************************************
 * EOF or error on .map file.  If error, exit, otherwise there should be a valid mapping
 * to store.
 ***************************************************************************************/
         case map_eof:
            if (ferror(map_fp)) { map_errno = MAP_FE; goto ext; }  /* error */
/*-------------------------------------------------------------------- store valid map */
            if (pmm && valid_if && valid_then)
            {  valid_map = 1;
               if (!pmap->pmm) pmap->pmm = pmm; else
               {  for (pmm2 = pmap->pmm; pmm2->pmm; pmm2 = pmm2->pmm) ;
                  pmm2->pmm = pmm;
               }
               pmm = NULL;
            }
            if (valid_map) map_errno = MAP_OK;
            else map_errno = MAP_NOMAPS;
            goto ext;
/*************************************************************** should never get here */
         default:
            map_errno = MAP_INT; goto ext;
      }
   }
/*---------------------------------------------------------------------------------------
 * This is the common exit point for the Setup routine.  If the error value is not "OK",
 * clear any memory which has been allocated to the map. Return map handle, or NULL.
 *-------------------------------------------------------------------------------------*/   
ext:
   if (map_fp) fclose(map_fp);
   FREE(map_pfn)
   FREE_PMTAG(pmtag)
   FREE_PMI(pmi)
   FREE_PMT(pmt)
   FREE_PMLT(pmlt)
   FREE_PMM(pmm)
   if (map_errno != MAP_OK && pmap)
   {  if (pmap->debug_fp) PrtMapErrorString(pmap->debug_fp,"mapSetup exit");
      FREE_PMAP(pmap); 
   }
   if (def_log_name)
   {  char msg[5000];
      if (pmap)
      {  strcpy(msg,"Maps initialized:\n");
         for (pmm2 = pmap->pmm; pmm2; pmm2 = pmm2->pmm)
         {  sprintf(em,"   %-20s %-10s %-10s\n", pmm2->name, pmm2->log_name ? pmm2->log_name : "",
               pmm2->category ? pmm2->category : "");
            strcat(msg,em);
         }
      }
      else sprintf(msg,"Map initialization failed: %s\n", GetMapErrorString());
      logMsg(def_log_name,msg);
      FREE(def_log_name)
   }
   return (pmap);
}
     
/****************************************************************************************
 * Set value of tag which is in "group 0", i.e., not in the DICOM header.  If first is
 * non-zero, the study_value and image_value will be set, otherwise just the image_value
 * will be set.
 ***************************************************************************************/
int mapTag(PMAP pmap, unsigned int element, char first, char * value)
{  PMTAG pmtag;
   map_errno = MAP_OK;
   if (!pmap)  { map_errno = MAP_NULLMAP; return MAP_R_FAIL; }
   for (pmtag = pmap->pmtag; pmtag; pmtag = pmtag->pmtag)
   {  if (!pmtag->group && pmtag->element == element)
      {  if (first) STR(pmtag->study_value,value)
         STR(pmtag->image_value,value)
         return MAP_R_OK;
      }
   }
   map_errno = MAP_TAGNOF;
   return MAP_R_FAIL;
}
/****************************************************************************************
 * process maps for an object
 ***************************************************************************************/
int mapProcess(PMAP pmap, DCM_OBJECT * obj, char first, char * map_category, char * map_name, int * changes)
{  PMM pmm;
   PMI pmi;
   PMT pmt;
   PMLT pmlt;
   char skip, *a = NULL, *b = NULL;
   int i,t, map_found = 0, ret;
   char * msg = NULL;

   map_errno = MAP_OK;
   *changes = 0;              /* returns 0=no changes made, 1=changes made */

   if (!pmap) { map_errno = MAP_NULLMAP; return MAP_R_FAIL; }
   if (!obj)  { map_errno = MAP_NULLOBJ; return MAP_R_FAIL; }

   ClearTags(pmap, first);
/*---------------------------------------------------------------------------------------
 * On the first image, all the keyword map flags are turned off, which is their default.
 * Only if a mapping exists which fits the image and turns them on do they have an on val
 *-------------------------------------------------------------------------------------*/
   if (first) for (flg = 0; flg < max_flag; flg++) FLAG.flag = 0;
/*--------------------------------------- Loop through all maps in this pmap structure */
   for (pmm = pmap->pmm; pmm; pmm = pmm->pmm)
/*---------------------------- If looking for specific map or category, match its name */
   {  if (map_category && strcasecmp(map_category,pmm->category)) continue;
      if (map_name && strcasecmp(map_name,pmm->name)) continue;
      map_found = 1;
/*------------------ Evaluation block (evaluate for first_object or if at image level) */
      if (first || pmm->level)
/*---------------------------------------------------- Loop through if lines for this map
 * The logical evaluation is done by dividing the IF lines into groups, with the last 
 * line of each group having an OR (||) logical or being the last IF line.  If all the
 * IF lines in any group are TRUE, the map evaluates TRUE and we do not process any of
 * the other IF lines.  If any line in the group evaluates FALSE, the group is FALSE and 
 * we skip to the next group. skip tells us we have found a FALSE line, and are skipping
 * the rest of the lines in that group.
 *-------------------------------------------------------------------------------------*/
      {  for (skip=0, pmi = pmm->pmi; pmi; pmi = pmi->pmi)    /* Step through IF lines */
         {  if (skip)                                /* skipping through current group */
            {  if (pmi->logical) continue;
               if (pmi->pmi) skip=0;
               continue;
            }
/*-------------- pull left tag, and right tag or string constant to evaluate this line */
            STR(a,GetTag(pmi->left_pmtag,obj, first, pmm->level))
            if (pmi->right_pmtag) STR(b,GetTag(pmi->right_pmtag,obj, first, pmm->level))
            else if(pmi->right_side) STR(b,pmi->right_side)
            else FREE(b)
/*-------------------------------------- Evaluation: t=1 line is true, 0 line is false */
            chomp(a); chomp(b);
            t = 0;
            if (!a) 
            {  if (!pmi->exists) t=1;
            }
            else if (b)
            {  i = strcmp(a,b);
               if (i < 0  && pmi->lt) t=1;
               if (i == 0 && pmi->eq) t=1;
               if (i > 0  && pmi->gt) t=1;
            }
            else if (pmi->right_exp) 
            {  ret = regexec(pmi->right_exp, a, 0, NULL, 0); 
               if (ret == 0 && pmi->eq) t=1;
               else if (ret == REG_NOMATCH && !pmi->eq) t=1;
               else 
               {  regerror(ret, pmi->right_exp, em, sizeof(em));
                  dbg(em);
               }
            }
/*---------- If line is true and we are at end of a group, we're done, otherwise loop */
            if (t)
            {  if (!pmi->logical || !pmi->pmi) break;
               continue;
            }
/*-------------------------------------------- Line wasn't true, so we're in skip mode */
            if (pmi->logical || !pmi->pmi) skip = 1;
         }
/*------- done evaluating.  If We're in skip mode, we never got a hit.  If not, we did */
         pmm->hit = !skip;
      }
/*---------------------------------- Allocate memory for log message if we are logging */
      if ((first || pmm->level) && pmm->log_name && (pmm->log_level == 2 || (pmm->log_level == 1 && pmm->hit)))
      {  msg = malloc(4096);
         sprintf(msg,"%s %s\n\n", pmm->log_msg ? pmm->log_msg : pmm->name, 
               (pmm->hit && pmm->log_level == 2) ? " ** Match **" : "");
         for (pmlt = pmm->pmlt; pmlt; pmlt = pmlt->pmlt)
         {  STR(a,GetTag(pmlt->pmtag, obj, first, pmm->level));
            STR(b,pmlt->pmtag->e.description)
            sprintf(em, "   %-30s - %s\n", b ? b : pmlt->pmtag->name, a ? a : "NOF"),
            strcat(msg,em);
         }
         if (pmm->hit) strcat(msg,"\nChanges:\n");
      }
/*---------------------------------------------------------------------------------------
 * If we are dealing with a keyword flag in the THEN section, we ALWAYS set the flag,
 * hit or miss, and log hits.
 *-------------------------------------------------------------------------------------*/
      if (!pmm->pmt->left_pmtag)
      {  for(flg = 0; flg < max_flag; flg++)
         {  if (STR_EQ(pmm->pmt->right_side,FLAG.keyword))
            {  FLAG.flag = pmm->hit;
               if (msg && pmm->hit)
               {  sprintf(em,"   %s\n", FLAG.log_msg); 
                  strcat(msg,em);
               }
               break;
            }
         }
      }
/*---------------------------------------------------------------------------------------
 * If we have tag modifications in the THEN block, we make them, but only if we hit.
 *-------------------------------------------------------------------------------------*/
      if (pmm->hit && pmm->pmt->left_pmtag)
/*------------------------------------------ Step through THEN lines, replacing values */
      {  for (pmt = pmm->pmt; pmt; pmt = pmt->pmt)
         {  if (msg) STR(a,GetTag(pmt->left_pmtag,obj, first, 1))
            if (pmt->right_pmtag) STR(b,GetTag(pmt->right_pmtag,obj, first, pmm->level))
            else STR(b,pmt->right_side)
            SetTag(pmt->left_pmtag,obj,b);
            *changes = 1;
            if (msg) 
            {  sprintf(em,"   %-30s: %s => %s\n", pmt->left_pmtag->e.description, a ? a: "NOF", b ? b : "NOF");
               strcat(msg,em);
            }
         }
      }
/*-------------------------------------------------------- log message, if we have one */
      if (msg) logMsg(pmm->log_name, msg);
      FREE(msg)
   } /* EO process a map structure */
   
   if ((map_category || map_name) && !map_found) map_errno = MAP_NF; 
   return MAP_R_OK; 
}
        
/****************************************************************************************
 * Get a tag value for an item. If this is a "first" image or the map is evaluated at 
 * the image level, we get the value from the obj - otherwise we return the value in the
 * pmtag struct. If Tag is not in object, return NULL.  This is an internal routine, so
 * we don't validate parameters. Group 0 tags always return the value from the struct.
 ***************************************************************************************/
static char * GetTag(PMTAG pmtag, DCM_OBJECT * obj, char first, char level)
/*---------------------------------------------------------- Get Element info from obj */
{  if (pmtag->group && (first || level))
   {  cond = DCM_GetElement(&obj, pmtag->e.tag, &(pmtag->e));
      if (cond == DCM_ELEMENTNOTFOUND) return NULL;                     /* Tag not in obj */
      if (cond != DCM_NORMAL) { map_errno = MAP_TAG; return NULL; }     /* Error */
/*------------------------------------------------------ Allocate memory for tag value */
      FREE(pmtag->e.d.string)
      pmtag->e.d.string = malloc(pmtag->e.length+1);
      if (!(pmtag->e.d.string)) { map_errno = MAP_MALLOC; return NULL; }
      memset(pmtag->e.d.string,0,pmtag->e.length+1);

      if (first)
      {  FREE(pmtag->study_value)
         pmtag->study_value = malloc(pmtag->e.length+1);
         if (!(pmtag->study_value)) { map_errno = MAP_MALLOC; return NULL; }
         memset(pmtag->study_value,0,pmtag->e.length+1);
      }
   
      FREE(pmtag->image_value)
      pmtag->image_value = malloc(pmtag->e.length+1);
      if (!(pmtag->image_value)) { map_errno = MAP_MALLOC; return NULL; }
      memset(pmtag->image_value,0,pmtag->e.length+1);
/*---------------------------------------------------------------------- Get Tag Value */
      ctx = NULL;
      cond = DCM_GetElementValue(&obj, &(pmtag->e), &ul, &ctx);
      if (cond != DCM_NORMAL) 
      {  map_errno = MAP_INT; 
         FREE(pmtag->image_value)
         FREE(pmtag->e.d.string)
         if (first) FREE(pmtag->study_value)
         return NULL; 
      }
      if (first) memmove(pmtag->study_value,pmtag->e.d.string,pmtag->e.length);
      memmove(pmtag->image_value,pmtag->e.d.string,pmtag->e.length);
   }
   if (level) return(pmtag->image_value);
   else return(pmtag->study_value);
}
     
/****************************************************************************************
 * Set a tag value in an object.
 ***************************************************************************************/
static void SetTag(PMTAG pmtag, DCM_OBJECT * obj, char * new_value)
{  STR(pmtag->e.d.string,new_value)
   if (!(pmtag->e.d.string)) { map_errno = MAP_MALLOC; return; }
   pmtag->e.length = strlen(new_value);
   DCM_RemoveElement(&obj, pmtag->e.tag);
   cond = DCM_AddElement(&obj, &(pmtag->e));
   if (cond != DCM_NORMAL) { COND_DumpConditions(); map_errno = MAP_SET; }
   return;
}

/*---------------------------------------------------------------------------------------
 * Clear all tag values for a mapping (Group 0 tags are not cleared)
 *-------------------------------------------------------------------------------------*/
static void ClearTags(PMAP pmap, char first)
{  PMTAG pmtag;
   for (pmtag = pmap->pmtag; pmtag; pmtag = pmtag->pmtag)
   {  if (pmtag->group)
      {  FREE(pmtag->e.d.string)
         if (first) FREE(pmtag->study_value)
         FREE(pmtag->image_value)
      }
   }
}

/*------------------------------------------------------------------- Return map value */
int mapGetFlag(int flg)
{  if ((flg >= 0) && (flg < max_flag)) {map_errno=MAP_OK; return FLAG.flag;}
   else {map_errno=MAP_INVFLG; return -1;}
}

/*-------------------------------------- Courtesy calls, map into mapProcess */
int mapFirstObject(PMAP pmap, DCM_OBJECT * obj, int * changes) {return mapProcess(pmap,obj,1,NULL,NULL, changes);}
int mapObject(PMAP pmap, DCM_OBJECT * obj, int * changes) {return mapProcess(pmap,obj,0,NULL,NULL, changes);}

/*------------------------------------------------------------------------ Free up map */
int mapFree(PMAP pmap) 
{  FREE_PMAP(pmap) 
   map_errno = MAP_OK;
   return MAP_R_OK;
}

/****************************************************************************************
 * return error message string for log_errno, or print to stdout
 ***************************************************************************************/
char * GetMapErrorString(void)
{	if (map_errno < max_e_msg) return e_msg[map_errno];
	else return no_msg;
}
void PrtMapErrorString(FILE * file, char * prefix)
{	if (!file) file = stdout;
   fprintf(file, "%s: %s\n", prefix, GetMapErrorString());
	return;
}

/*--------------------------------------------------------------------------------------
 * If argument is a map header line (format [name] starting in col 1) return non-zero,
 * otherwise return 0. a is unchanged. name must be non-zero length, non-blank string.
 *------------------------------------------------------------------------------------*/
static int is_header(char * a)
{  char * b = NULL, *p = NULL;
   int i;
/*------------------------------------------------------------- Must be format [name] */
   if (a[0] != '[') return 0;
   if (!strchr(a,']')) return 0;
   STR(b,a);
   p = strtok(b+1,"]");
   chomp(p);
   i = strlen(p);
   FREE(b)
   return i;
}

/*---------------------- Remove leading and trailing spaces and new lines from string */
static char * chomp(char * a)
{  int i;
   if (!a) return a;
   i = strspn(a," ");
   memmove(a,a+i,strlen(a+i)+1);
   while (LCHR(a) == ' ' || LCHR(a) == '\n') LCHR(a) = '\0';
   return a;
}

/*---------------------------------------------------------------------------------------
 * Routines to free up memory malloc'd for structs.  Most are recursive.
 *-------------------------------------------------------------------------------------*/
static void free_pmtag(PMTAG pmtag)
{  if (pmtag)
   {  FREE_PMTAG(pmtag->pmtag)
      FREE(pmtag->name)
      FREE(pmtag->e.d.string)
      FREE(pmtag->study_value)
      FREE(pmtag->image_value)
      FREE(pmtag)
   }
}
   
static void free_pmi(PMI pmi)
{  if (pmi)
   {  FREE_PMI(pmi->pmi)
      FREE_PMTAG(pmi->left_pmtag)
      FREE_PMTAG(pmi->right_pmtag)
      FREE(pmi->right_side)
      if (pmi->right_exp)
      {  regfree(pmi->right_exp);
         FREE(pmi->right_exp)
      }
      FREE(pmi)
   }
}

static void free_pmt(PMT pmt)
{  if (pmt)
   {  FREE_PMT(pmt->pmt)
      FREE_PMTAG(pmt->left_pmtag)
      FREE_PMTAG(pmt->right_pmtag)
      FREE(pmt->right_side)
      FREE(pmt)
   }
}

static void free_pmlt(PMLT pmlt)
{  if (pmlt)
   {  FREE_PMLT(pmlt->pmlt)
      FREE_PMTAG(pmlt->pmtag)
   }
}

static void free_pmm(PMM pmm)
{  if (pmm) 
   {  FREE_PMM(pmm->pmm)
      FREE(pmm->name)
      FREE(pmm->log_name)
      FREE(pmm->log_msg)
      FREE(pmm->category)
      FREE_PMLT(pmm->pmlt)
      FREE_PMI(pmm->pmi)
      FREE_PMT(pmm->pmt)
      FREE(pmm)
   }
}

static void free_pmap(PMAP pmap)
{  if (pmap)
   {  if (pmap->debug_fp) fclose(pmap->debug_fp);
      FREE(pmap->inipfn)
      FREE_PMTAG(pmap->pmtag)
      FREE_PMM(pmap->pmm)
      FREE(pmap)
   }
}


/*---------------------------------------------------------------------------------------
 * Routines to allocate and clear memory for mapping structs
 *-------------------------------------------------------------------------------------*/
static PMTAG malloc_pmtag()
{  PMTAG pmtag = NULL;
   if (pmtag = malloc(SMTAG))
   {  pmtag->pmtag = NULL;
      pmtag->name = NULL;
      pmtag->group = 0;
      pmtag->element = 0;
         pmtag->e.tag = 0;
         pmtag->e.representation = DCM_SH;
         strcpy(pmtag->e.description,"\0");
         pmtag->e.multiplicity = 0;
         pmtag->e.length = 0;
         pmtag->e.d.string = NULL;
      pmtag->study_value = NULL;
      pmtag->image_value = NULL;
   }
   return (pmtag);
}

static PMI   malloc_pmi()
{  PMI pmi = NULL;
   if (pmi = malloc(SMI))
   {  pmi->pmi = NULL;
      pmi->left_pmtag = NULL;
      pmi->right_pmtag = NULL;
      pmi->right_side = NULL;
      pmi->right_exp = NULL;
      pmi->lt = 0;
      pmi->eq = 0;
      pmi->gt = 0;
      pmi->exists = 0;
      pmi->logical = 0;
   }
   return (pmi);
}

static PMT   malloc_pmt()
{  PMT pmt = NULL;
   if (pmt = malloc(SMT))
   {  pmt->pmt = NULL;
      pmt->left_pmtag = NULL;
      pmt->right_pmtag = NULL;
      pmt->right_side = NULL;
   }
   return (pmt);
}

static PMLT  malloc_pmlt()
{  PMLT pmlt = NULL;
   if (pmlt = malloc(SMLT))
   {  pmlt->pmlt = NULL;
      pmlt->pmtag = NULL;
   }
   return (pmlt);
}

static PMM   malloc_pmm()
{  PMM pmm = NULL;
   if (pmm = malloc(SMM))
   {  pmm->pmm = NULL;
      pmm->name = NULL;
      pmm->level = 0;
      pmm->hit = 0;
      pmm->log_name = NULL;
      pmm->log_level = 0;
      pmm->log_msg = NULL;
      pmm->category = NULL;
      pmm->pmlt = NULL;
      pmm->pmi = NULL;
      pmm->pmt = NULL;
   }
   return (pmm);
}

static PMAP  malloc_pmap()
{  PMAP pmap = NULL;
   if (pmap = malloc(SMAP))
   {  pmap->debug_fp = NULL;
      pmap->inipfn = NULL;
      pmap->pmtag = NULL;
      pmap->pmm = NULL;
   }
   return (pmap);
}
