#ifndef _MAP_H_
#define _MAP_H_
/****************************************************************************************
 * mircs_mapping.h                                                              R Moulton
 * Header file for MIR Computer section DICOM header mapping functions.
 ***************************************************************************************/
#include <regex.h>

/*************************** structure definitions *************************************/

/*---------------------------------------------------------------------------------------
 * Map Tag structure.  One per Tag entry in the mapping file.
 *-------------------------------------------------------------------------------------*/
struct _map_tag_ {
   struct _map_tag_ * pmtag;
   char * name;                            /* Name of tag, from Tag line */
   unsigned int group;                     /* DICOM tag group & element from tag line */
   unsigned int element;
   DCM_ELEMENT e;
   char * study_value;                           /* Value of tag */
   char * image_value;

};
typedef struct _map_tag_ * PMTAG;
#define SMTAG sizeof(struct _map_tag_)

/*---------------------------------------------------------------------------------------
 * Map If structure.  One per IF line in the mapping.
 *-------------------------------------------------------------------------------------*/
struct _map_if_ {
   struct _map_if_ * pmi;                  /* ptr to next IF line, if any */
   PMTAG left_pmtag;                       /* left side tag pointer */
   PMTAG right_pmtag;                      /* right side tag pointer, if any */
   char * right_side;                      /* right side string, if any */
   regex_t * right_exp;                    /* right side regular expression, if any */
   char lt;                                /* is left < right "TRUE" */
   char eq;                                /* is left = right or left matches right "TRUE" */
   char gt;                                /* is left > right "TRUE" */
   char exists;                            /* is left exists in object "TRUE" */
   char logical;                           /* 1 = &&, 0 = || (pmi is NULL on last line) */ 
};
typedef struct _map_if_ * PMI;
#define SMI sizeof(struct _map_if_)

/*---------------------------------------------------------------------------------------
 * Map Then structure.  One per THEN line in the mapping.
 *-------------------------------------------------------------------------------------*/
struct _map_then_ {
   struct _map_then_ * pmt;                /* ptr to next IF line, if any */
   PMTAG left_pmtag;                       /* left side tag pointer */
   PMTAG right_pmtag;                      /* right side tag pointer, if any */
   char * right_side;                      /* right side string, if any */

};
typedef struct _map_then_ * PMT;
#define SMT sizeof(struct _map_then_)

/*---------------------------------------------------------------------------------------
 * Map Log tag structure.  One per LogTags entry in the mapping header.
 *-------------------------------------------------------------------------------------*/
struct _map_logtag_ {
   struct _map_logtag_ * pmlt;
   PMTAG pmtag;
};
typedef struct _map_logtag_ * PMLT;
#define SMLT sizeof(struct _map_logtag_)

/*---------------------------------------------------------------------------------------
 * Map structure.  One per active mapping in the file
 *-------------------------------------------------------------------------------------*/
struct _map_map_ {
   struct _map_map_ * pmm;   
   char * name;                            /* name of mapping, from [name] line */
   char level;                             /* evaluate at 0=study level, 1=image level */
   char hit;                               /* Did IF lines evaluate TRUE? */
   char * log_name;                        /* log name for map, or NULL */
   char log_level;                         /* logging level */
   char * log_msg;                         /* log message */
   char * category;                        /* Map category */
   PMLT  pmlt;                             /* Log tags */
   PMI   pmi;                              /* If lines */
   PMT   pmt;                              /* Then lines */
   
};
typedef struct _map_map_ * PMM;
#define SMM sizeof(struct _map_map_)
   
/*---------------------------------------------------------------------------------------
 * Overall map struct; pointer to this is passed from routines. All the sub structures 
 * are of the pointer link variety 
 * ------------------------------------------------------------------------------------*/
struct _map_whole_ {
   FILE * debug_fp;
   char * inipfn;
   PMTAG  pmtag;                            /* Tags list */ 
   PMM    pmm;                              /* Maps list */    
};
typedef struct _map_whole_ * PMAP;
#define SMAP sizeof(struct _map_whole_)

/*------------------------------------------------------------------------ error codes */
unsigned int map_errno;

#define MAP_R_OK   0
#define MAP_R_FAIL 1

#define MAP_OK        0      /* This is OK */
#define MAP_NOINI     1
#define MAP_ININOF    2
#define MAP_INIREAD   3
#define MAP_NOINISEC  4
#define MAP_SECNOF    5
#define MAP_INIERR    6
#define MAP_NOTON     7      /* This is OK */
#define MAP_MALLOC    8
#define MAP_MAPNOF    9
#define MAP_MAPREAD  10
#define MAP_MAPOPEN  11
#define MAP_INT      12
#define MAP_NOTAGS   13
#define MAP_INVTAG   14
#define MAP_TAGNAME  15
#define MAP_TAGDUP   16
#define MAP_TAGGRP   17
#define MAP_TAGELM   18
#define MAP_HDRINV   19
#define MAP_HDRLTAG  20
#define MAP_IFINV    21
#define MAP_IFTAG    22
#define MAP_IFREL    23
#define MAP_IFLOG    24
#define MAP_NOTHEN   25
#define MAP_THENINV  26
#define MAP_THENTAG  27
#define MAP_THENEQ   28
#define MAP_INVMAP   29
#define MAP_FE       30
#define MAP_NOMAPS   31
#define MAP_DICOMTAG 32
#define MAP_TAG      33
#define MAP_SET      34
#define MAP_NULLMAP  35
#define MAP_NULLOBJ  36
#define MAP_NF       37
#define MAP_TAGNOF   38
#define MAP_INVDEL   39
#define MAP_INVFLG   40
#define MAP_IEXP     41

/****************************** Map action flags ***************************************/
#define MAP_FLAG_DELETE_IMAGE 0
#define MAP_FLAG_NO_IDQC_HOLD 1

/*************************** function prototypes ***************************************/

/*---------------------------------------------------------------------------------------
 * Does setup for a mapping set based on information in "section" of .ini "inipfn", and 
 * .map file.  Sample .ini file entries:
 * 
 * [Mapping]
 * Status=OFF                       # Is mapping turned on? (YES|ON)
 * Debug=NULL   
 * LogName=MAPPING                  # Default logname (can be overidden by mapping rule)
 * MapFile=                         # Mapping rules file (def is mircs.map in working dir)
 * #----------------------------------------------------------------- Default mapping log
 * [LOG_MAPPING]
 * Status=ON
 * Dir=${LOG}
 * Prefix=MAP
 * Period=DAILY
 * Retention=30
 * #------------------------------------ Level 1 = log changes, Level 2 = log all studies
 * Level=1
 *
 * The .map file contains the information about the mapping(s).  Sample:
 *
 * #*************************************************************************************
 * # Simple storage mapping file.  Each Section corresponds to a mapping. At beginning, 
 * # define DICOM tag names, which are then referenced in the mappings. I try to use the
 * # same name as in the CTN software, but that is not required.
 * #*************************************************************************************
 * Tag IDINSTITUTIONNAME   (0008,0080) 
 * Tag IDSTATIONNAME       (0008,1010)
 * Tag IDMODALITY          (0008,0060)
 * Tag IDSTUDYDATE         (0008,0020)
 * Tag IDSTUDYTIME         (0008,0030)
 * Tag PATNAME             (0010,0010)
 * Tag PATID               (0010,0020)
 * Tag IDACCESSIONNUMBER   (0008,0050)
 * Tag RELSTUDYINSTANCEUID (0020,000D)
 * #*************************************************************************************
 * # OEC 9800 bad modality map to XA
 * #*************************************************************************************
 * [OEC_9800_MODALITY]
 * #--------------------------------------------------- Must be ON or YES to run this map
 * Status=ON
 * #------------------------------- Category, allows selection of groups of maps to apply
 * Category=IN-1
 * #------------------------------------------ Evaluate at STUDY level, or for each IMAGE
 * Level=STUDY
 * #--------------------------- Log name and level (1=chgs,2=all) or OFF|NO if no logging
 * LogName=MAP
 * LogLevel=1
 * LogMsg=OEC 9800 XA modality fix
 * LogTags=PATNAME,PATID,IDACCESSIONNUMBER,IDSTUDYDATE,IDSTUDYTIME,RELSTUDYINSTANCEUID
 * IF:
 *    IDINSTITUTIONNAME == "BARNES ERPC1" &&
 *    IDSTATIONNAME     == "OEC 9800"     &&
 *    IDMODALITY        != "XA"
 * THEN:
 *    IDMODALITY         =  "XA"
 *
 * returns a pmap pointer for the structures which control the mapping, which is then
 * passed to the other routines, and cleared using mapFree when you are done with it.
 * Returns NULL if mapping is turned off or on error. map-errno indicates error status.
 * A value of MAP_OK means the load worked, MAP_NOTON means that the status of the map is
 * not ON - others are errors.
 *-------------------------------------------------------------------------------------*/
PMAP mapSetup(char * section, char * inipfn);
/*---------------------------------------------------------------------------------------
 * Set value for a group 0 tag (not in DICOM header)
 *-------------------------------------------------------------------------------------*/
int mapTag(PMAP pmap, unsigned int element, char first_object, char * value);
/*---------------------------------------------------------------------------------------
 * Map a first object.  All maps in pmap are evaluated, changes are made and logged.
 *-------------------------------------------------------------------------------------*/
int mapFirstObject(PMAP pmap, DCM_OBJECT * obj, int * changes);
/*---------------------------------------------------------------------------------------
 * Map an other than first object. Image level maps are evaluated.  Changes are made and
 * logged.
 *-------------------------------------------------------------------------------------*/
int mapObject(PMAP pmap, DCM_OBJECT * obj, int * changes);
/*---------------------------------------------------------------------------------------
 * Process specific map for object. first_object <> 0 if this is first object in study,
 * map_name is name of specific map in .map file to process, or NULL to process all maps.
 * changes returns 0=no changes made, 1=changes made
 *-------------------------------------------------------------------------------------*/
int mapProcess(PMAP pmap, DCM_OBJECT * obj, char first_object, char * map_category, 
      char * map_name, int * changes);
/*-------------- Returns value of map flag (0 or 1) or -1 if invalid flag was selected */
int mapGetFlag(int mapFlag);
/*-------------------------------------- Releases memory assigned to a pmap structures */
int mapFree(PMAP pmap);

/*----------------------------------------- Standard library error processing routines */
char * GetMapErrorString(void);
void PrtMapErrorString(FILE * file, char * prefix);

#endif
