#!/bin/csh

#****************************************************************************
# Bounce.csh                                                        R Moulton
# This script provides a quick and relatively safe way to bounce or kill one
# of our cron scripts on a unix server.  Usage is: bounce scriptname, queue,
# [kill], where queue is one of the broadcast queues on that server (e.g. 
# bond1) which is to be bounced.  If the kill (or just k) parameter is 
# included, the lock file will not be deleted, and the script will not run 
# again until it is. Otherwise, bounce will clear the lock script, so that 
# the script will run again the next time it is invoked by cron.  If the
# parameter is killold (or just ko), Bounce is being called from an 
# invocation of the same script that is being bounced;  In this case, we are
# looking for multiple invocations and bouncing the oldest. The script
# functions by finding the invocation of the script with cron as its parent
# process and killing it, the killing any process which had that process as
# its parent, and repeating until no child process is found.  There are a
# number of checks to help avoid shooting oneself in the foot.
# Returns: 0 Bounced OK. 1 Error. > 1 Bounced OK, using killold; return value
# is number of incarnations of script that were running before bounce.
#----------------------------------------------------------------------------
# 11-01-01 R Moulton modified for old goldfinger configuration
# 12-05-01 R Moulton made general for different cron jobs; made recursive
# 12-24-01 R Moulton made lock file delete happen even if job not running
#          Added bounce entry in Script log file
# 5-8-2    R Moulton modified to use SNAME
# 2009 Aug 5 M.Suchko NOTE: this script fails to kill EchoTest.csh
#****************************************************************************
set old = 0
#----------------------------------------------------------------------------
# scripts are run under user spectrum, so you must be logged in as that
# to kill them
#----------------------------------------------------------------------------
if (`whoami` != spectrum) then
   echo Must be user spectrum to bounce scripts
   exit 1
endif

set SCRIPT = "$1"
if ("$SCRIPT" != Router && \
    "$SCRIPT" != Dealer && \
    "$SCRIPT" != EchoTest && \
    "$SCRIPT" != Pending && \
    "$SCRIPT" != Janitor) then
   echo "$SCRIPT not valid script name"
   exit 1
endif
set QUEUE  = "$2"
if ($SCRIPT == Dealer) set QUEUE = `echo "$QUEUE" | cut -d'.' -f2`
if ($SCRIPT == EchoTest) set QUEUE = ""

#-------------------- Validate that this script, queue are set up as cron job
crontab -l | egrep "$SCRIPT $QUEUE|$SCRIPT  $QUEUE" >& /dev/null
if ($status != 0) then
   echo "No job in crontab for: $SCRIPT $QUEUE"
   exit 1
endif

#------------------------------------------ Kill if param 3, otherwise bounce
set KILL = NO
if ("$3" == "k")       set KILL = YES
if ("$3" == "kill")    set KILL = YES
if ("$3" == "ko")      set KILL = OLD
if ("$3" == "killold") set KILL = OLD

set LOGS = $LOG

#------------------------------------------------------- Script log file name
set LOG = $LOGS/$SCRIPT.log
if ($SCRIPT == Dealer) then
   set LOG = $LOGS/Dealer.$SNAME$QUEUE.log
endif
#------------------------------------ Validate that expected lock file exists
set LOCK = $TMP/$SCRIPT.csh.lock
if ($SCRIPT == Dealer) then
   set LOCK = $LOCK.$SNAME$QUEUE
endif
if (!(-e $LOCK)) then
   echo script not running
   exit 1
endif
#----------------------------------------------- some temp files we will need
set T1 = $TMP/Bounce1.$QUEUE
set T2 = $TMP/Bounce2.$QUEUE
#------------------------------------list of running processes (newest first)
ps -ae -o etime,pid,ppid,user,args | grep spectrum | sort > $T1
#---------------------------------------------------- Now pull master process
cat $T1 | grep "$SCRIPT.csh $QUEUE" | grep " 1 spectrum" > $T2
if ($status != 0) then
   echo could not find script process
   if ($KILL != YES) then
      echo Lock file deleted
      rm $LOCK
   endif
   exit 1
endif
#-----------------------------------------------------------------------------
# If killing oldest, this subscript counts the number of lines in $T2 and
# returns in $T2 the last line (which is the oldest; see sort above)
#-----------------------------------------------------------------------------
if ($KILL == OLD) then
   cat $T2 | $BIN/BounceOld.csh > $T2
   @ old = $status 
   if ($old < 2) then
      echo "no older incarnation of $SCRIPT $QUEUE running"
      echo Lock file deleted
      rm $LOCK
      exit 1
   endif
endif
#----------------- Process repeatedly, killing script and all its descendants
while (1)
#---------------------------------------------- get PID and PPID of script job
   set PROCESS = `cat $T2`
   set PID = $PROCESS[2]
   if ($PID == "") break
   kill -9 $PID
   cat $T1 | grep " $PID spectrum" > $T2
   if ($status != 0) break
end
#-------------------------------------- Delete script lock file, if "bouncing"
if ($KILL == "YES") then
   echo script $Q killed 
   echo delete $LOCK to restart $SCRIPT $QUEUE
   echo "script $Q killed `date`" >> $LOG
   echo delete $LOCK to restart >> $LOG
else
   if (-e $LOCK) rm $LOCK
   echo $SCRIPT $QUEUE bounced
   echo will restart automatically
   echo "$SCRIPT $QUEUE bounced `date`" >> $LOG
   echo will restart automatically >> $LOG
endif
#------------------------- End script messages (Janitor doesn't have Messages)
if ($SCRIPT == Router) $BIN/Message Router.$SNAME.production $SNAME END_SCRIPT
if ($SCRIPT == Dealer) $BIN/Message Dealer.csh $SNAME$QUEUE END_SCRIPT
if ($SCRIPT == EchoTest) $BIN/Message EchoTest.csh ${SNAME}10 END_SCRIPT
if ($SCRIPT == Pending) $BIN/Message Pending.csh $SNAME END_SCRIPT

rm $T1
rm $T2
if ($KILL == OLD) exit $old
exit 0
