#!/bin/csh

#***************************************************************************
# DealStudy.csh               R Moulton, from a script originally by S Moore
#
#  This script deals images from a set of directories representing a
#  single study containing one or more series subdirectories, to one or more
#  DICOM storage SCPs.  This script should work for any deal which
#  uses the standard directory structure.  Specific deal scripts can
#  be derived from this one. Note:  It is assumed that parameters have been
#  validated by the calling script.  We just check for right number.
# Parameters:
#  1 root directory for images to be dealt
#  2 destination code from -cfg file (one or more AE titles)
#  3 server name
#  4 external queue number
#  5 internal queue number
#  6 base dir for server software
#  7 patient name
#  8 pid
#  9 accn
# 10 source AE title 
# 11 modality
#---------------------------------------------------------------------------
# 11-20-01 R Moulton - Modified to include server name in AE title
# 12-17-01 R Moulton - Modified to include study id in EXPORT Messages
# 06-04-02 R Moulton - Added SQL log
# 06-05-02 R Moulton - Added clear of .broadcast.rslt.dest.X files
# 07-09-03 R Moulton - Removed xfer_information log.
# 10-30-03 R Moulton - Added demographics to BEGIN_EXPORT
# 08-18-05 R Moulton - Modified to use MIRCS_dicom_broadcast
# 07-12-06 R Moulton - Modified to allow overide AE Titles (ovrAE.$DEST)
#***************************************************************************

#-------------------------------------- Check for right number of parameters
if ($11 == "" ) then
   echo "Usage: DealStudy.csh studydir dest server EQ IQ rootdir pname pid accn source mod"
   exit 1
endif

#----------------------------------------------------------- Load parameters
set DIR       = $1
set DEST      = $2
set S         = $3
set EQ        = $4
set IQ        = $5
set ROOT      = $6
set PNAME     = "$7"
set PID       = "$8"
set ACCN      = "$9"
set SOURCE    = "$10"
set MODALITY  = "$11"
set SQ        = $S$EQ

#--------------------------------------------------- Log and error log files
set TLOG = $ROOT/tmp/xmit_info.$SQ
touch $TLOG
set ERRORLOG = $ROOT/log/problem_log.$S
set ECLOG = $ROOT/log/EC.log

#-------------------------------------------- Server script/binary directory
set BIN = $ROOT/bin

#------------------------------------- Pull study out of directory tree name
@ i = 1
while (1)
   @ i += 1
   set s = `echo "$DIR" | cut -d'/' -f$i`
   if ($s == "") break
   set study = $s
end

#---------------------------------------- Make images base directory current         
cd $DIR

#------------------------------------------------- Log start of xmit message
set size = `du -k -s | cut -d'	' -f1` 
set unit = KB
if ($size > 1024) then
   @ size = $size / 1024
   set unit = MB
endif
echo "    Xmit $size $unit to $DEST -- `date`"

set dest = $DEST
set stdest = $dest
#---------------------------- Gets rid of ".X" files left by dicom_broadcast
#set dotx = ".broadcast.rslt.$stdest.X"
#if (-e $dotx) rm $dotx
#---------------------------------------------------------------------------
# The sendto file is used to hold data for the SQL log file, after the first
# xmit (attempt?).  Format is 'destcode',strt1stxmit,endlastxmit,pass,fail
# Here we load these values, or initialize them, if the file is blank.
#---------------------------------------------------------------------------
set xmtdest = `cat "dest.$DEST.sendto" | cut -d',' -f1`
if ("$xmtdest" != "") then
   set xmtsox  = `cat "dest.$DEST.sendto" | cut -d',' -f2`
   set xmteox  = `cat "dest.$DEST.sendto" | cut -d',' -f3`
   set xmtpass = `cat "dest.$DEST.sendto" | cut -d',' -f4`
   set xmtfail = `cat "dest.$DEST.sendto" | cut -d',' -f5`
   set xmtint  = `cat "dest.$DEST.sendto" | cut -d',' -f6`
   if ($xmtint == "") set xmtint = "0"
else
   set xmtdest = "'$DEST'"
   set xmteox  = "0"
   set xmtpass = "0"
   set xmtfail = "0"
   set xmtint  = "0"
   set xmtsox = `date '+%Y%m%d%H%M%S'`
endif
#---------------------------------------------------------------------------
# This tag file is used to show that a broadcast is in progress.  If it 
# exists at this point, the previous attempt was bounced or died.  Normally,
# we create it just before the broadcast and delete it just afterwards.
#---------------------------------------------------------------------------
set inprogress = "dest.$DEST.inprogress"
if (-e $inprogress) then
   @ xmtint += 1
   echo "$xmtdest,$xmtsox,$xmteox,$xmtpass,$xmtfail,$xmtint" > dest.$DEST.sendto
endif
touch $inprogress

#---------------------------------------------------------------------------
# If ovrAE.$DEST exists, it contains overide AE Title for transmit. 
# (Normally for studies being resent to distributor)
#---------------------------------------------------------------------------
set ovrdest = "nof"
set ovrae = "ovrAE.$DEST"
if (-e $ovrae) set ovrdest = `cat $ovrae`

if ("$ovrdest" == "nof") then
   $BIN/MIRCS_dicom_broadcast $DIR $stdest  > $TLOG
   set a = $status
else 
   $BIN/MIRCS_dicom_broadcast $DIR $stdest "$ovrdest" > $TLOG
   set a = $status
endif

set xmteox = `date '+%Y%m%d%H%M%S'`

#------------------------------------ Xmit OK, log and make xfer OK tag file
if ($a == 0) then
   @ xmtpass += 1
   echo "   " `tail -2 $TLOG`
   echo "$xmteox" > dest.$DEST.OK
#--------------------------------------------------------- problems occurred
else
   @ xmtfail += 1
   echo $DEST `date` >> $ECLOG
   echo $DEST `date` >> $ERRORLOG
   echo $DIR >> $ERRORLOG
   echo Transfer status: $a >> $ERRORLOG
   echo Some transfer problems in $study >> $ERRORLOG
   cat $TLOG >> $ERRORLOG
endif
#-------------------------------------------- update contents of sendto file
echo "$xmtdest,$xmtsox,$xmteox,$xmtpass,$xmtfail,$xmtint" > dest.$DEST.sendto

#--------------------------------------------- Give up on failures processing
if (($a != 0) && -e $BIN/EQuit.cfg) then
#------------------------------------------------------- based on destination
   set threshold = `cat $BIN/EQuit.cfg | grep "^Dest:$DEST" | cut -d',' -f2`
   if ($threshold != "") then
      if ($xmtfail >= $threshold) then
         rm dest.$DEST.sendto .broadcast*$DEST* 
         echo `date` "d t=$threshold f=$xmtfail $SOURCE/$MODALITY to $DEST Gave up - $PID $ACCN $PNAME $study" >> $LOG/EQuit.log
      endif
   endif
#------------------------------------------------------------ based on source
   set threshold = `cat $BIN/EQuit.cfg | grep "^Source:$SOURCE" | cut -d',' -f2`
   if ($threshold != "") then
      if ($xmtfail >= $threshold) then
         rm dest.$DEST.sendto .broadcast*$DEST*
         echo `date` "d t=$threshold f=$xmtfail $SOURCE/$MODALITY to $DEST Gave up - $PID $ACCN $PNAME $study" >> $LOG/EQuit.log
      endif
   endif
endif

if (-e $inprogress) rm $inprogress
if (-e $TLOG) rm $TLOG
exit 0
