#!/bin/csh

#************************************************************************
# Dealer.csh                                                    R Moulton
#
# This script examines image directories created by the dicom receiver
# for tag files placed in them by the router indicating that they are to
# be sent to the destination this queue is processing, with names in the 
# format dest.destination.sendto.  When found, these are sent to the
# destination, and if successful, the images are moved to the todelete 
# directory.
# The script has four parameters, the queue which is to be processed
# (default is bond1), the destination (default is All), a number of 
# seconds to sleep at the beginning of the run (used to offset seconds 
# from a cron call) and the processing mode (test or production, default 
# is production). A destination of "All" will process all destinations 
# for that queue in the distributor configuration file.  The distributor 
# configuration file Distributor.cfg is system in.  Command syntax is:
#
#  Dealer.csh queue destcode delay mode < Distributor.cfg        
#
# There are conceptually two different queue's being handled by the 
# Dealer, that seen by the monitor program and scripting tools, and the
# internal queue in Distributor.cfg.  Because there may be more than one
# cron process handling a given queue (e.g. two destinations and each
# one handled by a different cron invocation of Dealer), the queue 
# parameter may have two numbers, the external (monitor) and internal
# (.cfg) queues.  If these are the same, the queue parameter can be
# represented simply (e.g. bond.2).  If the two queues are different, the
# parameter is represented with the external queue then the internal
# queue, separated by a period "." (e.g. bond.2.1).  The external queue
# number must be unique for the server; The internal queue number must
# refer to the correct queue for that server in the .cfg.  If the queues
# are the same, the second period and internal queue can be omitted.
#
# To reduce the software overhead required to process all of the lines in
# Distributor.cfg, most of which do not apply to a particular invocation
# of Dealer, a script (c.csh) was developed to "compress" Distributor.cfg,
# creating a Server.cfg file (e.g. Bond.cfg) which contains just those
# lines which pertain to Bond, and internal queue Server#.cfg (e.g.
# Bond4.cfg) for each internal queue.  These files omit all the comments
# also, making the processing of a Dealer script faster.  For example,
# to invoke a Dealer for Bond4 queue you might say:
#
#  Dealer.csh bond.4 IMG4 < Bond4.cfg
#------------------------------------------------------------------------
# 11-14-01 R Moulton - Added support for live server in .cfg
# 11-15-01 R Moulton - Added Message queue different from .cfg queue
# 11-30-01 R Moulton - Modified to use .rslt files to determine which
#          images have been sent
# 02-21-02 R Moulton - Modified to enhance use of .rslt files; handle
#          cases where some images have been deleted, make sure that
#          images are deleted after last xmit.
# 05-08-02 R Moulton - Modified to allow two digit internal and external
#          queue numbers, support bigben
# 08-30-02 R Moulton - Modified to add relay1 and relay2
# 10-07-02 R Moulton - Commented out IMAGE_SENT message to monitor
#************************************************************************

#----------------------------------------------------- Validation strings
# These are used to validate entries in the distributor configuration
# file.  
#------------------------------------------------------------------------
set VALIDSERVERS = "bond goldfinger fleming bigben relay1 relay2 relay3 relay3"
set ONEMACHINE = DealStudy.csh

#------------------------------------------------------------------------
# The current live server is set by .cfg lines of the form 
# "Server.servername".  If the live server is set, and it is not our
# server, all .cfg lines will be skipped until another "Server" line or
# the end-of-file
#------------------------------------------------------------------------
set LIVESERVER = ""

#---------------------------- Must be logged in as spectrum to run Dealer
if (`whoami` != "spectrum") then
   echo "Must be user spectrum to run Dealer"
   exit 1
endif

#------------------------------------------------------------------------
# Parameter 1 is server and queue (e.g. bond.1) or server, external queue,
# "." and internal queue (e.g. bond.5.3).
#------------------------------------------------------------------------
set S = bond
set EQ = 1
set IQ = 1
set SQ = bond1
if ("$1" != "") then
   set S  = `echo "$1" | cut -d'.' -f1`
   set EQ = `echo "$1" | cut -d'.' -f2`
   set IQ = `echo "$1" | cut -d'.' -f3`
   if ("$IQ" == "") set IQ = "$EQ"
endif
set SQ = $S$EQ
#--------------------------------------------------- Validate server name
set FLAG = NO
foreach server ($VALIDSERVERS)
   if ("$S" == $server) then
      set FLAG = YES
      break
   endif
end
if ($FLAG == "NO") then
   echo "$1 is not a valid server name"
   exit 1
endif

#------------------------------------------------ Validate external queue
@ max = 98
set FLAG = NO
@ j = 0
while ($j < $max)
   @ j += 1
   if ($EQ == $j) then
      set FLAG = YES
      break
   endif
end
if ($FLAG == "NO") then
   echo "$1 has invalid external queue number"
   exit 1
endif

#------------------------------------------------ Validate internal queue
@ max = 98
set FLAG = NO
@ j = 0
while ($j < $max)
   @ j += 1
   if ($IQ == $j) then
      set FLAG = YES
      break
   endif
end
if ($FLAG == "NO") then
   echo "$1 has invalid internal queue number"
   exit 1
endif

#----------------------------------- Param 2 is destination, default All 
set DEST = All 
if ("$2" != "") set DEST = "$2"

#----------------------------- Param 3 is delay seconds (0-59, default 0)
set SLEEP = 0
if ("$3" != "") set SLEEP = "$3"
echo "$SLEEP" | egrep "[0-9]|[0-9][0-9]" > /dev/null
if ($status != 0) then
   echo "delay of $SLEEP - not valid number"
   exit 1
endif
if ($SLEEP < 0 || $SLEEP > 59) then
   echo " delay of $SLEEP not in range 0-59"
   exit 1
endif
if ($SLEEP != 0) sleep $SLEEP

#------------------ mode we are running in is param 4, production or test
set MODE = production
if ("$4" != "") set MODE = $4
if ("$MODE" != "production" && "$MODE" != "test") then
   echo "$MODE is not a valid production mode"
   exit 1
endif

#----------------------------- Validate appropriate server directory tree
set ROOT = $HOME/$MODE/$S
if (!(-d $ROOT)) then
   echo "$ROOT is not a valid directory"
   exit 1
endif

set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo "$BIN is not a valid directory"
   exit 1
endif

set LOGS = $ROOT/log
if (!(-d $LOGS)) then
   echo "$LOGS is not a valid directory"
   exit 1
endif

set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo "$TMP is not a valid directory"
   exit 1
endif

#---------------------------------------------------- Validate image root
set IMGS = /distributor/imgs
if ($MODE == "test") set IMGS = $ROOT/imgs
if (!(-d $IMGS)) then
   echo "$IMGS is not a valid directory"
   exit 1
endif
set STORE = $IMGS/todelete

#----------------------------------- Script lock and stoprouter tag files
set LOCK  = $TMP/Dealer.csh.lock.$SQ
set STOP  = $TMP/stopDealer
set QSTOP = $TMP/stopDealer.$SQ
set DSTOP = $TMP/stopDealer.$DEST
set QUIT  = "-e $STOP || -e $QSTOP || -e $DSTOP"
#----------------------------------------------------------- script names
set SCRIPT    = "Dealer.csh"
set DEAL = $BIN/$ONEMACHINE
#------------------------------------------------------------- temp files
set T1 = $TMP/Dealer.$SQ.temp1
set T2 = $TMP/Dealer.$SQ.temp2

#----------------------- If lock file exists, exit... otherwise create it
if (-e $LOCK) exit 0
touch $LOCK
if (!(-e $LOCK)) then
   echo "Could not create $LOCK file" 
   exit 1
endif

#--------------------------------------------------------- Open log files
set LOG = $LOGS/Dealer.$SQ.log
if (-e $LOG) rm $LOG
touch $LOG
if (!(-e $LOG)) then
   echo "Unable to create $LOG file"
   exit 1
endif
set MLOG = $LOGS/MoveStudy.$SQ.log
if (-e $MLOG) rm $MLOG
touch $MLOG
if (!(-e $MLOG)) then
   echo "Unable to create $MLOG file"
   exit 1
endif

set ERRLOG = $LOGS/Dealer_problems.log
touch $ERRLOG
if (!(-e $ERRLOG)) then
   echo "Unable to create $ERRLOG file"
   exit 1
endif

#------------------------ Are we to process MoveStudy on this queue?
set MOVEFLAG = "YES"

#------------------------------------- Begin Dealer message to monitor
$BIN/Message $SCRIPT $SQ BEGIN_SCRIPT
echo "Begin :: $SCRIPT $S.$EQ.$IQ $DEST $MODE -- `date`" >> $LOG

#--------------------------------------- Initial values of input defaults
set DEFISERVER = $S  
set DEFIQUEUE  = 1
set DEFIDEST1  = IMG4
set DEFIDEST2  = IMG2
set DEFIDEST3  = IMG5

#------------------------------------------------------------------------
# This loop processes one line from the config table, which is std in.
# The same table can be used for all servers being used as distributors.
#------------------------------------------------------------------------
@ linenumber = 0
@ blanklines = 0
@ xmits = 0
@ fails = 0
while (1)
#------------------------------------------- Check for stopDealer file
   if ($QUIT) then
      echo "Exiting early, stopDealer file found -- `date`" >> $LOG
      $BIN/Message $SCRIPT $SQ END_SCRIPT
      rm $LOCK
      exit 0
   endif
#---------------------------------------------------- Get next table line
   set TABLELINE = $<
   @ linenumber += 1

#---------------------- "end-of-file" on line signifies the table is done
   if ("$TABLELINE" == "end-of-file") break

#------------------------------------------------------------------------
# Blank lines and lines with "#" in column 1 are ignored as comments
#------------------------------------------------------------------------
   if ("$TABLELINE" == "") continue
   set COMCHAR = `echo "$TABLELINE" | cut -c1`
   if ("$COMCHAR" == "#") continue

#------------------------------------------------------------------------
# Server lines set live server.  If it is set, and it it not our server,
# we skip all .cfg lines until next Server line or end-of-file
#------------------------------------------------------------------------
   if (`echo "$TABLELINE" | cut -d',' -f1` == "Server") then
      set LIVESERVER = `echo "$TABLELINE" | cut -d',' -f2`
   endif
   if ("$LIVESERVER" != "" && "$LIVESERVER" != $S) continue

#------------------------------------------------------------------------
# SetDefaults record sets server, queue and one or two destinations
# NOTE: If there are errors in the SetDefaults line, the errors will be
# logged, but the erroneous values will be stored, and used if items on
# subsequent detail lines are defaulted.  This will result in errors on
# those detail lines, but the alternative is to use old defaults, which
# could lead to incorrect processing.
#------------------------------------------------------------------------
   if (`echo "$TABLELINE" | cut -d',' -f1` == "SetDefaults") then

#----------------------------------- pull and validate new default server
      set NEWDEFSERVER = `echo "$TABLELINE" | cut -d',' -f2`
      set FLAG = NO
      foreach server ($VALIDSERVERS)
         if ("$NEWDEFSERVER" == $server) then 
            set FLAG = YES
            break
         endif
      end
      if ($FLAG == "NO") then
         echo "config file error at line: $linenumber" >> $ERRLOG
         echo "  $TABLELINE" >> $ERRLOG
         echo "  invalid server name: $NEWDEFSERVER" >> $ERRLOG
      endif

#------------------------------------------------ validate default queue
      set max = 98
      set NEWDEFIQUEUE = `echo "$TABLELINE" | cut -d',' -f3`
      set FLAG = NO
      @ i = 0
      while ($i < $max)
         @ i += 1
         if ("$NEWDEFIQUEUE" == $i) then
            set FLAG = YES
            break
         endif
      end
      if ($FLAG == "NO") then
         echo "config file error at line: $linenumber" >> $ERRLOG
         echo "  $TABLELINE" >> $ERRLOG
         echo "  invalid queue: $NEWDEFIQUEUE" >> $ERRLOG
      endif

#-------------------------------------------------- Validate destinations
# We will look for two default destinations, which must match the
# valid destination list or be blank.  But if validation list is blank,
# we skip the validations.
#------------------------------------------------------------------------
      set NEWDEDEST1 = `echo "$TABLELINE" | cut -d',' -f4`
      set NEWDEDEST2 = `echo "$TABLELINE" | cut -d',' -f5`
      set NEWDEDEST3 = `echo "$TABLELINE" | cut -d',' -f6`

#------------------------------ All validations complete, update defaults
      set DEFISERVER = $NEWDEFSERVER 
      set DEFIQUEUE  = $NEWDEFIQUEUE 
      set DEFIDEST1  = $NEWDEDEST1
      set DEFIDEST2  = $NEWDEDEST2 
      set DEFIDEST3  = $NEWDEDEST3
#-------------------------------- End of processing for SetDefaults lines
      continue
   endif

#------------------------------------------------------------------------
# Here we are looking at a detail line from the configuration file, which
# is of the form source,modality,name,server,queue,dest1,dest2... with up
# to five destinations. Everything from server on will default if it is
# blank.  First, we validate the detail line fields
#------------------------------------------------------------------------

#-------------------------------------------- Pull values from table line
   set SOURCE   = `echo "$TABLELINE" | cut -d',' -f1`
   set MODALITY = `echo "$TABLELINE" | cut -d',' -f2`
   set NAME     = `echo "$TABLELINE" | cut -d',' -f3`
   set SERVER   = `echo "$TABLELINE" | cut -d',' -f4`
   set QUEUE    = `echo "$TABLELINE" | cut -d',' -f5`
   set D1       = `echo "$TABLELINE" | cut -d',' -f6`
   set D2       = `echo "$TABLELINE" | cut -d',' -f7`
   set D3       = `echo "$TABLELINE" | cut -d',' -f8`
   set D4       = `echo "$TABLELINE" | cut -d',' -f9`
   set D5       = `echo "$TABLELINE" | cut -d',' -f10`

#--------------------------------------------------------- Default values
   if ("$SERVER" == "") set SERVER = $DEFISERVER
   if ("$QUEUE" == "")  set QUEUE  = $DEFIQUEUE  
   if ("$D1" == "")     set D1     = $DEFIDEST1  
   if ("$D2" == "")     set D2     = $DEFIDEST2  
   if ("$D3" == "")     set D3     = $DEFIDEST3

#------------------------------ Destinations combined for loop processing
   set DESTINATIONS = "$D1 $D2 $D3 $D4 $D5"      

#------------------------------------------------------------------------
# Validations.  Source and modality will be validated by the existance of
# the appropriate directories in the image tree.  If the directory does
# not exist this is not an error; simply nothing is done.  So there is no
# explicit validation here.  The name field is passed to the monitor as
# part of the Message calls, and is not currently validated in this 
# script.  Detailed validation and logging is done by the Router.  Here
# we just look for what we are expecting.
#------------------------------------------------------------------------

#------------------------------------------------------------------------
# server.  If it is not this server then we skip this line.
#------------------------------------------------------------------------
   if ($SERVER != $S) continue

#------------------------------------------------------------------------
# queue.  If it is not this internal queue, then we skip this line.
#------------------------------------------------------------------------
   if ("$QUEUE" != $IQ) continue

#------------------------------------------------------------------------
# If our destination is not one of the destinations then skip this line
#------------------------------------------------------------------------
   if ("$DESTINATIONS" == "") continue
   set FLAG = NO
   if ("$DEST" == "All") set FLAG = YES
   foreach dest ($DESTINATIONS)
      if ("$dest" == $DEST) set FLAG = YES
   end
   if ($FLAG == "NO") continue

#------------------------------------------------------------------------
# We now have a line which is for our server, queue, and destination
#------------------------------------------------------------------------

#------------------------------------------------------------ Echo to log
   echo "==================================================" >> $LOG
   echo "Processing :: $TABLELINE -- `date`" >> $LOG

#----------- If source modality root directory does not exist, we're done
   set DIR = $IMGS/$SOURCE/$MODALITY
   set STOREDIR = $STORE/$SOURCE/$MODALITY
   if ($MODALITY == "") then
      set DIR = $IMGS/$SOURCE
      set STOREDIR = $STORE/$SOURCE
   endif
   if(!(-d $DIR)) then
     echo "  $DIR directory does not exist" >> $LOG
     continue
   endif

   cd $DIR

#--------------------------- If there are no images in modality, skip it
   ls | grep "^[1-9]" >& /dev/null
   if ($status != 0) then
      echo "  No images in $DIR" >> $LOG
      continue
   endif

#-------------- Notify monitor we are beginning Dealer pass for source
   $BIN/Message "$NAME" $SQ BEGIN_PASS $DEST

#----------------------------------------- Pass studies in base directory
   ls > $T2
   echo "end-of-file" >> $T2
   @ sc = 0
   while (1)
      @ sc += 1
      set study = `head -$sc $T2 | tail -1`
      if ("$study" == "end-of-file") then
         rm $T2
         break
      endif
      echo "$study" | grep "^[1-9]" >& /dev/null
      if ($status != 0) continue
      if (!(-d $study)) continue
#------------------- Quick check.  If no *.sendto files exist, we're done
      ls $study/*.sendto >& /dev/null
      if ($status != 0) continue

      echo "--study - $study" >> $LOG
#------------------------------------------------------------------------
# Now we are looking for destination tag files indicating that this study
# needs to be dealt to our destination: dest.destinationcode.sendto.
# If we find it, we do a deal.  
#------------------------------------------------------------------------
# Another issue is the "All" DEST, which means that we are to deal
# to all the destinations on the config file line which are not stopped
#------------------------------------------------------------------------
      set D = $DEST
      if ($D == "All") then
         set D = "$DESTINATIONS"
      else
         if (!(-e $study/dest.$D.sendto)) continue
      endif

#-----------------------------------------------------------------------
# To be safe, we'll check to make sure the images are still there
#------------------------------------------------------------------------
      ls $study/[1-9]* >& /dev/null
      if ($status != 0) then    
         echo "    No series in study: $study" >> $LOG
         continue
      endif
      set stillthere = NO 
      foreach series ($study/[1-9]*)
         ls $series/[1-9]* >& /dev/null
         if ($status != 0) continue
#-------------------------------------------- Grab an image for later
         set image = `ls $series/[0-9]* | tail -1`
         set simage = `echo "$image" | cut -d'/' -f2-3`
         set stillthere = YES
         break
      end
      if ($stillthere == NO) then    
         echo "    No images in study: $study" >> $LOG
         continue
      endif

      $BIN/dcm_dump_file $image | egrep "0010 0010|0010 0020|0008 0050|0008 0020|0008 0030|0008 0080|0008 1010" > $T1
      set PNAME = "`cat $T1 | grep '0010 0010' | cut -d'/' -f5 | $BIN/PFix`"
      set PID   = "`cat $T1 | grep '0010 0020' | cut -d'/' -f5 | $BIN/PFix`"
      set ACCN  = "`cat $T1 | grep '0008 0050' | cut -d'/' -f5 | $BIN/PFix`"
      set STDT  = "`cat $T1 | grep '0008 0020' | cut -d'/' -f5 | $BIN/PFix date`"
      set STTM  = "`cat $T1 | grep '0008 0030' | cut -d'/' -f5 | $BIN/PFix time`"
      set INSTN = "`cat $T1 | grep '0008 0080' | cut -d'/' -f5 | $BIN/PFix`"
      set STNM  = "`cat $T1 | grep '0008 1010' | cut -d'/' -f5 | $BIN/PFix`"
#------------------------------------------------------------------------
# send the images.  The foreach loop is because $D could be either a 
# single destination or all the destinations on the line.  For this same
# reason, we check for the stopDealer.dest & *.sendo again
#------------------------------------------------------------------------
      set XMITSOK = NONE
      foreach dst ($D)
         set dest = $dst
         if (!(-e $study/dest.$dest.sendto)) continue
         if (-e $TMP/stopDealer.$dest) continue
#------------------------------------------------------------------------------------
# For the standard destinations we add the servername to the destination, so the
# repositories can differentiate
#------------------------------------------------------------------------------------
         set stdest = $dest
         if (-e $BIN/Dealer.cfg) then
#------------------------------------- prefix lower case server name (e.g. bond_IMG6)
            set fnd = `cat $BIN/Dealer.cfg | grep "^server:$dest"`
            if ("$fnd" != "") set stdest = ${S}_$dest
#------------------------------------- prefix upper case server name (e.g. BOND_IMG6)
            set fnd = `cat $BIN/Dealer.cfg | grep "^SERVER:$dest"`
            if ("$fnd" != "") set stdest = ${ACSNAME}_$dest
#------------------------------------------- complete change (e.g. LTA => BWCH_CACHE)
            set newdest = `cat $BIN/Dealer.cfg | grep "^Dest:$dest" | cut -d'=' -f2`
            if ("$newdest" != "") set stdest = $newdest
#------------------------------------------ this dest & queue (e.g. IMG5.3 => relay2)
            set newdest = `cat $BIN/Dealer.cfg | grep "^DQueue:$dest.$IQ" | cut -d'=' -f2`
            if ("$newdest" != "") set stdest = $newdest
         endif 
#---------------------------------------------------- If already sent, skip this deal
         $BIN/SentTest.csh $DIR/$study $stdest $BIN $LOG
         if ($status == 2) then
	         set SKIPPED = YES
            echo "    $dest - already sent" >> $LOG
	         date '+%Y%m%d%H%M%S' > $study/dest.$dest.OK
         else 
	         set SKIPPED = NO 
#-------------------------------------------------- dicom xmit of all images in study
            $DEAL $DIR/$study $dest $S $EQ $IQ $ROOT $simage \
				"$PNAME" "$PID" "$ACCN" $SOURCE $MODALITY >> $LOG
            echo "    End xmit -- `date`" >> $LOG
#--------------------------------------- dest.D.OK file means send worked
            @ xmits +=1
	      endif
         if ($XMITSOK == NONE) set XMITSOK = YES
         if (!(-e $study/dest.$dest.OK)) then
            echo "    *** Xmit failed ***" >> $LOG
            @ fails +=1
            set XMITSOK = NO
            continue
         endif
         set EOT = `cat $study/dest.$dest.OK`
         rm $study/dest.$dest.OK

         if ($SKIPPED == NO) then
            echo "    Images from $NAME for $PNAME $PID $ACCN - `date` sent to $dest" >> $LOG
#           $BIN/Message "$NAME" $SQ IMAGE_SENT "$PNAME" "$PID" "$ACCN" $dest
	 endif
      end
#-------------------------- Here is substring for clear of completed files
      if ($XMITSOK == YES && $MOVEFLAG == YES) then
         $BIN/MoveStudy.csh $DIR/$study $STOREDIR/$study $S $SQ $ROOT $EOT \
            "$PNAME" "$PID" "$ACCN" $IQ $SOURCE $MODALITY \
	    "$STDT" "$STTM" "$INSTN" "$STNM" >> $MLOG & 
      endif
   end                           # End of foreach loop for studies
#-------------------------- Notify monitor of end of deal for source
   $BIN/Message "$NAME" $SQ END_PASS $DEST
end                              # End of while loop reading config table
#------------------------------------------------------------- We're done
echo "Dealer execution completed -- `date` $xmits - $fails" >> $LOG
$BIN/Message $SCRIPT $SQ END_SCRIPT
if (-e $LOCK) rm $LOCK
if (-e $T1) rm $T1
exit 0
