#!/bin/csh
#******************************************************************************
# ECheck.csh                                                          R Moulton
#
# This script checks for high error rates in transmissions from the distributor
# to specified destinations, using a special short term log of errors generated
# by the Dealer scripts.  ECheck.cfg gives the destinations to be tested, and
# contains lines of the form:
#    dest,# errors min to generate error,hours to page,hdr file to use
# where 5 is the default # of errors and ECheck.hdr is the default header file
# to use.  The hours to page represents the beginning and ending hour on
# a 24 hour clock during which pages should be sent, e.g. 8-17 being 8AM thru
# 5PM.  Default is 0-23 (all the time). An entry of "B" (business hours) is
# taken to mean 8-17.  Note that a page will be sent if the current hour is
# in the inclusive range given, so 8-17 will start paging at 8AM and stop 
# paging at 6PM (when hour 17 is over). 
# Invocation:
#    ECheck.csh < Echeck.cfg
#******************************************************************************

set DEFMIN    = 5     
set ECLOG     = $LOG/EC.log            
set LOGFILE   = $LOG/ECheck.log         
set DEFHFILE  = ECheck.hdr        
set ALERTLOG  = $LOG/ECheckAlerts.log
set T         = $TMP/ECheck.tmp

#------------------------------------------------ Must be logged in as spectrum 
if ($user != "spectrum") then
   echo "Must be user spectrum to run ECheck"
   exit 1
endif

echo "Begin :: ECheck.csh `date`" > $LOGFILE

#---------------------------------------- Quick check: No errors since last run
if (!(-e $ECLOG)) then
   echo "No errors since last check" >> $LOGFILE
   exit 0
endif

#-------------------------------------------------------- Pass ECheck.cfg input
while (1)
   set LINE = $<
   if ("$LINE" == "") break                    
   if (`echo "$LINE" | cut -c1` == "#") continue      

   set DEST = `echo "$LINE" | cut -d',' -f1`          
   set MIN  = `echo "$LINE" | cut -d',' -f2`          
   set TIM  = `echo "$LINE" | cut -d',' -f3`          
   set HDR  = `echo "$LINE" | cut -d',' -f4`          

   if ("$MIN" == "") set MIN = $DEFMIN
   if ("$HDR" == "") set HDR = $DEFHFILE

   if ("$TIM" == "") then
      @ SHR = 0
      @ EHR = 23
   else if ("$TIM" == "B") then
      @ SHR = 8
      @ EHR = 17
   else
      set SHR = `echo "$TIM" | cut -d'-' -f1`
      set EHR = `echo "$TIM" | cut -d'-' -f2`
   endif

   grep "^$DEST" $ECLOG > /dev/null
   if ($status == 1) then
      echo "$DEST - no errors found" >> $LOGFILE
      continue
   endif

   set A = (`grep "^$DEST" $ECLOG | cat -n | tail -1`)
   set E = $A[1]

   echo "$DEST - $E errors found, $MIN threshold" >> $LOGFILE

   if ($E < $MIN) continue

#------------------------------------------------------------------ Send Alert
   set CHR = `date "+%H"`
   if (($SHR <= $CHR) && ($EHR >= $CHR)) then
      echo "e=suchkom@mir.wustl.edu" > $T
      echo "t=Xmit Errors" >> $T
      if (-e $BIN/$HDR) cp $BIN/$HDR $T
      echo "Xmit errs: $SNAME" >> $T
      echo "sending to $DEST" >> $T
      echo "$E errs" >> $T
      $BIN/SendAlert $T

      echo "Alert for $DEST - $E errs, `date`" >> $ALERTLOG
   endif
end

echo "Completed at `date`" >> $LOGFILE
if (-e $T) rm $T
if (-e $ECLOG) rm $ECLOG
exit 0

