#!/bin/csh

#********************************************************************
# EchoTest.csh                                              R Moulton
#
# This script can be set up as a cron job to periodically test other
# dicom AE's using dicom_echo.  It sends messages to the monitor
# indicating that it is running, that tests are being performed, and
# what their results are.  The various echo tests which are to be
# run are stored in EchoTest.cfg, which is standard input to this
# string.  The string has one parameter, and the mode (production or
# test, def production).   Command format:
#
# EchoTest.csh mode < EchoTest.cfg
#--------------------------------------------------------------------
# 11-06-01 R Moulton Added Bounce to clear previous incarnation 
# 05-14-02 R Moulton Used SNAME parameters
# 06-29-07 R Moulton Removed Messages
# 2008 Jul 29 MSuchko change for standard error
# 2009 Aug  5 MSuchko empty line in cfg file now breaks out of read loop.
#********************************************************************

#--------------------------------------- must be spectrum to run this
if (`whoami` != spectrum) then
   echo "Must be user spectrum to run this script"
   exit 1
endif

#--------------------------------------------------------------- mode
set MODE = production
if ("$1" != "") set MODE = "$1"
if ("$MODE" != production && "$MODE" != test) then
   echo "$1 must be production or test"
   exit 1
endif

#------------------------------------------------ Validate directories
set ROOT = $home/$MODE/$SNAME
if (!(-d $ROOT)) then
   echo "$ROOT directory does not exist"
   exit 1
endif
set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo "$BIN directory does not exist"
   exit 1
endif
set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo "$TMP directory does not exist"
   exit 1
endif
set LOGS = $ROOT/log
if (!(-d $LOGS)) then
   echo "$LOGS directory does not exist"
   exit 1
endif

#-------------------------------------------------------------- files
set LOG    = $LOGS/EchoTest.log
set ERRLOG = $LOGS/EchoTest_problems.log
set LOCK   = $TMP/EchoTest.csh.lock
set T1     = $TMP/EchoTest.temp1

#--------------------------------------------------- Script lock file
# Note special case here.  dicom_echo may hang indefinitely, causing 
# EchoTest to find itself running.  In this case, we kill the 
# previous incarnation and continue.
#--------------------------------------------------------------------
if (-e $LOCK) $BIN/Bounce.csh EchoTest $SNAME killold >> $ERRLOG
touch $LOCK
if (!(-e $LOCK)) then
   echo "unable to create $LOCK file" >> $ERRLOG
   exit 1
endif

#----------------------------------------------------------- Log file
if (-e $LOG) rm $LOG
touch $LOG
if (!(-e $LOG)) then
   echo "Unable to create $LOG file" >> $ERRLOG
   exit 1
endif

#--------------------------- Monitor message for begin testing script
echo "EchoTest: $SNAME - Begin processing. `date`" >> $LOG

#--------------------------------------------- Pass EchoTest.cfg file
@ blanklines = 0
while (1)
   set TABLELINE = $<
#----------------------------------------- eof, blank lines, comments
   if ("$TABLELINE" == "end-of-file") break
   if ("$TABLELINE" == "") break
   if (`echo "$TABLELINE" | cut -c1` == "#") continue

#--------------------------------- param 1 is server, must match ours
   if (`echo "$TABLELINE" | cut -d',' -f1` != $SNAME) continue

#--------------------------------------------------------------------
# Now we have an echo.  param 2 is -a, param 3 is -c
#--------------------------------------------------------------------
   echo "==============================================" >> $LOG
   echo "$TABLELINE - `date`" >> $LOG

   set ae = `echo "$TABLELINE" | cut -d',' -f2`
   set a  = `echo "$TABLELINE" | cut -d',' -f3`
   set c  = `echo "$TABLELINE" | cut -d',' -f4`

#----------------------------------- About to test message to monitor
   echo "  Testing..." >> $LOG

#-------------------------------------------------------- actual test
   $BIN/dicom_echo -a $a -c $c >& $T1

#--------------- If "Successful Operation" found in output, it worked
   cat $T1 | grep "Successful operation" >& /dev/null
   if ($status == 0) then
      echo "  Passed." >> $LOG
   else set pf = FAIL
      echo "  Failed dicom echo, output follows:" >> $LOG
      cat $T1 >> $LOG
      echo "==== $TABLELINE failed test `date` output follows: ====" >> $ERRLOG
      cat $T1 >> $ERRLOG
      cp $BIN/EchoTest.hdr $T1
      echo "Echo Problem" >> $T1
      echo "on $ae" >> $T1
      $BIN/SendAlert $T1
   endif

#--------------------------------------- End of reading cfg file loop
end

if (-e $T1) rm $T1   
if (-e $LOCK) rm $LOCK
echo "Processing complete" >> $LOG
exit 0
