#!/bin/csh
#*******************************************************************************
# GetLogs.csh                                                          R Moulton
# Script designed to retrieve log files from PACS servers and organize them by
# days for analysis and review.  The files are retrieved from each host server
# using FTP, then processed to extract new entries.  The resulting file is
# greped to separate entries by days and store them in log files for each day.
# Files are stored in directories for each host server under the logs directory.
# The process is driven by the GetLogs.cfg file, which is standard input to this
# script.  Each host is represented by a line in the .cfg file.  It must also be
# listed in the server /etc/hosts file, and have an entry in the user .netrc
# file giving the id and password for the FTP login.  This script is run as a
# cron job repeatedly updating files as indicated in the cfg file.
#*******************************************************************************
set STORE = "/distributor/logs"

#---------------------------------- Must be logged in as spectrum to run GetLogs
if ($USER != spectrum) then
   echo "Must be user spectrum to run GetLogs"
   exit 1
endif

if (!(-d $STORE)) then
   echo "$STORE is not a valid directory"
   exit 1
endif

set ROOT = $HOME/production/$SNAME
if (!(-d $ROOT)) then
   echo "$ROOT is not a valid directory"
   exit 1
endif

set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo "$BIN is not a valid directory"
   exit 1
endif

set LOGS = $ROOT/log
if (!(-d $LOGS)) then
   echo "$LOGS is not a valid directory"
   exit 1
endif

set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo "$TMP is not a valid directory"
   exit 1
endif

set STOP = $TMP/stopGetLogs
set LOG = $LOGS/GetLogs.log
set ERRLOG = $LOGS/GetLogs_problems.log
set COMMAND = $TMP/GetLogs.cmd
set OUTPUT = $TMP/GetLogs.out
set T1 = GetLogs.tmp

if (-e $STOP) exit 0
   
touch $LOG
if (!(-e $LOG)) then
   echo "Unable to create $LOG file"
   exit 1
endif

touch $ERRLOG
if (!(-e $ERRLOG)) then
   echo "Unable to create $ERRLOG file"
   exit 1
endif

echo "Begin processing -- `date`" > $LOG

set TODAYHR = `date '+%Y%m%d%H'`
set TODAY   = `date '+%Y%m%d'`
set YESTERDAY = `$BIN/OTime 1 '%Y%m%d'`
#------------------------------------------------------------------------------
# This loop processes one line from the config table, which is std in.
#------------------------------------------------------------------------------
while (1)
#---------------------------------------------------------- Check for stop file
   if (-e $STOP) then
      echo "=========================================== `date`" >> $LOG
      echo "Exiting early, $STOP found" >> $LOG
      break
   endif

#------------------------------------------------ Load next file from .cfg file
   set TABLELINE = $<

#------------------------------------------------------------------ End of file
   if ("$TABLELINE" == "end-of-file") then
      echo "=========================================== `date`" >> $LOG
      echo "end-of-file found" >> $LOG
      break
   endif
   
#------------------------------------------------------------------------------
# Blank lines and lines with "#" in column 1 are ignored as comments
#------------------------------------------------------------------------------
   if ("$TABLELINE" == "") continue
   set COMCHAR = `echo "$TABLELINE" | cut -c1`
   if ("$COMCHAR" == "#") continue

   echo "=========================================== `date`" >> $LOG
   echo "Processing: $TABLELINE" >> $LOG
#------------------------------------------ Parse out components of format line
   set SERVER    = `echo "$TABLELINE" | cut -d',' -f1`
   set DIRECTORY = `echo "$TABLELINE" | cut -d',' -f2`
   set FILENAME  = `echo "$TABLELINE" | cut -d',' -f3`
   set SUBDIR    = `echo "$TABLELINE" | cut -d',' -f4`
	set HOURS     = `echo "$TABLELINE" | cut -d',' -f5`
	set PACSCODE  = `echo "$TABLELINE" | cut -d',' -f6`
	set FORMATSTRING = "%Y.%m.%d" 
	if ($PACSCODE == PPACS) set FORMATSTRING = "%Y-%m-%d"

#--------------------------------------------------- make sure directory exists
   set TARGETDIR = $STORE/$SERVER
   if (!(-d $TARGETDIR)) mkdir -p $TARGETDIR
   if ("$SUBDIR" != "") then
      set TARGETDIR = $TARGETDIR/$SUBDIR
      if (!(-d $TARGETDIR)) mkdir -p $TARGETDIR
   endif
#------------------------------------------------------------------------------
# a file "complete.YYYYMMDDHH" in the target directory indicates that logs have
# been saved through this point. This file will have data on the last 
# completed operation in it.
#------------------------------------------------------------------------------
   cd $TARGETDIR
#--------------------------- Pull date-hr done through (zero if file not found)
	set DONETHROUGH = `ls complete.* | cut -d'.' -f2`
	if ($DONETHROUGH == "") set DONETHROUGH = "0000000000"
	if (-e complete.$DONETHROUGH) cat complete.$DONETHROUGH >> $LOG
#------------------------------------------------------------------------------
# Look for the first (if any) to do time which is after the current done through
# time and not after the current time.  If found, we need to save logs
#------------------------------------------------------------------------------
   set DUE = NO
   foreach hh ($HOURS)
	   set NEXTDUE = ${TODAY}$hh
		if ($NEXTDUE <= $DONETHROUGH) continue
		if ($NEXTDUE > $TODAYHR) continue 

		set DUE = YES
		break
	end
	if ($DUE == NO) then
	   echo "Not due yet" >> $LOG
		continue
	endif
	echo "Saving logs" >> $LOG
#---------------------------------------------------- Generate ftp command file
   echo "cd $DIRECTORY" > $COMMAND
   echo "get $FILENAME" >> $COMMAND
   echo "quit" >> $COMMAND

#---------------------------------------------------------- Run the ftp session
   if (-e $OUTPUT) rm $OUTPUT
   if (-e $FILENAME) rm $FILENAME
   ftp $SERVER < $COMMAND > $OUTPUT

#----------------------------------------------------------------------- Log it
   echo "ftp to $SERVER" >> $LOG
#---------------------------------------- Error log if file not there, and loop
   cat $OUTPUT >> $LOG
   if (!(-e $FILENAME)) then
      echo "Error on ftp to $SERVER - `date`" >> $ERRLOG
      cat $OUTPUT >> $ERRLOG
		echo "Extract attempted `date`" >> complete.$DONETHROUGH
      echo "Error on ftp to $SERVER - `date`" >> complete.$DONETHROUGH
      cat $OUTPUT >> complete.$DONETHROUGH
      continue
   endif

#------------------------------------------------------------------------------ 
# Get last detail line logged, which is last line in the last file which is
# not empty.  We use this approach rather than cat-ing all the log files
#------------------------------------------------------------------------------
   set lsr = `ls -l | grep "log" | grep -v " 0 " | tail -1`
	if ("$lsr" != "") then
		foreach f ($lsr)
	   	set lastlogfile = $f
		end
		set lastline = "`tail -1 $lastlogfile`"
	else
		set lastline = " "
	endif
#--------------------- Pull out lines since previously saved ones from ftp file
   $BIN/LogProcess $PACSCODE "$lastline\n" < $FILENAME > $T1
#--------------------------------------- grep out yesterdays stuff to .log file
   set MASK = `$BIN/OTime 1 "$FORMATSTRING"`
   grep $MASK $T1 >> $YESTERDAY.log
#------------------------------------------ grep out today's stuff to .log file
   set MASK = `$BIN/OTime 0 "$FORMATSTRING"`
   grep $MASK $T1 >> $TODAY.log

   set COUNT = `cat -n "$T1" | tail -1 | cut -c1-6`
   echo "Log for $SERVER,  $FILENAME, extracted $COUNT records" >> $LOG
	if ($COUNT > 0) echo "`date '+%Y-%m-%d %H:%M:%S'` $SERVER, $COUNT records" >> extract_counts

   rm $FILENAME $T1

#------------------------ Put in complete file if everything has been processed
   if (-e complete.$DONETHROUGH) rm complete.$DONETHROUGH
	echo "Last extract for $SERVER at `date`" > complete.$TODAYHR
	echo "Extracted  $COUNT records" >> complete.$TODAYHR
end

echo "GetLog completed `date`" >> $LOG
exit 0
