#!/bin/csh
#*******************************************************************************
# GetPRISLogs.csh                                                      R Moulton
# Script designed to retrieve Philips Broker RIS logs and organize them by
# days for analysis and review.  The files are retrieved 
# using FTP after midnight, then renamed to previous day.  The host must be
# listed in the server /etc/hosts file and have an entry in the user .netrc
# file giving the id and password for the FTP login.  This script is run as a
# cron job.
#*******************************************************************************
set STORE = "/distributor/other_logs/bgdea1/risif"

#---------------------------------- Must be logged in as spectrum to run GetLogs
if ($USER != spectrum) then
   echo "Must be user spectrum to run GetLogs"
   exit 1
endif

rm $TMP/GetPRIS* >& /dev/null

if (!(-d $STORE)) then
   echo "$STORE is not a valid directory"
   exit 1
endif

cd $STORE

set STOP = $TMP/stopGetPRISLogs
set LOGF = $LOG/GetPRISLogs.log
set ERRLOG = $LOG/GetPRISLogs_problems.log
set COMMAND = $TMP/GetPRISLogs.cmd
set OUTPUT = $TMP/GetPRISLogs.out
set T1 = GetPRISLogs.tmp

set TODAY   = `date '+%Y%m%d'`
set YESTERDAY = `$BIN/OTime 1 '%Y%m%d'`
   
touch $LOGF
if (!(-e $LOGF)) then
   echo "Unable to create $LOGF file"
   exit 1
endif


if (-e $STOP) then
   echo "Found $STOP file, exiting early" >> $LOGF 
   exit 0
endif

touch $ERRLOG
if (!(-e $ERRLOG)) then
   echo "Unable to create $ERRLOG file" >> $LOGF
   exit 1
endif

if ((-e from_ris.log.$YESTERDAY) && (-e to_ris.log.$YESTERDAY) && (-e pacs_ris.log.$YESTERDAY)) then
   echo "`date` Already done" >> $LOGF
   exit 0
endif

if (!(-e from_ris.log.$YESTERDAY) && !(-e to_ris.log.$YESTERDAY) && !(-e pacs_ris.log.$YESTERDAY)) then
   echo "Begin processing -- `date`" > $LOGF
endif

rm *_ris.log.bak >& /dev/null

#---------------------------------------------------- Generate ftp command file
echo "cd /var/opt/sectra/risif" > $COMMAND
if (!(-e from_ris.log.$YESTERDAY)) echo "get from_ris.log.bak" >> $COMMAND
if (!(-e to_ris.log.$YESTERDAY  )) echo "get to_ris.log.bak"   >> $COMMAND
if (!(-e pacs_ris.log.$YESTERDAY)) echo "get pacs_ris.log.bak" >> $COMMAND
echo "quit" >> $COMMAND

#---------------------------------------------------------- Run the ftp session
if (-e $OUTPUT) rm $OUTPUT
ftp bgdea1 < $COMMAND > $OUTPUT

#----------------------------------------------------------------------- Log it
echo "ftp to bgdea1" >> $LOGF
cat $OUTPUT >> $LOGF

#------------------------------------- rename needed files, error if not there
set HAD_ERRORS = "NO"

if (!(-e from_ris.log.$YESTERDAY)) then
   if (-e from_ris.log.bak) then
      mv from_ris.log.bak from_ris.log.$YESTERDAY
      echo "Got from_ris.log.$YESTERDAY from bgdea1" >> $LOGF
   else
      set HAD_ERRORS = "YES"
      echo "Error: Couldn't get from_ris.log - `date`" >> $ERRLOG
   endif
endif

if (!(-e to_ris.log.$YESTERDAY)) then
   if (-e to_ris.log.bak) then
      mv to_ris.log.bak to_ris.log.$YESTERDAY
      echo "Got to_ris.log.$YESTERDAY from bgdea1" >> $LOGF
   else
      set HAD_ERRORS = "YES"
      echo "Error: Couldn't get to_ris.log - `date`" >> $ERRLOG
   endif
endif

if (!(-e pacs_ris.log.$YESTERDAY)) then
   if (-e pacs_ris.log.bak) then
      mv pacs_ris.log.bak pacs_ris.log.$YESTERDAY
      echo "Got pacs_ris.log.$YESTERDAY from bgdea1" >> $LOGF
   else
      set HAD_ERRORS = "YES"
      echo "Error: Couldn't get pacs_ris.log - `date`" >> $ERRLOG
   endif
endif

if ("$HAD_ERRORS" == "YES") then
   cat $OUTPUT >> $ERRLOG
   echo "Errors occurred" >> $LOGF
   rm $TMP/GetPRIS* >& /dev/null
   exit 1
endif

rm $TMP/GetPRIS* >& /dev/null
echo "Completed successfully" >> $LOGF
exit 0
