#!/bin/csh

#*****************************************************************
# Hold.csh                                               R Moulton
# This script is used to gently take a study out of the main 
# image directories and move it to the problem directory.  The
# purpose is to allow studies which have generated errors to be
# pulled from the queue.  Protections are placed in the script to
# help avoid mistakes, and a display of demographics information
# is made to assist in identifying the study, and in case it is
# needed to get the study resent.  The general form of the 
# command is: Hold.csh server mode studydirectory
# It is assumed that the script will normally be implemented 
# through a script hold in the spectrum home directory.
#-----------------------------------------------------------------
# 6-27-2 R Moulton - Modified to update SQL pending info
# 08-30-02 R Moulton - Modified to add relay1 and relay2
# 2008 Aug 07 MSuchko - Stop deleting broadcat files.
#*****************************************************************

#------------------------------------------- Must be user spectrum
if (`whoami` != spectrum) then
   echo "Must be user spectrum to run Hold"
   exit 1
endif

#------------------------------------ Validate server, directories
set S = $SNAME
set MODE = "production"

set ROOT = $HOME/$MODE/$S
if (!(-d $ROOT)) then
   echo "$ROOT is not a valid directory"
   exit 1
endif

set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo "$BIN is not a valid directory"
   exit 1
endif
 
set LOGS = $ROOT/log
if (!(-d $LOGS)) then
   echo "$LOGS is not a valid directory"
   exit 1
endif

set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo "$TMP is not a valid directory"
   exit 1
endif

set IMGS = "/distributor/imgs"
if ($MODE == test) set IMGS = $ROOT/imgs
if (!(-d $IMGS)) then
   echo "$IMGS not a valid directory"
   exit 1 
endif

#---------------------------------------------- Server error log
set ERRLOG = "$LOGS/problem_log.$S"
set T1 = Hold.csh.temp

set STUDY = "$3"

#------------------------------------------------ No such study
if (!(-d $STUDY)) then
   echo "No such study directory"
   exit 1
endif

#--------------------------- Validate study directory up to imgs
@ i = 1
while (1)
   set a = `echo "$IMGS"  | cut -d'/' -f$i`
   set b = `echo "$STUDY" | cut -d'/' -f$i`
   if ("$a" != "$b") then
      echo "entered path not in images directory tree"
      exit 1
   endif
   if ("$a" == imgs) break
   @ i += 1
end

#--------------------------------- Not in problems directory
@ i += 1
set b = `echo "$STUDY" | cut -d'/' -f$i`
if ("$b" == problems) then
   echo "Can't put a study in problems directory on hold"
   exit 1
endif

#-------------------------------------- build target directory
set TARGET = "$IMGS/problems"
set HEADER = ""
while (1)
   set a = `echo "$STUDY" | cut -d'/' -f$i`
   set ST = $a
   set b = `echo "$a" | cut -c1`
   if ($b == 1 || $b == 2 || $b == 3 || $b == 4 || $b == 5 || \
       $b == 6 || $b == 7 || $b == 8 || $b == 9) break
   set TARGET = "$TARGET/$a"
   set HEADER = "$HEADER $a"
   @ i += 1
end
#------- Add date directory to target to keep multiples separate
set A = `date '+%Y%m%d%H%M'`
set TARGET = "$TARGET/$A"

@ i += 1
set a = `echo $STUDY | cut -d'/' -f$i`
if ($a != "") then
   echo "last directory should be study"
   exit 1
endif

cd $STUDY
echo ===========================================================
echo "Place study in hold file?"
echo ===========================================================
echo "Source-modality $HEADER"
echo "Study: $ST"

pwd
ls [1-9]*/[1-9]* >& /dev/null
if ($status != 0) then
   echo "No images in study"
   set PNAME = ""
   set PID = ""
   set ACCN = ""
else
#-------------------------------------------- show size of study
   set size = `du -k -s | cut -d'	' -f1`
   set units = KB
   if ($size > 1024) then
      @ size = $size / 1024
      set units = MB
   endif
   echo "$size $units"
#--------------------------------------------- Show demographics
   set image = `ls [1-9]*/[1-9]* | tail -1`
   $BIN/dcm_dump_file $image | egrep "0010 0010|0010 0020|0008 0050" > $T1
   set PNAME = "`cat $T1 | grep '0010 0010' | cut -d'/' -f5`"
   set PID   = "`cat $T1 | grep '0010 0020' | cut -d'/' -f5`"
   set ACCN  = "`cat $T1 | grep '0008 0050' | cut -d'/' -f5`"
   echo "Patient Name: $PNAME"
   echo "Patient ID:   $PID"
   echo "Accession:    $ACCN"
endif
echo
echo "(Y)es to continue"
set RESP = $<
set RESP = `echo $RESP | cut -c1`
if ("$RESP" != "Y" && "$RESP" != "y") then
   echo "No action taken"
   exit 0
endif
#------------------------------ If .dba exists, delete SQL entries
if (-e .dba) then
   set dba = `cat .dba`
   if ("$dba" != "") then
      echo "use image_log_db;" > $T1
      echo "DELETE FROM studylog WHERE dba = $dba;" >> $T1
      echo "DELETE FROM pending WHERE pending_dba = $dba;" >> $T1
      echo "DELETE FROM xmits WHERE study_dba = $dba;" >> $T1
      /usr/local/mysql/bin/mysql -u dealer -pdlrpw < $T1
      rm .dba $T1
   endif
endif
rm dest.*
#--------------------------------------------------- Actual move
mkdir -p $TARGET
cd ..
mv $ST $TARGET
#----------------------------------------------------- log entry
echo "== Study moved to problems directory `date` ==" >> $ERRLOG
echo "  $HEADER $ST" >> $ERRLOG
if ("$PNAME" != "") then
   echo "  Patient Name: $PNAME" >> $ERRLOG
endif
if ("$PID" != "") then
   echo "  Patient ID:   $PID"   >> $ERRLOG
endif
if ("$ACCN" != "") then
   echo "  Accession:    $ACCN"  >> $ERRLOG
endif
echo "Study files moved to problems directory"
exit 0
