#!/bin/csh

#************************************************************************
# Janitor.csh                         R Moulton, from a script by S Moore
#
# This script purges the todelete image subdirectories and deletes files
# which are more than a threshold age.  It also deletes directories 
# which have not been touched in a threshold amount of time.  It drives
# from the actual directory tree, and does not need the distributor.cfg
# file.  The script has two parameters, the server (e.g. bond), and the
# mode (production or test, default is production.)  Command syntax is:
#
#   Janitor.csh server mode
#
#-----------------------------------------------------------------------
# 12-7-1 R Moulton reduced DIRAGE from 24 to 1
#************************************************************************
#------------------------- Age thresholds to delete directories and files
set DIRAGE  = 1
set FILEAGE = 24

#--------------------------- Must be logged in as spectrum to run Janitor
if (`whoami` != "spectrum") then
   echo "Must be user spectrum to run Janitor"
   exit 1
endif

#--------------------------------- Server we are running on, default bond
set S = $SNAME

#----------------------------- mode we are running in, production or test
set MODE = production
if ("$2" != "") set MODE = "$2"
if ($MODE != "production" && $MODE != "test") then
   echo "$MODE must be production or test"
   exit 1
endif

#----------------------------- Validate appropriate server directory tree
set ROOT = $home/$MODE/$S
if (!(-d $ROOT)) then
   echo "$ROOT is not a valid directory"
   exit 1
endif

set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo "$BIN is not a valid directory"
   exit 1
endif

set LOGS = $ROOT/log
if (!(-d $LOGS)) then
   echo "$LOGS is not a valid directory"
   exit 1
endif

set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo "$TMP is not a valid directory"
   exit 1
endif

#--------------------------------------------------- Validate images root
set IMGS = /distributor/imgs
if ($MODE == "test") set IMGS = $ROOT/imgs
if (!(-d $IMGS)) then
   echo "$IMGS is not a valid directory"
   exit 1
endif
   
#--------------------------------------- Script lock, stop, and log files
set LOCK = $TMP/Janitor.csh.lock
set STOP = $TMP/stopJanitor
set QUIT = "-e $STOP"
set LOG  = $LOGS/Janitor.log

#----------------------- If lock file exists, exit... otherwise create it
if (-e $LOCK) then
   echo "$LOCK file exists, Janitor assumed to be already running"
   exit 0
endif
touch $LOCK
if (!(-e $LOCK)) then
   echo "Could not create $LOCK file"
   exit 1
endif

#-------------------------------------------------------- Create log file
if (-e $LOG) rm $LOG
touch $LOG
if (!(-e $LOG)) then
   echo "Could not create $LOG file"
   exit 1
endif

#---------------------------------------- Log start of Janitor operations
echo "Beginning Janitor operations -- `date`" >> $LOG

#----------------------------------------------------- Look for stop file
if ($QUIT) then
   date >> $LOG
   echo "Exiting early, $STOP file found" >> $LOG
   if (-e $LOCK) rm $LOCK
   exit 0
endif

echo "=====================================================" >> $LOG
echo "Purging X files -- `date`" >> $LOG
find $IMGS -type f -name ".broad*.X" -exec rm {} \;
echo "Finished purging X files" >> $LOG

#----------------------------------------------------- Look for stop file
if ($QUIT) then
   date >> $LOG
   echo "Exiting early, $STOP file found" >> $LOG
   if (-e $LOCK) rm $LOCK
   exit 0
endif

echo "=====================================================" >> $LOG
echo "Purging NULL routed studies -- `date`" >> $LOG
find $IMGS -type f -name "dest.NULL.sendto" -exec $BIN/PurgeNULL.csh {} $STOP $LOG \;
echo "Finished purging NULL routed files" >> $LOG

#----------------------------------------------------- Look for stop file
if ($QUIT) then
   date >> $LOG
   echo "Exiting early, $STOP file found" >> $LOG
   if (-e $LOCK) rm $LOCK
   exit 0
endif

echo "=====================================================" >> $LOG
echo "Purging main image directories -- `date`" >> $LOG
find $IMGS -type d -exec $BIN/PurgeDirectory.csh {} $DIRAGE $STOP $LOG $BIN \;
echo "Finished purging main image directories" >> $LOG

#----------------------------------------------------- Look for stop file
if ($QUIT) then
   if (-e $LOCK.Q) exit 0
   date >> $LOG
   echo "Exiting early, $STOP file found" >> $LOG
   if (-e $LOCK) rm $LOCK
   exit 0
endif

echo "=====================================================" >> $LOG
echo "Purging todelete files -- `date`" >> $LOG
find $IMGS/todelete -type d -name "[1-9]*" -prune -exec $BIN/PurgeStudy.csh {} $FILEAGE $STOP $LOG $BIN \;
echo "Finished purging todelete files" >> $LOG

#----------------------------------------------------- Look for stop file
if ($QUIT) then
   if (-e $LOCK.Q) exit 0
   date >> $LOG
   echo "Exiting early, $STOP file found" >> $LOG
   if (-e $LOCK) rm $LOCK
   exit 0
endif

echo "=====================================================" >> $LOG
echo "Purging todelete directories -- `date`" >> $LOG
find $IMGS/todelete -type d -exec $BIN/PurgeDirectory.csh {} $DIRAGE $STOP $LOG $BIN \;
echo "Finished purging todelete directories" >> $LOG

echo "Janitor operations complete -- `date`" >> $LOG

#---------------------------------------------------- delete lock file
if (-e $LOCK) rm $LOCK
exit 0
