#!/bin/csh

#************************************************************************
# MonitorQueue.csh                                     R Moulton/C Barlow
#
# Modified from Pending.csh. Calculates number of pending studies on the
# server and matches that against a time based acceptable maximum. Logs,
# and generates e-mail warning if over threshold.
# Command syntax is:
#
#  MonitorQueue.csh < Servername.cfg   
#
# But is run from D.csh as:
#
#  D.csh MonitorQueue      
#------------------------------------------------------------------------
# 10-3-07 Cliff Barlow - Removed all output and added email message
#       also added the passing of the threshold number for the email.
# 7/21/09 R Moulton - Integrated threshold calculation into this script,
#         eliminating Transmission.csh. Modified to run via D.csh.
#         Added log back.
#************************************************************************
#------------------------------------------------------------------------
# The live server is set by .cfg lines "Server,servername".  If it is
# set, other servers will ignore all .cfg lines until the end-of-file or
# the next Server line.
#------------------------------------------------------------------------
set LIVESERVER = ""

#---------------------------- Must be logged in as spectrum to run 
if ($user != "spectrum") then
   echo Must be user spectrum
   exit 1
endif

#---------------------------------------------- Server we are running for
set S = $SNAME 
#----------------------------- Validate appropriate server directory tree
set ROOT = $HOME/production/$S
if (!(-d $ROOT)) then
   echo "$ROOT is not a valid directory"
   exit 1
endif

set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo "$BIN is not a valid directory"
   exit 1
endif

set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo "$TMP is not a valid directory"
   exit 1
endif

set LOG = $ROOT/log
if (!(-d $LOG)) then
   echo "$LOG is not a valid directory"
   exit 1
endif

set A = `date '+%Y%m'`
set LOGFILE = $LOG/MonitorQueue$A.log
touch $LOGFILE

#---------------------------------------------------- Validate image root
set IMGS = /distributor/imgs
if (!(-d $IMGS)) then
   echo "$IMGS is not a valid directory"
   exit 1
endif

set LOCK = $TMP/MonitorQueue.csh.lock
if (-e $LOCK) then
    echo "`date` $LOCK file exists" >> $LOGFILE
    exit 0
endif
touch $LOCK
if (!(-e $LOCK)) then
    echo "`date` Could not create $LOCK file" >> $LOGFILE
    exit 1
endif

#---------------------------------------------- Calculate study threshold
set TLINE = `grep "^$SNAME" $BIN/Pending.cfg`
set H = `date +%H` 

set S1 = `echo "$TLINE" | cut -d',' -f2`
set E1 = `echo "$TLINE" | cut -d',' -f3`
set T1 = `echo "$TLINE" | cut -d',' -f4`
set S2 = `echo "$TLINE" | cut -d',' -f5`
set E2 = `echo "$TLINE" | cut -d',' -f6`
set T2 = `echo "$TLINE" | cut -d',' -f7`
set S3 = `echo "$TLINE" | cut -d',' -f8`
set E3 = `echo "$TLINE" | cut -d',' -f9`
set T3 = `echo "$TLINE" | cut -d',' -f10`

set THRESHOLD = 0

#-------------------------------------------------- Threshold calculation 
# If End is less than start, they are assumed to wrap around midnight
#------------------------------------------------------------------------
while (1)
   if ($S1 < $E1) then
      if ($H >= $S1 && $H < $E1) then
         set THRESHOLD = $T1
         break
      endif
   else
      if ($H >= $S1 || $H < $E1) then
         set THRESHOLD = $T1
         break
      endif
   endif
   if ($S2 < $E2) then
      if ($H >= $S2 && $H < $E2) then
         set THRESHOLD = $T2
         break
      endif
   else
      if ($H >= $S2 || $H < $E2) then
         set THRESHOLD = $T2
         break
      endif
   endif
   if ($S3 < $E3) then
      if ($H >= $S3 && $H < $E3) then
         set THRESHOLD = $T3
         break
      endif
   else
      if ($H >= $S3 || $H < $E3) then
         set THRESHOLD = $T3
         break
      endif
   endif
   echo "`date` Threshold calculation error" >> $LOGFILE
   break
end

@ totalstudies = 0
#------------------------------------------------------------------------
# This loop processes one line from the config table, which is std in.
# The same table can be used for all servers being used as distributors.
#------------------------------------------------------------------------
while (1)
#---------------------------------------------------- Get next table line
   set TABLELINE = $<
#---------------------- "end-of-file" on line signifies the table is done
   if ("$TABLELINE" == "end-of-file") break   
#------------------------------------------------------------------------
# Blank lines and lines with "#" in column 1 are ignored as comments
#------------------------------------------------------------------------
   if ("$TABLELINE" == "") continue
   set COMCHAR = `echo "$TABLELINE" | cut -c1`
   if ("$COMCHAR" == "#") continue
   if (`echo "$TABLELINE" | cut -d',' -f1` == "SetDefaults") continue

#------------------------------------------------------------------------
# Here we are looking at a detail line from the configuration file, which
# is of the form source,modality,name,server,queue,dest1,dest2... with up
# to five destinations. 
#------------------------------------------------------------------------
   set SOURCE   = `echo "$TABLELINE" | cut -d',' -f1`
   set MODALITY = `echo "$TABLELINE" | cut -d',' -f2`
   set NAME     = `echo "$TABLELINE" | cut -d',' -f3`
   set SERVER   = `echo "$TABLELINE" | cut -d',' -f4`
   set QUEUE    = `echo "$TABLELINE" | cut -d',' -f5`
#--------------------------------------------------------- Default values
   if ("$SERVER" == "") set SERVER = $S
   if ("$QUEUE"  == "") set QUEUE  = 1

#----------- If source modality root directory does not exist, we're done
   set DIR = $IMGS/$SOURCE/$MODALITY
   set SRC = $SOURCE/$MODALITY
   if ($MODALITY == "") then
      set DIR = $IMGS/$SOURCE
      set SRC = $SOURCE
   endif
   if (!(-d $DIR)) continue

   cd $DIR

#--------------------------- Make sure there are some studies in modality
   ls [1-9]* >& /dev/null
   if ($status != 0) continue
#----------------------------------------- Pass studies in base directory
   foreach study ([1-9]*)
      set imagesexist = NO

#--------- Validate that it is a directory, and that files exist under it
      if (!(-d $study)) continue
      ls $study/[1-9]* >& /dev/null
      if ($status != 0) continue
#------------------------------ Watch for split routings not in this queue
      if (-e $study/DealerQueue) then
         if ($QUEUE != "`cat $study/DealerQueue`") continue
      endif
         
      foreach series ($study/[1-9]*)
         if (!(-d $series)) continue   
         ls $series/[1-9]* >& /dev/null
         if ($status == 0) then
            set imagesexist = YES
            break
         endif
      end
      if($imagesexist == "YES") then
          @ totalstudies += 1
      endif 
   end                           # End of foreach loop for studies
end                              # End of while loop reading config table



echo "`date` pending studies: $totalstudies - threshold: $THRESHOLD" >> $LOGFILE

if($totalstudies > $THRESHOLD) then
    cp $BIN/Pending.hdr $BIN/Pending.msg
    echo "$S has $totalstudies pending studies" >> $BIN/Pending.msg
    $BIN/SendAlert $BIN/Pending.msg
    echo "*** Alert sent ***" >> $LOGFILE
endif

endif

if (-e $LOCK) then
  rm $LOCK
endif

exit 0
