#!/bin/csh

#************************************************************************
# MoveStudy.csh                                                 R Moulton
#
# This string examines a study to determine what images in it can be
# moved to the todelete directory.  Method is as follows:
#  1. Must be .broadcast.rslt.code for each dest.code file
#  2. Must be entries in all of these files for this image
#  3. All entries must have xmit OK flag (0)
# If all these conditions are met, the image is moved to the todelete
# directory.  If all images are moved to todelete directory, script
# returns 0.  Otherwise 1 is returned.  This script is run only after
# a dicom_broadcast which returns an OK status, so we know that there
# are images, dest, and rslt files out there.
# If all the images are moved off a row is added to the studylog table.
#
# Input parameters are (1) images directory for study, (2) todelete
# directory for study, (3) server, (4) external queue name for Dealer, 
# (5) root dir, (6) time hack for end of transmission, (7) Patient Name,
# (8) Patient ID, (9) Accession number (10) internal queue number,
# (11) image source, (12) image modality, (13) Study date, (14) Study
# time, (15) Institution name, (16) Station Name, (16) DOB
#------------------------------------------------------------------------
# 01-10-02 R Moulton Added first and last image arrival time to STUDY ms
# 05-16-02 R Moulton Added studylog sql processing.
# 05-21-02 R Moulton Added xmits table to sql processing.
# 06-04-02 R Moulton Added columns to xmits table.
# 06-05-02 R Moulton Added existence check to image file mv command.
# 10-05-02 R Moulton Removed STUDY message
# 03-31-04 R Moulton Added cp of dest* files to todelete directory
# 05-31-05 R Moulton Added rm of .br* files when move completed.
# 02-21-06 R Moulton Added check for dest file non-empty before move
# 02-21-06 R Moulton Added check for dest file non-empty before move
# 08-27-08 M Suchko  Add info to error message
#************************************************************************
set S = $3
set SQ = $4
set EOT = $6
set PNAME = "$7"
set PID = "$8"
set ACCN = "$9"
set IQ = $10
set SOURCE = $11
set MODALITY = $12
set STDT = "$13$14"
set INSTN = "$15"
set STNM = "$16"
set DOB = "$17"
#----------------------------------------------- Validate study directory
set study = $1
if (!(-d $study)) then
   echo "    MoveStudy: $study directory does not exist"
   echo "    cannot clear xmitted images."
   exit 1
endif
#-------------------------------- MoveStudy lock file to avoid collisions
# We were just touching the file, but had problems with exact timing
# collisions.  Here we put our external queue in the file, wait three 
# seconds, then verify that it is still there.  Even if two processes
# create the same file at almost the same instant, one of them will "win"
# the race to have its external queue in the file; that's the one we run
#------------------------------------------------------------------------
set lockfile = $study/MoveStudy.lock
if (-e $lockfile) exit 0
echo "$SQ" > $lockfile
sleep 3
if (`cat $lockfile` != "$SQ") exit 0
#-------------------------------- Validate root, bin, and tmp directories
set ROOT = $5
if (!(-d $ROOT)) then
   echo "    MoveStudy: $ROOT directory does not exist"
   echo "    cannot clear xmitted images."
   goto exit1
endif

set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo "    MoveStudy: $BIN directory does not exist"
   echo "    cannot clear xmitted images."
   goto exit1
endif
set PFix = $BIN/PFix

set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo "    MoveStudy: $TMP directory does not exist"
   echo "    cannot clear xmitted images."
   goto exit1
endif
#------------------------------- Create/Validate study todelete directory
set move = $2
mkdir -p $move
if ($status != 0) then
   echo "    MoveStudy: unable to create $move directory"
   echo "    cannot clear xmitted images."
   goto exit1
endif
#---------------------------------- copy dest files to todelete directory
cp $study/dest.*.sendto $move
#--------------------------------------------------------------- tempfile
set temp = $study/MoveStudy.temp.$4
#------------------------------------------ image arrival time temp files
set fiaf = $study/MoveStudy.FIA.temp
set liaf = $study/MoveStudy.LIA.temp

cd $study
#-----------------------------------------------------------------------
# Here we are looking to see if .rslt files exists for all the dest
# files.  If not, all the xmits have not even been tried, and nothing
# can be moved yet. Also exit if dest file is zero length
#-----------------------------------------------------------------------
@ dest_count = 0
foreach DEST (dest.*.sendto)
   if (-z "$DEST") goto exit1
   @ dest_count += 1
   set dest = `echo "$DEST" | cut -d'.' -f2`
   if (!(-e .broadcast.rslt.$dest)) goto exit1
end

#-----------------------------------------------------------------------
# now we pass the images, looking for them in each rslt file, and 
# collecting the result records in the temp file.
#-----------------------------------------------------------------------
@ totalimages = 0
@ movedimages = 0
set FIA = "99999999999999"
set LIA = "00000000000000"
#-------------------------- If FIA and LIA have been cached, reload them
if (-e $fiaf) set FIA = `cat "$fiaf"`
if (-e $liaf) set LIA = `cat "$liaf"`

foreach series ([1-9]*)
   set MADESERIESDIRECTORY = NO
   pushd $series
   set POPPED = NO
   ls [1-9]* >& /dev/null
   if ($status != 0) then
      popd
      continue
   endif
   foreach image_file ([1-9]*)
      if ($POPPED == NO) then
         popd
         set POPPED = YES
      endif
      set image = "$series/$image_file"
      @ totalimages += 1
      set OKTOMOVE = YES
      if (-e $temp) rm $temp
      foreach rslt (.broadcast.rslt.*)
#-------------------------------------------- Skip (and delete .X files)
			if (`echo "$rslt" | cut -d'.' -f5` == "X") then
			   rm $rslt
				continue
			endif
#-----------------------------------------------------------------------
# In the grep, we are looking for two things: (1) the image, which is
# field one, and a result of 0, which is field 4.  We are not 
# guarranteed that another field might not start off with a zero, but
# this does not appear to happen.  At any rate, looking for the zero
# will reduce the number of lines XmitsOK has to process.  It will also
# cause most of the cases which do not have a successful xmit to break
# out here, another time saver.
#-----------------------------------------------------------------------
         cat $rslt | grep "$image" | grep '\^0' >> $temp
         if ($status != 0) then
            set OKTOMOVE = NO
            break
         endif
      end
      if ($OKTOMOVE == NO) continue
#--------- This subscript checks the tempfile to see if all xmits were OK
      $BIN/XmitsOK.csh < $temp
      if ($status != 0) continue
#---------------------------------------------------- image is OK to move
      if ($MADESERIESDIRECTORY == NO) then
         mkdir -p $move/$series
         if ($status != 0) then
            echo "    Not able to make $move/$series directory"
            echo "    Images not moved"
            goto exit1
         endif
         set MADESERIESDIRECTORY = YES
      endif
      if (-e $image) then
         set ctime = `"$BIN"/Itime "$image"`
	 $BIN/Iftime.csh $ctime $FIA
	 if ($status == 2) then
	    set FIA = $ctime
	    echo $FIA > $fiaf
	 endif
	 $BIN/Iftime.csh $ctime $LIA
	 if ($status == 1) then
	    set LIA = $ctime
	    echo $LIA > $liaf
	 endif
         if (-e $image) mv $image $move/$series
         if ($status != 0) then
            echo "    Unable to move $image"
         else
            @ movedimages += 1
         endif
      endif
   end
end
if (-e $temp) rm $temp
if ($totalimages != $movedimages) then
   echo "    $movedimages of $totalimages images moved to todelete directory"
   goto exit1
endif
#----------------------------------------- All images now sent & moved off
echo "    All $totalimages images moved to todelete directory"
#-------------------------------------------------------------------------
# Send STUDY message to monitor. 
#---------------------------------- Time hacks divided into date and time 
 set fadate = `echo "$FIA" | cut -c1-8`
 set fatime = `echo "$FIA" | cut -c9-14`
 set ladate = `echo "$LIA" | cut -c1-8`
 set latime = `echo "$LIA" | cut -c9-14`
 set edate = `echo "$EOT" | cut -c1-8`
 set etime = `echo "$EOT" | cut -c9-14`
#------------------------------------------------------------------------
# Now get the number of images, using one of the results files
#------------------------------------------------------------------------
set map = `ls .broadcast.map.* | tail -1`
set dest = `echo "$map" | cut -d'.' -f4`
set ndests = `cat -n "$map" | tail -1 | cut -c1-7`
set nxmits = `cat -n .broadcast.rslt."$dest" | tail -1 | cut -c1-7` 
@ nimages = $nxmits / $ndests
#--------------------------------------------- Done with .broadcast files
rm .broadcast.*
#-------------------------------- Pull study id from study directory name
@ i = 4
while (1)
   @ i += 1
   set st = `echo "$study" | cut -d'/' -f$i`
   if ("$st" == "") break
   set study_id = "$st"
end

#------------------------------------------------------------------------
# This code will send a Check Study e-mail if the tag file is present.
# Sends only for studies which have valid looking pid and accn, a DOS of
# today, and which are being sent to Imageweb
#------------------------------------------------------------------------
if (-e $TMP/CheckStudy) then
   if (-e dest.BH_LTA.sendto) then
      $BIN/CheckStudy $PID $ACCN $STDT
      if ($status == 0) then
         rm $TMP/CheckStudy
         set MSG = $BIN/CheckStudy.msg
         cp $BIN/CheckStudy.hdr $MSG
         echo "Pname = $PNAME" >> $MSG
         echo "PID   = $PID" >> $MSG
         echo "DOB   = $DOB" >> $MSG
         echo "ACCN  = $ACCN" >> $MSG
         echo "MOD   = $MODALITY" >> $MSG
         echo "DOS   = $STDT" >> $MSG
         $BIN/SendAlert $MSG
      endif
   endif 
endif
#------------------------------------------------------------------------
# This code will add a row to a Mysql database table to log the study
# completion.  First, look for a stopSQL tmp file.
#------------------------------------------------------------------------
if (-e $TMP/stopSQL) goto exit0
#------------------------------------------------------------------------
# We make a file in the move directory to hold the SQL insert command.  
# If the file already exists, this study has already been logged.  If
# not, make it quickly, to avoid conflicts with other Dealers.
#------------------------------------------------------------------------
# sql consists of a single insert into studylog for the study and an
# insert into xmits for each destination the study went to.
#------------------------------------------------------------------------
set SQL = $move/StudySQL.tmp
if (-e $SQL) goto exit0
touch $SQL
#------------------------------------ Pull study size from move directory
set kbytes = `du -k -s $move | cut -d'	' -f1`
#-------------------------------------------------------------- Build SQL
echo "USE image_log_db;" > $SQL
#----------------- If .dba exists, do conditional delete of these records
if (-e .dba) then
   set dba = `cat .dba`
   if ("$dba" != "") then
      echo "DELETE FROM studylog WHERE dba = $dba;" >> $SQL
      echo "DELETE FROM pending WHERE pending_dba = $dba;" >> $SQL
      echo "DELETE FROM xmits WHERE study_dba = $dba;" >> $SQL
   endif
endif
#----------------------------------------------- Insert main study record
echo "INSERT INTO studylog VALUES (NULL," >> $SQL
echo "'$SNAME',$IQ,'$study_id','$INSTN','$STNM'," >> $SQL
echo "'$SOURCE','$MODALITY'," >> $SQL
echo "'$PID','$PNAME','$ACCN'," >> $SQL
echo "$nimages,$kbytes,$STDT," >> $SQL
echo "$FIA,$LIA,$EOT);" >> $SQL
#------------------------------------- First line of multiple dest insert
echo "INSERT INTO xmits VALUES " >> $SQL
@ i = 0
foreach dest (dest.*.sendto)
   @ i += 1
   set XMT = `cat "$dest"`
   if ($i == $dest_count) then
      echo "(LAST_INSERT_ID(),$XMT);" >> $SQL
   else
      echo "(LAST_INSERT_ID(),$XMT)," >> $SQL
   endif
end
#---------------------------------------------- insert row into SQL table
/usr/local/mysql/bin/mysql -u dealer -pdlrpw image_log_db < $SQL
#------------------------------------------ Log and Alert on mysql errors
if ($status != 0) then
   echo "`date` ============== MySQL error in MoveStudy" >> $LOG/MysqlError.log
   echo " " >> $LOG/MysqlError.log
   cat $SQL  >> $LOG/MysqlError.log
   echo " " >> $LOG/MysqlError.log
   cp $BIN/MySQL.hdr $BIN/MySQL.msg
   echo "on $SNAME" >> $BIN/MySQL.msg
   echo "$SQL" >> $BIN/MySQL.msg
   $BIN/SendAlert $BIN/MySQL.msg
endif 

#----------------------------------------------- Clear lock file and exit
exit0:
   if (-e $lockfile) rm $lockfile
   exit 0
exit1:
   if (-e $lockfile) rm $lockfile
   exit 1
