#!/bin/csh

#************************************************************************
# Pending.csh                                                  R Moulton
#
# In live feed, this script examines image directories created by the 
# dicom receiver and outputs a list of pending studies in them to std 
# out.  In monitor feed, it sends status messages including the number
# of pending images and total bytes (on a source-modality basis) to the
# monitor.  The script has three parameters, the server name which is to
# be processed, (default is bond), the feed (live or monitor), and the 
# processing mode (test or production, default is production. The 
# distributor configuration file distributor.cfg is system in.  When 
# Pending is set up as a cron job, the standard output should be 
# re-directed to a log file.
# Command syntax is:
#
#  Pending.csh servername feed mode < Distributor.cfg        
#------------------------------------------------------------------------
# 12-3-01 R Moulton - Added monitor feed
#************************************************************************
#------------------------------------------------------------------------
# The live server is set by .cfg lines "Server,servername".  If it is
# set, other servers will ignore all .cfg lines until the end-of-file or
# the next Server line.
#------------------------------------------------------------------------
set LIVESERVER = ""

#---------------------------- Must be logged in as spectrum to run Router
if ($user != "spectrum") then
   echo Must be user spectrum
   exit 1
endif

#---------------------------------------------- Server we are running for
set S = $SNAME 

#--------------------------------------- Feed is param 2, live or monitor
set FEED = live
if ("$2" != "") set FEED = $2
if ("$FEED" != "live" && "$FEED" != "monitor") then
   echo "$FEED must be live or monitor"
   exit 1
endif
#-------------------------------------------------------------- shorthand
set LIVE = "$FEED == live" 
set MONITOR = "$FEED == monitor"

#------------------ mode we are running in is param 3, production or test
set MODE = production
if ("$3" != "") set MODE = $3
if ("$MODE" != "production" && "$MODE" != "test") then
   echo "$MODE must be production or test"
   exit 1
endif

#----------------------------- Validate appropriate server directory tree
set ROOT = $HOME/$MODE/$S
if (!(-d $ROOT)) then
   echo "$ROOT is not a valid directory"
   exit 1
endif

set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo "$BIN is not a valid directory"
   exit 1
endif

set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo "$TMP is not a valid directory"
   exit 1
endif

#---------------------------------------------------- Validate image root
set IMGS = /distributor/imgs
if ($MODE == "test") set IMGS = $ROOT/imgs
if (!(-d $IMGS)) then
   echo "$IMGS is not a valid directory"
   exit 1
endif

#--------------------------------------------- Lock file, if monitor feed
if ($MONITOR) then
   set LOCK = $TMP/Pending.csh.lock
   if (-e $LOCK) then
      echo "$LOCK file exists, Pending assumed to be running"
      exit 0
   endif
   touch $LOCK
   if (!(-e $LOCK)) then
      echo "Could not create $LOCK file"
      exit 1
   endif
endif

#------------------------------------------------------------- temp files
set T1 = $TMP/Pending.$MODE.temp1

#--------------------------------------- Initial values of input defaults
set DEFISERVER = bond
set DEFIQUEUE  = 1

if ($MONITOR) then
   echo "Begin Pending.$S  `date`"
   $BIN/Message "Pending.csh" $S BEGIN_SCRIPT
else
   echo "Searching for pending images..."
endif
@ totalstudies = 0
@ substudies = 0
@ totalkbytes = 0
@ subkbytes = 0
#------------------------------------------------------------------------
# This loop processes one line from the config table, which is std in.
# The same table can be used for all servers being used as distributors.
#------------------------------------------------------------------------
while (1)
#---------------------------------------------------- Get next table line
   set TABLELINE = $<
#---------------------- "end-of-file" on line signifies the table is done
   if ("$TABLELINE" == "end-of-file") break   
#------------------------------------------------------------------------
# Blank lines and lines with "#" in column 1 are ignored as comments
#------------------------------------------------------------------------
   if ("$TABLELINE" == "") continue
   set COMCHAR = `echo "$TABLELINE" | cut -c1`
   if ("$COMCHAR" == "#") continue
#------------------------------------------------------------------------
# Server line sets live server.  If live server is set and is not our
# server, skip all other processing.
#------------------------------------------------------------------------
   if (`echo "$TABLELINE" | cut -d',' -f1` == "Server") then
      set LIVESERVER = `echo "$TABLELINE" | cut -d',' -f2`
   endif
   if ("$LIVESERVER" != "" && "$LIVESERVER" != $S) continue
#------------------------------------ SetDefaults sets new default server
   if (`echo "$TABLELINE" | cut -d',' -f1` == "SetDefaults") then
      set DEFISERVER = `echo "$TABLELINE" | cut -d',' -f2`
      set DEFIQUEUE = `echo "$TABLELINE" | cut -d',' -f3`
      echo 
      echo ==============================================
      echo "  $DEFISERVER $DEFIQUEUE"
      echo ==============================================
      continue
   endif

#------------------------------------------------------------------------
# Here we are looking at a detail line from the configuration file, which
# is of the form source,modality,name,server,queue,dest1,dest2... with up
# to five destinations. 
#------------------------------------------------------------------------
   set SOURCE   = `echo "$TABLELINE" | cut -d',' -f1`
   set MODALITY = `echo "$TABLELINE" | cut -d',' -f2`
   set NAME     = `echo "$TABLELINE" | cut -d',' -f3`
   set SERVER   = `echo "$TABLELINE" | cut -d',' -f4`
   set QUEUE    = `echo "$TABLELINE" | cut -d',' -f5`
#--------------------------------------------------------- Default values
   if ("$SERVER" == "") set SERVER = $DEFISERVER
   if ("$QUEUE"  == "") set QUEUE  = $DEFIQUEUE 

   if ($SERVER != $S) continue

#----------- If source modality root directory does not exist, we're done
   set DIR = $IMGS/$SOURCE/$MODALITY
   set SRC = $SOURCE/$MODALITY
   if ($MODALITY == "") then
      set DIR = $IMGS/$SOURCE
      set SRC = $SOURCE
   endif
   if (!(-d $DIR)) continue

   echo "Searching $SRC ..."
   cd $DIR

#--------------------------- Make sure there are some studies in modality
   ls [1-9]* >& /dev/null
   if ($status != 0) continue
#----------------------------------------- Pass studies in base directory
   foreach study ([1-9]*)
      set imagesexist = NO
      set image = "empty"

#--------- Validate that it is a directory, and that files exist under it
      if (!(-d $study)) continue
      ls $study/[1-9]* >& /dev/null
      if ($status != 0) continue
#------------------------------ Watch for split routings not in this queue
      if (-e $study/DealerQueue) then
         if ($QUEUE != "`cat $study/DealerQueue`") continue
      endif
         
      foreach series ($study/[1-9]*)
         if (!(-d $series)) continue   
         ls $series/[1-9]* >& /dev/null
         if ($status == 0) then
            set imagesexist = YES
#---------------------------------------- Save an image name for later use
            set image = `ls $series/[1-9]* | tail -1` 
            break
         endif
      end
      if ($imagesexist == NO) continue
#------------------------------------------------------------------------
# Get one of the images, extract dicom demographics, display Message
#------------------------------------------------------------------------
      $BIN/dcm_dump_file $image | egrep "0010 0010|0010 0020|0008 0050" > $T1
      set PNAME = "`cat $T1 | grep '0010 0010' | cut -d'/' -f5`"
      set PID   = "`cat $T1 | grep '0010 0020' | cut -d'/' -f5`"
      set ACCN  = "`cat $T1 | grep '0008 0050' | cut -d'/' -f5`"
      if ("$PNAME" == "")  set PNAME = blank
      if ("$PID" == "")    set PID   = blank
      if ("$ACCN" == "")   set ACCN  = blank

      set kbytes = `du -k -s $study | cut -d'	' -f1`
      set temp = "`ls -l $image`"
      set DATE = ""
      @ i = 0
      foreach word ($temp)
         @ i += 1
         if ($i == 6 || $i == 7 || $i == 8) set DATE = "$DATE $word"
      end

#-------------------------------------------------- convert to visual units
      @ size = $kbytes
      set units = KB    
      if ($size > 1024) then
         @ size = $size / 1024
         set units = MB
      endif
      if ($size > 1024) then
         @ g = $size / 1024
         @ d = $size - ($g * 1024)
         @ d = $d / 102
         set size = $g.$d
         set units = GB
      endif

      echo "=== Study :: $study"
      echo "  $size $units arrived=$DATE from: $NAME for Name=$PNAME PID=$PID AN=$ACCN" 
      pushd $study >& /dev/null
      ls dest.* >& /dev/null
      if ($status != 0) then
         echo "  Not routed"
      else
         ls dest.*
      endif
      popd >& /dev/null
      @ totalstudies += 1
      @ substudies +=1
      @ totalkbytes += $kbytes
      @ subkbytes += $kbytes
   end                           # End of foreach loop for studies
   if ($MONITOR) $BIN/Message "$NAME" $SERVER$QUEUE PENDING $substudies $subkbytes
   @ substudies = 0
   @ subkbytes = 0
end                              # End of while loop reading config table

if (-e $T1) rm $T1
echo "Finished searching"

#-------------------------------------------------- convert to visual units
      @ size = $totalkbytes
      set units = KB    
      if ($size > 1024) then
         @ size = $size / 1024
         set units = MB
      endif
      if ($size > 1024) then
         @ g = $size / 1024
         @ d = $size - ($g * 1024)
         @ d = $d / 102
         set size = $g.$d
         set units = GB
      endif

if ($totalstudies == 0) then
    echo "  No pending images found"
else
   echo "$totalstudies studies, $size $units"
endif
#-------------------------------------------------------------- gross estimate
@ totk = `du -k -s /distributor/imgs | cut -d'	' -f1`
@ delk = `du -k -s /distributor/imgs/todelete | cut -d'	' -f1`
@ prbk = `du -k -s /distributor/imgs/problems | cut -d'	' -f1`
@ size = $totk - $delk - $prbk
set units = KB    
if ($size > 1024) then
   @ size = $size / 1024
   set units = MB
endif
if ($size > 1024) then
   @ g = $size / 1024
   @ d = $size - ($g * 1024)
   @ d = $d / 102
   set size = $g.$d
   set units = GB
endif
echo "usage = $size $units"

set units = KB    
if ($delk > 1024) then
   @ delk = $delk / 1024
   set units = MB
endif
if ($delk > 1024) then
   @ g = $delk / 1024
   @ d = $delk - ($g * 1024)
   @ d = $d / 102
   set delk = $g.$d
   set units = GB
endif
echo "todelete = $delk $units"

if ($MONITOR) then
   echo "End Pending.$S  `date`"
   $BIN/Message "Pending.csh" $S END_SCRIPT
   rm $LOCK
endif
exit 0
