#!/bin/csh
#***********************************************************************
# PurgeStudy.csh                                               R Moulton
#
# This script deletes files from a study which have not been touched in
# a given age threshold.  The age threshold has a default provided on 
# the command line, and can be overidden if an entry in the Janitor.cfg
# file matches a dest* file in the study directory.  This script is 
# called from other scripts; the parameters are assumed valid.
#
# PurgeStudy.csh studydir age lockfile logfile bin
#-----------------------------------------------------------------------
#***********************************************************************

#--------------------------------- Load command line args into mnemonics
set STUDY = $1
set AGE   = $2
set STOP  = $3
set LOG   = $4
set BIN   = $5

#-------------------------------------------------- Check for early stop
# Because this is called from a find command, the find will continue to
# pass entries after the stop file has been created, so we put in a "Q"
# stop file to silence subsequent log file entries.  If the stop file
# is then removed, the Q file will be removed also.
#-----------------------------------------------------------------------
if (-e $STOP) then 
   if (-e $STOP.Q) exit 1
   echo "*** Stopping early, $STOP file found -- `date` ***" >> $LOG
   touch $STOP.Q
   exit 1
endif
rm -f $STOP.Q 

cd $STUDY
if (-e $BIN/Janitor.cfg) then        # If Janitor.cfg exists...
#-----------------------------------------------------------------------
# We pull the source from the pathname and look for a match in .cfg file
# This may raise (but never lower) the age threshold for files in the 
# study.  If the new age threshold is "hold", we don't clear the study 
# files at all.
#-----------------------------------------------------------------------
   set source = `pwd | cut -d'/' -f5`
   set newage = `cat $BIN/Janitor.cfg | grep "^Source:$source" | cut -d'=' -f2`
   if ("$newage" == "hold") exit 0           # hold studies indefinitely
   if ($newage > $AGE) set AGE = $newage     # hold longer
#-----------------------------------------------------------------------
# We look for dest* files.  If found, we pull the destinations and look
# for a match in the Janitor.cfg file. 
#-----------------------------------------------------------------------
   ls dest.*.sendto >& /dev/null     # If there are any dest* files...
   if ($status == 0) then
      foreach ddir (dest.*.sendto)
         set dst = `echo "$ddir" | cut -d'.' -f2`
         set newage = `cat $BIN/Janitor.cfg | grep "^Dest:$dst" | cut -d'=' -f2`
         if ("$newage" == "hold") exit 0           # hold studies indefinitely
         if ($newage > $AGE) set AGE = $newage     # hold longer
      end
   endif
endif

echo "=== $AGE - $STUDY" >> $LOG
#------------------------------------------ Purge the files in the study
find . -type f -name "[1-9]*" -exec $BIN/PurgeFile.csh {} $AGE $STOP $LOG $BIN \;

exit 0

