#!/bin/csh

#************************************************************************
# Router.csh                                                    R Moulton
#
# This script examines image directories created by the dicom receiver
# and places tag files in them indicating the destinations to which they
# are to be broadcast, with names in the format D-destination code.tosend
# This is done for all queues by one script.
# The script has one parameter the processing mode (test or production, 
# default is production. The distributor configuration file distributor.cfg  
# is system in.  Command syntax is:
#
#  Router.csh mode  
#------------------------------------------------------------------------
# 11-14-01 R Moulton - Added support for live server line in .cfg
# 5-8-2    R Moulton - Modified to eliminate servername parameter in 
#                      invocation and in call to Analyze.csh.  Also to
#                      handle two digit queue numbers
# 6-26-2 R Moulton - Added support for SQL entry at Routing time by
#                    passing additional parameters to Analyze
# 08-30-02 R Moulton - Modified to add relay1 and relay2
# 09-26-02 R Moulton - Fixed bug caused by very large studies
# 10-07-02 R Moulton - Commented out IMAGE_RECEIVED message to monitor
# 07-09-03 R Moulton - Rewrote to process directories against .cfg
#                      instead of .cfg against directories.
# 11-14-03 R Moulton - Reactivated IMAGE_RECEIVED and added new items
# 06-29-07 R Moulton - Removed Messages
# 08-19-08 R Moulton - Change grep to prevent searching directories unnecessarily.
# 2009 02 25 M.Suchko - square bracket fix for INSTN
#************************************************************************
#---------------------------- Must be logged in as spectrum to run Router
if (`whoami` != "spectrum") then
   echo "Must be user spectrum to run Router"
   exit 1
endif

set S = $SNAME

#------------------ mode we are running in is param 2, production or test
set MODE = production
if ("$1" != "") set MODE = $1
if ("$MODE" != "production" && "$MODE" != "test") then
   echo "$MODE must be production or test"
   exit 1
endif

#----------------------------- Validate appropriate server directory tree
set ROOT = $HOME/$MODE/$S
if (!(-d $ROOT)) then
   echo "$ROOT is not a valid directory"
   exit 1
endif

set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo "$BIN is not a valid directory"
   exit 1
endif

set LOGS = $ROOT/log
if (!(-d $LOGS)) then
   echo "$LOGS is not a valid directory"
   exit 1
endif

set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo "$TMP is not a valid directory"
   exit 1
endif

#---------------------------------------------------- Validate image root
set IMGS = /distributor/imgs
if ($MODE == "test") set IMGS = $ROOT/imgs
if (!(-d $IMGS)) then
   echo "$IMGS is not a valid directory"
   exit 1
endif

#----------------------------------------------------- Validate .cfg file
set CFG = $BIN/$CSNAME.cfg
if (!(-e $CFG)) then
   echo "$CFG does not exist"
	exit 1
endif

#----------------------------------- Script lock and stoprouter tag files
set LOCK = $TMP/Router.csh.lock   
set STOP = $TMP/stopRouter
set QUIT = "-e $STOP"

#----------------------- If lock file exists, exit... otherwise create it
if (-e $LOCK) exit 0
touch $LOCK
if (!(-e $LOCK)) then
   echo "Could not create $LOCK file"
   exit 1
endif

#--------------------------------------------------------- Open log files
set LOG = $LOGS/Router.log
if (-e $LOG) rm $LOG
touch $LOG
if (!(-e $LOG)) then
   echo "Unable to create $LOG file"
   exit 1
endif

set ERRLOG = $LOGS/Router_problems.log
touch $ERRLOG
if (!(-e $ERRLOG)) then
   echo "Unable to create $ERRLOG file"
   exit 1
endif

#------------------------------------------------------------- temp files
set T1 = $TMP/Router.temp1

echo "Begin Router.$S -- `date`" >> $LOG

#------------------------------------------------------------------------
# This loop processes one directory in the imgs tree, looking for a 
# matching entry in the configuration file. It is actually two loops, but
# we treat it as one (indent-wise)
#------------------------------------------------------------------------
@ routings = 0
cd $IMGS
#----------------------------------------------------- source directories
foreach sdir (*)
if (!(-d $sdir)) continue
grep "^$sdir," $CFG >& /dev/null
if ($status == 1) continue
#--------------------------------------------------- modality directories
ls $IMGS/$sdir/* >& /dev/null
if ($status != 0) continue
cd $IMGS/$sdir
foreach mdir (*)
   cd $IMGS/$sdir
   if (!(-d $mdir)) continue
   set TABLELINE = `$BIN/RouterLine.csh $sdir $mdir $CFG NO`
   set STAT = $status
   set QFILE = NO
   if ($STAT == 1) continue
   if ($STAT == 2) then
      set QFILE = YES
   endif
#---------------------------------------------- Check for stoprouter file
   if ($QUIT) then
      echo "Exiting early, $QUIT file found"
      rm $LOCK
      exit 0
   endif

#------------------------------------------------------------------------
# Here we are looking at a detail line from the configuration file, which
# is of the form source,modality,name,server,queue,dest1,dest2... with up
# to five destinations. First, we validate the detail line fields
#------------------------------------------------------------------------

#-------------------------------------------- Pull values from table line
   set SOURCE   = `echo "$TABLELINE" | cut -d',' -f1`
   set MODALITY = `echo "$TABLELINE" | cut -d',' -f2`
   set NAME     = `echo "$TABLELINE" | cut -d',' -f3`
   set SERVER   = `echo "$TABLELINE" | cut -d',' -f4`
   set QUEUE    = `echo "$TABLELINE" | cut -d',' -f5`
   set D1       = `echo "$TABLELINE" | cut -d',' -f6`
   set D2       = `echo "$TABLELINE" | cut -d',' -f7`
   set D3       = `echo "$TABLELINE" | cut -d',' -f8`
   set D4       = `echo "$TABLELINE" | cut -d',' -f9`
   set D5       = `echo "$TABLELINE" | cut -d',' -f10`

#------------------------------ Destinations combined for loop processing
   set DESTINATIONS = "$D1 $D2 $D3 $D4 $D5"      

#------------------------------------------------------------------------
# server.  If it is not this server or all then we skip this line.
#------------------------------------------------------------------------
   if ($SERVER != $S && $SERVER != "all") continue

#--------------------------------- There must be at least one destination
   if ("$DESTINATIONS" == "") continue

#------------------------------------------------------------ Echo to log
   echo "===================================================" >> $LOG
   echo "Processing: $TABLELINE -- `date`" >> $LOG

#----------- If source modality root directory does not exist, we're done
   set DIR = $IMGS/$SOURCE/$MODALITY
   cd $DIR

#--------------------------- Make sure there are some studies in modality
   ls [1-9]* >& /dev/null
   if ($status != 0) then
      echo "  No studies in modality" >> $LOG
      continue
   endif
#----------------------------------------- Pass studies in base directory
   foreach study ([1-9]*)
      echo "  study - $study" >> $LOG

#--------- Validate that it is a directory, and that files exist under it
      if (!(-d $study)) continue
      set imagesexist = NO
      ls $study/[1-9]* >& /dev/null
      if ($status != 0) then
         echo "    No series in study" >> $LOG
         continue
      endif
      foreach series ($study/[1-9]*)
         if (!(-d $series)) continue
         ls $series/[1-9]* >& /dev/null
         if ($status == 0) then
            set imagesexist = YES
#---------------------------------- Save an image file name for later use
            set image = `ls $series/[1-9]* | tail -1`
            break
         endif
      end
      if ($imagesexist == NO) then
         echo "    No images in study" >> $LOG
         continue
      endif
#---------------------------------------- Destination files already exist
      ls $study/dest.* >& /dev/null
      if ($status == 0) then
         echo "    Already routed" >> $LOG
         continue
      endif

#------------------------------------------------------------------------
# Get one of the images, extract dicom demographics
#------------------------------------------------------------------------
      $BIN/dcm_dump_file $image | egrep \
         "0010 0010|0010 0020|0010 0021|0010 0040|0010 0030|0008 0050|0008 0020|0008 0030|0008 1010|0008 0080" > $T1
      set PNAME = "`cat $T1 | grep '0010 0010' | cut -d'/' -f5`"
      set PID   = "`cat $T1 | grep '0010 0020' | cut -d'/' -f5`"
      set FACILITY = "`cat $T1 | grep '0010 0021' | cut -d'/' -f5`"
      set DOB   = "`cat $T1 | grep '0010 0030' | cut -d'/' -f5`"
      set SEX   = "`cat $T1 | grep '0010 0040' | cut -d'/' -f5`"
      set ACCN  = "`cat $T1 | grep '0008 0050' | cut -d'/' -f5`"
      set STDDT = "`cat $T1 | grep '0008 0020' | cut -d'/' -f5 | cut -d'.' -f1`"
      set STDTM = "`cat $T1 | grep '0008 0030' | cut -d'/' -f5 | cut -d'.' -f1`"
      set STID  = "`cat $T1 | grep '0008 1010' | cut -d'/' -f5`"
      set INSTN = "`cat $T1 | grep '0008 0080' | cut -d'/' -f5 | sed 's/\[//' | sed 's/\]//'`"
      if ("$PNAME" == "") set PNAME = blank
      if ("$PID" == "")   set PID   = blank
      if ("$FACILITY" == "")   set FACILITY   = blank
      if ("$DOB" == "")   set DOB   = blank
      if ("$SEX" == "")   set SEX   = blank
      if ("$ACCN" == "")   set ACCN  = blank
      if ("$STDDT" == "")   set STDDT = blank
      if ("$STDTM" == "")   set STDTM = blank
      if ("$STID" == "")   set STID = blank
      if ("$INSTN" == "")   set INSTN = blank

      set DATE = `date`
      echo "    New images from $NAME for Name=$PNAME PID=$PID AN=$ACCN - $DATE" >> $LOG
      @ routings += 1
#------------------------------ If split routing, put in dealerqueue file
      echo "QFILE = $QFILE" >> $LOG
      if ($QFILE == YES) then
         set TABLELINE = `$BIN/RouterLine.csh $sdir $mdir $CFG YES`
         set QUEUE    = `echo "$TABLELINE" | cut -d',' -f5`
         echo "$QUEUE" > $study/DealerQueue
         echo "    Split Routing, sent to queue $QUEUE" >> $LOG
      endif
#--------------------------------- Check for possible study abnormalities
# This is a separate process; Router continues on
#------------------------------------------------------------------------
      $BIN/AnalyzeStudy.csh "$DIR/$study" "$SOURCE" "$MODALITY" "$study" "$QUEUE" &

#----------------------------------------------------------------------------------------
# Studies may be routed to new destinations or an additional destination based on their
# facility,institution name or station name, but not EQCweb sources.  Order is facility,
# station, institution, and within that replace (=), add (+), subtract (-).  No more than
# one match per type and action.
#----------------------------------------------------------------------------------------
      set DESTS = "$DESTINATIONS"
      if ("$SOURCE" != "EQCweb") then
         set cfg = $BIN/Routing.cfg
         if (-e $cfg) then
#************************************************************************** Facility Code
            set FACILITY = ($FACILITY)
            set prefix = "Facility:${FACILITY}:"
            @ cp = `echo $prefix | wc -c` + 1
#---------------------------------------------------------- Facility code replace routing
            set rdest = `cat $cfg | grep "${prefix}=" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$rdest"
#----------------------------------------------------------- Facility code add routing(s)
            set rdest = `cat $cfg | grep "${prefix}+" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$DESTS $rdest"
#------------------------------------------------------ Facility code subtract routing(s)
# If the subtract is "IMAGEWEB", all Imageweb destinations (IMG?) will be subtracted.  
# If the subtract is "MAGICSTORE" all Magic Store destinations (MSRD?) will be 
# subtracted. Otherwise, only destinations exactly matching one of the subtracts will be
# subtracted.
#---------------------------------------------------------------------------------------
            set rdest = `cat $cfg | grep "${prefix}-" | cut -c$cp-`
            if ("$rdest" != "") then
               foreach r ($rdest)
                  set nd = ""
                  foreach f ($DESTS)
                     if ($r == IMAGEWEB   && `echo "$f" | cut -c1-3` == IMG)  continue
                     if ($r == MAGICSTORE && `echo "$f" | cut -c1-4` == MSRD) continue
                     if ($r == $f) continue
                     set nd = "$f $nd"
                  end
                  set DESTS = "$nd"
               end
            endif
#************************************************ same as Facility Code, but station name
            set STID = ($STID)
            set prefix = "Station:${STID}:"
            @ cp = `echo "$prefix" | wc -c` + 1
            set rdest = `cat $cfg | grep "${prefix}=" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$rdest"
            set rdest = `cat $cfg | grep "${prefix}+" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$DESTS $rdest"
            set rdest = `cat $cfg | grep "${prefix}-" | cut -c$cp-`
            if ("$rdest" != "") then
               foreach r ($rdest)
                  set nd = ""
                  foreach f ($DESTS)
                     if ($r == IMAGEWEB   && `echo "$f" | cut -c1-3` == IMG)  continue
                     if ($r == MAGICSTORE && `echo "$f" | cut -c1-4` == MSRD) continue
                     if ($r == $f) continue
                     set nd = "$f $nd"
                  end
                  set DESTS = "$nd"
               end
            endif
#******************************************** same as Facility Code, but Institution name
            set INSTN = ($INSTN)
            set prefix = "Inst:${INSTN}:"
            @ cp = `echo "$prefix" | wc -c` + 1
            set rdest = `cat $cfg | grep "${prefix}=" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$rdest"
            set rdest = `cat $cfg | grep "${prefix}+" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$DESTS $rdest"
            set rdest = `cat $cfg | grep "${prefix}-" | cut -c$cp-`
            if ("$rdest" != "") then
               foreach r ($rdest)
                  set nd = ""
                  foreach f ($DESTS)
                     if ($r == IMAGEWEB   && `echo "$f" | cut -c1-3` == IMG)  continue
                     if ($r == MAGICSTORE && `echo "$f" | cut -c1-4` == MSRD) continue
                     if ($r == $f) continue
                     set nd = "$f $nd"
                  end
                  set DESTS = "$nd"
               end
            endif
#************************************************ same as Facility Code, but modality
            set MODALITY = ($MODALITY)
            set prefix = "Modality:${MODALITY}:"
            @ cp = `echo $prefix | wc -c` + 1
            set rdest = `cat $cfg | grep "${prefix}=" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$rdest"
            set rdest = `cat $cfg | grep "${prefix}+" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$DESTS $rdest"
            set rdest = `cat $cfg | grep "${prefix}-" | cut -c$cp-`
            if ("$rdest" != "") then
               foreach r ($rdest)
                  set nd = ""
                  foreach f ($DESTS)
                     if ($r == IMAGEWEB   && `echo "$f" | cut -c1-3` == IMG)  continue
                     if ($r == MAGICSTORE && `echo "$f" | cut -c1-4` == MSRD) continue
                     if ($r == $f) continue
                     set nd = "$f $nd"
                  end
                  set DESTS = "$nd"
               end
            endif
#******************************************** same as Facility Code, but Source AND Station Name
            set prefix = "SourceStation:${SOURCE}:${STID}:"
            @ cp = `echo "$prefix" | wc -c` + 1
            set rdest = `cat $cfg | grep "${prefix}=" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$rdest"
            set rdest = `cat $cfg | grep "${prefix}+" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$DESTS $rdest"
            set rdest = `cat $cfg | grep "${prefix}-" | cut -c$cp-`
            if ("$rdest" != "") then
               foreach r ($rdest)
                  set nd = ""
                  foreach f ($DESTS)
                     if ($r == IMAGEWEB   && `echo "$f" | cut -c1-3` == IMG)  continue
                     if ($r == MAGICSTORE && `echo "$f" | cut -c1-4` == MSRD) continue
                     if ($r == $f) continue
                     set nd = "$f $nd"
                  end
                  set DESTS = "$nd"
               end
            endif

#******************************************** same as Facility Code, but Source AND Modality
            set prefix = "SourceModality:${SOURCE}:${MODALITY}:"
            @ cp = `echo "$prefix" | wc -c` + 1
            set rdest = `cat $cfg | grep "${prefix}=" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$rdest"
            set rdest = `cat $cfg | grep "${prefix}+" | cut -c$cp-`
            if ("$rdest" != "") set DESTS = "$DESTS $rdest"
            set rdest = `cat $cfg | grep "${prefix}-" | cut -c$cp-`
            if ("$rdest" != "") then
               foreach r ($rdest)
                  set nd = ""
                  foreach f ($DESTS)
                     if ($r == IMAGEWEB   && `echo "$f" | cut -c1-3` == IMG)  continue
                     if ($r == MAGICSTORE && `echo "$f" | cut -c1-4` == MSRD) continue
                     if ($r == $f) continue
                     set nd = "$f $nd"
                  end
                  set DESTS = "$nd"
               end
            endif
         endif
      endif
#------------------------------------------------------------------------
# For each destination, put in tag file.  
#------------------------------------------------------------------------
      if ("$DESTS" == "") continue
      foreach rdest ($DESTS)
         set dest = $rdest
         if (-e $BIN/Router.cfg) then
            set newdest = `cat $BIN/Router.cfg | grep "^Dest:$rdest" | cut -d'=' -f2`
            if ("$newdest" == "off") continue
            if ("$newdest" != "") set dest = $newdest
         endif
         if ($dest != "") touch $study/dest.$dest.sendto
      end
   end                           # End of foreach loop for studies
end                              # End of foreach loop for modality directories
cd $IMGS
end                              # End of foreach loop for source directories

#------------------------------------------------------------- We're done
echo "Router execution completed -- `date`  $routings" >> $LOG
rm $LOCK
if (-e $T1) rm $T1
exit 0
