#!/bin/csh

#*****************************************************************
# StudyHold.csh                                          R Moulton
# This script is used to permit the Repair utility to hold a study
# in the image directories and move it to the problem directory.  
# The purpose is to allow studies which have generated errors to be
# pulled from the queue.  An e-mail is sent to the PACS group to 
# let them know that the study has been "held" and information is
# logged.  Command invocation format (see HoldStudy.csh) is:
# find /distributor/imgs -type d -name "studyid" > tmpfile
# StudyHold.csh server studyid < tmpfile >> log file 
# rm tmpfile
#-----------------------------------------------------------------
# 08-30-02 R Moulton - Modified to add relay1 and relay2
#*****************************************************************
set VALIDSERVERS = "bond goldfinger fleming bigben relay1 relay2"
#set PACS = "kleypasd@mir.wustl.edu"
set PACS = "moultonr@mir.wustl.edu"

#------------------------------------ Validate server, directories
set S = $SNAME

set MODE = production

set ROOT = $HOME/$MODE/$S
if (!(-d $ROOT)) then
   echo $ROOT is not a valid directory
   exit 1
endif

set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo $BIN is not a valid directory
   exit 1
endif
 
set LOGS = $ROOT/log
if (!(-d $LOGS)) then
   echo $LOGS is not a valid directory
   exit 1
endif

set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo $TMP is not a valid directory
   exit 1
endif

set IMGS = /distributor/imgs
if ($MODE == test) set IMGS = $ROOT/imgs
if (!(-d $IMGS)) then
   echo $IMGS not a valid directory
   exit 1 
endif

#---------------------------------------------- Server error log
set T1 = HoldStudy.csh.temp

#---------------------------------------- parameter 2 is study id
set studyid = "$2"

#----------------------------------------------------------------
# stdin is output from find command for directories with name
# studyid.  These could be the right one, or one in todelete or
# problems trees.  Eliminate other ones
#----------------------------------------------------------------
while (1)
   set STUDY = $< 
#---------------------------------------------- No such directory
   if (!(-d $STUDY)) continue

   set a = `echo $STUDY | cut -d'/' -f4`
   if (a == todelete) continue
   if (a == problems) continue
#--------------------------- Validate study directory up to imgs
   @ i = 1
   while (1)
      set a = `echo $IMGS  | cut -d'/' -f$i`
      set b = `echo $STUDY | cut -d'/' -f$i`
      if ("$a" != "$b") then
         echo entered path not in images directory tree
         exit 1
      endif
      @ i += 1
      if ("$a" == imgs) break
   end

#-------------------------------------- build target directory
   set TARGET = $IMGS/problems
   set HEADER = ""
   while (1)
      set a = `echo $STUDY | cut -d'/' -f$i`
      set ST = $a
      set b = `echo $a | cut -c1`
      if ($b == 1 || $b == 2 || $b == 3 || $b == 4 || $b == 5 || \
         $b == 6 || $b == 7 || $b == 8 || $b == 9) break
      set TARGET = "$TARGET/$a"
      set HEADER = "$HEADER $a"
      @ i += 1
   end
#------- Add date directory to target to keep multiples separate
   set A = `date '+%Y%m%d%H%M'`
   set TARGET = "$TARGET/$A"

   @ i += 1
   set a = `echo $STUDY | cut -d'/' -f$i`
   if ($a != "") then
      echo last directory should be study
      exit 1
   endif

   cd $STUDY
   echo ===========================================================
   echo "Study placed in hold file"
   echo ===========================================================
   echo "Source-modality $HEADER"
   echo "Study: $ST"

   pwd
   ls [1-9]*/[1-9]* >& /dev/null
   if ($status != 0) then
      echo No images in study
      set PNAME = ""
      set PID = ""
      set ACCN = ""
   else
#-------------------------------------------- show size of study
      set size = `du -k -s | cut -d'	' -f1`
      set units = KB
      if ($size > 1024) then
         @ size = $size / 1024
         set units = MB
      endif
      echo "$size $units"
#--------------------------------------------- Show demographics
      set image = `ls [1-9]*/[1-9]* | tail -1`
      $BIN/dcm_dump_file $image | egrep "0010 0010|0010 0020|0008 0050" > $T1
      set PNAME = "`cat $T1 | grep '0010 0010' | cut -d'/' -f5`"
      set PID   = "`cat $T1 | grep '0010 0020' | cut -d'/' -f5`"
      set ACCN  = "`cat $T1 | grep '0008 0050' | cut -d'/' -f5`"
      echo "Patient Name: $PNAME"
      echo "Patient ID:   $PID"
      echo "Accession:    $ACCN"
   endif
#--------------------------------------------------- Actual move
   mkdir -p $TARGET
   cd ..
   mv $ST $TARGET
#---------------------------------------------------------------
# Email.  Note: do not change subject line, as this is used by
# pacs group to generate a forwarding rule.
#--------------------------------------------------------------- 
set DT = `date`
/usr/bin/mail $PACS << _EOD_
Subject: Study moved to problems directory
== Study moved to problems directory == 
$DT
$HEADER $ST" 
Patient Name: $PNAME" 
Patient ID:   $PID"  
Accession:    $ACCN"
_EOD_

exit 0
end
