#!/bin/csh

#************************************************************************
# Dealer.csh                                                    R Moulton
#
# This script examines image directories created by the dicom receiver
# for tag files placed in them by the router indicating that they are to
# be sent to the destination this queue is processing, with names in the 
# format dest.destination.sendto.  When found, these are sent to the
# destination, and if successful, the images are moved to the todelete 
# directory.
# There are conceptually two different queue's being handled by the 
# Dealer, that seen by the monitor program and scripting tools, and the
# internal queue in Distributor.cfg.  Because there may be more than one
# cron process handling a given queue (e.g. two destinations and each
# one handled by a different cron invocation of Dealer), the queue 
# parameter may have two numbers, the external (monitor) and internal
# (.cfg) queues.  
# Calling syntax:
#   Dealer.csh External_queue Internal_queue Destination Delay
# e.g.
#   Dealer.csh 14 4 IMG2 10
#------------------------------------------------------------------------
# 11-14-01 R Moulton - Added support for live server in .cfg
# 11-15-01 R Moulton - Added Message queue different from .cfg queue
# 11-30-01 R Moulton - Modified to use .rslt files to determine which
#          images have been sent
# 02-21-02 R Moulton - Modified to enhance use of .rslt files; handle
#          cases where some images have been deleted, make sure that
#          images are deleted after last xmit.
# 05-08-02 R Moulton - Modified to allow two digit internal and external
#          queue numbers, support bigben
# 08-30-02 R Moulton - Modified to add relay1 and relay2
# 10-07-02 R Moulton - Commented out IMAGE_SENT message to monitor
# 07-10-03 R Moulton - Modified to scan directories in imgs tree and 
#                      match them to .cfg with grep, rather than passing
#                      grep and looking for corresponding directories.
# 05-04-05 R Moulton - Modified to pick up additional sourcings not
#                      present in the .cfg file.
# 05-31-05 R Moulton - Always drops MoveStudy on LTA queues
# 07-22-05 R Moulton - Added support for Resend Destinations
# 05-16-06 R Moulton - Added support for Dealer.cfg file
#************************************************************************

set ONEMACHINE = DealStudy.csh

#---------------------------- Must be logged in as spectrum to run Dealer
if (`whoami` != "spectrum") then
   echo "Must be user spectrum to run Dealer"
   exit 1
endif

#------------------------------------------------------------- Parameters
set S = $SNAME
set EQ = $1
set IQ = $2
set SQ = $S$EQ

#----------------------------------- Param 3 is destination, default All 
set DEST = All 
if ("$3" != "") set DEST = "$3"

#----------------------------- Param 4 is delay seconds (0-59, default 0)
set SLEEP = 0
if ("$4" != "") set SLEEP = "$4"
echo "$SLEEP" | egrep "[0-9]|[0-9][0-9]" > /dev/null
if ($status != 0) then
   echo "delay of $SLEEP - not valid number"
   exit 1
endif
if ($SLEEP < 0 || $SLEEP > 59) then
   echo " delay of $SLEEP not in range 0-59"
   exit 1
endif
if ($SLEEP != 0) sleep $SLEEP

#------------------ mode we are running in is param 5, production or test
set MODE = production
if ("$5" != "") set MODE = $5
if ("$MODE" != "production" && "$MODE" != "test") then
   echo "$MODE is not a valid production mode"
   exit 1
endif

#----------------------------- Validate appropriate server directory tree
set ROOT = $HOME/$MODE/$S
if (!(-d $ROOT)) then
   echo "$ROOT is not a valid directory"
   exit 1
endif

set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo "$BIN is not a valid directory"
   exit 1
endif

set LOGS = $ROOT/log
if (!(-d $LOGS)) then
   echo "$LOGS is not a valid directory"
   exit 1
endif

set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo "$TMP is not a valid directory"
   exit 1
endif

#---------------------------------------------------- Validate image root
set IMGS = /distributor/imgs
if ($MODE == "test") set IMGS = $ROOT/imgs
if (!(-d $IMGS)) then
   echo "$IMGS is not a valid directory"
   exit 1
endif
set STORE = $IMGS/todelete

#---------------------------------------------- Configuration file to use
set CFG = $BIN/$CSNAME$IQ.cfg
if (!(-e $CFG)) then
   echo "$CFG file does not exist"
	exit 1
endif

#----------------------------------- Script lock and stoprouter tag files
set LOCK  = $TMP/Dealer.csh.lock.$SQ
set STOP  = $TMP/stopDealer
set QSTOP = $TMP/stopDealer.$SQ
set DSTOP = $TMP/stopDealer.$DEST
set QUIT  = "-e $STOP || -e $QSTOP || -e $DSTOP"
#----------------------------------------------------------- script names
set SCRIPT    = "Dealer.csh"
set DEAL = $BIN/$ONEMACHINE
#------------------------------------------------------------- temp files
set T1 = $TMP/Dealer.$SQ.temp1

#----------------------- If lock file exists, exit... otherwise create it
if (-e $LOCK) exit 0
touch $LOCK
if (!(-e $LOCK)) then
   echo "Could not create $LOCK file" 
   exit 1
endif

#--------------------------------------------------------- Open log files
set LOG = $LOGS/Dealer.$SQ.log
if (-e $LOG) rm $LOG
touch $LOG
if (!(-e $LOG)) then
   echo "Unable to create $LOG file"
   exit 1
endif
set MLOG = $LOGS/MoveStudy.$SQ.log
if (-e $MLOG) rm $MLOG
touch $MLOG
if (!(-e $MLOG)) then
   echo "Unable to create $MLOG file"
   exit 1
endif

set ERRLOG = $LOGS/Dealer_problems.log
touch $ERRLOG
if (!(-e $ERRLOG)) then
   echo "Unable to create $ERRLOG file"
   exit 1
endif

#------------------------ Are we to process MoveStudy on this queue?
set MOVEFLAG = `echo "$MOVEQUEUES" | cut -d',' -f"$EQ"`
if ("$MOVEFLAG" != "NO") set MOVEFLAG = "YES"
if ("$DEST" == "LTA") set MOVEFLAG = "YES"
if ("$DEST" == "IMG6") set MOVEFLAG = "YES"

#------------------------------------- Begin Dealer message to monitor
$BIN/Message $SCRIPT $SQ BEGIN_SCRIPT
echo "Begin :: $SCRIPT $S.$EQ.$IQ $DEST $MODE -- `date`" >> $LOG

#--------------------------------------- Initial values of input defaults
set DEFISERVER = $S  
set DEFIQUEUE  = 1
set DEFIDEST1  = IMG4
set DEFIDEST2  = IMG2
set DEFIDEST3  = IMG5

#------------------------------------------------------------------------
# This loop processes one directory in the imgs tree, looking for a 
# matching entry in the configuration file. It is actually two loops, but
# we treat it as one (indent-wise)
#------------------------------------------------------------------------
@ xmits = 0
@ fails = 0
cd $IMGS
#----------------------------------------------------- source directories
foreach sdir (*)
if (!(-d $sdir)) continue
grep "^$sdir" $CFG >& /dev/null
if ($status == 1) continue
#--------------------------------------------------- modality directories
ls $IMGS/$sdir/* >& /dev/null
if ($status != 0) continue
cd $IMGS/$sdir
foreach mdir (*)
   cd $IMGS/$sdir
   if (!(-d $mdir)) continue
	set TABLELINE = `grep "^$sdir,$mdir" $CFG`
	if ($status == 1) continue
	
#------------------------------------------- Check for stopDealer file
   if ($QUIT) then
      echo "Exiting early, stopDealer file found -- `date`" >> $LOG
      $BIN/Message $SCRIPT $SQ END_SCRIPT
      rm $LOCK
      exit 0
   endif

#------------------------------------------------------------------------
# Here we are looking at a detail line from the configuration file, which
# is of the form source,modality,name,server,queue,dest1,dest2... with up
# to five destinations. Everything from server on will default if it is
# blank.  First, we validate the detail line fields
#------------------------------------------------------------------------

#-------------------------------------------- Pull values from table line
   set SOURCE   = `echo "$TABLELINE" | cut -d',' -f1`
   set MODALITY = `echo "$TABLELINE" | cut -d',' -f2`
   set NAME     = `echo "$TABLELINE" | cut -d',' -f3`
   set SERVER   = `echo "$TABLELINE" | cut -d',' -f4`
   set QUEUE    = `echo "$TABLELINE" | cut -d',' -f5`
   set D1       = `echo "$TABLELINE" | cut -d',' -f6`
   set D2       = `echo "$TABLELINE" | cut -d',' -f7`
   set D3       = `echo "$TABLELINE" | cut -d',' -f8`
   set D4       = `echo "$TABLELINE" | cut -d',' -f9`
   set D5       = `echo "$TABLELINE" | cut -d',' -f10`

#------------------------------ Destinations combined for loop processing
   set DESTINATIONS = "$D1 $D2 $D3 $D4 $D5"      

#------------------------------------------------------------------------
# Validations.  Source and modality will be validated by the existance of
# the appropriate directories in the image tree.  If the directory does
# not exist this is not an error; simply nothing is done.  So there is no
# explicit validation here.  The name field is passed to the monitor as
# part of the Message calls, and is not currently validated in this 
# script.  Detailed validation and logging is done by the Router.  Here
# we just look for what we are expecting.
#------------------------------------------------------------------------

#------------------------------------------------------------------------
# server.  If it is not this server then we skip this line.
#------------------------------------------------------------------------
   if ($SERVER != $S) continue

#------------------------------------------------------------------------
# queue.  If it is not this internal queue, then we skip this line.
#------------------------------------------------------------------------
   if ("$QUEUE" != $IQ) continue

#------------------------------------------------------------------------
# We now have a line which is for our server and queue
#------------------------------------------------------------------------

#------------------------------------------------------------ Echo to log
   echo "==================================================" >> $LOG
   echo "Processing :: $TABLELINE -- `date`" >> $LOG

   set DIR = $IMGS/$SOURCE/$MODALITY
   set STOREDIR = $STORE/$SOURCE/$MODALITY
   cd $DIR

#--------------------------- If there are no studies in modality, skip it
   ls [1-9]* >& /dev/null
   if ($status != 0) then
      echo "  No studies in $DIR" >> $LOG
      continue
   endif

#----------------------------------------- Pass studies in base directory
   foreach study ([1-9]*)
      if (!(-d $study)) continue
#------------------- Quick check.  If no *.sendto files exist, we're done
      ls $study/*.sendto >& /dev/null
      if ($status != 0) continue
#---------------- If split routing, is this study for our internal queue?
      if (-e $study/DealerQueue) then
         if ($IQ != "`cat $study/DealerQueue`") continue
      endif

      echo "--study - $study" >> $LOG
#------------------------------------------------------------------------
# Now we are looking for destination tag files indicating that this study
# needs to be dealt to our destination: dest.destinationcode.sendto.
# If we find it, we do a deal.  
#------------------------------------------------------------------------
      set D = $DEST
#------------------------------------------------------------------------
# The "All" DEST means that we are to deal to all the destinations on the
# config file line which are not stopped
#------------------------------------------------------------------------
      if ($D == "All") then
         set D = "$DESTINATIONS"
#------------------------------------------------------------------------
# The Dealer.cfg file allows group destination codes, which will be sent
# to all the destinations listed for that group.  
#------------------------------------------------------------------------
      else if ($D != "All") then
         if (-e $BIN/Dealer.cfg) then
            grep "^${D}:" $BIN/Dealer.cfg >& /dev/null
            if ($status == 0) then
               set D = (`grep "^${D}:" "$BIN/Dealer.cfg" | cut -d':' -f2`)
            endif
         endif
      endif
#-----------------------------------------------------------------------
# If there is only one listed destination, and no dest* file exists for
# that destination, we're done.
#-----------------------------------------------------------------------
      if (${#D} == 1)then
         if (!(-e $study/dest.$D.sendto)) continue
      endif
#-----------------------------------------------------------------------
# To be safe, we'll check to make sure the images are still there
#------------------------------------------------------------------------
      ls $study/[1-9]* >& /dev/null
      if ($status != 0) then    
         echo "    No series in study: $study" >> $LOG
         continue
      endif
      set stillthere = NO 
      foreach series ($study/[1-9]*)
         ls $series/[1-9]* >& /dev/null
         if ($status != 0) continue
#-------------------------------------------- Grab an image for later
         set image = `ls $series/[0-9]* | tail -1`
         set simage = `echo "$image" | cut -d'/' -f2-3`
         set stillthere = YES
         break
      end
      if ($stillthere == NO) then    
         echo "    No images in study: $study" >> $LOG
         continue
      endif

      $BIN/dcm_dump_file $image | egrep "0010 0010|0010 0020|0010 0030|0008 0050|0008 0020|0008 0030|0008 0080|0008 1010" > $T1
      set PNAME = "`cat $T1 | grep '0010 0010' | cut -d'/' -f5 | $BIN/PFix`"
      set PID   = "`cat $T1 | grep '0010 0020' | cut -d'/' -f5 | $BIN/PFix`"
      set DOB   = "`cat $T1 | grep '0010 0030' | cut -d'/' -f5 | $BIN/PFix`"
      set ACCN  = "`cat $T1 | grep '0008 0050' | cut -d'/' -f5 | $BIN/PFix`"
      set STDT  = "`cat $T1 | grep '0008 0020' | cut -d'/' -f5 | $BIN/PFix date`"
      set STTM  = "`cat $T1 | grep '0008 0030' | cut -d'/' -f5 | $BIN/PFix time`"
      set INSTN = "`cat $T1 | grep '0008 0080' | cut -d'/' -f5 | $BIN/PFix`"
      set STNM  = "`cat $T1 | grep '0008 1010' | cut -d'/' -f5 | $BIN/PFix`"
#------------------------------------------------------------------------
# send the images.  The foreach loop is because $D could be either a 
# single destination or all the destinations on the line.  For this same
# reason, we check for the stopDealer.dest & *.sendo again
#------------------------------------------------------------------------
      set XMITSOK = NONE
      foreach dst ($D)
         set dest = $dst
         if (!(-e $study/dest.$dest.sendto)) continue
         if (-e $TMP/stopDealer.$dest) continue
#------------------------------------------------------------------------------------
# We use the Dealer.cfg file to determine if we add the server name, the all caps    
# server name, or make a name change for this destination or this queue
#------------------------------------------------------------------------------------
         set stdest = $dest
         if (-e $BIN/Dealer.cfg) then
#------------------------------------- prefix lower case server name (e.g. bond_IMG6)
            set fnd = `cat $BIN/Dealer.cfg | grep "^server:$dest"`
            if ("$fnd" != "") set stdest = ${S}_$dest
#------------------------------------- prefix upper case server name (e.g. BOND_IMG6)
            set fnd = `cat $BIN/Dealer.cfg | grep "^SERVER:$dest"`
            if ("$fnd" != "") set stdest = ${ACSNAME}_$dest
#------------------------------------------- complete change (e.g. LTA => BWCH_CACHE)
            set newdest = `cat $BIN/Dealer.cfg | grep "^Dest:$dest" | cut -d'=' -f2`
            if ("$newdest" != "") set stdest = $newdest
#------------------------------------------ this dest & queue (e.g. IMG5.3 => relay2)
            set newdest = `cat $BIN/Dealer.cfg | grep "^DQueue:$dest.$IQ" | cut -d'=' -f2`
            if ("$newdest" != "") set stdest = $newdest
         endif 
#---------------------------------------------------- If already sent, skip this deal
         $BIN/SentTest.csh $DIR/$study $stdest $BIN $LOG
         if ($status == 2) then
				set SKIPPED = YES
				echo "    $dest - already sent" >> $LOG
				date '+%Y%m%d%H%M%S' > $study/dest.$dest.OK
         else 
				set SKIPPED = NO 
#-------------------------------------------------- dicom xmit of all images in study
            csh -x $DEAL $DIR/$study $dest $S $EQ $IQ $ROOT \
				"$PNAME" "$PID" "$ACCN" $SOURCE $MODALITY 
            echo "    End xmit -- `date`" >> $LOG
#--------------------------------------- dest.D.OK file means send worked
            @ xmits +=1
			endif
         if ($XMITSOK == NONE) set XMITSOK = YES
         if (!(-e $study/dest.$dest.OK)) then
            echo "    *** Xmit failed ***" >> $LOG
            @ fails +=1
            set XMITSOK = NO
            continue
         endif
         set EOT = `cat $study/dest.$dest.OK`
         rm $study/dest.$dest.OK

         if ($SKIPPED == NO) then
            echo "    Images from $NAME for $PNAME $PID $ACCN - `date` sent to $dest" >> $LOG
            $BIN/Message "$NAME" $SQ IMAGE_SENT "$PNAME" "$PID" "$ACCN" $dest "$study" $SOURCE $MODALITY
	 		endif
      end
#-------------------------- Here is substring for clear of completed files
      if ($XMITSOK == YES && $MOVEFLAG == YES) then
         $BIN/MoveStudy.csh $DIR/$study $STOREDIR/$study $S $SQ $ROOT $EOT \
            "$PNAME" "$PID" "$ACCN" $IQ $SOURCE $MODALITY \
	    "$STDT" "$STTM" "$INSTN" "$STNM" "$DOB" >> $MLOG & 
      endif
   end                           # End of foreach loop for studies
#-------------------------- Notify monitor of end of deal for source
#  $BIN/Message "$NAME" $SQ END_PASS $DEST
end                        # End of foreach loop for modality directories
cd $IMGS
end                        # End of foreach loop for source directories
#------------------------------------------------------------- We're done
echo "Dealer execution completed -- `date` $xmits - $fails" >> $LOG
$BIN/Message $SCRIPT $SQ END_SCRIPT
if (-e $LOCK) rm $LOCK
if (-e $T1) rm $T1
exit 0
