#!/bin/csh

#*****************************************************************
# Xmit.csh                                               R Moulton
# This script will broadcast a study to some selected destination.
# The script will work for studies in standard /distributor/imgs
# subdirectories or in non-standard directories, and has controls
# to help avoid sending to destinations which are not set up.  It
# will also dump some of the demographics on a study for the user
# to verify.
#-----------------------------------------------------------------
# Function is as follows: (1) Find the study, dump demographics,
# and verify with user. (2) If in a non-standard directory, xmit
# directly from that directory and done.  (3) If in a standard
# /distributor/imgs directory, copy images in study from std dir
# and and std todelete directory (if any) to special directory,
# xmit from there, then delete special directory images.
# In all cases, destination will be validated by a grep on the 
# load control config file. These xmits are not logged, and they 
# are not subject to control by stop or lock files.
#-----------------------------------------------------------------
# Xmit.csh server studydirectory destination   
# It is assumed that the script will normally be implemented 
# through a script xmit in the spectrum home directory.
#-----------------------------------------------------------------
# 08-30-02 R Moulton - Modified to add relay1 and relay2
#*****************************************************************

#------------------------------------------- Must be user spectrum
if (`whoami` != spectrum) then
   echo Must be user spectrum to run Xmit
   exit 1
endif

#------------------------------------ Validate server, directories
set S = $SNAME
set CFILE = $CSNAME

set ROOT = $HOME/production/$S
if (!(-d $ROOT)) then
   echo $ROOT is not a valid directory
   exit 1
endif

set BIN = $ROOT/bin
if (!(-d $BIN)) then
   echo $BIN is not a valid directory
   exit 1
endif
 
set TMP = $ROOT/tmp
if (!(-d $TMP)) then
   echo $TMP is not a valid directory
   exit 1
endif

set IMGS = /distributor/imgs
if (!(-d $IMGS)) then
   echo $IMGS not a valid directory
   exit 1 
endif

set T1 = Xmit.csh.temp1
set STUDY = "$2"
#------------------------------------------------ No such study
if ("$STUDY" == "") then
   echo No study specified
   exit 1
endif
if (!(-d $STUDY)) then
   echo No such study directory
   exit 1
endif

#---------------------------- Validate destination (default IMG6)
set DEST = IMG6
if ("$3" != "") set DEST = "$3"

cat $BIN/MIRCS_dicom_broadcast.ini | grep "^$DEST=" >& /dev/null
if ($status != 0) then
   echo "$DEST not valid destination code"
   exit 1
endif 

#--------------------------------------------------------------
# Determine if study is in a standard broadcast directory for
# this server.  If not, the target directory (from which we 
# will broadcast) is the study directory.
#--------------------------------------------------------------
set STANDARD = NO
set TARGET = $STUDY
while (1)
#---------------- base directory tree must be /distributor/imgs
   if (`echo $STUDY | cut -d'/' -f2` != distributor) break
   if (`echo $STUDY | cut -d'/' -f3` != imgs) break
   set SRC = `echo $STUDY | cut -d'/' -f4`
   set MOD = `echo $STUDY | cut -d'/' -f5`
   set STD = `echo $STUDY | cut -d'/' -f6`
   if ($STD == "") then 
      set STD = $MOD
      set MOD = ""
   endif
   set PATTERN = "$SRC,$MOD,"
   cat $BIN/$CFILE.cfg | grep "$PATTERN"
   if ($status != 0) break
#--------------------------------------------------------------
# Images are in a standard directory for this distributor.
# This means that they may be in process of being distributed
# using the normal scripts, or in process of being moved off
# to the todelete directory.  In this case, we will broadcast
# from a study directory in /distributor/special, and we will
# copy all images from the study directory and the study 
# todelete directory to this directory first.
#--------------------------------------------------------------
   set STANDARD = YES
   set TARGET = /distributor/special/$STD
   set TODELETE = /distributor/imgs/todelete/$SRC/$MOD/$STD
   if ($MOD == "") set TODELETE = /distributor/imgs/todelete/$SRC/$STD
   break
end

#---------------------------------------------------------------
# print out demographics and get user validation for broadcast
#---------------------------------------------------------------
cd $STUDY
ls [1-9]*/[1-9]* >& /dev/null
if ($status != 0) then
   echo No images in study
   exit 1
endif

echo ===========================================================
echo "Broadcast study to $DEST immediately?"
echo ===========================================================
echo "Study: $STUDY"

ls [1-9]*/[1-9]* >& /dev/null
if ($status != 0) then
   echo No images in study
   set PNAME = ""
   set PID = ""
   set ACCN = ""
else
#-------------------------------------------- show size of study
   set size = `du -k -s | cut -d'	' -f1`
   set units = KB
   if ($size > 1024) then
      @ size = $size / 1024
      set units = MB
   endif
   echo "$size $units"
#--------------------------------------------- Show demographics
   set image = `ls [1-9]*/[1-9]* | tail -1`
   $BIN/dcm_dump_file $image | egrep "0010 0010|0010 0020|0008 0050" > $T1
   set PNAME = "`cat $T1 | grep '0010 0010' | cut -d'/' -f5`"
   set PID   = "`cat $T1 | grep '0010 0020' | cut -d'/' -f5`"
   set ACCN  = "`cat $T1 | grep '0008 0050' | cut -d'/' -f5`"
   echo "Patient Name: $PNAME"
   echo "Patient ID:   $PID"
   echo "Accession:    $ACCN"
endif
echo
echo "(Y)es to continue"
set RESP = $<
set RESP = `echo $RESP | cut -c1`
if ("$RESP" != "Y" && "$RESP" != "y") then
   echo No action taken
   exit 0
endif

#-------------------------------------------------------------
# If study is in standard directory, copy images from that
# directory and todelete directory to special target directory
#-------------------------------------------------------------
if ($STANDARD == YES) then
   echo Copying study for broadcast
   mkdir -p /distributor/special
   if ($status != 0) then
      echo "Unable to make /distributor/special directory"
      exit 1
   endif
   cp -r $STUDY /distributor/special
   if ($status != 0) then
      echo "Unable to copy from $STUDY directory"
      exit 1
   endif
   if (-d $TODELETE) then
      cp -r $TODELETE /distributor/special
      if ($status != 0) then
         echo "Unable to copy from $TODELETE directory"
         exit 1
      endif
   endif
endif

#-------------------------------------------------- Broadcast 
cd $TARGET
if (-e .broadcast.rslt.$DEST) rm .broadcast.rslt.$DEST
echo Broadcasting study
$BIN/MIRCS_dicom_broadcast $TARGET $DEST 
if ($status == 0) then
   echo "Study sent successfully"
else
   echo "Some errors occurred"
endif

#--------------------- Delete image files if they were copied
cd
if ($STANDARD == YES) rm -rf $TARGET
exit 0
