#!/bin/csh

#*****************************************************************
# XmitOther.csh                                          R Moulton
# This script will broadcast a test study to the DICOM receiver on
# other distributors and log the results.  If the test fails, 
# it will immediately retry the test.  If the retry also fails,
# it will send an alert message and signal the other server to
# restart its DICOM receiver.  If the XMITOTHER variable is set,
# it overides the DESTINATIONS list 
#-----------------------------------------------------------------
# 05-10-03 R Moulton - Initial write
# 05-12-04 R Moulton - Modified to use small CT, no OK log
# 03-17-06 R Moulton - Modified to user XMITOTHER, do restart
# 03-24-17 R Moulton - Modified to do retry on failed xmit, make
#                      OK log optional, make self-test optional
#*****************************************************************

#------------------------- Do we want to log tests with no errors?
set OKLOG = "YES"
#-------------------------------- Do we want to test to ourselves?
set SELFTEST = "NO"
#-------------------------------- Default dest is all distributors
set DESTINATIONS = "BIGBEN RELAY1 RELAY2 RELAY3 RELAY4"

#------------ If XMITOTHER exists in env, it overides DESTINATIONS
if (${?XMITOTHER}) then
   set DESTINATIONS = "$XMITOTHER"
#-------- We assume if XMITOTHER is set & has selftest, that is OK
   set SELFTEST = "YES"
endif

set STUDY  = "CT/1.test.ct.study"
set TARGET = "$IMGS/TEST_IMGS/$STUDY"
set ELOG   = "$LOG/XmitOtherError.log"
set OLOG  = "$LOG/XmitOther.log"
set MSG    = $TMP/XmitTest.msg
set T1     = $TMP/XmitTest1.tmp
set T2     = $TMP/XmitTest2.tmp

#------------------------------------------- Must be user spectrum
if (`whoami` != spectrum) then
   echo Must be user spectrum 
   exit 1
endif

#------------------------------------- Make sure test study exists
if (!(-d $TARGET)) then
   echo $TARGET directory not found
   exit 1
endif

cd $TARGET

#----------------------------------- pass each destination in list 
foreach DEST ($DESTINATIONS)
#---------------------------------------------- test to ourselves?
   if ("$DEST" == "$ACSNAME" && "$SELFTEST" != "YES") continue

#------------------------------------------------------ First test
   if (-e .broadcast.rslt.$DEST) rm .broadcast.rslt.$DEST
   if (-e .broadcast.map.$DEST) rm .broadcast.map.$DEST
   $BIN/MIRCS_dicom_broadcast $TARGET $DEST > $T1
#--------------------------------------------------------- xmit Ok
   if ($status == 0) then
      if ("$OKLOG" == "YES") echo "`date` Xmit OK to $DEST" >> $OLOG
      continue
   endif

#------------------------------------------------------ retry test
   if (-e .broadcast.rslt.$DEST) rm .broadcast.rslt.$DEST
   if (-e .broadcast.map.$DEST) rm .broadcast.map.$DEST
   $BIN/MIRCS_dicom_broadcast $TARGET $DEST > $T2
#--------------------------------------------------------- xmit Ok
   if ($status == 0) then
#------------------------------------------------------- Error log
      echo "`date` Single Xmit failure to $DEST" >> $ELOG
      echo "-----------------------------------------" >> $ELOG
      cat $T1 >> $ELOG
      echo "-----------------------------------------" >> $ELOG
#--------------------------------------------------- Alert message
      cp $BIN/XmitTest.hdr $MSG
      echo "t=1 Xmit fail" >> $MSG
      echo "Fr: $ACSNAME" >> $MSG
      echo "To: $DEST" >> $MSG
      echo "retry xmit OK" >> $MSG
      $BIN/SendAlert $MSG
      continue
   endif

########################### Double Failure #######################

#----------------------------------------------- Send alert message 
   echo "`date` ***** Double Xmit Failure to $SNAME *****" >> $ELOG
   echo "---------- First attempt ----------------" >> $ELOG
   cat $T1 >> $ELOG
   echo "---------- Second attempt ---------------" >> $ELOG
   cat $T2 >> $ELOG
   echo "-----------------------------------------" >> $ELOG
#--------------------------------------------------- Alert message
   cp $BIN/XmitTest.hdr $MSG
   echo "t=2 Xmit fail" >> $MSG
   echo "Fr: $ACSNAME" >> $MSG
   echo "To: $DEST" >> $MSG
   echo "Sending restart" >> $MSG
   $BIN/SendAlert $MSG

#-------------------------------- send restart to other distributor
   if ("$DEST" != "TEST") then
      echo "s=$DEST" > $MSG
      echo "p=restart" >> $MSG
      $BIN/Run $MSG
   endif

   if (-e $T1) rm $T1
   if (-e $T2) rm $T2
   if (-e $MSG) rm $MSG
end
