#!/usr/local/bin/perl
# Script scans the files "transfer_information.bond[n].log" [n=1,2,etc] 
#	which are produced by bond's scripts moving images to the
#	Easyweb and/or Mitra servers.

# Output is used to help monitor bond's transfer activity.
 
# Output is sent via STDOUT and consists of tab-delimited lines
#	suitable for input to an Excel spreadsheet.
#	If the script appears as a crontab task, the output may be mailed.
#
# The same output is also written to a file bond[n].[yyyymmdd]
#	where [n=1,2,etc] and [yyyymmdd = a date]
#

# This version is designed to be run as a crontab task at 5 minutes after
#	midnight each day, looking for and summarizing yesterday's activity.

#***********************
#*                     *
#*  Input Description  *
#*                     *
#***********************
# Here's sample of a 9-line group from source transfer_information.bond[1-]log:
# [line 1]  0010 0010       18 //               PAT Patient Name//NEWMAN,ALFRED E 
# [line 2]  0010 0020        8 //                 PAT Patient ID//2075653 
# [line 3]  0008 0050        8 //            ID Accession Number//5123478
# [line 4]  0008 0060        2 //                    ID Modality//CT
# [line 5]  136 images and Kbytes: 71896 .
# [line 6]  drwxrwsr-x 29 root	spectrum	2048 Jan 29 23:45 ..
# [line 7]  START: 01/23/01  17:21:20
# [line 8]  END: 01/23/01  17:27:57
# [line 9]  All images transferred from 1.2.124.113532.128.252.220.117.20010117.155318.5348093
#
# Explanation of Input Lines
# [line 1] Patient Name
# [line 2] Patient ID
# [line 3] Accession Number
# [line 4] Modality
# [line 5] Number of Images and Kilobytes
# [line 6] Time stamp of arrival of study's last image from modality
# [line 7] Start time to move images to Web browser / Image archive
# [line 8] Stop time of moving images
# [line 9] Message reporting success/trouble for named Study UID

#************************
#*                      *
#*  Output Description  *
#*                      *
#************************
# Output fields [A,B, etc would be Excel columns]:
#	A -- Index number
#	B -- Patient name
#	C -- Patient ID
#	D -- Accession Number
#	E -- Modality (CT, MR, CR, NM, etc)
#	F -- Number of Images
#	G -- Number of Kilobytes
#	H -- Time of study's last image from modality (input-end)
#	I -- Elapsed time, input-end to output-start (min)
#	J -- Output Start Date
#	K -- Output Start Time
#	L -- Output End Time
#	M -- Elapsed Time, output-start to output-end (secs)
#	N -- Error list

# If source Patient ID is blank, output is "ID Missing"
# If source Accession Number is blank, output is "Accession Missing"
# If source Patient ID does not begin with D or 0-9, output is "Bad ID"
# If source Patient ID's next 6 characters are not 0-9, output is "Bad ID"
# If source Patient ID is longer than 7 characters, output is "ID too Long"
# If source Accession Number does not contain exactly 7 alphanumerics,
#	output is "Bad Accession"
# If source Patient ID and Accession Number are equal, 
#	output is "Accession=ID"
# Illegal ID's and Accession numbers are printed in last output field,
#	also called the error list.

# These variables should be set according the user and the number of 
#	logs ($logmax)of the form "transfer_information.bond[1-].log"
#	For spectrum, these logs are in /bonda/spooler.
#	For anyone else, set appropriately.
#
# Output is filename  of the form "bond[1-].yyyymmdd"
#	where the yyyymmdd timestamp is yesterday's date.
#	These files are saved in the TEST_DAY_LOGS subdirectory.
#
# Run the routine "main" for each input log.

# $input_file_base  = "/opt/home/erl/kwc/transfer_information.bond";
# $output_file_base = "/opt/home/erl/kwc/BOND_DAY_LOGS/bond";

$output_file_base = "/distributor/logs/TEST_DAY_LOGS/test";
$input_file_base  = "/distributor/logs/xfer_information.host.";

@queue_names = ("queue");

foreach $q (@queue_names) {
	main ($q, $input_file_base, $output_file_base);
}

##########
#        #
#  main  #
#        #
##########

sub main {

$lognum = $_[0];	# Log number (as in 1,2, etc as in bond1)
$ifb = $_[1];		# Input file base
$ofb = $_[2];		# Output file base

# The subroutine "yesterday" computes scalars $tyy, $tmm, $tdd
#	which are used later to compare dates in the transfer_logs.
#	Here, the text representations of $tyy, $tmm, $tdd are returned
#	and concatenated to form the output file suffix.

$xferout = "$ofb"."$lognum".".".yesterday();	# Form the output filename

$til = "$ifb"."$lognum".".log";			# Form the input filename

$months = "JanFebMarAprMayJunJulAugSepOctNovDec";

##################################
# Set index j = number of events #
##################################
$j = 0;

##################################
# Open the output file as STDOUT #
##################################
open (OUT, ">$xferout")
	|| die "Cannot open ouput file $xferout\n";

################################################
# Open the transfer_intformation.log for input #
################################################
open (XFRLOG, $til)
	|| die "Cannot open input file $til\n";

###################################################
# Process groups of 9 lines until EOF encountered #
###################################################
# Here's sample of 9-line group from the log:
#  0010 0010       18 //               PAT Patient Name//NEWMAN,ALFRED E 
#  0010 0020        8 //                 PAT Patient ID//2075653 
#  0008 0050        8 //            ID Accession Number//5123478
#  0008 0060        2 //                    ID Modality//CT
#  136 images and Kbytes: 71896 .
#  drwxrwsr-x 29 root	spectrum	2048 Jan 29 23:45 ..
#  START: 01/23/01  17:21:20
#  END: 01/23/01  17:27:57
#  All images transferred from 1.2.124.113532.128.252.220.117.20010117.155318.5348093

# Sometimes, the input file gets corrupted and a line is missing.
# The following are checks to make sure the 9 lines contain certain
#	expected words or phrases.  If they, the program does a "next",
#	aborts the loop, restarts the loop looking for a line
#	containing Patient Name so things can get back in synch.

while ($line1 = (<XFRLOG>)) {		# Line 1 -- Patient Name
  if (!($line1 =~ /PAT Patient Name/)) {
	print "Line 1 does not contain PAT Patient Name\n";
	next;
  }
  $line2 = (<XFRLOG>);			# Line 2 -- Patient ID
  if (!($line2 =~ /PAT Patient ID/)) {
	print "Line 2 does not contain PAT Patient ID\n";
	next;
  }

  $line3 = (<XFRLOG>);			# Line 3 -- Accession Number
  if (!($line3 =~ /ID Accession Number/)) {
	print "Line 3 does not contain ID Accession Number\n";
	next;
  }
  $line4 = (<XFRLOG>);			# Line 4 -- Modality
  if (!($line4 =~ /ID Modality/)) {
	print "Line 4 does not contain ID Modality\n";
	next;
  }
  $line5 = (<XFRLOG>);			# Line 5 -- Images and Kbytes
  if (!($line5 =~ /images and Kbytes/)) {
	print "Line 5 does not contain images and Kbytes\n";
	next;
  }
  $line6 = (<XFRLOG>);			# Line 6 -- Input Start (from modality) DTG
  if (!($line6 =~ /spectrum/)) {
	print "Line 6 does not contain spectrum\n";
	next;
  }
  $line7 = (<XFRLOG>);			# Line 7 -- Output Start (to image servers) DTG
  if (!($line7 =~ /START/)) {
	print "Line 7 does not contain START\n";
	next;
  }
  $line8 = (<XFRLOG>);			# Line 8 -- Output End (to image servers) DTG
  if (!($line8 =~ /END/)) {
	print "Line 8 does not contain END\n";
	print "Line 1 = $line1\n";
	next;
  }
  $line9 = (<XFRLOG>);			# Line 9 -- DICOM source
  if (! (($line9 =~ /All images transferred/) || 
	 ($line9 =~ /Some transfer problems/)) ) {
	print "Line 9 does not contain Some tranfer problems or All images transferred\n";
	next;
  }


  #################################
  #                               #
  # Process Good Group of 9 Lines #
  #                               #
  #################################

  # Check date; choose only yesterday

  @xxxx = split (/ /, $line7);

  $SDATE = @xxxx[1];
  ($sm,$sdy,$syr) = split (/\//, $SDATE);

  if ( ($sdy != $tdd) || ($sm != $tmm) || ($syr != $tyy) ) {next;}

  # Reset the error list
  $errlist = "";

  ################
  # Patient Name #
  ################
  @xxxx = split (/\/\//,$line1);	# Split line by sets of double slashes
  $patient = @xxxx[2];	# Patient name is after 2nd set
  chop $patient;
  $patient = strip_trailing_spaces ($patient);

  ##############
  # Patient ID #
  ##############
  @xxxx = split (/\/\//,$line2);
  $ptid = @xxxx[2];	# Patient ID
  chop $ptid;
  if ( (@xxxx[2] eq "\n") || (@xxxx[2] eq "") ){
    $ptid = "Missing ID";
  }
  # First digit must be D or 0-9
  # Remaining 6 digits must be 0-9
  elsif (!($ptid =~ /[D_0-9][0-9]{6}/)) {
    $errlist .= $ptid." ";
    $ptid = "Bad ID";
  }
  else {
    $ptid = strip_trailing_spaces ($ptid);
    if (length($ptid) != 7) {
	$ptid = "ID too long";
    }
  }

  ####################
  # Accession Number #
  ####################
  @xxxx = split (/\/\//,$line3);
  chop @xxxx[2];
  $acclen = length(@xxxx[2]);
  if ( (@xxxx[2] eq "\n") || (@xxxx[2] eq "") ){
    $accession = "Missing Accession";
  }
  elsif (!(@xxxx[2] =~ /\w{7,7}/)){ # contains exactly 7 alphanum
    $accession = @xxxx[2];     # Accession Number
    $errlist .= $accession." ";
    $accession = "Bad Accession";
  }
  elsif (@xxxx[2] == $ptid) {
    $accession = "Accession=ID";
  }
  else {
    $accession = @xxxx[2];     # Accession Number
    $accession = strip_trailing_spaces ($accession);
  }

  ############
  # Modality #
  ############
  @xxxx = split (/\/\//,$line4);
  $modality = @xxxx[2];       # Modality
  chop $modality;

  #############################
  # Image Count and Kilobytes #
  #############################
  @xxxx = split (/ /, $line5);
  $images = @xxxx[0];	# Number of images
  $kbytes = @xxxx[4];	# Number of K-bytes

  ###########################
  # Input-End Date and Time #
  ###########################
  ($mode,$uid,$owner,$group,$dirsize,$input_month,$input_date,$input_hhmm,$remainder) = split (/\s+/, $line6);
  ($input_hh,$input_mm) = split (/:/, $input_hhmm);
  $imonth = (index($months,$input_month))/3;

  #######################
  # Start Date and Time #
  #######################
  @xxxx = split (/ /, $line7);

  $SDATE = @xxxx[1];
  ($sm,$sdy,$syr) = split (/\//, $SDATE);
  $STIME = @xxxx[3];
  ($shr,$smin,$ssec) = split (/:/, $STIME);
  $SHHMM = substr ($STIME, 0, 5);
 

  #####################
  # End Date and Time #
  #####################
  @xxxx = split (/ /, $line8);

  $EDATE = @xxxx[1];
  ($em,$edy,$eyr) = split (/\//, $EDATE);
  $ETIME = @xxxx[3];
  ($ehr,$emin,$esec) = split (/:/, $ETIME);
  $EHHMM = substr ($ETIME, 0, 5);

  ################
  # Instance UID #
  ################
  # Check for transfer problems and report if so.

  if ($line9 =~ /transfer problems/) {
	$errlist .= " Transfer Problems ";
  }

  if ($SDATE eq $EDATE) {

	##################################################################
	# Compute output-to-servers start and end time difference (secs) #
	##################################################################
	$s = ($shr*3660)+($smin*60)+$ssec;
	$e = ($ehr*3660)+($emin*60)+$esec;
	$d = $e - $s;

	###########################################################################
	# Compute input-complete to output-to-server-start time difference (mins) #
	###########################################################################
	$ic = ($input_hh*60)+$input_mm;
	$os = ($shr*60)+$smin;
	$dw = $os - $ic;
	while ($dw < 0) { $dw += 24*60;} # handle midnight crossover
	
	##########
        # Output #
	##########
	print (OUT    "$j\t$patient\t$ptid\t$accession\t$modality\t$images\t$kbytes\t$input_hhmm\t$dw\t$SDATE\t$SHHMM\t$EHHMM\t$d\t$errlist\n");
	print (STDOUT    "$j\t$patient\t$ptid\t$accession\t$modality\t$images\t$kbytes\t$input_hhmm\t$dw\t$SDATE\t$SHHMM\t$EHHMM\t$d\t$errlist\n");
  }
  else {
	"Start date != End Date; fix this later\m";
  }


  #############################
  # Increment the Event Count #
  #############################

  $j++;
}
close (XFRLOG);		# Close the input file
close (OUT);		# Close the output file

}			# End main

######################################
#                                    #
# Subroutine:  strip_trailing_spaces #
#                                    #
######################################

sub strip_trailing_spaces {

        $s = $_[0];
        $n = 0;
        $len = $pos = length $s;        # Length of source string

        while ( substr($s, --$pos, 1) eq " " )  { $n++; }
        if ($n == 0)    { return $s; }
        else            { return substr($s, 0, $len-$n); }
}

##########################
#                        #
# Subroutine:  yesterday #
#                        #
##########################

sub yesterday {

($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) =
         localtime(time);
$yr = $year-100;
$mo = $mon+1;
#printf ("Todays date = %02d %02d %02d\n", $mo, $mday, $yr);
 
# Compute the target date (yesterday)

  $tyy = $yr; $tmm = $mo; $tdd = $mday;
  if (--$tdd == 0) {  
	if (--$tmm == 0) {	
		$tmm = 12;
	}
	if ($tmm == 1) {$tdd =31;}
	if ($tmm == 2) {$tdd =28;}
	if ($tmm == 3) {$tdd =31;}
	if ($tmm == 4) {$tdd =30;}
	if ($tmm == 5) {$tdd =31;}
	if ($tmm == 6) {$tdd =30;}
	if ($tmm == 7) {$tdd =31;}
	if ($tmm == 8) {$tdd =31;}
	if ($tmm == 9) {$tdd =30;}
	if ($tmm ==10) {$tdd =31;}
	if ($tmm ==11) {$tdd =30;}
	if ($tmm ==12) {$tdd =31;}
  }

$ty = sprintf "%02d", $tyy;
$tm = sprintf "%02d", $tmm;
$td = sprintf "%02d", $tdd;
return ("20"."$ty"."$tm"."$td");
}

