/***********************************************************************
 * Program sends messages to specified host based on command line
 * parameters.  Format of command line call is as follows: 
 *
 *    Message = constant, program name
 *
 *    job_id = Any alpha string describing job running (e.g., "CT Scan")
 *             (default is "unknown")
 *
 *    distributor_queue = machine running distributor (bond,goldfinger,fleming)
 *                        (default is bond), optionally followed by queue number
 *                        (default is 1)
 *
 *    message_type = BEGIN_SCRIPT          END_SCRIPT
 *                   BEGIN_PASS            END_PASS
 *                   BEGIN_EXPORT          END_EXPORT
 *                   IMAGE_RECEIVED        IMAGE_SENT
 *                   PENDING               BOOT
 *                   STUDY
 *                   (default is BEGIN_PASS)
 *
 *    destination_id = this is the broadcast title (destination code)
 *                     from the broadcast script, and the cfg file
 *                     (default is "BOND")
 *                     This parameter is only present for message types
 *                     BEGIN_PASS, END_PASS, BEGIN_EXPORT, and 
 *                     END_EXPORT
 *
 *    transmit_status = status of the transmit (0=success, 1=failure)
 *                      (default is 0)
 *                      This parameter is only present for END_EXPORT
 *
 *    study_id = study ID#, Present only in BEGIN_EXPORT, END_EXPORT & STUDY
 *
 *    patient_name, patient_id, patient_accn (alpha from dicom image)
 *                     Present only in STUDY, IMAGE_RECEIVED & IMAGE_SENT
 *
 *    studies, kbytes (only in PENDING)
 *
 *    images = number of images in study. fadate, fatime, ladate, latime,
 *             edate, etime. Present only in STUDY
 *
 *       ( these last two are almost always defaulted)
 *
 *    target_host  = host name where message will be sent
 *                   (default is rentgen)
 *
 *    target_port  = port on host where message will be sent
 *                   (default is 45000)
 *
 ************************************************************************
 *
 * For BEGIN_EXPORT standard input is assumed to be the output of a
 * du command of the form du -k dir, where dir is the directory which is
 * about to be distributed.  The first item on this line will be
 * interpreted as an integer number and placed in the message with the
 * tag kbytes.  If this is done, the command line format should be:
 *
 * du -k dirname | Message ...(as above)
 *
 ************************************************************************
 *
 * Compile and link with: gcc Message.c -o Message -lsocket -lnsl
 *
 *-----------------------------------------------------------------------
 * 5-8-2 R Moulton - Modified to allow 2 digit queue numbers
 ***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>

/***********************************************************************
 * Valid distributor names, message types
 **********************************************************************/
#define BOND           "bond"
#define GOLDFINGER     "goldfinger"
#define FLEMING        "fleming"
#define BIGBEN         "bigben"
#define RELAY1         "relay1"
#define RELAY2         "relay2"
#define BEGIN_SCRIPT   "BEGIN_SCRIPT"
#define END_SCRIPT     "END_SCRIPT"
#define BEGIN_PASS     "BEGIN_PASS"
#define END_PASS       "END_PASS"
#define BEGIN_EXPORT   "BEGIN_EXPORT"
#define END_EXPORT     "END_EXPORT"
#define IMAGE_RECEIVED "IMAGE_RECEIVED"
#define IMAGE_SENT     "IMAGE_SENT"
#define PENDING        "PENDING"
#define BOOT           "BOOT"
#define STUDY          "STUDY"

/***********************************************************************
 * Command line parameters, preset with default values
 **********************************************************************/
char job_id[24]            = "unknown";
char destination_id[24]    = "BOND";
char distributor_queue[24] = "bond1";
char distributor_host[24]  = "bond";
char message_type[24]      = "BEGIN_PASS";
char transmit_status[24]   = "0";
char target_host[24]       = "rentgen";
char target_ip[]           = "128.252.118.40";
char target_port[24]       = "45000";
char patient_name[48]      = "blank";
char patient_id[24]        = "blank";
char patient_accn[24]      = "blank";
char studies[24]           = "0";
char kbytes[24]            = "0";
char images[24]            = "0";
char study_id[128]         = "unknown";
char fadate[24]            = "20010101";
char fatime[24]            = "000000";
char ladate[24]            = "20010101";
char latime[24]            = "000000";
char edate[24]             = "20010101";
char etime[24]             = "000000";

/***********************************************************************
 * Other global working storage
 **********************************************************************/
char msg[512];
char a[512];
time_t timer;
time_t * ptimer = &timer; 
struct tm * ptime;
int monitor_socket, i, j, retval;
struct hostent *pmonitorent, monitorent;
struct sockaddr_in monitor_sock_name;
fd_set socket_set;
struct timeval timeout;
   
/***********************************************************************
 * Get current time
 **********************************************************************/
get_time_hack()
{  timer = time(ptimer);
   ptime = localtime(ptimer);
}

/***********************************************************************
 * Contatenate current date/time on msg;
 **********************************************************************/
cat_date()
{  char cdate[10]; memset(cdate,0,sizeof(cdate));
   strftime(cdate, 9, "%Y%m%d", ptime);
   strcat(msg, cdate);
}

cat_time()
{  char ctime[10]; memset(ctime,0,sizeof(ctime));
   strftime(ctime, 9, "%H%M%S", ptime);
   strcat(msg, ctime);
}

/***********************************************************************
 * close out socket
 **********************************************************************/
cleanup()
{  retval = shutdown(monitor_socket, 2);
   if(retval != 0) perror("shutdown");
   retval = close(monitor_socket);
   if(retval) perror("close");
}

/***********************************************************************
 * Replace all &'s in string with hex FF
 **********************************************************************/
char * fix (char *s)
{  char * p = s;   
   while (*p) { if (*p == '&') *p = 255; p++; }
   return s;
}
/***********************************************************************
 * Set up message. &'s are used as field delimiters.  Too avoid errors
 * in parsing at the monitor, fields which could conceivably contain
 * &'s are scanned, any &'s that are found are replaced with hex FF.
 * The monitor will need to replace hex FF's with &'s. 
 **********************************************************************/
build_message()
{  int i, c; 
/*---------------------------------------------------- Message header */
   memset(msg, 0, sizeof(msg));
   msg[0] = 11;
/*------------------------------------------------ msg=[message_type] */
   strcat(msg, "msg=");
   strcat(msg, message_type);
/*-------------------------------------- host running the distributor */
   strcat(msg, "&host=");
   strcat(msg, distributor_host);
/*----------------------------------------------------------------------
 * For most messages, the current system time is used for a time hack,
 * being put in start or end as appropriate. For the STUDY message, 
 * both start and end date and time are params. 
 *--------------------------------------------------------------------*/
   if(!strcmp(message_type, STUDY))
   {  strcat(msg, "&start1date=");
      strcat(msg, fadate);
      strcat(msg, "&start1time=");
      strcat(msg, fatime);
      strcat(msg, "&start2date=");
      strcat(msg, ladate);
      strcat(msg, "&start2time=");
      strcat(msg, latime);
      strcat(msg, "&enddate=");
      strcat(msg, edate);
      strcat(msg, "&endtime=");
      strcat(msg, etime);
   }
   else
   {  get_time_hack();
      if(!strcmp(message_type, BEGIN_EXPORT) ||
         !strcmp(message_type, BEGIN_PASS) ||
         !strcmp(message_type, BEGIN_SCRIPT) ||
         !strcmp(message_type, PENDING) ||
         !strcmp(message_type, BOOT) ||
         !strcmp(message_type, IMAGE_RECEIVED))
         strcat(msg, "&startdate=");
      else
         strcat(msg, "&enddate=");
      cat_date();
      if(!strcmp(message_type, BEGIN_EXPORT) ||
         !strcmp(message_type, BEGIN_PASS) ||
         !strcmp(message_type, BEGIN_SCRIPT) ||
         !strcmp(message_type, PENDING) ||
         !strcmp(message_type, BOOT) ||
         !strcmp(message_type, IMAGE_RECEIVED))
         strcat(msg, "&starttime=");
      else
         strcat(msg, "&endtime=");  
      cat_time();
   }
/*------------------------------------------------------------- queue */
   strcat(msg, "&queue=");
   strcat(msg, distributor_queue);
/*------------------------------------------------------------ source */
   strcat(msg, "&jobsequence1=");
   strcat(msg, fix(job_id));
/*----------------------------------------------------------------------
 * Demographic data for image messages
 *--------------------------------------------------------------------*/
   if (!strcmp(message_type, IMAGE_RECEIVED) ||
       !strcmp(message_type, IMAGE_SENT) ||
       !strcmp(message_type, STUDY))
      {  strcat(msg, "&patientname=");
         strcat(msg, fix(patient_name));
         strcat(msg, "&patientid=");
         strcat(msg, fix(patient_id));
         strcat(msg, "&accession=");
         strcat(msg, fix(patient_accn));
      }
/*--------------------------------------- destination (from cfg file) */
   if(!strcmp(message_type, BEGIN_PASS) ||
      !strcmp(message_type, BEGIN_EXPORT) ||
      !strcmp(message_type, END_EXPORT) ||
      !strcmp(message_type, END_PASS))  
      {  strcat(msg, "&jobsequence2=");
         strcat(msg, fix(destination_id));
      }
/*---------------------------------------------------------- study id */
   if(!strcmp(message_type, STUDY) ||
      !strcmp(message_type, BEGIN_EXPORT) ||
      !strcmp(message_type, END_EXPORT))  
      {  strcat(msg,"&study_id=");
         strcat(msg, fix(study_id));
      }
/*-------------------------------------------------- number of images */
   if(!strcmp(message_type, STUDY))
   {  strcat(msg, "&images=");
      strcat(msg, images);
   }
/*--------------------------------------------------- transmit status */
   if(!strcmp(message_type, END_EXPORT))
      if(!strcmp(transmit_status, "0")) strcat(msg, "&status=SUCCESS");
      else strcat(msg, "&status=FAILURE");
   strcat(msg, "&");
/*----------------------------------------------------------------------
 * For BEGIN_EXPORT stdin should be redirected input from "du -k dir"
 * pull up to 10 digits for kbytes parameter
 *--------------------------------------------------------------------*/
   if(!strcmp(message_type, BEGIN_EXPORT))
   {  i = 0;
      while ((c = getchar()) != EOF)
      {  if((char)c == 'k' && i == 0) continue;
         if(!isdigit(c) || i > 9) break;
         if(i++ == 0) strcat(msg, "kbytes=");
         msg[strlen(msg)] = c;
      }
      if(i) strcat(msg, "&");
   }
/*------------------------------------- studies & kbytes, for PENDING */
   if(!strcmp(message_type, PENDING))
   {  strcat(msg, "&studies=");
      strcat(msg, studies);
      strcat(msg, "&kbytes=");
      strcat(msg, kbytes);
   }
/*--------------------------------------------------- add closing 28  */
   msg[strlen(msg)] = 28;
/*-------------------------------- Uncomment to debug message formatting
 * memset(a,0,sizeof(a)); strncpy(a, msg+1, strlen(msg) -2);
 * strcat(a,"\n"); printf(a); 
 *--------------------------------------------------------------------*/
}

/***********************************************************************
 * Validate command line parameters, loading into values if non-blank
 * return non-zero if validation failed
 **********************************************************************/
int validate_params(int numpars, char * par[])
/*----------------------------------------------------------------------
 * i is parameter number we are looking at.  Numpars is the total number
 * of parameters.  Since the program name is the first parameter in C,
 * we decrement numpars to get the total number of "real" parameters.
 *--------------------------------------------------------------------*/
{  int i = 0; numpars--;
/*----------------------------------------------------------------------
 * First parameter is job_id, no validation
 *--------------------------------------------------------------------*/
   i++;
   if (numpars < i) return 0;
   strcpy(job_id, par[i]);
/*----------------------------------------------------------------------
 * Next parameter is distributor, matched against valid names
 * host may (but not must) have one or two digit queue number at end
 *--------------------------------------------------------------------*/
   i++;
   if(numpars < i) return 0;
   strcpy(distributor_queue, par[i]);
   strcpy(distributor_host, par[i]);
   if (strcmp(distributor_host, BOND) &&
       strcmp(distributor_host, GOLDFINGER) &&
       strcmp(distributor_host, FLEMING) &&
       strcmp(distributor_host, RELAY1) &&
       strcmp(distributor_host, RELAY2) &&
       strcmp(distributor_host, BIGBEN))
   {  distributor_host[strlen(distributor_host) -1] = 0;
      if(strcmp(distributor_host, BOND) &&
         strcmp(distributor_host, GOLDFINGER) &&
         strcmp(distributor_host, FLEMING) &&
         strcmp(distributor_host, RELAY1) &&
         strcmp(distributor_host, RELAY2) &&
	 strcmp(distributor_host, BIGBEN))
      {  distributor_host[strlen(distributor_host) -1] = 0;
         if(strcmp(distributor_host, BOND) &&
            strcmp(distributor_host, GOLDFINGER) &&
            strcmp(distributor_host, FLEMING) &&
            strcmp(distributor_host, RELAY1) &&
            strcmp(distributor_host, RELAY2) &&
	    strcmp(distributor_host, BIGBEN))
         {  printf (par[i]);
            printf("invalid distributor host name on command line\n");
            return 1;
         }
      }
   }
/*----------------------------------------------------------------------
 * Next parameter is message type code, matched against valid codes
 *--------------------------------------------------------------------*/
   i++;
   if(numpars < i) return 0;
   strcpy(message_type, par[i]);
   if(strcmp(message_type, BEGIN_SCRIPT))
     if(strcmp(message_type, END_SCRIPT))
       if(strcmp(message_type, BEGIN_PASS))
         if(strcmp(message_type, END_PASS))
           if(strcmp(message_type, BEGIN_EXPORT))
             if(strcmp(message_type, END_EXPORT))
               if(strcmp(message_type, IMAGE_RECEIVED))
                 if(strcmp(message_type, IMAGE_SENT))
                   if(strcmp(message_type, PENDING))
                     if(strcmp(message_type, BOOT))
                       if(strcmp(message_type, STUDY))
                       {  printf (message_type);
                          printf("\ninvalid message type on command line\n");
                          return 1;
                       }
/*----------------------------------------------------------------------
 * Next parameter is destination id, which is the broadcast title
 * from the script, referenced in the cfg file.  No validation
 * However, this parameter is only present on begin and end pass and
 * export.
 *--------------------------------------------------------------------*/
   if(!strcmp(message_type, BEGIN_PASS) ||
      !strcmp(message_type, BEGIN_EXPORT) ||
      !strcmp(message_type, END_EXPORT) ||
      !strcmp(message_type, END_PASS))  
      {   i++;
          if (numpars < i) return 0;
          strcpy(destination_id, par[i]);
      }
/*----------------------------------------------------------------------
 * Next parameter is the transmit status code, 0=success,1=failure,
 * however this parameter is only present on end export.
 *--------------------------------------------------------------------*/
   if(!strcmp(message_type, END_EXPORT))
   {  i++;
      if (numpars < i) return 0;
      strcpy(transmit_status, par[i]);
      if(strcmp(transmit_status, "0"))
         if(strcmp(transmit_status, "1"))
         {   printf("invalid transmit status on END EXPORT\n");
             return 1;
         }
   }
/*----------------------------------------------------------------------
 * Next parameter is the study id code (directory name).  Not validated.
 * Present on begin and end export and study
 *--------------------------------------------------------------------*/
   if(!strcmp(message_type, STUDY) ||
      !strcmp(message_type, BEGIN_EXPORT) ||
      !strcmp(message_type, END_EXPORT))  
      {   i++;
          if (numpars < i) return 0;
          strcpy(study_id, par[i]);
      }
/*----------------------------------------------------------------------
 * For image received and sent and study, next parameters are 
 * demographics, treated simply as alpha
 *--------------------------------------------------------------------*/
   if(!strcmp(message_type, IMAGE_RECEIVED) ||
      !strcmp(message_type, IMAGE_SENT) ||
      !strcmp(message_type, STUDY))
   {  i++;
      if (numpars < i) return 0;
      strcpy(patient_name, par[i]);
      i++;
      if (numpars < i) return 0;
      strcpy(patient_id, par[i]);
      i++;
      if (numpars < i) return 0;
      strcpy(patient_accn, par[i]);
   }
/*----------------------------------------------------------------------
 * studies and kbytes, only for PENDING
 *--------------------------------------------------------------------*/
   if(!strcmp(message_type, PENDING))
   {  i++;
      if(numpars < i) return 0;
      strcpy(studies, par[i]);
      i++;
      if(numpars < i) return 0;
      strcpy(kbytes, par[i]);
   }
/*----------------------------------------------------------------------
 * images, fadate, fatime, ladate, latime, edate and etime, only for STUDY
 *--------------------------------------------------------------------*/
   if(!strcmp(message_type, STUDY))
   {  i++;
      if(numpars < i) return 0;
      strcpy(images, par[i]);
      i++;
      if(numpars < i) return 0;
      strcpy(fadate, par[i]);
      i++;
      if(numpars < i) return 0;
      strcpy(fatime, par[i]);
      i++;
      if(numpars < i) return 0;
      strcpy(ladate, par[i]);
      i++;
      if(numpars < i) return 0;
      strcpy(latime, par[i]);
      i++;
      if(numpars < i) return 0;
      strcpy(edate, par[i]);
      i++;
      if(numpars < i) return 0;
      strcpy(etime, par[i]);
   }
/*----------------------------------------------------------------------
 * Next parameter is host name, validated later
 *--------------------------------------------------------------------*/
   i++;
   if(numpars < i) return 0;
   strcpy(target_host, par[i]);
/*----------------------------------------------------------------------
 * Next parameter is host port, which must be numeric
 *--------------------------------------------------------------------*/
   i++;
   if(numpars < i) return 0;
   strcpy(target_port, par[i]);
   return 0;
}

/***********************************************************************
 * main function
 **********************************************************************/
int main(int numpars, char *par[])
{  if(validate_params(numpars, par)) exit(1);
   build_message();
/*----------------------------------------------------------------------
 * get a socket
 *--------------------------------------------------------------------*/
   monitor_socket = socket(AF_INET, SOCK_STREAM, 0);
   if(monitor_socket == -1) {perror("socket"); exit(1);}
/*----------------------------------------------------------------------
 * Set socket to non-blocking.  This is done so that later, when we
 * connect, we can avoid a long wait if the monitor is down.
 *--------------------------------------------------------------------*/
   i = fcntl(monitor_socket, F_GETFL, 0);
   fcntl(monitor_socket, F_SETFL, O_NONBLOCK|i);
/*----------------------------------------------------------------------
 * get host by name
 *--------------------------------------------------------------------*/
   pmonitorent = gethostbyname(target_host);
   if(pmonitorent == NULL) pmonitorent = gethostbyname(target_ip);
   if(pmonitorent == NULL) 
   {  perror("gethost");
      cleanup();
      exit(1);
   }
/*----------------------------------------------------------------------
 * load socket address struct
 *--------------------------------------------------------------------*/
   monitorent = *pmonitorent;
   monitor_sock_name.sin_family = monitorent.h_addrtype;
   monitor_sock_name.sin_port = htons(atoi(target_port));
   monitor_sock_name.sin_addr = *((struct in_addr *)monitorent.h_addr);
/*----------------------------------------------------------------------
 * connect to monitor host.  The socket has been set not to block, but
 * will not connect instantly. We therefore use select to insure that
 * the connection was made.  However, we only wait 10 seconds, so that
 * broadcasts will not be slowed down if monitor is not responding.
 *--------------------------------------------------------------------*/
   i = sizeof(monitor_sock_name);
   retval = connect(monitor_socket, (struct sockaddr *)&monitor_sock_name, i);
   FD_ZERO (&socket_set);
   FD_SET (monitor_socket, &socket_set);
   timeout.tv_sec = 30;
   timeout.tv_usec = 0;
   j = monitor_socket + 1;
   retval = select(j, NULL, &socket_set, NULL, &timeout);
   if (retval < 1)
   {  perror("send select");
      cleanup();
      exit(1);
   }
/*----------------------------------------------------------------------
 * send message
 *--------------------------------------------------------------------*/
   i = strlen(msg);
   retval = send(monitor_socket, msg, i, 0);
   if(retval == -1)
   {  perror("send");
      cleanup();
      exit(1);
   }
/*----------------------------------------------------------------------
 * receive OK; we don't even look at it 
 *--------------------------------------------------------------------*/
   timeout.tv_sec = 10;
   FD_ZERO (&socket_set);
   FD_SET (monitor_socket, &socket_set);
   retval = select(j, &socket_set, NULL, NULL, &timeout);
   if (retval < 1)
   {  memset(a, 0, sizeof(a));
      strcpy(a, message_type);
      strcat(a, ": recv select"); 
      perror(a);
      cleanup();
      exit(1);
   }
   retval = recv(monitor_socket, a, sizeof(a), 0);
   if (retval <= 0)
   {  perror("select");
      cleanup();
      exit(1);
   }
   cleanup();
}
