/**************************************************************************************************
 * PACS_WARN.C                                                                            R Moulton
 * PACS Warning server.
 * Receives warning messages from Unix servers and sends them to printer queues, e-mails and/or
 * pager e-mails, as appropriate.  Also receives signal requests and sends signals to specified
 * processes.
 *-------------------------------------------------------------------------------------------------
 * 10-23-03 R Moulton Initial Write
 * 02-01-05 R Moulton Added Logging, use of .ini file
 * 04-25-05 R Moulton Added signal messages.
 *************************************************************************************************/

#define _VMS_MODE_

#include stdio
#include unistd
#include stdlib
#include string
#include errno
#include types
#include unixio
#include fcntl
#include dirent

#include time

#include ssdef
#include starlet
#include iodef
#include descrip
#include rms

#include socket
#include in
#include netdb
#include inet

#include "[source.common]queue_file.h"
#include "[source.common]intermail.h"
#include "[source.common]dbg.h"
#include "[source.common]ini.h"

/*--------------------------------------------------------------------------------------- Macros */
#define MAKE_STR(A,B) if (A=malloc(strlen(B)+3)) strcpy(A,B);
#define FREE_STR(A) if (A) { free(A); A=NULL; }
#define STR_EQ(A,B) !strcasecmp(A,B)

/*------------------------------------------------------------------------------------ Constants */
#define SB 11
#define EB 28
#define PROCESS_NAME        "PACS_WARN"
#define BUFFER_SIZE         10240       /* Buffer sizes */
#define MAX_PRT             2           /* Maximum number of printer queues for a message */
#define MAX_EMAILS          5           /* Maximum number of e-mails for a message */
#define MAX_PRS             2           /* Maximum number of processes for a signal */

/*----------------------------------------------------- Environment variables, storage, defaults */
static int port = 52226;

static FILE * debug_fp;
static int file_number = 0;

static char log_dir[256] = "[.log]";
static char log_pfn_format[128] = "%s%4d%02d%02d.LOG";
static char log_pfn[256];
static int log_purge_hour;                            /* hour to purge logs (0-23) */
static time_t last_purge_date;								/* date last purge was done (yyyymmdd) */
static int log_retain_days;							      /* # of days to retain log files */

/*------------------------------ last CheckStudy message (we cut out duplicate ones) */
static time_t check_study_basis = 0;


/*-------------------------------------------------------------------------- Function prototypes */
void read_ini(void);
void count_transaction(void);           /* Increment transaction counts, displays */
void tcpserver (void);                    /* main TCP/IP server loop */
void accept_connect(void);               /* Accept connection on listen socket */
void process_tcp_read(int slot);         /* read data on socket slot*/
void process_tcp_write(int slot);        /* write data on socket slot */
void clear_socket(int slot);             /* Clear socket table slot*/
void close_listen_socket(void);                /* close listen socket */
int shutdown_link(void);
void set_program_name (void);                /* Set process id */
void clear_program_name(void);            /* Clears process name */
void dbgm(char *msg, char *buff);
void ldbgm(char *msg, char *buff);
void create_listen_socket(void);          /* open PACS listen socket */
void process_message(int slot);
void prt_log (char *msg, char * buff);
void purge_logs(void);
void ldbg(char * msg);

static char em[256];
static char ib[BUFFER_SIZE];
/*----------------------------------------------------------------------
 * Socket data structure
 *--------------------------------------------------------------------*/
#define MAX_CONNECTIONS     25       /* Maximum allowable connections */
/*---------------------------------------------- Socket status codes */
#define SS_INACTIVE         0           /* No socket active this slot */
#define SS_READ             1           /* Waiting for input from PACS */
#define SS_WRITE            2           /* Waiting for output to PACS */

struct _sockets_ 
{  int status;                       /* Current socket status */
   int socket;                       /* Socket number */
   char * message_buffer;            /* Completed message */
   char * ip;                        /* IP address for this connection */
};
struct _sockets_ sockets[MAX_CONNECTIONS]; /* PACS inbound sockets */

/*----------------------------------------------------------------------
 * Valid connecting IP addresses (enter -1 for subnet, matches anything)
 *--------------------------------------------------------------------*/
struct _ipa_ {short int addr[4];}; 
static struct _ipa_ ips[] = {
   { 128, 252, -1, -1 },	 
   { 159, 251, -1, -1 },
};
#define MAX_IPS sizeof(ips) / sizeof(struct _ipa_)


static struct timeval timeval = {60, 0};
static int new_socket, listen_sock = 0;
static struct sockaddr_in listen_sock_name;

static int main_loop = 1, tcp_loop = 1;
/***********************************************************************
 ***********************************************************************
 *********************** Main function *********************************
 ***********************************************************************
 **********************************************************************/
int main (int argc, char **argv)
{  int i;
   read_ini();
   ldbg("PACS Warning Server started");
   while (main_loop)
   {  tcp_loop = 1;
      set_program_name ();                          /* Set process ID */
      create_listen_socket();
/*--------------------------------------- Clear PACS connection table */
      for(i=0; i < MAX_CONNECTIONS; i++)
      {  sockets[i].status = SS_INACTIVE;
			sockets[i].socket = 0;
			sockets[i].message_buffer = NULL;
			sockets[i].ip = NULL;
      }
/*----------------------------------------- main loop is in tcpserver */
      sprintf(em,"Calling tcpserver, port %d", port); ldbg(em);
      tcpserver();
   }
   ldbg("Exiting main loop");
   clear_program_name();
   exit(0);
}

/***********************************************************************
 ***********************************************************************
 **************** Main TCP server processing loop **********************
 ***********************************************************************
 **********************************************************************/

void tcpserver (void)
{  int rc, i, a;
   fd_set rsock, wsock, esock;
/*----------------------------------------------------------------------
 * Main socket processing loop
 * -------------------------------------------------------------------*/
	while (tcp_loop)
	{	dbg ("tcp_loop");
      purge_logs();
/*---------------------------------- identify connections for testing */
		FD_ZERO(&rsock);
      FD_ZERO(&wsock);
      FD_ZERO(&esock);
      FD_SET(listen_sock, &rsock);
/*----------------------------------------------------------------------
 * All sockets are set for read and error, those with waiting message 
 * are also set to write
 * -------------------------------------------------------------------*/
      for (i = 0; i < MAX_CONNECTIONS; i++)
      {  if (sockets[i].socket)
			{  FD_SET(sockets[i].socket, &rsock);
				FD_SET(sockets[i].socket, &esock);
            if (sockets[i].status == SS_WRITE) FD_SET(sockets[i].socket, &wsock);
			}
		}
         
/*----------------------------------------------- see if any activity */
      rc = select (FD_SETSIZE, &rsock, &wsock, &esock, &timeval);
		sprintf(em,"select returned %d", rc); dbg(em); 
/*----------------------------------------------------------- timeout */
      if (!rc) continue;

/*--------------- activity on listen socket is request for connection */
      if (FD_ISSET (listen_sock, &rsock)) accept_connect();
			
/*------------------------------------------------------ PACS sockets */
      for (i = 0; i < MAX_CONNECTIONS; i++)
		{  if (sockets[i].socket)
			{  if(FD_ISSET(sockets[i].socket, &esock)) { clear_socket(i); continue; }
				if(FD_ISSET(sockets[i].socket, &rsock)) process_tcp_read(i);
				if(FD_ISSET(sockets[i].socket, &wsock)) process_tcp_write(i);
			}
		}
	}
/*--------------------------------------------------- Shutdown server */
   ldbg("Exiting tcp loop");
   for (i = 0; i < MAX_CONNECTIONS; i++) clear_socket(i);
   close_listen_socket();
}

/***********************************************************************
 * Setup and open server listen socket 
 **********************************************************************/
void create_listen_socket (void)
{  struct hostent *hostentptr, hostentstruct;
   char hostname[256];
   int rc, on = 1, flag = 0, count = 15;
   size_t sl;

   while (count--)
   {  if(flag) sleep(60); flag = 1;
/*----------------------------------------------------- make a socket */
		if ((listen_sock = socket (AF_INET, SOCK_STREAM, 0)) == -1)
      {	perror ("listen socket");
			continue;
      }

      if (setsockopt (listen_sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (on)))
      {	perror ("listen socket setsockopt");
			continue;
      }
/*----------------------------------------------- get local host name */
      rc = gethostname (hostname, sizeof (hostname));
      if (rc) {perror("gethostname"); close_listen_socket(); continue;}
/*----------------------------- get pointer to network data structure */
      hostentptr = gethostbyname (hostname);
      if (!hostentptr) {perror("gethostbyname"); close_listen_socket(); continue;}

      hostentstruct = *hostentptr;
      listen_sock_name.sin_family = hostentstruct.h_addrtype;
      listen_sock_name.sin_port = htons(port);
      listen_sock_name.sin_addr = * ((struct in_addr *)hostentstruct.h_addr);
/*------------------------------------------------ bind listen socket */
      sl = sizeof(listen_sock_name);
      rc = bind (listen_sock, (struct sockaddr*)&listen_sock_name, sl);
      if (rc) {perror("bind"); close_listen_socket(); continue;}
/*-------------------------------------------- listen for connections */
      rc = listen (listen_sock, 5);
      if (rc) {perror("listen"); close_listen_socket(); continue;}
      sprintf(em, "Listen socket created on port %d", port); ldbg(em);
      return;
   }
   ldbg("Unable to create listen socket"); 
   clear_program_name();
   exit(1);
}

/*----------------------------------------------------------------------
 * Accept new connection on listen socket.  
 *--------------------------------------------------------------------*/
void accept_connect (void)
{  int rc, i, j, flag, on = 1;
   unsigned char *ptr;
   size_t sl;
   char alpha_ip[20];
/*------------------------------------------------- Accept connection */
   sl = sizeof (listen_sock_name);
   new_socket = accept (listen_sock, (struct sockaddr*)&listen_sock_name, &sl);
   if (new_socket == -1) { perror ("accept"); return; }

/*------- If this socket number already in the table, clear its entry */
   for (i = 0; i < MAX_CONNECTIONS; i++)
      if(new_socket == sockets[i].socket) clear_socket(i);

   rc = setsockopt (new_socket, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof (on));
   if (rc == -1) perror ("setsockopt");

   ptr = (unsigned char*)&listen_sock_name.sin_addr;
   memset(alpha_ip,0,sizeof(alpha_ip));
   sprintf(alpha_ip,"%d.%d.%d.%d",ptr[0],ptr[1],ptr[2],ptr[3]);

/*------------------------------------ Validate IP against ips table */
   for (i = 0; i < MAX_IPS; i++)
   {  flag = 1;
      for (j = 0; j < 4; j++)
         if (ips[i].addr[j] != -1 && ips[i].addr[j] != ptr[j]) { flag = 0; break; }
      if (flag) break;
   }
   if (!flag)
   {  sprintf(em, "Attempt to connect from: %s rejected", alpha_ip); ldbg(em);
      close(new_socket);
      return;
   }
/*---------------------------------------------------------------------
 * Then we look for an empty slot to put the new connection in. Failing 
 * this, we have exceeded connections limit
 * ------------------------------------------------------------------*/
   j = -1;
   for (i = 0; i < MAX_CONNECTIONS; i++)
   {  if (sockets[i].status == SS_INACTIVE) { j = i; break; }
   }
   if (j == -1)
   {  ldbg("Connection limit exceeded");
      sprintf(em, "Attempt to connect from: %s rejected", alpha_ip); ldbg(em);
      close(new_socket);
      return;
   }
/*---------------------------------------------------------------------
 * Slot j is our slot, initialize and set ready to read.
 *-------------------------------------------------------------------*/
   if (sockets[j].status != SS_INACTIVE) clear_socket(j);
   sockets[j].socket = new_socket;
   sockets[j].message_buffer = malloc(BUFFER_SIZE);
   MAKE_STR(sockets[j].ip,alpha_ip)
   sprintf (em,"Connected to: %s", alpha_ip); dbg(em);
/*---------------------------------------- Make sure malloc's worked */
   if (!sockets[j].message_buffer || !sockets[j].ip)
   {  ldbg("Memory allocation error: new socket");
      clear_program_name();
      exit(1);
   }
/*---------------------------- Set ready to read, clear input buffer */
   sockets[j].status = SS_READ;
   memset(sockets[j].message_buffer,0,BUFFER_SIZE);
   return;
}


/*----------------------- Close out socket entry at param table slot */
void clear_socket(int i)
{  if (sockets[i].socket) 
   {  shutdown(sockets[i].socket,2); 
      close(sockets[i].socket); 
   }
   FREE_STR(sockets[i].message_buffer);
   FREE_STR(sockets[i].ip);
   sockets[i].status = SS_INACTIVE;
   sockets[i].socket = 0;
}

/***********************************************************************
 ***********************************************************************
 * Process TCP/IP read on PACS socket slot i
 ***********************************************************************
 **********************************************************************/
void process_tcp_read(int i)
{  int j, feed, facility, rc;
   char * sb_ptr, * eb_ptr;
   char id[128], * id_ptr, buffer[BUFFER_SIZE];
/*----------------------------------------------------------------------
 * If we are not waiting for input, clear out current message
 * ------------------------------------------------------------------ */
   if (sockets[i].status != SS_READ)
   {  dbg ("PACS socket reset");
      memset(sockets[i].message_buffer,0,BUFFER_SIZE);
      sockets[i].status == SS_READ;
   }
/*-------------------------------------------------------- read input */
   memset (buffer, 0, BUFFER_SIZE);
   rc = recv (sockets[i].socket, buffer, BUFFER_SIZE, 0);
   if (rc == 0){dbg("PACS socket closed by user"); clear_socket(i); return;}
   if (rc < 0)
   {  sprintf(em,"PACS socket error %d on read", rc); ldbg(em); 
      perror("PACS receive"); clear_socket(i); return;
   }
   sprintf(em,"read on PACS socket %d, returned %d",sockets[i].socket,rc); dbg(em);
/*----------------------------------------------------------------------
 * At beginning of message, wait for SB; Otherwise, if an SB is found, 
 * dump whatever message we have and restart with the SB
 * -------------------------------------------------------------------*/
   sb_ptr = strchr (buffer, SB);
   if (!strlen(sockets[i].message_buffer) && !sb_ptr) return;
   if (sb_ptr)
   {  memset(sockets[i].message_buffer,0,BUFFER_SIZE);
      strcpy(sockets[i].message_buffer, sb_ptr);
   }
/*----------------------------------------------------------------------
 * Check for overflow; if found, clear buffer and exit; wait for next
 * -------------------------------------------------------------------*/
   else if (strlen(sockets[i].message_buffer) + strlen(buffer) > BUFFER_SIZE)
   {  memset(sockets[i].message_buffer,0,BUFFER_SIZE);
      return;
   }
/*-------------------------------------------------------- Copy it in */
   else strcat(sockets[i].message_buffer, buffer);

/*------------------------------------------- see if there is 28 <EB> */
   eb_ptr = strchr(sockets[i].message_buffer, EB);
   if (!eb_ptr) return;
/*----------------------------------------------------------------------
 * We may have a complete message.  Check for SB before EB 
 *--------------------------------------------------------------------*/
   sb_ptr = strchr(sockets[i].message_buffer, SB);
   if (!sb_ptr || (sb_ptr > eb_ptr))
   {  memset (sockets[i].message_buffer, 0, BUFFER_SIZE);
      return;
   }
   ldbgm("Message received", sockets[i].message_buffer);
/*----------------------------------------------------------------------
 * We now have a completed message; Process it
 *--------------------------------------------------------------------*/
   process_message(i);
}

/*----------------------------------------------------------------------
 * Process message received on socket slot.
 * Messages are "&" separated fields, of the form:
 *		<SB>message
 *		print=printerqueue           (queue message to this printer queue)
 *		e-mail=e-mailaddress         (e-mail to this address)
 *		process=processname          (process to send e-mail to)
 *		title=titleline              (e-mail title, if not blank)
 *		signal=signalname            (signal to send to processes)
 *		msg=messagetext              (message to be printed/e-mailed)
 *		<EB>
 *	
 *	A single message may be sent to up to MAX_PRT printer queues and up to
 *	MAX_EMAILS e-mail addresses.  If there are multiple title fields, the 
 *	first one will be used.  The msg field must be the last one in the 
 *	message.  messagetext is the text of the message, including needed
 *	formatting characters.  Ampersands in the messagetext are considered 
 *	to be part of the message, not field delimiters.  messagetext ends at 
 *	the <EB> The server responds to the client with <SB>ACK<EB> if the 
 *	message format was valid, with <SB>NAK<EB> if there was a formatting 
 *	problem with the message. Validation of e-mail addresses, printer 
 *	queue names, and message and title format is the responsibility of the
 *	client.
 *---------------------------------------------------------------------*/
void process_message(int slot)
{	char * printer_queue[MAX_PRT] = { NULL };
	char * e_mail_address[MAX_EMAILS] = { NULL };
   char * processes[MAX_PRS] = { NULL };
	char * title_line = NULL;
   char * signal_name = NULL;
	char * message_text = NULL;
	char * ack = "\013ACK\034";
	char * nak = "\013NAK\034";
	char * p, * q, * message = NULL;
	int i, rc, pc = 0, ec = 0, qc = 0;
	FILE * fp;
	char pfn[128];
	
/*-------------------------------------------- Pull message, load nak */
	MAKE_STR(message,sockets[slot].message_buffer)
	strcpy(sockets[slot].message_buffer,nak);
	p = message;

	if (*p == SB) p++; else goto finish;    /* Doesn't start with <SB> */

	p = strtok(p,"&");                               /* Pull "message" */
	if (!p || strcasecmp(p,"message")) goto finish;

/*-------------------------------- This loop processes all the fields */
	while (1)
	{	p = strtok(NULL, "=");             /* Pull next field type code */
		if (!p) goto finish;
/*------------------------------------------------ print=printerqueue */
		if (STR_EQ(p,"print"))
		{	p = strtok(NULL,"&");
			if (!p) goto finish;
			if (pc < MAX_PRT)
			{	MAKE_STR(printer_queue[pc],p);
				pc++;
				continue;
			}
		}
/*---------------------------------------------------- e-mail=address */
		if (STR_EQ(p,"e-mail"))
		{	p = strtok(NULL,"&");
			if (!p) goto finish;
			if (ec < MAX_EMAILS)
			{	MAKE_STR(e_mail_address[ec],p);
				ec++;
				continue;
			}
		}
/*---------------------------------------------------- process=prs_name */
		if (STR_EQ(p,"process"))
		{	p = strtok(NULL,"&");
			if (!p) goto finish;
			if (ec < MAX_PRS)
			{	MAKE_STR(processes[qc],p);
				qc++;
				continue;
			}
		}
/*----------------------------- title=titleline (only first one used) */
		if (STR_EQ(p,"title"))
		{	p = strtok(NULL,"&");
			if (!p) goto finish;
			if (!title_line) MAKE_STR(title_line,p)
			continue;
		}
/*----------------------------- signal=signal_name (only first one used) */
		if (STR_EQ(p,"signal"))
		{	p = strtok(NULL,"&");
			if (!p) goto finish;
			if (!signal_name) MAKE_STR(signal_name,p)
			continue;
		}
/*------------------------------------------------- msg=message_text */
		if (STR_EQ(p,"msg"))
		{	p = strtok(NULL,"\034");
			if (!p) goto finish;
			MAKE_STR(message_text,p);
			break;
		}
	}
	if (!pc && !ec && !qc) goto finish;
	if (!title_line) MAKE_STR(title_line,"\0");
   if (!signal_name) MAKE_STR(signal_name,"SIGUSR1");

/*------------------------------- We cut out duplicate CheckStudy msgs */
   if (STR_EQ(title_line,"CheckStudy"))
   {  time_t now = time(NULL);
      if ((now - check_study_basis) > 6300) check_study_basis = now;
      else 
      {  sprintf(em,"Only %d minutes since last CheckStudy, skipping", (now - check_study_basis)/60); ldbg(em);
         goto finish;
      }
   }
/*---------------------------------------- Write message text to file */
   if (file_number++ > 25) file_number = 1;
	sprintf(pfn,"[.MSG]MSG%02d.DAT;1", file_number);
	delete(pfn);
	fp = fopen(pfn, "wb");
	if (!fp)
	{	sprintf(em,"Could not open %s", pfn); ldbg(em);
		goto finish;
	}
	i = fputs(message_text,fp);
	if (i < 0)
	{	sprintf(em,"Write to %s failed", pfn); ldbg(em);
		fclose(fp);
		goto finish;
	}
	fclose(fp);
/*------------------------------------------------- queue to printers */
	for (i=0; i < pc; i++) 
	{	
		rc = queue_file(printer_queue[i], pfn, 0, 0);
		sprintf(em,"queue_file queue=%s, pfn=%s, rc=%d\n", printer_queue[i], pfn, rc); dbg(em);
		sprintf(em,"queued to %s, rc=%d", printer_queue[i], rc); prt_log(em,NULL);
	}
/*--------------------------------------------------- send to e-mails */
	for (i=0; i < ec; i++) 
   {  rc = mail_message(e_mail_address[i], title_line, pfn); 
      sprintf(em,"e-mail %s, title %s, pfn %s, rc=%d\n", e_mail_address[i], title_line, pfn, rc); dbg(em);
      sprintf(em,"e-mailed to %s, rc=%d", e_mail_address[i], rc); prt_log(em,NULL);
   }
/*-------------------------------------------------- generate signals */
   for (i=0; i < qc; i++)
   {  FILE * cmd;
      p = tmpnam(NULL);
      sprintf(em,"ps -u spectrum -o pid,comm | grep \"%s\" | cut -c1-6 > %s", processes[i], p);
      system(em);
      if (!access(p,F_OK) && cmd = fopen(p))
      {  while (q = fgets(em, sizeof(em), cmd))
         {  pid_t pid = atoi(p);
            /* Add other signals as needed */
            if(STR_EQ(signal_name,"SIGUSR1")) 
            {  kill(pid,SIGUSR1);
               sprintf(em,"Sent %s signal to %d-%s", signal_name, pid, processes[i]); 
               dbg(em); prt_log(em,NULL);
            }
            if(STR_EQ(signal_name,"SIGUSR2")) 
            {  kill(pid,SIGUSR2);
               sprintf(em,"Sent %s signal to %d-%s", signal_name, pid, processes[i]); 
               dbg(em); prt_log(em,NULL);
            }
         }
         fclose(cmd);
         delete(cmd);
      }
   }
/*------------------------------------------------- change nak to ack */
	strcpy(sockets[slot].message_buffer,ack);

/*---------------------------------------------------------- Clean up */
finish:
	sockets[slot].status = SS_WRITE;
/*------------------------------------------------------- Free memory */
	for (i=0; i < pc; i++) FREE_STR(printer_queue[i])
	for (i=0; i < ec; i++) FREE_STR(e_mail_address[i])
	FREE_STR(title_line)
	FREE_STR(message_text)
	FREE_STR(message)
	return;
}
/***********************************************************************
 ***********************************************************************
 * Process PACS TCP/IP write
 ***********************************************************************
 **********************************************************************/
static char tb[BUFFER_SIZE];
void process_tcp_write(int i)
{  int ml, rc;
   if (sockets[i].status != SS_WRITE) return;
   dbgm("TCP write", sockets[i].message_buffer);
   memset(tb,0,BUFFER_SIZE);
   strcpy(tb,sockets[i].message_buffer);
   ml = strchr(tb, EB) - strchr(tb, SB) +2;
   rc = send(sockets[i].socket, tb, ml, 0);
   if (rc < 0)
   {  perror ("PACS send");
      clear_socket(i);
      return;
   }
   sprintf(em, "%d characters sent to PACS", rc); dbg(em);
   memset(sockets[i].message_buffer,0,BUFFER_SIZE);
   sockets[i].status = SS_READ;
}

void close_listen_socket(void)
{  if (listen_sock <= 0) return;
   if (shutdown(listen_sock,2)) {perror("listen socket shutdown"); return;}
   if (close(listen_sock)) perror("listen socket close");
   return;
}

/***************************************************** Set process ID */

void set_program_name (void)
{  char pn[] = PROCESS_NAME;
   int rc;
   $DESCRIPTOR(pn_startup, pn);
   rc = sys$setprn (&pn_startup);
   if (rc == SS$_NORMAL)
   {  sprintf(em,"Process name set to %s", pn); 
      ldbg(em);
   }
   else
   {  printf("Process is already running\n");
      sys$exit(1);
   }
   return;
}
void clear_program_name (void)
{  int status, count = 1;
   char name[] = "VMS_001";
   $DESCRIPTOR(pn, name);
   while (1)
   {  sprintf (name, "VMS_%03d", count++);
      status = sys$setprn (&pn);
      if (status == SS$_NORMAL) break;
   }
}

/*----------------------------------------- Debug to dump buffer area */
void ldbgm(char *msg, char * buff) { dbgm(msg,buff); prt_log(msg,buff); }
void dbgm(char*msg, char*buff)
{	char p[53], *ptr;
	int piece = sizeof(p) - 3, cont = 1;
	if (!debug_is_on()) return;
	if (strlen(msg) > (sizeof(p) -3)) dbg(msg);
	else
	{	 memset(p,0,sizeof(p)); strcpy(p,msg);
		 if (p[strlen(p)-1] != ' ')p[strlen(p)] = ' ';
		 memset(p+strlen(p),'-',(sizeof(p)-strlen(p)-1));
		 dbg(p);
	}
  
	if (!buff) { dbg("NULL buffer\n"); return; }
	while (cont)
	{	memset(p,0,sizeof(p)); strcpy(p,"> ");
		if (strlen(buff) < piece) { strcat(p,buff); cont = 0; }
		else { memcpy(p+2,buff,piece); buff += piece; }
		if (ptr = strchr(p,SB)) *ptr = '.';
		if (ptr = strchr(p,EB)) *ptr = '.';
		while (ptr = strchr(p,'\r')) *ptr = '~';
		dbg(p);
	}
	dbg(" ");
}

/*-----------------------------------------------------------------------------
 * Returns pointer to string which is cat of s1 and s2, but s1 can be a 
 * constant.
 *---------------------------------------------------------------------------*/
char * strcmb(char *s1, char *s2)
{  static char ws[128];
   memset(ws,0,sizeof(ws));
   strcpy(ws,s1);
   return strcat(ws,s2);
}

void read_ini(void)
{  char a[1024], sname[] = "[MAIN]", pfn[] = "PACS_WARN.INI";
   int sa = sizeof(a);

/*------------------------------------------------------------------------------ debug */
   if (!GetIniString(sname, "Debug", "stdout", a, sa, pfn))
      dbg_exit(strcmb("Debug", GetIniErrorString()), 1);
   debug_fp = debug_on(a);
/*---------------------------------------------------------------------------- Logging */
   if (!GetIniString(sname, "LogDir", log_dir, a, sa, pfn))
      dbg_exit(strcmb("LogDir", GetIniErrorString()),1);
   strcpy(log_dir,a);
   if (!GetIniInt(sname, "LogRetainDays", 30, &log_retain_days, pfn))
      dbg_exit(strcmb("LogRetainDays", GetIniErrorString()),1);
   if (!GetIniInt(sname, "LogPurgeHour", 2, &log_purge_hour, pfn))
      dbg_exit(strcmb("LogPurgeHour", GetIniErrorString()),1);
}

/*----------------------------------------------------------------- Print log messages */   
void ldbg(char * msg) { dbg(msg); prt_log(msg,NULL); }

void prt_log(char *msg, char * buff)
{  time_t now;
   struct tm *tm;
   FILE *fp; 
   char * text;
   char nlb[12] = "\n", nla[64] = " ----------------------------\n";

   if (!msg && !buff) return;
   if (!msg) msg = "\0";
   text = malloc(strlen(msg) + 128);
   
   now = time(NULL);
   tm = localtime(&now);
   
   sprintf(log_pfn, log_pfn_format, log_dir, tm->tm_year + 1900, tm->tm_mon +1, tm->tm_mday);
   fp = fopen(log_pfn, "a");
   if (!fp) 
   {  printf("LOG FILE I/O ERR: %s on %s\n", strerror(errno), log_pfn);
      return;
   }
   if (!buff) { strcpy(nlb,"  "); strcpy(nla," \n"); }
   sprintf(text, "%s%04d/%02d/%02d %02d:%02d:%02d %s%s", nlb, tm->tm_year + 1900, tm->tm_mon +1,
         tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec, msg, nla);
   fprintf(fp,text); 
   free (text);
   
   if (buff)
   {  char p[70], *ptr;
	   int piece = sizeof(p) - 3, cont = 1;
	   while (cont)
	   {	memset(p,0,sizeof(p)); strcpy(p,"> ");
		   if (strlen(buff) < piece) { strcat(p,buff); cont = 0; }
		   else { memcpy(p+2,buff,piece); buff += piece; }
		   if (ptr = strchr(p,SB)) *ptr = '.';
		   if (ptr = strchr(p,EB)) *ptr = '.';
		   while (ptr = strchr(p,'\n')) *ptr = '.';
		   while (ptr = strchr(p,'\r')) *ptr = '~';
		   fprintf(fp,"%s\n",p);
	   }
   }
   fclose(fp);
   return;
}

/*----------------------------------------------------------------------
 * Checks to see if purge has been run yet today.  If not, runs it, and
 * purges log files older than retain days.  This routine is run at a
 * timeout, so will lag during busy times.
 *--------------------------------------------------------------------*/
void purge_logs(void)
{  time_t now;
   struct tm tdy, *lp;
   char pfn[128], cutoff_pfn[128], *p;
   int month, year;
   DIR * d;
   struct dirent * e;
/*------------------------------------- Determine if we need to purge */
   now = time(NULL);
   memcpy(&tdy,localtime(&now),sizeof(tdy));
   lp = localtime(&last_purge_date);
   if (tdy.tm_year == lp->tm_year && tdy.tm_yday == lp->tm_yday) return;
   if (tdy.tm_hour < log_purge_hour) return;
   last_purge_date = now;     /* We are going to purge; Update last purge date */
   now = time(NULL) - (log_retain_days * 86400);
   lp = localtime(&now);
   sprintf(cutoff_pfn, log_pfn_format, "\0", lp->tm_year + 1900, lp->tm_mon +1, lp->tm_mday);
   sprintf(em,"Purging log & err files before %4d/%02d/%02d", lp->tm_year + 1900, lp->tm_mon +1, lp->tm_mday); 
   prt_log(em,NULL);
/*------------------------------ Pass all files, purging those before cutoff */
   if (d=opendir(log_dir))
   {  while (e=readdir(d))
      {  p = e->d_name;
         if (!strstr(p,".LOG")) continue;
         if (strcmp(p, cutoff_pfn) < 0)
         {  sprintf(em,"Purging file %s", p); prt_log(em,NULL);
            strcpy(pfn,log_dir); strcat(pfn,p); remove(pfn); 
         }
      }
      closedir(d);
   }
}

