/***********************************************************************
 * Program sends status messages to specified host based on command line
 * parameters.  Format of command line call is:
 *
 * SMessage distributor_queue receiver_AE message_type target_host target_port
 *
 *    distributor_queue = machine running distributor (bond,goldfinger,fleming,bigben)
 *                        (default is bond), followed by one or two digit queue number
 *                        (default is 0)
 *
 *    receiver_AE = machine whose status was checked (imageweb, etc)
 *
 *    message_type = STATUS_TEST    
 *                   STATUS_PASS    
 *                   STATUS_FAIL
 *
 *    target_host  = host name where message will be sent
 *                   (default is rentgen)
 *
 *    target_port  = port on host where message will be sent
 *                   (default is 45000)
 *
 ************************************************************************
 *
 * Compile and link with: gcc SMessage.c -o SMessage -lsocket -lnsl
 *
 *-----------------------------------------------------------------------
 * 5-8-2 R Moulton - Modified to allow 2 digit queue numbers
 ***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>

/***********************************************************************
 * Valid distributor names, message types
 **********************************************************************/
#define BOND         "bond"
#define GOLDFINGER   "goldfinger"
#define FLEMING      "fleming"
#define BIGBEN       "bigben"
#define RELAY1       "relay1"
#define RELAY2       "relay2"
#define STATUS_TEST  "STATUS_TEST" 
#define STATUS_PASS  "STATUS_PASS" 
#define STATUS_FAIL  "STATUS_FAIL"

/***********************************************************************
 * Command line parameters, preset with default values
 **********************************************************************/
char queue[24]             = "bond0";
char sender_AE[24]         = "bond";
char receiver_AE[24]       = "imageweb";
char message_type[24]      = "STATUS_TEST";
char target_host[24]       = "rentgen";
char target_ip[]           = "128.252.118.40";
char target_port[24]       = "45000";

/***********************************************************************
 * Other global working storage
 **********************************************************************/
char msg[256];
char a[256];
time_t timer;
time_t * ptimer = &timer; 
struct tm * ptime;
int monitor_socket, i, j, retval;
struct hostent *pmonitorent, monitorent;
struct sockaddr_in monitor_sock_name;
fd_set socket_set;
struct timeval timeout;
   
/***********************************************************************
 * Get current time
 **********************************************************************/
get_time_hack()
{  timer = time(ptimer);
   ptime = localtime(ptimer);
}

/***********************************************************************
 * Contatenate current date/time on msg;
 **********************************************************************/
cat_date()
{  char cdate[10]; memset(cdate,0,sizeof(cdate));
   strftime(cdate, 9, "%Y%m%d", ptime);
   strcat(msg, cdate);
}

cat_time()
{  char ctime[10]; memset(ctime,0,sizeof(ctime));
   strftime(ctime, 9, "%H%M%S", ptime);
   strcat(msg, ctime);
}

/***********************************************************************
 * close out socket
 **********************************************************************/
cleanup()
{  retval = shutdown(monitor_socket, 2);
   if(retval != 0) perror("shutdown");
   retval = close(monitor_socket);
   if(retval) perror("close");
}

/***********************************************************************
 * Set up message
 **********************************************************************/
build_message()
{  int i, c; 
/*---------------------------------------------------- Message header */
   memset(msg, 0, sizeof(msg));
   msg[0] = 11;
/*------------------------------------------------ msg=[message_type] */
   strcat(msg, "msg=");
   strcat(msg, message_type);
/*-------------------------------------------------------------- host */
   strcat(msg, "&host=");
   strcat(msg, sender_AE);
/*--------------------------------------------------- sender AE title */
   strcat(msg, "&sender_AE=");
   strcat(msg, sender_AE);
/*------------------------------------------------- receiver AE title */
   strcat(msg, "&receiver_AE=");
   strcat(msg, receiver_AE);
/*------------------------------------------- start or stop time hack */
   get_time_hack();
   strcat(msg, "&startdate=");
   cat_date();
   strcat(msg, "&starttime=");
   cat_time();
/*------------------------------------------------- distributor_queue */
   strcat(msg, "&queue=");
   strcat(msg, queue);
   strcat(msg, "&");
/*--------------------------------------------------- add closing 28  */
   msg[strlen(msg)] = 28;
/*-------------------------------- Uncomment to debug message formatting
 * memset(a,0,sizeof(a)); strncpy(a, msg+1, strlen(msg) -2);
 * strcat(a,"\n"); printf(a); 
 *--------------------------------------------------------------------*/
}

/***********************************************************************
 * Validate command line parameters, loading into values if non-blank
 * return non-zero if validation failed
 **********************************************************************/
int validate_params(int numpars, char * par[])
/*----------------------------------------------------------------------
 * i is parameter number we are looking at.  Numpars is the total number
 * of parameters.  Since the program name is the first parameter in C,
 * we decrement numpars to get the total number of "real" parameters.
 *--------------------------------------------------------------------*/
{  int i = 0; numpars--;
/*----------------------------------------------------------------------
 * Next parameter is distributor queue, matched against valid names
 * This parameter also gives us the sender AE (which is also the host)
 *--------------------------------------------------------------------*/
   i++;
   if(numpars < i) return 0;
   strcpy(queue, par[i]);
   strcpy(sender_AE, par[i]);
   sender_AE[strlen(sender_AE) -1] = 0;
   if(strcmp(sender_AE, BOND) &&
      strcmp(sender_AE, GOLDFINGER) &&
      strcmp(sender_AE, FLEMING) &&
      strcmp(sender_AE, RELAY1) &&
      strcmp(sender_AE, RELAY2) &&
      strcmp(sender_AE, BIGBEN))
   {  sender_AE[strlen(sender_AE) -1] = 0;
      if(strcmp(sender_AE, BOND) &&
         strcmp(sender_AE, GOLDFINGER) &&
         strcmp(sender_AE, FLEMING) &&
         strcmp(sender_AE, RELAY1) &&
         strcmp(sender_AE, RELAY2) &&
	 strcmp(sender_AE, BIGBEN))
         {  printf("invalid sender AE name on command line\n");
            return 1;
         }
   }
/*----------------------------------------------------------------------
 * Next parameter is receiver_AE, not validated
 *--------------------------------------------------------------------*/
   i++;
   if (numpars < i) return 0;
   strcpy(receiver_AE, par[i]);
/*----------------------------------------------------------------------
 * Next parameter is message type code, matched against valid codes
 *--------------------------------------------------------------------*/
   i++;
   if (numpars < i) return 0;
   strcpy(message_type, par[i]);
   if (strcmp(message_type, STATUS_PASS))
      if (strcmp(message_type, STATUS_FAIL))
         if (strcmp(message_type, STATUS_TEST))
            {  printf("invalid message type on command line\n");
               return 1;
            }
/*----------------------------------------------------------------------
 * Next parameter is host name, validated later
 *--------------------------------------------------------------------*/
   i++;
   if(numpars < i) return 0;
   strcpy(target_host, par[i]);
/*----------------------------------------------------------------------
 * Next parameter is host port, which must be numeric
 *--------------------------------------------------------------------*/
   i++;
   if(numpars < i) return 0;
   strcpy(target_port, par[i]);
   return 0;
}

/***********************************************************************
 * main function
 **********************************************************************/
int main(int numpars, char *par[])
{  if(validate_params(numpars, par)) exit(1);
   build_message();
/*----------------------------------------------------------------------
 * get a socket
 *--------------------------------------------------------------------*/
   monitor_socket = socket(AF_INET, SOCK_STREAM, 0);
   if(monitor_socket == -1) {perror("socket"); exit(1);}
/*----------------------------------------------------------------------
 * Set socket to non-blocking.  This is done so that later, on the
 * connect, we can avoid a long wait if the monitor is down.
 *--------------------------------------------------------------------*/
   i = fcntl(monitor_socket, F_GETFL, 0);
   fcntl(monitor_socket, F_SETFL, O_NONBLOCK|i);
/*----------------------------------------------------------------------
 * get host by name
 *--------------------------------------------------------------------*/
   pmonitorent = gethostbyname(target_host);
   if(pmonitorent == NULL) pmonitorent = gethostbyname(target_ip);
   if(pmonitorent == NULL) 
   {  perror("gethost");
      cleanup();
      exit(1);
   }
/*----------------------------------------------------------------------
 * load socket address struct
 *--------------------------------------------------------------------*/
   monitorent = *pmonitorent;
   monitor_sock_name.sin_family = monitorent.h_addrtype;
   monitor_sock_name.sin_port = htons(atoi(target_port));
   monitor_sock_name.sin_addr = *((struct in_addr *)monitorent.h_addr);
/*----------------------------------------------------------------------
 * connect to monitor host
 *--------------------------------------------------------------------*/
   i = sizeof(monitor_sock_name);
   retval = connect(monitor_socket, (struct sockaddr *)&monitor_sock_name, i);
   FD_ZERO (&socket_set);
   FD_SET  (monitor_socket, &socket_set);
   timeout.tv_sec = 5;
   timeout.tv_usec = 0;
   j = monitor_socket + 1;
   retval = select(j, NULL, &socket_set, NULL, &timeout);
   if(retval < 1)
   {  perror("select");
      cleanup();
      exit(1);
   }
/*----------------------------------------------------------------------
 * send message
 *--------------------------------------------------------------------*/
   i = strlen(msg);
   retval = send(monitor_socket, msg, i, 0);
   if(retval == -1) perror("send");
   cleanup();
}
