/****************************************************************************************
 * configure.c                                                                  R Moulton
 * This takes over the function of c.csh, which loads the distributor configuration file
 * Distributor.cfg into the various configuration files for the distributor.  It is 
 * designed to be run in the $BIN directory. 
 * Compile: gcc -g configure.c -o configure `pkg-config --cflags --libs configure`
 * --------------------------------------------------------------------------------------
 * 08-16-05 R Moulton - Initial write
 ***************************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

#include "dbg.h"
#include "ini.h"
#include "log.h"

/*---------------------------- function prototypes ------------------------------------*/
int read_line(void);
void write_line(char * buf, int queue);
void all_line(void);
void params(void);
int server_line(void);

/*---------------------------- macros -------------------------------------------------*/
#define FREE(A) {free(A);A=NULL;}
#define STR_EQ(A,B) !strcasecmp(A,B)
#define STR(A,B) {free(A);A=strdup(B);}
#define LCHR(A) A[strlen(A) - 1]
#define CLEAR(A) memset(A,0,sizeof(A))

#define EOFSTRING "end-of-file"

static FILE * Distributor_cfg;
static char * Distributor_cfg_pfn = "Distributor.cfg";
#define MAX_FILE 99
FILE * Queue_file[MAX_FILE];
static char * sname;
static char * csname;

static char * def_sname;
static char def_queue;
#define MAX_DESTS 5
static char * def_dest[MAX_DESTS];
static int line;

#define FS 48
static char source[FS];
static char modality[FS];
static char desc[FS];
static char server[FS];
static int queue;
static char dest[MAX_DESTS][FS];

#define BS 512
static char buf[BS];
static char iline[BS];

static time_t now;
static char flag = 0;
FILE * fp;
static char pfn[256];


int main(int argc, char *argv[])
{  char * p;
   int i;

   now = time(NULL);

   debug_on("stdout");
   
   printf("Processing server configuration for ");
   
   if(p=getenv("SNAME")) {STR(sname,p) printf("%s\n",sname); }
   else dbg_exit("\nenv var 'sname' not defined",1);
   
   if (p=getenv("CSNAME")) STR(csname,p)
   else dbg_exit("env var 'csname' not defined",1);

   if (access(Distributor_cfg_pfn,F_OK) || access(Distributor_cfg_pfn,R_OK))
      dbge_exit("Distributor.cfg:",1);

   Distributor_cfg = fopen(Distributor_cfg_pfn,"r");
   if (!Distributor_cfg) dbge_exit("Distributor.cfg open:",1);

   printf("Creating compressed files for %s...\n", csname);

/****************************************************************************************
 * Distributor.cfg begins with the Routing rules, until we see a line beginning "SetFile"
 ***************************************************************************************/
   STR(def_sname,sname)                           /* defaults for routing rules */
   def_queue = 1;
   for (i=0; i < MAX_DESTS; i++) STR(def_dest[i],"")

   line = 0;
   while (1)
   {  if (read_line()) break;
      line++;
      if (!strncmp(buf,"SetFile,",8)) break;    /* switching to SetFile mode */ 
      printf("%4d: %s\n", line, buf);           /* echo line to stdout */
      
      if (!strlen(buf) || buf[0] == '#') continue;  /* comments and blank lines */

      if (server_line()) continue;
      all_line();

/*------------------------------------ SetDefaults,def server,def queue,def_dest(s)... */
      if (!strncmp(buf,"SetDefaults,",12))
      {  p = strtok(buf,",");
         if(p = strtok(NULL,",")) STR(def_sname,p)     /* default server */
         else STR(def_sname,sname)
         if(p=strtok(NULL,",")) def_queue = atoi(p);   /* default queue */
         else def_queue = 1;
         for (i=0; i < MAX_DESTS; i++)
         {  if (p=strtok(NULL,"<")) STR(def_dest[i],p)
            else STR(def_dest[i],"")
         }
/*----------------------------------------------------------- validate default queue # */
         if (def_queue < 1 || def_queue > 98) 
         {  printf("\nError at line %d: invalid default queue\n", line);
            return (1);
         }
         write_line(iline, def_queue);
         continue;
      }

/*------------------------ Detail lines. break down, update with defaults, re-assemble */
      CLEAR(source); 
      CLEAR(modality);
      CLEAR(desc);
      CLEAR(server);
      queue = 0;
      for (i=0; i < MAX_DESTS; i++) CLEAR(dest[i]);

      i = sscanf(buf,"%[^,],%[^,],%[^,],%[^,],%d,%[^,],%[^,],%[^,],%[^,],%[^, ]", 
         &source, &modality, &desc, &server, &queue,
         &(dest[0]), &(dest[1]), &(dest[2]), &(dest[3]), &(dest[4])); 

      switch (i) 
      {  case EOF:
         case 0:
         case 1:
         case 2:
            printf("*** invalid detail line ***\n");
            continue;
         case 3:
            strcpy(server,def_sname);
         case 4:
            queue = def_queue;
         case 5:
            strcpy(dest[0],def_dest[0]);
         case 6:
            strcpy(dest[1],def_dest[1]);
         case 7:
            strcpy(dest[2],def_dest[2]);
         case 8:
            strcpy(dest[3],def_dest[3]);
         case 9:
            strcpy(dest[4],def_dest[4]);
         default:
            break;
      }

      sprintf(iline,"%s,%s,%s,%s,%d,%s,%s,%s,%s,%s", 
          source, modality, desc, server, queue,
          dest[0], dest[1], dest[2], dest[3], dest[4]); 
      while (LCHR(iline) == ',') LCHR(iline) = '\0';
            
      write_line(iline, queue);
      continue;
   }

/*---------------------------------------------------------------------------------------
 * Done with routing files.  output end-of-file and close those which are open.
 *-------------------------------------------------------------------------------------*/
   for (i=0; i < MAX_FILE; i++)
   {  if (Queue_file[i])
      {  if (fputs("end-of-file\n",Queue_file[i]) == EOF) dbge_exit("eof write", 1);
         if(fclose(Queue_file[i])) dbge_exit("eof write", 1);
      }
   }
   printf("Dealer/Router Files created successfully\n");

   if (STR_EQ(buf,EOFSTRING)) return (0);

/****************************************************************************************
 * Now we enter the SetFile mode, where we are creating various configuration files.
 ***************************************************************************************/

   printf("\nEntering SetFile mode...\n\n");   
      
   while (1)
   {  if (flag) 
         if (read_line()) break;
         else line++;
      flag = 1;

      printf("%4d: %s\n", line, buf);

      if (buf[0] == '#') continue;
      if (server_line()) continue;
      while(strlen(buf) && (LCHR(buf) == ' ')) LCHR(buf) = '\0';

      params();

      if (!strncmp(buf,"SetFile,",8))      /* new file */
      {  strcpy(pfn,buf+8);
         if (!strlen(pfn)) strcpy(pfn,"CONFIG.ERRS");
         if (fp) fclose(fp);
         fp = fopen(pfn,"w");
         if (!fp) dbge_exit(pfn,1);
         continue;
      }


      strcat(buf,"\n");
      if (fputs(buf,fp) == EOF) dbge_exit("write:", 1);
      continue;
   }

   if (fp) fclose(fp);
   printf("Distributor.cfg loaded successfully\n");
   return (0);
}
      
      
/*---------------------------------------------------------------------------------------
 * reads line from Distributor.cfg, aborts if error, returns 0=OK, 1=EOF.
 * If eof, returns "end-of-file" in buffer, treats that line as EOF. 
 * strips off excess spaces and new-lines at eol
 *-------------------------------------------------------------------------------------*/
int read_line(void)
{  if (!fgets(buf,BS,Distributor_cfg))
      if (ferror(Distributor_cfg)) dbge_exit("Distributor.cfg read:", 1);
      else strcpy(buf,EOFSTRING);
   while(strlen(buf) && (LCHR(buf) == ' ' || LCHR(buf) == '\n')) LCHR(buf) = '\0';
   if (STR_EQ(buf,EOFSTRING)) return (1);
   return (0);
}

/*---------------------------------------------------------------------------------------
 * write buf to queue file and server file. Abort on any error.  Open files if not open
 *-------------------------------------------------------------------------------------*/
void write_line(char * buf, int queue)
{  char pfn[24], a[BS + 1];
/*------------------------------------------------------------------- server file open */
   if (!Queue_file[0])
   {  sprintf(pfn,"%s.cfg", csname);
      Queue_file[0] = fopen(pfn,"w");
      if (!Queue_file[0]) dbge_exit("server .cfg file open:", 1);
      sprintf(a,"#--- %s -- %s", pfn, ctime(&now));
      if (fputs(a,Queue_file[0])== EOF) dbge_exit("server .cfg file write:", 1);
   }
/*-------------------------------------------------------------------- queue file open */
   if (!Queue_file[queue])
   {  sprintf(pfn,"%s%d.cfg", csname, queue);
      Queue_file[queue] = fopen(pfn,"w");
      if (!Queue_file[queue]) dbge_exit("queue .cfg file open:", 1);
      sprintf(a,"#--- %s -- %s", pfn, ctime(&now));
      if (fputs(a,Queue_file[queue])== EOF) dbge_exit("queue .cfg file write:", 1);
   }
/*--------------------------------------------------------------------- Actual writes */
   sprintf(a,"%s\n",buf);
   if (fputs(a,Queue_file[0])== EOF)     dbge_exit("server .cfg file write:", 1);
   if (fputs(a,Queue_file[queue])== EOF) dbge_exit("queue .cfg file write:", 1);
}

/*---------------------------- Look for ",all," and replace with ",sname,", then copy */
void all_line(void)
{  char * q;
   q = strstr(buf,",all,");
   if (!q) { strcpy(iline,buf); return; }
   CLEAR(iline);
   strncpy(iline,buf,(q - buf) + 1);
   strcat(iline,sname);
   strcat(iline,q+4);
   strcpy(buf,iline);
   return;
}

/*---------------------------------------------------------------------------------------
 * Substitute environment parameters of form "${VAR}".  Print error messages, but do not
 * abort 
 *-------------------------------------------------------------------------------------*/
void params(void)
{  char a[BS], b[BS], e[BS], v[BS], *p, *q;
   strcpy(a,buf);
   while ((p=strstr(a,"${")) && (q=strstr(p,"}")))
   {  memset(b,0,BS);
      memset(e,0,BS);
      memset(v,0,BS);
      strncpy(b,a,p-a);
      strcpy(e,q+1);
      strcpy(v,strtok(p+2,"}"));
      if (p=getenv(v)) strcpy(v,p);
      else printf("*** no such variable: %s\n", v);
      strcpy(a,b);
      strcat(a,v);
      strcat(a,e);
   }
   strcpy(buf,a);
   return;
}

/*---------------------------------------------------------------------------------------
 * Lines for particular servers.  Look for <server1,server2...> indicating lines only for 
 * those servers, or <-server1,server2...> on lines which are to be omitted on those 
 * servers.  <...> construct is removed before writing line to file.  Return 0 if line
 * is OK for this server, 1 if not.
 *-------------------------------------------------------------------------------------*/
int server_line(void)
{  char a[BS];
   char * p, * q;
   char minus = 0;
   strcpy(a,buf);
   p = strchr(a,'<');
   q = strchr(a,'>');
   if (!p || !q || (p > q)) return (0);   /* no <...> construct in line */
   
   CLEAR(buf);
   strncpy(buf,a,p-a);             /* remove <...> construct from buf */
   strcat(buf,q+1);

   *(q+1) = '\0';                  /* p is now <...> construct */
   if (*(p+1) == '-') minus = 1;   /* minus if we are working by omission */
   q = strstr(p,sname);            /* non-NULL if server is in construct */
   if ((q && !minus) || (!q && minus)) return (0);
   return (1);
}
