/******************************************************************************
 * hl7query.c                                                         R Moulton
 * Library routines which process a variety of HL7 queries to the RAD.
 *-----------------------------------------------------------------------------
 * 8-4 R Moulton Initial write.
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <time.h>
#include <arpa/inet.h>
#include <errno.h>
#include <netdb.h>

#include "hl7.h"
#include "hl7query.h"
#include "dbg.h"

/*-------------------------------- Macros --------------------------------------*/
#define MAKE_STR(A,B) {if (A=malloc(strlen(B)+3)) strcpy(A,B);}
#define FREE(A) if (A) { free(A); A=NULL; }
#define STR_EQ(A,B) !strcasecmp(A,B)
#define CLEAR(A) memset(A,0,sizeof(A))
#define SB 11
#define EB 28

/*************************** Function Prototypes ********************************/
static char * time_stamp(time_t time, unsigned char include_seconds);
static void cleanup(void);

/*-------------------------------- Error msg strings. Match to ini_errno values */
static char * e_msg[] = 
{	"No error",
   "No message",
   "Invalid server #",
   "Exceeded wait limit",
   "Can't create socket",
   "Can't id RAD host",
   "Can't connect",
   "Send select error",
   "Send error",
   "Receive select error",
   "Receive error",
   "Buffer too small",
   "HL7 Parsing error",
   "RAD AE'd query",
   "Memory allocation error",
   "Invalid MPI #",
   "Exam is CANCELLED",
};
static int max_e_msg = sizeof(e_msg) / sizeof(char *);
static char * no_msg = "Unknown hl7q_errno value";


#define TRUE   1
#define FALSE  0
#define HL7Q_R_OK   0
#define HL7Q_R_FAIL 1

/*-------------------------- Query server table -------------------------------
 * Allows selection of query link to use (Def PACSI) and related parameters
 *---------------------------------------------------------------------------*/
#define RAD_HOST "rentgn"
static char rad_host[24] = RAD_HOST;
#define RAD_IP   "10.33.162.75"
static char rad_ip[24] = RAD_IP;
#define DEFAULT_SERVER 0
#define DEFAULT_WAIT -1
#define DEFAULT_PERSIST 0
#define WAIT (wait_limit >= 0)
struct _query_server_ {
   char * name;
   int  port;
};
const static struct _query_server_ servers[] = {
   { "PACSI",    10001 },
   { "BQSTCP00",  6680 },
   { "BQSTCP01",  6681 },
   { "BQSTCP02",  6682 },
   { "BQSTCP03",  6683 },
   { "BQSTCP04",  6699 },
};
static unsigned char max_server = sizeof(servers) / sizeof(struct _query_server_);

static int port = 0;
static int wait_limit = -1; /* Minutes to wait for connect/msg prs (neg = forever) */
static char on_idx_server;                   /* Are we on the IDXRad server? */ 
static char keep_socket;           /* Should we hold socket between queries? */
static int sckt;
static char srvr;
static char include_canceled_exams = 1;  /* Should cancelled exams be included in results? */

/*-----------------------------------------------------------------------------
 * For Patient Query: Should exams be sorted by accn?  And how many consequtive
 * exams before the cutoff date should stop query.
 *---------------------------------------------------------------------------*/
#define SORT_ACCESSIONS 0
#define NUM_EXAMS_CUTOFF 3

/*********************** Select host to connect to (if NULL, rentgn is used) */
int hl7SelectHost(char * host, char * ip)
{  if (host) strcpy(rad_host,host);
   else strcpy(rad_host,RAD_HOST);
   if (ip) strcpy(rad_ip,ip);
   else strcpy(rad_ip,RAD_IP);
   return (HL7Q_R_OK);
}
/********************************** Select query server (link) to connect to */
const char * hl7QSelectServer(char s, int wait, char persist)
{  char a[128], *p = NULL;
/*---------------------------------- Find out if we are on the IDXrad server */
   if (!port)
   {  on_idx_server = 0;
      if (!gethostname(a,sizeof(a)) && (p=strtok(a,"."))&& STR_EQ(p,rad_host)) 
         on_idx_server = 1; 
   }
   hl7q_errno = HL7Q_OK;
   if (s < 0 || s >= max_server) { hl7q_errno = HL7Q_INVSRV; return (NULL); }
   if (port != servers[s].port && sckt) cleanup();
   port = servers[s].port;
   wait_limit = wait;
   keep_socket = persist;
   srvr = s;
   return (servers[s].name);
}

int hl7IncludeCanceledExams(char boolean)
{  include_canceled_exams = 0;
   if (boolean) include_canceled_exams = 1;
   hl7q_errno = HL7Q_OK;
   return HL7Q_R_OK;
}

static HL7MSG hl7msg;

/*---------------------- Standard MSH segment to begin query (Fill in stamp) */
#define MSH "MSH|^~\\&|HL7QRY||RAD QRY||%s||QRY^Q01|1.10|P|2.1|||||\r"

/******************************************************************************
 * Plain text query, returns actual response.  If no msg, connects to server.
 * This is the base routine which actually processes the query to the RAD; 
 * The other routines call this one.
 *****************************************************************************/
int hl7MsgQuery(char * msg, char * reply, int size)
{  static struct hostent *psrvent, srvent;
   static struct sockaddr_in srvname; 
   static fd_set sock, esock;
   static struct timeval timeout = {60, 0};
   static char buffer[1024 * 20];
   time_t start;
   int i, j, s, b = sizeof(buffer);
   char rts = 0, rtr = 0, *p = NULL, *q = NULL;

   hl7q_query_failed = 1;
   
   if (msg) 
   {  rts = 1;                           /* If we have a message, we are rts */
/*----------------------------- If msg does not have wrapper bytes, add them */
      if (msg[0] != SB) { memmove(msg+1,msg,strlen(msg)); msg[0]=SB;}
      if (msg[strlen(msg)-1] != EB) msg[strlen(msg)]=EB;
   }
      
   if (WAIT) start = time(NULL);       /* If we have time limit, start clock */
   
/*-------------------------- If user didn't select a query link, use default */
   if (!port) hl7QSelectServer(DEFAULT_SERVER, DEFAULT_WAIT, DEFAULT_PERSIST); 

   while (1)
   {  if (WAIT)                                          /* Check wait limit */
      {  if ((int)((time(NULL)-start)/60) > wait_limit) 
         {  hl7q_errno = HL7Q_WAITLIM;
            if (!keep_socket) cleanup();
            return (HL7Q_R_FAIL);
         }
      }
      if (!sckt)
      {  dbg("Making Socket");
         sckt = socket(AF_INET, SOCK_STREAM, 0);
         if (sckt == -1)
         {  dbge ("socket:");
            hl7q_errno = HL7Q_SOCKET;
            sckt = 0;
            return (HL7Q_R_FAIL);
         }
         
         dbg("Setting to non-blocking");
         i = fcntl(sckt, F_GETFL, 0);
         fcntl(sckt, F_SETFL, O_NONBLOCK|i);
         
         dbg("Getting host data");
         psrvent = gethostbyname(rad_host);
         if (!psrvent) psrvent = gethostbyname(rad_ip);
         if (!psrvent)
         {  dbge ("gethostbyname");
            hl7q_errno = HL7Q_GETHOST;
            cleanup();
            return (HL7Q_R_FAIL);
         }
         
         dbg("Setting socket name entries");
         srvent = *psrvent;
         srvname.sin_family = srvent.h_addrtype;
         srvname.sin_port = htons(port);
         srvname.sin_addr = *((struct in_addr *)srvent.h_addr);
         
         dbg("Connect to RAD server");
         s = sizeof(srvname);
         i = connect(sckt, (struct sockaddr *)&srvname, s);
         if (i < 0 && errno != EINPROGRESS) /* We expect the inprogress error*/
         {  dbge("connect");
            hl7q_errno = HL7Q_CONNECT;
            cleanup();
            return (HL7Q_R_FAIL);
         }
      }
      if (rts)                                       /* Ready to send */
      {  dbg("Select to write");
         FD_ZERO(&sock); FD_SET(sckt, &sock);
         FD_ZERO(&esock); FD_SET(sckt, &esock);
         i = sckt + 1;
         j = select(i, NULL, &sock, &esock, &timeout);
         if (j == 0) { dbg("Timed out"); continue; } /* Timed out; loop */
         if (j < 0)
         {  dbge("send select");
            hl7q_errno = HL7Q_SSELECT;
            cleanup();
            return (HL7Q_R_FAIL);
         }
         if (FD_ISSET(sckt, &esock))
         {  dbg("dropped");
            cleanup();
            continue;
         }
         dbg("send msg");
         i = send(sckt, msg, strlen(msg), 0);
         if (i < 0)
         {  dbge("send");
            hl7q_errno = HL7Q_SEND;
            cleanup();
            return (HL7Q_R_FAIL);
         }

         rts = 0;
         rtr = 1;
         memset(reply,0,size);
      }
      if (rtr)                                        /* Ready to receive */
      {  dbg("Select to read");
         FD_ZERO(&sock);  FD_SET(sckt, &sock);
         FD_ZERO(&esock); FD_SET(sckt, &esock);
         i = sckt + 1;
         j = select(i, &sock, NULL, &esock, &timeout);
         if (j < 0)
         {  dbge("recv select");
            hl7q_errno = HL7Q_RSELECT;
            cleanup();
            return (HL7Q_R_FAIL);
         }
         if (FD_ISSET(sckt, &esock))
         {  dbg("dropped");
            cleanup();
            rts = 1; rtr = 0;
            continue;
         }
         dbg("recv msg");
         memset(buffer,0,b);
         i = recv(sckt, buffer, b, 0);
         if (i < 0)
         {  dbge("recv");
            hl7q_errno = HL7Q_RECV;
            cleanup();
            return (HL7Q_R_FAIL);
         }
         if (i == 0)
         {  dbg("dropped");
            cleanup();
            rts = 1, rtr = 0;
            continue;
         }
         if (p = strchr(buffer,SB))
         {  memset(reply,0,size);
            if (strlen(p) > (size -1))
            {  hl7q_errno = HL7Q_SIZE;
               dbg("buffer overflow");
               return (HL7Q_R_FAIL);
            }      
            strcpy(reply,p);
         }
         else if (strlen(reply))
         {  if ((strlen(reply) + strlen(buffer) > (size -1)))
            {  hl7q_errno = HL7Q_SIZE;
               dbg("buffer overflow");
               return (HL7Q_R_FAIL);
            }      
            strcat(reply,buffer);
         }
/*-------------------------- We're done if there is an SB and a following EB */
         p = strchr(reply,SB);
         q = strchr(reply,EB);
         if (p && q && p < q) 
         {  *(q+1) = 0; 
            hl7q_query_failed = 0; 
            break; 
         }
         continue;
      }
   }
   if (!keep_socket) cleanup();
   hl7q_errno = HL7Q_OK;
   return (HL7Q_R_OK);
}

/***********************************************************************
 * close out socket
 **********************************************************************/
void cleanup(void)
{  if (sckt)
   {  shutdown(sckt, 2);
      close(sckt);
      sckt = 0;
   }
}

/******************************************************************************
 * Struct based query on accession number.  Fills in and returns remainder of
 * items in the struct.  
 *****************************************************************************/

#define AQ_FORMAT "QRD|%s|R|I|HL7Q|||150^RD||RES|%7d||\r"

int hl7AccnQuery(struct _hl7_accn_query_ * query)
{  char msg[1024], reply[1024 * 20], segid[5];
   time_t now = time(NULL);
   int i, segnum;
   sprintf(em,"Accn query: %d", query->accn); dbg(em);
   hl7q_errno = HL7Q_OK;
   
/*---------------------------------------------------- Clear query structure */   
   query->mpi = 0;
   FREE(query->visit)
   FREE(query->dnumber)
   FREE(query->pname)
   FREE(query->alias)
   FREE(query->dob)
   FREE(query->sex)
   FREE(query->ssn)
   FREE(query->exam)
   FREE(query->exam_desc)
   FREE(query->status)
   FREE(query->start_time)
   FREE(query->time)
   FREE(query->r_physician)
   if (query->accn < 0) { query->accn = 0; return (HL7Q_R_OK); }
   
/*-------------------------------------------------------------------- Query */ 
   memset(msg,0,sizeof(msg));  
   sprintf(msg, MSH AQ_FORMAT, time_stamp(now, 0), time_stamp(now, 1), query->accn);
   if(hl7MsgQuery(msg, reply, sizeof(reply)))  
   {  hl7q_errno = HL7Q_NAK;
      return (HL7Q_R_FAIL);
   }
/*-------------------------------------------------------------- Parse reply */
   if (hl7ReadMsg(&hl7msg, reply, strlen(reply)))
   {  PrtHl7ErrorString(debug_is_on(), "hl7ReadMsg");
      hl7q_errno = HL7Q_HL7;
      return (HL7Q_R_FAIL);
   }
/*-------------------- If not including canceled exams, check for that first */
   if (!include_canceled_exams)
   {  if (hl7FindSeg(&hl7msg, "ZOS", &segnum) == HL7Q_R_OK)
         if (STR_EQ(hl7GetFld(&hl7msg,1),"X"))
         {  hl7q_errno = HL7Q_CANCELED;
            return (HL7Q_R_FAIL);
         }
   }
/*--------------------------------- Scan segments, pulling out relevant data */
   for (segnum = 0; segnum < hl7NumSegs(&hl7msg); segnum++)
   {  hl7SegNum(&hl7msg, segid, &segnum);

/*-------------------------------------------------------------- MSA segment */
      if (STR_EQ(segid,"MSA"))
/*------------------------------------------------------- RAD reply not "AA" */
      {  if (strcasecmp(hl7GetFld(&hl7msg,1),"AA"))
         {  hl7q_errno = HL7Q_NAK;
            sprintf(em, "Query failed: %s", hl7GetFld(&hl7msg,3)); dbg(em);
            return (HL7Q_R_FAIL);
         }
      }

/*-------------------------------------------------------------- PID segment */
      if (STR_EQ(segid,"PID"))
      {  query->mpi = atoi(hl7GetFld(&hl7msg,3));
         MAKE_STR(query->dnumber,hl7GetFld(&hl7msg,4))
         MAKE_STR(query->pname,hl7GetFld(&hl7msg,5))
         MAKE_STR(query->dob,hl7GetFld(&hl7msg,7))
         MAKE_STR(query->sex,hl7GetFld(&hl7msg,8))
         MAKE_STR(query->alias,hl7GetFld(&hl7msg,9))
         MAKE_STR(query->visit,hl7GetFld(&hl7msg,18))
         MAKE_STR(query->ssn,hl7GetFld(&hl7msg,19))
      }

/*-------------------------------------------------------------- RAD segment */
      if (STR_EQ(segid,"RAD"))
      {  char * t=NULL, *p = NULL;
         MAKE_STR(query->exam,hl7GetFld(&hl7msg,1))
         MAKE_STR(query->exam_desc,hl7GetFld(&hl7msg,2))
/*----------------------------------------------- Referring physician for order
 * This is of form: PIN^LAST^FIRST... so if it is there, we look for the first
 * "^" and point just beyond that, transferring only the name
 *---------------------------------------------------------------------------*/
         MAKE_STR(t,hl7GetFld(&hl7msg,8))
         if (t)
         {  if (p=strchr(t,'^')) p++;
            else p=t;
            MAKE_STR(query->r_physician,p)
            FREE(t)
         }
      }

/*-------------------------------------------------------------- ORH segment */
      if (STR_EQ(segid,"ORH"))
      {  MAKE_STR(query->start_time,hl7GetFld(&hl7msg,11))
      }
/*-------------------------------------------------------------- ZOS segment */
      if (STR_EQ(segid,"ZOS"))
      {  MAKE_STR(query->status,hl7GetFld(&hl7msg,1))
         MAKE_STR(query->time,hl7GetFld(&hl7msg,2))
      }
   }
   return (HL7Q_R_OK);
}

/******************************************************************************
 * Struct based query on MPI number.  Computes number of exams and allocates
 * memory for exam struct.  Releases any memory assigned to struct on entry; 
 * If MPI number is negative, returns without actual query.  NOTE: This query
 * must be sent to a Barnes query link (not PACSLINKI) in order to return the
 * facility code
 *****************************************************************************/

#define MPI_FORMAT "QRD|%s|D|I|IWCD|||150^LI|%07d|RES|LIST||\r"

int hl7MPIQuery(struct _hl7_mpi_query_ * query)
{  char msg[256], reply[1024 * 40], segid[5], *p = NULL;
   time_t now = time(NULL);
   int i, j, segnum, accn;
   struct _hl7_mpi_query_exam_ * x;
   sprintf(em,"MPI query: %d", query->mpi); dbg(em);
   hl7q_errno = HL7Q_OK;
/*---------------------------- Here we clear the struct, including any exams */
   if (query->exam)
   {  for (x=query->exam; x->code; x++)
      {  FREE(x->code)
         FREE(x->desc)
         FREE(x->facility)
         FREE(x->status)
         FREE(x->date_time)
      }
      FREE(query->exam)
   }
   query->num_exams = 0;
   if (query->mpi < 0) return (HL7Q_R_OK);

/*-------------------------------------------------------------------- Query */
   memset(msg,0,sizeof(msg));
   sprintf(msg, MSH MPI_FORMAT, time_stamp(now, 0), time_stamp(now, 1), query->mpi);
   if (hl7MsgQuery(msg, reply, sizeof(reply))) 
   {  hl7q_errno = HL7Q_NAK;
      return (HL7Q_R_FAIL);
   }
/*-------------------------------------------------------------- Parse reply */
   if (hl7ReadMsg(&hl7msg, reply, strlen(reply)))
   {  PrtHl7ErrorString(debug_is_on(), "hl7ReadMsg");
      hl7q_errno = HL7Q_HL7;
      return (HL7Q_R_FAIL);
   }

   printf("check for AE\n");
/*---------------------------------------------------- Make sure reply is AA */
   if (hl7FindSeg(&hl7msg, "MSA", &segnum))
   {  PrtHl7ErrorString(debug_is_on(), "FindSeg MSA");
      hl7q_errno = HL7Q_NAK;
      return (HL7Q_R_FAIL);
   }
   if (strcasecmp(hl7GetFld(&hl7msg,1),"AA"))
   {  hl7q_errno = HL7Q_NAK;
      sprintf(em, "Query failed: %s", hl7GetFld(&hl7msg,3)); dbg(em);
      return (HL7Q_R_FAIL);
   }
   printf("count exams\n");
/*------------------------------------------------------- Count number of exams
 * We also look here for invalid mpi #
 * We do not include exams with accn's less than seven digits.
 *---------------------------------------------------------------------------*/
   for (segnum=0, j=0; segnum < hl7NumSegs(&hl7msg); segnum++)
   {  hl7SegNum(&hl7msg, segid, &segnum);
      if (STR_EQ(segid,"DSP") && STR_EQ(hl7GetFld(&hl7msg,2),"1")) 
      {  if (strstr(hl7GetFld(&hl7msg,3),"Unknown patient #"))
         {  hl7q_errno = HL7Q_MPI; 
            return (HL7Q_R_FAIL);
         }
/*---------------------------- We ignore exams with accn less than 7 digits */
         accn = atoi(hl7GetSub(&hl7msg, 5, 1, 1, 1));
         if(accn < 1000000) continue;   
/*------------------------------ We ignore exams with status "X" (cancelled) */
         if (STR_EQ(hl7GetComp(&hl7msg, 5, 1, 4),"X")) continue;
         else j++;
      }
   }

/*----------------------------------------------------------------------------- 
 * Allocate memory for exam table in struct.  We allocate 1 extra slot, so user
 * can use the num_exams or look for a null pointer for EOL
 *---------------------------------------------------------------------------*/
   query->num_exams = j;
   query->exam = calloc(j+1, sizeof(struct _hl7_mpi_query_exam_));
   if (!query->exam) { hl7q_errno = HL7Q_MEM; return (HL7Q_R_FAIL); }
   
/*--------------------------- Now pass again, loading information from exams */
   for (segnum=0, x=query->exam; segnum < hl7NumSegs(&hl7msg); segnum++)
   {  hl7SegNum(&hl7msg, segid, &segnum);
      if (STR_EQ(segid,"DSP") && STR_EQ(hl7GetFld(&hl7msg,2),"1"))
      {  accn = atoi(hl7GetSub(&hl7msg, 5, 1, 1, 1));
         if (accn < 1000000) continue;
         if (STR_EQ(hl7GetComp(&hl7msg, 5, 1, 4),"X")) continue;
         x->accn = accn;
         MAKE_STR(x->code,hl7GetSub(&hl7msg, 5, 1, 3, 1))
/*-------------------------------- Exam description is cols 22-39 in display */
         p = hl7GetFld(&hl7msg, 3);
         if (p)
         {  memset(msg,0,sizeof(msg));
               memcpy(msg,p+21,18);
            while (msg[strlen(msg)-1] == ' ') msg[strlen(msg)-1] = '\0';
            MAKE_STR(x->desc,msg)
/*-----------------------------------------------------------------------------
 * Exam date & time are in cols 40-62: "06/01/2005     10:01 AM"
 * format into yyyymmddhhmm00
 *---------------------------------------------------------------------------*/
            {  int r, mo = 0, da = 0, yr = 0, hr = 0, mi = 0;
               char am[10] = "";
               r = sscanf(p+39, "%d/%d/%d %d:%d %s", &mo, &da, &yr, &hr, &mi, am);
               if (!strcasecmp(am,"PM") && hr < 12) hr += 12;
               x->date_time = malloc(17);
               memset(x->date_time,0,17);
               sprintf(x->date_time,"%04d%02d%02d%02d%02d00", yr, mo, da, hr, mi);
            }
         }
         MAKE_STR(x->facility,hl7GetComp(&hl7msg, 5, 1, 5))
         MAKE_STR(x->status,hl7GetComp(&hl7msg, 5, 1, 4))
         x++;
      }
   }
   return (HL7Q_R_OK);
}

/******************************************************************************
 * Combination query, pulls all exams for given MPI number. Returns the MPI
 * query data in the first parameter, and an array of accn query structs in the 
 * 2nd parameter.  The mpi struct should already be allocated, but the accn
 * struct is allocated by the routine.  The structs must be kept paired in
 * calls, as the function uses information in the mpi struct to clear the accn
 * struct.
 *****************************************************************************/
int hl7PatientQuery(struct _hl7_mpi_query_ * mpi_query_data, 
                    struct _hl7_accn_query_ ** aq_data, char * cutoff_date)
{  int ret, i, j, flag, num_exams_cutoff = 0;
   struct _hl7_accn_query_ * accn_query_data = *aq_data;
   char cutoff[15] = "00000000";
/*----------------------------------------------- Items held to reset at end */
   char hold_srvr = srvr;
   char hold_keep_socket = keep_socket;
   keep_socket = 1;

   if (cutoff_date && (strlen(cutoff_date) < 15)) strcpy(cutoff,cutoff_date);

/*------- Set server to other than PACSI so mpi query can get facility codes */
   if (STR_EQ(servers[srvr].name,"PACSI")) hl7QSelectServer(srvr+1, -1, 1);

/*----------------------------------------------- clear the accn matrix data */
   for (i = 0; i < mpi_query_data->num_exams; i++) 
   {  accn_query_data[i].accn = -1;
      hl7AccnQuery(&accn_query_data[i]);
   }
   FREE(accn_query_data)
/*---------------------------------------------------- Do the MPI query call */
   ret = hl7MPIQuery(mpi_query_data);
   if (mpi_query_data->mpi < 0) goto eopq;  /* just a clear the memory call */
   if (ret == HL7Q_R_FAIL) goto eopq;       /* mpi query failed */
   if (!mpi_query_data->num_exams) goto eopq; /* no exams for this mpi */

/*------------------------------- sort the exams in descending order of accn */
   if (SORT_ACCESSIONS)
   {  do
      {  int sa = sizeof(struct _hl7_mpi_query_exam_);
         char a[sa];
         flag = 0;
         for (i=0; (i+1) < mpi_query_data->num_exams; i++)
         {  if (mpi_query_data->exam[i].accn < mpi_query_data->exam[i+1].accn)
            {  memmove(a,&(mpi_query_data->exam[i]),sa);
               memmove(&(mpi_query_data->exam[i]),&(mpi_query_data->exam[i+1]),sa);
               memmove(&(mpi_query_data->exam[i+1]),a,sa);
               flag = 1;
            }
         }
      } while (flag);
   }
/*------------------------------------- Set server to PACSI for accn queries */
   hl7QSelectServer(0, -1, 1);

/*-----------------------------------------------------------------------------
 * Allocate and clear memory for exam data.  We allocate one extra struct, so
 * that the user can use num_exams or look for null ptrs to find the end.
 *---------------------------------------------------------------------------*/
   i = (mpi_query_data->num_exams +1) * sizeof(struct _hl7_accn_query_);
   accn_query_data = malloc(i);
   if (!accn_query_data) {hl7q_errno=HL7Q_MEM; ret=HL7Q_R_FAIL; goto eopq;}
   memset(accn_query_data,0,i);

/*---------------------------------------------- Query for each of the exams */
   for (i=0; i < mpi_query_data->num_exams; i++)
   {  accn_query_data[i].accn = mpi_query_data->exam[i].accn;
      ret = hl7AccnQuery(&accn_query_data[i]);
      if (ret == HL7Q_R_FAIL) goto eopq;
/*------------------------------------------- Check for exam before cutoff date
 * if cutoff is found, we clear the exam we just retrieved, and clear the mpi
 * entries also, setting the number of exams to the new count. 
 *---------------------------------------------------------------------------*/
      if (strncmp(cutoff,accn_query_data[i].start_time,8) > 0)
      {  accn_query_data[i].accn = -1;
         hl7AccnQuery(&accn_query_data[i]);
         accn_query_data[i].accn = 0;
         for (j=i; j < mpi_query_data->num_exams; j++)
         {  FREE(mpi_query_data->exam[j].code)
            FREE(mpi_query_data->exam[j].desc)
            FREE(mpi_query_data->exam[j].facility)
            FREE(mpi_query_data->exam[j].status)
         }
         mpi_query_data->num_exams = i;
         goto eopq;
      }
   }
/*---------------------------------------------------- successful completion */
   hl7q_errno = HL7Q_OK;
   ret = HL7Q_R_OK;

/*----------------- end of routine catch goto, restore server params and end */
eopq:
   i = hl7q_errno;
   hl7QSelectServer(hold_srvr, wait_limit, hold_keep_socket);
   hl7q_errno = i;
   *aq_data = accn_query_data;
   return (ret);
}

/******************************************************************************
 * return error message string for ini_errno, or print to stdout
 *****************************************************************************/
char * GetHl7QErrorString(void)
{	if (hl7q_errno < max_e_msg) return e_msg[hl7q_errno];
	else return no_msg;
}
void PrtHl7QErrorString(FILE * file, char * prefix)
{	if (!file) file = stdout;
   fprintf(file, "%s: %s\n", prefix, GetHl7QErrorString());
	return;
}

char * time_stamp(time_t tim, unsigned char include_seconds)
{  static char tstamp[15];
   time_t now;
   char format[16] = "%Y%m%d%H%M";
   if (include_seconds) strcat(format,"%S");
   if (tim) now = tim;
   else now = time(NULL);
   strftime(tstamp, sizeof(tstamp), format, localtime(&now));
   return (tstamp);
}
