#ifndef _HL7Q_H_
#define _HL7Q_H_

/******************************************************************************
 * hl7query.h                                                         R Moulton
 * Header file for hl7 query routines, which facilitate processing HL7 queries
 * against the RAD.
 *-----------------------------------------------------------------------------
 * 8-4 R Moulton Initial write.
 *****************************************************************************/

/*-------------------------------------------------------------- error codes */
unsigned int hl7q_errno;
/*-----------------------------------------------------------------------------
 * This flag indicates if the actual query to the RAD worked, i.e. a message
 * was sent and a reply was received.  It is used to test if the query link is
 * up or down. 0=query worked, 1 = query failed 
 *---------------------------------------------------------------------------*/
unsigned int hl7q_query_failed;      

#define HL7Q_OK         0  /* Returns OK */
#define HL7Q_NOMSG      1  /* May return OK */
#define HL7Q_INVSRV     2
#define HL7Q_WAITLIM    3
#define HL7Q_SOCKET     4
#define HL7Q_GETHOST    5
#define HL7Q_CONNECT    6
#define HL7Q_SSELECT    7
#define HL7Q_SEND       8
#define HL7Q_RSELECT    9
#define HL7Q_RECV      10
#define HL7Q_SIZE      11
#define HL7Q_HL7       12
#define HL7Q_NAK       13
#define HL7Q_MEM       14
#define HL7Q_MPI       15
#define HL7Q_CANCELED  16

#endif

/***************************** Query structures *******************************
 * Note that struct names have same general format as function calls.  User
 * creates struct and passes reference to function.  Function will release any
 * memory allocated in struct; If .acc is negative, function will release any
 * memory allocated in struct and return without attempting a query.   
 *****************************************************************************/
struct _hl7_accn_query_ {
   int accn;	   	/* requested accession */
	int mpi;			   /* returned mpi number - PID 3.1 */
   char *dnumber;    /* department number   - PID 4.1 */
	char *visit;		/* returned visit      - PID 18  */
	char *pname;		/* returned name       - PID 5   */
   char *alias;      /* returned alias name - PID 9   */
	char *dob;			/* returned dob        - PID 7   */
   char *sex;        /* patient gender      - PID 8   */
   char *ssn;        /* patient SSN         - PID 19  */
	char *exam;			/* returned exam       - RAD 1  (OBR 4.1) */
	char *exam_desc;	/* returned exam desc  - RAD 2  (OBR 4.2) */
	char *status;		/* returned status     - ZOS 1   */
   char *start_time; /* sched start time    - ORH 11 (OBR 27) */
	char *time;       /* status chg time     - ZOS 2   */
   char *r_physician;/* referring physician - RAD 8 (part 2 on) */
};

/******************************************************************************
 * MPI query returns information on all exams for the patient.  The struct 
 * pointer in _hl7_mpi_query_ is allocated x times memory, where x is the numb
 * of exams the patient has, then loaded for each one.  Step through with a
 * _hl7_mpi_query_exam_ pointer, incremented to point to each exam in turn.
 *****************************************************************************/
struct _hl7_mpi_query_exam_ {
   int accn;
   char * code;
   char * desc;
   char * facility;
   char * status;
   char * date_time;
};
struct _hl7_mpi_query_ {
   int mpi;
   int num_exams;
   struct _hl7_mpi_query_exam_ * exam;
};
/***************************** Prototypes ************************************/
char * GetHl7QErrorString(void);
void PrtHl7QErrorString(FILE * file, char * prefix);
int hl7SelectHost(char * host, char * ip);
const char * hl7QSelectServer(char feed, int wait, char persist);
int hl7IncludeCanceledExams(char boolean);
int hl7MsgQuery(char * msg, char * reply, int size);
int hl7AccnQuery(struct _hl7_accn_query_ * query_data);
int hl7MPIQuery(struct _hl7_mpi_query_ * query_data);
int hl7PatientQuery(struct _hl7_mpi_query_ * mpi_query_data, 
    struct _hl7_accn_query_ ** accn_query_data, char * cutoff_date);
