/******************************************************************************
 * ini.c                                                              R Moulton
 * Library of routines which allow storage and retrieval of program parameters
 * in an ini file for the application, similar to the Windows 3.x 
 * "PrivateProfile" calls.
 * The ini file pfn must be of the form: filename.ini.
 * Its contents must be of the form:
 *   [section]
 *   key=value
 *     .
 *     .
 * Leading spaces and tabs are ignored.  Lines which begin with any character 
 * other than isalnum or "[" are treated as comments.
 * Unlike Windows, these routines are case sensitive in regards to file, 
 * section, and key names.  Values are stored and retrieved without changing
 * case. If the value is enclosed in single or double quotation marks, these 
 * will be discarded before the value is returned.  All routines which update
 * the ini file lock the file.  All routines block until they can access the
 * ini file.  Note: ini files are meant to contain parameter settings, not
 * status or other dynamic information.
 *-----------------------------------------------------------------------------
 * Compile using:  gcc -c ini.c
 *-----------------------------------------------------------------------------
 * 9-3 R Moulton Initial write.
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include <sys/time.h>

#include "ini.h"
#include "dbg.h"

#ifdef _VMS_MODE_
#include <ssdef.h>
#include <descrip.h>
#include <starlet.h>
#endif


/*----------------------------------------------------------------------- Macros */
#define MAKE_STR(A,B) { if (A=malloc(strlen(B)+3)) strcpy(A,B); }
#define NEW_STR(A,B) { if (A) free(A); if (A=malloc(strlen(B)+3)) strcpy(A,B); }
#define FREE_STR(A) if (A) { free(A); A=NULL; }
#define STR_EQ(A,B) !strcasecmp(A,B)

/*-------------------------------- Error msg strings. Match to ini_errno values */
static char * e_msg[] = 
{	"No error",
	"No result buffer specified",
	"pfn not *.ini format",
	"Could not open ini file",
	"I/O error on ini",
	"Result buffer overflow",
	"No sections found",
	"Section not found",
	"Key not found",
	"Invalid section name",
	"Invalid key name",
	"Internal error",
   "Get cwd failed",
   "chdir failed"
};
static int max_e_msg = sizeof(e_msg) / sizeof(char *);
static char * no_msg = "Unknown ini_errno value";

#define TRUE   1
#define FALSE  0
#define INI_R_OK   1
#define INI_R_FAIL 0
#define MAX_LINE 1024

/*-------------------- Function Prototypes ----------------------------------*/
static void trim(char * str, char * extra_delimiters);

/******************************************************************************
 * GetIniString retrieves a string from the specified section of the ini file.
 * sname is the section name containing the desired key.  If NULL, function
 * returns all the section names in the file, up to the size limit.
 * kname is the name of the key whose string is to be retrieved.  If NULL,
 * function returns all key names in the specified section in the file, up to
 * the size limit.
 * dvalue is the default value, returned if the indicated value can not be
 * found.  If NULL, function will return a zero length string if the value 
 * cannot be found.
 * rvalue is a pointer to a user supplied buffer in which function will place
 * the result of the request.
 *****************************************************************************/
int GetIniString(char * sname, char * kname, char * dvalue, char * rvalue, int size, char * pfn)
{	FILE * ini;
	char line[MAX_LINE], * p, * k, * v, * found_equal_sign;
	char isname[MAX_LINE], ikname[MAX_LINE], idvalue[MAX_LINE], rdvalue[MAX_LINE];
	int right_section = FALSE, found = FALSE, ret = INI_R_OK;
/*--------------------------------------------------------------- Inbound debug */
   memset(em,0,sizeof(em));
   strcpy(em, "Ini: s="); if (sname)  strcat(em,sname);  else strcat(em,"NULL");
   strcat(em," k=");      if (kname)  strcat(em,kname);  else strcat(em,"NULL");
   strcat(em," d=");      if (dvalue) strcat(em,dvalue); else strcat(em,"NULL");
   strcat(em," p=");      if (pfn)    strcat(em,pfn);    else strcat(em,"NULL");
   dbg(em); 

	ini_errno = INI_OK;                              /* Clear any previous error */
/*--------------------------------------- Trim parameter values for matching */
	if (sname) { strcpy(isname,sname); trim(isname, "[]"); }
	if (kname) { strcpy(ikname,kname); trim(ikname, NULL); }
	if (dvalue){ strcpy(idvalue,dvalue); trim(idvalue, NULL); strcpy(rdvalue,dvalue);}	
	
	if (!rvalue) 
	{	ini_errno=INI_NORSLT;
		return(INI_R_FAIL);
	}
	memset(rvalue,0,size);                                    /* Clear return */

/*----------------------------------- Validate that pfn is of the form *.ini */
	if (strcasecmp(pfn + strlen(pfn) - 4,".ini"))
	{	ini_errno=INI_PFN; 
		if (dvalue) strncpy(rvalue, rdvalue, size - 1);
		return(INI_R_FAIL);
	}
	
	ini = fopen(pfn, "r");                                   /* Open ini file */
	if (!ini) 
	{ 	ini_errno = INI_OPEN; 
		if (dvalue) strncpy(rvalue, rdvalue, size - 1);
		return (INI_R_FAIL); 
	}

/*-------------------------------------------- Pass lines in file until done */
	while (1)
	{	p = fgets(line, MAX_LINE, ini);
		if (!p)
		{	if (feof(ini)) break;
			ini_errno = INI_IO; 
			ret = INI_R_FAIL;
			break;
		}
		trim(p, NULL);
		if (!strlen(p)) continue;                  /* Ignore blank/empty lines */
		if (!isalnum(*p) && *p != '[') continue;                   /* Comments */ 

/******************************************************************* Section */
		if (*p == '[')
/*---------------------------------------------- Copy all sections to result */
		{	if (!sname)      
			{	found = TRUE;
				if ((strlen(p) + strlen(rvalue)) < size) 
				{	strcat(rvalue,p); 
					continue; 
				}
				else
				{	ini_errno = INI_OVFRSLT;
					break;
				}
			}
/*--------------------------------------------- Look for the section we want */
			if (right_section == TRUE) break;  /* already in right section, so we're done */
			trim(p, "[]");     /* Pull section name from between [] */
			if (!strcmp(p,isname)) right_section = TRUE;
			continue;
		}
		else
/*********************************************************************** Key */
		{	if (!sname) continue;          /* If copy all sections, ignore keys */
			if (right_section == FALSE) continue;  /* Skip if not in right section */
         found_equal_sign = strchr(p,'=');
			k = strtok(p,"="); trim(k,NULL);                    /* Pull key name*/
/*--------------------------------------- Copy all keys in section to result */
			if (!kname)
			{	found = TRUE;
				if ((strlen(k) + strlen(rvalue) + 1) < size)
				{	strcat(rvalue,k); strcat(rvalue,"=");
					continue;
				}
				else
				{	ini_errno = INI_OVFRSLT;
					break;
				}
			}
/*----------------------------------------- Match specific key, return value */
			if (strcmp(k,ikname)) continue;
			v = strtok(NULL,"\0");
			if (v)
			{	trim(v,NULL);
/*--------------------------------- Remove enclosing single or double quotes */
				if (*v == '\'' && *(v + strlen(v) -1) == '\'') trim(v,"\'");
				else if (*v == '\"' && *(v + strlen(v) -1) == '\"') trim(v,"\"");
				strcpy(rvalue,v);
			}
			if (found_equal_sign) found = TRUE;
			break;
		}
	}
	fclose(ini);
	if (found == TRUE) 
   {  sprintf(em," r=%s", rvalue); dbg(em);
      return (ret);                              /* success */
   }
	if (dvalue) { strcpy(rvalue,rdvalue); trim(rvalue,NULL); }
   sprintf(em," r=%s", rvalue); dbg(em);
	if (ini_errno != INI_OK) return (ret);                  /* codes already set */
/*---------------------------------- Load of section names found no sections */
	if (!sname) { ini_errno = INI_NOSEC; return (INI_R_OK); } 
/*------------------------------------------------ Never found right section */
	if (!right_section) { ini_errno = INI_SECNOF; return (INI_R_FAIL); }
	ini_errno = INI_KEYNOF; return(INI_R_OK);
}

/******************************************************************************
 * These routines are just convenience calls to GetIniString with slight
 * variations:
 * GetIniInt returns a specific key value as an integer.
 * GetIniSection returns all the key names in a specified section.
 * GetIniSectionNames returns all the section names in the ini file.
 *****************************************************************************/
int GetIniInt(char * sname, char * kname, int dvalue, int * rvalue, char * pfn)
{	char irvalue[MAX_LINE];
	char idvalue[MAX_LINE];
	int ret;
	if (!sname || !strlen(sname)) { ini_errno = INI_INVSEC; return (INI_R_FAIL); }
	if (!kname || !strlen(kname)) { ini_errno = INI_INVKEY; return (INI_R_FAIL); }
	sprintf(idvalue, "%d", dvalue);
	ret = GetIniString(sname, kname, idvalue, irvalue, MAX_LINE, pfn);
	*rvalue = atoi(irvalue);
	return (ret);
}
int GetIniSection(char * sname, char * rvalue, int size, char * pfn)
{	int ret;
	if (!sname || !strlen(sname)) { ini_errno = INI_INVSEC; return (INI_R_FAIL); }
	ret = GetIniString(sname, NULL, NULL, rvalue, size, pfn);
	return (ret);
}
int GetIniSectionNames(char * rvalue, int size, char * pfn)
{	int ret;
	ret = GetIniString(NULL, NULL, NULL, rvalue, size, pfn);
	return (ret);
}

/******************************************************************************
 * PutIniString inserts a key and corresponding value into the designated 
 * section of the ini file. If the section or key do not exist, they are added.
 * sname is the section name of the desired entry.  Cannot be NULL.
 * kname is the key name of the desired entry.  If NULL, the section and all
 * keys in it are deleted from the file.
 * nvalue is the new value for the desired entry. If NULL, the key entry is
 * deleted from the file.
 *****************************************************************************/
int PutIniString(char * sname, char * kname, char * nvalue, char * pfn)
{	FILE * ini, * tmp;
	char line[MAX_LINE], o[MAX_LINE], *p, *k;
	char isname[MAX_LINE], ikname[MAX_LINE], invalue[MAX_LINE];
	int right_section = FALSE;    /* Are we in the right section of the file? */
	int found = FALSE;              /* Have we found what we are looking for? */
	int copy_back = FALSE;          /* Will we need to copy back tmp at EOF? */
	int done = FALSE;                          /* Are we done making changes? */
	int ret = INI_R_OK;                          /* Return value for function */
	ini_errno = INI_OK;                           /* Clear any previous error */
/*-----------------------------------------------------------------------------
 * Must have section name.  Trim off brackets, and any enclosing quotes.
 * Result must be non-blank.
 *---------------------------------------------------------------------------*/
	if (!sname) {ini_errno=INI_INVSEC; return(INI_R_FAIL); }
	strcpy(isname,sname); trim(isname, "[]");
	if (*isname == '\'' && *(isname + strlen(isname) -1) == '\'') trim(isname,"\'");
	else if (*isname == '\"' && *(isname + strlen(isname) -1) == '\"') trim(isname,"\"");
	if (!strlen(isname)) { ini_errno=INI_INVSEC; return(INI_R_FAIL); }
/*-----------------------------------------------------------------------------
 * If keyname exists, trim as above, and result must be non-blank
 *---------------------------------------------------------------------------*/
	if (kname) 
	{	strcpy(ikname,kname); trim(ikname, "[]");
		if (*ikname == '\'' && *(ikname + strlen(ikname) -1) == '\'') trim(ikname,"\'");
		else if (*ikname == '\"' && *(ikname + strlen(ikname) -1) == '\"') trim(ikname,"\"");
		if (!strlen(ikname)) { ini_errno=INI_INVKEY; return(INI_R_FAIL); }
	}
/*-----------------------------------------------------------------------------
 * If nvalue exists, trim as above.  This may be blank.
 *---------------------------------------------------------------------------*/
	if (nvalue)
	{	strcpy(invalue,nvalue); trim(invalue, "[]");
		if (*invalue == '\'' && *(invalue + strlen(invalue) -1) == '\'') trim(invalue,"\'");
		else if (*invalue == '\"' && *(invalue + strlen(invalue) -1) == '\"') trim(invalue,"\"");
	}
	
/*----------------------------------- Validate that pfn is of the form *.ini */
	if (strcasecmp(pfn + strlen(pfn) - 4,".ini"))
	{	ini_errno=INI_PFN; 
		return(INI_R_FAIL);
	}
	
	ini = fopen(pfn, "r+");                                   /* Open ini file */
	if (!ini) 
	{ 	ini_errno = INI_OPEN; 
		return (INI_R_FAIL); 
	}
/*----------------------------------------- Open tmp used to update ini file */
	tmp = tmpfile();
	if (!tmp) {	fclose(ini); ini_errno=INI_INT; return(INI_R_FAIL); }

/*-------------------------------------------- Pass lines in file until done */
	while (TRUE)
	{	p = fgets(line, MAX_LINE, ini);
		if (!p)
		{	if (feof(ini)) break;
/*--------------------------------------------------------------- File error */
			ini_errno = INI_IO;
			fclose(ini); fclose(tmp); 
			return (INI_R_FAIL);
		}
/*----------------- If we're done with changes, just copy the rest until EOF */
		if (done == TRUE) { fputs(p, tmp); continue; }
/*-----------------------------------------------------------------------------
 * We save the original line read from the file.  This is the line that will be
 * passed to tmp to be copied back later.  That way we will not lose any
 * special formatting the user might have put in for their own readability.
 *---------------------------------------------------------------------------*/
		strcpy(o,p);
		trim(p, NULL);
		if (!strlen(p)) { fputs(o, tmp); continue; }    /* ignore blank lines */
		if (!isalnum(*p) && *p != '[') { fputs(o, tmp);  continue; } /* Comments */

/******************************************************************* Section */
		if (*p == '[')
/*-----------------------------------------------------------------------------
 * If we were already in the section we want, we are now at the end of that 
 * section. If inserting a key, we insert it here. In any case, we are done
 *---------------------------------------------------------------------------*/
		{	if (right_section)
			{	if (kname && nvalue)
				{	memset(line,0,MAX_LINE);
					sprintf(line, "%s=%s\n", ikname, invalue);
					fputs(line, tmp);
					copy_back = TRUE;
					found = TRUE;
				}
				fputs(o, tmp);
				done = TRUE;
				continue;
			}
/*----------------------- Look for right section, just copy until we find it */
			trim(p, "[]");
			if (strcmp(p,isname)) { fputs(o, tmp); continue; }
			right_section = TRUE;
/*-----------------------------------------------------------------------------
 * Here we are looking at the section header for the right section.  If kname
 * is NULL, we're deleting the section (not copying to tmp).  Otherwise, we
 * copy and continue, looking for the key we want.
 *---------------------------------------------------------------------------*/
			if (kname)  { fputs(o, tmp); continue; }
			copy_back = TRUE;
			found = TRUE;
			continue;
		}
/*********************************************************************** Key */
/*--------------------- If not in right section, just copy and keep on going */
		if (right_section == FALSE) { fputs(o, tmp); continue; }
/*--------------------------------- In right section, deleting whole section */
		if (!kname) continue;
/*---------------------------------- In right section, looking for right key */
		k = strtok(p,"="); trim(k,NULL);
		if (strcmp(k,ikname)) { fputs(o,tmp); continue; }  /* Not right key */
/*-------------------------------------- Found the right key... replacing it */
		if (nvalue)
		{	memset(line,0,MAX_LINE);
			sprintf(line, "%s=%s\n", ikname, invalue);
			fputs(line, tmp);
		}
/*-------------------------------------- Or just deleting it, by not copying */
		copy_back = TRUE;
		found = TRUE;
		done = TRUE;
		continue;
	}
/*-----------------------------------------------------------------------------
 * At EOF on ini file.  If we were adding/modifying a key, and we haven't done
 * it yet, it may be because we never found the right section (in which case we
 * add it) or the right section was the last section and they key was not in it
 * (in which case we add the key).  In both of these cases, we can just re-open
 * the ini file append, and skip the copy back.
 *---------------------------------------------------------------------------*/
	if (kname && nvalue && (found == FALSE))
	{	fclose(ini); fclose(tmp); 
		ini = fopen(pfn, "a");
		if (!ini)
		{	ini_errno = INI_OPEN;
			return (INI_R_FAIL);
		}
		if (right_section == FALSE)
		{	ret = fprintf(ini, "[%s]\n", isname);
			if (ret < 0)
			{	ini_errno = INI_IO;
				fclose(ini);
				return (INI_R_FAIL);
			}
		}
		ret = fprintf(ini, "%s=%s\n", ikname, invalue);
		if (ret < 0)
		{	ini_errno = INI_IO;
			fclose(ini);
			return (INI_R_FAIL);
		}
		fclose(ini);
		return (INI_R_OK);
	}
/*-----------------------------------------------------------------------------
 * If we're doing a delete and we haven't found it yet, then whatever we were
 * trying to delete wasn't there.  This is an OK return but we set ini_errno to
 * indicate.  We can just close the files.
 *---------------------------------------------------------------------------*/
	if (found == FALSE)
	{	fclose(ini); fclose(tmp); 
		if (!kname) ini_errno = INI_NOSEC;
		else ini_errno = INI_KEYNOF;
		return (INI_R_OK);
	}
/*-----------------------------------------------------------------------------
 * Whatever change is wanted has been made.  If we need a copy back, do it.
 *---------------------------------------------------------------------------*/
	if (copy_back == TRUE)
	{	ini = freopen(pfn, "w", ini);
		if (!ini) 
		{ 	ini_errno = INI_OPEN; 
			return (INI_R_FAIL); 
		}
		rewind(tmp);
		while (TRUE)
		{	if (fgets(line, MAX_LINE, tmp))
			{	if (fputs(line, ini) != EOF) continue;
				else { ini_errno = INI_IO; ret = INI_R_FAIL; break; }
			}
			else
			{	if (ferror(tmp)) { ini_errno = INI_IO; ret = INI_R_FAIL; }
				break;
			}
		}
	}
	fclose(ini); fclose(tmp); 
	return (ret);
}

/******************************************************************************
 * Key value add/update, but the nvalue is provided here as an integer
 *****************************************************************************/
int PutIniInt(char * sname, char * kname, int nvalue, char * pfn)
{	char value[MAX_LINE];
	int ret;
	memset(value,0,MAX_LINE);
	sprintf(value,"%d", nvalue);
	ret = PutIniString(sname, kname, value, pfn);
	return (ret);
}

/******************************************************************************
 * Replaces an entire section with the key value combinations in the nvalue
 * buffer.  existing section is deleted if it exists.  nvalue consists of a 
 * series of NULL terminated strings of the form "key=value", followed by a
 * single NULL value.  Stops at first failure point.
 *****************************************************************************/
int PutIniSection(char * sname, char * nvalue, char * pfn)
{	char str[MAX_LINE], kname[MAX_LINE], value[MAX_LINE], * p = nvalue, *k, *v;
	int ret;
	ret = PutIniString(sname, NULL, NULL, pfn);     /* Clear existing section */
	if (ret == INI_R_FAIL) return (ret);
	while (strlen(p))
	{	strcpy(str,p);
		k = strtok(str,"=");
		v = strtok(NULL,"\x00");
		strcpy(kname,k);
		strcpy(value,v);
		ret = PutIniString(sname, kname, value, pfn);
		if (ret == INI_R_FAIL) return (ret);
		p += (strlen(p) + 1);
	}
	return (INI_R_OK);
}

/******************************************************************************
 * return error message string for ini_errno, or print to stdout
 *****************************************************************************/
char * GetIniErrorString(void)
{	if (ini_errno < max_e_msg) return e_msg[ini_errno];
	else return no_msg;
}
void PrtIniErrorString(FILE * file, char * prefix)
{	if (!file) file = stdout;
	fprintf(file, "%s: %s\n", prefix, GetIniErrorString());
	return;
}

/******************************************************************************
 * Returns str with leading and trailing spaces, horizontal and vertical tabs, 
 * form feeds, newlines, and carriage returns, plus any characters in 
 * extra_delimiters, removed.
 *****************************************************************************/
static void trim(char * str, char * extra_delimiters)
{	char a[MAX_LINE], * p = str;
	char delimiters[] = " \f\n\r\t\v";
	unsigned int i, j;
	if (extra_delimiters) strcpy(delimiters,extra_delimiters);
	p += strspn(p,delimiters);
	strcpy(a,p);
	i = strlen(a);
	while (i-- > 0)
	{	for (j = 0; j <= strlen(delimiters); j++)
		{	if (!delimiters[j]) {i = 0; break; }     /* Didn't find chr in delim string */
			if (a[i] == delimiters[j]) {a[i] = 0; break; }   /* Did find it */
		}
	}
	strcpy(str,a);
	return;
}

/*-------------------------------------------------------------------------------------------------
 * Makes directory program is running from current.  Par0 is the first command line parameter
 * Returns pointer to path with trailing "/" (just tack on file name) or NULL if failure.
 *-----------------------------------------------------------------------------------------------*/
static char cwd[1024] = "\0";                                    /* static, so we can return cwd */
char * MakeProgDirCurrent(char * par0)
{  char *p = par0;
   memset(cwd,0,sizeof(cwd));
#ifdef _VMS_MODE_
/*---------------------------------------------------------------------------------------- VMS mode
 * par0 is always full path name.  Cut off after "]".  If no "]" just return current directory.
 *-----------------------------------------------------------------------------------------------*/
   if (strchr(p,']')) memcpy(cwd,p,(strchr(p,']') - p)+1);      
   else getcwd(cwd,1);  /* second parameter is DEC extension, forces OpenVMS format */
#else
/*--------------------------------------------------------------------------------------- UNIX mode
 * Make the directory program is running from current
 *-----------------------------------------------------------------------------------------------*/
   if (!getcwd(cwd,sizeof(cwd)))               /* start with cwd */
   {  ini_errno = INI_CWDF;
      return(NULL);
   } 
   if (*p == '/') strcpy(cwd,p);                                        /* Prog called absolute */
   else if (*p == '.') strcat(cwd,p+1);                               /* Prog called . relative */
   else {strcat(cwd,"/"); strcat(cwd,p); }                              /* Prog called relative */
   *(strrchr(cwd,'/')) = 0;                                          /* Insert eos for last "/" */
   if (chdir(cwd) == -1)                                                     /* Make it current */
   {  ini_errno = INI_CDF;
      return (NULL);
   }   
   strcat(cwd,"/");                                                         /* Tack on ending / */
#endif
   return (cwd);
}

/*-------------------------------------------------------------------------------------------------
 * Routine verifies that process is not already running
 *-------------------------------------------------------------------------------------------------
 * In VMS, this is done with a process name, which must be 7-15 characters in length.  Calling the
 * routine with a NULL process name will clear the process name by replacing it with one of the
 * form VMS_###.
 *-----------------------------------------------------------------------------------------------*/
#ifdef _VMS_MODE_
int unique_process_check(char * process_name)
{	int status;
	char * edesc;
	char p7[8], p8[9], p9[10], p10[11], p11[12], p12[13], p13[14], p14[15], p15[16];
   $DESCRIPTOR(pn7,  p7);
   $DESCRIPTOR(pn8,  p8);
   $DESCRIPTOR(pn9,  p9);
   $DESCRIPTOR(pn10, p10);
   $DESCRIPTOR(pn11, p11);
   $DESCRIPTOR(pn12, p12);
   $DESCRIPTOR(pn13, p13);
   $DESCRIPTOR(pn14, p14);
   $DESCRIPTOR(pn15, p15);
   if (process_name)
   {  sprintf(em,"Process: %s", process_name); dbg(em);
	   switch (strlen(process_name))
	   {	case 7:
		   	strcpy(p7,process_name);
			   status = sys$setprn (&pn7);
			   break;
		   case 8:
			   strcpy(p8,process_name);
			   status = sys$setprn (&pn8);
			   break;
		   case 9:
			   strcpy(p9,process_name);
			   status = sys$setprn (&pn9);
			   break;
		   case 10:
			   strcpy(p10,process_name);
			   status = sys$setprn (&pn10);
			   break;
		   case 11:
			   strcpy(p11,process_name);
			   status = sys$setprn (&pn11);
			   break;
		   case 12:
			   strcpy(p12,process_name);
			   status = sys$setprn (&pn12);
			   break;
		   case 13:
			   strcpy(p13,process_name);
			   status = sys$setprn (&pn13);
			   break;
		   case 14:
			   strcpy(p14,process_name);
			   status = sys$setprn (&pn14);
			   break;
		   case 15:
			   strcpy(p15,process_name);
			   status = sys$setprn (&pn15);
			   break;
		   default:
			   sprintf(em,"Process name: %s is %d chars", process_name, strlen(process_name)); dbg(em);
			   dbg("Must be 7-15 chars"); 
            sys$exit (1);
			   break;
	   }
      if (status != SS$_NORMAL)
      {  if (status == SS$_ACCVIO) edesc = "ACCVIO";
		   if (status == SS$_DUPLNAM) edesc = "DUPLNAM";
		   if (status == SS$_IVLOGNAM) edesc = "IVLOGNAM";
		   dbg(edesc);
		   sprintf(em,"%s is already running", process_name); dbg(em); 
         sys$exit (1);
      }
      sprintf(em,"%s started", process_name); dbg(em);
      return (0);
   }
   else
   {	int status, count = 1;
      char name[] = "VMS_001";
      $DESCRIPTOR(pn, name);
      while (1)
      {  sprintf (name, "VMS_%03d", count++);
         status = sys$setprn (&pn);
         if (status == SS$_NORMAL) break;
      }
	   exit(0);
   }
}

/*-------------------------------------------------------------------------------------------------
 * In Unix, the unique process check is made by opening a fifo to write.  If this fails, the fifo
 * is opened read-write.  If it succeeds, the fifo is already open read-write, meaning that the 
 * process is already running.  The fd is returned
 *-----------------------------------------------------------------------------------------------*/
#else
int unique_process_check(char * file)
{  char pfn[256];
   int fd;
   if (!file) return(0);
   strcpy(pfn,cwd);
   strcat(pfn,file);
/*--------------- Open file to write; If this works, the process is already running, and we quit */
	fd = open(pfn, O_WRONLY | O_NONBLOCK);
	if (fd >= 0)
	{	close(fd);
		sprintf(em,"%s already running", file); dbg(em);
		exit(1);
	}
/*-------------------------------------------------------------------------------------------------
 * Now we open the file rw.  If another process is run, it will fail first test.
 *-----------------------------------------------------------------------------------------------*/
	fd = open(pfn, O_RDWR | O_NONBLOCK);
	if (fd < 0)
	{	sprintf(em, "Couldn't open %s", pfn); dbg(em);
		exit(1);
	}
   return (fd);
}
#endif

/*-------------------------------------------------------------------------------------------------
 * Opens file for echo/logging, and returns FILE *, based on file name.  May have no name (returns
 * NULL), a standard name (returns std FILE *), or a pfn (open and return FILE *).  pfn may have
 * date format entries, in which case they are resolved using strftime, if permitted.
 *-----------------------------------------------------------------------------------------------*/
FILE * open_log_file(char * file_name, unsigned char dates_allowed)
{  time_t now;
   struct tm * tm;
   char * pfn;
   FILE * fp;
   size_t s;
   if (!file_name || !strlen(file_name)) return NULL;   /* No file */
   if (STR_EQ(file_name,"stdout")) return stdout;       /* standard files */
   if (STR_EQ(file_name,"stderr")) return stderr;
/*------------------------------------------------------------------------------------------ pfn */
   s = strlen(file_name) + 10;
   pfn = malloc(s);
   strcpy(pfn,file_name);
   if (strchr(file_name,'%') && dates_allowed)                    /* If date formatting, resolve */
   {  now = time(NULL);
      tm = localtime(&now);
      strftime(pfn,s,file_name,tm);
   }
   fp = fopen(pfn,"a");
   if (!fp) dbge(pfn);
   free(pfn);
   return fp;
}

void close_log_file(FILE * fp)
{  if (fp != stdout && fp != stderr) fclose(fp);
   return;
}

