/****************************************************************************************
 * msrdLogPrs.c                                                                 R Moulton
 * Processes logs harvested from magic store archive.  Raw log file is in $TMP/msrd#.log
 * and has been extracted using expect script and cleaned up using dos2unix.  Resulting
 * messages are to be appended to /distributor/logs/msrd#/YYYYMMDD.log files. Message are
 * in format:
 *
 * date         :2005.10.17 23:08:33.36
 * message      :I REC, Receiving study of patient 'BOHNERT,DIANNA' from 001S02OT01RCV0
 * 
 * and need to be converted to format:
 *
 * 2005.10.17 23:08:33 I REC, Receiving study of patient 'BOHNERT,DIANNA' from 001S02OT01RCV0
 *
 * Log entries for today's date will not have date in them, so it will have to be added.
 * Their date line will be in format:
 *
 *  date         :15:08:15
 *
 * Program will look in $TMP directory for msrd#.log files, process and delete those it 
 * finds. Other data may be in files, only " date" lines followed by " message" lines 
 * will be processed.  Several days worth of information may be in an msrd#.log file.
 * For each date found, program will create the YYYYMMDD.log file if it does not exist, 
 * or insert lines into the existing file.  If the files does already exist, program will
 * look at the last line already there, and add only those which come after it in time
 * sequence.  Note that multiple messages may be logged in a single second.  The hundreth
 * of a second is not used uniformly, and it is not present in other log files, so it
 * is deleted (as shown in the example above).  Messages with the same second should be
 * matched by text.
 * Invocation is by program name alone.
 *---------------------------------------------------------------------------------------
 * Compile with gcc -g msrdLogPrs.c -o msrdLogPrs `pkg-config --cflags --libs msrdLogPrs`
 *---------------------------------------------------------------------------------------
 * 2005-10-18 R Moulton - Initial write
 ***************************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>

#include <dbg.h>

/*---------------------------- macros -------------------------------------------------*/
#define FREE(A) if(A){free(A);A=NULL;}
#define STR_EQ(A,B) !strcasecmp(A,B)
#define STR(A,B) {if(A)free(A);A=strdup(B);}
#define LCHR(A) A[strlen(A) - 1]
#define CLEAR(A) memset(A,0,sizeof(A));

/*----------------------------- function prototypes -----------------------------------*/


/*------------------------------ working storage --------------------------------------*/
#define DATE_STRING "DATE_STRING"
#define BUFSIZE 10240
static FILE * debug_fp;
static char * tmp;
static char raw_log_pfn[256];
static FILE * raw_log_fp;
static char log_dir[64];
static char log_pfn[64];
static FILE * log_fp;

static char * skip[] = {
   "DRIVE_STATUS",
   "DB_SIZE_OK",
   "BAD_OBJECT_CNT"
};
static char sk, skips = sizeof(skip) / sizeof(char *);




/****************************************************************************************
 ****************************************************************************************
 ********************************** main function ***************************************
 ****************************************************************************************
 ***************************************************************************************/
int main(int argc, char * argv[])
{  char * p;
   DIR * tmp_dp;
   struct dirent * tmp_de;
   time_t now;
   struct tm * tm;
   int mo = 0, da = 0, yr = 0, r;
   char raw_buff[BUFSIZE], buff[BUFSIZE], *rst, todays_date[16], last_record[BUFSIZE],
      log_file_name[14], new_log_file_name[14], flag;
   
   now = time(NULL);
   tm = localtime(&now);
   sprintf(todays_date,"%04d.%02d.%02d", tm->tm_year + 1900, tm->tm_mon +1, tm->tm_mday);
   CLEAR(buff)
   CLEAR(log_file_name)
   umask(007);
   
/*-------------------------------------------------- If par 1 exists, it is debug file */
   if (argc > 1 && strlen(argv[1])) debug_fp = debug_on(argv[1]);

/*--------------------- find tmp directory, validate that it exists and is a directory */
   if (!(p=getenv("TMP"))) dbg_exit("TMP environment variable not defined", 1);
   STR(tmp,p)
   if (!(tmp_dp=opendir(tmp))) dbge_exit("TMP opendir",1);

/****************************************************************************************
 * This loop processes each msrd#.log file in the TMP directory
 ***************************************************************************************/
   while (tmp_de = readdir(tmp_dp))
   {
/*------------------------------ Skip any files in directory which are not "msrd#.log" */
      if (strlen(tmp_de->d_name) != 9 ||        
          memcmp(tmp_de->d_name,"msrd",4) ||
          memcmp(tmp_de->d_name+5,".log",4) ||
          !isdigit(tmp_de->d_name[4])) continue;
/*------------------------------------------------------------------------ raw log pfn */
      sprintf(raw_log_pfn,"%s/%s",tmp,tmp_de->d_name); 
      if (access(raw_log_pfn, R_OK)) 
      {  sprintf(em,"Can't read %s", raw_log_pfn); dbg(em);
         continue;
      }
      if (!(raw_log_fp=fopen(raw_log_pfn,"r")))
      {  sprintf(em,"open %s", raw_log_pfn), dbge(em);
         continue;
      }
/*------------------------------------------------------------- Processed log directory */
      sprintf(log_dir,"/distributor/logs/%s",strtok(tmp_de->d_name,"."));
      if (access(log_dir, F_OK) && mkdir(log_dir,0770)) 
      {  sprintf(em,"mkdir %s", log_dir); dbge(em);
         continue;
      }
      strcat(log_dir,"/");

/****************************************************************************************
 * This loop processes each line in a single msrd#.log raw log file
 ***************************************************************************************/
      while (1)
      {  
         if (!fgets(raw_buff,BUFSIZE,raw_log_fp)) break;
         while (LCHR(raw_buff) == '\r' || LCHR(raw_buff) == '\n' || LCHR(raw_buff) == ' ') LCHR(raw_buff) = '\0';

/*---- Call to date, inserted in dialog, can give us date script was run, or use today */
         if (!memcmp(raw_buff,DATE_STRING,strlen(DATE_STRING)))
         {  p = strtok(raw_buff," ");
            yr = atoi(strtok(NULL," "));
            mo = atoi(strtok(NULL," "));
            da = atoi(strtok(NULL," "));
            sprintf(todays_date,"%04d.%02d.%02d", yr, mo, da);
            continue;
         }

/*-------------------------------------------------------------------------- date line */
         if (!memcmp(raw_buff," date     ",10))   
         {  CLEAR(buff)
            strtok_r(raw_buff,":",&rst);
/*-------------------------- We get date and time, adding date if it isn't on the line */
            if (rst[2] == ':')sprintf(buff,"%s %s", todays_date, strtok(rst,"."));
            else strcpy(buff,rst);
            continue;
         }

/*---------------------------------- message line, but only if we just got a date line */
         if (strlen(buff))
         {  if (!memcmp(raw_buff," message    ",12)) 
            {  strcat(buff," ");
               strtok_r(raw_buff,":",&rst);
               strcat(buff,rst);
               strcat(buff,"\n");
            }
            else 
            {  CLEAR(buff) 
               continue; 
            }
         }
         else continue;

/*----------------------------------------- Some message types we don't bother to save */
         for (sk = 0; sk < skips; sk++) if(strstr(buff,skip[sk])){CLEAR(buff) break;}
         if (!strlen(buff)) continue; 

/****************************************************************************************
 * buff now contains a message ready to add to the appropriate log file
 ***************************************************************************************/
         dbg(buff);
/*------------------- Get the date out of the message and build pfn of needed log file */
         r = sscanf(buff,"%d.%d.%d ", &yr, &mo, &da);
         if (r != 3) { sprintf(em,"date sscanf r=%d",r); dbg_exit(em, 1); }
         sprintf(new_log_file_name,"%04d%02d%02d.log", yr, mo, da);
/*----------------- If names don't match we need to open and position the new log file */
         if (strcmp(new_log_file_name,log_file_name))
         {  if (log_fp) fclose(log_fp);             /* Close previous log file if open */
/*---------------------------------------------------------------------------- new pfn */
            strcpy(log_file_name,new_log_file_name);
            strcpy(log_pfn,log_dir); strcat(log_pfn,log_file_name);

/*---------------------------------------------------------------------------------------
 * If this yyymmdd.log file does not exist, we open to write, create, and set last_record
 * to low values. If the file exists, we need to pull the last record from the file to 
 * compare with the records we are generating. 
 *-------------------------------------------------------------------------------------*/
            if (!access(log_pfn, F_OK))
            {  if (!(log_fp = fopen(log_pfn, "r"))) dbge_exit("yyyymmdd.log open",1);
               while (fgets(raw_buff,BUFSIZE,log_fp)) strcpy(last_record,raw_buff);
               if (ferror(log_fp)) dbge_exit("yyyymmdd.log read", 1);
               else fclose(log_fp);
               flag = 0;
            }
            else 
            {  CLEAR(last_record)
               flag = 1;
            }
            if (!(log_fp = fopen(log_pfn, "a"))) dbge_exit("yyyymmdd.log open",1);
         }

/*---------------------------------------------------------------------------------------
 * Now we check against the last record, to see if we need to add this one.
 *-------------------------------------------------------------------------------------*/
         if (!flag)
         {  r = memcmp(buff,last_record,19);
            if (r > 0) flag = 1;
            if (r == 0 && !strcmp(buff,last_record)) 
            {  CLEAR(buff) 
               flag = 1; 
               continue; 
            }
         }
         if (flag) if(fputs(buff,log_fp) == EOF) dbge_exit("log write",1);
         CLEAR(buff)
      }
            
/*------------------------------------------------- finished with current raw log file */
      if (ferror(raw_log_fp)) dbge("raw log I/O");
      fclose(raw_log_fp);
      remove(raw_log_pfn);
      if (log_fp) fclose(log_fp);
      CLEAR(log_file_name)
   }
   closedir(tmp_dp);
   return (0);
}
