/*****************************************************************************
 * PACSO_CLIENT							      B Clayton
 * PACS Out to Broker TCP/IP client.  This works in conjunction with the PACSO
 * IDX feed, which receives the outbound messages from IDX and puts them in 
 * high and low priority directories.  This program maintains a TCP/IP 
 * connection with the Broker and sends the messages on, high priority one's
 * first, then low priority.  There may be more than one instance of this 
 * program running
 *-----------------------------------------------------------------------------
 * 8/3 R Moulton Modified to run multiple instances, add logging and error
 * logging, add control file bounce and down, add status file update, add
 * print, e-mail, and pager warnings.
 * 5/7 R Moulton. Modified to allow overide e-mail and pager addresses at the
 * feed level.
 *-----------------------------------------------------------------------------
 * VMS: compile in src_pacso
 * Unix: gcc -g pacso_client.c dbg.c ini.c hl7.c alert.c -lsocket -lnsl -lz -lm -o pacso_client
 ******************************************************************************/

/*  #define _VMS_MODE_ */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <ctype.h>
#include <dirent.h>

#ifdef _VMS_MODE_

#include <iodef.h>
#include <unixio.h>
#include <ssdef.h>
#include <starlet.h>
#include <descrip.h>
#include <socket.h>
#include <in.h>
#include <inet.h>

#include "[source.common]alert.h"
#include "[source.common]ini.h"
#include "[source.common]dbg.h"
#include "[source.common]hl7.h"

#else

#include "alert.h"
#include "ini.h"
#include "dbg.h"
#include "hl7.h"

#endif

/*------------------------------------------------------------- Macros */
#define MAKE_STR(A,B) if (A=malloc(strlen(B)+1)) strcpy(A,B);
#define FREE_STR(A) if (A) { free(A); A=NULL; }
#define STR_EQ(A,B) !strcasecmp(A,B)

/*---------------------------------------------------------- Constants */
#define SB 11
#define EB 28

#define LOG 0
#define ERRLOG 1

/**************************** Prototypes *************************************/
void read_ini(char * feed);
static int send_file (char *fn, int mode);
void connect_to_server(void);
static int cleanup (int *open_sock);
int PACS_send (char *fn, int mode);
static int PACS_recv (int mode);
#define WAIT 0
#define NO_WAIT 1
void write_log (char *msg);
static int read_pacs (void);
static int write_pacs (void);
int print_inout_msg (char *msg, int dash);
void log_error(char * msg);
void update_status(int force);
void count_transaction(char * fn);
void purge_logs(void);
int control_check(void);
void link_error_check(int error_code);
char * strcmb(char *s1, char *s2);
void shutdown_connection_status(void);
/*-------------------------------------------------------- Error checking ws */
#define CONNECTION 0
#define MESSAGES   1
#define XMIT	   2
#define RECV	   3
struct _err_cond_
{  time_t basis;	       /* Time basis for this error */
   int idx_down;	       /* Do we process this error during IDX down window? */
   int print_mins;	       /* Minutes until print warning (<= 0 turns off) */
   int print_interval;	       /* Minutes between print warnings */
   time_t last_print;	       /* Time of last print warning */
   int page_mins;	       /* Same, for pager */
   int page_interval;
   time_t last_page;
   char * suffix;	       /* env variable suffix */
   char * name;		       /* Error description */
};
#define ERR e[err]
static struct _err_cond_ e[] = {
   {  0, 1,  30, 15, 0, 30, 30, 0, "Connection", "No peer connection" },
   {  0, 0,  30, 15, 0, 30, 30, 0, "Messages",   "No messages to send" },
   {  0, 1,  30, 15, 0, 30, 30, 0, "Xmit",       "Transmit delay" },
   {  0, 1,  30, 15, 0, 30, 30, 0, "Recv",       "Receive delay" },
};
static char err;
static char max_err = sizeof(e) / sizeof(struct _err_cond_);
static char * bu_e;

/******************************** ini variables ******************************/
#define INIFILE "PACSO.INI"
#define PROGRAM_TYPE "CLIENT"

static char * cwd = NULL;   /* Current working directory */

static char process_name[15];
static char lock_name[24];
static char peer_ip[24];
static int  port;

/*---------------------------------------------- Message and log directories */
#ifdef _VMS_MODE_
static char * msg_dir[] = { "[.high]", "[.low]", "[.A08]"};
#else
static char * msg_dir[] = { "./high/", "./low/", "./A08/"};
#endif
static char * bu_msg_dir;
static char CombineA08s = 0;

#define MDIR dirp[dir]
static DIR * dirp[3];
static char dir;
static max_dir = 3;
struct dirent * entry;

/*------------------------------------------ process control file parameters */
static char control_pfn[128];
static time_t last_control_check = 0;
static int control_check_seconds = 60;

static char status_pfn[128];
static int status_update_seconds = 60;

static char log_dir[256];
static char log_pfn_format[128] = "%s%4d%02d.LOG";
static char log_pfn[256];
static char elog_pfn_format[128] = "%s%4d%02d.ERR";
static char elog_pfn[256];
static int log_purge_hour = 2;				 /* hour to purge logs (0-23) */
static time_t last_purge_date = 0;			 /* date last purge was done (yyyymmdd) */
static int log_retain_mons = 3;				 /* # of months to retain log files */
static int files_purged = 0;				 /* # of files purged last purge */
static int efiles_purged = 0;

static char error_printer[128] = "mirp1$printer";         /* MIR error notification printer */
static int print_window_start = 6;		     /* message print time window, 24 hr clock */
static int print_window_end = 18;

static int e_mail_warning_messages;		      /* Should error and warning messages email? */
static char e_mail_user_name[128] = "mirhelpdesk@mir.wustl.edu";

static char pager1[128] = "3144241687@myairmail.com";     /* Pager #, Def = PACS primary pages */
static char pager2[128] = "\0";
static int pager_window_start = 6;			 /* page window, 24 hr clock */
static int pager_window_end = 2;

/*----------------------------------------- IDXrad system down window */
static int idx_down_dow = 3;
static int idx_down_shr = 2;
static int idx_down_smin = 0;
static int idx_down_mins = 45;

#define MAX_TCP_BUFFER 40 * 1024

#define LOWS_TO_LOAD 100
#define LOWS_TO_SEND 2
#define SECONDS_TO_SLEEP 5

static char IDXrad_buffer[MAX_TCP_BUFFER];
static char PACS_buffer[MAX_TCP_BUFFER];

static int sock = 0;
struct hostent *psrvent, srvent;
static struct sockaddr_in sock_name;
static size_t sock_length = sizeof(struct sockaddr_in); 

static int need_more;

static char *fn_high[25000];
static char *fn_low[25000];

/******************************************************************************
 * debug
 *****************************************************************************/
extern int errno;
static FILE * debug_fp;

void logg(char *msg) {write_log(msg); dbg(msg);}
void ldbg(char *msg) {log_error(msg); dbg(msg);}

/*----------------------------------------- Debug to dump buffer area */
void dbgm(char*msg, char*buff)
{  char p[53], *ptr;
   int piece = sizeof(p) - 3, cont = 1;
   if (!debug_is_on()) return;
   if (strlen(msg) > (sizeof(p) -3)) dbg(msg);
   else
   {   memset(p,0,sizeof(p)); strcpy(p,msg);
      if (p[strlen(p)-1] != ' ')p[strlen(p)] = ' ';
      memset(p+strlen(p),'-',(sizeof(p)-strlen(p)-1));
      dbg(p);
   }

   if (!buff) { dbg("NULL buffer\n"); return; }
   while (cont)
   {  memset(p,0,sizeof(p)); strcpy(p,"> ");
      if (strlen(buff) < piece) { strcat(p,buff); cont = 0; }
      else { memcpy(p+2,buff,piece); buff += piece; }
      if (ptr = strchr(p,SB)) *ptr = '.';
      if (ptr = strchr(p,EB)) *ptr = '.';
      while (ptr = strchr(p,'\r')) *ptr = '~';
      dbg(p);
   }
   dbg(" ");
}

/*-------------------------------------------- History items for status file */
static time_t interface_started_at;
static time_t interface_bounced_at = 0;
static time_t last_connect_time = 0;
static time_t last_disconnect_time = 0;
static char peer[24];
unsigned int connect_count = 0;
static int transaction_day = -1, transactions_this_day;
static int transaction_hour = -1, transactions_this_hour;
static char msg1[128] = "\0", msg2[128] = "\0", msg3[128] = "\0";
static time_t last_trans_time = 0;

/******************************************************************************
 ******************************************************************************
 **************************** Main function ***********************************
 ******************************************************************************
 *****************************************************************************/
static int main_loop = 1, tcp_loop = 1;

int main (int argc, char *argv[])
{  int count = 30;
   int a, b, zero_low, count_high = 0, count_low = 0, remain_low;
   char fn[256], wstr[256];
   int no_con_count = 0, was_sleeping = 0;

   if (!(cwd = MakeProgDirCurrent(argv[0])))
   {  PrtIniErrorString(NULL, "cwd");
      exit(1);
   }

   if (argc < 2) 
   {  printf("Feed ID must be parameter 1\n");
      return (1);
   }

   read_ini(argv[1]);
   unique_process_check(process_name);

/*--------------------------------------------------------------- Validate directories */
   if (!CombineA08s) msg_dir[2] = "NULL";
   for (dir = 0; dir < max_dir; dir++) 
   {  MDIR = NULL;
      if (STR_EQ(msg_dir[0],"NULL")) continue;  
      MDIR = opendir(msg_dir[dir]);
      if (MDIR == NULL) dbge_exit("msg directory error",1);
   }


   logg("PACSO Client started");
   interface_started_at = e[CONNECTION].basis = e[MESSAGES].basis = time(NULL);
   update_status(1);
   while (main_loop)
   {  dbg("main loop");
      /*-----------------------------------------------------------------------------
       * main processing loop. Get file from high/low directory, put there by 
       * pacso_feed.  Send it to broker, accept response from broker.
       *---------------------------------------------------------------------------*/
      while (tcp_loop)
      {	if (control_check()) continue;
         update_status(0);
         /*----------------------------------------------------------- Try to connect */
         connect_to_server();
         if (!sock) {link_error_check(CONNECTION); continue; }

/*--------------------------------------------- Get next file to send, if any */         
         memset(fn,0,sizeof(fn));                          /* clear file name */
         for (dir = 0; dir < max_dir; dir++)              /* pass directories */
         {  if (MDIR == NULL) continue;          /* skip dirs not implemented */
            rewinddir(MDIR);                                /* start dir read */
            while (entry = readdir(MDIR))                        /* get entry */
            {  if (!strstr(entry->d_name,".DAT")) continue;     /* .DAT file? */
               strcpy(fn,msg_dir[dir]);                           /* load pfn */
               strcat(fn,entry->d_name);
               break;                     /* break (both) loops if file found */
            }
            if (strlen(fn)) break;
         }
         if (!strlen(fn))                      /* nothing found, sleep & loop */
         {  link_error_check(MESSAGES);
            sleep(60); 
            continue; 
         }      
         if (!send_file(fn, WAIT)) unlink(fn);
      }
   }
   dbg("Exiting main loop");
   shutdown_connection_status();
   unique_process_check(NULL);
}
/*******************************************************************************
 * send selected file to server
 ******************************************************************************/
static int send_file (char *fn, int mode)
{  FILE *fp;
   unsigned int rc, sec_start, sec_stop;

   while (1)
   {  if (!sock) return(1);
/*------------------------------------------------------------ try to send it */
      sec_start = time (NULL);

      sprintf (em, "SENDING: %s", fn); dbg(em); 

      rc = PACS_send (fn, mode);
      if (rc == 2) return (0); /* bad file */
      if (rc)
      {	sprintf(em,"TX FAIL: %s", fn); dbg(em); write_log(em);
         if (mode == NO_WAIT) return (1);
         else continue;
      }
/*------------------------------------------------------------ get the answer */
      if (PACS_recv (mode))
      {	 sprintf(em,"RX FAIL: %s", fn); dbg(em); 
         if (mode == NO_WAIT) return (1);
         else continue;
      }
/*------------------------------------------------------ all is ok, were done */
      last_trans_time = sec_stop = time (NULL);

      sprintf (em,"SUCCESS: %s\n", fn); dbg(em);

      count_transaction(fn);
      update_status(0);
      purge_logs();
      return (0);
   }
}

/*------------------------------------------------------- Send file to server */
int PACS_send (char *fn, int mode)
{  FILE *fp;
   int rc;
   struct timeval timeval = {60, 0};	
   fd_set esock, wsock;
   char *ptr;

   if (!sock) { dbg("PACS_send: no socket"); return(1); }

   memset(IDXrad_buffer,0,sizeof(IDXrad_buffer));

   fp = fopen (fn, "rb");
   if (!fp) {ldbg("PACS_send: Couldn't open file"); return (2);}
   rc = fread (IDXrad_buffer, 1, sizeof (IDXrad_buffer), fp);
   fclose (fp);

/*------------------------------ check for bad file, 0 bytes, no <SB> or <EB> */
   if (!rc) {ldbg("PACS_send: 0 byte file"); return (2);}
   if (!strchr(IDXrad_buffer,SB) || !strchr(IDXrad_buffer,EB)) 
   {ldbg("PACS_send: wrapper byte(s) missing"); return (2); }

   e[XMIT].basis = time(NULL);
   while (1)
   {  if (!sock) break;
      FD_ZERO(&wsock); FD_SET(sock, &wsock);
      FD_ZERO(&esock); FD_SET(sock, &esock);
/*------------------------------------------------------- see if any activity */
      if(!select (FD_SETSIZE, NULL, &wsock, &esock, &timeval))
/*----------------------------------- if no activity in 15 seconds, try again */
      {	 dbg("TX WAIT");
         link_error_check(XMIT);
         if (mode == NO_WAIT) return (1);
         continue;
      }
/*---------------------------------------------------------- check for errors */
      if (FD_ISSET (sock, &esock))
      {	 cleanup (&sock);
         return (1);
      }
/*----------------------------------------------------- see if ready to write */
      if (FD_ISSET (sock, &wsock))
      {	 dbg("PACS_send: socket ready for write");
         if (write_pacs ()) return (1);
         else return (0);
      }
   }
   return (1);
}

/*--------------------------------------------------- receive ack from server */
static int PACS_recv (int mode)
{  int rc, rx_count = 2;
   struct timeval timeval = {15, 0};	
   unsigned int beg_clock, end_clock;
   fd_set esock, rsock;

   if (!sock) { dbg("PACS_recv no socket"); return(1); }

   e[RECV].basis = time(NULL);
   while (rx_count--)
      /*----------------------------------------------- identify for read and error */
   {  if (!sock) break;
      FD_ZERO(&rsock); FD_SET(sock, &rsock);
      FD_ZERO(&esock); FD_SET(sock, &esock);
      /*------------------------------------------------------ see if any activity */
      if(!select (FD_SETSIZE, &rsock, NULL, &esock, &timeval))
      {	 dbg("RX WAIT:");
         link_error_check(RECV);
         if (mode == NO_WAIT) return (1);
         continue;
      }
      /*--------------------------------------------------------- check for errors */
      if (FD_ISSET(sock,&esock))
      {	 cleanup (&sock);
         return(1);
      }
      /*----------------------------------------------------- see if ready to read */
      if (FD_ISSET(sock,&rsock))
      {	 if (read_pacs()) return (1);
         if (need_more) continue;
         return(0);
      }
   }
   return (1);
}

static char read_buffer[MAX_TCP_BUFFER];

static int read_pacs (void)
{  int rc;
   char *ptr_SB, *ptr_EB;
   int len_pacs, len_read, len_total;

   /*--------------------------- see if anything there or they dropped the line */
   memset (read_buffer, 0, sizeof (read_buffer));

   rc = recv (sock, read_buffer, sizeof (read_buffer), MSG_PEEK);
   if (!rc) {dbge("receive 1"); cleanup (&sock); return (1); }

   memset (read_buffer, 0, sizeof (read_buffer));

   rc = recv (sock, read_buffer, sizeof (read_buffer), 0);
   sprintf(em,"recv returned %d", rc); dbg(em);
   if (rc <= 0) {  dbge("receive 2"); cleanup (&sock); return (1); }
   e[CONNECTION].basis = time(NULL);
   /*---------------------------------------------- make sure a place to put it */
   ptr_SB = strchr (read_buffer, SB);
   if (ptr_SB)
   {  memset ((char *) &PACS_buffer, 0, sizeof (PACS_buffer));
      strcpy (PACS_buffer, read_buffer);
   }
   else strcat (PACS_buffer, read_buffer);

   ptr_EB = strchr (read_buffer, EB);
   if (ptr_SB) {dbg("read complete"); need_more = 0; return (0);}
   else {dbg("read not complete"); need_more = 1; return (0);}
}

/*-----------------------------------------------------------------------------
 * Send message; return 0=OK, 1=error
 *---------------------------------------------------------------------------*/
static int write_pacs (void)
{  int rc;
   dbg("write_pacs");
   rc = send (sock, IDXrad_buffer, strlen (IDXrad_buffer), 0);
   if (rc < 0)
   {  dbge("write_pacs");
      cleanup (&sock);
      return (1);
   }
   sprintf(em,"write_pacs: send returned %d",rc); dbg(em);
   e[CONNECTION].basis = time(NULL);
   return (0);
}

static int cleanup (int *open_sock)
{  int rc, sock = *open_sock;
   last_disconnect_time = time(NULL); 
   switch (*open_sock)
   {  case -1:
      case 0:
         *open_sock = 0;
         return (0);
      default:
         sprintf(em,"Close: %d", *open_sock); dbg(em);
         *open_sock = 0;
         break;
   }
   /*------------------------------------ shut down socket if it was connected. */
   if (sock)
   {  rc = shutdown (sock, 2);
      if (rc == -1) dbge("shutdown");
   }
   /*------------------------------------------------------------- close socket */
   rc = close (sock);
   if (rc) dbge("close");
   return (0);
}

/*----------------------------------------------------------------------
 * Update status info to show latest transaction.
 *--------------------------------------------------------------------*/
void count_transaction(char * fn)
{  time_t now = time(NULL);
   struct tm * tm = localtime(&now);
   if (transaction_day != tm->tm_yday)	/* Reset counters if needed */
   {  transaction_day = tm->tm_yday;
      transaction_hour = -1;
      transactions_this_day = 0;
   }
   if (transaction_hour != tm->tm_hour)
   {  transaction_hour = tm->tm_hour;
      transactions_this_hour = 0;
   }
   transactions_this_day++;		/* Increment counters */
   transactions_this_hour++;

   strcpy(msg1,msg2);			/* Scroll back messages */
   strcpy(msg2,msg3);
   sprintf(msg3, "Sent %s at %s", fn, asctime(localtime(&last_trans_time)));
}

/***********************************************************************
 ***********************************************************************
 * Periodically updates connection status to a file, 
 ***********************************************************************
 **********************************************************************/
static time_t last_status_update = 0;

void update_status(int force)
{  int i;
   time_t now;
   struct tm * tm;
   char tim[35];
   FILE * fp;
   /*--------------------------------------------------- Is it time yet? */
   now = time(NULL);
   if (!force && (now - last_status_update) < status_update_seconds) return;
   last_status_update = now;
   /*------------------------------------------------- Clear update file */
   delete(status_pfn);
   fp = fopen(status_pfn, "wb");
   if (!fp) 
   {  sprintf(em,"Unable to create status file: %s", status_pfn); ldbg(em);
      return;
   }
   /*------------------------------------------------------------ Header */
   fprintf(fp, "PACS Out server: %s, current status: %s", process_name, asctime(localtime(&now)));
   fprintf(fp, "----------------------------------------------------------------\n");
   fprintf(fp, "Link started: %s", asctime(localtime(&interface_started_at)));
   if(interface_bounced_at)fprintf(fp, "Link last bounced: %s", asctime(localtime(&interface_bounced_at)));
   if(last_purge_date)
      fprintf(fp, "Last log file purge: %s   %d log files purged\n   %d err files purged\n", 
            asctime(localtime(&last_purge_date)), files_purged, efiles_purged);
   /*------------------------------------------------------------ Listen socket */
   fprintf(fp, "Server: %s, port %d\n", peer_ip, port);
   if (sock) 
      fprintf(fp, "Connected to Broker: %s, at %s", peer, asctime(localtime(&last_connect_time)));
   else if (last_disconnect_time) 
      fprintf(fp, "Disconnected from Broker: %s, at %s", peer, asctime(localtime(&last_disconnect_time)));
   else fprintf(fp, "Never connected to Broker\n");
   fprintf(fp, "Connections: %d\n", connect_count);

   if (msg1 || msg2 || msg3) fprintf(fp, "Most recent messages:\n");
   else fprintf(fp, "No messages processed.\n");
   if (msg1) fprintf(fp, msg1);
   if (msg2) fprintf(fp, msg2);
   if (msg3) fprintf(fp, msg3);
   fprintf(fp, "Transactions: today %d, this hour %d\n", transactions_this_day, transactions_this_hour);
   fclose(fp);	/* Done */
}

void shutdown_connection_status(void)
{  time_t now;
   FILE * fp;
   /*--------------------------------------------------- Is it time yet? */
   now = time(NULL);
   /*------------------------------------------------- Clear update file */
   delete(status_pfn);
   fp = fopen(status_pfn, "wb");
   if (!fp) {printf("Unable to create status file: %s\n", status_pfn); return;}
   /*------------------------------------------------------------ Header */
   fprintf(fp, "PACS Inbound Link shut down at: %s", asctime(localtime(&now)));
   fclose(fp);	/* Done */
}

void write_log(char *msg)
{  time_t now;
   struct tm *tm;
   FILE *fp; 

   now = time(NULL);
   tm = localtime(&now);
   sprintf(log_pfn, log_pfn_format, log_dir, tm->tm_year + 1900, tm->tm_mon +1);
   fp = fopen(log_pfn, "a");
   if (!fp) 
   {  printf("LOG FILE I/O ERR: %s on %s\n", strerror(errno), log_pfn);
      return;
   }
   fprintf(fp, "%04d/%02d/%02d %02d:%02d:%02d %s\n", tm->tm_year + 1900, tm->tm_mon +1,
         tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec, msg);
   fclose(fp);
}

void log_error(char *msg)
{  time_t now;
   struct tm *tm;
   FILE *fp; 

   now = time(NULL);
   tm = localtime(&now);
   sprintf(elog_pfn, elog_pfn_format, log_dir, tm->tm_year + 1900, tm->tm_mon +1);
   fp = fopen(elog_pfn, "a");
   if (!fp) 
   {  printf("LOG FILE I/O ERR: %s on %s\n", strerror(errno), log_pfn);
      return;
   }
   fprintf(fp, "%04d/%02d/%02d %02d:%02d:%02d %s\n", tm->tm_year + 1900, tm->tm_mon +1,
         tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec, msg);
   fclose(fp);
}


void prt_log(char *msg, char * buff, char type)
{  time_t now;
   struct tm *tm;
   FILE *fp; 
   char * ext[] = {"LOG","ERR"};
   char * text;

   if (!msg && !buff) return;
   if (!msg) msg = "\0";
   text = malloc(strlen(msg) + 24);
   
   now = time(NULL);
   tm = localtime(&now);
   
   if (type) type = 1;
   sprintf(log_pfn, log_pfn_format, log_dir, tm->tm_year + 1900, tm->tm_mon +1, ext[type]);
   fp = fopen(log_pfn, "a");
   if (!fp) 
   {  printf("LOG FILE I/O ERR: %s on %s\n", strerror(errno), log_pfn);
      return;
   }
   
   sprintf(text, "%04d/%02d/%02d %02d:%02d:%02d %s\n", tm->tm_year + 1900, tm->tm_mon +1,
         tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec, msg);
   fprintf(fp,text); dbg(text);
   free (text);
   
   if (buff)
   {  char p[70], *ptr;
	   int piece = sizeof(p) - 3, cont = 1;
      fprintf(fp,"--------------------\n");
	   while (cont)
	   {	memset(p,0,sizeof(p)); strcpy(p,"> ");
		   if (strlen(buff) < piece) { strcat(p,buff); cont = 0; }
		   else { memcpy(p+2,buff,piece); buff += piece; }
		   if (ptr = strchr(p,SB)) *ptr = '.';
		   if (ptr = strchr(p,EB)) *ptr = '.';
		   while (ptr = strchr(p,'\r')) *ptr = '~';
		   fprintf(fp,"%s\n",p);
	   }
	   fprintf(fp," ");
   }
   fclose(fp);
   return;
}

/*----------------------------------------------------------------------
 * Checks to see if purge has been run yet today.  If not, runs it, and
 * purges log files older than retain days.  This routine is run at a
 * timeout, so will lag during busy times.
 *--------------------------------------------------------------------*/
void purge_logs(void)
{  time_t now;
   struct tm tdy, *lp;
   char pfn[128], cutoff_pfn[128], *p;
   int month, year;
   DIR * d;
   struct dirent * e;
/*------------------------------------- Determine if we need to purge */
   now = time(NULL);
   memcpy(&tdy,localtime(&now),sizeof(tdy));
   lp = localtime(&last_purge_date);
   if (tdy.tm_year == lp->tm_year && tdy.tm_yday == lp->tm_yday) return;
   if (tdy.tm_hour < log_purge_hour) return;
   last_purge_date = now;     /* We are going to purge; Update last purge date */
/*----------------------------------------------------------------------
 * First purge is message logs.
 * Set up wildcard pfn for lookup, purge cutoff pfn for comparison 
 *--------------------------------------------------------------------*/
   now = time(NULL);
   lp = localtime(&now);
   year = lp->tm_year + 1900; month = lp->tm_mon + 1 - log_retain_mons;
   while (month < 1) { year--; month += 12; }
   sprintf(cutoff_pfn, log_pfn_format, "\0", year, month, "LOG");
   sprintf(em,"Purging log & err files before %4d%02d", year, month); prt_log(em,NULL,LOG);
/*------------------------------ Pass all files, purging those before cutoff */
   files_purged = 0;
   if (d=opendir(log_dir))
   {  while (e=readdir(d))
      {  p = e->d_name;
         if (!strstr(p,".LOG")) continue;
         if (strcmp(p, cutoff_pfn) < 0)
         {  sprintf(em,"Purging file %s", p); prt_log(em,NULL,LOG);
            strcpy(pfn,log_dir); strcat(pfn,p); remove(pfn); 
            files_purged++;
         }
      }
      closedir(d);
   }
/*----------------------------------------------- Second purge is error logs */
   sprintf(cutoff_pfn, log_pfn_format, "\0", year, month, "ERR");
/*----------------------- Pass all files, purging those before cutoff */
   efiles_purged = 0;
   if (d=opendir(log_dir))
   {  while (e=readdir(d))
      {  p = e->d_name;
         if (!strstr(p,".ERR")) continue;
         if (strcmp(p, cutoff_pfn) < 0)
         {  sprintf(em,"Purging file %s", p); prt_log(em,NULL,LOG);
            strcpy(pfn,log_dir); strcat(pfn,p); remove(pfn); 
            efiles_purged++;
         }
      }
      closedir(d);
   }
}

/*-----------------------------------------------------------------------------
 * Read ini file
 *---------------------------------------------------------------------------*/
void read_ini(char * feed)
{  char a[1024], sname[24], kname[64];
   int sa = sizeof(a);
   int i, j;
   
/*---------------------------------------------------------------------------------------
 * static structs which are set to default values and modified by .ini file settings are
 * saved here on the first read_ini call, and restored to their initial value on 
 * subsequent calls, which will happen when the server is bounced.
 *-------------------------------------------------------------------------------------*/
   if (!bu_msg_dir) 
   {  bu_msg_dir=malloc(sizeof(msg_dir)); 
      memmove(bu_msg_dir,msg_dir,sizeof(msg_dir)); 
   }
   else memmove(msg_dir,bu_msg_dir,sizeof(msg_dir));
   
   if (!bu_e) 
   {  bu_e=malloc(sizeof(e)); 
      memmove(bu_e,e,sizeof(e)); 
   }
   else memmove(e,bu_e,sizeof(e));

   /*---------------------------------------- Create section name for this feed */
   strcpy(sname,"FEED_"); strcat(sname,feed);
   for(i=0;i<strlen(sname);i++) sname[i] = toupper(sname[i]);  /* Upper case it */

   /*----------- Pull debug for this feed, also checks for existence of section */
   if (!GetIniString(sname, "Debug", "stdout", a, sa, INIFILE))
   {  if (ini_errno == INI_SECNOF) dbg_exit(strcmb(sname," section not found in ini file"), 1);
      else dbg_exit(strcmb("Debug", GetIniErrorString()), 1);
   }
   debug_fp = debug_on(a);

   /*----------------------------------------------- Feed Status (must be "ON") */
   if (!GetIniString(sname, "Status", "OFF", a, sa, INIFILE))
      dbg_exit(strcmb("Status", GetIniErrorString()),1);
   if (strcasecmp(a,"ON")) dbg_exit("Feed Status not 'ON', terminating", 1);

   /*---------------------- Now pull the process type and match to program type */
   if (!GetIniString(sname, "ProcessType", "NONE", a, sa, INIFILE)) 
      dbg_exit(strcmb("ProcessType", GetIniErrorString()),1);
   if (strcasecmp(PROGRAM_TYPE,a))
   {  sprintf(em,"Progam type: %s, Process Type: %s - Terminating", PROGRAM_TYPE, a);
      dbg_exit(em, 1);
   }

/*-------------------------------------------------------- Compute process name, all caps
 * This is PACSO_ followed by the first character of the process type (S for Server)
 * and the feed section name. On VMS this is used as the process ID, and so cannot be
 * longer than 15 characters. On Unix, .prs is added, as the lock is a fifo.
 *-------------------------------------------------------------------------------------*/
   memset(process_name,0,sizeof(process_name));
   strcpy(process_name,"PACSO_");
   process_name[strlen(process_name)] = a[0];   
   strcat(process_name,feed);
   for(i=0;i<strlen(process_name);i++) process_name[i] = toupper(process_name[i]);  
   strcpy(lock_name,process_name);
#ifdef _VMS_MODE_
   if (strlen(process_name) > 15)
   {  sprintf(em,"process name %s too long", process_name); 
      dbg_exit(em, 1);
   }
#else
   strcat(lock_name,".prs");
#endif

   /*---------------------------------------------- Process specific file names */
#ifdef _VMS_MODE_
   sprintf(log_dir, "[.%s_log]", process_name);
   sprintf(control_pfn,"%s_control.dat;1", process_name);
   sprintf(status_pfn,"%s_status.dat;1", process_name);
#else
   sprintf(log_dir, "./%s_log/", process_name);
   sprintf(control_pfn,"%s_control.dat", process_name);
   sprintf(status_pfn,"%s_status.dat", process_name);
#endif


   /*---------------------------------------------------------- Connection info */
   if (!GetIniString(sname, "PeerIP", NULL, a, sa, INIFILE))
      dbg_exit(strcmb("PeerIP", GetIniErrorString()),1);
   strcpy(peer_ip,a);
   if (!GetIniInt(sname, "Port", 0, &port, INIFILE))
      dbg_exit(strcmb("Port", GetIniErrorString()),1);

   /*-------------------------------------------------------------- Directories */
   if (!GetIniString(sname, "HighMsgDir", msg_dir[0], a, sa, INIFILE))
      dbg_exit(strcmb("HighMsgDir", GetIniErrorString()),1);
   strcpy(msg_dir[0],a);
   if (!GetIniString(sname, "LowMsgDir", msg_dir[1], a, sa, INIFILE))
      dbg_exit(strcmb("LowMsgDir", GetIniErrorString()),1);
   strcpy(msg_dir[1],a);
   if (!GetIniString(sname, "LogDir", log_dir, a, sa, INIFILE))
      dbg_exit(strcmb("LogDir", GetIniErrorString()),1);
   strcpy(msg_dir[2],a);

   /*------- Log purge months, control file check secs, status file update secs */
   if (!GetIniInt(sname, "ControlCheckSeconds", control_check_seconds, &control_check_seconds, INIFILE))
      dbg_exit(strcmb("ControlCheckSeconds", GetIniErrorString()),1);
   if (!GetIniInt(sname, "StatusUpdateSeconds", status_update_seconds, &status_update_seconds, INIFILE))
      dbg_exit(strcmb("StatusUpdateSeconds", GetIniErrorString()),1);
   if (!GetIniInt(sname, "LogRetainMonths", log_retain_mons, &log_retain_mons, INIFILE))
      dbg_exit(strcmb("LogRetainMonths", GetIniErrorString()),1);
   if (!GetIniInt(sname, "LogPurgeHour", log_purge_hour, &log_purge_hour, INIFILE))
      dbg_exit(strcmb("LogPurgeHour", GetIniErrorString()),1);

   /*-------------------------------------------------------- Error log printer */
   if (!GetIniString("[MAIN]", "ErrorPrinter", error_printer, a, sa, INIFILE))
      dbg_exit(strcmb("ErrorPrinter", GetIniErrorString()),1);
   strcpy(error_printer,a);
   /*----------------------------- Warning print window (Mon-Fri) 24 hour clock */
   if (!GetIniInt("[MAIN]", "PrintWindowStart", print_window_start, &print_window_start, INIFILE))
      dbg_exit(strcmb("PrintWindowStart", GetIniErrorString()),1);
   if (!GetIniInt("[MAIN]", "PrintWindowEnd", print_window_end, &print_window_end, INIFILE))
      dbg_exit(strcmb("PrintWindowEnd", GetIniErrorString()),1);
   /*----------- Should warning messages be Echoed to e-mail and if so, where? */
   if (!GetIniString("[MAIN]", "EmailWarningMessages", "NO", a, sa, INIFILE))
      dbg_exit(strcmb("EmailWarningMessages", GetIniErrorString()),1);  
   if (STR_EQ(a,"YES")) 
   {  e_mail_warning_messages = 1; 
      if (!GetIniString("[MAIN]", "ErrorEmail", e_mail_user_name, a, sa, INIFILE))
         dbg_exit(strcmb("ErrorEmail", GetIniErrorString()),1);  
      strcpy(e_mail_user_name,a);
      if (!GetIniString(sname, "ErrorEmail", e_mail_user_name, a, sa, INIFILE))
         dbg_exit(strcmb("ErrorEmail", GetIniErrorString()),1);
      strcpy(e_mail_user_name,a);
   }

   /*-------------- Pager number for page errors window (Mon-Fri) 24 hour clock */
   if (!GetIniString("[MAIN]", "Pager", pager1, pager1, sizeof(pager1), INIFILE))
      dbg_exit(strcmb("Pager", GetIniErrorString()),1);
   if (!GetIniString(sname, "Pager", pager1, pager1, sizeof(pager1), INIFILE))
      dbg_exit(strcmb("Pager", GetIniErrorString()),1);
   GetIniString("[MAIN]", "Pager2", pager2, pager2, sizeof(pager2), INIFILE);
   GetIniString(sname, "Pager2", pager2, pager2, sizeof(pager2), INIFILE);
   if (!GetIniInt("[MAIN]", "PagerWindowStart", pager_window_start, &pager_window_start, INIFILE))
      dbg_exit(strcmb("PagerWindowStart", GetIniErrorString()),1);
   if (!GetIniInt("[MAIN]", "PagerWindowEnd", pager_window_end, &pager_window_end, INIFILE))
      dbg_exit(strcmb("PagerWindowEnd", GetIniErrorString()),1);

   /*------------------ IDX down day, start hr and min, and minutes long */
   if (!GetIniInt("[MAIN]", "IDXDownDOW", idx_down_dow, &idx_down_dow, INIFILE))
      dbg_exit(strcmb("IDXDownDOW", GetIniErrorString()),1);
   if (!GetIniInt("[MAIN]", "IDXDownSHr", idx_down_shr, &idx_down_shr, INIFILE))
      dbg_exit(strcmb("IDXDownSHr", GetIniErrorString()),1);
   if (!GetIniInt("[MAIN]", "IDXDownSMin", idx_down_smin, &idx_down_smin, INIFILE))
      dbg_exit(strcmb("IDXDownSMin", GetIniErrorString()),1);
   if (!GetIniInt("[MAIN]", "IDXDownMins", idx_down_mins, &idx_down_mins, INIFILE))
      dbg_exit(strcmb("IDXDownMins", GetIniErrorString()),1);

   /*-------------------------------------------- error warnings/pages settings */
   for (err = 0; err < max_err; err++)
   {  if (!GetIniInt(sname, strcmb("PrintMins",ERR.suffix), ERR.print_mins, &(ERR.print_mins), INIFILE))
      dbg_exit(strcmb("PrintMins", GetIniErrorString()),1);
      if (!GetIniInt(sname, strcmb("PrintInterval",ERR.suffix), ERR.print_interval, &(ERR.print_interval), INIFILE))
         dbg_exit(strcmb("PrintInterval", GetIniErrorString()),1);
      if (!GetIniInt(sname, strcmb("PageMins",ERR.suffix), ERR.page_mins, &(ERR.page_mins), INIFILE))
         dbg_exit(strcmb("PageMins", GetIniErrorString()),1);
      if (!GetIniInt(sname, strcmb("PageInterval",ERR.suffix), ERR.page_interval, &(ERR.page_interval), INIFILE))
         dbg_exit(strcmb("PageInterval", GetIniErrorString()),1);
   }
}

/*----------------------------------------------------------------------
 * Check control file for user commands and carry them out.
 *--------------------------------------------------------------------*/
int control_check(void)
{  time_t current_time;
   FILE * fp;
   char buffer[256], cmd[256];

   /*----------------------------------- Check control file at intervals */
   current_time = time(NULL);
   if (difftime(current_time, last_control_check) < control_check_seconds) return(0);
   last_control_check = current_time;

   /*-------------------------- Open control file, read command (if any) */
   fp = fopen(control_pfn, "r");
   if (!fp) return(0);
   fgets(buffer, sizeof(buffer), fp);
   if (feof(fp))
   {  fclose(fp);
      delete(control_pfn);
      return(0);
   }
   fclose(fp);
   delete(control_pfn);
   sprintf(em,"Control file read: %s", buffer); dbg(em);
   strtok(buffer, " \r\n");

   /*----------------------------------------- Shut down and exit server */
   if(STR_EQ(buffer, "EXIT"))
   {  logg("Control exit");
      main_loop = tcp_loop = 0;
      return(1);
   }
   /*----------------------------------------------------- Bounce server */
   if(STR_EQ(buffer, "BOUNCE"))
   {  logg("Control bounce");
      tcp_loop = 0;
      interface_bounced_at = time(NULL);
      return(1);
   }
   return (0);
}


/*----------------------------------------------------------------------
 * Check for delays in processing on IDX or PACS connection.  Print or
 * page errors as appropriate.
 *--------------------------------------------------------------------*/
void link_error_check(int err)
{  time_t now, base;
   struct tm *tm;
   char msg[4096];
   int interval, minutes, i, j;
   /*------------------------------------------------- Pull current time */
   now = time(NULL);
   tm = localtime(&now);
   /*-----------------------------------------------------------------------------
    * If error is not processed during IDX down time, check to see if we are now
    * in that time, and if so, update basis and return.
    *---------------------------------------------------------------------------*/
   if (!ERR.idx_down)
   {  if (tm->tm_wday == idx_down_dow &&
         ((i = tm->tm_hour * 60 + tm->tm_min)) >= ((j = idx_down_shr * 60 + idx_down_smin)) &&
         i <= (j + idx_down_mins))
   {	 ERR.basis = now;
      dbg("Error not processed, in IDX down period");
      return;
   }
   }
   /*----------------------------------------------------------- print warnings */
   do
   {  if (!ERR.print_mins) break;      /* 0 in print_mins "turns off" prints */
      /*-- If basis is later than last print, we're on a "new" wait, clear old one */
      if (ERR.last_print && (ERR.basis > ERR.last_print)) ERR.last_print = 0;
      /*---------------------------------------------- Are we in the print window? */
      if (print_window_start <= print_window_end)  
      {	 if (tm->tm_hour < print_window_start || tm->tm_hour > print_window_end) break;}
      else 
         if (tm->tm_hour < print_window_start && tm->tm_hour > print_window_end) break; 
      /*-------------- If first print, go from basis, otherwise go from last print */
      if (ERR.last_print) 
      {	 base = ERR.last_print; 
         interval = ERR.print_interval; 
      }
      else 
      {	 base = ERR.basis; 
         interval = ERR.print_mins;
      }
      /*------------------------------------------------- Is it time to print yet? */
      if (now - base < interval * 60) break; 
      minutes = (now - ERR.basis) / 60;			   /* How long total */
      sprintf(em, "sending print message %s, %d minutes", ERR.name, minutes); dbg(em);
      /*------------------------------------------------- generate printed message */
      sprintf(msg, "\f************************************************************\n"
            "* %s"
            "* %s Warning:\n"
            "*   %s\n"
            "*   %d minutes\n"
            "* Contact PACS group\n"
            "************************************************************\n\n\n\n\n"
            , asctime(tm), process_name, ERR.name, minutes);
      if (!alertSendPrint(error_printer, msg, NULL))PrtAlertErrorString(debug_fp, "PACSO Warning");
      if (e_mail_warning_messages) if (!alertSendEmail(e_mail_user_name, "PACSO", msg, NULL))
         PrtAlertErrorString(debug_fp, "PACSO Warning");
      ERR.last_print = now;
   } while (0);

   /*----------------------------------------------------------- Pager warnings */
   do
   {  if (!ERR.page_mins) break;   /* Not paging messages */
      /*-- If basis is later than last print, we're on a "new" wait, clear old one */
      if (ERR.last_page && (ERR.basis > ERR.last_page)) ERR.last_page = 0;
      if (pager_window_start <= pager_window_end)     /* Not in page window */
      {	 if (tm->tm_hour < pager_window_start || tm->tm_hour > pager_window_end) break;}
      else 
         if (tm->tm_hour < pager_window_start && tm->tm_hour > pager_window_end) break; 
      /*-------------- If first print, go from basis, otherwise go from last print */
      if (ERR.last_page) 
      {	 base = ERR.last_page; 
         interval = ERR.page_interval; 
      }
      else 
      {	 base = ERR.basis; 
         interval = ERR.page_mins;
      }
      /*------------------------------------------------- Is it time to print yet? */
      if (now - base < interval * 60) break; 
      minutes = (now - ERR.basis) / 60;			   /* How long total */
      sprintf(em, "sending page %s, %d minutes", ERR.name, minutes); dbg(em);
      /*--------------------------------------------------- Generate pager message */
      sprintf(msg, "%s\n"
            "%s\n"
            "%d minutes\n"
            , process_name, ERR.name, minutes);
      if (!alertSendEmail(pager1, "PACSO", msg, NULL))
         PrtAlertErrorString(debug_fp, "PACSO Warning");
      if (strlen(pager2))
         if (!alertSendEmail(pager2, "PACSO", msg, NULL))
            PrtAlertErrorString(debug_fp, "PACSO Warning");
      ERR.last_page = now;
   } while (0);
}

/*-----------------------------------------------------------------------------
 * Returns pointer to string which is cat of s1 and s2, but s1 can be a 
 * constant.
 *---------------------------------------------------------------------------*/
char * strcmb(char *s1, char *s2)
{  static char ws[128];
   memset(ws,0,sizeof(ws));
   strcpy(ws,s1);
   return strcat(ws,s2);
}

void connect_to_server(void)
{  int tries = 5, i, rc;
   if (sock) return;
   while (tries--)
   {  sprintf(em, "Connect PACSO Client %s, port %d", peer_ip, port); dbg(em);

      dbg("Creating socket");
      sock = socket(AF_INET, SOCK_STREAM, 0);
      if(sock == -1) {dbge("socket"); exit(1);}

      dbg("Get host by IP");
      psrvent = gethostbyname(peer_ip);
      if(!psrvent){dbge("gethost"); cleanup(&sock); exit(1); }

      dbg("Setting socket name entries");
      srvent = *psrvent;
      sock_name.sin_family = srvent.h_addrtype;
      sock_name.sin_port   = htons(port);
      sock_name.sin_addr   = *((struct in_addr *)srvent.h_addr);

      dbg("Connect to server");
      i = sizeof(sock_name);
      rc = connect(sock, (struct sockaddr *)&sock_name, i);
      sprintf(em, "Connect return = %d", rc); dbg(em);
      if (rc) dbge("connect"); 
      if (!rc || errno == EINPROGRESS) { connect_count++; return; }
      cleanup(&sock);
      sleep(60);
   }
}
