/**************************************************************************************************
 * PACS_WARN.C                                                                            R Moulton
 * PACS Warning server.
 * Receives warning messages from Unix servers and sends them to printer queues, e-mails and/or
 * pager e-mails, as appropriate.  Also receives signal requests and sends signals to specified
 * processes. Also receives run requests and executes scripts or programs
 *-------------------------------------------------------------------------------------------------
 * Unix: gcc -g qc_signal_server.c -o qc_signal_server `pkg-config --cflags --libs qc_signal_server`
 *-------------------------------------------------------------------------------------------------
 * 10-23-03 R Moulton Initial Write
 * 02-01-05 R Moulton Added Logging, use of .ini file
 * 04-25-05 R Moulton Added signal messages.
 * 03-15-06 R Moulton Added run process/script.
 *************************************************************************************************/

/*  #define _VMS_MODE_   Include this when compiling in VMS */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <dirent.h>
#include <fcntl.h>
#include <time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <signal.h>

#ifdef _VMS_MODE_

#include <iodef.h>
#include <unixio.h>
#include <ssdef.h>
#include <starlet.h>
#include <descrip.h>

#include "[source.common]queue_file.h"
#include "[source.common]intermail.h"
#include "[source.common]dbg.h"
#include "[source.common]ini.h"
#include "[source.common]log.h"

#else

#include "dbg.h"
#include "ini.h"
#include "log.h"

#endif

/*--------------------------------------------------------------------------------------- Macros */
#define MAKE_STR(A,B) if (A=(char*)malloc(strlen(B)+3)) strcpy(A,B);
#define FREE_STR(A) if (A) { free(A); A=NULL; }
#define STR_EQ(A,B) !strcasecmp(A,B)

/*------------------------------------------------------------------------------------ Constants */
#define SB 11
#define EB 28

#ifdef _VMS_MODE_
#define PROCESS_NAME        "PACS_WARN"
#define MSG_FILE_FORMAT     "[.MSG]MSG%02d.DAT;1"
#define QCS_INI_PFN             "PACS_WARN.INI";
#else
#define PROCESS_NAME        "qc_signal_server.prs"
#define MSG_FILE_FORMAT     "msg%02d.dat"
#define QCS_INI_PFN             "qc_signal_server.ini"
#endif

#define BUFFER_SIZE         10240       /* Buffer sizes */
#define MAX_PRT             2           /* Maximum number of printer queues for a message */
#define MAX_EMAILS          5           /* Maximum number of e-mails for a message */
#define MAX_PRS             2           /* Maximum number of processes for a signal */
#define MAX_PAR             5           /* Max par substitions on a run */

/*----------------------------------------------------- Environment variables, storage, defaults */

static int port = 52226;

static FILE * debug_fp;
static int file_number = 0;

static int message_files = 1;
static int logging = 1;

#ifdef _VMS_MODE_
static char log_dir[256] = "[.log]";
static char log_pfn_format[128] = "LOG;1";
#else
static char log_dir[256] = "./log";
static char log_pfn_format[128] = "%s/%4d%02d%02d.full";
#endif

static char log_pfn[256];
static int log_purge_hour;                            /* hour to purge logs (0-23) */
static time_t last_purge_date;								/* date last purge was done (yyyymmdd) */
static int log_retain_days;							      /* # of days to retain log files */

/*------------------------------ last CheckStudy message (we cut out duplicate ones) */
static time_t check_study_basis = 0;
static pid_t child = 0;                 /* Currently running child process (or 0) */

/*------------------------- Linked list of currently running processes */
struct _prs_
{  pid_t pid;
   char * name;
   char * log_name;
   struct _prs_ * nxt;
};
static struct _prs_ * prs_list = NULL, * prs, *pprs;


/*-------------------------------------------------------------------------- Function prototypes */
void read_ini(void);
void count_transaction(void);           /* Increment transaction counts, displays */
void tcpserver (void);                    /* main TCP/IP server loop */
void accept_connect(void);               /* Accept connection on listen socket */
void process_tcp_read(int slot);         /* read data on socket slot*/
void process_tcp_write(int slot);        /* write data on socket slot */
void clear_socket(int slot);             /* Clear socket table slot*/
void close_listen_socket(void);                /* close listen socket */
int shutdown_link(void);
void dbgm(char *buff);
void ldbgm(char *buff);
void create_listen_socket(void);          /* open PACS listen socket */
void process_message(int slot);
void ldbg(char * msg);
static void wait_pid();

static char ib[BUFFER_SIZE];
/*----------------------------------------------------------------------
 * Socket data structure
 *--------------------------------------------------------------------*/
#define MAX_CONNECTIONS     25       /* Maximum allowable connections */
/*---------------------------------------------- Socket status codes */
#define SS_INACTIVE         0           /* No socket active this slot */
#define SS_READ             1           /* Waiting for input from PACS */
#define SS_WRITE            2           /* Waiting for output to PACS */

struct _sockets_ 
{  int status;                       /* Current socket status */
   int socket;                       /* Socket number */
   char * message_buffer;            /* Completed message */
   char * ip;                        /* IP address for this connection */
};
struct _sockets_ sockets[MAX_CONNECTIONS]; /* PACS inbound sockets */

/*----------------------------------------------------------------------
 * Valid connecting IP addresses (enter -1 for subnet, matches anything)
 *--------------------------------------------------------------------*/
struct _ipa_ {short int addr[4];}; 
static struct _ipa_ ips[] = {
   { 10, 33, 162, -1 },	 
   { 159, 251, -1, -1 },
   { 172, 16, 219, 32 },
   { 172, 16, 219, 40 },
   { 172, 16, 219, 33 },
};
#define MAX_IPS sizeof(ips) / sizeof(struct _ipa_)


static struct timeval timeval = {60, 0};
static int new_socket, listen_sock = 0;
static struct sockaddr_in listen_sock_name;

static int main_loop = 1, tcp_loop = 1;
/***********************************************************************
 ***********************************************************************
 *********************** Main function *********************************
 ***********************************************************************
 **********************************************************************/
int main (int argc, char **argv)
{  int i;
   read_ini();
   unique_process_check (PROCESS_NAME);                          /* Set process ID */
   dbg("PACS Warning Server started");
   while (main_loop)
   {  tcp_loop = 1;
      create_listen_socket();
/*--------------------------------------- Clear PACS connection table */
      for(i=0; i < MAX_CONNECTIONS; i++)
      {  sockets[i].status = SS_INACTIVE;
			sockets[i].socket = 0;
			sockets[i].message_buffer = NULL;
			sockets[i].ip = NULL;
      }
/*----------------------------------------- main loop is in tcpserver */
      sprintf(em,"Calling tcpserver, port %d", port); dbg(em);
      tcpserver();
   }
   dbg("Exiting main loop");
   unique_process_check (NULL);
   exit(0);
}

/***********************************************************************
 ***********************************************************************
 **************** Main TCP server processing loop **********************
 ***********************************************************************
 **********************************************************************/

void tcpserver (void)
{  int rc, i, a;
   fd_set rsock, wsock, esock;
/*----------------------------------------------------------------------
 * Main socket processing loop
 * -------------------------------------------------------------------*/
	while (tcp_loop)
	{	logPurge("MAIN");
      wait_pid();
/*---------------------------------- identify connections for testing */
		FD_ZERO(&rsock);
      FD_ZERO(&wsock);
      FD_ZERO(&esock);
      FD_SET(listen_sock, &rsock);
/*----------------------------------------------------------------------
 * All sockets are set for read and error, those with waiting message 
 * are also set to write
 * -------------------------------------------------------------------*/
      for (i = 0; i < MAX_CONNECTIONS; i++)
      {  if (sockets[i].socket)
			{  FD_SET(sockets[i].socket, &rsock);
				FD_SET(sockets[i].socket, &esock);
            if (sockets[i].status == SS_WRITE) FD_SET(sockets[i].socket, &wsock);
			}
		}
         
/*----------------------------------------------- see if any activity */
      dbg("select");
      rc = select (FD_SETSIZE, &rsock, &wsock, &esock, &timeval);
/*----------------------------------------------------------- timeout */
      if (!rc) continue;
		sprintf(em,"select returned %d", rc); dbg(em); 

/*--------------- activity on listen socket is request for connection */
      if (FD_ISSET (listen_sock, &rsock)) accept_connect();
			
/*------------------------------------------------------ PACS sockets */
      for (i = 0; i < MAX_CONNECTIONS; i++)
		{  if (sockets[i].socket)
			{  if(FD_ISSET(sockets[i].socket, &esock)) { clear_socket(i); continue; }
				if(FD_ISSET(sockets[i].socket, &rsock)) process_tcp_read(i);
				if(FD_ISSET(sockets[i].socket, &wsock)) process_tcp_write(i);
			}
		}
	}
/*--------------------------------------------------- Shutdown server */
   dbg("Exiting tcp loop");
   for (i = 0; i < MAX_CONNECTIONS; i++) clear_socket(i);
   close_listen_socket();
}

/***********************************************************************
 * Setup and open server listen socket 
 **********************************************************************/
void create_listen_socket (void)
{  struct hostent *hostentptr, hostentstruct;
   char hostname[256];
   int rc, on = 1, flag = 0, count = 15;
   size_t sl;

   while (count--)
   {  if(flag) sleep(60); flag = 1;
/*----------------------------------------------------- make a socket */
		if ((listen_sock = socket (AF_INET, SOCK_STREAM, 0)) == -1)
      {	dbge ("listen socket");
			continue;
      }

      if (setsockopt (listen_sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (on)))
      {	dbge ("listen socket setsockopt");
			continue;
      }
/*----------------------------------------------- get local host name */
      rc = gethostname (hostname, sizeof (hostname));
      if (rc) {dbge("gethostname"); close_listen_socket(); continue;}
/*----------------------------- get pointer to network data structure */
      hostentptr = gethostbyname (hostname);
      if (!hostentptr) {dbge("gethostbyname"); close_listen_socket(); continue;}

      hostentstruct = *hostentptr;
      listen_sock_name.sin_family = hostentstruct.h_addrtype;
      listen_sock_name.sin_port = htons(port);
      listen_sock_name.sin_addr = * ((struct in_addr *)hostentstruct.h_addr);
/*------------------------------------------------ bind listen socket */
      sl = sizeof(listen_sock_name);
      rc = bind (listen_sock, (struct sockaddr*)&listen_sock_name, sl);
      if (rc) {dbge("bind"); close_listen_socket(); continue;}
/*-------------------------------------------- listen for connections */
      rc = listen (listen_sock, 5);
      if (rc) {dbge("listen"); close_listen_socket(); continue;}
      sprintf(em, "Listen socket created on port %d", port); dbg(em);
      return;
   }
   ldbg("Unable to create listen socket"); 
   unique_process_check (NULL);
   exit(1);
}

/*----------------------------------------------------------------------
 * Accept new connection on listen socket.  
 *--------------------------------------------------------------------*/
void accept_connect (void)
{  int rc, i, j, flag, on = 1;
   unsigned char *ptr;
   size_t sl;
   char alpha_ip[20];
/*------------------------------------------------- Accept connection */
   sl = sizeof (listen_sock_name);
   new_socket = accept (listen_sock, (struct sockaddr*)&listen_sock_name, &sl);
   if (new_socket == -1) { dbge ("accept"); return; }

/*------- If this socket number already in the table, clear its entry */
   for (i = 0; i < MAX_CONNECTIONS; i++)
      if(new_socket == sockets[i].socket) clear_socket(i);

   rc = setsockopt (new_socket, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof (on));
   if (rc == -1) dbge ("setsockopt");

   ptr = (unsigned char*)&listen_sock_name.sin_addr;
   memset(alpha_ip,0,sizeof(alpha_ip));
   sprintf(alpha_ip,"%d.%d.%d.%d",ptr[0],ptr[1],ptr[2],ptr[3]);

/*------------------------------------ Validate IP against ips table */
   for (i = 0; i < MAX_IPS; i++)
   {  flag = 1;
      for (j = 0; j < 4; j++)
         if (ips[i].addr[j] != -1 && ips[i].addr[j] != ptr[j]) { flag = 0; break; }
      if (flag) break;
   }
   if (!flag)
   {  sprintf(em, "Attempt to connect from: %s rejected", alpha_ip); ldbg(em);
      close(new_socket);
      return;
   }
/*---------------------------------------------------------------------
 * Then we look for an empty slot to put the new connection in. Failing 
 * this, we have exceeded connections limit
 * ------------------------------------------------------------------*/
   j = -1;
   for (i = 0; i < MAX_CONNECTIONS; i++)
   {  if (sockets[i].status == SS_INACTIVE) { j = i; break; }
   }
   if (j == -1)
   {  ldbg("Connection limit exceeded");
      sprintf(em, "Attempt to connect from: %s rejected", alpha_ip); ldbg(em);
      close(new_socket);
      return;
   }
/*---------------------------------------------------------------------
 * Slot j is our slot, initialize and set ready to read.
 *-------------------------------------------------------------------*/
   if (sockets[j].status != SS_INACTIVE) clear_socket(j);
   sockets[j].socket = new_socket;
   sockets[j].message_buffer = (char *)malloc(BUFFER_SIZE);
   MAKE_STR(sockets[j].ip,alpha_ip)
   sprintf (em,"Connected to: %s", alpha_ip); dbg(em);
/*---------------------------------------- Make sure malloc's worked */
   if (!sockets[j].message_buffer || !sockets[j].ip)
   {  ldbg("Memory allocation error: new socket");
      unique_process_check (NULL);
      exit(1);
   }
/*---------------------------- Set ready to read, clear input buffer */
   sockets[j].status = SS_READ;
   memset(sockets[j].message_buffer,0,BUFFER_SIZE);
   return;
}


/*----------------------- Close out socket entry at param table slot */
void clear_socket(int i)
{  if (sockets[i].socket) 
   {  shutdown(sockets[i].socket,2); 
      close(sockets[i].socket); 
   }
   FREE_STR(sockets[i].message_buffer);
   FREE_STR(sockets[i].ip);
   sockets[i].status = SS_INACTIVE;
   sockets[i].socket = 0;
}

/***********************************************************************
 ***********************************************************************
 * Process TCP/IP read on PACS socket slot i
 ***********************************************************************
 **********************************************************************/
void process_tcp_read(int i)
{  int j, feed, facility, rc;
   char * sb_ptr, * eb_ptr;
   char id[128], * id_ptr, buffer[BUFFER_SIZE];
/*----------------------------------------------------------------------
 * If we are not waiting for input, clear out current message
 * ------------------------------------------------------------------ */
   if (sockets[i].status != SS_READ)
   {  dbg ("PACS socket reset");
      memset(sockets[i].message_buffer,0,BUFFER_SIZE);
      sockets[i].status == SS_READ;
   }
/*-------------------------------------------------------- read input */
   memset (buffer, 0, BUFFER_SIZE);
   rc = recv (sockets[i].socket, buffer, BUFFER_SIZE, 0);
   if (rc == 0){dbg("PACS socket closed by user"); clear_socket(i); return;}
   if (rc < 0)
   {  sprintf(em,"PACS socket error %d on read", rc); ldbg(em); 
      dbge("PACS receive"); clear_socket(i); return;
   }
   sprintf(em,"read on PACS socket %d, returned %d",sockets[i].socket,rc); dbg(em);
/*----------------------------------------------------------------------
 * At beginning of message, wait for SB; Otherwise, if an SB is found, 
 * dump whatever message we have and restart with the SB
 * -------------------------------------------------------------------*/
   sb_ptr = strchr (buffer, SB);
   if (!strlen(sockets[i].message_buffer) && !sb_ptr) return;
   if (sb_ptr)
   {  memset(sockets[i].message_buffer,0,BUFFER_SIZE);
      strcpy(sockets[i].message_buffer, sb_ptr);
   }
/*----------------------------------------------------------------------
 * Check for overflow; if found, clear buffer and exit; wait for next
 * -------------------------------------------------------------------*/
   else if (strlen(sockets[i].message_buffer) + strlen(buffer) > BUFFER_SIZE)
   {  memset(sockets[i].message_buffer,0,BUFFER_SIZE);
      return;
   }
/*-------------------------------------------------------- Copy it in */
   else strcat(sockets[i].message_buffer, buffer);

/*------------------------------------------- see if there is 28 <EB> */
   eb_ptr = strchr(sockets[i].message_buffer, EB);
   if (!eb_ptr) return;
/*----------------------------------------------------------------------
 * We may have a complete message.  Check for SB before EB 
 *--------------------------------------------------------------------*/
   sb_ptr = strchr(sockets[i].message_buffer, SB);
   if (!sb_ptr || (sb_ptr > eb_ptr))
   {  memset (sockets[i].message_buffer, 0, BUFFER_SIZE);
      return;
   }
   ldbgm(sockets[i].message_buffer);
/*----------------------------------------------------------------------
 * We now have a completed message; Process it
 *--------------------------------------------------------------------*/
   process_message(i);
}

/*----------------------------------------------------------------------
 * Process message received on socket slot.
 * Messages are "&" separated fields, of the form:
 *		<SB>message
 *		print=printerqueue           (queue message to this printer queue)
 *		e-mail=e-mailaddress         (e-mail to this address)
 *		process=processname          (process to send e-mail to)
 *		title=titleline              (e-mail title, if not blank)
 *		signal=signalname            (signal to send to processes)
 *		msg=messagetext              (message to be printed/e-mailed)
 *		<EB>
 *	
 *	A single message may be sent to up to MAX_PRT printer queues and up to
 *	MAX_EMAILS e-mail addresses.  If there are multiple title fields, the 
 *	first one will be used.  The msg field must be the last one in the 
 *	message.  messagetext is the text of the message, including needed
 *	formatting characters.  Ampersands in the messagetext are considered 
 *	to be part of the message, not field delimiters.  messagetext ends at 
 *	the <EB> The server responds to the client with <SB>ACK<EB> if the 
 *	message format was valid, with <SB>NAK<EB> if there was a formatting 
 *	problem with the message. Validation of e-mail addresses, printer 
 *	queue names, and message and title format is the responsibility of the
 *	client.
 *---------------------------------------------------------------------*/
void process_message(int slot)
{	char * printer_queue[MAX_PRT] = { NULL };
	char * e_mail_address[MAX_EMAILS] = { NULL };
   char * processes[MAX_PRS] = { NULL };
   char * run_name = NULL;
	char * title_line = NULL;
   char * signal_name = NULL;
	char * message_text = NULL;
	char * ack = "\013ACK\034";
	char * nak = "\013NAK\034";
	char * p, * q, * message = NULL;
	int i, rc, pc = 0, ec = 0, qc = 0;
	FILE * fp;
	char pfn[128];
	
/*-------------------------------------------- Pull message, load nak */
	MAKE_STR(message,sockets[slot].message_buffer)
	strcpy(sockets[slot].message_buffer,nak);
	p = message;

	if (*p == SB) p++; else goto finish;    /* Doesn't start with <SB> */

	p = strtok(p,"&");                               /* Pull "message" */
	if (!p || strcasecmp(p,"message")) goto finish;

/*-------------------------------- This loop processes all the fields */
	while (1)
	{	p = strtok(NULL, "=");             /* Pull next field type code */
		if (!p) goto finish;
/*------------------------------------------------ print=printerqueue */
		if (STR_EQ(p,"print"))
		{	p = strtok(NULL,"&");
			if (!p) goto finish;
			if (pc < MAX_PRT)
			{	MAKE_STR(printer_queue[pc],p);
				pc++;
				continue;
			}
		}
/*---------------------------------------------------- e-mail=address */
		if (STR_EQ(p,"e-mail"))
		{	p = strtok(NULL,"&");
			if (!p) goto finish;
			if (ec < MAX_EMAILS)
			{	MAKE_STR(e_mail_address[ec],p);
				ec++;
				continue;
			}
		}
/*---------------------------------------------------- process=prs_name */
		if (STR_EQ(p,"process"))
		{	p = strtok(NULL,"&");
			if (!p) goto finish;
			if (ec < MAX_PRS)
			{	MAKE_STR(processes[qc],p);
				qc++;
				continue;
			}
		}
/*----------------------------- title=titleline (only first one used) */
		if (STR_EQ(p,"title"))
		{	p = strtok(NULL,"&");
			if (!p) goto finish;
			if (!title_line) MAKE_STR(title_line,p)
			continue;
		}
/*----------------------------- signal=signal_name (only first one used) */
		if (STR_EQ(p,"signal"))
		{	p = strtok(NULL,"&");
			if (!p) goto finish;
			if (!signal_name) MAKE_STR(signal_name,p)
			continue;
		}
/*------------------------------------------------- msg=message_text */
		if (STR_EQ(p,"msg"))
		{	p = strtok(NULL,"\034");
			if (!p) goto finish;
			MAKE_STR(message_text,p);
			break;
		}
/*--------------------------------------------- run=process name to run */
      if (STR_EQ(p,"run"))
      {  p = strtok(NULL,"&");
         if (!p) goto finish;
         MAKE_STR(run_name,p)
         break;
      }
	}
/*----------------------------------------------- run script or process */
   if (run_name)
   {  char a[1024], sname[64] = "RUN_", dname[64] = "Dir_";
      int sa = sizeof(a);
      char name[64];
      char dir[64];
      char * p;
      char log_name[64];
      if (p=strchr(run_name,' ')) *p = 0;       /* Trim trailing spaces */
      strcpy(name,run_name);                 /* default prs/script name */
      p = run_name; while (*p) *p++ = toupper(*p);          /* all caps */
      strcat(sname,run_name);                    /* run section in .ini */
      strcpy(log_name,run_name);       /* this is default log name also */
/*------------------------------------------------------------------------
 * script/process name from .ini file - default is run name from msg
 *----------------------------------------------------------------------*/
      if (!GetIniString(sname, "Name", name, name, sizeof(name), QCS_INI_PFN)) goto finish;
/*------------------------------------------------------------------------
 * home directory - default is "home", then looked up in "[RUN] section
 *----------------------------------------------------------------------*/
      if (!GetIniString(sname, "Dir", "home", dir, sizeof(dir), QCS_INI_PFN)) goto finish;
      strcat(dname,dir);
      if (!GetIniString("[RUN]", dname, dir, dir, sizeof(dir), QCS_INI_PFN)) goto finish;
/*------------------------------ Log name defaults to run name from msg */
      if (!GetIniString(sname, "Log", log_name, log_name, sizeof(log_name), QCS_INI_PFN)) goto finish;
/*------------------------------------------------------------------------
 * Handle the fork.  Run the script or program in the child process, 
 * return in the main process.  Set flag to do a wait pid later to avoid 
 * zombie processes.
 *----------------------------------------------------------------------*/
      logPurge(log_name);
      sprintf(em,"running %s - %s", run_name, name);
      logNoteLevel(log_name,em, 1);
      
      child = fork();
/*-------------------------------------------------------- fork failed */
      if (child == -1) 
      {  sprintf(em, "Attempt to run %s, fork failed error %d - %s", 
            name, errno, strerror(errno));
         logNote(log_name, em);
         child = 0;
         goto finish;
      }
/*------------------------- fork succeeded, parent adds process to list */
      if (child) 
      {  prs = malloc(sizeof(struct _prs_));       /* make new link struct */
         if (!prs) dbg_exit("malloc failure", 1);
         prs->pid  = child;
         prs->name = strdup(run_name);
         prs->log_name = strdup(log_name);
         prs->nxt  = NULL;
/*------------------------------------------------- change nak to ack */
	      strcpy(sockets[slot].message_buffer,ack);

         if (!prs_list) { prs_list = prs; goto finish; } /* first link in list */
/*-------------------------------- find eo list and add this link there */
         for (pprs = prs_list; pprs->nxt; pprs = pprs->nxt) ;
         pprs->nxt = prs;
         
         goto finish;
      }
/*------------------------------------------------------------------------
 * child: make selected directory current, then exec 
 *----------------------------------------------------------------------*/
      if (chdir(dir) == -1)
      {  sprintf(em, "Attempt to run %s, chdir failed error %d - %s", 
            name, errno, strerror(errno));
         logNote(log_name, em);
         exit(1);
      }
      sprintf(em,"Running %s", name); dbg(em);
      execlp(name, name, NULL);
/*-------------------------------------- Only way you get here is error */
      sprintf(em, "Attempt to run %s, execlp failed error %d - %s", 
         name, errno, strerror(errno));
      logNote(log_name, em);
      exit(1);
   }

	if (!pc && !ec && !qc) goto finish;
	if (!title_line) MAKE_STR(title_line,"\0");
   if (!signal_name) MAKE_STR(signal_name,"SIGUSR1");

/*------------------------------- We cut out duplicate CheckStudy msgs */
   if (STR_EQ(title_line,"CheckStudy"))
   {  time_t now = time(NULL);
      if ((now - check_study_basis) > 6300) check_study_basis = now;
      else 
      {  sprintf(em,"Only %d minutes since last CheckStudy, skipping", (now - check_study_basis)/60); dbg(em);
         goto finish;
      }
   }
/*---------------------------------------- Write message text to file */
   if (message_files)
   {  if (file_number++ > 25) file_number = 1;
	   sprintf(pfn, MSG_FILE_FORMAT, file_number);
	   remove(pfn);
	   fp = fopen(pfn, "wb");
	   if (!fp)
	   {	sprintf(em,"Could not open %s", pfn); ldbg(em);
		   goto finish;
	   }
	   i = fputs(message_text,fp);
	   if (i < 0)
	   {	sprintf(em,"Write to %s failed", pfn); ldbg(em);
		   fclose(fp);
		   goto finish;
	   }
	   fclose(fp);
   }

/*------------------------------------ e-mail and printer only implemented in VMS */
#ifdef _VMS_MODE_
/*------------------------------------------------- queue to printers */
	for (i=0; i < pc; i++) 
	{	
		rc = queue_file(printer_queue[i], pfn, 0, 0);
		sprintf(em,"queue_file queue=%s, pfn=%s, rc=%d\n", printer_queue[i], pfn, rc); dbg(em);
		sprintf(em,"queued to %s, rc=%d", printer_queue[i], rc); logNoteLevel("MAIN",em,2);
	}
/*--------------------------------------------------- send to e-mails */
	for (i=0; i < ec; i++) 
   {  rc = mail_message(e_mail_address[i], title_line, pfn); 
      sprintf(em,"e-mail %s, title %s, pfn %s, rc=%d\n", e_mail_address[i], title_line, pfn, rc); dbg(em);
      sprintf(em,"e-mailed to %s, rc=%d", e_mail_address[i], rc); logNoteLevel("MAIN",em,2);
   }
#endif
   
/*-------------------------------------------------- generate signals */
   dbg("generating signals");
   dbg(signal_name);
   for (i=0; i < qc; i++)
   {  FILE * cmd;
      p = tmpnam(NULL);
      sprintf(em,"ps -u spectrum -o pid,comm | grep \"%s\" | cut -c1-6 > %s", processes[i], p);
      dbg(em); system(em);
      if (!access(p,F_OK) && (cmd = fopen(p,"r")))
      {  while (q = fgets(em, sizeof(em), cmd))
         {  dbg(em);
            pid_t pid = atoi(em);
            /* Add other signals as needed */
            if(STR_EQ(signal_name,"SIGUSR1")) 
            {  dbg("sending sigusr1");
               kill(pid,SIGUSR1);
               sprintf(em,"Sent %s signal to %d-%s", signal_name, pid, processes[i]); 
               dbg(em); logNoteLevel("MAIN",em,2);
            }
            if(STR_EQ(signal_name,"SIGUSR2")) 
            {  kill(pid,SIGUSR2);
               sprintf(em,"Sent %s signal to %d-%s", signal_name, pid, processes[i]); 
               dbg(em); logNoteLevel("MAIN",em,2);
            }
         }
         fclose(cmd);
         remove(p);
      }
   }
/*------------------------------------------------- change nak to ack */
	strcpy(sockets[slot].message_buffer,ack);

/*---------------------------------------------------------- Clean up */
finish:
	sockets[slot].status = SS_WRITE;
/*------------------------------------------------------- Free memory */
	for (i=0; i < pc; i++) FREE_STR(printer_queue[i])
	for (i=0; i < ec; i++) FREE_STR(e_mail_address[i])
	FREE_STR(title_line)
	FREE_STR(message_text)
	FREE_STR(message)
	return;
}
/***********************************************************************
 ***********************************************************************
 * Process PACS TCP/IP write
 ***********************************************************************
 **********************************************************************/
static char tb[BUFFER_SIZE];
void process_tcp_write(int i)
{  int ml, rc;
   if (sockets[i].status != SS_WRITE) return;
   dbgm(sockets[i].message_buffer);
   memset(tb,0,BUFFER_SIZE);
   strcpy(tb,sockets[i].message_buffer);
   ml = strchr(tb, EB) - strchr(tb, SB) +2;
   rc = send(sockets[i].socket, tb, ml, 0);
   if (rc < 0)
   {  dbge ("PACS send");
      clear_socket(i);
      return;
   }
   sprintf(em, "%d characters sent to PACS", rc); dbg(em);
   memset(sockets[i].message_buffer,0,BUFFER_SIZE);
   sockets[i].status = SS_READ;
}

void close_listen_socket(void)
{  if (listen_sock <= 0) return;
   if (shutdown(listen_sock,2)) {dbge("listen socket shutdown"); return;}
   if (close(listen_sock)) dbge("listen socket close");
   return;
}

/*----------------------------------------- Debug to dump buffer area */
void ldbgm(char * buff) 
{  dbgm(buff); 
   logMsg("MAIN",buff); 
}

void dbgm(char*buff)
{	char p[53], *ptr;
	int piece = sizeof(p) - 3, cont = 1;
	if (!debug_is_on()) return;
  
	if (!buff) { dbg("NULL buffer\n"); return; }
	while (cont)
	{	memset(p,0,sizeof(p)); strcpy(p,"> ");
		if (strlen(buff) < piece) { strcat(p,buff); cont = 0; }
		else { memcpy(p+2,buff,piece); buff += piece; }
		if (ptr = strchr(p,SB)) *ptr = '.';
		if (ptr = strchr(p,EB)) *ptr = '.';
		while (ptr = strchr(p,'\r')) *ptr = '~';
		dbg(p);
	}
	dbg(" ");
}

/*-----------------------------------------------------------------------------
 * Returns pointer to string which is cat of s1 and s2, but s1 can be a 
 * constant.
 *---------------------------------------------------------------------------*/
char * strcmb(char *s1, char *s2)
{  static char ws[128];
   memset(ws,0,sizeof(ws));
   strcpy(ws,s1);
   return strcat(ws,s2);
}

void read_ini(void)
{  char a[1024], sname[] = "[MAIN]";
   int sa = sizeof(a);

/*------------------------------------------------------------------------------ debug */
   if (!GetIniString(sname, "Debug", "stdout", a, sa, QCS_INI_PFN))
      dbg_exit(strcmb("Debug", GetIniErrorString()), 1);
   debug_fp = debug_on(a);
/*------------------------------------------------- Generate message files? default ON */
   GetIniString(sname,"MessageFiles", "ON", a, sa, QCS_INI_PFN);
   if (STR_EQ(a,"OFF") || STR_EQ(a,"NO")) message_files = 0;
   else message_files = 1;
/*---------------------------------------------------------------------------- Logging */
   GetIniString(sname,"Logging", "ON", a, sa, QCS_INI_PFN);
   if (STR_EQ(a,"OFF") || STR_EQ(a,"NO")) logging = 0;
   else logging = 1;
   
   if (!GetIniString(sname, "LogDir", log_dir, a, sa, QCS_INI_PFN))
      dbg_exit(strcmb("LogDir", GetIniErrorString()),1);
   strcpy(log_dir,a);
   if (!GetIniInt(sname, "LogRetainDays", 30, &log_retain_days, QCS_INI_PFN))
      dbg_exit(strcmb("LogRetainDays", GetIniErrorString()),1);
   if (!GetIniInt(sname, "LogPurgeHour", 2, &log_purge_hour, QCS_INI_PFN))
      dbg_exit(strcmb("LogPurgeHour", GetIniErrorString()),1);
/*------------------------------------------------------------- Logging initialization */
   if (logInitialize(QCS_INI_PFN)) dbg(GetLogErrorString());

}

/*----------------------------------------------------------------- Error logging */   
void ldbg(char * msg) { dbg(msg); logNote("MAIN", msg); }

/*----------------------------------------------------------------------------------
 * Here we are looking for return statuses on processes which we have run, so that 
 * we can avoid zombies, and so that we can log the return status of the processes.
 *--------------------------------------------------------------------------------*/
static void wait_pid()
{  pid_t pid;
   int status, found;
   while (prs_list)                                 /* No child processes running */
   {  status = 0;
      pid = waitpid((pid_t) -1, &status, WNOHANG);
/*--------------------------------- waitpid was interrupted or generated an error */
      if (pid == -1)
      {  if (errno == EINTR) return;
         dbge_exit("waitpid failure:", 1);
      }
      if (pid == 0) return;                          /* nothing done at this time */
/*--------------------------------------- find process which has finished on list */
      found = 0;
      for (pprs = prs_list, prs = NULL; pprs; prs = pprs, pprs = pprs->nxt)
      {  if (pid == pprs->pid)
         {  found = 1;
/*---------------------------------------------- Log process completed and status */
            sprintf(em,"pid %d, %s exit status = %d", pid, pprs->name, status);
            if (status) logNote(pprs->log_name,em);
            else logNoteLevel(pprs->log_name,em,1);
/*------------------------------------------------------------------- remove link */
            if (prs) prs->nxt = pprs->nxt;
            else prs_list = pprs->nxt;
            FREE_STR(pprs->name)
            FREE_STR(pprs->log_name)
            FREE_STR(pprs)
            break;
         }
      }
      if (!found)
      {  sprintf(em,"pid %d - unknown process - exit status = %d", pid, status);
         logNote("MAIN",em);
      }   
   }  
}
