/****************************************************************************************
 * Emageon qc resend_qc program.                                                R Moulton
 * Fixes DICOM headers of qc'd studies and resends them, IAW study table in image_qc_db.
 * This is a batch process which runs on each server which processes Emageon qc, and
 * periodically checks the image_qc_db to see if there are any qc'd studies to be fixed
 * and resent for this server
 *---------------------------------------------------------------------------------------
 * 07/06 R Moulton Added ability to reprocess studies held when RIS query is down.
 *---------------------------------------------------------------------------------------
 * gcc -g resend_qc.c -o resend_qc `pkg-config --cflags --libs resend_qc`
 ***************************************************************************************/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <mysql.h>

#include "dicom_platform.h"
#include "ctn_os.h"
#include "dicom.h"
#include "utility.h"
#include "dicom_uids.h"
#include "lst.h"
#include "condition.h"
#include "dulprotocol.h"
#include "dicom_objects.h"
#include "dicom_messages.h"
#include "dicom_services.h"

#include "dbg.h"
#include "ini.h"
#include "sql.h"
#include "log.h"
#include "hl7query.h"
#include "EQCerrors.h"

/*----------------------------------- Function prototypes ------------------------------*/
int  read_ini(void);
void EQC_init(void);
void free_study_struct(void);
void catch_signal(int signo);
int  connect_to_db(void);
int  get_study_list(int flag);
int  process_study_list(void);
int  process_nq_study_list(void);
void delete_study(void);
int  process_image(char * pfn, char * new_pfn);
void process_first_image(void);
int  create_needed_directories(char * series_UID);
void nap(void);
void dbg_mysql(char * msg);
char * eliminate_characters(char * input, char * characters_to_eliminate);
int elapsed_time(char * start_time_string);

/*---------------------------- macros -------------------------------------------------*/
#define FREE(A) {free(A);A=NULL;}
#define STR_EQ(A,B) !strcasecmp(A,B)
#define STR(A,B) A=strdup(B)
#define LCHR(A) A[strlen(A) - 1]
#define CLEAR(A) memset(A,0,sizeof(A))
#define CONDERR {sprintf(em,"COND=%x",cond); dbg(em); COND_PopCondition(TRUE);}
#define CLEARCOND COND_PopCondition(TRUE);

/*------------------------------------------------------------------ study query types */
#define QCD_STUDIES 1
#define NQ_STUDIES  2

/*-------------------------------------------------------------- study table structure */
#define STUDY Study[study]
struct _study_ {
   unsigned long int dba;
   char qc_status;
   char * server;
   char * study_id;
   char * source;
   char * modality;
   char * station_name;
   char * institution_name;
   char * pid;
   char * dnumber;
   char * pname;
   char * dob;
   char * sex;
   char * visit;
   char * accn;
   char * description;
   char * exam_code;
   char * body_part;
   char * scheduled;
   char * study_done;
   char * report_done;
   char * first_arrive;
   unsigned long errors;
   char * qc_done;
   char * user_id;
   char * dest_code;
};
static struct _study_ * Study = NULL;
static int study, max_study;

/*-------------------------- Structs used to hold information from DICOM & HL7 sources */
static struct {
   char accn;              /* (0008,0050) OBR 3.1 */
   char description;       /* (0008,1030) OBR 4.2 */
   char pname;             /* (0010,0010) PID 5   */
   char pid;               /* (0010,0020) PID 3.1 */
   char facility;          /* (0010,0021) PID 3.4 (in mpi query)*/
   char dob;               /* (0010,0030) PID 7   */
   char sex;               /* (0010,0040) PID 8   */
   char other_id;          /* (0010,1000) PID 4   */
   char body_part;         /* (0018,0015) */
   char study_uid;         /* (0020,000d) */
   char visit;             /* (0038,0010) PID 18  */
   char sched_date;        /* (0040,0002) OBR 27  */
   char sched_time;        /* (0040,0003) OBR 27  */
   char exam_code;         /* (0008,1032) -> (0008,0100) OBR 4.1 */
   char exam_coding_scheme;/* (0008,1032) -> (0008,0102) IDX constant */
   char exam_code_meaning; /* (0008,1032) -> (0008,0104) OBR 4.2 */
} EQC;
static struct _EQC_elements_ {
   char * ptr;    /* pointer to list above */
   char status;   /* 0=not there,1=there,2=add it,3=modify it */
   char inhl7;    /* Is this something that should be in the RAD? */
   char required; /* Is this required for Emageon? */
   DCM_TAG seqTag; /* Tag of SQ this is part of, or NULL (tags in same SQ should be grouped)*/
   U32 length;     /* Length of this type of item */
   DCM_ELEMENT e; 
} EQCItem[] = {
   { &EQC.accn,             0, 1, 1, 0, DICOM_SH_LENGTH,
      DCM_IDACCESSIONNUMBER,             DCM_SH, "accession   ", 1, DICOM_SH_LENGTH, NULL},
   { &EQC.description,      0, 1, 0, 0, DICOM_LO_LENGTH,
      DCM_IDSTUDYDESCRIPTION,            DCM_LO, "description ", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.pname,            0, 1, 1, 0, DICOM_PN_LENGTH,
      DCM_PATNAME,                       DCM_PN, "patient name", 1, DICOM_PN_LENGTH, NULL},
   { &EQC.pid,              0, 1, 1, 0, DICOM_LO_LENGTH,
      DCM_PATID,                         DCM_LO, "patient id  ", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.facility,         0, 1, 1, 0, DICOM_LO_LENGTH,
      DCM_ISSUERPATIENTID,               DCM_LO, "facility    ", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.dob,              0, 1, 1, 0, DICOM_DA_LENGTH,
      DCM_PATBIRTHDATE,                  DCM_DA, "patient DOB ", 1, DICOM_DA_LENGTH, NULL},
   { &EQC.sex,              0, 1, 1, 0, DICOM_CS_LENGTH,
      DCM_PATSEX,                        DCM_CS, "patient sex ", 1, DICOM_CS_LENGTH, NULL},
   { &EQC.other_id,         0, 1, 0, 0, DICOM_LO_LENGTH,
      DCM_PATOTHERIDS,                   DCM_LO, "d number    ", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.body_part,        0, 0, 0, 0, DICOM_CS_LENGTH,
      DCM_ACQBODYPARTEXAMINED,           DCM_CS, "body part   ", 1, DICOM_CS_LENGTH, NULL},
   { &EQC.visit,            0, 1, 1, 0, DICOM_LO_LENGTH,
      DCM_VISADMISSIONID,                DCM_LO, "visit number", 1, DICOM_LO_LENGTH, NULL},
   { &EQC.sched_date,       0, 1, 1, 0, DICOM_DA_LENGTH,
      DCM_PRCSCHEDULEDPROCSTEPSTARTDATE, DCM_DA, "sched date  ", 1, DICOM_DA_LENGTH, NULL},
   { &EQC.sched_time,       0, 1, 1, 0, DICOM_TM_LENGTH,
      DCM_PRCSCHEDULEDPROCSTEPSTARTTIME, DCM_TM, "sched time  ", 1, DICOM_TM_LENGTH, NULL},
   { &EQC.exam_code,        0, 1, 1, DCM_IDPROCEDURECODESEQUENCE, DICOM_SH_LENGTH,
      DCM_IDCODEVALUE,                   DCM_SH, "exam code   ", 1, DICOM_SH_LENGTH, NULL},
   { &EQC.exam_coding_scheme, 0, 1, 1, DCM_IDPROCEDURECODESEQUENCE, DICOM_SH_LENGTH,
      DCM_IDCODINGSCHEMEDESIGNATOR,      DCM_SH, "coding sch  ", 1, DICOM_SH_LENGTH, NULL},
   { &EQC.exam_code_meaning,  0, 1, 1, DCM_IDPROCEDURECODESEQUENCE, DICOM_LO_LENGTH,
      DCM_IDCODEMEANING,                 DCM_LO, "exam code ds", 1, DICOM_LO_LENGTH, NULL},
};
#define EQCITEM EQCItem[item]
#define EQCTAG(A) EQCItem[EQC.A]
static char item;
static char max_item = sizeof(EQCItem) / sizeof(struct _EQC_elements_);

/*-------------------------------------------------------------------------------- ctn */
DCM_OBJECT * obj = NULL;
CONDITION cond;

/*----------------------------------------------------------------------- MySQL access */
static char * dbhost;
static MYSQL * conn = NULL;
static MYSQL_RES * res = NULL;
static MYSQL_ROW row = NULL;
static char sql_query[1024];

static int image_count;

#define NQ_INTERVAL 60 * 10
static char sigusr2_received = 0;
static void sig2_rtn(int sig) { sigusr2_received = 1; }
static time_t last_try = 0;

/*-------------------------------------------------------------------------- .ini file */
#define PROCESS_FIFO "resend_qc.prs"
#define DEF_INI_PFN "qc.ini"
static char * cwd;                      /* Current working directory (program dir) */
static char inipfn[256];                /* .ini file pfn (def = cwd/qc.ini) */
FILE * debug_fp;                        /* debug file pointer, or NULL if debug is off */
static char * distributor_name;         /* distributor name for server we are on */
static char * acsname;                  /* All caps distributor name */
static int sleep_seconds = 600;         /* seconds to sleep when nothing to fix */
static char log_msg[4096];
static time_t elapsed_seconds;
static char elapsed_string[128];

static char hold_dir[256];
static char resend_dir[256];
/****************************************************************************************
 ****************************************************************************************
 ********************************** main function ***************************************
 ****************************************************************************************
 ***************************************************************************************/
int main(int argc, char * argv[])
{  char * p;     
/*---------------------------------------------------------------------------------------
 * SIGUSR1 is used to wake up this program if it is sleeping, and is sent to it by a qc
 * program that has just qc'd a study for this server.  If this program is not sleeping,
 * SIGUSR1 is ignored.
 * SIGUSR2 is used to let the program know that IDXrad query was down and is now up, so 
 * that any studies which have been held for IDXrad processing can be processed through.
 * When SIGUSR2 is received, a flag is set which is periodically checked.
 *-------------------------------------------------------------------------------------*/
   signal(SIGUSR2, sig2_rtn);
   if (sigignore(SIGUSR1)) dbge_exit("sigignore", 1);
/*---------------------------------------------------------------------------------------
 * Program runs out of directory binary is in.  Make that current.  Then use parameter 1
 * or default to complete .ini file pfn.
 *-------------------------------------------------------------------------------------*/
   if (p = MakeProgDirCurrent(argv[0])) STR(cwd,p);
   else { PrtIniErrorString(stdout,"Startup"); exit(1); }
   strcpy(inipfn,cwd); strcat(inipfn,DEF_INI_PFN);

   unique_process_check(PROCESS_FIFO);   /* make sure only one of process is running */
   if (read_ini()) return(1);            /* process .ini file */
/*---------------------------------------------------------------------------------------
 * Main process loop: queries db for studies ready to fix and fixes any that it finds.
 *-------------------------------------------------------------------------------------*/
   while (1)
   {  if (connect_to_db()) break;             /* Make sure we can get to qc db */
/*--------------------------------------------------------------- Process qc'd studies */
      if (get_study_list(QCD_STUDIES)) break;
      if (max_study == 0) nap(); 
      else if (process_study_list()) break;
/*-------------------------------------- Process studies held because RIS query failed */ 
      if (get_study_list(NQ_STUDIES)) break;
      if (process_nq_study_list()) break;
   }
/*-------------------------------------------------------------------- exiting program */
   PutIniInt("MAIN", "ResendPID", 0, inipfn);  /* clear PID from .ini file */
   exit (1);
}

/*--------------- Connect, reconnect to image_qc_db, return 0 if OK, non-zero on error */
int connect_to_db(void)
{  if (conn) 
   {  if(mysql_ping(conn)) mysql_close(conn);
      else return (0);
   }
   conn = mysql_init(conn);
   conn = mysql_real_connect(conn, dbhost, "qcprs", "farsi", "image_qc_db", 0, NULL, 0);
   if (!conn) { dbg_mysql("connect"); return (1); }
   return (0);
}

/*---------------------------------------------------------------------------------------
 * query db for studies for this server which are ready to qc, and load in struct.
 * Return 0 if OK, non-zero on error
 *-------------------------------------------------------------------------------------*/
int get_study_list(int flag)
{  free_study_struct();
   int i;

   CLEAR(sql_query);
   switch (flag)
/*------------------------------------------------------------- Query for qc'd studies */
   {  case QCD_STUDIES:
         sprintf(sql_query, "SELECT * FROM study WHERE qc_status > 0 AND server = \"%s\"",
            distributor_name);
         break;
/*--------------------------------------------- Query for studies which need RIS query */
      case NQ_STUDIES:
         sprintf(sql_query, "SELECT * FROM study WHERE qc_status = 0 AND server = \"%s\" AND "
            "(errors & %lu) > 0", distributor_name, EQCGetErrorBit(EQCE_QUERY_FAILED));
         break;
      default:
         dbg("get_study_list flag error");
         return (1);
   }

   if (mysql_query(conn,sql_query)) {dbg_mysql("query"); return(1);}
   res = mysql_store_result(conn);
   if (!res) {dbg_mysql("store result"); return(2);}

   max_study = mysql_num_rows(res);
   i = max_study * sizeof(struct _study_);
   Study = malloc(i);
   memset((void *)Study,0,i);
   for (study = 0; study < max_study; study++)
   {  row = mysql_fetch_row(res);
      STUDY.dba = atoi(row[0]);
      STUDY.qc_status = atoi(row[1]);
      STR(STUDY.server,row[2]);
      STR(STUDY.study_id,row[3]);
      STR(STUDY.source,row[4]);
      STR(STUDY.modality,row[5]);
      STR(STUDY.station_name,row[6]);
      STR(STUDY.institution_name,row[7]); 
      STR(STUDY.pid,row[8]);
      STR(STUDY.dnumber,row[9]);
      STR(STUDY.pname,row[10]);
      eliminate_characters(row[11],"-");
      STR(STUDY.dob,row[11]); 
      STR(STUDY.sex,row[12]);
      STR(STUDY.visit,row[13]);
      STR(STUDY.accn,row[14]);
      STR(STUDY.description,row[15]);
      STR(STUDY.exam_code,row[16]);
      STR(STUDY.body_part,row[17]);
      eliminate_characters(row[18],"- :");
      STR(STUDY.scheduled,row[18]);
      eliminate_characters(row[19],"- :");
      STR(STUDY.study_done,row[19]);
      eliminate_characters(row[20],"- :");
      STR(STUDY.report_done,row[20]);
      STR(STUDY.first_arrive,row[21]);
      STUDY.errors = strtoul(row[22], (char **)NULL, 10);
      eliminate_characters(row[23],"- :");
      STR(STUDY.qc_done,row[23]);
      STR(STUDY.user_id,row[24]);
      STR(STUDY.dest_code,row[25]);
   }
   mysql_free_result(res);
   return (0);
}
   
void free_study_struct(void)
{  if (!Study) return;
   for (study = 0; study < max_study; study++)
   {  FREE(STUDY.server)  
      FREE(STUDY.study_id)
      FREE(STUDY.source)
      FREE(STUDY.modality)
      FREE(STUDY.station_name)
      FREE(STUDY.institution_name)
      FREE(STUDY.pid)
      FREE(STUDY.dnumber)
      FREE(STUDY.pname)
      FREE(STUDY.dob)
      FREE(STUDY.sex)
      FREE(STUDY.visit)
      FREE(STUDY.accn)
      FREE(STUDY.description)
      FREE(STUDY.exam_code)
      FREE(STUDY.body_part)
      FREE(STUDY.scheduled)
      FREE(STUDY.study_done)
      FREE(STUDY.report_done)
      FREE(STUDY.first_arrive)
      FREE(STUDY.qc_done)
      FREE(STUDY.user_id)
   }
   FREE(Study)
   return;
}

/*---------------------------------------------------------------------------------------
 * "naps" for sleep_seconds, allowing for SIGUSR1 to wake it up early.
 *-------------------------------------------------------------------------------------*/
static void sig_rtn(int sig) { return; }
void nap(void)
{  int early;
   sprintf(em,"Sleeping for %d seconds", sleep_seconds); dbg(em);
   signal(SIGUSR1,sig_rtn);
   early = sleep(sleep_seconds);
   signal(SIGUSR1,SIG_IGN);
   if (early > 0) { sprintf(em,"Woke up %d seconds early", early); dbg(em); }
   else dbg("Woke up on time");
   return;
}

/*---------------------------------------------------------------------------------------
 * Initialize EQC struct. Done before each study.
 *-------------------------------------------------------------------------------------*/
void EQC_init(void)
{  for (item = 0; item < max_item; item++) 
   {  EQCITEM.e.length = EQCITEM.length;          /* set element length to size for type */
      if (!EQCITEM.e.d.string)                    /* first time items */
      {  *(EQCITEM.ptr) = item;                     /* loads subscrips into EQC struct */
         EQCITEM.e.d.string = malloc(EQCITEM.e.length+1);          /* data value areas */
      }
      memset(EQCITEM.e.d.string,0,EQCITEM.e.length);
   }
}

/*---------------------------------------------------------------------------------------
 * Process studies in the study list.  Return 0 if OK, non-zero on error.
 *-------------------------------------------------------------------------------------*/
static char first_image, first_in_series, ok_to_clear_study;
int process_study_list(void)
{  char study_path[256], series_path[256], series_UID[128], image_UID[128], pfn[256], new_pfn[256];
   DIR * study_dir, * series_dir;
   struct dirent * dir_ent;
   struct stat statbuf;
   sprintf(em,"Processing %d studies", max_study); dbg(em);
/*------------------------------------------------------- Process each study in struct */
   for (study = 0; study < max_study; study ++)
   {  EQC_init();                           /* initialize EQC struct items */
      if (STUDY.qc_status == 2) { delete_study(); continue; }
      sprintf(em,"Fixing study: pid = %s, accn = %s, pname = %s", STUDY.pid, STUDY.accn, STUDY.pname); dbg(em);
/*-------------------------------------------- create study directory name and open it */
      first_image = 1;   /* first image for this study */
      ok_to_clear_study = 1;  /* OK to clear this study when we're done (so far) */
/*-------------------------------------------------------------------- study directory */
      CLEAR(study_path);
      sprintf(study_path, "%s/%s/%s/%s", hold_dir, STUDY.source, STUDY.modality, STUDY.study_id);
/*-------------------------------- Check if exists and is directory, or we just delete */
      if (access(study_path, F_OK) || stat(study_path, &statbuf) || !(S_ISDIR(statbuf.st_mode)))
      {  CLEAR(sql_query);
         sprintf(sql_query, "DELETE FROM study WHERE DBA = %d", STUDY.dba);
         if (mysql_query(conn,sql_query)) {dbg_mysql("query"); continue;}
         STUDY.dba = -1;
         continue;
      }
      study_dir = opendir(study_path);
      if (!study_dir) {dbge("open study dir"); continue;}
/*---------------------------------------------------------------------------------------
 * scan study directory looking for series directories.  We do not allow any other 
 * directories in a study directory except a series directory. Files in the study
 * directory other than QCStudy.dat are deleted.
 *-------------------------------------------------------------------------------------*/
      while (dir_ent = readdir(study_dir))
      {  if (STR_EQ(dir_ent->d_name,".") || STR_EQ(dir_ent->d_name,".."))  continue;
         CLEAR(series_path);
         sprintf(series_path,"%s/%s", study_path, dir_ent->d_name);
         if (stat(series_path, &statbuf)) { dbge("series dir stat"); continue; }
         if (!S_ISDIR(statbuf.st_mode)) 
         {  if (strcmp(dir_ent->d_name,"QCStudy.dat"))
            {  sprintf(pfn,"%s/%s", study_path, dir_ent->d_name);
               remove(pfn);
            }
            continue;
         }
         strcpy(series_UID,dir_ent->d_name);
/*---------------------------------------------------------------------------------------
 * scan series directory looking for images.  We do not allow any other files except
 * image files in series directories
 *-------------------------------------------------------------------------------------*/
         series_dir = opendir(series_path);
         if (!series_dir) {dbge("open series dir"); continue;}
         first_in_series = 1;
         while (dir_ent = readdir(series_dir))
         {  if (*(dir_ent->d_name) == '.') continue;
            strcpy(image_UID,dir_ent->d_name);
            sprintf(pfn, "%s/%s/%s", study_path, series_UID, image_UID);
            sprintf(new_pfn, "%s/%s/%s/%s/%s", resend_dir, STUDY.modality, STUDY.study_id,
               series_UID, image_UID);
            create_needed_directories(series_UID);
            process_image(pfn, new_pfn);
            first_image = first_in_series = 0;
         }
         closedir(series_dir);
         if (rmdir(series_path)) ok_to_clear_study = 0;
      }
/*--- If all series directories rmdir'd, do another scan to be sure study dir is clear */
      if (!ok_to_clear_study) { closedir(study_dir); continue; }
      rewinddir(study_dir);
      while (dir_ent = readdir(study_dir))
      {  if (*(dir_ent->d_name) == '.' || !strcmp(dir_ent->d_name,"QCStudy.dat")) continue;
         ok_to_clear_study = 0;
         break;
      }
      closedir(study_dir);
      if (!ok_to_clear_study) continue;
/*------- dir clear except for QCStudy.dat file; delete it and try to delete study dir */
      sprintf(pfn,"%s/QCStudy.dat", study_path); 
      if (remove(pfn)) continue;
      if (rmdir(study_path)) continue;
/*------------------------------------------------------------ Remove study from qc db */
      CLEAR(sql_query);
      sprintf(sql_query, "DELETE FROM study WHERE DBA = %d", STUDY.dba);
      if (mysql_query(conn,sql_query)) {dbg_mysql("query"); continue;}
      STUDY.dba = -1;

/*----------------------------------------------------------------- resend log message */
      CLEAR(log_msg);
      if (STR_EQ(STUDY.user_id,"_resend_qc_") && STR_EQ(STUDY.dest_code,"SELF"))
         sprintf(log_msg, "Study held for query resent to %s :%s\n", STUDY.dest_code, STUDY.source);
      else
         sprintf(log_msg, "Study updated and resent to %s\n", STUDY.dest_code);
      sprintf(em,"   pid %s  accn %s  name %s\n", STUDY.pid, STUDY.accn, STUDY.pname); 
      strcat(log_msg,em);
      sprintf(em,"   source %s  modality %s  study done %s\n", STUDY.source, STUDY.modality, STUDY.study_done); 
      strcat(log_msg,em);
      sprintf(em,"   station %s  inst %s\n", STUDY.station_name, STUDY.institution_name); 
      strcat(log_msg,em);
      sprintf(em,"   study UID.. %s\n", STUDY.study_id);         
      strcat(log_msg,em);

      elapsed_time(STUDY.first_arrive);
      sprintf(em,"   arrived.... %s - time to qc and resend %s\n", STUDY.first_arrive, elapsed_string);
      strcat(log_msg,em);
   
      logMsgLevel("EQChold",log_msg,2);

   }
   return (0);
}

/*---------------------------------------------------------------------------------------
 * Update image file (pfn) with STUDY data, then output to new_pfn, delete pfn when done
 *-------------------------------------------------------------------------------------*/
int process_image(char * pfn, char * new_pfn)
{  DCM_OBJECT * seqObj, * ob;
   DCM_TAG seqObj_tag = 0;
   LST_HEAD * seqList;
   DCM_SEQUENCE_ITEM * seqItem;

/*-------------------------------------------------- Get rid of previous image, if any */
   if (obj) DCM_CloseObject(&obj); CLEARCOND
/*-------------------------------------------------------------------- Load image file */
   cond = DCM_OpenFile(pfn, DCM_ORDERLITTLEENDIAN, &obj);
   if (cond != DCM_NORMAL) {
      CLEARCOND
      cond = DCM_OpenFile(pfn, DCM_PART10FILE, &obj);
      if (cond != DCM_NORMAL) { CONDERR return; }
   }
   if (first_image) process_first_image(); 

   image_count++;
   sprintf(em,"---- Image %d", image_count); dbg(em);
/*------------------------------------------------------ Add or update items as needed */
   for (item = 0; item < max_item; item++)
   {  if (EQCITEM.status < 2) continue;
/*---------------------------------------------------------------------------------------
 * For items which are part of a sequence, make the sequence if needed, and pull the 
 * DICOM object pointer from the list.  Note that the table groups objects in the same
 * SQ tag, we can just use the one we have if it is the same as the last item's.
 *-------------------------------------------------------------------------------------*/
      if (EQCITEM.seqTag && (!seqObj || (EQCITEM.seqTag != seqObj_tag)))
      {  cond = DCM_GetSequenceList(&obj, EQCITEM.seqTag, &seqList); 
         if (cond != DCM_NORMAL)
         {  CLEARCOND
            DCM_ELEMENT e = {EQCITEM.seqTag, DCM_SQ, "", 1, 0, 0};
            e.d.sq = seqList = LST_Create();
            seqItem = (DCM_SEQUENCE_ITEM *) malloc(sizeof(*seqItem));
            DCM_CreateObject(&seqItem->object, 0);
            LST_Enqueue(&seqList, seqItem);
            cond = DCM_AddSequenceElement(&obj, &e);
            if (cond != DCM_NORMAL)
            {  sprintf(em,"Failed to add SQ for %s", EQCITEM.e.description); dbg(em);
               CONDERR
               continue;
            }
         }
         seqItem = (DCM_SEQUENCE_ITEM*) LST_Head(&seqList);
         seqObj = seqItem->object;
         seqObj_tag = EQCITEM.seqTag;
      }
      if (EQCITEM.seqTag) ob = seqObj; else ob = obj;
      
      if (EQCITEM.status == 2)
      {  sprintf(em,"Adding %s: v=%s", EQCITEM.e.description, EQCITEM.e.d.string); dbg(em);
         cond = DCM_AddElement(&ob, &(EQCITEM.e));
         if (cond != DCM_NORMAL) 
         {  sprintf(em,"Failed to add %s: v=%s", EQCITEM.e.description, EQCITEM.e.d.string);
            CONDERR
         }
      }
      if (EQCITEM.status == 3)
      {  sprintf(em,"Modifying %s: v=%s", EQCITEM.e.description, EQCITEM.e.d.string); dbg(em);
         cond = DCM_RemoveElement(&ob, EQCITEM.e.tag);
         if (cond != DCM_NORMAL) dbg("DCM_RemoveElement failed");
         cond = DCM_AddElement(&ob, &(EQCITEM.e));
         if (cond != DCM_NORMAL) 
         {  sprintf(em,"Failed to modify %s: v=%s", EQCITEM.e.description, EQCITEM.e.d.string);
            CONDERR
         }
      }
   }
   remove(new_pfn);
   cond = DCM_WriteFile(&obj, DCM_ORDERLITTLEENDIAN, new_pfn);
   if (cond != DCM_NORMAL) CONDERR
   else remove(pfn);
}

/*---------------------------------------------------------------------------------------
 * First image for a study. reset and set up EQC struct items.
 *-------------------------------------------------------------------------------------*/
void process_first_image(void)
{  unsigned long ul;
   void * ctx;
   image_count = 0;
/*-------------------------------------------------- clear struct, load from DICOM obj */
   for (item = 0; item < max_item; item++) 
   {  EQCITEM.status=0; 
      memset(EQCITEM.e.d.string,0,EQCITEM.e.length+1);
      if (EQCITEM.seqTag)      
/*-------------------------------------------------------- If item is part of sequence */
      {  LST_HEAD * seqList = NULL;
         DCM_SEQUENCE_ITEM * itm = NULL;
/*--------------------------------------------- Pull the list out of the sequence item */
         cond = DCM_GetSequenceList(&obj, EQCITEM.seqTag, &seqList);
         if (cond == DCM_NORMAL)
/*-------------- Pull object item out of list (this works only for non-repeating sq's) */
         {  itm = (DCM_SEQUENCE_ITEM*) LST_Head(&seqList);
            if (itm && itm->object)
/*------------------------------------------------------------ Pull item out of object */
            {  ctx = NULL; 
               cond = DCM_GetElementValue(&(itm->object), &EQCITEM.e, &ul, &ctx);
               if (cond == DCM_NORMAL || cond == DCM_GETINCOMPLETE) EQCITEM.status = 1;
            }
         }
      }
      else
/*--------------------------------------------------------------------- top level tags */
      {  ctx = NULL;
         cond = DCM_GetElementValue(&obj, &EQCITEM.e, &ul, &ctx);
         if (cond == DCM_NORMAL || cond == DCM_GETINCOMPLETE) EQCITEM.status = 1; 
      }                                                                                                   
      sprintf(em,"%s: s=%2d, v=%s", EQCITEM.e.description, EQCITEM.e.length, EQCITEM.e.d.string); dbg(em);
   }
/*-------------------------------- Load STUDY items into struct element string fields */ 
   strcpy(EQCTAG(accn).e.d.string,STUDY.accn); EQCTAG(accn).status += 2;
   strcpy(EQCTAG(pid).e.d.string,STUDY.pid); EQCTAG(pid).status += 2;
   strcpy(EQCTAG(other_id).e.d.string,STUDY.dnumber); EQCTAG(other_id).status += 2;
   strcpy(EQCTAG(visit).e.d.string,STUDY.visit); EQCTAG(visit).status += 2;
   strcpy(EQCTAG(pname).e.d.string,STUDY.pname); EQCTAG(pname).status += 2;
   strcpy(EQCTAG(dob).e.d.string,STUDY.dob); EQCTAG(dob).status += 2;
   strcpy(EQCTAG(sex).e.d.string,STUDY.sex); EQCTAG(sex).status += 2;
   strcpy(EQCTAG(exam_code).e.d.string,STUDY.exam_code); EQCTAG(exam_code).status += 2;
   if (strlen(EQCTAG(exam_code).e.d.string)) 
   {  strcpy(EQCTAG(exam_coding_scheme).e.d.string,"99IDX"); 
      EQCTAG(exam_coding_scheme).status += 2;
   }
   strcpy(EQCTAG(exam_code_meaning).e.d.string,STUDY.description); EQCTAG(exam_code_meaning).status += 2;
   strcpy(EQCTAG(description).e.d.string,STUDY.description); EQCTAG(description).status += 2;
/*----------------------------------------- Split sched date & time (one field in hl7) */
   if (strlen(STUDY.scheduled) >= 8)
   {  memset(em,0,sizeof(em)); 
      memcpy(em,STUDY.scheduled,8);
      strcpy(EQCTAG(sched_date).e.d.string,em); EQCTAG(sched_date).status += 2;
      if (strlen(STUDY.scheduled) >= 12)
      {  memset(em,0,sizeof(em)); 
         memcpy(em,STUDY.scheduled+8,4);
         strcpy(EQCTAG(sched_time).e.d.string,em); EQCTAG(sched_time).status += 2;
      }
   }
/*------------------------------------------------------- Compute new lengths of items */
   for (item = 0; item < max_item; item++) EQCITEM.e.length = strlen(EQCITEM.e.d.string);
   return;
}
   
/*---------------------------------------------------------------------------------------
 * Create directories as needed for resend of study and series in study. Return 0 if OK,
 * non-zero on error
 *-------------------------------------------------------------------------------------*/
int create_needed_directories(char * series_UID)
{  char path_name[256], tag[64], ovrAE_name[256];
   int fd, i;
   mode_t hold, c = (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
/*---------------------------------------- Create study directory, put in routing file */
   if (first_image)
   {  sprintf(path_name, "%s/%s/%s", resend_dir, STUDY.modality, STUDY.study_id);
      cond = UTL_VerifyCreatePath(path_name);
      if (cond != UTL_NORMAL) { dbg("Unable to create study directory"); return (1); }
      strcpy(ovrAE_name,path_name);
      sprintf(tag,"/dest.%s.sendto", strcasecmp(STUDY.dest_code,"SELF") ? STUDY.dest_code : acsname);
      strcat(path_name,tag);
      hold = umask(0);
      fd = creat(path_name, c); 
      if (fd < 0) { sprintf(em,"Unable to create %s", path_name); dbge(em); return(1); }
      close(fd);
/*------------------------------ for SELF routed studies, put in overide AE title file */
      if (!strcasecmp(STUDY.dest_code,"SELF"))
      {  sprintf(tag,"/ovrAE.%s", acsname);
         strcat(ovrAE_name,tag);
         fd = open(ovrAE_name, O_WRONLY | O_CREAT | O_TRUNC, c);
         if (fd < 0) { sprintf(em,"Unable to create %s", ovrAE_name); dbge(em); return(1); }
         i = write(fd, STUDY.source, strlen(STUDY.source));
         if (i != strlen(STUDY.source)) { dbge("Unable to write ovrAE file"); return(1); }
         close(fd);
      }
            
      umask(hold);
   }
/*------------------------------------------------------------ Create series directory */
   if (first_in_series)
   {  sprintf(path_name, "%s/%s/%s/%s", resend_dir, STUDY.modality, STUDY.study_id, series_UID);
      cond = UTL_VerifyCreatePath(path_name);
      if (cond != UTL_NORMAL) { dbg("Unable to create series directory"); return (1); }
   }
   return (0);
}

/*---------------------------------------------------------------------------------------
 * Study is to be deleted.
 *-------------------------------------------------------------------------------------*/
void delete_study(void)
{  sprintf(em,"rm -r %s/%s/%s/%s/", hold_dir, STUDY.source, STUDY.modality, STUDY.study_id);
   if(system(em)) dbge("study delete rm call");
   CLEAR(sql_query);
   sprintf(sql_query,"DELETE FROM study WHERE dba = %d", STUDY.dba);
   if (mysql_query(conn,sql_query)) dbg_mysql("query");

/*----------------------------------------------------------------- delete log message */
      strcpy(log_msg, "Study deleted\n");
      sprintf(em,"   pid %s  accn %s  name %s\n", STUDY.pid, STUDY.accn, STUDY.pname); 
      strcat(log_msg,em);
      sprintf(em,"   source %s  modality %s  study done %s\n", STUDY.source, STUDY.modality, STUDY.study_done); 
      strcat(log_msg,em);
      sprintf(em,"   station %s  inst %s\n", STUDY.station_name, STUDY.institution_name); 
      strcat(log_msg,em);
      sprintf(em,"   study UID.. %s\n", STUDY.study_id);         
      strcat(log_msg,em);

      elapsed_time(STUDY.first_arrive);
      sprintf(em,"   arrived.... %s - time to qc and delete %s\n", STUDY.first_arrive, elapsed_string);
      strcat(log_msg,em);
   
      logMsgLevel("EQChold",log_msg,3);

   return;
}

/****************************************************************************************
 * Process "Need ris Query" (NQ) studies
 ***************************************************************************************/
int process_nq_study_list(void)
{  char flag = 0;
   static struct _hl7_accn_query_ accn_query;
   char msg[4096];
/*---------------------------------------- If no nq studies, clear flags and return OK */
   if (max_study == 0) 
   {  sigusr2_received = 0;
      last_try = 0;
      return (0);
   }   
/*----------------------- Check if we are going to try to process, or clear and return */
   if (!last_try) last_try = time(NULL);
   if (!sigusr2_received && (difftime(time(NULL), last_try) < NQ_INTERVAL ))
   {  free_study_struct();
      return (0);
   }
   sprintf(msg,"Process %d studies on hold for RIS query\n", max_study);
   if (sigusr2_received) strcat(msg,"   Signal received\n");
   else 
   {  sprintf(em, "   %d second interval expired\n", sleep_seconds);
      strcat(msg,em);
   }
/*---------------------------------------------------------------- We are going to try */
   last_try = time(NULL);
/*------------------------------ First step is to do a test RIS query, to see if it works
 * This uses an accession from William Shakespeare, but it doesn't really matter.
 *-------------------------------------------------------------------------------------*/
   memset(&accn_query,0,sizeof(accn_query));
   accn_query.accn = 4398551;
   hl7AccnQuery(&accn_query);
   if (hl7q_query_failed)
   {  free_study_struct();
      strcat(msg,"   Query test failed");
      logMsg("EQChold", msg);
      return (0);
   }
   accn_query.accn = -1;
   hl7AccnQuery(&accn_query);
/*---------------------------------------------------------------------------------------
 * query worked, so select all the studies for resend and set destination to "SELF",
 * which will send them to their own server using their original presenting AE Title.
 *-------------------------------------------------------------------------------------*/
   free_study_struct();
   CLEAR(sql_query);
   sprintf(sql_query,"UPDATE study SET qc_status = 1, user_id = \"_resend_qc_\", dest_code = \"SELF\" "
      "WHERE qc_status = 0 AND server = \"%s\" AND (errors & %lu) > 0", 
      distributor_name, EQCGetErrorBit(EQCE_QUERY_FAILED));
   if (mysql_query(conn,sql_query)) 
   {  strcat(msg,"   Update qc_status failed");
      logMsg("EQChold", msg);
      return (0);
   }
   strcat(msg,"   studies set to resend to SELF");
   logMsg("EQChold", msg);
   return (0);
}

/*--------------------------------------------------------- Read .ini file information */
int read_ini(void)
{  char a[1024], main[] = "[MAIN]";
   int sa = sizeof(a);
   char * p;
   
   if (access(inipfn, F_OK)) { printf("%s file not found\n", inipfn); return(1); }
   
   if (!GetIniString(main, "ResendDebug", "stdout", a, sa, inipfn))
   {  PrtIniErrorString(stdout, "ini file error");
      return(1);
   }
   debug_fp = debug_on(a);
/*--------------------------------------------------------- server running image_qc_db */
   GetIniString(main, "DBHost", "localhost", a, sa, inipfn);
   STR(dbhost,a);
/*----------------------------------- distributor name we are on, defaults to hostname */
   gethostname(a, sa);
   GetIniString(main, "DistributorName", a, a, sa, inipfn);
   STR(distributor_name,a);
   STR(acsname,a);
   for (p = acsname; *p; p++) *p = toupper(*p);

/*------------------------------------------------------------------------ directories */
   GetIniString(main,"ImageDirectory", "/distributor/imgs", a, sa, inipfn);
   if (LCHR(a) != '/') strcat(a,"/");
   strcpy(hold_dir,a); strcat(hold_dir,"EQChold");
   strcpy(resend_dir,a); strcat(resend_dir,"EQCresend");
   
/*-------------------------------- Seconds to sleep when nothing to fix in image_qc_db */
   GetIniInt(main, "SleepSeconds", sleep_seconds, &sleep_seconds, inipfn);

/*------------------------------- Store pid for qc programs to kick start this process */
   PutIniInt(main, "ResendPID", getpid(), inipfn);
   
/*------------------------------------------------------------- Logging initialization */
   if (logInitialize(inipfn)) dbg(GetLogErrorString());

   return 0;
}

/*--------------------------------------------------------------- dbg for mysql errors */
void dbg_mysql(char * msg) {sprintf(em,"%s: %s",msg,mysql_error(conn)); dbg(em);}

/*---------------------------------------------------------------------------------------
 * Eliminate characters in second string from first string, closing up empty spaces.
 * return pointer to first string.
 *-------------------------------------------------------------------------------------*/
char * eliminate_characters(char * input, char * characters_to_eliminate)
{  char work[strlen(input) + 1], *p = work, *q = input;
   strcpy(work,input); 
   memset(input,0,strlen(input));
   while (*p)
   {  if (!strchr(characters_to_eliminate,*p)) { *q = *p; q++; }
      p++;
   }
   return input;
}

/*---------------------------------------------------------------------------------------
 * Compares a mysql format date-time stamp with the current time. Puts the elapsed time
 * in elapsed_seconds, and a character version in elapsed_string.  Returns zero if OK,
 * non-zero on error.
 *-------------------------------------------------------------------------------------*/
int elapsed_time(char * start_time_string)
{  time_t a;
   int i;
   struct tm tm;
/*------------------------------------------------------ Pull date out of mysql format */
   sscanf(start_time_string,"%4u%*c%2u%*c%2u%*c%2u%*c%2u%*c%2u", &tm.tm_year, &tm.tm_mon,
      &tm.tm_mday, &tm.tm_hour, &tm.tm_min, &tm.tm_sec);
   tm.tm_year -= 1900; tm.tm_mon--;     /* adjust year and month bases */
   tm.tm_isdst = -1;                    /* Let mktime determine if DST is in effect */
/*----------------------------------------------------- compute elapsed time until now */
   a = mktime(&tm);
   if (a == -1) 
   {  CLEAR(elapsed_string);
      elapsed_seconds = 0;
      return 1;
   }
   a = elapsed_seconds = difftime(time(NULL), a);
/*------------------------------------------------------------- character elapsed time */
   CLEAR(elapsed_string);
   i = a / (60 * 60);
   if (i) 
   {  sprintf(elapsed_string,"%d:", i);
      a = a - (i * 60 * 60);
   }
   i = a / 60;
   sprintf(em,"%02d:", i); strcat(elapsed_string, em);
   a = a - (i * 60);
   sprintf(em,"%02d", a); strcat(elapsed_string, em);
   return 0;
}
