/******************************************************************************
 * PACSO_Query.c                                                      R Moulton
 * Query processing server for PACSO Merge process.  Listens for query coming
 * from rentgn which will contain an MPI number or a D number (which has been
 * eliminated in an IDXrad merge). Queries all the distributors, looking for 
 * any which have sent studies using that number as the PID. 
 * Inbound message: <SB>QRY|1|pidtocheck|<EB>
 * Outbound message: <SB>QRY|1|pidtocheck|AA|count|<EB> (success)
 * Outbound message: <SB>QRY|AE|text: reason for failure|<EB> (failure)
 * Where: pidtocheck is the MPI or D number being checked, result is YES if
 * studies were found with that number, NO if no studies were found with that
 * number, or UNK if one or more of the distributor databases could not be
 * querried.
 *-----------------------------------------------------------------------------
 * Compile gcc PACSO_Query.c dbg.o ini.o sql.o -I/usr/local/mysql/include -L/usr/local/mysql/lib -lmysqlclient -lsocket -lnsl -lz -lm -o PACSO_Query
 *****************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>

#include "dbg.h"
#include "ini.h"
#include "sql.h"

/*------------------------------------------------------------------- Macros */
#define FREE(A) {free(A);A=NULL;}
#define STR_EQ(A,B) !strcasecmp(A,B)
#define STR(A,B) {if(B){if(A=realloc(A,strlen(B)+3))strcpy(A,B);}else{free(A);A=NULL;}}
#define LCHR(A) A[strlen(A) - 1]
#define CLEAR(A) memset(A,0,sizeof(A))

#define SB 11
#define EB 28
#define SEB "\034"

/*------------------------------------------------------ Function prototypes */
void cleanup(int *sock);
void wrap(char *msg);
void unwrap(char *msg);
int read_ini(void);
void dbgm(char*msg, char*buff);
void process_msg(void);

char message[256]          = "null";
char host_name[256]        = "\0";
char peer_name[256]        = "rentgn";
char peer_ip[24] = "128.252.118.40";
int listen_port          = 45002;
struct hostent *phostent, shostent;
struct sockaddr_in listen_socket_name;
struct sockaddr_in * sptr = &listen_socket_name;
time_t timer;
time_t * ptimer = &timer; 
struct tm * ptime;
int listen_socket = 0, rts, new_socket, i, j, rc;
int on = 1;
int socket_num = 0;
FILE * fp;
char pfn[] = "server_messages.txt";
char server_message[256];
char * p;
#define DEF_INI_PFN "PACSO_Query.ini"
char inipfn[256];
char sqlini[256] = "sql.ini";
#define PROCESS_FIFO "PACSO_Query.prs"
FILE * debug_fp;
static char * db[25];
fd_set rsock, wsock, esock;
struct timeval timeout = {15, 0 }, *to;
char msg[512], buff[512];
char * cwd;
   
int main(int numpars, char * par[])
{  int srvr = 1, slct;
   char * p;
/*----------------------------------------------------------------------
 * Program runs out of directory binary is in.  Make that current.
 *--------------------------------------------------------------------*/
   if (p = MakeProgDirCurrent(par[0])) STR(cwd,p)
   else { PrtIniErrorString(stdout,"Startup"); exit(1); }
/*------------------------- If parameter 1 exists, it is ini file pfn */
   strcpy(inipfn,cwd);
   if (numpars > 1) strcat(inipfn,par[1]);
   else strcat(inipfn,DEF_INI_PFN);
   
   if (read_ini()) return(1);
   unique_process_check(PROCESS_FIFO);
   
/*---------------------------------------------- Set up server socket */
   while (srvr)
   {  dbg("Initializing server");
      cleanup(&listen_socket);
      listen_socket = socket(AF_INET, SOCK_STREAM, 0);
      if (listen_socket == -1) dbge_exit("socket",1);

      if (setsockopt(listen_socket, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)))
         dbge_exit("setsockopt", 1); 

      if (!strlen(host_name)) if (gethostname(host_name, sizeof(host_name))) dbge_exit("gethostname", 1);

      if (!(phostent = gethostbyname(host_name))) dbge_exit("gethostbyname", 1); 
   
      shostent = *phostent;
      listen_socket_name.sin_family = shostent.h_addrtype;
      listen_socket_name.sin_port = htons(listen_port);
      listen_socket_name.sin_addr = *((struct in_addr *)shostent.h_addr);

      if (bind(listen_socket, (struct sockaddr *)sptr, sizeof(listen_socket_name))) dbge_exit("bind", 1);

      if (listen(listen_socket, 2)) dbge_exit("listen", 1);

      sprintf(em, "Server initialized: %s, %d", host_name, listen_port); dbg(em);
      slct = 1;
      cleanup(&socket_num); rts = 0;
/*--------------------------------------------------------- Select loop */
      while (slct)
      {  dbg("select loop");
         if (!socket_num) dbg("Waiting for connection");
         
         FD_ZERO(&rsock);
         FD_ZERO(&wsock);
         FD_ZERO(&esock);
         FD_SET(listen_socket, &rsock);
         if (socket_num)
         {  FD_SET(socket_num,&rsock);
            FD_SET(socket_num,&esock);
            if (rts) FD_SET(socket_num,&wsock);
         }
         
         i = ((listen_socket > socket_num) ? listen_socket : socket_num) + 1;
         timeout.tv_sec = debug_fp ? 15 : 3600;  /* Timeout shorter in debug mode */
         rc = select(i, &rsock, &wsock, &esock, &timeout);

         if (!rc) continue;  /* timeout */
         if (rc < 0) {dbge("select"); slct = 0; continue; } /* Error, bounce server */
         
/*-----------------------------------------------------------------------------
 * Activity on listen socket is attempt to connect.  Validate that this is from
 * peer, and close old socket if not already closed.
 *---------------------------------------------------------------------------*/
         while (FD_ISSET(listen_socket, &rsock))
         {  unsigned char * p;
            char pip[24];
            int on = 1;
/*-------------------------------------------------------- Accept connection */
            i = sizeof(struct sockaddr_in);
            new_socket = accept(listen_socket, (struct sockaddr*)&listen_socket_name, &i);
            if (new_socket == -1) { dbge("accept"); break; }
            rc = setsockopt(new_socket, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(on));
            if (rc) { dbge("setsockopt"); break; }
/*------------------------------------------------- Validate peer IP address */
            p = (unsigned char *)&listen_socket_name.sin_addr;
            sprintf(pip, "%d.%d.%d.%d", p[0], p[1], p[2], p[3]);
            if (strcasecmp(pip,peer_ip))
            {  sprintf(em, "Inv IP: %s, rejected", pip); dbg(em);
               break;
            }
            sprintf(em,"Connected to %s", pip); dbg(em);
/*----------------------------------- close old socket, replace with new one */
            cleanup(&socket_num);
            socket_num = new_socket;
            CLEAR(msg);
            rts = 0; 
            break;
         }
            
/*-----------------------------------------------------------------------------
 * Error on socket; clear out and loop
 *---------------------------------------------------------------------------*/
         while (FD_ISSET(socket_num, &esock))
         {  dbg("socket errored out");
            cleanup(&socket_num);
            CLEAR(msg);
            rts = 0;
            break;
         }

/*--------------------------------------------------------- Write on socket */
         while (FD_ISSET(socket_num, &wsock))
         {  dbgm("Responce", msg);
            rc = send(socket_num, msg, strlen(msg), 0);
            if (rc < 0) { dbge("send"); cleanup(&socket_num); break; }
            sprintf(em, "Send %d bytes", rc); dbg(em);
            CLEAR(msg);
            rts = 0;
            break;
         }

/*----------------------------------------------------------------------------
 * Read on socket.  Get message, then process
 *--------------------------------------------------------------------------*/
         while (FD_ISSET(socket_num, &rsock))
         {  char * p;
            dbg("Reading socket");
            CLEAR(buff);
            rc = recv(socket_num, buff, sizeof(buff),0);
            if (rc == 0) 
            {  dbg("client dropped connection"); 
               cleanup(&socket_num);
               CLEAR(msg);
               break;
            }
            if (rc < 0)
            {  dbge("recv");
               cleanup(&socket_num);
               CLEAR(msg);
               break;
            }
            sprintf(em, "%d chars read", rc); dbg(em);
            if (p = strchr(buff,SB)) { CLEAR(msg); strcpy(msg,p); }
            else strcat(msg,buff);
            if (strchr(msg,EB)) process_msg();
            break;
         }
      }
      dbg("Server bouncing");
   }
   dbg("Terminating");
}

/******************************************************************************
 * process message. requests count of studies for pid
 * Inbound msg: QRY|1|pid| 
 * Success Outbound msg: QRY|1|pid|AA|count|
 * Failure Outbound msg: QRY|AE|text message, reason for failure|
 * This query will return either one row (if any studies) or no rows if there
 * are no studies.
 *****************************************************************************/
#define QRY "SELECT COUNT(*) as \"cnt\", pname, pid FROM studylog where pid = \"%s\" group by pid"
#define ERR_EXIT(A) { strcpy(msg,"QRY|AE|"); strcat(msg,A); strcat(msg,"|"); wrap(msg); rts = 1; return; }
void process_msg(void)
{  MYSQL     * conn = NULL;
   MYSQL_RES * res = NULL;
   MYSQL_ROW row = NULL;
   char *p, *pid = NULL;
   int i, j, total_studies = 0, error_flag = 0;
   unwrap(msg);
   p = strtok(msg,"|");
   if (!p || strcasecmp(p,"QRY")) ERR_EXIT("Not QRY msg")
   p = strtok(NULL, "|");
   if (!p || strcasecmp(p,"1")) ERR_EXIT("Invalid QRY #")
   p = strtok(NULL, "|");
   if (!p || !strlen(p)) ERR_EXIT("Missing pid")
   STR(pid,p)
   CLEAR(buff);
   sprintf(buff, QRY, pid);

/*------------------------------------------- Pass all distributor databases */
   for (i = 0; db[i]; i++)
/*----------------------- Open database connection, using .ini file settings */
   {  conn = sqlOpen(db[i], "spectrum");
      if (!conn)
      {  sprintf(em,"%s open err: %s", db[i], GetSqlErrorString(conn)); dbg(em);
         ERR_EXIT(em)
      }
/*------------------------------------------------------------- Do the query */
      if (mysql_query(conn, buff))
      {  sprintf(em, "%s query: %s", db[i], mysql_error(conn)); dbg(em);
         sqlClose(conn);
         ERR_EXIT(em)
      }
/*---------------------------------------------------------- Pull result set */
      res = mysql_store_result(conn);
      if (!res)
      {  sprintf(em,"%s store_result: %s", db[i], mysql_errno(conn)); dbg(em);
         sqlClose(conn);
         ERR_EXIT(em)
      }
/*------------------------------------------------------ Pull row, get count */
      if (mysql_num_rows(res))                  /* are there any studies */
      {  row = mysql_fetch_row(res);
         if (!row)
         {  sprintf(em,"%s fetch_row: %s", db[i], mysql_errno(conn)); dbg(em);
            mysql_free_result(res);
            sqlClose(conn);
            ERR_EXIT(em)
         }
         if (debug_fp)
         {  dbg(db[i]);
            for (j = 0; j < mysql_num_fields(res); j++) dbg(row[j] ? row[j] : "NULL");
         }
         if (row[0]) total_studies += atoi(row[0]);
      }
      mysql_free_result(res);
      sqlClose(conn);
   }
/*-------------------------------------------------------- Make return query */
   CLEAR(msg);
   sprintf(msg,"QRY|1|%s|AA|%d|", pid, total_studies); dbg(msg);
   wrap(msg);
   rts = 1;
   return;
}

/*------------------------------------------------ Load .ini file parameters */
int read_ini(void)
{  char a[1024], main[] = "[MAIN]";
   int sa = sizeof(a);
   if (access(inipfn, F_OK)) { printf("%s file not found\n", inipfn); return(1); }
   
   if (!GetIniString(main, "Debug", "stdout", a, sa, inipfn))
   {  PrtIniErrorString(stdout, "ini file error");
      return(1);
   }
   debug_fp = debug_on(a);
   
   GetIniString(main, "HostName", host_name, host_name, sizeof(host_name), inipfn);
   GetIniInt(main, "ListenPort", listen_port, &listen_port, inipfn);
   GetIniString(main, "PeerName", peer_name, peer_name, sizeof(peer_name), inipfn);
   GetIniString(main, "PeerIP",   peer_ip,   peer_ip,   sizeof(peer_ip),   inipfn);
   GetIniString(main, "SQLini", sqlini, sqlini, sizeof(sqlini), inipfn);
   if (sqlSetIniPfn(sqlini)) PrtSqlErrorString(NULL, debug_fp, "read_ini");
   GetIniString(main, "Query", NULL, a, sa, inipfn);
   for (i=0, p=strtok(a,","); p; p=strtok(NULL,","), i++) STR(db[i],p)
   if (!db[0]) { printf("No Query entry in ini\n"); return (1); }
   return (0);
}

/*----------------------------------------------------------- Close a socket */
void cleanup(int * sock)
{  int rc;
   if(*sock) 
   {  rc = shutdown(*sock, 2);
      if (rc) dbge("shutdown socket");
      rc = close(*sock);
      if (rc) dbge("close socket");
      FD_CLR(*sock, &rsock);
      FD_CLR(*sock, &wsock);
      FD_CLR(*sock, &esock);
      *sock = 0;
   }
}

/******************************************************************************
 * Add/Remove start and stop message characters (11 & 28)
 * Presume that msg is null terminated string with enough room to add
 * the two characters if needed.
 *****************************************************************************/
void wrap(char * msg)
{  if (msg[0] != SB) { memmove(msg+1,msg,strlen(msg)+1); msg[0] = SB; }
   if (LCHR(msg) != EB) strcat(msg,SEB);
}
void unwrap(char *msg)
{  if (msg[0] == SB) memmove(msg,msg+1,strlen(msg)+1);
   if (LCHR(msg) == EB) LCHR(msg) = 0;
}

/*----------------------------------------- Debug to dump buffer area */
void dbgm(char*msg, char*buff)
{  char p[53], *ptr;
   int piece = sizeof(p) - 3, cont = 1;
   if (!debug_fp) return;
   if (strlen(msg) > (sizeof(p) -3)) dbg(msg);
   else
   {   CLEAR(p); strcpy(p,msg);
       if (p[strlen(p)-1] != ' ')p[strlen(p)] = ' ';
       memset(p+strlen(p),'-',(sizeof(p)-strlen(p)-1));
       dbg(p);
   }
  
   if (!buff) { dbg("NULL buffer\n"); return; }
   while (cont)
   {  CLEAR(p); strcpy(p,"> ");
      if (strlen(buff) < piece) { strcat(p,buff); cont = 0; }
      else { memcpy(p+2,buff,piece); buff += piece; }
      if (ptr = strchr(p,SB)) *ptr = '.';
      if (ptr = strchr(p,EB)) *ptr = '.';
      while (ptr = strchr(p,'\r')) *ptr = '~';
      dbg(p);
   }
   dbg(" ");
}

