/***********************************************************************
 * Repair.c                                                    R Moulton
 *
 * The purpose of Repair is to carry out basic "fix" actions on the 
 * Image distributor when problems are recognized by the monitor program 
 * on Rentgn.  Repair runs on Image Distributors and communicates with
 * Rentgn as a TCP/IP client.  It identifies itself to Rentgn and waits
 * for a command message.  When messages are received, they are carried
 * out.  Currently supported messages are:
 *
 * To query for servernames: 
 * format:   msg=SERVERNAMES&
 *
 * To bounce a Dealer script:  (Use external queue #)
 * format:   msg=BOUNCE&dealer=queuename&
 * example:  msg=BOUNCE&dealer=bond3&
 *
 * To bounce a Router script:
 * format:   msg=BOUNCE&router=servername&
 * example:  msg=BOUNCE&router=fleming&
 *
 * To bounce an EchoTest script:
 * format:   msg=BOUNCE&echotest=servername&
 * example:  msg=BOUNCE&echotest=bond&
 *
 * To bounce a dicom receiver:
 * format:   msg=BOUNCE&receiver=servername&
 * example:  msg=BOUNCE&receiver=bond&
 *
 * To reboot the server: (This one is coded, but disabled for now)
 * format:   msg=BOUNCE&server=servername&
 * example:  msg=BOUNCE&server=bond&
 *
 * To hold a study:
 * format:   msg=HOLD&study=studyid&
 * example:  msg=HOLD&study=1.3.12.2.1107.5.8.2.4848...&
 *
 * Program invocation:  Repair 
 *
 * Repair looks for the following in the indicated environment variables,
 * which may be defined by the user or defaulted:
 *
 * HOSTNAME   = host computer name, default = rentgn
 * HOSTIP     = host computer IP address, default = 128.252.118.40
 * PORTNUM    = port number on host, default = 45001
 * SNAME      = Distributor server name (e.g. fleming)
 *              If this environment variable is not set, Repair will 
 *              extract the server name from the DISPLAY system variable
 * SERVERNAMES= Comma delimited list of all logical Distributor servers
 *              currently running on this machine (default SNAME)
 *              used to accept messages for more than one logical server
 *              if they are running together.
 * SENDSERVERS= Send servernames to server on connect (1,Y, or y = yes
 *              other = no, default = no)
 * HOMEDIR    = The user home directory, default is:
 *              /opt/home/erl/spectrum
 * LOGGING    = is logging turned on (1,Y, or y = yes, other = no)
 *              default is on.
 *
 * Repair looks for the following files in the production tmp directory,
 * taking the indicated action if they are found:
 *
 * holdRepair = continues to run and receives messages, but does not
 *              process them.  Prints them to stdout and logs them.
 * exitRepair = terminates the program.
 *
 ************************************************************************
 *
 * Compile and link with: gcc Repair.c -o Repair -lsocket -lnsl
 *
 *-----------------------------------------------------------------------
 ***********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

/***********************************************************************
 * Valid distributor names, message types, environ var names
 **********************************************************************/
#define BOND           "bond"
#define GOLDFINGER     "goldfinger"
#define FLEMING        "fleming"
#define BIGBEN         "bigben"
#define RELAY1         "relay1"
#define RELAY2         "relay2"
#define BOUNCE         "BOUNCE"
#define HOLD           "HOLD"
#define MSG            "msg"
#define DEALER         "dealer"
#define ROUTER         "router"
#define ECHOTEST       "echotest"
#define RECEIVER       "receiver"
#define SERVER         "server"
#define STUDY          "study"

#define HOSTNAME       "HOSTNAME"
#define HOSTIP         "HOSTIP"
#define PORTNUM        "PORTNUM"
#define SNAME          "SERVERNAME"
#define SERVERNAMES    "SERVERNAMES"
#define SENDSERVERS    "SENDSERVERS"
#define HOMEDIR        "HOMEDIR"
#define LOGGING        "LOGGING"
#define DISPLAY        "DISPLAY"

/***********************************************************************
 * Message area and parameters
 **********************************************************************/
char message[256]          = "null";
char parameter_id[10]      = "null";
char parameter_value[256]  = "null";
int  msg_code              = 0;     /* 1=BOUNCE, 2=HOLD, 3=SENDSERVERS, 0=UNKNOWN */
int  bounce_code           = 0;     /* 1=DEALER, 2=ROUTER, 3=ECHOTEST, 4=RECEIVER, 5=SERVER, 0=UNKNOWN */
char bounce_id[15]         = "null";
char study_id[128]         = "unknown";

char host_name[24]         = "rentgen";
char host_ip[]             = "128.252.118.40";
char port_num[10]          = "45001";
char home_dir[128]         = "/opt/home/erl/spectrum";
char bin_dir[128]          = " ";
char log_dir[128]          = " ";
char log_pfn[128]          = " ";
char tmp_dir[128]          = " ";
char hold_pfn[128]         = " ";
char exit_pfn[128]         = " ";
char job_pfn[128]          = " ";
char server_name[16]       = "bond";
char server_names[128]     = "bond";
int send_servers           = 0;     /* 1=on, 0 = off */
int  logging               = 0;     /* 1=on, 0 = off */
/***********************************************************************
 * Other global working storage
 **********************************************************************/
char a[512];
time_t timer;
time_t * ptimer = &timer; 
struct tm * ptime;
int monitor_socket, i, j, retval;
struct hostent *pmonitorent, monitorent;
struct sockaddr_in monitor_sock_name;
fd_set read_socket_set, write_socket_set, error_socket_set;
struct timeval timeout;

/***********************************************************************
 * Get current time
 **********************************************************************/
get_time_hack()
{  timer = time(ptimer);
   ptime = localtime(ptimer);
}

/***********************************************************************
 * Add/Remove start and stop message characters (11 & 28)
 **********************************************************************/
Wrap()
{  char buffer[256];
   memset(buffer,0,sizeof(buffer));
   buffer[0] = 11;
   strcat(buffer,message);
   buffer[strlen(buffer)] = 28;
   strcpy(message,buffer);
}
Unwrap()
{  char buffer[256];
   int i, j;
   memset(buffer,0,sizeof(buffer));
   for (i=0,j=0; message[i]; i++)
      if (message[i] != 11 && message[i] != 28) buffer[j++] = message[i];
   strcpy(message,buffer);
}

/***********************************************************************
 * Log ASCII message in argument to log file, if logging is turned on.
 * Log files are created in the server production log directory, using
 * pfn = RepairYYYYMM.log.  Date and time is added to front and 
 * carriage return to end.
 **********************************************************************/
log (const char * message)
{  FILE* fp;
   char ctime[24];
   printf("log: %s\n", message);
   if (!logging) return;
/*------------------------------------------ Get current log file pfn */
   get_time_hack();
   strftime(ctime,7, "%Y%m", ptime);
   sprintf(log_pfn,"%s/Repair%s.log", log_dir, ctime); 
/*---------------------------------------------- Open file for append */
   fp = fopen(log_pfn, "a");
   if (!fp) { perror("open log:"); return; }
/*------------------------------------------------ log time component */
   strftime(ctime,21 , "%Y-%m-%d %H:%M:%S", ptime);
   fprintf(fp, "%s %s\n", ctime, message);
   fclose(fp);
   return;
}

/***********************************************************************
 * Connect socket, disconnect and reconnect 
 **********************************************************************/
connect_socket(int flag)
{  int size;
   size = sizeof (monitor_sock_name);
   if (flag) 
   {  retval = shutdown(monitor_socket, 2);
      if (retval) { perror("shutdown"); }
   }
   while (1)
   {   retval = connect(monitor_socket, (struct sockaddr *)&monitor_sock_name, size);
       if (!retval || errno == EISCONN) break; 
       else { perror("connect"); sleep(60); }
   }
}

/***********************************************************************
 * close out socket
 **********************************************************************/
cleanup()
{  retval = shutdown(monitor_socket, 2);
   if(retval != 0) perror("shutdown");
   retval = close(monitor_socket);
   if(retval) perror("close");
   log ("shutting down");
}

/***********************************************************************
 * Pull environment variables and validate command line parameters
 **********************************************************************/
validate_params()
{  char * value;
   char chr;
/*--------------------------- Pull environment parameters, if present */
   value = getenv(HOSTNAME); if (value) strcpy(host_name,value);
   value = getenv(HOSTIP);   if (value) strcpy(host_ip,value);
   value = getenv(PORTNUM);  if (value) strcpy(port_num,value);
   value = getenv(HOMEDIR);  if (value) strcpy(home_dir,value);
/*----------------------------------------------------------------------
 * Logging is based on the first character of the environment variable,
 * with 1, Y, or y signifying that logging is on.
 *--------------------------------------------------------------------*/
   logging = 1;
   value = getenv(LOGGING);
   if (value)
   {  chr = *value;
      if (chr == '1' || chr == 'Y' || chr == 'y') logging = 1;
      else logging = 0;
   }
/*----------- For server name, the default is in the DISPLAY variable */
   value = getenv(SNAME);
   if (value) strcpy(server_name,value);
   else
   {  value = getenv(DISPLAY);
      if (value)
      {  strtok(value,":");
	 strcpy(server_name,value);
      }
      else
      {  printf("Unable to determine server name.\n");
	 exit (1);
      }
   }
/*------------------------------- SERVERNAMES, defaults to SNAME */
   strcpy(server_names,server_name);
   value = getenv(SERVERNAMES);
   if (value) strcpy(server_names,value);
/*----------------------------- Send server names on initial connect? */
   value = getenv(SENDSERVERS);
   if (value)
   {  chr = * value;
      if (chr == '1' || chr == 'Y' || chr == 'y') send_servers = 1;
   } 
/*-------------------------------------------  bin directory pathname */
   sprintf(bin_dir, "%s/production/%s/bin", home_dir, server_name);

/*-------------------------------------------  log directory pathname */
   sprintf(log_dir, "%s/production/%s/log", home_dir, server_name);

/*------------------  tmp directory pathname, hold and exit filenames */
   sprintf(tmp_dir, "%s/production/%s/tmp", home_dir, server_name);
   sprintf(hold_pfn, "%s/holdRepair", tmp_dir); 
   sprintf(exit_pfn, "%s/exitRepair", tmp_dir);
   sprintf(job_pfn,  "%s/Repair.csh", tmp_dir);
   return;
}

/***********************************************************************
 * process message from host
 **********************************************************************/
process_message()
{  char * token;
   FILE * fp;
   int retv;
/*----------------- If holdRepair present, log and print message only */
   if (!access(hold_pfn, F_OK))
   {  log("holdRepair file found, message not processed:");
      log(message);
      printf("%s\n", message);
      return;
   }	 
/*------------------------- parse the message, log and exit on errors */
   sprintf(a, "message received: %s", message); log(a);
/*-------------------------------------------------------------- msg= */
   token = strtok(message, "=");
   if (!token || strcmp(token, MSG)){log("Error: msg not found"); return;}
/*------------------------------------------------- Message type code */
   token = strtok(NULL, "&");
   if (!token){log("Error: msg code not found"); return;}
   msg_code = 0;
   if (!strcmp(token, BOUNCE)) msg_code = 1;
   else if (!strcmp(token, HOLD)) msg_code = 2;
   else if (!strcmp(token, SENDSERVERS)) msg_code = 3;
   else {log("Error: invalid msg code"); return;}
/*-------------------------------------- Rest depends on message code */
   switch (msg_code)
   {  case 1:             /* bounce */
	 bounce_code = 0;
         token = strtok(NULL, "=");
	 if (!token){log("Error: bounce type code not found"); return;}
	 if (!strcmp(token, DEALER)) bounce_code = 1;
	 else if (!strcmp(token, ROUTER)) bounce_code = 2;
	 else if (!strcmp(token, ECHOTEST)) bounce_code = 3;
	 else if (!strcmp(token, RECEIVER)) bounce_code = 4;
	 else if (!strcmp(token, SERVER)) bounce_code = 5;
         else {log("Error: invalid bounce type code"); return;}

	 token = strtok(NULL, "&");
	 if (!token){log("Error: bounce id not found"); return;}
	 strcpy(bounce_id, token);
	 break;
      case 2:            /* hold */
	 token = strtok(NULL, "=");
	 if (!token){log("Error: study code not found"); return;}
	 token = strtok(NULL, "&");
	 if (!token){log("Error: study id not found"); return;}
	 strcpy(study_id, token);
	 break;
      case 3:           /* servernames */
         break;
      default:
	 log("Error: invalid message_code value");
	 return;
   }
   
/*----------------------------------------------------------------------
 * Message parsed OK.  Process it. Build csh command call, or send
 * response back to server.
 *--------------------------------------------------------------------*/
   memset(message,0,sizeof(message));
   switch (msg_code)
   {  case 1:                                   /* bounce */
   
         switch (bounce_code)
	 {  case 1:                             /* bounce dealer */
	       strcpy(a, bounce_id);
	       token = strtok(a,"0123456789");
	       sprintf(message, "%s/production/%s/bin/Bounce.csh Dealer %s",
		       home_dir, token, bounce_id);
	       break;
            case 2:                            /* bounce router */
	       sprintf(message, "%s/production/%s/bin/Bounce.csh Router %s",
		       home_dir, bounce_id, bounce_id);
               break;
            case 3:                           /* bounce echotest */
	       sprintf(message, "%s/production/%s/bin/Bounce.csh EchoTest %s",
		       home_dir, bounce_id, bounce_id);
               break;
	    case 4:                             /*bounce receiver */
	       sprintf(message, "%s/BounceReceiver.csh", bin_dir);
	       break;
	    case 5:                             /*bounce server */
/*--------------------------------------------------------------- Disabled */
               log("Bounce server message processing disabled"); return;
	       sprintf(message, "%s/BounceServer.csh", bin_dir);
	       break;
	    default:
	       log ("Error: invalid bounce_code");
	       break;
	 }
	 break;
      case 2:                                  /*hold study*/
	 sprintf(message, "%s/HoldStudy.csh %s %s >> %s", bin_dir, server_name, study_id, log_pfn);
	 break;
      case 3:                      /* send servernames back to server */
         memset(message,0,sizeof(message));
         sprintf(message, "msg=SERVERNAMES&%s&", server_names);
         send_message();
         return;
      default:
	 log("Error: invalid bounce_code");
	 break;
   } 
   if (!message[0]) {log("Error: command build failed"); return; }

/*---------------------------------------------------- Open job cl file */
   fp = fopen(job_pfn, "w");
   if (!fp) {log("Error: job file open failed"); return; }
   
/*------------------------------------------- write the csh file lines */
   fprintf(fp, "#!/bin/csh\n");
   fprintf(fp, "%s >>& %s\n", message, log_pfn);
   fclose(fp);
   sprintf(a, "chmod 775 %s", job_pfn); system(a);
   log("Command submitted");
/*------------------------------------- Temp just show it; for testing */
   sprintf(a, "cat %s", job_pfn); system(a);
   /* retv = system(job_pfn); */
   sprintf(a, "Return status: %d", retv); log(a);
}
   
/***********************************************************************
 * Send message to socket
 **********************************************************************/
send_message()
{  while (1)
   {  sprintf(a, "write message: %s", message); log(a);
/*------------------------------------------- Set ten minute time out */
      timeout.tv_sec = 600;
      timeout.tv_usec = 0;
/*-------------------------------------------- select for read, error */
      FD_ZERO (&write_socket_set);
      FD_SET (monitor_socket, &write_socket_set);
      FD_ZERO (&error_socket_set);
      FD_SET (monitor_socket, &error_socket_set);

      retval = select(j, NULL, &write_socket_set, &error_socket_set, &timeout);

/*----------------------------------- select error (not socket error) */
      if (retval < 0)
      {  log ("select error");
	 perror("select");
         cleanup();
	 exit(1);
      }
      
/*---------------------------------------------------------- time out */
      if (retval == 0)
      {  log("select timeout");
         continue;
      }

/*---------------------------------------------- socket error pending */
      if (FD_ISSET(monitor_socket, &error_socket_set))
      {  log("socket error");
	 connect_socket(1);
	 continue;
      }

/*------------------------------------------- socket read NOT pending */
      if (!FD_ISSET(monitor_socket, &write_socket_set))
      {  log("socket write error");
	 connect_socket(1);
	 continue;
      }
/*-----------------------------------------------------  send message */
      Wrap();
      retval = send(monitor_socket, message, strlen(message) + 1, 0);
      if(retval == -1)
      {  log("send error");
         perror("send");
         connect_socket(1);
         continue;
      }
      return;
   }
}

/***********************************************************************
 * main function
 **********************************************************************/
int main(int numpars, char *par[])
{  
   log ("Repair starting");
/*----------------------------------------------------------------------
 * get a socket
 *--------------------------------------------------------------------*/
   monitor_socket = socket(AF_INET, SOCK_STREAM, 0);
   if(monitor_socket == -1) 
   {  log("socket error"); 
      perror("socket"); 
      exit(1);
   }
/*----------------------------------------------------------------------
 * Set socket to non-blocking.  This is done so that later, when we
 * connect, we can avoid a long wait if the monitor is down.
 *--------------------------------------------------------------------*/
   i = fcntl(monitor_socket, F_GETFL, 0);
   fcntl(monitor_socket, F_SETFL, O_NONBLOCK|i);
/*----------------------------------------------------------------------
 * get host by name, or by IP if name is unknown
 *--------------------------------------------------------------------*/
   pmonitorent = gethostbyname(host_name);
   if(pmonitorent == NULL) pmonitorent = gethostbyname(host_ip);
   if(pmonitorent == NULL) 
   {  log("gethosterror");
      perror("gethost");
      cleanup();
      exit(1);
   }
/*----------------------------------------------------------------------
 * load socket address struct
 *--------------------------------------------------------------------*/
   monitorent = *pmonitorent;
   monitor_sock_name.sin_family = monitorent.h_addrtype;
   monitor_sock_name.sin_port = htons(atoi(port_num));
   monitor_sock_name.sin_addr = *((struct in_addr *)monitorent.h_addr);
/*----------------------------------------------------------------------
 * initial connection to monitor host. 
 *--------------------------------------------------------------------*/
   connect_socket(0);
   j = monitor_socket + 1;

/*--------------------------------- Send server names message to host */
   if (send_servers)
   {  memset(message,0,sizeof(message));
      sprintf(message, "msg=SERVERNAMES&%s&", server_names);
      send_message();
   }
/***********************************************************************
 * Main processing loop, looks for messages and processes them when it
 * finds them.
 **********************************************************************/
   while (1)
   {
/*------------------------------------------- Set ten minute time out */
      timeout.tv_sec = 600;
      timeout.tv_usec = 0;
/*-------------------------------------------- select for read, error */
      FD_ZERO (&read_socket_set);
      FD_SET (monitor_socket, &read_socket_set);
      FD_ZERO (&error_socket_set);
      FD_SET (monitor_socket, &error_socket_set);

      retval = select(j, &read_socket_set, NULL, &error_socket_set, &timeout);

/*----------------------------- If exitRepair file present, shut down */
      if (!access(exit_pfn, F_OK))
      {  log ("exit file found, shutting down");
	 cleanup();
	 exit(0);
      }

/*----------------------------------- select error (not socket error) */
      if (retval < 0)
      {  log ("select error");
	 perror("select");
         cleanup();
	 exit(1);
      }
      
/*---------------------------------------------------------- time out */
      if (retval == 0)
      {  log("select timeout");
         continue;
      }

/*---------------------------------------------- socket error pending */
      if (FD_ISSET(monitor_socket, &error_socket_set))
      {  log("socket error");
	 connect_socket(1);
	 continue;
      }

/*------------------------------------------- socket read NOT pending */
      if (!FD_ISSET(monitor_socket, &read_socket_set))
      {  log("socket read error");
	 connect_socket(1);
	 continue;
      }
/*--------------------------------------------------  receive message */
      memset(message,0,sizeof(message));
      retval = recv(monitor_socket, message, sizeof(message), 0);
      if(retval == -1)
      {  log("receive error");
         perror("recv");
         connect_socket(1);
         continue;
      }
/*--------------------------------- Zero length packet, server closed */
      if(retval == 0)
      {  sleep(60);
         connect_socket(1);
         continue;
      }
/*------------------------------------------------------ Real message */
      Unwrap();
      process_message();
   }
}
