/****************************************************************************************
 * map.c                                                                        R Moulton
 * Library routines to allow simple mapping functions on DICOM headers of DCM_OBJECT.
 ***************************************************************************************/

/*---------------------------------------------------------------------- Unix includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/*--------------------------------------------------------------- CTN library includes */
#include "ctn_os.h"
#include "dicom.h"
#include "utility.h"
#include "dicom_uids.h"
#include "lst.h"
#include "condition.h"
#include "dulprotocol.h"
#include "dicom_objects.h"
#include "dicom_messages.h"
#include "dicom_services.h"
#include "dbquery.h"
#include "iap.h"

/*------------------------------------------------------------- MIRCS library includes */
#include "dbg.h"
#include "ini.h"
#include "log.h"

#include "map.h"

/*------------------------------ function prototypes ----------------------------------*/
static int is_header(char * line);
static void chomp(char * a);
static void free_pmtag(PMTAG);
static void free_pmi(PMI);
static void free_pmt(PMT);
static void free_pmlt(PMLT);
static void free_pmm(PMM);
static void free_pmap(PMAP);


/*----------------------------------------------------------------------------- Macros */
#define STR(A,B) { if (A) free(A); A=strdup(B); }
#define FREE(A)       if (A) { free(A);       A=NULL; }
#define FREE_PMTAG(A) if (A) { free_pmtag(A); A=NULL; }
#define FREE_PMI(A)   if (A) { free_pmi(A);   A=NULL; }
#define FREE_PMT(A)   if (A) { free_pmt(A);   A=NULL; }
#define FREE_PMLT(A)  if (A) { free_pmlt(A);  A=NULL; }
#define FREE_PMM(A)   if (A) { free_pmm(A);   A=NULL; }
#define STR_EQ(A,B) !strcasecmp(A,B)
#define ISON(A) (!strcasecmp(A,"ON") || !strcasecmp(A,"YES"))
#define LCHR(A) A[strlen(A) -1]

/*-------------------------------- Error msg strings. Match to ini_errno values */
static char * e_msg[] = 
{	"No error",
   "No .ini file name",
   ".ini file not found",
   "Can't read .ini file",
   "No .ini section name",
   "No such .ini section",
   ".ini file error",
   "mapping not on",
   "memory allocation error",
   ".map file not found",
   "Can't read .map file",
   "Can't open .map file",
   "Internal error",
   "No Tags defined in .map file",
   "'Tag' line: invalid format",
   "'Tag' line: invalid name",
   "'Tag' line: duplicate name",
   "'Tag' line: invalid group",
   "'Tag' line: invalid element",
   "Header line invalid format",
   "'LogTag' tag undefined",
   "'If' line: invalid format",
   "'If' line: tag undefined",
   "'If' line: invalid relation",
   "'If' line: invalid logical",
   "'THEN:' expected, not found",
   "'Then' line: invalid format",
   "'Then' line: tag undefined",
   "'Then' line: '=' expected, not found",
   "Incomplete map definition",
   "Error reading .map file"
/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*/
};
static int max_e_msg = sizeof(e_msg) / sizeof(char *);
static char * no_msg = "Unknown map_errno value";


/****************************************************************************************
 * Mapping setup. Loads all the mapping structures and prepares for mapping functions.
 * Returns a pointer to the map struct if mapping is on, NULL if not (or if an error 
 * occurred.)
 ***************************************************************************************/


PMAP mapSetup(char * section, char * inipfn)
{  char a[1024];
   int sa = sizeof(a);
/*---------------------------------------------------------------------------------------
 * struct pointers.  First is one we are building, and which may have to be cleared at
 * exit.  tag2 is used for stepping through pointer lists, and will not have to be 
 * cleared.
 *-------------------------------------------------------------------------------------*/
   PMTAG pmtag = NULL, pmtag2;
   PMI   pmi   = NULL, pmi2;
   PMT   pmt   = NULL, pmt2;
   PMLT  pmlt  = NULL, pmlt2;
   PMM   pmm   = NULL, pmm2;
   PMAP  pmap  = NULL;
   
   FILE * map_fp = NULL;
   char * map_pfn = NULL, * def_log_name = NULL;
   char * p;
   enum _state_ { dicom_tags, map_header, map_if, map_ifovr, map_then, map_off, map_eof } state;
   char valid_map, valid_tag, valid_if, valid_then, keep_reading;
/*@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@*/
   
   map_errno = MAP_OK;
   
/*----------------------------------------------- Validate .ini file and section name */
   if (!inipfn) { map_errno = MAP_NOINI; goto ext; }
   if (access(inipfn, F_OK)) { map_errno = MAP_ININOF; goto ext; }
   if (access(inipfn, R_OK)) { map_errno = MAP_INIREAD; goto ext; }
   if (!section) { map_errno = MAP_NOINISEC; goto ext; }
/*-------------------------------------------------------------------- Is mapping on? */
   if (!GetIniString(section, "Status", "OFF", a, sa, inipfn))
   {  if (ini_errno == INI_SECNOF) { map_errno = MAP_SECNOF; goto ext; }
      map_errno = MAP_INIERR; goto ext; 
   }
   if (!ISON(a)) { map_errno = MAP_NOTON; goto ext; }
   
/*--------------------------------------------------------- Create overall map struct */
   pmap = malloc(SMAP);
   if (!pmap) { map_errno = MAP_MALLOC; goto ext; }
   memset ((void *)pmap,0,SMAP);
 
   if (!GetIniString(section, "Debug", "stdout", a, sa, inipfn))
   {  PrtIniErrorString(stdout, "ini file error");
      map_errno = MAP_INIERR; goto ext;
   }
   pmap->debug_fp = debug_on(a);

   STR(pmap->inipfn,inipfn)
   
/*--------- Default log name, can be overidden in a mapping, if nof no default logging */
   GetIniString(section,"LogName","NOF",a,sa,inipfn);
   if (strcasecmp(a,"NOF")) STR(def_log_name,a)

   GetIniString(section,"MapFile","mircs.map",a,sa,inipfn);
   if (access(a, F_OK)) { map_errno = MAP_MAPNOF;  goto ext; }
   if (access(a, R_OK)) { map_errno = MAP_MAPREAD; goto ext; }
   STR(map_pfn,a)
   if (!(map_fp = fopen(map_pfn,"r"))) { map_errno = MAP_MAPOPEN; goto ext; }   

/*---------------------------------------------------------------------------------------
 * .map file is processed using a "next state" paradigm. States are:
 *    dicom_tags - Processing dicom tag entries at beginning of file.
 *    map_header - Processing a new map header, from [name] up to IF:
 *    map_if     - Processing conditional lines, from IF: to last if line before THEN:
 *    map_ifovr  - We have seen the last if, and now need to find the THEN:
 *    map_then   - Processing changes, from THEN: to next map or EOF
 *    map_off    - This map is not ON, skipping
 *    map_eof    - End of file or error has been found on .map file
 * --------------------------------------------------------------------------------------
 * Flags are:
 *    valid_map  - True if one+ valid mappings have been found/loaded.
 *    valid_tag  - True if one+ valid tag entries have been found/processed.
 *    valid_if   - True is one+ valid conditional lines found/prs for this mapping.
 *    valid_then - True if one+ valid change lines found/processed for this mapping.
 * --------------------------------------------------------------------------------------
 * Any errors in the .map file will result in it not being processed.  However, maps 
 * which are not ON are not parsed, just skipped to the next map or EOF
 *-------------------------------------------------------------------------------------*/

   state = dicom_tags;
   valid_tag = valid_map = 0;
   keep_reading = 1;
   while (keep_reading)
   {  if (!fgets(a,sa,map_fp)) 
      {  strcpy(a,"... end of file ...");
         state = map_eof;
         keep_reading = 0;
      }
      dbg(a);
      chomp(a);
      if (a[0] == '#') continue

      switch (state)
      {
/****************************************************************************************
 * dicom_tags state.  Need to find one or more Tags lines, then the first mapping section
 * name [mapname].
 ***************************************************************************************/
         case dicom_tags:
/*----------------------------------------------- First mapping header line: [mapname] */
            if (is_header(a))
            {  if (valid_tag)
               {  pmm = malloc(SMM);
                  if (!pmm) { map_errno = MAP_MALLOC; goto ext; }
                  memset((void *)pmm,0,SMM);
                  p = strtok(a+1,"]");
                  STR(pmm->name,p)
                  STR(pmm->log_name,def_log_name)
                  pmm->log_level = 1;
                  state = map_header;
                  valid_if = valid_then = 0;
                  break;
               }
               else { map_errno = MAP_NOTAGS; goto ext; }
            }
/*------------------------- Should be looking at a line like: Tag name (group,element) */
            {  char n[strlen(a)] = "\0";
               unsigned int group = 0, element = 0;
               if(sscanf(a,"Tag %s (%x,%x)",n,&group,&element) != 3){map_errno=MAP_INVTAG; goto ext;}
/*--------------------------------------------------------------------------- tag name */
               if (!strlen(n)) { map_errno = MAP_TAGNAME; goto ext; }
/*--------------------------------------------------------------- Check for duplicates */
               for (pmtag2 = pmap->pmtag; pmtag2; pmtag2 = pmtag2->pmtag)
               {  if (STR_EQ(p,pmtag2->name))
                  {  map_errno = MAP_TAGDUP; 
                     goto ext;
                  }
               }
               if (!group)   { map_errno = MAP_TAGGRP; goto ext; }
               if (!element) { map_errno = MAP_TAGELM; goto ext; }
/*-------------------------------------------------------------------- Make tag struct */
               pmtag = malloc(SMTAG);
               if (!pmtag) { map_errno = MAP_MALLOC; goto ext; }
               memset ((void *)pmtag,0,SMTAG);
               STR(pmtag->name,n)
               pmtag->group = group;
               pmtag->element = element;
               pmtag->e.tag = DCM_MAKETAG(group,element);
               
/*--------------------------------------- Valid tag, add to list in overall map struct */
               valid_tag = 1;                        
               if (!pmap->pmtag) pmap->pmtag = pmtag; else
               {  for (pmtag2 = pmap->pmtag; pmtag2->pmtag; pmtag2 = pmtag2->pmtag) ;
                  pmtag2->pmtag = pmtag;
               }
               pmtag = NULL;
               break;
            }
/****************************************************************************************
 * map_header state.  We have a pmm struct, and we need to fill in the header items. 
 * All of these entries have defaults, so we don't need them, and if multiple entries are
 * present, we will just use the last one.  They are:
 *    Status=[ON|OFF].  (If OFF, we clear the struct and go to MAP_OFF state)
 *    Level=[STUDY|IMAGE]  
 *    LogName=name
 *    LogLevel=#
 *    LogMsg=message
 *    LogTags=tagname,tagname,tagname
 *
 * The LogTag entries all have to be in the tags section, and each one is entered in the 
 * map header
 *    
 * Then we should find the "IF:" line.
 ***************************************************************************************/
         case map_header:
            if (STR_EQ(a,"IF")) { status = map_if; break; }
            if (!strchr(a,"=")) { map_errno = MAP_HDRINV; goto ext; }
            p = strtok(a,"=");
/*------------------- Status=[ON|OFF]  If not ON, clear struct and skip to next header */
            if (STR_EQ(p,"Status"))
            {  p = strtok(NULL," ");
               if (!p || !ISON(p))
               {  FREE_PMM(pmm)
                  state = map_off;
                  break;
               }
            }
/*--------------------------------- Evaluate at Study(0) or Image(1) level (def study) */
            if (STR_EQ(p,"Level"))
            {  pmm->level = 0;
               p = strtok(NULL," ");
               if (p && !strcasecmp(p,"IMAGE")) pmm->level = 1;
               break;
            }
/*---------------------------------------------------------------------------- LogName */
            if (STR_EQ(p,"LogName"))
            {  p = strtok(NULL," ");
               if (p && strlen(p)) STR(pmm->log_name,p)
               break;
            }
/*------------------------------------------------------ LogLevel (integer, default 1) */
            if (STR_EQ(p,"LogLevel"))
            {  p = strtok(NULL," ");
               if (p) pmm->log_level = atoi(p);
               break;
            }
/*------------------------------------------------------------------------ Log message */
            if (STR_EQ(p,"LogMsg"))
            {  p = strtok(NULL,"\n");
               FREE(pmm->log_msg)
               if (p) STR(pmm->log_msg,p)
               break;
            }
/*------------------------------------------------------------------------------ Log Tags
 * Each one must be one of the Tags entries.  What we store is the pointer to that Tags
 * entry, from which we will get the value to print at logging time.
 * For LogTags, tags in an additional entry are added to those already there.
 *-------------------------------------------------------------------------------------*/
            if (strcasecmp(p,"LogTags")) { map_errno = MAP_HDRINV; goto ext; }
/*------------------------------------------- Process each tag in comma separated list */
            for (p=strtok(p,","); p ; p=strtok(NULL,","))
/*--------------------------------------------------- Find it in tag list or error out */
            {  for (pmtag = pmap->pmtag; pmtag; pmtag = pmtag->pmtag)
               {  if (STR_EQ(p,pmtag->name)) break;
               }
               if (!pmtag) { map_errno = MAP_HDRLTAG; goto ext; }
/*--------------------------------------------------------------- Create logtag struct */
               pmlt = malloc(SMLT);
               if (!pmlt) { map_errno = MAP_MALLOC; goto ext; }
               memset((void *)pmlt,0,SMLT);
               pmlt->pmtag = pmtag;
/*------------------------------------------------------ Add it to the map logtag list */
               if (!pmap->pmlt) pmap->pmlt = pmlt; else
               {  for (pmlt2 = pmap->pmlt; pmlt2->pmlt; pmlt2 = pmlt2->pmlt) ;
                  pmlt2->pmlt = pmlt;
               }
               pmlt = NULL;
            }
            break;
/****************************************************************************************
 * Processing lines in the IF: section.  These are of the format:
 *    TAG [==|!=|>|>=|<|<=] [TAG|"string"] [&& or ||]
 * All the tags must be defined.  The last line before the "THEN:" line does not have a
 * logical && or ||. Once that is seen, the next line has to be the THEN:.  We do not
 * currently handle \" or "" in the strings.
 ***************************************************************************************/
         case map_if:
            {  char left_side[strlen(a)] = "\0", relation[strlen(a)] = "\0", 
                  right_side[strlen(a)] = "\0", and_or[strlen(a)] = "\0";
               char quoted = 0;
/*---------------------------------------------------------------------------------------
 * Can't use sscanf, because the right side may be a quoted string parse line 
 *-------------------------------------------------------------------------------------*/
               if (p=strtok(a," ")) strcpy(left_side,p);
               else {map_errno = MAP_IFINV; goto ext; }
               
               if (p=strtok(NULL," ")) strcpy(relation,p);
               else {map_errno = MAP_IFINV; goto ext; }
               
               if (!(p=strtok(NULL," "))) {map_errno = MAP_IFINV; goto ext; }
               if (*p == '"')
               {  p = strtok(NULL,"\"");
                  p = strtok(NULL,"\""):
                  if (p && strlen(p)) {  strcpy(right_side,p); quoted = 1; }
                  else {  map_errno = MAP_IFINV; goto ext; }
               }
               else strcpy(right_side,p);
               
               if (p=strtok(NULL," ")) strcpy(and_or,p);
               else {map_errno = MAP_IFINV; goto ext; }
/*------------------------------------------------------------------- Create if struct */
               pmi = malloc(SMI);
               if (!pmi) { map_errno = MAP_MALLOC; goto ext; }
               memset((void *)pmi,0,SMI);
/*----------------------------- Left side must be a defined tag; We store it's pointer */
               for (pmtag2 = pmap->pmtag; pmtag2; pmtag2 = pmtag2->pmtag)
                  if (STR_EQ(left_side,pmtag2->name)) break;
               if (!pmtag2) { map_errno = MAP_IFTAG; goto ext; }
               pmi->left_pmtag = pmtag2;
/*---------------------------------------------------------------------------------------
 * Relations: == != > < >= < <= (and also => and =<) are mapped to lt, eq, and gt to 
 * indicate which one's of "less than", "equal to", and "greater than" are "TRUE" for 
 * this relation.
 *-------------------------------------------------------------------------------------*/
               pmi->lt = pmi->eq = pmi->gt = 0;
               if (STR_EQ(relation,"==")) { pmi->eq = 1; } else
               if (STR_EQ(relation,"!=")) { pmi->lt = pmi->gt = 1; } else
               if (STR_EQ(relation,">"))  { pmi->gt = 1; } else
               if (STR_EQ(relation,"<"))  { pmi->lt = 1; } else
               if (STR_EQ(relation,">=") || STR_EQ(relation,"=>")) { pmi->gt pmi->eq = 1; } else
               if (STR_EQ(relation,"<=") || STR_EQ(relation,"=<")) { pmi->lt pmi->eq = 1; } else
               { map_errno = MAP_IFREL; goto ext; }
/*---------------------------------------------------------------------------------------
 * Right side is either a tag name or a double quoted string constant
 *-------------------------------------------------------------------------------------*/
               if (quoted) STR(pmi->right_side,right_side)
               else
               {  for (pmtag2 = pmap->pmtag; pmtag2; pmtag2 = pmtag2->pmtag)
                     if (STR_EQ(right_side,pmtag2->name)) break;
                  if (!pmtag2) { map_errno = MAP_IFTAG; goto ext; }
                  pmi->right_pmtag = pmtag2;
               }
/*---------------------------------------------------------------------------------------
 * Logical is either && or ||, except for last (or maybe only) if line, which is blank
 *-------------------------------------------------------------------------------------*/
               if (STR_EQ(and_or,"&&")) { pmi->logical = 1;  } else
               if (STR_EQ(and_or,"||")) { pmi->logical = 0;  } else
               if (!strlen(and_or))     { state = map_ifovr; } else
               { map_errno = MAP_IFREL; goto ext; }
/*---------------------------------------------------------------------------------------
 * Valid IF line, add it to the mapping
 *-------------------------------------------------------------------------------------*/
               valid_if = 1;
               if (!pmm->pmi) pmmm->pmi = pmi; else
               {  for (pmi2 = pmm->pmi; pmi2->pmi; pmi2 = pmi2->pmi) ;
                  pmi2->pmi = pmi;
               }           
               pmi = NULL;
            }
            break;  
/****************************************************************************************
 * This state is for when we have found the last IF line (the one without a logical).
 * The next line must be "THEN:".
 ***************************************************************************************/
         case map_ifovr:
            if (strcasecmp(a,"THEN:")) { map_errno = MAP_NOTHEN; goto ext; }
            state = map_then;
            break;
/****************************************************************************************
 * Processing lines in the THEN: section.  These must be of the form:
 *    TAG = [TAG|"string"]
 * We know we're done when we hit eof or the next header string, but when we do, we need
 * to store the map we are working on now and start a new one.
 ***************************************************************************************/
         case map_then:
/*------------------------------------------------------------------- next header line */
            if (is_header(a))
/*--------------------------------------- store current map if valid, error out if not */
            {  if (pmm && valid_if && valid_then)
               {  valid_map = 1;
                  if (!pmap->pmm) pmap->pmm = pmm else
                  {  for (pmm2 = pmap->pmm; pmm2->pmm; pmm2 = pmm2->pmm) ;
                     pmm2-pmm = pmm;
                  }
                  pmm = NULL;
/*--------------------------------------------------------------------- set up new map */
                  pmm = malloc(SMM);
                  if (!pmm) { map_errno = MAP_MALLOC; goto ext; }
                  memset((void *)pmm,0,SMM);
                  p = strtok(a+1,"]");
                  STR(pmm->name,p)
                  STR(pmm->log_name,def_log_name)
                  pmm->log_level = 1;
                  state = map_header;
                  valid_if = valid_then = 0;
                  break;
               }
               else { map_errno = MAP_INVMAP; goto ext; }
            }
/*************************************************************************** Then line */
            {  char left_side[strlen(a)] = "\0", relation[strlen(a)] = "\0", 
                  right_side[strlen(a)] = "\0";
               char quoted = 0;
/*---------------------------------------------------------------------------------------
 * parse line - Can't use sscanf, because the right side may be a quoted string  
 *-------------------------------------------------------------------------------------*/
               if (p=strtok(a," ")) strcpy(left_side,p);
               else {map_errno = MAP_THENINV; goto ext; }
               
               if (p=strtok(NULL," ")) strcpy(relation,p);
               else {map_errno = MAP_THENINV; goto ext; }
               
               if (!(p=strtok(NULL," "))) {map_errno = MAP_THENINV; goto ext; }
               if (*p == '"')
               {  p = strtok(NULL,"\"");
                  p = strtok(NULL,"\""):
                  if (p && strlen(p)) {  strcpy(right_side,p); quoted = 1; }
                  else { map_errno = MAP_THENINV; goto ext; }
               }
               else strcpy(right_side,p);
/*----------------------------------------------------------------- Create then struct */
               pmt = malloc(SMT);
               if (!pmt) { map_errno = MAP_MALLOC; goto ext; }
               memset((void *)pmt,0,SMT);
/*----------------------------- Left side must be a defined tag; We store it's pointer */
               for (pmtag2 = pmap->pmtag; pmtag2; pmtag2 = pmtag2->pmtag)
                  if (STR_EQ(left_side,pmtag2->name)) break;
               if (!pmtag2) { map_errno = MAP_IFTAG; goto ext; }
               pmt->left_pmtag = pmtag2;
/*---------------------------------------------------------------------------------------
 * Relation must be =
 *-------------------------------------------------------------------------------------*/
               if (strcmp(relation,"=")) { map_errno = MAP_THENEQ; goto ext; }
/*---------------------------------------------------------------------------------------
 * Right side is either a tag name or a double quoted string constant
 *-------------------------------------------------------------------------------------*/
               if (quoted) STR(pmt->right_side,right_side)
               else
               {  for (pmtag2 = pmap->pmtag; pmtag2; pmtag2 = pmtag2->pmtag)
                     if (STR_EQ(right_side,pmtag2->name)) break;
                  if (!pmtag2) { map_errno = MAP_THENTAG; goto ext; }
                  pmt->right_pmtag = pmtag2;
               }
/*---------------------------------------------------------------------------------------
 * Valid IF line, add it to the mapping
 *-------------------------------------------------------------------------------------*/
               valid_then = 1;
               if (!pmm->pmt) pmmm->pmt = pmt; else
               {  for (pmt2 = pmm->pmt; pmt2->pmt; pmt2 = pmt2->pmt) ;
                  pmt2->pmt = pmt;
               }           
               pmt = NULL;
            }
            break;  
            
/****************************************************************************************
 * This mapping is turned off.  Skip all lines until you find the next mapping, or eof
 ***************************************************************************************/
         case map_off:
/*------------------------------------------------------------------- next header line */
            if (is_header(a))
/*--------------------------------------------------------------------- set up new map */
            {  pmm = malloc(SMM);
               if (!pmm) { map_errno = MAP_MALLOC; goto ext; }
               memset((void *)pmm,0,SMM);
               p = strtok(a+1,"]");
               STR(pmm->name,p)
               STR(pmm->log_name,def_log_name)
               pmm->log_level = 1;
               state = map_header;
               valid_if = valid_then = 0;
            }
            break;
            
/****************************************************************************************
 * EOF or error on .map file.  If error, exit, otherwise there should be a valid mapping
 * to store.
 ***************************************************************************************/
         case map_eof:
            if (ferror(map_fp)) { map_errno = MAP_FE; goto ext; }  /* error */
/*-------------------------------------------------------------------- store valid map */
            if (pmm && valid_if && valid_then)
            {  valid_map = 1;
               if (!pmap->pmm) pmap->pmm = pmm else
               {  for (pmm2 = pmap->pmm; pmm2->pmm; pmm2 = pmm2->pmm) ;
                  pmm2-pmm = pmm;
               }
               pmm = NULL;
            }
            if (valid_map) map_errno = MAP_OK;
            else map_errno = MAP_NOMAPS;
            goto ext;
/*************************************************************** should never get here */
         default:
            map_errno = MAP_INT; goto ext;
      }
   }
   
      
/*---------------------------------------------------------------------------------------
 * This is the common exit point for the Setup routine.  If the error value is not "OK",
 * clear any memory which has been allocated to the map. Return map handle, or NULL.
 *-------------------------------------------------------------------------------------*/   
ext:
   FREE(def_log_name)
   if (map_fp) fclose(fp);
   FREE(map_pfn)
   FREE_PMTAG(pmtag)
   FREE_PMI(pmi)
   FREE_PMT(pmt)
   FREE_PMLT(pmlt)
   FREE_PMM(pmm)
   if (map_errno != MAP_OK && pmap)
   {  if (pmap->debug_fp) PrtMapErrorString(pmap->debug_fp,"mapSetup exit");
      FREE_PMAP(pmap); 
   }
   return (pmap);
}
   
  
   
int mapFirstObject(PMAP, DCM_OBJECT * obj) {}
int mapObject(PMAP, DCM_OBJECT * obj) {}
int mapFree(PMAP pmap) { FREE_PMAP(pmap) }

/******************************************************************************
 * return error message string for log_errno, or print to stdout
 *****************************************************************************/
char * GetMapErrorString(void)
{	if (map_errno < max_e_msg) return e_msg[map_errno];
	else return no_msg;
}
void PrtMapErrorString(FILE * file, char * prefix)
{	if (!file) file = stdout;
   fprintf(file, "%s: %s\n", prefix, GetMapErrorString());
	return;
}

/*--------------------------------------------------------------------------------------
 * If argument is a map header line (format [name] starting in col 1) return non-zero,
 * otherwise return 0. a is unchanged. name must be non-zero length, non-blank string.
 *------------------------------------------------------------------------------------*/
static int is_header(char * a)
{  char * b = NULL;
   int i;
/*------------------------------------------------------------- Must be format [name] */
   if (a[0] != '[') return 0;
   if (!strchr(a,']')) return 0;
   STR(b,a);
   p = strtok(b+1,"]");
   CHOMP(p)
   i = strlen(p);
   FREE(b)
   return i;
}

/*---------------------- Remove leading and trailing spaces and new lines from string */
static void chomp(char * a)
{  int i = strspn(a," ");
   memmove(a,a+i,strlen(a+i)+1);
   while (LCHR(a) == ' ' || LCHR(a) == '\n') LCHR(a) = '\0';
}

/*---------------------------------------------------------------------------------------
 * Routines to free up memory malloc'd for structs.  Most are recursive.
 *-------------------------------------------------------------------------------------*/
static void free_pmtag(PMTAG pmtag)
{  if (pmtag)
   {  FREE_PMTAG(pmtag->pmtag)
      FREE(pmtag->name)
      FREE(pmtag->e.d.string)
      FREE(pmtag->study_value)
      FREE(pmtag->image_value)
      FREE(pmtag)
   }
}
   
static void free_pmi(PMI pmi)
{  if (pmi)
   {  FREE_PMI(pmi->pmi)
      FREE_PMTAG(pmi->left_pmtag)
      FREE_PMTAG(pmi->right_tag)
      FREE(pmi->right_side)
      FREE(pmi)
   }
}

static void free_pmt(PMT pmt)
{  if (pmt)
   {  FREE_PMT(pmt->pmt)
      FREE_PMTAG(pmt->left_pmtag)
      FREE_PMTAG(pmt->right_pmtag)
      FREE(pmt->right_side)
      FREE(pmt)
   }
}

static void free_pmlt(PMLT pmlt)
{  if (pmlt)
   {  FREE_PMLT(pmlt->pmlt)
      FREE_PMTAG(pmlt->pmtag)
   }
}

static void free_pmm(PMM pmm)
{  if (pmm) 
   {  FREE_PMM(pmm->pmm)
      FREE(pmm->name)
      FREE(pmm->log_name)
      FREE(pmm->log_msg)
      FREE_PMI(pmm->pmi)
      FREE_PMT(pmm->pmt)
      FREE(pmm)
   }
}

static void free_pmap(PMAP pmap)
{  if (pmap)
   {  if (pmap->debug_fp) fclose(debug_fp);
      FREE(pmap->inipfn)
      FREE(pmap->log_name)
      FREE_PMTAG(pmap->pmtag)
      FREE_PMLT(pmap->pmlt)
      FREE_PMM(pmap->pmm)
      FREE(pmap)
   }
}
