/**************************************************************************************************
 * alert.c                                                                               R Moulton
 * Library of routines which allow sending print, e-mail and pager alerts via a server on rentgen
 *-------------------------------------------------------------------------------------------------
 * Compile using:  gcc -c alert.c
 * Programs linked with this will require -lresolv -lsocket -lnsl
 *-------------------------------------------------------------------------------------------------
 * 10-3 R Moulton Initial write.
 *************************************************************************************************/

#include <arpa/inet.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <unistd.h>

#include "alert.h"
#include "dbg.h"

/*--------------------------------------------------------------------------------------- Macros */
#define MAKE_STR(A,B) if (A=malloc(strlen(B)+3)) strcpy(A,B);
#define FREE_STR(A) if (A) { free(A); A=NULL; }
#define STR_EQ(A,B) !strcasecmp(A,B)

/*------------------------------------------------- Error msg strings. Match to ini_errno values */
static const char * e_msg[] = 
{	"No error",
	"Missing or empty message",
	"No destinations",
	"Too many e-mail addresses",
	"Too many printer queues",
   "Too many process alerts",
	"Memory allocation failure",
	"Message size limit exceeded",
	"Unable to create TCP/IP socket",
	"Unable to gethost by name or ip",
	"Unable to connect to host",
	"Unable to select host for send",
	"Send error",
	"Unable to select host for recv",
	"Receive error",
	"Server returned a NAK",
	"gethostbyname: ",
   "Missing/Invalid signal name",
   "Invalid distributor name"
};
static int max_e_msg = sizeof(e_msg) / sizeof(char *);
static const char * no_msg = "Unknown alert_errno value";

#define TRUE   1
#define FALSE  0
#define SB 11
#define EB 28

static char host_name[24] = "rentgn";
static char host_ip[17] = "128.252.118.40";
static int  host_port = 52226;
static int  secs_to_wait = 60;
static int debug = 1;

static void cleanup(int * socket);

/**************************************************************************************************
 * This is the general purpose routine, which connects to the PACS_WARN server on rentgn and sends
 * the alert message.  e_mail_addresses and printer_queues are character pointer arrays, with a
 * null pointer at the end.
 *************************************************************************************************/
int alertSend(char * e_mail_addresses[], char * printer_queues[], char * process_names[], 
    char * title, char * signal_name, char * message, FILE * fp)
{	char * buff, * p, a[24];
	struct hostent *psrvent, srvent;
	struct sockaddr_in srv_sock_name;
	fd_set socket_set;
	struct timeval timeout;
	int i, j, ss, sock;
	alert_errno = ALERT_OK;
/*-------------------------------------------------------------------- Some quick failure checks */
	if (!message || !strlen(message)) 
	{	alert_errno = ALERT_INVMSG; 
		return (ALERT_R_FAIL);
	}
	if ((!e_mail_addresses[0] || !strlen(e_mail_addresses[0])) && 
		 (!printer_queues[0] || !strlen(printer_queues[0])) &&
       (!process_names[0]  || !strlen(process_names[0])))
	{	alert_errno = ALERT_NODEST; 
		return (ALERT_R_FAIL);
	}
   if (process_names[0] && ! signal_name)
   {  alert_errno = ALERT_SIGNAL;
      return (ALERT_R_FAIL);
   }
/*------------------------------ Further validation, and calculate size of message buffer needed */
	i = 12;                                                       /* <SB>message& ... <EB>\0 */
	for (j=0; e_mail_addresses[j]; j++)
	{	if (j == ALERT_MAX_EMAIL)
		{	alert_errno = ALERT_EMLIM; 
			return (ALERT_R_FAIL);
		}
		i += 8 + strlen(e_mail_addresses[j]);                      /* e-mail=address& */
	}
	for (j=0; printer_queues[j]; j++)
	{	if (j == ALERT_MAX_PRT)
		{	alert_errno = ALERT_PRTLIM; 
			return (ALERT_R_FAIL);
		}
		i += 7 + strlen(printer_queues[j]);                      /* print=queues& */
	}
   for (j=0; process_names[j]; j++)
	{	if (j == ALERT_MAX_PRS)
		{	alert_errno = ALERT_PRSLIM; 
			return (ALERT_R_FAIL);
		}
		i += 9 + strlen(process_names[j]);                      /* process=name& */
	}

   if (signal_name && strlen(signal_name))
      i += 8 + strlen(signal_name);                            /* signal=signal_name& */
      
	if (title && strlen(title)) i += 7 + strlen(title);         /* title=titlemessage& */
	i += 4 + strlen(message);                                   /* msg=message */

	if (i > ALERT_MAX_MSG_SIZE)
	{	alert_errno = ALERT_MAXMSG;  
		return (ALERT_R_FAIL);
	}
/*-------------------------------------------------------------------------------------------------
 * If a FILE * was passed, we log the message to that file
 *-----------------------------------------------------------------------------------------------*/
   if (fp)
   {  for (j=0; e_mail_addresses[j]; j++) fprintf(fp, "email to %s\n", e_mail_addresses[j]);
      for (j=0; printer_queues[j]; j++)   fprintf(fp, "print to %s\n", printer_queues[j]);
      for (j=0; process_names[j]; j++)    fprintf(fp, "sent %s to %s\n", signal_name, process_names[j]);
      if (title) fprintf(fp,"title %s\n", title);
      fprintf(fp, "\n%s\n", message);
   }
/*------------------------------------------------------------------------ Now we build the message
 * Messages are "&" separated fields, of the form:
 *		<SB>message
 *		print=printerqueue                  (queue message to this printer queue)
 *		e-mail=e-mailaddress                (e-mail to this address)
 *		process=process_names               (signal these processes)
 *		signal=signal_name                  (send this signal)
 *		title=titleline                     (e-mail title, if not blank)
 *		msg=messagetext                     (message to be printed/e-mailed)
 *		<EB>
 *	
 *	A single message may be sent to up to MAX_PRT printer queues and up to MAX_EMAILS e-mail 
 *	addresses.  If there are multiple title fields, the first one will be used.  The msg field
 *	must be the last one in the message.  messagetext is the text of the message, including needed
 *	formatting characters.  Ampersands in the messagetext are considered to be part of the message,
 *	not field delimiters.  messagetext ends at the <EB>
 *-----------------------------------------------------------------------------------------------*/
	buff = malloc(i);
	if (!buff)
	{	alert_errno = ALERT_MEM; 
		return (ALERT_R_FAIL);
	}
	memset(buff,0,i);
	
	*buff = SB;
	
	strcat (buff,"message&");
	
	for (j=0; e_mail_addresses[j]; j++)
	{	if (strlen(e_mail_addresses[j]))
		{	strcat(buff,"e-mail=");
			strcat(buff,e_mail_addresses[j]);
			strcat(buff,"&");
		}
	}
	for (j=0; printer_queues[j]; j++)
	{	if (strlen(printer_queues[j]))
		{	strcat(buff,"print=");
			strcat(buff,printer_queues[j]);
			strcat(buff,"&");
		}
	}
   if (!signal_name || strcasecmp(signal_name,"RUN"))
   {  for (j=0; process_names[j]; j++)
	   {	if (strlen(process_names[j]))
		   {	strcat(buff,"process=");
			   strcat(buff,process_names[j]);
			   strcat(buff,"&");
		   }
	   }
   }

	if (title && strlen(title))
	{	strcat(buff,"title=");
		strcat(buff,title);
		strcat(buff,"&");
	}

   if (signal_name && strlen(signal_name))
   {  if (strcasecmp(signal_name,"RUN"))
	   {	strcat(buff,"signal=");
		   strcat(buff,signal_name);
		   strcat(buff,"&");
	   }
      else
      {  strcat(buff,"run=");
         strcat(buff,process_names[0]);
         strcat(buff,"&");
      }
   }

	if (!signal_name || strcasecmp(signal_name,"RUN")) strcat(buff,"msg=");
	strcat(buff,message);

	buff[strlen(buff)] = EB;

/*----------------------------------------------------------- Create socket to connect to server */
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock == -1)
	{	alert_errno = ALERT_SOCKET;
		FREE_STR(buff)
		dbge("socket");
		return (ALERT_R_FAIL);
	}
	i = fcntl(sock, F_GETFL, 0);
	fcntl(sock, F_SETFL, O_NONBLOCK|i);

   dbg("gethostbyname");   
	psrvent = gethostbyname(host_name);
	if (!psrvent)
   {  dbg("gethostbyaddr");
      in_addr_t addr = inet_addr(host_ip);
      if (addr != -1) psrvent = gethostbyaddr((char *)&addr, sizeof(addr), AF_INET);
   }
	if (!psrvent)
	{	alert_errno = ALERT_HOST;
		FREE_STR(buff)
		cleanup(&sock);
		return (ALERT_R_FAIL);
	}
	dbg("got host");
	srvent = *psrvent;
   srv_sock_name.sin_family = srvent.h_addrtype;
   srv_sock_name.sin_port = htons(host_port);
   srv_sock_name.sin_addr = *((struct in_addr *)srvent.h_addr);

/*---------------------------------------------------------------------------- Connect to server */
   dbg("connect");
	ss = sizeof(srv_sock_name);
   i = connect(sock, (struct sockaddr *)&srv_sock_name, ss);
   sprintf(em,"connect returned %d, errno = %d (IP = %d)", i, errno, EINPROGRESS); dbg(em);
   if (i < 0 && errno != EINPROGRESS)         /* We expect the "Operation now in progress error" */
	{	alert_errno = ALERT_CONNECT;
		FREE_STR(buff)
		dbge("connect");
		cleanup(&sock);
		return (ALERT_R_FAIL);
	}

   dbg("connected");
/*----------------------------------------------------------------------- Send message to server */
	FD_ZERO (&socket_set);
   FD_SET (sock, &socket_set);
	timeout.tv_sec = secs_to_wait;
	timeout.tv_usec = 0;
	
   while (1) {
   dbg("send select");
   j = sock + 1; errno = 0;
	i = select(j, NULL, &socket_set, NULL, &timeout);
   sprintf(em,"select returned %d, errno = %d", i, errno); dbg(em);
   if (!errno) break;
   sleep (1);
   }
	if (i <= 0)
	{	alert_errno = ALERT_SELECT;
		FREE_STR(buff)
		dbge("select");
		cleanup(&sock);
		return (ALERT_R_FAIL);
	}

   dbg("send");
   dbg(buff);
	ss = strlen(buff);
	i = send(sock, buff, ss, 0);
   sprintf(em,"send ret = %d", i); dbg(em);
	if (i < 0)
	{	alert_errno = ALERT_SEND;
		FREE_STR(buff)
		dbge("send");
		cleanup(&sock);
		return (ALERT_R_FAIL);
	}
	
   dbg("recv select");
/*----------------------------------------------------------------------------------- ACK or NAK */
	FD_ZERO (&socket_set);
   FD_SET (sock, &socket_set);

	i = select(j, &socket_set, NULL, NULL, &timeout);
	if (i <= 0)
	{	alert_errno = ALERT_RELECT;
		FREE_STR(buff)
		dbge("select 2");
		cleanup(&sock);
		return (ALERT_R_FAIL);
	}
	
   dbg("recv");
	i = recv(sock, a, sizeof(a), 0);
	if (i <= 0)
	{	alert_errno = ALERT_RECV;
		FREE_STR(buff)
		dbge("recv");
		cleanup(&sock);
		return (ALERT_R_FAIL);
	}
   sprintf(em,"recv %s, %d chars", a, i); dbg(em);

   dbg("cleanup");
	cleanup(&sock);
	FREE_STR(buff);
	if (!memcmp(a+1,"ACK",3)) 
	{	alert_errno = ALERT_OK;
		return (ALERT_R_OK);
	}
	else
	{	alert_errno = ALERT_NAK;
		return (ALERT_R_FAIL);
	}
}

void cleanup(int * socket)
{	shutdown(*socket, 2);
	close(*socket);
	*socket = 0;
}
/**************************************************************************************************
 * Courtesy calls, which use alertSend.
 *************************************************************************************************/
static char * null[1] = {NULL};
int alertSendEmail(char * e_mail_address, char * title, char * message, FILE * fp)
{	char * e[2] = {e_mail_address, NULL};
	return alertSend(e, null, null, title, NULL, message, fp);
}
int alertSendPrint(char * printer_queue, char * message, FILE * fp)
{	char * p[2] = {printer_queue, NULL };
	return alertSend(null, p, null, NULL, NULL, message, fp);
}
int alertSendPrintEmail(char * printer_queue, char * e_mail_address, char * title, char * message, FILE * fp)
{	char * e[2] = {e_mail_address, NULL};
	char * p[2] = {printer_queue, NULL};
	return alertSend(e, p, null, title, NULL, message, fp);
}

/**************************************************************************************************
 * Signal call, which permits simple server selection for distributors
 *************************************************************************************************/
#define DIST distributors[dist]
struct _distributors_ {
   char * name;
   char * server_name;
   char * ip;
   int  port;
};
const static struct _distributors_ distributors[] = {
   {  "bond",   NULL,      "10.33.162.35", 52226 },
   {  "bigben", NULL,      "10.33.162.2",  52226 },
   {  "relay1", NULL,      "10.33.162.32", 52226 },
   {  "relay2", NULL,      "10.33.162.33", 52226 },
   {  "relay3", NULL,      "10.33.162.51", 52226 },
   {  "relay4", NULL,      "10.33.162.53", 52226 },
   {  "BOND",   NULL,      "10.33.162.35", 52226 },
   {  "BIGBEN", NULL,      "10.33.162.2",  52226 },
   {  "RELAY1", NULL,      "10.33.162.32", 52226 },
   {  "RELAY2", NULL,      "10.33.162.33", 52226 },
   {  "RELAY3", NULL,      "10.33.162.51", 52226 },
   {  "RELAY4", NULL,      "10.33.162.53", 52226 },
};
static unsigned char max_distributor = sizeof(distributors) / sizeof(struct _distributors_);
static unsigned char dist;
int alertSendSignal(char * signal_name, char * process_name, char * distributor)
{  char msg[128], * p, * prs[2] = { process_name, NULL};
   char hold_host[24], hold_ip[24];
   int hold_port, hold_secs, ret;
/*------------------------------------------------------------------ quick parameter validations */
   if (!signal_name  || !strlen(signal_name)) {alert_errno = ALERT_SIGNAL; return (ALERT_R_FAIL);}
   if (!process_name || !strlen(process_name)){alert_errno = ALERT_PRSLIM; return (ALERT_R_FAIL);}
   if (!distributor  || !strlen(distributor)) {alert_errno = ALERT_DIST;   return (ALERT_R_FAIL);}

/*-------------------------------------------------------- save current alert server parameters */
   strcpy(hold_host,host_name);
   strcpy(hold_ip,host_ip);
   hold_port = host_port;
   hold_secs = secs_to_wait;
/*-------------------------------------------------------- select distributor server from table */
   for (dist = 0; dist < max_distributor; dist++)
   {  if (STR_EQ(DIST.name,distributor))
      {  p = DIST.name;
         if (DIST.server_name) p = DIST.server_name;
         alertSetServer(p, DIST.ip, DIST.port, 15);
         break;
      }
   }
   if (dist == max_distributor) {alert_errno = ALERT_DIST; return (ALERT_R_FAIL); }

/*------------------------------------------------------------------------- Call general routine */
   sprintf(msg,"send %s signal to process %s on %s&", signal_name, process_name, distributor);
   ret = alertSend(null, null, prs, NULL, signal_name, msg, NULL);

/*--------------------------------------------------------------------------------- reset server */
   alertSetServer(hold_host, hold_ip, hold_port, hold_secs);

/*---------------------------------------------------------- return what alertSend call gave you */
   return (ret);
}
int alertSendRun(char * process_name, char * distributor)
{  char msg[128], * p, * prs[2] = { process_name, NULL};
   char hold_host[24], hold_ip[24];
   int hold_port, hold_secs, ret, hold_errno;
/*------------------------------------------------------------------ quick parameter validations */
   if (!process_name || !strlen(process_name)){alert_errno = ALERT_PRSLIM; return (ALERT_R_FAIL);}
   if (!distributor  || !strlen(distributor)) {alert_errno = ALERT_DIST;   return (ALERT_R_FAIL);}

/*-------------------------------------------------------- save current alert server parameters */
   strcpy(hold_host,host_name);
   strcpy(hold_ip,host_ip);
   hold_port = host_port;
   hold_secs = secs_to_wait;
/*-------------------------------------------------------- select distributor server from table */
   for (dist = 0; dist < max_distributor; dist++)
   {  if (STR_EQ(DIST.name,distributor))
      {  p = DIST.name;
         if (DIST.server_name) p = DIST.server_name;
         alertSetServer(p, DIST.ip, DIST.port, 15);
         break;
      }
   }
   if (dist == max_distributor) {alert_errno = ALERT_DIST; return (ALERT_R_FAIL); }

/*------------------------------------------------------------------------- Call general routine */
   sprintf(msg,"send run %s to %s", process_name, distributor);
   ret = alertSend(null, null, prs, NULL, "RUN", msg, NULL);
   hold_errno = alert_errno;
/*--------------------------------------------------------------------------------- reset server */
   alertSetServer(hold_host, hold_ip, hold_port, hold_secs);

/*---------------------------------------------------------- return what alertSend call gave you */
   alert_errno = hold_errno;
   return (ret);
}

/**************************************************************************************************
 * Allows changing alert parameters.  NULL/0 items are not changed
 *************************************************************************************************/
int alertSetServer(char * host, char * ip, int port, int wait_seconds)
{	if (host) strcpy(host_name,host);
	if (ip) strcpy(host_ip,ip);
	if (port) host_port = port;
	if (wait_seconds) secs_to_wait = wait_seconds;
	alert_errno = ALERT_OK;
	return (ALERT_R_OK);
}

/**************************************************************************************************
 * return error message string for alert_errno, or print to open stream
 *************************************************************************************************/
const char * GetAlertErrorString(void)
{	if (alert_errno == ALERT_GHBN) return strerror(h_errno);
	else if (alert_errno < max_e_msg) return e_msg[alert_errno];
	else return no_msg;
}

void PrtAlertErrorString(FILE * file, char * prefix)
{	if (!file) file = stdout;
	fprintf(file, "%s: %s\n", prefix, GetAlertErrorString());
	return;
}

