/**************************************************************************************************
 * fifo.c                                                                                 R Moulton
 * Library of routines supporting pseudo fifo's buffered to stream files on disk.
 * Programming note: This library is loaded once, and used for all fifo's the process accesses.
 * Static items must be generic; All state information must be in the FIFO structure.
 *-------------------------------------------------------------------------------------------------
 * Compile using gcc -c fifo.c ini.c dbg.c -lsocket -lnsl
 *-------------------------------------------------------------------------------------------------
 * 2-4 R Moulton  Initial write
 *************************************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "fifo.h"
#include "ini.h"
#include "dbg.h"

static struct sockaddr_in * setup_server(int port, int * sock);
static int connect_as_client(char * target_host, char * target_ip, int port, int * socket);
static void accept_new_connection(FIFO * fifo);
static int tcpPut(FIFO * fifo, char * msg);
static int tcpGetServer(FIFO * fifo, char * msg, size_t size);
static int tcpGetClient(FIFO * fifo, char * msg, size_t size);
static void close_socket(int * socket);
static FIFO * make_fifo(void);
static void move_out(char * out, char * in, char sbeb);
static int check_child(pid_t pid);

#define PROG_NAME "FifoBuffer"

#define SB 11
#define EB 28

/*--------------------------------------------------------------------------------------- Macros */
#define STR(A,B) {if(A)free(A); if(B && (A=malloc(strlen(B)+10)))strcpy(A,B);}
#define FREE(A) if(A){free(A);A=NULL;}
#define STR_EQ(A,B) !strcasecmp(A,B)
#define STR_NOT_NULL(A,B) {if(A)free(A); if(B && strcasecmp(B,"NULL") && (A=malloc(strlen(B)+10)))strcpy(A,B);}
#define LCHR(A) A[strlen(A) - 1]
#define FNF(A) (access(A, F_OK))

#define CLIENT fifo->clients[i]
#define CLIENTIP fifo->clients[i].ip[j]

/*------------------------------------------- Error message strings; Match to fifo_errno values */
static char * e_msg[] =
{  "No error",
   "NULL/invalid FIFO ptr",
   "FIFO FNF",
   "Invalid I/O direction",
   "Memory allocation error",
   "Wrong permissions",
   "Fork failure",
   PROG_NAME " load failure",
   "pipe open failure",
   "invalid msg area",
   "end of file",
   "I/O error",
   "Connect failure",
   "TCP/IP select failure",
   "TCP/IP retries",
   "Can't launch fifo",
   "No valid server name in .ini",
   ".ini file error",
   "inv # of clients",
   "Can't init server"
};
static int max_e_msg = sizeof(e_msg) / sizeof(char *);
static char * no_msg = "Unknown fifo_errno value";

static const char snd_msg[] = "\013SND\n\034";
static const char rtr_msg[] = "\013RTR\n\034";
static const char rel_msg[] = "\013REL\n\034";
static const char ack_msg[] = "\013ACK\034";

/**************************************************************************************************
 * Opens a buffered fifo. Returns a pointer to the FIFO struct if successful, NULL otherwise.
 *************************************************************************************************/
FIFO * fifoOpen(char * name, char direction, unsigned char flags)
{  FIFO * fifo;
   struct sockaddr_in * ptr;
   FILE * fp, * fp_rtr = NULL;
   int i, j, r, fd, fd_rtr = 0, blocking, hostport, bufsiz, sbeb = 0;
   pid_t pid;
   char pfn[512], ini_pfn[512], a[1024], connect_as, host, hostn[MAXHOSTNAMELEN], hostip[18];
   char * cltn[FIFO_MAX_CLIENTS + 1], * cltip[FIFO_MAX_CLIENTS + 1], * p, * h;
   int sa = sizeof(a);
   if (flags & FIFO_BLOCK) blocking = 0; else blocking = O_NONBLOCK;
      
   fifo_errno = FIFO_OK;                                       /* Clear any previous error value */

   if(direction != FIFO_INPUT && direction != FIFO_OUTPUT)              /* Validate IO direction */
   {  fifo_errno = FIFO_INVDIRECTION; 
      return (NULL);
   }
/*------------------------------------------------------------------ Must have .ini for the fifo */
   strcpy(ini_pfn,name); strcat(ini_pfn,".ini");
   if (FNF(ini_pfn)) { fifo_errno = FIFO_INI; return (NULL); }
/*--------------------------------------------------- Size of buffer, from max record size entry */
   if (!GetIniInt("[Main]","MaxRecordSize",1024,&bufsiz,ini_pfn))
   {  fifo_errno = FIFO_INI;
      return (NULL);
   }
   bufsiz += 8;
/*------------------------------------ On output, should SB & EB characters be added to message? */
   if (direction == FIFO_OUTPUT)
   {  if (!GetIniString("[Main]","SBEB","Yes",a,sa,ini_pfn))
      {  fifo_errno = FIFO_INI;
         return (NULL);
      }
      if (*a == 'Y' || *a == 'y') sbeb = 1;
   }
/*-------------------------------------------------------------------------------------------------
 * Are we the "host" server (i.e., the one running the FifoBuffer prs for this fifo)?
 *-----------------------------------------------------------------------------------------------*/ 
   if (!GetIniString("[Main]","Server","Unknown",a,sa,ini_pfn)) 
   {  fifo_errno=FIFO_SERVER; 
      return(NULL); 
   }
   gethostname(hostn,MAXHOSTNAMELEN);
   host = STR_EQ(a,hostn);
/*-------------------------------------------------------------------------------------------------
 * The .put fifo must exist and be writable, even if it is not used, on the host server, as it 
 * doubles as the process fifo. 
 *-----------------------------------------------------------------------------------------------*/
   if (host)
   {  strcpy(pfn,name); strcat(pfn,".put");
      if (FNF(pfn)) { fifo_errno = FIFO_NOSUCHFIFO; return (NULL); }
      if (access(pfn, W_OK)) { fifo_errno = FIFO_PERMISSIONS; return (NULL); }
/*-------------------------------------------------------------------------------------------------
 * If it exists, open it for write, non-blocking. If this does not work, and AutoLaunch is set, we
 * need to launch the FifoBuffer process for this pipe.  When launching, we hold until we can open 
 * the .put, or until we get a message that the buffer handler failed to load.  
 *-----------------------------------------------------------------------------------------------*/ 
      fd = open(pfn, O_WRONLY | O_NONBLOCK);
      if (fd < 0)
      {  if (!GetIniString("[Main]","AutoLaunch","No",a,sa,ini_pfn)){fifo_errno=FIFO_INI; return(NULL);}
         if (*a != 'Y' && *a != 'y') { fifo_errno = FIFO_LAUNCH; return (NULL); }
         pid = fork();
         if (pid < 0) { fifo_errno = FIFO_FORK; return (NULL); }
         if (pid == 0) if(execlp(PROG_NAME, PROG_NAME, name, (char *)0) < 0) dbge_exit("execlp", 1);
         while (fd < 0)
         {  if (check_child(pid)) { fifo_errno = FIFO_BFAIL; return (NULL); }
            fd = open(pfn, O_WRONLY | O_NONBLOCK);
         }
      }
      close (fd);
   }
/*-------------------------------------------------------------------------------------------------
 * If we are doing puts, and we are on the host, we use the fifo.  If we are not on the host, we
 * connect to the host using TCP/IP.
 *-----------------------------------------------------------------------------------------------*/
   if (direction == FIFO_OUTPUT)
   {  if (host)                                                     /* put via fifo */
      {  strcpy(pfn,name); strcat(pfn,".put");
         fd = open(pfn, O_WRONLY | blocking);
         if (fd < 0) { fifo_errno = FIFO_OPEN; return (NULL); }
         fp = fdopen(fd, "a");
         if (!fp) { fifo_errno = FIFO_OPEN; return (NULL); }
         setvbuf(fp, NULL, _IOLBF, 0);

         fifo = make_fifo();                                     /* make struct, fifo put */
         if (!fifo) { fifo_errno = FIFO_MEM; return (NULL); }
         fifo->fp = fp;
         STR(fifo->name, name)
         fifo->direction = FIFO_OUTPUT;
         if (flags & FIFO_BLOCK) fifo->blocking = 1;
         fifo->bufsiz = bufsiz;
         fifo->sbeb = sbeb;
         return (fifo);
      }
/*------------------------------------------------------------------------ put via TCP/IP client */
      if (!GetIniString("[PutTCPIP]","HostName",NULL,hostn,sizeof(hostn),ini_pfn) ||
          !GetIniString("[PutTCPIP]","HostIP",NULL,hostip,sizeof(hostip),ini_pfn) ||
          !GetIniInt("[PutTCPIP]","Port",0,&hostport,ini_pfn)){fifo_errno=FIFO_INI; return(NULL);}
      
      if (connect_as_client(hostn,hostip,hostport,&fd)){fifo_errno=FIFO_CONNECT; return(NULL);}

      fifo = make_fifo();                                      /* Make struct, TCP/IP put */
      if (!fifo) { fifo_errno = FIFO_MEM; return (NULL); }
      fifo->sock = fd;
      STR(fifo->name,name)
      STR(fifo->host, hostn)
      STR(fifo->ip, hostip)
      fifo->port = hostport;
      fifo->direction = FIFO_OUTPUT;
      if (flags & FIFO_BLOCK) fifo->blocking = 1;
      fifo->bufsiz = bufsiz;
      fifo->sbeb = sbeb;
      fifo->max_client = 0;         /* This indicates that we are the client */
      return (fifo);
   }
         
/*-------------------------------------------------------------------------------------------------
 * .get direction.  If the port is 0, we are using fifo's.  For get, there are two fifo's, the
 * .get fifo, and the .rtr fifo, which is used to signal the FifoBuffer prs that we are ready for
 * another record.
 *-----------------------------------------------------------------------------------------------*/
   if (!GetIniInt("[PutTCPIP]","Port",0,&hostport,ini_pfn)) {fifo_errno=FIFO_INI; return(NULL);}

   if (!hostport)                                                /* get via fifo */
   {  strcpy(pfn,name); strcat(pfn,".get");                                 /* get fifo */
      if (FNF(pfn)) { fifo_errno = FIFO_NOSUCHFIFO; return (NULL); }
      if (access(pfn, R_OK)) { fifo_errno = FIFO_PERMISSIONS; return (NULL); }
      fd = open(pfn, O_RDWR);
      if (fd < 0) { fifo_errno = FIFO_OPEN; return (NULL); }
      fp = fdopen(fd, "r");
      if (!fp) { fifo_errno = FIFO_OPEN; return (NULL); }
      
      strcpy(pfn,name); strcat(pfn,".rtr");                                 /* rtr fifo */
      if (FNF(pfn)) { fifo_errno = FIFO_NOSUCHFIFO; return (NULL); }
      if (access(pfn, W_OK)) { fifo_errno = FIFO_PERMISSIONS; return (NULL); }
      fd_rtr = open(pfn, O_WRONLY | blocking);
      if (fd_rtr < 0) { fifo_errno = FIFO_OPEN; return (NULL); }
      fp_rtr = fdopen(fd_rtr, "a");
      if (!fp_rtr) { fifo_errno = FIFO_OPEN; return (NULL); }
      setvbuf(fp_rtr, NULL, _IOLBF, 0);                           

      fifo = make_fifo();                                                 /* make struct fifo get */
      if (!fifo) { fifo_errno = FIFO_MEM; return (NULL); }
      strcpy(fifo->rtr_msg,snd_msg);
      fifo->fp = fp;
      fifo->fp_rtr = fp_rtr;
      STR(fifo->name, name)
      fifo->direction = direction; 
      if (flags & FIFO_BLOCK) fifo->blocking = 1;
      fifo->bufsiz = bufsiz;
      fifo->sbeb = sbeb;
      return (fifo);
   }
   
/*-------------------------------------------------------------------------------------------------
 * .get, using TCP/IP.  Pull rest of parameters
 * Note: the ConnectAs entry is from the point of view of the FifoBuffer prs, so from here (fifo.c)
 * it would be the opposite.
 *-----------------------------------------------------------------------------------------------*/
   if (!GetIniString("[GetTCPIP]","ConnectAs","S",a,sa,ini_pfn)){fifo_errno=FIFO_INI; return(NULL);}
   connect_as = 's';
   if (!a == 'c' || *a == 'C') connect_as = 'c'; 
   
/*-------------------------------------------------------------------------------------------------
 * .get TCP/IP, with them (FifoBuffer prs) as the server.  We get host and ip and try to connect
 * as a client.
 *-----------------------------------------------------------------------------------------------*/
   if (connect_as == 's')
   {  if (!GetIniString("[GetTCPIP]","PeerName",NULL,hostn,sizeof(hostn),ini_pfn) ||
          !GetIniString("[GetTCPIP]","PeerIP",NULL,hostip,sizeof(hostip),ini_pfn)){fifo_errno=FIFO_INI; return(NULL);}
          
      if (connect_as_client(hostn,hostip,hostport,&fd)){fifo_errno=FIFO_CONNECT; return(NULL);}
      
      fifo = make_fifo();                                      /* Make struct, TCP/IP get client */
      if (!fifo) { fifo_errno = FIFO_MEM; return (NULL); }
      strcpy(fifo->rtr_msg,snd_msg);
      fifo->rts = 1;
      fifo->sock = fd;
      STR(fifo->name,name)
      STR(fifo->host, hostn)
      STR(fifo->ip, hostip)
      fifo->port = hostport;
      fifo->direction = FIFO_INPUT;
      if (flags & FIFO_BLOCK) fifo->blocking = 1;
      fifo->bufsiz = bufsiz;
      fifo->sbeb = sbeb;
      fifo->buffer = malloc(bufsiz);
      memset(fifo->buffer,0,bufsiz);
      fifo->max_client = 0;         /* This indicates that we are the client */
      return (fifo);
   }

/*-------------------------------------------------------------------------------------------------
 * We are the server. In this case, we need to load the valid clients list from the .ini file,
 * loading their ip addresses from the /etc/hosts table if needed.
 *-----------------------------------------------------------------------------------------------*/   
/*-------------------------------------- Load all clients, key names in the [GetClients] section */
   if (!GetIniSection("[GetClients]",a,sa,ini_pfn)) {fifo_errno=FIFO_INI; return(NULL);}
   for (p=strtok(a,"="), i=0; p && i<FIFO_MAX_CLIENTS; p=strtok(NULL,"="), i++) STR(cltn[i],p)
   if (i == FIFO_MAX_CLIENTS){fifo_errno=FIFO_CLIENTS; return(NULL);}

/*---------------------------- Load ip's from the .ini file (these take precedence over /etc/hosts)
 * r flags that at least one ip was not in the ini file, so we will need to look at /etc/hosts
 *-----------------------------------------------------------------------------------------------*/
   for (i=0, r=0; cltn[i]; i++)
   {  if (!GetIniString("[GetClients]",cltn[i],"NULL",a,sizeof(a),ini_pfn)){fifo_errno=FIFO_INI; return(NULL);}
      STR_NOT_NULL(cltip[i],a)
      if (!cltip[i]) r=1;
   }
/*--------------------------------------------------------- Attempt to look up ips in /etc/hosts */
   if (r)
   {  if (fp = fopen("/etc/hosts", "r")) 
      {  while (fgets(a,sa,fp))
         {  if (p=strchr(a,'#')) *p = '\0';            /* handles comment data */
            p = strtok(a," \t");                       /* ip address, if any */
            h = strtok(NULL," \t");                    /* host name, if any */
            if (!p || !h) continue;                    /* not a data line */
            for (i=0; cltn[i]; i++)                    /* pass hosts from .ini */
            {  if (STR_EQ(cltn[i],h) && !cltip[i])     /* match host to file */
               {  STR(cltip[i],p)                      /* load ip from file */
                  break;
               }
            }
         }
         fclose(fp);
      }
   }
/*------------------------------------------------------------- Note & delete clients with no ip */
   for (i=0; cltn[i]; i++)
   {  if (!cltip[i]) 
      {  sprintf(em,"client %s, no ip", cltn[i]); dbg(em);
         FREE(cltn[i])
         for (j=i; j==i || cltn[j]; j++) {cltn[j]=cltn[j+1]; cltip[j]=cltip[j+1];}
         i--;
      }
   }
   if (!cltn[0]) { fifo_errno = FIFO_CLIENTS; return(NULL); } 
/*------------------------------------------------------------------------- Try to set up server */
   if (!(ptr=setup_server(hostport, &fd))) { fifo_errno = FIFO_SERVER_INIT; return (NULL); }

   fifo = make_fifo();                                        /* fifo for TCP/IP get, server */
   if (!fifo) { fifo_errno = FIFO_MEM; return (NULL); }
   strcpy(fifo->rtr_msg,snd_msg);
   STR(fifo->name,name)
   fifo->sock = fd;
   fifo->port = hostport;
   fifo->direction = direction;
   if (flags & FIFO_BLOCK) fifo->blocking = 1;
   fifo->bufsiz = bufsiz;
   fifo->sbeb = sbeb;
   fifo->ptr = ptr;
   for (i = fifo->max_client = 0; cltn[i]; i++)
   {  STR(CLIENT.name,cltn[i])
      for (j=0, p=strtok(cltip[i],"="); j < 4; j++, p=strtok(NULL,"=")) if (p) CLIENTIP = atoi(p);
      CLIENT.buffer = malloc(bufsiz); 
      memset(CLIENT.buffer,0,bufsiz);
      CLIENT.sock = 0;
      CLIENT.rts = 1;
      fifo->max_client++;
   }
   return (fifo);
}
         
/*-------------------------------------------------------------------------------------------------
 * Look to see if generated process has terminated.  Return non-zero if terminated, 0 otherwise.
 *-----------------------------------------------------------------------------------------------*/
static int check_child(pid_t pid)
{  int status;
   return (int) waitpid(pid, &status, WNOHANG);
}

/****************************************************** Closes open fifo and de-allocates memory */
int fifoClose(FIFO * fifo)
{  int i;
   char msg[2];
   fifo_errno = FIFO_OK;
   if (!fifo) { fifo_errno = FIFO_INVFIFO; return (FIFO_R_FAIL); }
/*----------------------------------------------------- For get, send rel (if snd has been sent) */
   if (fifo->direction == FIFO_INPUT && strcmp(fifo->rtr_msg,snd_msg))
   {  strcpy(fifo->rtr_msg,rel_msg);
      fifoGet(fifo, msg, sizeof(msg));
   }
   if (fifo->fp) { fclose(fifo->fp); fifo->fp = NULL; }
   if (fifo->fp_rtr) { fclose(fifo->fp_rtr); fifo->fp_rtr = NULL;}
   close_socket(&fifo->sock);
   FREE(fifo->buffer)
   FREE(fifo->name)
   FREE(fifo->host)
   FREE(fifo->ip)
   for (i=0; i < fifo->max_client; i++)
   {  FREE(CLIENT.name)
      FREE(CLIENT.buffer)
      close_socket(&CLIENT.sock);
   }
   FREE(fifo->ptr);
   FREE(fifo)
   return (FIFO_R_OK);
}

/************************************************************************************** Get data */
int fifoGet(FIFO * fifo, char * msg, size_t size) 
{  char * p;
   int r;
   fifo_errno = FIFO_OK;
   if (!fifo) { fifo_errno = FIFO_INVFIFO; return (FIFO_R_FAIL); }
   if (!msg)  { fifo_errno = FIFO_INVMSG;  return (FIFO_R_FAIL); }
   if (fifo->direction != FIFO_INPUT) { fifo_errno = FIFO_INVDIRECTION; return(FIFO_R_FAIL); }
/*------------------------------------------------------------------------------- Get using fifo */
   if (fifo->fp)
/*---------------------------------------------------------- Send RTR (Ready To Receive) message */
   {  clearerr(fifo->fp_rtr);
      if (fputs(fifo->rtr_msg, fifo->fp_rtr) < 0){fifo_errno=FIFO_ERROR; return(FIFO_R_FAIL);}
/*-------------------------------------------- If this was rel_msg, we are closing, and are done */
      if (STR_EQ(fifo->rtr_msg,rel_msg)) return (FIFO_R_OK);
      strcpy(fifo->rtr_msg,rtr_msg);
/*---------------------------------------------------------------------------------- Read record */
      clearerr(fifo->fp);
      memset(msg,0,size);
      p = fgets(msg, size, fifo->fp);
      if (p)
      {  while (LCHR(msg) == '\n') LCHR(msg) = '\0';
         return (FIFO_R_OK);
      }
      if (feof(fifo->fp)) { fifo_errno = FIFO_EOF; return (FIFO_R_OK); }
      fifo_errno = FIFO_ERROR;
      return (FIFO_R_FAIL);
   }
/*---------------------------------------------------------------------------------- TCP/IP Gets */
   if (fifo->ptr) return (tcpGetServer(fifo, msg, size));                    /* We're the server */
   else return (tcpGetClient(fifo, msg, size));                              /* We're the client */
}   

/*-------------------------------------------------------------------------------------------------
 * TCP/IP get - We're the server
 *-----------------------------------------------------------------------------------------------*/
static int tcpGetServer(FIFO * fifo, char * msg, size_t size)
{  int i, j, set_size, rc;
   static fd_set read_set, write_set, error_set;
   char buff[fifo->bufsiz], *p;
/*-------------------------------------------------------------------------------------------------
 * We have a timeout value of 5 seconds, but only if non-blocking.  If blocking, we NULL out the
 * pointer, so the selects will block indefinitely.  Thus, if a select returns with a timeout, we
 * just return an EOF
 *-----------------------------------------------------------------------------------------------*/
   static struct timeval tv = {5, 0};
   struct timeval * ptv = &tv;
   if (fifo->blocking) ptv = NULL;

   while (1)
   {  FD_ZERO(&read_set);
      FD_ZERO(&write_set);
      FD_ZERO(&error_set);
/*------------------------------------ We're server, sock is listen socket, and we're only read. */
      FD_SET(fifo->sock, &read_set);   
      set_size = fifo->sock;
/*-------------------------------------------------------------------------------------------------
 * Any of the connected clients could be ready to send a 
 * message to, or ready to receive a message from 
 *-----------------------------------------------------------------------------------------------*/
      for (i=0; i < fifo->max_client; i++)
      {  if (CLIENT.sock) 
         {  FD_SET(CLIENT.sock, &read_set);
            FD_SET(CLIENT.sock, &error_set);
            if (CLIENT.rts) FD_SET(CLIENT.sock, &write_set);
            if (CLIENT.sock > set_size) set_size = CLIENT.sock;
         }
      }

      set_size++;
      rc = select(set_size, &read_set, &write_set, &error_set, ptv);
      if (!rc) { fifo_errno = FIFO_EOF; return (FIFO_R_OK); } 
      if (rc < 0) {dbge("select"); fifo_errno = FIFO_SELECT; return (FIFO_R_FAIL); }

/*------------------------------------------------------------------------------- New connection */
      if (FD_ISSET(fifo->sock, &read_set)) accept_new_connection(fifo);

/*----------------------------------------------------------------------- Process client sockets */
      for (i=0; i < fifo->max_client; i++)
      {  if (FD_ISSET(CLIENT.sock, &error_set))           /* Error */
         {  FD_CLR(CLIENT.sock, &read_set);
            FD_CLR(CLIENT.sock, &write_set);
            close_socket(&CLIENT.sock);
            sprintf(em,"Error on %s client socket; reset", CLIENT.name); dbg(em);
         }
/*-------------------------------------------------------------- Write - send rtr_msg (SND, RTR) */
         if (FD_ISSET(CLIENT.sock, &write_set))
         {  rc = send(CLIENT.sock, CLIENT.rtr_msg, strlen(CLIENT.rtr_msg), 0);
            if (rc == -1)
            {  close_socket(&CLIENT.sock);
               sprintf(em,"Send error on %s; reset", CLIENT.name); dbg(em);
            }
            else
            {  sprintf(em,"Sent %3s to %s", CLIENT.rtr_msg +1, CLIENT.name); dbg(em);
               strcpy(CLIENT.rtr_msg,rtr_msg);
               CLIENT.rts = 0;
            }
         }
/*----------------------------------------------------------------------------------------- Read */
         if (FD_ISSET(CLIENT.sock, &read_set))
         {  memset(buff,0,sizeof(buff));
            rc = recv(CLIENT.sock, buff, sizeof(buff), 0);
            if (!rc)
            {  close_socket(&CLIENT.sock);
               sprintf(em,"%s dropped connection; reset", CLIENT.name); dbg(em);
               continue;
            }
            if (rc < 0)
            {  sprintf(em,"%s recv", CLIENT.name); dbge(em);
               close_socket(&CLIENT.sock);
               continue;
            }
            if (p=strchr(buff,SB)) strcpy(CLIENT.buffer,p);  /* SB starts message */
            else strcat(CLIENT.buffer,buff);                 /* additional message text */
            
            if (*CLIENT.buffer == SB && (p=strchr(CLIENT.buffer,EB)))   /* complete message? */
            {  *(p+1) = '\0';                                /* Truncate anything after EB */
               move_out(msg, CLIENT.buffer, fifo->sbeb);
               memset(CLIENT.buffer,0,fifo->bufsiz);
               strcpy(CLIENT.rtr_msg,rtr_msg);
               CLIENT.rts = 1;
               return (FIFO_R_OK);
            }
         }
      }
   }
}


/*-------------------------------------------------------------------------------------------------
 * TCP/IP get - We're the client
 *-----------------------------------------------------------------------------------------------*/
static int tcpGetClient(FIFO * fifo, char * msg, size_t size)
{  int i, j, set_size, rc;
   static fd_set read_set, write_set, error_set;
   char buff[fifo->bufsiz], *p;
/*-------------------------------------------------------------------------------------------------
 * We have a timeout value of 5 seconds, but only if non-blocking.  If blocking, we NULL out the
 * pointer, so the selects will block indefinitely.  Thus, if a select returns with a timeout, we
 * just return an EOF
 *-----------------------------------------------------------------------------------------------*/
   static struct timeval tv = {5, 0};
   struct timeval * ptv = &tv;
   if (fifo->blocking) ptv = NULL;

   while (1)
/*-------------------------- If not connected, try to connect; Return EOF on failure to connect */
   {  if (!fifo->sock) 
      {  if (connect_as_client(fifo->host, fifo->ip, fifo->port, &fifo->sock)) 
         {  fifo_errno = FIFO_EOF; 
            return (FIFO_R_OK);
         }
      }        

      FD_ZERO(&read_set);   
      FD_ZERO(&write_set);  
      FD_ZERO(&error_set);  
      
      if (fifo->rts) FD_SET(fifo->sock, &write_set);
      else FD_SET(fifo->sock, &read_set);
      FD_SET(fifo->sock, &error_set);
      set_size = fifo->sock +1;
      
      rc = select(set_size, &read_set, &write_set, &error_set, ptv);
      if (!rc) { fifo_errno = FIFO_EOF; return (FIFO_R_OK); } 
      if (rc < 0) {dbge("select"); fifo_errno = FIFO_SELECT; return (FIFO_R_FAIL); }

/*-------------------------------------------------------------------------- Error: reset socket */
      if (FD_ISSET(fifo->sock, &error_set)) 
      {  dbg("Error on client socket; reset");
         close_socket(&fifo->sock); 
         continue; 
      }
/*-------------------------------------------------------------- Write - send rtr_msg (SND, RTR) */
      if (FD_ISSET(fifo->sock, &write_set))
      {  rc = send(fifo->sock, fifo->rtr_msg, strlen(fifo->rtr_msg), 0);
         if (rc == -1)
         {  close_socket(&fifo->sock);
            dbg("Send error; reset");
         }
         else
         {  sprintf(em,"Sent %3s to server", fifo->rtr_msg +1); dbg(em);
            strcpy(fifo->rtr_msg,rtr_msg);
            fifo->rts = 0;
         }
         continue;
      }
/*----------------------------------------------------------------------------------------- Read */
      memset(buff,0,sizeof(buff));
      rc = recv(fifo->sock, buff, sizeof(buff), 0);
      if (!rc)
      {  close_socket(&fifo->sock);
         dbg("Dropped connection; reset"); 
         continue;
      }
      if (rc < 0)
      {  dbge("recv");
         close_socket(&fifo->sock);
         continue;
      }
      if (p=strchr(buff,SB)) strcpy(fifo->buffer,p);  /* SB starts message */
      else strcat(fifo->buffer,buff);                 /* additional message text */
            
      if (*fifo->buffer == SB && (p=strchr(fifo->buffer,EB)))   /* complete message? */
      {  *(p+1) = '\0';                                /* Truncate anything after EB */
         move_out(msg, fifo->buffer, fifo->sbeb);
         memset(fifo->buffer,0,fifo->bufsiz);
         strcpy(fifo->rtr_msg,rtr_msg);
         fifo->rts = 1;
         return (FIFO_R_OK);
      }
   }
}
   
/************************************************************************************** Put data */
int fifoPut(FIFO * fifo, char * msg) 
{  int r, flag = 0;
   fifo_errno = FIFO_OK;
   if (!fifo) { fifo_errno = FIFO_INVFIFO; return (FIFO_R_FAIL); }
   if (!msg)  { fifo_errno = FIFO_INVMSG;  return (FIFO_R_FAIL); }
   if (fifo->direction != FIFO_OUTPUT) { fifo_errno = FIFO_INVDIRECTION; return(FIFO_R_FAIL); }

/*--------------------------------------------------------------------------------- Using a fifo */
   if (fifo->fp)
   {  if (LCHR(msg) != '\n') {strcat(msg,"\n\0"); flag=1;}
      clearerr(fifo->fp);
      r = fputs(msg, fifo->fp);
      if (flag) LCHR(msg) = '\0';
      if (r < 0) {fifo_errno=FIFO_ERROR; return(FIFO_R_FAIL);}
      return (FIFO_R_OK);
   }  
   return (tcpPut(fifo, msg));
}   

/*-------------------------------------------------------------------------------------------------
 * Put for tcp/ip connections (on the put, we are always a client)
 *-----------------------------------------------------------------------------------------------*/
static int tcpPut(FIFO * fifo, char * msg)
{  int r, i, bflag = 0, eflag = 0, flag = 0, retries = 15, ret = retries, set_size;
   struct timeval timeout = {60, 0};
   fd_set read_set, write_set, error_set;
   char buff[64], rtr_msg[64] = "\0", *p;

/*------------------------------- Add SB & EB if needed; flags are used to remove them when done */
   if (*msg != SB) {bflag = 1; memmove(msg +1, msg, strlen(msg)); *msg = SB; }   
   if (LCHR(msg) != EB) {eflag = 1; strcat(msg,"\034\0"); }

   while (retries--)
/*------------------------------------------------------------ If you can't connect, return fail */
   {  if (!fifo->sock)
      {  if (connect_as_client(fifo->host, fifo->ip, fifo->port, &fifo->sock)) 
         {  sprintf(em,"Couldn't connect to %s (%s:%s)", fifo->host, fifo->ip, fifo->port); dbg(em);
            return (FIFO_R_FAIL);
         }
      }
      
      FD_ZERO(&read_set);
      FD_ZERO(&write_set); 
      FD_ZERO(&error_set); 
/*----------------------------- flag is set when the message is sent and we are awaiting the ACK */
      if (flag) FD_SET(fifo->sock, &read_set);
      else FD_SET(fifo->sock, &write_set);
      FD_SET(fifo->sock, &error_set);
      set_size = fifo->sock + 1;

      r = select(set_size, &read_set, &write_set, &error_set, &timeout);

      if (r < 0)                                                               /* Error on select*/
      {  dbge("send select"); 
         close_socket(&fifo->sock); 
         fifo_errno = FIFO_SELECT;
         return (FIFO_R_FAIL);
      }
      
      if (!r) continue;                                                        /* Timeout; retry */

      if (FD_ISSET(fifo->sock, &error_set))                                   /* error on socket */
      {  close_socket(&fifo->sock);
         continue;
      }

/*-------------------------------------------------------------------------------- Write message */
      if (FD_ISSET(fifo->sock, &write_set))                                           
      {  r = send(fifo->sock, msg, strlen(msg), 0);
         sprintf(em,"%d = send", r); dbg(em);
         if (r < 0) {dbge("send"); close_socket(&fifo->sock); continue;}
         flag = 1;
         retries = ret;
         continue;
      }

/*----------------------------------------------------------------------------- Read rtr message */
      if (FD_ISSET(fifo->sock, &read_set))
      {  memset(buff,0,sizeof(buff));
         r = recv(fifo->sock, buff, sizeof(buff), 0);
         sprintf(em,"%d = recv", r); dbg(em);
         if (r < 0) {dbge("recv"); close_socket(&fifo->sock); continue;}
/*--------------------------------------- store response until complete, then match with ACK msg */
         if (p=strchr(buff,SB)) strcpy(rtr_msg,p);
         else strcat(rtr_msg,buff);
         if (rtr_msg[0] != SB || !(p=strchr(rtr_msg,EB))) continue;
         *(p+1) = '\0';
/*----------------------------------------------------------------------- If a match, we're done */
         if (STR_EQ(rtr_msg,ack_msg))
         {  dbg("message ACK'd by server");
            if (bflag) memmove(msg,msg+1,strlen(msg));
            if (eflag) LCHR(msg) = '\0';
            return (FIFO_R_OK);
         }
/*-------------------------------------------------------------------------- Not matched, resend */
         dbg("server did not ACK message; resend");
         retries = ret;
         flag = 0;
         continue;
      }
   }
   fifo_errno = FIFO_RETRIES;
   return (FIFO_R_FAIL);
}


/*------------------------------------------------------------------------------- Error routines */
char * GetFifoErrorString(void)
{  if (fifo_errno < max_e_msg) return e_msg[fifo_errno];
   return no_msg;
}
void PrtFifoErrorString (FILE * file, char * prefix)
{	if (!file) file = stdout;
	fprintf(file, "%s: %s\n", prefix, GetFifoErrorString());
	return;
}

/**************************************************************************************************
 * Open server listen socket. On success, return pointer to sockaddr_in struct which has been
 * created for listen socket.  On failure, return NULL;
 *************************************************************************************************/
static struct sockaddr_in * setup_server(int port, int * sock)
{  struct hostent *hostentptr, hostentstruct;
   char hostname[MAXHOSTNAMELEN];
   struct sockaddr_in sock_name, * ptr;
   struct sockaddr * sptr = (struct sockaddr*)&sock_name;
   int rc, on = 1;
   size_t sl = sizeof(sock_name);

   sprintf(em, "Creating listen socket on port %d", port); dbg(em);

   if (*sock) close_socket(sock);
   
	if ((*sock=socket(AF_INET,SOCK_STREAM,0)) == -1){dbge("listen socket"); return(NULL);}
   if (setsockopt(*sock,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on))){dbge("listen opt"); return(NULL);}
   
   if (gethostname(hostname,sizeof(hostname))) {dbge("gethostname"); return(NULL);}
   hostentptr = gethostbyname (hostname);
   if (!hostentptr){dbge("gethostbyname"); return(NULL);}

   hostentstruct = *hostentptr;
   sock_name.sin_family = hostentstruct.h_addrtype;
   sock_name.sin_port = htons(port);
   sock_name.sin_addr = *((struct in_addr *)hostentstruct.h_addr);

   if (bind(*sock,(struct sockaddr*)&sock_name, sl) == -1) {dbge("bind"); return(NULL);}
   if (listen(*sock, 10) == -1) {dbge("listen"); return(NULL);}
      
   dbg("Listen socket created"); 
/*---------------------------------------------------------- Send back ptr to malloc'd structure */
   ptr = malloc(sizeof(struct sockaddr_in) + 3);
   memcpy(ptr,&sock_name,sizeof(struct sockaddr_in));
   return (ptr);
}


/**************************************************************************************************
 * Attempt to connect as client to server on target host, ip & port.  Returns 0 if successful, 1
 * otherwise.  If successful, also returns socket in sock. 
 *************************************************************************************************/
static int connect_as_client(char * target_host, char * target_ip, int port, int * sock)
{  struct hostent *phent, hent;
   struct sockaddr_in hsn;
   struct timeval timeout = {30,0};
   fd_set set;
   int i, j, r, retries = 15, ret = retries;
   while (retries--)
   {  if (*sock) close_socket(sock);                             /* clear previous socket if any */
      
      *sock = socket(AF_INET, SOCK_STREAM, 0);                                 /* get new socket */
      if (*sock == -1) {dbge("socket"); return (1); }
      
      i = fcntl(*sock, F_GETFL, 0);                                          /* set non blocking */
      fcntl(*sock, F_SETFL, O_NONBLOCK|i);
      
      phent = gethostbyname(target_host);                                            /* get host */
      if (!phent) phent = gethostbyname(target_ip);
      if (!phent) { dbge("gethost"); close_socket(sock); return(1); }
      
      hent = *phent;                                               /* load socket address struct */
      hsn.sin_family = hent.h_addrtype;
      hsn.sin_port   = htons(port);
      hsn.sin_addr   = *((struct in_addr *)hent.h_addr);
      
      i = sizeof(hsn);                                                /* connect to monitor host */
      if (connect(*sock, (struct sockaddr *)&hsn, i) == -1 && errno != EINPROGRESS)
      {  dbge("connect");
         fifo_errno = FIFO_CONNECT; 
         close_socket(sock);
         return (1);
      }
      return (0);
   }
   sprintf(em, "Connect failed %d times", ret); dbg(em);
   close_socket(sock);
   return (1);
}

/**************************************************************************************************
 * Accept a new client connection for server, matching it against valid clients
 *************************************************************************************************/
static void accept_new_connection(FIFO * fifo)
{	int ret, new_socket, i, j;	
	char ip[18];
	unsigned char * p;
	unsigned char flag;
   static size_t sl = sizeof(struct sockaddr_in);
   static int on = 1;
   static fd_set read_set, write_set, error_set;
/*------------------------------------------------------- Accept new connection on listen_socket */
	new_socket = accept(fifo->sock, (struct sockaddr*)fifo->ptr, &sl);
	if (new_socket == -1) {dbge("accept"); return;}
	
	ret = setsockopt (new_socket, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof (on));
	if (ret == -1) { dbge("setsockopt"); return; }

	p = (unsigned char *) &(fifo->ptr->sin_addr);
	sprintf(ip, "%d.%d.%d.%d", p[0], p[1], p[2], p[3]);
	
/*-------------------------------------------------------------------------------------------------
 * Scan the clients. The new socket's ip address must match one of them, and the connection is for 
 * that client.  If that client already has a connection, we need to clear it and set this up as 
 * the new one.
 *-----------------------------------------------------------------------------------------------*/
	for(i = 0; i < fifo->max_client; i++)
	{	for (j=0, flag=1; j < 4; j++) if (CLIENTIP != p[j]) flag=0;
      if (flag)
		{	if (CLIENT.sock) 
			{	FD_CLR(CLIENT.sock, &read_set);
				FD_CLR(CLIENT.sock, &write_set);
				FD_CLR(CLIENT.sock, &error_set);
				close_socket(&CLIENT.sock);
			}
			CLIENT.sock = new_socket;
         strcpy(CLIENT.rtr_msg,snd_msg);
         CLIENT.rts = 1;
         memset(CLIENT.buffer,0,fifo->bufsiz);
			sprintf(em,"Connected to %s", CLIENT.name); dbg(em);
			return;
		}
	}
	sprintf(em,"Attempt to connect from %s rejected", ip); dbg(em);
	return;
}

static void close_socket(int *s){if(*s){shutdown(*s,2);close(*s);*s=0;}}

/*------------------------------------------------------- Allocate and default fifo struct items */
static FIFO * make_fifo(void)
{  int i, j;
   FIFO * fifo;
   fifo = malloc(sizeof(struct _fifo_) + 3);
   if (!fifo) return (NULL);
   fifo->fp = NULL;
   fifo->fp_rtr = NULL;
   strcpy(fifo->rtr_msg,snd_msg);
   fifo->name = NULL;
   fifo->sock = 0;
   fifo->host = NULL;
   fifo->ip = NULL;
   fifo->port = 0;
   fifo->direction = 0;
   fifo->blocking = 0;
   fifo->bufsiz = 0;
   fifo->rts = 0;
   fifo->sbeb = 0;
   fifo->ptr = NULL;
   for (i = 0; i < FIFO_MAX_CLIENTS; i++)
   {  CLIENT.name = NULL;
      for (j = 0; j < 4; j++) CLIENTIP = 0;
      strcpy(CLIENT.rtr_msg,snd_msg);
      CLIENT.sock = 0;
      CLIENT.rts = 0;
      CLIENT.buffer = NULL;
   }
   fifo->max_client = 0;
   return (fifo);
}

/*----------------------------------------------------------------- in => out, with/without sbeb */
static void move_out(char * out, char * in, char sbeb)
{  if (sbeb)
   {  *out = '\0';
      if (*in != SB) strcat(out,"\013");
      strcat(out,in);
      if (LCHR(out) != EB) strcat(out,"\034");
      
   }
   else
   {  if (LCHR(in) == EB) LCHR(in) = '\0';
      if (*in == SB) in++;
      strcpy(out,in);
   }
}
