#ifndef _LOG_H_
#define _LOG_H_

/******************************************************************************
 * log.h                                                              R Moulton
 * Header file for log routines, which define a standard approach to generating
 * and maintaining log files for a program. Unless otherwise stated, all int 
 * returns are 0 =OK, non-zero =Failed.  All pointer returns are pointer=OK, 
 * NULL=Failed.  log_errno is set at each call to indicate the specifics of the 
 * error. GetLogErrorString will return a pointer to a string describing the 
 * error, * PrtLogErrorString will print the string to the passed FILE * or 
 * stdout. 
 *-----------------------------------------------------------------------------
 * Basic operation.  Set up each log in application ini file or another ini 
 * file, as desired.  Call logInitialize to pass ini file you will be using.
 * Use logOpen and logClose for log if you are generating your own entries.
 * Use Log Note for short, one line entries with time stamp on line (50 chars
 * or so).  Use logMsg for longer messages, where the time stamp will be on a
 * separate line, followed by the msg.
 * Note that, if Status is off, the log will be turned off, but purging will
 * still function. All calls will still function the same, except that writes 
 * will go to /dev/null instead of the directory and file specified.
 *-----------------------------------------------------------------------------
 * logNote and logMsg will generate log entries if Status is ON
 * logNoteLevel and logMsgLevel will generate log entries if Status is ON and
 *    Level is 0 (meaning all levels) or the level in the function call is >=
 *    Level. The concept here is that log message are hierarchial, with Level 1
 *    being the most important, Level 2 the next and so on.  The Level entry in
 *    the .ini would then give the user the ability to set the level of detail 
 *    in the logging to meet their needs.
 * logNoteType and logMsgType will generate log entries if Status is ON and
 *    the type code in the function call matches an entry in the section which
 *    is not OFF. (They default to ON).  The concept here is that logging is
 *    broken down into categories, and the user can turn off categories of 
 *    logging which are not needed.
 *-----------------------------------------------------------------------------
 * Each log_name corresponds to a section in inipfn with entries as follows:
 * [LOG_log_name]
 * Status=[ON|OFF] (is this logging enabled (default = ON))
 * Level=0 (level of logging to generate (default 0))
 * TYPE_ID=[ON|OFF]
 * Dir=path name of log directory (default cwd)
 * Prefix=log pfn name prefix (if any)
 * Suffix=log pfn name suffix (default = .log)
 * Period=[MONTHLY|DAILY] (default DAILY)
 * Retention=number of log files to keep (default MONTHLY 6, DAILY 30)
 * PurgeHour=hour of day to perform purge (0-23, default 2)
 * LastPurge=MM-DD-YYYY (this is maintained by purge process)
 *-----------------------------------------------------------------------------
 * 5-5 R Moulton Initial write.
 *****************************************************************************/

#include "hl7.h"

/*-------------------------------------------------------------- Error codes */
extern unsigned int log_errno;

#define LOG_OK          0
#define LOG_DISABLED    1  /* Status=OFF in .ini - returns OK */
#define LOG_LEVEL       2  /* level lower than Level in .ini - returns OK */
#define LOG_TYPE        3  /* log type turned off in .ini -returns OK */
#define LOG_INIINV      4
#define LOG_ININOF      5
#define LOG_ININOTRW    6
#define LOG_NOTINIT     7
#define LOG_NAMEINV     8
#define LOG_DIRNOF      9
#define LOG_DIRSTAT    10
#define LOG_DIRNOTDIR  11
#define LOG_DIRPERMS   12
#define LOG_OPEN       13
#define LOG_NULL       14
#define LOG_EMPTY      15
#define LOG_DIROPEN    16
#define LOG_SECINV     17
#define LOG_HL7ERR     18
#define LOG_MESGERR    19

int logInitialize(char * inipfn);
FILE * logOpen(char * log_name);
FILE * logClose(void);
int logNote(char * log_name, char * note);
int logMsg(char * log_name, char * msg);
int logHl7Msg(char * log_name, char * prefix, PMSG, int par[]);
int logNoteLevel(char * log_name, char * note, char level);
int logMsgLevel(char * log_name, char * msg, char level);
int logHl7MsgLevel(char * log_name, char * prefix, PMSG, int par[], char level);
int logLevel(char * log_name, char level);
int logNoteType(char * log_name, char * note, char * type);
int logMsgType(char * log_name, char * msg, char * type);
int logPurge(char * log_name);
char * GetLogErrorString(void);
void PrtLogErrorString(FILE * file, char * prefix);

#endif
