/**************************************************************************************************
 * sql.c                                                                                  R Moulton
 * Library of routines for mysql database access.
 *-------------------------------------------------------------------------------------------------
 * When compiling, use: gcc -c -I/usr/local/mysql/include sql.c
 * When linking, use:   -L/usr/local/mysql/lib -lmysqlclient -lsocket -lnsl
 *-------------------------------------------------------------------------------------------------
 * 12-3 R Moulton Initial write
 *************************************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "sql.h"
#include "ini.h"
#include "dbg.h"

/*--------------------------------------------------------------------------------------- Macros */
#define FREE(A) {free(A);A=NULL;}
#define STR_EQ(A,B) !strcasecmp(A,B)
#define STR(A,B) {if(B){if(A=realloc(A,strlen(B)+3))strcpy(A,B);}else{free(A);A=NULL;}}
#define LCHR(A) A[strlen(A) - 1]

#define DEFAULT_INI_FILE_PFN "sql.ini"   
#define SQL_R_OK    0
#define SQL_R_FAIL  1

static unsigned int quoted = FIELD_TYPE_DATE | FIELD_TYPE_TIME | FIELD_TYPE_DATETIME | FIELD_TYPE_STRING |
                             FIELD_TYPE_BLOB | FIELD_TYPE_ENUM | FIELD_TYPE_SET;

/*------------------------------------------------ Error msg strings.  Match to sql_errno values */
static char * e_msg[] =
{  "No error",
   "Null database name",
   "Null user name",
   "Null ini pfn",
   "ini file error",
   "mysql init failure",
   "ini file not found",
   "ini db section not found",
   "Missing/empty fields struct",
   "No db connection",
   "No sql query",
   "Query doesn't return results",
   "None of columns in query res",
   "Memory allocation error"
};
static int max_e_msg = sizeof(e_msg) / sizeof(char *);
static char * no_msg = "Unknown sql_errno value";

static char * inipfn = NULL;

static FILE * debug_fp = NULL;

static char oemsg[256] = "NULL conn";

/*---------------------------------------------------------------------------- Sets ini file pfn */
int sqlInitialize(char * new_pfn)
{  sql_errno = SQL_OK;
   if (!new_pfn) { sql_errno = SQL_NOINIPFN; return SQL_R_FAIL; }
   STR(inipfn,new_pfn)
   return SQL_R_OK;
}

/*-------------------------------------------------------------------------------------------------
 * Connects to a db server and makes a database current.  db is the section in the .ini file, and 
 * will be used as the database name unless there is a Database entry in that section which 
 * overides it. Returns pointer to new connection structure.
 *-----------------------------------------------------------------------------------------------*/
MYSQL * sqlOpen(char * db, char * user)
{  char *host_name = NULL, *password = NULL, *socket_name = NULL, *db_name = NULL;
   char a[2048], *p;
   int i, sa = sizeof(a);
   unsigned int flags, port_num;
   static MYSQL * conn = NULL;
   if (!inipfn) inipfn = DEFAULT_INI_FILE_PFN;
   sql_errno = SQL_OK;
   if (!db)   { sql_errno = SQL_NODBNAME; return NULL; }                  /* Validate parameters */
   if (!user) { sql_errno = SQL_NOUSER;   return NULL; }
   STR(db_name,db)
   flags = 0;

/*--------------------------------------------- Ini file must exist and have section matching db */
   if (access(inipfn, F_OK)) { sql_errno = SQL_ININOF; return NULL; }
   if (!GetIniSectionNames(a, sa, inipfn)) { sql_errno = SQL_INI; return NULL; }
   for (i=0, p=strtok(a,"[]"); p; p=strtok(NULL,"[]")) if (STR_EQ(p,db)) i=1;
   if (!i) {sql_errno = SQL_SECNOF; return NULL; }
   
/*----------If there is a DataBase entry, the db was a nickname; if not, it is the database name */
   GetIniString(db, "DataBase", db, a, sizeof(a), inipfn);
   STR(db_name,a);

/*--------------------------------------------------------------------- Pull items from ini file */
   GetIniString(db,"HostName", "NULL", a, sizeof(a), inipfn);
   if(strcasecmp(a,"NULL")) STR(host_name,a)
   GetIniInt(db,"PortNum", 0, &i, inipfn);
   port_num = i;
   GetIniString(db,"SocketName", "NULL", a, sizeof(a), inipfn);
   if (strcasecmp(a,"NULL")) STR(socket_name,a)
      
/*------------------------------------ User name (provided in call) is kname, password is kvalue */
   GetIniString(db, user, "NULL", a, sizeof(a), inipfn);
   if(strcasecmp(a,"NULL")) STR(password,a)

/*---------------------------------------------------------------------------------------- Flags */
   GetIniInt(db,"Flags", 0, &i, inipfn);
   flags = i;

/*------------------------------------ Make new structure and open connection to server/database */
   conn = mysql_init(NULL);
   if (!conn) { sql_errno = SQL_INITFAIL; return NULL; }

   if (mysql_real_connect(conn, host_name, user, password, db_name, port_num, socket_name, flags) == NULL)
   {  sql_errno = mysql_errno(conn);
      strcpy(oemsg,mysql_error(conn));
      return NULL;
   }
   return conn;
}

/*----------------------------------------------------------------- Close an existing connection */
int sqlClose(MYSQL * conn)
{  mysql_close(conn);
   sql_errno = SQL_OK;
   return SQL_R_OK;
}

/*-------------------------------------------------------------------------------------------------
 * Perform a query using a _result_field_ structure array.  
 * 
 * The caller defines a structure of char pointers for the items they want from a query, and a
 * pointer to that struct, e.g.:
 * struct _study_ {
 *    char dba;
 *    char qc_status;
 *     "
 *     "
 * };
 * static struct _study_ * Study = NULL;
 *
 * They then define the db items in the struct using an array of _result_field_ structures, 
 * initialized with the db item macros from the database definition header, followed by the
 * end of results macro, e.g.:
 * struct _result_field_ Study_def[] = {
 *    DBA_STUDY_QC,
 *    QCSTATUS_STUDY_QC,
 *       "
 *       "
 *    EO_RESULTS_DEF,   -- This is at end, and is all NULL's
 * };
 *
 * They then open a connection, and develop an appropriate SELECT query (or other query generating
 * results, and call this function, e.g.:
 *
 * Study = (struct _study_ *) sqlSelect(conn, sql_query, Study_def, &max_study);
 *
 * The function will return a void pointer to an array of structs of char pointers matching the
 * definition, containing the data returned from the query, one array element per result row.  If
 * an item in the definition is not present in the query results, its pointers will be NULL.  Items
 * in the result but not in the definition will be ignored.  The rows parameter (max_study) in the
 * example, will return the number of rows.
 * If there are no rows returned by the query rows will return 0, and the function will return NULL.
 * If an error occurs, these will return the same, but sql_errno will not be SQL_OK.
 * If old is non-NULL, it is presumed to be a void pointer to a previous struct of the same type,
 * which is freed up.  Thus new queries of a type can be done simply with another call to the
 * function; Otherwise the user must free the struct memory themselves.  
 *-----------------------------------------------------------------------------------------------*/
void * sqlSelect(MYSQL * conn, char * sql_query, struct _result_field_ * fields, int * rows, void * old)
{  int max_field = 0, field, max_col, col, r, flag;
   char * fld, *p;
   size_t s;
   static char eof[] = "$EOF";   /* string added at end of others in struct */
   MYSQL_RES * res;
   MYSQL_FIELD * cols;
   MYSQL_ROW row;
   sql_errno = SQL_OK;
   *rows = 0;
/*-------------------------------------------------------------------------------------------------
 * If old is non-NULL, it is a previous struct, and we free up the memory allocated to it as a
 * courtesy.  Note that all structs allocated by this function have one extra pointer at the end,
 * which we put eof in.  We then look for this address (not the string actually) to indicate that
 * we have gotten to the end of the allocated space.  We do not rely on the current definition to
 * determine the size of old.
 *-----------------------------------------------------------------------------------------------*/
   if (old)
   {  fld = (char *) old;
      while (fld != eof) { if (fld) free(fld); fld++; }
      free (old);
   }
/*---------------------------------------------------------------------------------- validations */
   if (!conn) { sql_errno = SQL_NOCONN; return NULL; }
   if (!sql_query || !strlen(sql_query)) { sql_errno = SQL_NOQUERY; return NULL; }
/*---------------------------------------------------------------------- validate & count fields */
   if (!fields) { sql_errno = SQL_NOFIELDS; return NULL; }
   while (fields[max_field].name) { fields[max_field].row = -1; max_field++; }
   if (!max_field) { sql_errno = SQL_NOFIELDS; return NULL; }
/*-------------------------------------------------------------------------- now we do the query */
   if (mysql_query(conn,sql_query))
   {  sql_errno = mysql_errno(conn);
      strcpy(oemsg,mysql_error(conn));
      return NULL;
   }
   res = mysql_store_result(conn);
   if (!res)
   {  if (mysql_errno(conn))
      {  sql_errno = mysql_errno(conn);    /* query had an error */
         strcpy(oemsg,mysql_error(conn));
      }
      else sql_errno = SQL_NORESULTS;      /* query not type that returns result set */
      return NULL;
   }
   *rows = mysql_num_rows(res);            /* Number of rows returned */
   if (!(*rows)) mysql_free_result(res); return NULL;  /* No rows, but this is OK */
/*-------------------------------------------------------------------------------------------------
 * Now we match the fields in the result set with those in the fields struct to find those which 
 * will are included, and determine their column locations.  The db match is not available until
 * release 4.1.x of MySQL.  Since this is a compile time check, we must have 4.1 on all servers
 * which will run the program if it is compiled on a server which has 4.1
 *-----------------------------------------------------------------------------------------------*/
   max_col = mysql_num_fields(res);
   cols = mysql_fetch_fields(res);
   for (col = 0, flag = 0; col < max_col; col++)
   {  for (field = 0; field < max_field; field++)
      {  if (strcmp(fields[field].name,cols[col].name)) continue;
         if (strcmp(fields[field].table,cols[col].table)) continue;
#if MYSQL_VERSION_ID > 41000
         if (strcmp(fields[field].db,cols[col].db))continue;
#endif
         fields[field].row = col;
         flag = 1;
      }
   }
/*---------------------------------------------------------- no match between fields and results */
   if (!flag) 
   {  sql_errno = SQL_NOCOLS; 
      *rows = 0; 
      mysql_free_result(res); 
      return NULL; 
   }   
/*-------------------------------------------------------------------------------------------------
 * Now we allocate the memory for the struct array, which we treat as array of char pointers. The
 * one extra slot is for our eof pointer.
 *-----------------------------------------------------------------------------------------------*/
   s = *rows * max_field * sizeof(char *) + 1;
   fld = malloc(s);
   if (!fld) 
   {  sql_errno = SQL_MALLOC; 
      *rows = 0; 
      mysql_free_result(res);
      return NULL; 
   }
   memset(fld,0,s);
   fld[*rows * max_field] = eof;
/*--------------------------------------------------- Process the rows from the query result set */
   for (r = 0; r < *rows; r++)
   {  row = mysql_fetch_row(res);
      for (field = 0; field < max_field; field++)
      {  if (fields[field].row != -1) fld[(r * max_field) + field] = strdup(row[fields[field].row]);
      }
   }
   mysql_free_result(res);
   return (void *) fld;
}

char * GetSqlErrorString(MYSQL * conn)
{  static char msg[256];
   if (sql_errno > 999) 
   {  if (conn) 
      {  strcpy(msg,mysql_error(conn));
         return msg;
      }
      else return oemsg;
   }
   if (sql_errno < max_e_msg) return e_msg[sql_errno];
   return no_msg;
}
void PrtSqlErrorString (MYSQL * conn, FILE * file, char * prefix)
{	if (!file) file = stdout;
	fprintf(file, "%s: %s\n", prefix, GetSqlErrorString(conn));
	return;
}

