/***********************************************************************
 * PACS_IN_TCP.C																R Moulton
 * PACS inbound feed to RAD. 
 * This program was originally modified from PACS_IN_FEED, to use a
 * TCP client to access RAD rather than DECNet.
 *----------------------------------------------------------------------
 * This TCP/IP socket server listens on a specified port and accepts up
 * to MAX_CONNECTIONS.  Connections are assumed to be PACS inbound HL7 
 * feeds.  Connections are accepted only from specified ip's and subnets
 * and only one connection at a time is accepted from any one ip.  A
 * status file is maintained for open connections.  Messages are sent to
 * RAD without processing and responses are returned.  Both
 * messages and responses are logged.
 *----------------------------------------------------------------------
 * gcc -g pacsi_feed.c alert.o dbg.o ini.o -lsocket -lnsl -o pacsi_feed
 *----------------------------------------------------------------------
 * 12/20/2 R Moulton Added error logging
 * 12/23/2 R Moulton Fixed bug; hanging when RAD link goes down.
 * 1/2/3	  R Moulton Added error printout, upgraded ctrl display.
 * 1/31/3  R Moulton Fixed bug I tried to fix on 12/23.
 * 2/5/3	  R Moulton Improved warning prints, added warning pages,
 *							improved status display.
 * 2/11/3  R Moulton Added warning e-mails.
 * 8/5/3   R Moulton Modified to use .ini, handle warnings on particular
 *                   Peers, handle multiple connects from a given peer
 **********************************************************************/

/* #define _VMS_MODE_ */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <time.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>

#ifdef _VMS_MODE_

#include <iodef.h>
#include <unixio.h>
#include <ssdef.h>
#include <starlet.h>
#include <descrip.h>

#include "[source.common]alert.h"
#include "[source.common]dbg.h"
#include "[source.common]ini.h"

#else

#include "alert.h"
#include "dbg.h"
#include "ini.h"

#endif

/*------------------------------------------------------------- Macros */
#define FREE(A) if (A){free(A);A=NULL;}
#define STR_EQ(A,B) !strcasecmp(A,B)
#define STR(A,B) {if(B){if(A=realloc(A,strlen(B)+3))strcpy(A,B);}else{free(A);A=NULL;}}
#define MAX(A,B) if (B > A) A = B;
#define LCHR(A) A[strlen(A) -1]

/*---------------------------------------------------------- Constants */
#define SB 11
#define EB 28

#ifdef _VMS_MODE_
#define PROCESS_ID			 "pacsi_feed"
#else
#define PROCESS_ID			 "pacsi_feed.prs"
#endif

#define BUFFER_SIZE			 61440		 /* Buffer sizes */
static char * cwd = NULL;
static char inipfn[] = "pacsi.ini";

/*---------------------------------- ini variables, storage, defaults */
static int listen_port = 52201;

#ifdef _VMS_MODE_
static char log_dir[128] = "[.log]";
static char log_pfn_format[128] = "%s%4d%02d%02d.full;1";
static char elog_pfn_format[128] = "%s%4d%02d.err;1";
static char status_file_pfn[] = "pacsi_status.dat;1";
static char control_pfn[] = "pacsi_control.dat;1";
#else
static char log_dir[128] = "./log/";
static char log_pfn_format[128] = "%s%4d%02d%02d.full";
static char elog_pfn_format[128] = "%s%4d%02d.err";
static char status_file_pfn[] = "pacsi_status.dat";
static char control_pfn[] = "pacsi_control.dat";
#endif

static char elog_pfn[256];
static char log_pfn[256];

static FILE * debug_fp;

static int status_update_seconds = 30;  

static char rad_host_name[32] = "rentgn";
static char host_name[32] = "\0";
static char rad_host_ip[20] = "128.252.118.40";
static int rad_host_port = 10001;

static time_t last_control_check = 0;
static int control_check_seconds = 180;

static int log_purge_hour = 2;         /* hour of day to trigger purge check */
static time_t last_purge_date = 0;		/* date last purge was done (yyyymmdd) */
static int log_retain_days = 30;	      /* # of days to retain general logs */
static int files_purged = 0;           /* # of logs purged last purge */
static int elog_retain_months = 6;     /* # of months to retain error logs */
static int efiles_purged = 0;          /* # of error logs purged last purge */

static char error_printer[128] = "mirp1$printer";         /* MIR error notification printer */
static int print_warning_messages = 0;     /* Should printed warnings be generated? */
static int print_window_start = 6;         /* Start/end hour of print warning window (0-23) */
static int print_window_end = 18;
static int e_mail_warning_messages = 0;
static char error_email[128] = "mirhelpdesk@mir.wustl.edu";

static int pager_warning_messages = 0;             /* Should pager warnings be generated? */
static char pager[48] = "3145388960@myairmail.com"; /* Def = Ralph */
static int pager_window_start = 6;                 /* Start/end hour of pager warning window */
static int pager_window_end = 2;

/*----------------------------------------- RAD system down window */
static int rad_down_dow = 3;    /* day of week (0-6) */
static int rad_down_shr = 2;    /* Start hour (0-23 */
static int rad_down_smin = 0;   /* Start minute in hour */
static int rad_down_mins = 45;  /* Length of time in minutes */

/*------------- Minutes before warning, waiting on RAD, waiting for PACS
 * RAD warning only if not in system down window.  PACS warning only if
 * not waiting on RAD.
 *--------------------------------------------------------------------*/
static time_t rad_basis = 0;   /* Time hack basis for rad warnings */
static int rad_warning_mins = 5;
static int rad_warning_interval = 15;
static int rad_page_mins = 15;
static int rad_page_interval = 60;

/*-------------------------------------------------- A bit of history */
static time_t interface_started_at;
static time_t interface_bounced_at = 0;
static char last_trans_ip[24];
static time_t last_msg_received = 0, last_trans_time = 0, oldest_message = 0;
static int message_waiting;
static time_t last_RAD_open = 0, last_RAD_close = 0;
static int last_slot_sent;
static int transaction_day = -1, transactions_this_day;
static int transaction_hour = -1, transactions_this_hour;
static char msg1[128] = "\0", msg2[128] = "\0", msg3[128] = "\0";
static char msg4[128] = "\0", msg5[128] = "\0", msg6[128] = "\0";
static char rej1[128] = "\0", rej2[128] = "\0", rej3[128] = "\0";
/*------------ "Current" time, used for time hacks, updated each pass */
static time_t now;
static struct tm * tm;
static char stm[30];

/*----------------------------------------------- Function prototypes */
void count_transaction(void);				 /* Increment transaction counts, displays */
void read_ini(void);							  /* Load environment variables */
void tcpserver (void);							/* main TCP/IP server loop */
void accept_connect(void);					  /* Accept connection on listen socket */
void process_pacs_read(void);			  /* read data on socket sock*/
void process_pacs_write(void);		  /* write data on socket sock */
void clear_socket(int sock);	         			  /* Clear socket table sock*/
void close_listen_socket(void);					  /* close listen socket */
void update_connection_status(int force);		 /* Update connection status information */
void link_error_check(void);				 /* Check for errors or warnings on link */
void shutdown_connection_status(void);
int shutdown_link(void);
void ldbg(char *msg);
void ldbge(char *msg);
void dbgm(char *msg, char *buff);
void log_message(char * note, char *msg);
void log_error(char *msg);
void connect_to_rad(void);					 /* attempt to connect to RAD, if not connected */
void create_listen_socket(void);		 /* open PACS listen socket */
void clear_rad_socket(void);
void process_rad_read(int slot);
void process_rad_write(int slot);
int control_check(void);				 /* check/process control file command*/
void check_purge_logs(void);			 /* check/purge old log files */
void time_hack(void);
char * strcmb(char *s1, char *s2);

static char em[256];
static char ib[BUFFER_SIZE];
/*----------------------------------------------------------------------
 * Socket data structure
 *--------------------------------------------------------------------*/
#define SPEER peers[sockets[sock].peer]
#define SOCKET sockets[sock]
#define MAX_CONNECTIONS		 20			 /* Maximum allowable connections */
/*---------------------------------------------- Socket status codes */
#define SS_INACTIVE			 0				 /* No socket active this slot */
#define SS_READ				 1				 /* Waiting for input from PACS */
#define SS_WRITE				 2				 /* Waiting for output to PACS */
#define SS_RAD_WRITE			 3				 /* Waiting for output to RAD */
#define SS_RAD_READ			 4				 /* Waiting for input from RAD */
#define SS_WAITING			 5				 /* RAD socket, nothing to process */

struct _sockets_ 
{	int status;								 /* Current socket status */
	int socket;								 /* Socket number */
	char * message_buffer;				 /* Completed message */
	time_t time_message_arrived;      /* message arrived (used to FIFO process msgs) */		 
   time_t basis;                     /* Last activity this socket (used for timeout) */
	char peer;								 /* slot in peer table for this socket */
};
static struct _sockets_ sockets[MAX_CONNECTIONS]; /* PACS inbound sockets */
static char sock;
static char max_sock = MAX_CONNECTIONS;

/*----------------------------------------------------------------------
 * Valid peers, loaded from .ini file.  Connections are only permitted
 * from these IP addresses, and warnings can be set up.
 * Note that more than one application can send queries from a given
 * IP address.  Feeds with status other than ON are not included in the
 * table.
 *--------------------------------------------------------------------*/
#define PEER peers[peer]
#define MAX_PEERS 10
struct _ipa_ {
   char * sname;             /* Ini section name */
   char * description;
   char * ip;
   int multiple_connections; /* Permit simultaneous connections from this peer? */
   int persistant_connections; /* keep connections until client closes them? */
   unsigned int timeout_seconds; /* close if inactive this long (0=never)*/
   time_t basis;                       /* time of last valid activity */
   time_t last_print;
   time_t last_page;
   unsigned int transactions_this_day;
   unsigned int transactions_this_hour;
   char * msg1;
   char * msg2;
   char * msg3;
/*------------------------- All of these default to main (RAD) values */
   int print_warning_messages;
   int warning_mins;
   int warning_interval;
   int pager_warning_messages;
   int page_mins;
   int page_interval;
}; 
static struct _ipa_ peers[MAX_PEERS];
static char max_peer = 0, peer;

struct hostent *rad_hostentptr, rad_hostent;
static char rad_hostname[256];
static struct sockaddr_in rad_sock_name;
static int rad_socket = 0;
static int rad_socket_status = SS_INACTIVE;

static int slot;	/* sockets slot currently processing with RAD */

static struct timeval timeval = {60, 0};
static int new_socket, listen_sock = 0;
static struct sockaddr_in listen_sock_name;
static char log_note[24];

static int main_loop = 1, tcp_loop = 1;
/***********************************************************************
 ***********************************************************************
 *********************** Main function *********************************
 ***********************************************************************
 **********************************************************************/
int main (int argc, char *argv[])
{	time_hack();
   if (!(cwd = MakeProgDirCurrent(argv[0])))
   {  PrtIniErrorString(NULL, "cwd"); 
      exit(1);
   }
   read_ini();									 /* Load environment variables */
	unique_process_check(PROCESS_ID);				      			  
	ldbg("PACSI interface started");
	interface_started_at = now;
	while (main_loop)
	{	time_hack();
      last_control_check = rad_basis = now;
		tcp_loop = 1;
		create_listen_socket();
		connect_to_rad();								 /* Connect to RAD */
/*--------------------------------------- Clear PACS connection table */
		for(sock=0; sock < max_sock; sock++) SOCKET.peer = -1;
		update_connection_status(1);
/*----------------------------------------- main loop is in tcpserver */
		sprintf(em,"Calling tcpserver, port %d", listen_port); dbg(em);
		tcpserver();
	}
	dbg("Exiting main loop");
	unique_process_check(NULL);
	shutdown_connection_status();
	exit(0);
}

/***********************************************************************
 ***********************************************************************
 **************** Main TCP server processing loop **********************
 ***********************************************************************
 **********************************************************************/

void tcpserver (void)
{	int rc, highest_socket, a, oldest_message_slot;
	fd_set rsock, wsock, esock;
/*----------------------------------------------------------------------
 * Main socket processing loop
 * -------------------------------------------------------------------*/
		while (tcp_loop)
		{	dbg ("tcp_loop"); 
/*------------------------------------------------------------- Housekeeping */
         time_hack();
			if(control_check()) continue;
			check_purge_logs();
			update_connection_status(0);
			link_error_check();
/*---------------------------------- identify connections for testing */
         highest_socket = listen_sock;
			FD_ZERO(&rsock);
			FD_ZERO(&wsock);
			FD_ZERO(&esock);
			FD_SET(listen_sock, &rsock);
			message_waiting = oldest_message_slot = 0;
			oldest_message = 0;
/*------------------------------------------------- PACS inbound sockets 
 * All sockets are set for read and error, those with waiting message
 * are also set to write
 * -------------------------------------------------------------------*/
			for (sock = 0; sock < max_sock; sock++)
			{	if (SOCKET.socket)
				{	if (SPEER.timeout_seconds && ((now - SOCKET.basis) > SPEER.timeout_seconds))
               {  clear_socket(sock);
                  continue;
               }
               MAX(highest_socket,SOCKET.socket)
               FD_SET(SOCKET.socket, &rsock);
					FD_SET(SOCKET.socket, &esock);
					if (SOCKET.status == SS_WRITE) FD_SET(SOCKET.socket, &wsock);
/*- This code finds oldest PACS message waiting to be sent to RAD, if any */
					if (SOCKET.status == SS_RAD_WRITE) 
					{	message_waiting = 1;
						SPEER.basis = now;
						if (!oldest_message || (SOCKET.time_message_arrived < oldest_message))
						{	oldest_message = SOCKET.time_message_arrived;
							oldest_message_slot = sock;
						}
					}
				}
			}
/*--------------------------------------------------- RAD host socket */
			if (!rad_socket && (message_waiting || (rad_socket_status != SS_WAITING))) connect_to_rad();
			if (message_waiting && rad_socket && rad_socket_status == SS_WAITING)
			{	slot = oldest_message_slot;
				rad_socket_status = SS_WRITE;
			}
			if (rad_socket)
			{	MAX(highest_socket,rad_socket)
            FD_SET(rad_socket, &esock); 
				FD_SET(rad_socket, &rsock); 
				if (rad_socket_status == SS_WRITE) FD_SET(rad_socket, &wsock); 
			}
/*----------------------------------------------- see if any activity */
			rc = select (highest_socket + 1, &rsock, &wsock, &esock, &timeval);
			sprintf(em,"select returned %d", rc); dbg(em); 
/*----------------------------------------------------------- timeout */
			if (!rc) continue;

/*--------------- activity on listen socket is request for connection */
			if (FD_ISSET (listen_sock, &rsock)) accept_connect();
/*----------------------------------------------------------------------
 * PACS inbound sockets
 *------------------------------------------------------------------- */
			for (sock = 0; sock < max_sock; sock++)
			{	if (SOCKET.socket)
				{	if(FD_ISSET(SOCKET.socket, &esock)) { clear_socket(sock); continue; }
					if(FD_ISSET(SOCKET.socket, &rsock)) process_pacs_read();
					if(FD_ISSET(SOCKET.socket, &wsock)) process_pacs_write();
				}
			}
/*----------------------------------------------------------------------
 * RAD host socket
 *--------------------------------------------------------------------*/
			if (rad_socket)
			{	if (FD_ISSET(rad_socket, &esock)) 
				{	ldbg("Main select: RAD socket error found");
					clear_rad_socket();
					connect_to_rad();
					continue;
				}
				if (FD_ISSET(rad_socket, &rsock)) process_rad_read(slot);
				if (FD_ISSET(rad_socket, &wsock)) process_rad_write(slot);
			}
		}
/*--------------------------------------------------- Shutdown server */
		ldbg("Exiting tcp loop");
		for (sock = 0; sock < max_sock; sock++) clear_socket(sock);
		clear_rad_socket();
		close_listen_socket();
}

/***********************************************************************
 * Setup and open server listen socket 
 **********************************************************************/
void create_listen_socket (void)
{	struct hostent *hostentptr, hostentstruct;
	int rc, on = 1, flag = 0, count = 15;
	size_t sl;

	while (count--)
	{	if(flag) sleep(60); flag = 1;
/*----------------------------------------------------- make a socket */
		if ((listen_sock = socket (AF_INET, SOCK_STREAM, 0)) == -1)
		{	ldbge("listen socket");
			continue;
		}

		if (setsockopt (listen_sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof (on)))
		{	ldbge("listen socket setsockopt");
			continue;
		}
/*----------------------------------------------- get local host name */
		if (!strlen(host_name))
      {  rc = gethostname (host_name, sizeof (host_name));
		   if (rc) 
         {  ldbge("gethostname"); 
            close_listen_socket(); 
            continue;
         }
      }
/*---------------------------- get pointer to network data structure */
		hostentptr = gethostbyname (host_name);
		if (!hostentptr) {ldbge("gethostbyname"); close_listen_socket(); continue;}

		hostentstruct = *hostentptr;
		listen_sock_name.sin_family = hostentstruct.h_addrtype;
		listen_sock_name.sin_port = htons(listen_port);
		listen_sock_name.sin_addr = * ((struct in_addr *)hostentstruct.h_addr);
/*----------------------------------------------- bind listen socket */
		sl = sizeof(listen_sock_name);
		rc = bind (listen_sock, (struct sockaddr*)&listen_sock_name, sl);
		if (rc) {ldbge("bind"); close_listen_socket(); continue;}
/*-------------------------------------------- listen for connections */
		rc = listen (listen_sock, 5);
		if (rc) {ldbge("listen"); close_listen_socket(); continue;}
		sprintf(em, "Listen socket created on port %d", listen_port); dbg(em);
		return;
	}
	ldbg("Unable to create listen socket"); 
	unique_process_check(NULL);
	shutdown_connection_status();
	exit(1);
}

/*----------------------------------------------------------------------
 * Accept new connection on listen socket.  
 *--------------------------------------------------------------------*/
void accept_connect (void)
{	int rc, i, j, on = 1;
	unsigned char *ptr;
	time_t oldest_time;
	size_t sl;
	char alpha_ip[20];
/*------------------------------------------------- Accept connection */
	sl = sizeof (listen_sock_name);
	new_socket = accept (listen_sock, (struct sockaddr*)&listen_sock_name, &sl);
	if (new_socket == -1) { dbge ("accept"); return; }

/*---------- If this socket number already in the table, clear its entry */
	for (sock = 0; sock < max_sock; sock++)
		if(new_socket == SOCKET.socket) clear_socket(sock);

	rc = setsockopt (new_socket, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof (on));
	if (rc == -1) dbge ("setsockopt");

	ptr = (unsigned char*)&listen_sock_name.sin_addr;
	memset(alpha_ip,0,sizeof(alpha_ip));
	sprintf(alpha_ip,"%d.%d.%d.%d",ptr[0],ptr[1],ptr[2],ptr[3]);

/*--------------------------------------- Validate IP against peer table
 * If found, peer is the peer table entry
 *--------------------------------------------------------------------*/
   for (peer=0; peer < max_peer; peer++) if (STR_EQ(alpha_ip,PEER.ip)) break;
	if (peer == max_peer)
	{	sprintf(em, "Attempt to connect from: %s rejected", alpha_ip); ldbg(em);
      strcpy(rej1,rej2);
      strcpy(rej2,rej3);
      sprintf(rej3, "%s, %s", alpha_ip, stm); 
		close(new_socket);
		return;
	}
/*----------------------------------------------------------------------
 * If multiple connections are not permitted for this peer, look for any
 * currently existing connections and clear them.
 *--------------------------------------------------------------------*/
   if (!PEER.multiple_connections)
   {  for (sock = 0; sock < max_sock; sock++)
	   {	if (SOCKET.peer == peer)
		   {	sprintf(em,"Multiple connection on %s", PEER.description); ldbg(em);
			   clear_socket(sock);
		   }
	   }
   }
/*----------------------------------------------------------------------
 * Look for an empty slot to put the new connection in. Failing this, we
 * have exceeded connections limit
 * -------------------------------------------------------------------*/
	for (sock = 0; sock < max_sock; sock++) if(SOCKET.status == SS_INACTIVE) break; 
	if (sock == max_sock)
	{	ldbg("Connection limit exceeded");
		sprintf(em, "Attempt to connect from: %s rejected", alpha_ip); ldbg(em);
		close(new_socket);
		return;
	}
/*----------------------------------------------------------------------
 * initialize and set ready to read.
 *--------------------------------------------------------------------*/
	if (SOCKET.status != SS_INACTIVE) clear_socket(sock);
	SOCKET.socket = new_socket;
	if (!(SOCKET.message_buffer = malloc(BUFFER_SIZE)))
	{	ldbg("Memory allocation error: new socket");
		unique_process_check(NULL);
		shutdown_connection_status();
		exit(1);
	}
	memset(SOCKET.message_buffer,0,BUFFER_SIZE);
   
	SOCKET.time_message_arrived = 0;
   SOCKET.basis = now;
   SOCKET.peer = peer;
	sprintf (em,"Connected to: %s", alpha_ip); dbg(em);
/*----------------------------- Set ready to read, clear input buffer */
	SOCKET.status = SS_READ;
	return;
}
/*------------------------ Close out socket entry at param table slot */
void clear_socket(int sock)
{	if (SOCKET.socket) 
	{	shutdown(SOCKET.socket,2); 
		close(SOCKET.socket); 
	}
	FREE(SOCKET.message_buffer);
	SOCKET.status = SS_INACTIVE;
	SOCKET.socket = 0;
	SOCKET.time_message_arrived = 0;
   SOCKET.basis = 0;
   SOCKET.peer = -1;
}

/***********************************************************************
 ***********************************************************************
 * Process TCP/IP read on PACS sock
 ***********************************************************************
 **********************************************************************/
void process_pacs_read(void)
{	int j, feed, facility, rc;
	char * sb_ptr, * eb_ptr;
	char id[128], * id_ptr, buffer[BUFFER_SIZE];
/*----------------------------------------------------------------------
 * If we are not waiting for input, clear out current message
 * ------------------------------------------------------------------ */
   SOCKET.basis = now;
	if (SOCKET.status != SS_READ)
	{	dbg ("PACS socket reset");
		memset(SOCKET.message_buffer,0,BUFFER_SIZE);
		SOCKET.status == SS_READ;
	}
/*-------------------------------------------------------- read input */
	memset (buffer, 0, BUFFER_SIZE);
	rc = recv (SOCKET.socket, buffer, BUFFER_SIZE, 0);
	if (rc == 0){dbg("PACS socket closed by user"); clear_socket(sock); return;}
	if (rc < 0)
	{	sprintf(em,"PACS socket error %d on read", rc); dbg(em); 
		dbge("PACS receive"); clear_socket(sock); return;
	}
	sprintf(em,"read on PACS socket %d, returned %d",SOCKET.socket,rc); dbg(em);
/*----------------------------------------------------------------------
 * At beginning of message, wait for SB; Otherwise, if an SB is found,
 * dump whatever message we have and restart with the SB
 * -------------------------------------------------------------------*/
	sb_ptr = strchr (buffer, SB);
	if (!strlen(SOCKET.message_buffer) && !sb_ptr) return;
	if (sb_ptr)
	{	memset(SOCKET.message_buffer,0,BUFFER_SIZE);
		strcpy(SOCKET.message_buffer, sb_ptr);
	}
/*----------------------------------------------------------------------
 * Check for overflow; if found, clear buffer and exit; wait for next
 * -------------------------------------------------------------------*/
	else if (strlen(SOCKET.message_buffer) + strlen(buffer) > BUFFER_SIZE)
	{	memset(SOCKET.message_buffer,0,BUFFER_SIZE);
		return;
	}
/*-------------------------------------------------------- Copy it in */
	else strcat(SOCKET.message_buffer, buffer);

/*------------------------------------------- see if there is 28 <EB> */
	eb_ptr = strchr(SOCKET.message_buffer, EB);
	if (!eb_ptr) return;
/*----------------------------------------------------------------------
 * We may have a complete message.  Check for SB before EB and required
 * "MSH|" message header.
 *--------------------------------------------------------------------*/
	sb_ptr = strchr(SOCKET.message_buffer, SB);
	if (!sb_ptr || (sb_ptr > eb_ptr) || !strstr(SOCKET.message_buffer, "MSH|"))
	{	memset (SOCKET.message_buffer, 0, BUFFER_SIZE);
		return;
	}
	SOCKET.time_message_arrived = last_msg_received = now;
	dbg("Message received");
	SPEER.basis = now;
/*---------------------------------------------------------------------
 * We now have a completed message; Process the message through RAD
 *-------------------------------------------------------------------*/
	SOCKET.status = SS_RAD_WRITE;
}

/***********************************************************************
 ***********************************************************************
 * Process RAD TCP/IP read for PACS sock
 ***********************************************************************
 **********************************************************************/
void process_rad_read(int sock)
{	int j, feed, facility, rc;
	char * sb_ptr, * eb_ptr;
	char id[128], * id_ptr, buffer[BUFFER_SIZE];
	sprintf(em, "process_rad_read: idxs=%d, pacs=%d", rad_socket_status, SOCKET.status);
	dbg(em);
/*-------------------------------------------------------- read input */
	memset (buffer, 0, BUFFER_SIZE);
	rc = recv (rad_socket, buffer, BUFFER_SIZE, 0);
	if (rc == 0)
	{	ldbg("RAD has closed socket"); 
		last_RAD_close = time(NULL);
		clear_rad_socket(); 
		if (SOCKET.status == SS_RAD_READ)
		{	dbg("RAD msg reset");
			strcpy(SOCKET.message_buffer,ib);
			SOCKET.status = SS_RAD_WRITE;
		}
		return;
	}
	if (rc < 0)
	{	ldbge("RAD socket error on read");  
		clear_rad_socket(); 
		if (SOCKET.status == SS_RAD_READ)
		{	dbg("RAD msg reset");
			strcpy(SOCKET.message_buffer,ib);
			SOCKET.status = SS_RAD_WRITE;
		}
		return;
	}
	sprintf(em,"read on RAD socket %d, returned %d",SOCKET.socket,rc); dbg(em);
	if (rad_socket_status != SS_READ) return;
	if (SOCKET.status != SS_RAD_READ) { rad_socket_status = SS_WAITING; return; }
/*----------------------------------------------------------------------
 * At beginning of message, wait for SB; Otherwise, if an SB is found,
 * dump whatever message we have and restart with the SB
 * -------------------------------------------------------------------*/
	sb_ptr = strchr (buffer, SB);
	if (!strlen(SOCKET.message_buffer) && !sb_ptr) return;
	if (sb_ptr)
	{	memset(SOCKET.message_buffer,0,BUFFER_SIZE);
		strcpy(SOCKET.message_buffer, sb_ptr);
	}
/*----------------------------------------------------------------------
 * Check for overflow; if found, clear buffer and exit; wait for next
 * -------------------------------------------------------------------*/
	else if (strlen(SOCKET.message_buffer) + strlen(buffer) > BUFFER_SIZE)
	{	memset(SOCKET.message_buffer,0,BUFFER_SIZE);
		return;
	}
/*-------------------------------------------------------- Copy it in */
	else strcat(SOCKET.message_buffer, buffer);

/*------------------------------------------- see if there is 28 <EB> */
	eb_ptr = strchr(SOCKET.message_buffer, EB);
	if (!eb_ptr) return;
/*----------------------------------------------------------------------
 * We may have a complete message.  Check for SB before EB and required
 * "MSH|" message header.
 *--------------------------------------------------------------------*/
	sb_ptr = strchr(SOCKET.message_buffer, SB);
	if (!sb_ptr || (sb_ptr > eb_ptr) || !strstr(SOCKET.message_buffer, "MSH|"))
	{	memset (SOCKET.message_buffer, 0, BUFFER_SIZE);
		return;
	}
	dbg("Message received");
	log_message("RAD", SOCKET.message_buffer);
/*---------------------------------------------------------------------
 * We now have a completed message; return to pacs
 *-------------------------------------------------------------------*/
	SOCKET.status = SS_WRITE;
	rad_socket_status = SS_WAITING;
	rad_basis = now;
}

/***********************************************************************
 ***********************************************************************
 * Process PACS TCP/IP write from socket sock
 ***********************************************************************
 **********************************************************************/
static char tb[BUFFER_SIZE];
void process_pacs_write(void)
{	int rc;
   SOCKET.basis = now;
	if (SOCKET.status != SS_WRITE) return;
	dbgm("TCP write", SOCKET.message_buffer);
	memset(tb,0,BUFFER_SIZE);
	strcpy(tb,SOCKET.message_buffer);
	rc = send(SOCKET.socket, tb, strlen(tb), 0);
	if (rc < 0)
	{	dbge ("PACS send");
		clear_socket(sock);
		return;
	}
	sprintf(em, "%d characters sent to PACS", rc); dbg(em);
	memset(SOCKET.message_buffer,0,BUFFER_SIZE);
	SOCKET.status = SS_READ;
/*---------------------------------------------------- History update */
	strcpy(last_trans_ip,SPEER.ip);
	last_trans_time = SPEER.basis = now;
	count_transaction();
   if (!SPEER.persistant_connections) clear_socket(sock);
	update_connection_status(1);
}

/***********************************************************************
 ***********************************************************************
 * Process RAD TCP/IP write of data from socket sock
 ***********************************************************************
 **********************************************************************/
void process_rad_write(int sock)
{	int ml, rc;
	if (SOCKET.status != SS_RAD_WRITE) 
	{	rad_socket_status = SS_WAITING;
		return;
	}
	dbgm("RAD write", SOCKET.message_buffer);
	sprintf(log_note, "%s - IP=%s", SPEER.description, SPEER.ip);
	log_message(log_note, SOCKET.message_buffer);
	memset(ib,0,BUFFER_SIZE);
	strcpy(ib,SOCKET.message_buffer);
	rc = send(rad_socket, ib, strlen(ib), 0);
	if (rc < 0)
	{	ldbge("RAD send");
		clear_rad_socket();
		connect_to_rad();
		return;
	}
	sprintf(em, "%d characters sent to RAD", rc); dbg(em);
	memset(SOCKET.message_buffer,0,BUFFER_SIZE);
	SOCKET.status = SS_RAD_READ;
	rad_socket_status = SS_READ;
	last_slot_sent = sock;
	rad_basis = now;
}

void close_listen_socket(void)
{	if (listen_sock > 0)
   {  if (close(listen_sock)) dbge("listen socket close");
      listen_sock = 0;
   }
	return;
}

/***********************************************************************
 ***********************************************************************
 * RAD client socket setup.  Connect attempts to make connection if one 
 * does not already exist.
 ***********************************************************************
 **********************************************************************/
void connect_to_rad(void)
{	int i, rc;
	if (rad_socket) return;
	while (1)
	{	if (!rad_host_port) 
		{	ldbg("No RAD host port"); 
			unique_process_check(NULL);
			shutdown_connection_status();
			exit(1); 
		}
		sprintf(em, "Opening RAD client socket on %s, port %d", rad_host_name, rad_host_port); dbg(em);
/*-------------------------------------------------------- Get socket */
		rad_socket = socket(AF_INET, SOCK_STREAM, 0);
		if (rad_socket == -1) 
		{	ldbge("RAD socket"); 
			rad_socket = 0; 
			break;
		}
/*----------------------------------------------- Set to non-blocking */
		i = fcntl(rad_socket, F_GETFL, 0);
		fcntl(rad_socket, F_SETFL, O_NONBLOCK|i);
/*----------------------------------------------------- Get host name */
		rad_hostentptr = gethostbyname(rad_host_name);
		if (!rad_hostentptr) rad_hostentptr = gethostbyname(rad_host_ip);
		if (!rad_hostentptr)
		{	clear_rad_socket();	 
			ldbge("RAD gethost");
			break;
		}
/*------------------------------------------------- Build socket name */
		rad_hostent = *rad_hostentptr;
		rad_sock_name.sin_family = rad_hostent.h_addrtype;
		rad_sock_name.sin_port = htons(rad_host_port);
		rad_sock_name.sin_addr = *((struct in_addr *)rad_hostent.h_addr);

/*------------------------------------------------- Connect to server */
		i = sizeof(rad_sock_name);
		rc = connect(rad_socket, (struct sockaddr *)&rad_sock_name, i);
		if (rc && errno != EINPROGRESS)
		{	ldbge("RAD connect");
			clear_rad_socket();
			break;
		}
		sprintf(em, "RAD client socket opened on port %d", rad_host_port); ldbg(em);
		last_RAD_open = time(NULL);
		break;
	}

	if (rad_socket) rad_socket_status = SS_WAITING;	  
	else rad_socket_status = SS_INACTIVE;
}

void clear_rad_socket(void)
{	if (rad_socket == 0) return;
	if (shutdown(rad_socket,2)) dbge("RAD shutdown"); 
	if (close(rad_socket)) dbge("RAD close");
	rad_socket = 0;
	rad_socket_status = SS_INACTIVE;
	update_connection_status(1);
}

/*----------------------------------------------------------------------
 * Update status info to show latest transaction.
 *--------------------------------------------------------------------*/
void count_transaction(void)
{	if (transaction_day != tm->tm_yday)	 /* Reset counters if needed */
	{	transaction_day = tm->tm_yday;
		transaction_hour = -1;
		transactions_this_day = 0;
      for (peer = 0; peer < max_peer; peer++) PEER.transactions_this_day = 0;
   }
	if (transaction_hour != tm->tm_hour)
	{	transaction_hour = tm->tm_hour;
		transactions_this_hour = 0;
      for (peer = 0; peer < max_peer; peer++) PEER.transactions_this_hour = 0;
	}
	transactions_this_day++;				 /* Increment counters */
	transactions_this_hour++;
   SPEER.transactions_this_day++;
   SPEER.transactions_this_hour++;

	strcpy(msg1,msg2);						 /* Scroll back messages */
	strcpy(msg2,msg3);
	strcpy(msg3,msg4);
	strcpy(msg4,msg5);
	strcpy(msg5,msg6);
	sprintf(msg6, " Message from %s at %s", SPEER.ip, stm);

   STR(SPEER.msg1,SPEER.msg2)
   STR(SPEER.msg2,SPEER.msg3)
   STR(SPEER.msg3,msg6)   
}

/***********************************************************************
 ***********************************************************************
 * Periodically updates connection status to a file, 
 ***********************************************************************
 **********************************************************************/
static time_t last_status_update = 0;

void update_connection_status(int force)
{	int i, peer, sock;
	char tim[35];
	FILE * fp;
/*--------------------------------------------------- Is it time yet? */
	if (!force && (now - last_status_update) < status_update_seconds) return;
	last_status_update = now;
/*------------------------------------------------- Clear update file */
	if (remove(status_file_pfn)) dbge(status_file_pfn);
	fp = fopen(status_file_pfn, "wb");
	if (!fp) {printf("Unable to create status file: %s\n", status_file_pfn); return;}
/*------------------------------------------------------------ Header */
	fprintf(fp, "PACS Inbound Link current status: %s", asctime(tm));
	fprintf(fp, "----------------------------------------------------------------\n");
	fprintf(fp, "Link started: %s", asctime(localtime(&interface_started_at)));
	if(interface_bounced_at)fprintf(fp, "Link last bounced: %s", asctime(localtime(&interface_bounced_at)));
	if(last_purge_date)
		fprintf(fp, "Last log file purge: %s   %d log files purged\n   %d err files purged\n", 
			asctime(localtime(&last_purge_date)), files_purged, efiles_purged);
/*--------------------------------------------- Process each connection */
	if (rad_socket)
	{	fprintf(fp, "Connected to RAD socket\n");
		fprintf(fp, "on %s (%s), port %d\n\n", rad_host_name, rad_host_ip, rad_host_port);
	}
	else fprintf(fp, "Not connected to RAD socket\n\n");
	if (last_RAD_close) fprintf(fp, "Last RAD socket close: %s", asctime(localtime(&last_RAD_close)));
	if (last_RAD_open)  fprintf(fp, "Last RAD socket open: %s", asctime(localtime(&last_RAD_open)));
	if (listen_sock)    fprintf(fp, "Listen socket open on port %d\n\n", listen_port);
	else fprintf(fp, "No listen socket open\n\n");
	if (strlen(msg6)) fprintf(fp, "Most recent messages:\n");
	else fprintf(fp, "No messages processed.\n");
	if (strlen(msg1)) fprintf(fp, msg1);
	if (strlen(msg2)) fprintf(fp, msg2);
	if (strlen(msg3)) fprintf(fp, msg3);
	if (strlen(msg4)) fprintf(fp, msg4);
	if (strlen(msg5)) fprintf(fp, msg5);
	if (strlen(msg6)) fprintf(fp, msg6);
   if (strlen(rej3)) fprintf(fp,"Most recent rejections:\n");
   else fprintf(fp,"No rejections processed.\n");
   if (strlen(rej1)) fprintf(fp, rej1);
   if (strlen(rej2)) fprintf(fp, rej2);
   if (strlen(rej3)) fprintf(fp, rej3);
	fprintf(fp, "Total Transactions: today %d, this hour %d\n", transactions_this_day, transactions_this_hour);

/*--------------------------------- Now we do the stuff for each peer */
   for (peer = 0; peer < max_peer; peer++)
   {  fprintf(fp, "\n--- Peer: %s %s %s\n", PEER.sname, PEER.ip, PEER.description);
      for (i=0,sock=0;sock < max_sock; sock++) if (SOCKET.peer == peer) i++;
      if (i) fprintf(fp, "    %d current connections\n", i);
      else fprintf(fp, "    No current connections\n");
	   fprintf(fp, "    Transactions: today %d, this hour %d\n", PEER.transactions_this_day, PEER.transactions_this_hour);
	   if (PEER.msg3 && strlen(PEER.msg3)) fprintf(fp, "    Most recent messages:\n");
	   else fprintf(fp, "    No messages processed.\n");
      if (PEER.msg1 && strlen(PEER.msg1)) fprintf(fp,PEER.msg1);
      if (PEER.msg2 && strlen(PEER.msg2)) fprintf(fp,PEER.msg2);
      if (PEER.msg3 && strlen(PEER.msg3)) fprintf(fp,PEER.msg3);
      
   }
	fclose(fp);	 /* Done */
}

void shutdown_connection_status()
{	FILE * fp;
/*------------------------------------------------- Clear update file */
	if(remove(status_file_pfn)) dbge(status_file_pfn);
	fp = fopen(status_file_pfn, "wb");
	if (!fp) {printf("Unable to create status file: %s\n", status_file_pfn); return;}
/*------------------------------------------------------------ Header */
	fprintf(fp, "PACS Inbound Link shut down at: %s", asctime(tm));
	fclose(fp);	 /* Done */
}

/*----------------------------------------------------------------------
 * Check for delays in processing on RAD or PACS connection.  Print or
 * page errors as appropriate.
 *--------------------------------------------------------------------*/
void link_error_check(void)
{	static time_t last_error_check = 0;
	int i, j, minutes, peer;
   char msg[4096];
/*---------------------------------------------- Check every 5 minutes*/
   if (!last_error_check) last_error_check = now;
	if ((now - last_error_check) < 300) return; 
	last_error_check = now;
/*------------------ We do not error check during the idx down period */
	if (!rad_socket && tm->tm_wday == rad_down_dow &&
		((i = tm->tm_hour * 60 + tm->tm_min)) >= ((j = rad_down_shr * 60 + rad_down_smin)) &&
		i <= (j + rad_down_mins))
	{  rad_basis = now; 
      for (peer=0;peer<max_peer;peer++) PEER.basis = now;
      return; 
   }
	dbg("Not in down period");
/******************************************************************************
 * RAD link warnings
 *****************************************************************************/
	do
/*---------- RAD link warnings ignored if no messages waiting for rad */
   {  for (sock = 0, i=0; sock < max_sock; sock++) 
         if (SOCKET.status == SS_RAD_WRITE || SOCKET.status == SS_RAD_READ) i = 1;
      if (!i) break;
/*------------------------------------------------ Printer / e-mail warnings */
      do
	   {	static time_t last_print = 0;
         if (!last_print) last_print = now;
         if (!print_warning_messages && !e_mail_warning_messages) break;			/* Not printing messages */
         minutes = (now - rad_basis) / 60;  /* minutes since last basis point */
		   if (minutes < rad_warning_mins) break;		 /* haven't reached limit */
		   if ((now - last_print) < (rad_warning_interval * 60)) break; /* haven't reached print interval */
         
		   if (print_window_start <= print_window_end)		/* Not in print window */
            {if (tm->tm_hour < print_window_start || tm->tm_hour > print_window_end) break;}
		   else 
            if (tm->tm_hour < print_window_start && tm->tm_hour > print_window_end) break; 
         
/*--------------------------------------------------- Generate print message */
		   sprintf(em, "Waiting %d minutes on RAD link", minutes); ldbg(em);
		   sprintf(msg, "\f************************************************************\n"
		          	      "* %s - %s"
		                  "* PACS Broker - RAD Link Warning:\n" 
		                  "* RAD connection problem\n"
		                  "* No activity in %d minutes, message waiting\n" 
		                  "* Contact PACS group\n"
		                  "************************************************************\n\n\n\n\n"
                        , rad_host_name, asctime(tm), minutes ); 
		   if (e_mail_warning_messages) 
            if (!alertSendEmail(error_email, "PACSI", msg, NULL))PrtAlertErrorString(debug_fp, "RAD Warning");
		   if (print_warning_messages) 
            if (!alertSendPrint(error_printer, msg, NULL))PrtAlertErrorString(debug_fp, "RAD e-mail");
		   last_print = now;
	   } while (0);   /* EO printer / e-mail messages */

/*------------------------------------------------------------ Pager message */
	   do
	   {	static time_t last_page = 0;
         if (!last_page) last_page = now;
         if (!pager_warning_messages) break;		/* Not paging messages */
		   if (minutes < rad_page_mins) break;		/* haven't reached limit */
		   if ((now - last_page) < (rad_page_interval * 60)) break; /* haven't reached page interval */
         
		   if (pager_window_start <= pager_window_end)		/* Not in pager window */
			   {if (tm->tm_hour < pager_window_start || tm->tm_hour > pager_window_end) break;}
		   else 
			   if (tm->tm_hour < pager_window_start && tm->tm_hour > pager_window_end) break; 
/*-------------------------------------------- Generate pager message */
		   ldbg("Sending pager message - RAD");
		   sprintf(msg, "PACS Brkr Lnk In\nRAD connection\nInactive %d mins\nBounce PACSI\n", minutes);
		   if (!alertSendEmail(pager, "PACSI", msg, NULL))PrtAlertErrorString(debug_fp, "RAD page");
		   last_page = now;
	   } while (0);     /* EO pager messages */ 
      
   } while (0);        /* EO RAD messages */
   
/******************************************************************************
 * PACS link warnings
 *****************************************************************************/
   for (peer = 0; peer < max_peer; peer++)
/*---------------------------------------------------------- Printed message */
   {	do
	   {	if (!PEER.print_warning_messages) break;	 /* Not printing messages */
         minutes = (now - PEER.basis) / 60;
		   if (minutes < PEER.warning_mins) break;	 /* haven't reached limit */
		   if ((now - PEER.last_print) < (PEER.warning_interval * 60)) break; /* haven't reached print interval */
		   if (print_window_start <= print_window_end)		/* Not in print window */
			   {if (tm->tm_hour < print_window_start || tm->tm_hour > print_window_end) break;}
		   else 
			   if (tm->tm_hour < print_window_start && tm->tm_hour > print_window_end) break; 
/*-------------------------------------------- Generate print message */
		   sprintf(em, "Waiting %d minutes on PACS link", minutes); ldbg(em);
		   sprintf(msg, "\f************************************************************\n"
		                  "* %s - %s"
		                  "* %s Warning:\n"
		                  "* PACSI connection problem\n"
		                  "* No activity in %d minutes\n"
		                  "* Contact PACS group\n"
		                  "************************************************************\n\n\n\n\n"
                        , rad_host_name, asctime(tm), PEER.description, minutes);
		   if (!alertSendPrint(error_printer, msg, NULL))PrtAlertErrorString(debug_fp, "PACS print");
		   if (e_mail_warning_messages) 
            if (!alertSendEmail(error_email, "PACSI", msg, NULL))PrtAlertErrorString(debug_fp, "PACS e-mail");
		   PEER.last_print = now;
	   } while (0);

/****************************************************** Pager message */
	   do
	   {	if (!PEER.pager_warning_messages) break;		/* Not paging messages */
		   if (minutes < PEER.page_mins) break;	 /* haven't reached limit */
		   if ((now - PEER.last_page) < (PEER.page_interval * 60)) break; /* haven't reached page interval */
		   if (pager_window_start <= pager_window_end)		/* Not in pager window */
            {if (tm->tm_hour < pager_window_start || tm->tm_hour > pager_window_end) break;}
		   else 
			   if (tm->tm_hour < pager_window_start && tm->tm_hour > pager_window_end) break; 
/*-------------------------------------------- Generate pager message */
		   ldbg("Sending pager message - PACS");
		   sprintf(msg, "%s\nPACSI connection\nInactive %d mins\n", PEER.description, minutes);
		   if (!alertSendEmail(pager, "PACSI", msg, NULL))PrtAlertErrorString(debug_fp, "PACS page");
		   PEER.last_page = now;
	   } while (0);  /* EO pager messages */
   }                /* EO for loop for peers */
}

void log_error(char *msg)
{	FILE *fp; 
	sprintf(elog_pfn, elog_pfn_format, log_dir, tm->tm_year + 1900, tm->tm_mon +1);
	fp = fopen(elog_pfn, "a");
	if (!fp) {dbge(elog_pfn); return;}
	fprintf(fp, "%04d/%02d/%02d %02d:%02d:%02d %s\n", tm->tm_year + 1900, tm->tm_mon +1,
		tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec, msg);
	fclose(fp);
}

void log_message(char * note, char *msg)
{	char buffer[125], *ptr = buffer;
	FILE *fp;
	int a, cc = 0;

	sprintf(log_pfn, log_pfn_format, log_dir, tm->tm_year + 1900, tm->tm_mon +1, tm->tm_mday);
	fp = fopen(log_pfn, "a");
	if (!fp) {dbge(log_pfn); return;}
	fprintf(fp, "\n%04d/%02d/%02d %02d:%02d:%02d %s --------------------------\n",
		tm->tm_year + 1900, tm->tm_mon +1, tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec, note);
	if (!strlen(msg)) fprintf (fp, "No data\n");

	while (*msg)
	{	if(*msg == 11){msg++; if(cc)fprintf(fp, "\n"); fprintf(fp, "<SB>\n"); continue;}
		if(*msg == 28){msg++; if(cc)fprintf(fp, "\n"); fprintf(fp, "<EB>\n"); continue;}
		if(*msg == '\n') {msg++; fprintf(fp, "<N>\n"); continue;}
		if(*msg == '\r') {strcpy(ptr, "<R>"); fprintf(fp, "%s\n", buffer); ptr = buffer; cc = 0;}
		else
		{	if (cc == 90)
			{	*ptr = 0;
				fprintf (fp, "%s\n", buffer);
				ptr = buffer;
				for (a = 0, cc = 0; a < 4; a++, cc++) *ptr++ = ' ';
			}
			*ptr++ = *msg;
			cc++;
		}
		msg++;
	}
	*ptr = 0;

	if (strlen(buffer)) fprintf (fp, "%s\n", buffer);
	fclose(fp);
}

/*----------------------------------------------------------------------
 * Checks to see if purge has been run yet today.  If not, runs it, and
 * purges log files older than retain days.  This routine is run at a
 * timeout, so will lag during busy times.
 *--------------------------------------------------------------------*/
void check_purge_logs(void)
{	struct tm tdy, *lp;
	char cut_pfn[128], pfn[128];
	int month, year;
   time_t cut;
   DIR * dp;
   struct dirent * de;
/*------------------------------------- Determine if we need to purge */
	memcpy(&tdy,tm,sizeof(tdy));
	lp = localtime(&last_purge_date);
	if (tdy.tm_year == lp->tm_year && tdy.tm_yday == lp->tm_yday) return;
	if (tdy.tm_hour < log_purge_hour) return;
	last_purge_date = now;		/* We are going to purge; Update last purge date */
/*------------------------------------ Open log directory for reading */
   if (!(dp = opendir(log_dir))) { dbge("purge, couldn't open log dir"); return; }
/*----------------------------------------------------------------------
 * First purge is message logs.
 * Set up wildcard pfn for lookup, purge cutoff pfn for comparison 
 *--------------------------------------------------------------------*/
	cut = now - (log_retain_days * 86400);
	lp = localtime(&cut);
	sprintf(cut_pfn, log_pfn_format, "", lp->tm_year + 1900, lp->tm_mon +1, lp->tm_mday);
	sprintf(em,"Purging log files before %s", cut_pfn); ldbg(em);
/*----------------------- Pass all files, purging those before cutoff */
	files_purged = 0;
	while (de = readdir(dp))
	{	strcpy(pfn,de->d_name);
      if (STR_EQ(strchr(pfn,'.'),".FULL") && (strcmp(pfn, cut_pfn) < 0))
		{	sprintf(em,"Purging file %s", pfn); ldbg(em);
         strcpy(pfn,log_dir); strcat(pfn,de->d_name);
			if (remove(pfn)) dbge(pfn); 
         rewinddir(dp);
			files_purged++; 
		}
	}
/*----------------------------------------------------------------------
 * Second purge is error logs.  Same as above, but by months
 *--------------------------------------------------------------------*/
   rewinddir(dp);
	lp = localtime(&now);
	year = lp->tm_year + 1900; month = lp->tm_mon + 1 - elog_retain_months;
	while (month < 1) { year--; month += 12; }
	sprintf(cut_pfn, elog_pfn_format, "", year, month);
	sprintf(em,"Purging error log files before %s", cut_pfn); ldbg(em);
/*----------------------- Pass all files, purging those before cutoff */
	efiles_purged = 0;
	while (de = readdir(dp))
	{	strcpy(pfn,de->d_name);
      if (STR_EQ(strchr(pfn,'.'),".ERR") && (strcmp(pfn, cut_pfn) < 0))
		{	sprintf(em,"Purging file %s", pfn); ldbg(em);
         strcpy(pfn,log_dir); strcat(pfn,de->d_name);
			if (remove(pfn)) dbge(pfn); 
         rewinddir(dp);
			files_purged++; 
		}
	}
   if (closedir(dp)) dbge("purge, couldn't close log dir"); 
}

/*---------------------------------------- Load environment variables */
void read_ini(void)
{	char a[1024], b[128], *p;
   char main[] = "[MAIN]";
   int sa = sizeof(a), i;
/*------------------------------------------------------- Pull debug */
   if (!GetIniString(main, "Debug", "stdout", a, sa, inipfn))
      dbg_exit(strcmb("Debug ", GetIniErrorString()), 1);
   debug_fp = debug_on(a);
   
/*--------------------------------- Host (RAD) server identification */
   GetIniString(main, "RADHostName", rad_host_name, rad_host_name, sizeof(rad_host_name), inipfn);
   GetIniString(main, "RADHostIP", rad_host_ip, rad_host_ip, sizeof(rad_host_ip), inipfn);
   GetIniInt(main, "RADHostPort", rad_host_port, &rad_host_port, inipfn);
/*------------------------------------------------------- status file */
   GetIniInt(main, "StatusUpdateSeconds", status_update_seconds, &status_update_seconds, inipfn); 
   GetIniInt(main, "ControlCheckSeconds", control_check_seconds, &control_check_seconds, inipfn);
/*------------------------------------------------------- listen port */
   GetIniString(main, "HostName", host_name, host_name, sizeof(host_name), inipfn);
   GetIniInt(main, "ListenPort", listen_port, &listen_port, inipfn);
/*------------------------------------------------ Log file directory */
   GetIniString(main, "LogDir", log_dir, log_dir, sizeof(log_dir), inipfn);
/*-------------------------- Log purge hour, days to retain log files */
   GetIniInt(main,"LogPurgeHour", log_purge_hour, &log_purge_hour, inipfn);
   GetIniInt(main,"LogRetainDays", log_retain_days, &log_retain_days, inipfn);
/*---------------------------------- Months to retain error log files */
   GetIniInt(main,"ELogRetainMons", elog_retain_months, &elog_retain_months, inipfn);
/*------------------------------------------------- Error log printer */
   GetIniString(main,"ErrorPrinter", error_printer, error_printer, sizeof(error_printer), inipfn);
/*---------------------------- Should error & warning messages print? */
   GetIniString(main, "PrintWarningMessages", "YES", a, sa, inipfn);
   if (STR_EQ(a,"ON") || STR_EQ(a,"YES")) print_warning_messages = 1;
/*--------------------------- Should error & warning messages e-mail? */
   GetIniString(main, "EmailWarningMessages", "YES", a, sa, inipfn);
   if (STR_EQ(a,"ON") || STR_EQ(a,"YES")) e_mail_warning_messages = 1;
   GetIniString(main, "ErrorEmail", error_email, error_email, sizeof(error_email), inipfn);
/*----------------------------- Warning print window (Mon-Fri) 24 hour clock */
   GetIniInt(main, "PrintWindowStart", print_window_start, &print_window_start, inipfn);
   GetIniInt(main, "PrintWindowEnd",   print_window_end,   &print_window_end, inipfn);
/*------------------------------------ Should error & warning messages page? */
   GetIniString(main, "PagerWarningMessages", "YES", a, sa, inipfn);
   if (STR_EQ(a,"ON") || STR_EQ(a,"YES")) pager_warning_messages = 1;
   GetIniString(main, "Pager", pager, pager, sizeof(pager), inipfn); 
/*------------------------------------- pager window (Mon-Fri) 24 hour clock */
   GetIniInt(main, "PagerWindowStart", pager_window_start, &pager_window_start, inipfn);
   GetIniInt(main, "PagerWindowEnd",   pager_window_end,   &pager_window_end,   inipfn);
/*------------------ RAD down day, start hr and min, and minutes long */
   GetIniInt(main, "RADDownDOW", rad_down_dow, &rad_down_dow, inipfn);
   GetIniInt(main, "RADDownSHr", rad_down_shr, &rad_down_shr, inipfn);
   GetIniInt(main, "RADDownSMin", rad_down_smin, &rad_down_smin, inipfn);
   GetIniInt(main, "RADDownMins", rad_down_mins, &rad_down_mins, inipfn);
/*----------------------------------- Timing for RAD warnings, pages */
   GetIniInt(main, "WarningMins", rad_warning_mins, &rad_warning_mins, inipfn);
   GetIniInt(main, "WarningInterval", rad_warning_interval, &rad_warning_interval, inipfn);
   GetIniInt(main, "PageMins", rad_page_mins, &rad_page_mins, inipfn);
   GetIniInt(main, "PageInterval", rad_page_interval, &rad_page_interval, inipfn);
     
/*----------------------- Load FEED_ section names which are Status=ON */
   GetIniSectionNames(a, sa, inipfn);   
   for (peer=0, p=strtok(a,"[]");p;p=strtok(p,"[]")) 
   {  if (strstr(p,"FEED_"))
      {  GetIniString(p, "Status", "OFF", b, sizeof(b), inipfn);
         if (STR_EQ(b,"ON"))
         {  STR(PEER.sname,p) 
            max_peer = ++peer; 
         }
      }
      p += strlen(p) + 1;
   }
/*------------------------------------- Now load data for those feeds */
   for (peer = 0; peer < max_peer; peer++)
   {  GetIniString(PEER.sname, "Description", PEER.sname, a, sa, inipfn);
      STR(PEER.description,a)
      GetIniString(PEER.sname, "PeerIP", " ", a, sa, inipfn);
      STR(PEER.ip,a)
         
      GetIniString(PEER.sname, "MultipleConnections", "NO", a, sa, inipfn);
      if (STR_EQ(a,"YES") || STR_EQ(a,"ON")) PEER.multiple_connections = 1;
      else PEER.multiple_connections = 0;
      GetIniString(PEER.sname, "PersistantConnections", "NO", a, sa, inipfn);
      if (STR_EQ(a,"YES") || STR_EQ(a,"ON")) PEER.persistant_connections = 1;
      else PEER.persistant_connections = 0;
      GetIniInt(PEER.sname, "TimeoutSeconds", 0, &i, inipfn); 
      PEER.timeout_seconds = i;
      
/*--------- Main section items are defaults for these feed parameters */
      GetIniString(PEER.sname, "PrintWarningMessages", (print_warning_messages ? "YES" : "NO"), a, sa, inipfn);
      if (STR_EQ(a,"YES") || STR_EQ(a,"ON")) PEER.print_warning_messages = 1;
      else PEER.print_warning_messages = 0;

      GetIniInt(PEER.sname, "WarningMins",     rad_warning_mins,     &PEER.warning_mins,     inipfn);
      GetIniInt(PEER.sname, "WarningInterval", rad_warning_interval, &PEER.warning_interval, inipfn);
      
      GetIniString(PEER.sname, "PagerWarningMessages", (pager_warning_messages ? "YES" : "NO"), a, sa, inipfn);
      if (STR_EQ(a,"YES") || STR_EQ(a,"ON")) PEER.pager_warning_messages = 1;
      else PEER.pager_warning_messages = 0;

      GetIniInt(PEER.sname, "PageMins",     rad_page_mins,     &PEER.page_mins,     inipfn);
      GetIniInt(PEER.sname, "PageInterval", rad_page_interval, &PEER.page_interval, inipfn);

      PEER.basis = now;
      PEER.transactions_this_day = PEER.transactions_this_hour = 0;
   }
}

/*----------------------------------------------------------------------
 * Check control file for user commands and carry them out.
 *--------------------------------------------------------------------*/
int control_check(void)
{  FILE * fp;
	char buffer[256], cmd[256];
	
/*-------------------------------- Check control file every x seconds */
	if (difftime(now, last_control_check) < control_check_seconds) return(0);
	last_control_check = now;

/*-------------------------- Open control file, read command (if any) */
	fp = fopen(control_pfn, "r");
	if (!fp) return(0);
	fgets(buffer, sizeof(buffer), fp);
	if (feof(fp))
	{	fclose(fp);
		if(remove(control_pfn)) dbge(control_pfn);
		return(0);
	}
	fclose(fp);
	if(remove(control_pfn)) dbge(control_pfn);
	sprintf(em,"Control file read: %s", buffer); dbg(em);
	strtok(buffer, " \r\n");

/*----------------------------------------- Shut down and exit server */
	if(!strcasecmp(buffer, "EXIT"))
	{	ldbg("Control exit");
		main_loop = tcp_loop = 0;
		return(1);
	}
/*----------------------------------------------------- Bounce server */
	if(!strcasecmp(buffer, "BOUNCE"))
	{	ldbg("Control bounce");
		tcp_loop = 0;
		interface_bounced_at = time(NULL);
		return(1);
	}
	return (0);
}

/***********************************************************************
 * Debug prints
 **********************************************************************/
void ldbg(char *msg) {log_error(msg); dbg(msg);}
void ldbge(char * msg)
{  char * p, * def = "No errno text available";
   if (!(p=strerror(errno))) p = def;
   sprintf(em,"%s: %s", msg, p);
   ldbg(em);
}
/*----------------------------------------- Debug to dump buffer area */
void dbgm(char *msg, char *buff)
{	char p[53], *ptr;
	int piece = sizeof(p) - 3, cont = 1;
	if (!debug_fp) return;
	if (strlen(msg) > (sizeof(p) -3)) dbg(msg);
	else
	{	 memset(p,0,sizeof(p)); strcpy(p,msg);
		 if (p[strlen(p)-1] != ' ')p[strlen(p)] = ' ';
		 memset(p+strlen(p),'-',(sizeof(p)-strlen(p)-1));
		 dbg(p);
	}
	if (!buff) { dbg("NULL buffer\n"); return; }
	while (cont)
	{	memset(p,0,sizeof(p)); strcpy(p,"> ");
		if (strlen(buff) < piece) { strcat(p,buff); cont = 0; }
		else { memcpy(p+2,buff,piece); buff += piece; }
		if (ptr = strchr(p,SB)) *ptr = '<';
		if (ptr = strchr(p,EB)) *ptr = '>';
		while (ptr = strchr(p,'\r')) *ptr = '~';
		dbg(p);
	}
	dbg(" ");
}

/*---------------------------------------------- Update current time */
void time_hack()
{  now=time(NULL); 
   tm=localtime(&now);
   strcpy(stm,asctime(tm));
}

/*-----------------------------------------------------------------------------
 * Returns pointer to string which is cat of s1 and s2, but s1 can be a 
 * constant.
 *---------------------------------------------------------------------------*/
char * strcmb(char *s1, char *s2)
{	static char ws[128];
	memset(ws,0,sizeof(ws));
	strcpy(ws,s1);
	return strcat(ws,s2);
}
