/**************************************************************************************************
 * sql.c                                                                                  R Moulton
 * Library of routines for mysql database access.
 *-------------------------------------------------------------------------------------------------
 * When compiling, use: gcc -c -I/usr/local/mysql/include sql.c
 * When linking, use:   -L/usr/local/mysql/lib -lmysqlclient -lsocket -lnsl
 *-------------------------------------------------------------------------------------------------
 * 12-3 R Moulton Initial write
 *************************************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

#include "sql.h"
#include "ini.h"
#include "dbg.h"

/*--------------------------------------------------------------------------------------- Macros */
#define FREE(A) {free(A);A=NULL;}
#define STR_EQ(A,B) !strcasecmp(A,B)
#define STR(A,B) {if(B){if(A=realloc(A,strlen(B)+3))strcpy(A,B);}else{free(A);A=NULL;}}
#define LCHR(A) A[strlen(A) - 1]

#define DEFAULT_INI_FILE_PFN "sql.ini"   
#define SQL_R_OK    0
#define SQL_R_FAIL  1

static unsigned int quoted = FIELD_TYPE_DATE | FIELD_TYPE_TIME | FIELD_TYPE_DATETIME | FIELD_TYPE_STRING |
                             FIELD_TYPE_BLOB | FIELD_TYPE_ENUM | FIELD_TYPE_SET;

/*------------------------------------------------ Error msg strings.  Match to sql_errno values */
static char * e_msg[] =
{  "No error",
   "Null database name",
   "Null user name",
   "Null ini pfn",
   "ini file error",
   "mysql init failure",
   "No valid db connection",
   "No valid table",
   "No columns in result set",
   "Memory allocation error",
   "Missing/invalid field name",
   "Operation failed",
   "ini file not found",
   "ini db section not found"
};
static int max_e_msg = sizeof(e_msg) / sizeof(char *);
static char * no_msg = "Unknown sql_errno value";

static char * inipfn = NULL;

static FILE * debug_fp = NULL;

   static char * oemsg = "NULL conn";

/*---------------------------------------------------------------------------- Sets ini file pfn */
int sqlSetIniPfn(char * new_pfn)
{  sql_errno = SQL_OK;
   if (!new_pfn) { sql_errno = SQL_NOINIPFN; return SQL_R_FAIL; }
   STR(inipfn,new_pfn)
   return SQL_R_OK;
}

/*-------------------------------------------------------------------------------------------------
 * Connects to a db server and makes a database current.  db is the section in the .ini file, and 
 * will be used as the database name unless there is a Database entry in that section which 
 * overides it. Returns pointer to new connection structure.
 *-----------------------------------------------------------------------------------------------*/
MYSQL * sqlOpen(char * db, char * user)
{  char *host_name = NULL, *password = NULL, *socket_name = NULL, *db_name = NULL;
   char a[2048], *p;
   int i, sa = sizeof(a);
   unsigned int flags, port_num;
   static MYSQL * conn = NULL;
   if (!inipfn) inipfn = DEFAULT_INI_FILE_PFN;
   sql_errno = SQL_OK;
   if (!db)   { sql_errno = SQL_NODBNAME; return NULL; }                  /* Validate parameters */
   if (!user) { sql_errno = SQL_NOUSER;   return NULL; }
   STR(db_name,db)
   flags = 0;

/*--------------------------------------------- Ini file must exist and have section matching db */
   if (access(inipfn, F_OK)) { sql_errno = SQL_ININOF; return NULL; }
   if (!GetIniSectionNames(a, sa, inipfn)) { sql_errno = SQL_INI; return NULL; }
   for (i=0, p=strtok(a,"[]"); p; p=strtok(NULL,"[]")) if (STR_EQ(p,db)) i=1;
   if (!i) {sql_errno = SQL_SECNOF; return NULL; }
   
/*----------If there is a DataBase entry, the db was a nickname; if not, it is the database name */
   GetIniString(db, "DataBase", db, a, sizeof(a), inipfn);
   STR(db_name,a);

/*--------------------------------------------------------------------- Pull items from ini file */
   GetIniString(db,"HostName", "NULL", a, sizeof(a), inipfn);
   if(strcasecmp(a,"NULL")) STR(host_name,a)
   GetIniInt(db,"PortNum", 0, &i, inipfn);
   port_num = i;
   GetIniString(db,"SocketName", "NULL", a, sizeof(a), inipfn);
   if (strcasecmp(a,"NULL")) STR(socket_name,a)
      
/*------------------------------------ User name (provided in call) is kname, password is kvalue */
   GetIniString(db, user, "NULL", a, sizeof(a), inipfn);
   if(strcasecmp(a,"NULL")) STR(password,a)

/*---------------------------------------------------------------------------------------- Flags */
   GetIniInt(db,"Flags", 0, &i, inipfn);
   flags = i;

/*------------------------------------ Make new structure and open connection to server/database */
   conn = mysql_init(NULL);
   if (!conn) { sql_errno = SQL_INITFAIL; return NULL; }

   if (mysql_real_connect(conn, host_name, user, password, db_name, port_num, socket_name, flags) == NULL)
   {  sql_errno = mysql_errno(conn);
      p = mysql_error(conn);
      STR(oemsg,p)
      return NULL;
   }
   return conn;
}

#define VALUE tbl[i].value
#define FIELD tbl[i].field
/**************************************************************************************************
 * Load the definition of a table, returning pointer to allocated structure, array of field structs
 * The array is created at one more than the number of fields in the record, with NULLs in the last
 * slot.  The default value is also loaded, using mysql_list_fields.
 *************************************************************************************************/
struct _sql_table_ * sqlLoadRecordDef(MYSQL * conn, char * table_name)
{  char a[1024];
   int rc, num_fields, i;
   MYSQL_RES * rslt, * rslt2;
   MYSQL_FIELD * fld;
   static struct _sql_table_ * tbl;
   sql_errno = SQL_OK;
/*---------------------------------------------------------------------------------- Validations */
   if (!conn) { sql_errno = SQL_CONN; return NULL; }
   if (!table_name) { sql_errno = SQL_NOTABLE; return NULL; }
/*----------------------------------------------------- SQL query to return all columns in table */
   if (!(rslt2 = mysql_list_fields(conn, table_name, NULL))) { sql_errno = mysql_errno(conn); return NULL; }
   sprintf(a,"SELECT * FROM %s LIMIT 1;", table_name);
   if (mysql_real_query(conn, a, strlen(a))) { sql_errno = mysql_errno(conn); return NULL; }
   if (!(rslt = mysql_store_result(conn)))   { sql_errno = mysql_errno(conn); return NULL; }
/*--------------------------------------------------------------------- Allocate and build table */
   if (!(num_fields = mysql_field_count(conn))) { sql_errno = SQL_NOFIELDS; return NULL; }
   tbl = malloc(sizeof(struct _sql_table_) * (num_fields + 1) + 3);
   if (!tbl) { sql_errno = SQL_MALLOC; return NULL; }
   for (i = 0; i < num_fields; i++)
   {  FIELD = mysql_fetch_field_direct(rslt, i);
      VALUE = NULL;
      fld = mysql_fetch_field_direct(rslt2, i);
      STR(FIELD->def,fld->def)
   }
   FIELD = NULL;
   VALUE = NULL;
   mysql_free_result(rslt);
   
   return (tbl);
}
/*-------------------------------------------- Frees memory allocated to a _sql_table_ structure */
int sqlFreeRecord(struct _sql_table_ * tbl)
{  int i;
   sql_errno = SQL_OK;
   if (!tbl) { sql_errno = SQL_NOTABLE; return (SQL_R_FAIL); }
   for (i = 0; FIELD; i++)
   {  FREE(VALUE)
      FREE(FIELD->def)
      free(FIELD);
   }
   return (SQL_R_OK);
}
/*---------------------- Clears value fields in _sql_table_ structure and sets to default values */
int sqlClearRecord(struct _sql_table_ * tbl)
{  int i;
   sql_errno = SQL_OK;
   if (!tbl) { sql_errno = SQL_NOTABLE; return (SQL_R_FAIL); }
   for (i = 0; FIELD; i++)
   {  STR(VALUE, FIELD->def)
   }
   return (SQL_R_OK);
}
/*-------------------------------------------------- Set field "name" in record "tbl" to "value" */
int sqlSetValue(struct _sql_table_ * tbl, char * name, char * value)
{  int i;
   sql_errno = SQL_OK;
   if (!tbl)  { sql_errno = SQL_NOTABLE; return (SQL_R_FAIL); }
   if (!name) { sql_errno = SQL_FIELD;   return (SQL_R_FAIL); }
   for (i = 0; FIELD; i++)
   {  if (STR_EQ(FIELD->name,name))
      {  STR(VALUE, value)
         return (SQL_R_OK);
      }
   }
   sql_errno = SQL_FIELD;
   return (SQL_R_FAIL);
}

/*-------------------------------------------------------------------------------------------------
 * Generates insert for record with values set up for values set up in _sql_table_ structure.
 * Checks to insure that the record was added and returns dba value in dba.  
 *-----------------------------------------------------------------------------------------------*/
int sqlInsertRecord(MYSQL * conn, struct _sql_table_ * tbl, unsigned int flags, unsigned long * dba)
{  int i = 0, auto_dba = 0, rc;
   char a[4096], b[2046];
   sql_errno = SQL_OK;
   *dba = 0;
/*---------------------------------------------------------------------------------- Validations */
   if (!conn) { sql_errno = SQL_CONN; return (SQL_R_FAIL); }
   if (!tbl || !FIELD)  { sql_errno = SQL_NOTABLE; return (SQL_R_FAIL); }
/*----------------------------------------------------------------------------- Build SQL INSERT */
   memset(a,0,sizeof(a));
   strcpy(a,"INSERT ");
   if (flags & SQLF_LOW_PRIORITY) strcat(a,"LOW_PRIORITY");
   else if (flags & SQLF_DELAYED) strcat(a,"DELAYED ");
   if (flags & SQLF_IGNORE) strcat(a,"IGNORE ");
   strcat(a,FIELD->table); 
   strcat(a," VALUES (");
   for (i = 0; FIELD; i++)
   {  if (FIELD->flags | AUTO_INCREMENT_FLAG) auto_dba = 1;
      if (i) strcat(a,", ");
      if (!VALUE) { strcat(a,"NULL"); continue; }
      if (FIELD->type & quoted)
      {  strcat(a,"'");
         mysql_real_escape_string(conn, b, VALUE, strlen(VALUE));
         strcat(a,b);
         strcat(a,"'");
      }
      else strcat(a,VALUE);
   }
   strcat(a,"0;");
/*-------------------------------------------------------------------------------- Do the insert */
   if (mysql_real_query(conn, a, strlen(a))) { sql_errno = mysql_errno(conn); return (SQL_R_FAIL); }
   if ( flags & (SQLF_LOW_PRIORITY | SQLF_DELAYED)) return (SQL_R_OK);
   *dba = mysql_insert_id(conn);   
   if(mysql_affected_rows(conn) != 1) { sql_errno = SQL_OPFAILED; return (SQL_R_FAIL); }
   return (SQL_R_OK);
}

/*----------------------------------------------------------------- Close an existing connection */
int sqlClose(MYSQL * conn)
{  mysql_close(conn);
   sql_errno = SQL_OK;
   return SQL_R_OK;
}

char * GetSqlErrorString(MYSQL * conn)
{  if (sql_errno > 999) if (conn) return mysql_error(conn); else return oemsg;
   if (sql_errno < max_e_msg) return e_msg[sql_errno];
   return no_msg;
}
void PrtSqlErrorString (MYSQL * conn, FILE * file, char * prefix)
{	if (!file) file = stdout;
	fprintf(file, "%s: %s\n", prefix, GetSqlErrorString(conn));
	return;
}

