/****************************************************************************************
 * tce.c                                                                        R Moulton
 * Program is designed to emulate a Telnet client, log into a host, and process a 
 * pre-defined script, loading desired portions into an output file for processing.
 ****************************************************************************************
 * Calling format: tce [inipfn=tce.ini]
 ****************************************************************************************
 * Compile and link with: gcc -g tce.c -o tce `pkg-config --cflags --libs tce`
 ***************************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>

#include <dbg.h>
#include <ini.h>
#include <log.h>

/*----------------------------------------------------------------------------- Macros */
#define STR(A,B) { if (A) free(A); A=strdup(B); }
#define FREE(A) if (A) { free(A); A=NULL; }
#define STR_EQ(A,B) !strcasecmp(A,B)
#define ISON(A) (!strcasecmp(A,"ON") || !strcasecmp(A,"YES"))
#define LCHR(A) A[strlen(A) -1]
#define CLEAR(A) memset(A,0,sizeof(A))

/****************************** Function Prototypes ************************************/
static int read_ini(int numpars, char *par[]);
static int hex_it(char * str);
static char * de_hex(char * str);
static void write_str(char * str, int ts);
static void read_until(char * str);

/****************************************************************************************
 * .ini file items
 ***************************************************************************************/
static char * cwd;
#define DEF_INI_PFN "tce.ini"
static char inipfn[256] = "\0";
static FILE * debug_fp;
static char * target_host;
static char * target_ip;
static int    target_port;
static char dialog_pfn[256];
static FILE * dialog_fp;
static char capture_pfn[256];
static char over_write[2] = "a";
static FILE * capture_fp;
static char capture_on;

/***********************************************************************
 * Other global working storage
 **********************************************************************/
char a[512];
int sock, i, j, retval;
struct hostent *psrvent, srvent;
struct sockaddr_in sock_name;
fd_set socket_set, eset;
struct timeval timeout;

/***********************************************************************
 * close out socket
 **********************************************************************/
cleanup()
{  retval = shutdown(sock, 2);
   if(retval != 0) perror("shutdown");
   retval = close(sock);
   if(retval) perror("close");
}

/****************************************************************************************
 * main function
 ***************************************************************************************/
int main(int numpars, char *par[])
{  
/*-------------------------------------------------------------- Load .ini file values */
   if (read_ini(numpars, par)) return (1);
   
   sprintf(em, "Telnet Client Emulator\n  %s %s port %d", target_host, target_ip, target_port); dbg(em); 
   sprintf(em, "  Dialog: %s\n  Capture: %s", dialog_pfn, capture_pfn); dbg(em);
   
   if(!(dialog_fp = fopen(dialog_pfn,"r"))) dbge_exit("dialog open", 1);
   if(!(capture_fp = fopen(capture_pfn,over_write))) dbge_exit("capture open", 1);
   
   dbg("Creating socket");
   if ((sock=socket(AF_INET, SOCK_STREAM, 0)) == -1) dbge_exit("socket",1);

   dbg("Get host by name");
   psrvent = gethostbyname(target_host);
   if(!psrvent) 
   {  dbg("Get host by IP");
      psrvent = gethostbyname(target_ip);
   }
   if(!psrvent) dbge_exit("gethost",1);
   
   dbg("Setting socket name entries");
   srvent = *psrvent;
   sock_name.sin_family = srvent.h_addrtype;
   sock_name.sin_port = htons(target_port);
   sock_name.sin_addr = *((struct in_addr *)srvent.h_addr);
   
   dbg("Connect to host");
   i = sizeof(sock_name);
   if (connect(sock, (struct sockaddr *)&sock_name, i) < 0) dbge_exit("Connect",1);
   j = sock + 1;

/****************************************************************************************
 * Main processing loop, driven by reading commands from the dialog file
 ***************************************************************************************/
   while (1)
   {  char dialog_line[256], *p, *t;
      int ts;

      CLEAR(dialog_line);                                       /* read dialog line */
      if (!fgets(dialog_line, sizeof(dialog_line), dialog_fp))
      {  if (feof(dialog_fp)) break;
         dbge_exit("dialog read",1);
      }
/*---------------------------------------------- Trim NL's and spaces, then debug echo */
      while (LCHR(dialog_line) == '\n' || LCHR(dialog_line) == ' ' || LCHR(dialog_line) == '\r') 
         LCHR(dialog_line) = '\0';
      dbg(dialog_line);
      
      if (dialog_line[0] == '#') continue;                             /* Comments */
      if (!strlen(dialog_line)) continue;                              /* blank lines */

      if (!(p=strtok(dialog_line," "))) { dbg("no command code"); continue; }
      t = strtok(NULL,"|");
      
/*-------------------------------------------------------------------- end of session */
      if (STR_EQ(p,"END")) break;
      
/*--------------------------------------------------------------- Turn capture on/off */
      if (STR_EQ(p,"CAPTURE"))
      {  if (ISON(t)) capture_on = 1;
         else capture_on = 0;
         continue;
      }

/*------------------------------------------------- Check for hex to ascii conversion */
      ts = strlen(t);
      if (strstr(p,"HEX")) ts = hex_it(t);

/*------------------------------------------------------------------ Reads and writes */
      if (strstr(p,"READ"))  { read_until(t); continue; }
      if (strstr(p,"WRITE")) { write_str(t, ts); continue; }

      dbg("Invalid command code");
      continue;
   }
   dbg("End of dialog");
   cleanup();
   return (0);
}

/*--------------------------------------------------------------------------------------
 * Convert hex coded text to actual hex string in place. Skips non-hex characters
 *------------------------------------------------------------------------------------*/
static int hex_it(char * str)
{  unsigned char work[128];
   char * p, l;
   unsigned char * o, w;
   int ts = 0;
   CLEAR(work);
   for (p=str, o=work, l=1; *p; p++)
   {   if (*p >= '0' && *p <= '9') w = *p - '0';
       else if (*p >= 'a' && *p <= 'f') w = *p - 'a' + 10;
       else if (*p >= 'A' && *p <= 'F') w = *p - 'A' + 10;
       else continue; 
       if (l) { *o = 16 * w; l = 0; }
       else { *o += w; ts++; o++; l = 1; }
   }
   memmove(str,work,strlen(work)+1);
   return (ts);
}

static char * de_hex(char * str)
{  unsigned char work[128], w, *p;
   static char ws[350], *o;
   CLEAR(work);
   CLEAR(ws);
   memmove(work,str,125);
   for (p=work, o=ws; *p || *(p+1) || *(p+2); p++)
   {  if (*p == 255) { *o = ' '; o++; }
      w = *p / 16; 
      if (w < 10) *o = w + '0'; 
      else *o = (w - 10) + 'a'; 
      o++;
      w = *p % 16;  
      if (w < 10) *o = w + '0'; 
      else *o = (w - 10) + 'a'; 
      o++;
   }
   return ws;
}
/*---------------------------------------------------------------------------------------
 * Send string to telnet host
 *-------------------------------------------------------------------------------------*/
static void write_str(char * str, int ts)
{  
   FD_ZERO (&socket_set);
   FD_SET (sock, &socket_set);

   retval = select(j, NULL, &socket_set, NULL, NULL);
   if (retval < 1) dbge_exit("write select", 1);
   
   retval = send(sock, str, ts, 0);
   if (retval == -1) dbge_exit("send", 1);
   sprintf(em, "sending %d, returned %d", ts, retval); dbg(em);

   return;
}

/*--------------------------------------------------------------------------------------
 * read from host until output includes string in param
 * If capture is on, write everything you read to capture file
 * Note: The string we are matching could be partly in one packet and partly in the 
 * next, so we hang on to a bit of each packet so we can still find the match.
 *------------------------------------------------------------------------------------*/
static void read_until(char * str)
{  static char b[20 * 1024], *p;
   static size_t sb = sizeof(b);
   static struct timeval tv = {60, 0};

   CLEAR(b);

   while (1)
   {  
      FD_ZERO(&socket_set); FD_SET(sock, &socket_set);
      FD_ZERO(&eset); FD_SET(sock, &eset);
      
      retval = select(j, &socket_set, NULL, &eset, &tv);

      if (!retval) continue; 
      if (retval < 0) dbge_exit("read select", 1);

      if (FD_ISSET(sock, &eset)) dbg_exit("read select error", 1);

      i = strlen(b);
      retval = recv(sock, b + i, sb - i, 0);
      sprintf(em,"recv returned %d", retval); dbg(em);
      if (!retval) dbg_exit("session closed by host", 1);
      if (retval < 0) dbge_exit("recv", 1);

/*-------------------------------------------------------------------- handle capture */
      if (capture_on)
      {  retval = fputs(b + i, capture_fp);
         if (retval <= 0) dbge_exit("capture file", 1);
      }
      dbg(de_hex(b+i));
/*------------------------------------------------------------- Check for done string */ 
      if (strstr(b,str)) { dbg("Matched"); return; }
/*-------------------------------------- Store enough for match and loop to next read */
      memmove(b,b + strlen(b) - strlen(str), strlen(str));
      memset(b+strlen(str),0,sb - strlen(str));
   }
}

/*************************************************************** Read .ini file values */
static int read_ini(int numpars, char * par[])
{  char * p, ws[1024];
   int sws = sizeof(ws);

   if(p=MakeProgDirCurrent(par[0])) STR(cwd,p) else return (1);
   LCHR(cwd) = '\0';
/*--------------------------------------------- par[1] is .ini file pfn, or we default */
   strcpy(inipfn,cwd);
   if ((numpars > 1) && strlen(par[1]))
   {  if (*(par[1]) == '/') strcpy(inipfn,par[1]);
      if (*(par[1]) == '.') strcat(inipfn,par[1] +1);
   }
   else {strcat(inipfn,"/"); strcpy(inipfn,DEF_INI_PFN);}
/*------------------------------------------------------------------ Validate .ini pfn */
   if (access(inipfn, F_OK)) return (1);

   if (!GetIniString("[MAIN]", "Debug", "NULL", ws, sws, inipfn)) return (1);
   debug_fp = debug_on(ws);

   if (logInitialize(inipfn)) dbg(GetLogErrorString());

/*-------------------------------------------------------------- Connection parameters */
   GetIniString("[HOST]", "Name", "NOF", ws, sws, inipfn);
   if (strcmp(ws,"NOF")) STR(target_host,ws) else { dbg("Host name missing in .ini file"); return (1); }
   GetIniString("[HOST]", "IP", "NOF", ws, sws, inipfn);
   if (strcmp(ws,"NOF")) STR(target_ip,ws) else { dbg("Host IP missing in .ini file"); return (1); }
   GetIniInt("[HOST]", "Port", 0, &target_port, inipfn);
   if (!target_port) { dbg("Host port missing in .ini file"); return (1); }

/*----------------------------------------------------------- Dialog and capture files */
   GetIniString("[FILES]", "Dialog", "NOF", ws, sws, inipfn);
   if (strcmp(ws,"NOF")) 
   {  strcpy(dialog_pfn,cwd);
      if (ws[0] == '/') strcpy(dialog_pfn,ws);
      else if (ws[0] == '.') strcat(dialog_pfn,ws+1);
      else {strcat(dialog_pfn,"/"); strcat(dialog_pfn,ws); }
   } else { dbg("Dialog file name missing in .ini file"); return (1); }
    
   GetIniString("[FILES]", "Capture", "NOF", ws, sws, inipfn);
   if (strcmp(ws,"NOF")) 
   {  strcpy(capture_pfn,cwd);
      if (ws[0] == '/') strcpy(capture_pfn,ws);
      else if (ws[0] == '.') strcat(capture_pfn,ws+1);
      else {strcat(capture_pfn,"/"); strcat(capture_pfn,ws); }
   } else { dbg("Capture file name missing in .ini file"); return (1); }

   GetIniString("[FILES]", "Overwrite", "NO", ws, sws, inipfn);
   if (ISON(ws)) strcpy(over_write,"w");

   return(0);
}
